Package: HealthMarkers
Type: Package
Title: Clinical and Metabolic Biomarker Calculation Toolkit
Version: 0.1.4
Authors@R: 
    person(given = "Sufyan", family = "Suleman",
           email = "sufyansuleman@hotmail.com",
           role = c("aut", "cre"),
           comment = c(ORCID = "0000-0001-6612-6915"))
Description: Computes specialist biomarker indices and risk scores for metabolic,
    cardiovascular, renal, hepatic, inflammatory, frailty, and psychiatric
    health assessment. Includes fasting and OGTT insulin sensitivity/resistance
    indices, ASCVD/QRISK3/KFRE risk equations, liver and kidney markers,
    frailty and comorbidity indices, biofluid marker panels, and utilities for
    column mapping, normalization, imputation, and combined marker dispatch.
License: MIT + file LICENSE
URL: https://sufyansuleman.github.io/HealthMarkers/,
        https://github.com/sufyansuleman/HealthMarkers
BugReports: https://github.com/sufyansuleman/HealthMarkers/issues
Depends: R (>= 4.0)
Imports: dplyr (>= 1.0.0), rlang (>= 1.0.0), stats, tibble (>= 3.0.0),
        utils, vctrs, Rdpack
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), pkgload, CVrisk, di,
        rspiro, PooledCohort, QRISK3, RiskScorescvd, usethis, withr,
        mice, missForest
RdMacros: Rdpack
Encoding: UTF-8
Language: en-US
RoxygenNote: 7.3.2
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-06-16 21:20:10 UTC; au773972
Author: Sufyan Suleman [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6612-6915>)
Maintainer: Sufyan Suleman <sufyansuleman@hotmail.com>
Repository: CRAN
Date/Publication: 2026-06-17 07:20:20 UTC
Built: R 4.6.0; ; 2026-06-17 10:15:14 UTC; unix

# Generated by roxygen2: do not edit by hand

export(adipo_is)
export(adiposity_sds)
export(adiposity_sds_strat)
export(all_health_markers)
export(all_insulin_indices)
export(allostatic_load)
export(alm_bmi_index)
export(asrs_score)
export(atherogenic_indices)
export(bis_score)
export(bode_index)
export(bone_markers)
export(calc_sds)
export(charlson_index)
export(ckd_stage)
export(cognitive_score)
export(corrected_calcium)
export(cvd_marker_aip)
export(cvd_marker_ldl_particle_number)
export(cvd_risk)
export(cvd_risk_ascvd)
export(cvd_risk_qrisk3)
export(cvd_risk_scorescvd)
export(cvd_risk_stroke)
export(fasting_is)
export(frailty_index)
export(frax_score)
export(gad7_score)
export(ghq12_score)
export(glycemic_markers)
export(health_summary)
export(hm_col_report)
export(hm_normalize)
export(hormone_markers)
export(iAge)
export(impute_mice)
export(impute_missforest)
export(impute_missing)
export(infer_cols)
export(inflammatory_markers)
export(isi_score)
export(k10_score)
export(k6_score)
export(kidney_failure_risk)
export(kyn_trp_ratio)
export(lipid_markers)
export(liver_fat_markers)
export(liver_markers)
export(marker_summary)
export(mdq_score)
export(metabolic_markers)
export(metabolic_risk_features)
export(metss)
export(nfl_marker)
export(normalize_vec)
export(nutrient_markers)
export(obesity_indices)
export(ogtt_is)
export(oxidative_markers)
export(phq9_score)
export(plot_frailty_age)
export(psych_dx_flags)
export(psych_markers)
export(psych_med_flags)
export(pulmo_markers)
export(renal_markers)
export(saliva_markers)
export(sarc_f_score)
export(spirometry_markers)
export(spq_score)
export(sweat_markers)
export(tracer_dxa_is)
export(urine_markers)
export(validate_inputs)
export(vitamin_d_status)
export(vitamin_markers)
export(who5_score)
import(Rdpack)
importFrom(Rdpack,reprompt)
importFrom(dplyr,across)
importFrom(dplyr,bind_cols)
importFrom(dplyr,case_when)
importFrom(dplyr,everything)
importFrom(dplyr,if_else)
importFrom(dplyr,mutate)
importFrom(dplyr,transmute)
importFrom(rlang,":=")
importFrom(rlang,abort)
importFrom(rlang,enquo)
importFrom(rlang,inform)
importFrom(rlang,quo_is_null)
importFrom(rlang,quo_name)
importFrom(rlang,warn)
importFrom(stats,complete.cases)
importFrom(stats,lm)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,prcomp)
importFrom(stats,qnorm)
importFrom(stats,resid)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(vctrs,vec_as_names)

HealthMarkers
================

- [HealthMarkers](#healthmarkers)
  - [Installation](#installation)
  - [Quick start](#quick-start)
  - [Package overview](#package-overview)
  - [Which function do I need?](#which-function-do-i-need)
  - [All markers dispatcher
    `all_health_markers()`:](#all-markers-dispatcher-all_health_markers)
  - [Column mapping](#column-mapping)
  - [Multi-biobank automatic column
    recognition](#multi-biobank-automatic-column-recognition)
  - [Selected function families](#selected-function-families)
  - [Common problems and solutions](#common-problems-and-solutions)
  - [Missing data](#missing-data)
  - [Normalisation](#normalisation)
  - [Output utilities](#output-utilities)
  - [Citation policy](#citation-policy)
  - [Articles and further reading](#articles-and-further-reading)
  - [Development status and validated
    publications](#development-status-and-validated-publications)
  - [Contributing](#contributing)
  - [Citation](#citation)
  - [License](#license)
  - [AI use disclaimer](#ai-use-disclaimer)

<!-- badges: start -->

[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/HealthMarkers)](https://cran.r-project.org/package=HealthMarkers)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/HealthMarkers)](https://cran.r-project.org/package=HealthMarkers)
[![GitHub release
downloads](https://img.shields.io/github/downloads/sufyansuleman/HealthMarkers/total.svg)](https://github.com/sufyansuleman/HealthMarkers/releases)
[![R-CMD-check](https://github.com/sufyansuleman/HealthMarkers/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sufyansuleman/HealthMarkers/actions/workflows/R-CMD-check.yaml)
[![License:
MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)
[![DOI](https://zenodo.org/badge/1012367183.svg)](https://doi.org/10.5281/zenodo.20668334)
<!-- badges: end -->

# HealthMarkers

Computing derived clinical indices manually is slow, error-prone, and
rarely documented. Researchers working with UK Biobank, NHANES, or any
large clinical or phenotypic dataset routinely need HOMA-IR, eGFR,
FIB-4, TyG index, ASCVD risk, NLR, Matsuda index, and WHtR, indices that
each require a different formula, different input columns, and careful
handling of missing values. HealthMarkers takes care of all of that. You
pass in your data frame and get the markers back, each implemented from
its primary source publication. No formula lookup, no unit checking, no
silent NA propagation.

HealthMarkers is an R package for computing, standardising, and
summarising derived health indices from routine laboratory and
phenotypic data. More than 50 specialist functions cover over 290
validated biomarkers across insulin sensitivity, lipid and
cardiovascular risk, renal and hepatic function, inflammation and aging,
body composition, psychiatric scales, and alternate biofluids. The
unified dispatcher all_health_markers() runs every group at once and
returns a single wide tibble; individual domain functions are available
when you only need one panel.

> **Full documentation, function reference, and articles** are available
> at:\
> <https://sufyansuleman.github.io/HealthMarkers/>

**Key features:**

- **Broad coverage.** One `all_health_markers()` call returns glycaemic,
  lipid, liver, renal, pulmonary, inflammatory, hormonal, bone,
  psychiatric, nutritional, and frailty markers as a single wide tibble.
- **Works with your column names as-is.** The built-in synonym
  dictionary covers 15+ cohorts and biobanks (UK Biobank, NHANES, HUNT,
  Tromsø, FinnGen, Estonian Biobank, LifeLines, Generation Scotland,
  Danish NPU codes, OMOP/LOINC codes, and more). In most cases you do
  not need to rename anything.
- **Safe defaults.** NA handling, input validation, and column-name
  inference are built in. A failed marker group is skipped with a
  warning; your pipeline never crashes.
- **Fully traceable.** Every function cites its primary source paper.
  Full bibliography in `inst/REFERENCES.bib`.

**Typical workflow (four steps):**

``` r
library(HealthMarkers)

# 1. See which columns in your data are recognised automatically
hm_col_report(my_data)

# 2. (If needed) fill in any unmatched keys
col_map <- list(eGFR = "GFR_CKD_EPI", G0 = "fasting_glucose_mmol")

# 3. Compute the markers you need
results <- all_health_markers(my_data,
                               which   = c("glycemic", "lipid", "renal"),
                               col_map = col_map)

# 4. Inspect, summarise, or export the new columns
new_cols <- setdiff(names(results), names(my_data))
health_summary(results[, new_cols])
```

------------------------------------------------------------------------

## Installation

Install the released version from CRAN:

``` r
install.packages("HealthMarkers")
```

For the development version, install from GitHub:

``` r
remotes::install_github("sufyansuleman/HealthMarkers")
```

Some marker groups need optional packages. Install only those you use:

``` r
install.packages(c(
  "CVrisk",        # Framingham / basic CVD risk
  "PooledCohort",  # ASCVD Pooled Cohort Equations + stroke risk
  "QRISK3",        # QRISK3
  "RiskScorescvd", # SCORE2 / SCORE2-OP
  "rspiro",        # Spirometry GLI 2012 z-scores
  "di",            # frailty and DXA-based insulin sensitivity
  "mice",          # multiple imputation
  "missForest"     # random-forest imputation
))
```

If an optional package is missing, its corresponding marker group is
skipped safely. Use `verbose = TRUE` to see which groups were computed
and which were omitted.

------------------------------------------------------------------------

## Quick start

A simulated dataset is included so you can explore the package without
your own data.

``` r
library(HealthMarkers)

sim_path  <- system.file("extdata", "simulated_hm_data.rds", package = "HealthMarkers")
sim       <- readRDS(sim_path)
sim_small <- sim[1:50, ]   # small subset for speed
```

**Step 1: inspect auto-detected columns**

``` r
hm_col_report(sim_small)
```

**Step 2: compute key marker groups in one call**

``` r
out <- all_health_markers(
  data    = sim_small,
  which   = c("glycemic", "lipid", "renal", "inflammatory"),
  verbose = FALSE
)

new_cols <- setdiff(names(out), names(sim_small))
head(out[, new_cols])
```

> This example is shown with `eval=FALSE` so README rendering stays
> fast. Run it interactively to generate the full demo output.

**Step 3: call a single-purpose function when you need one biomerker or
one gorup only:**

``` r
# Atherogenic index of plasma (log TG/HDL): two columns in, one index out
aip <- cvd_marker_aip(sim_small, col_map = list(TG = "TG", HDL_c = "HDL_c"),
                      na_action = "keep")
head(aip[, setdiff(names(aip), names(sim_small))])
```

------------------------------------------------------------------------

## Package overview

| Domain | Functions | Key outputs |
|----|----|----|
| **Insulin sensitivity** | `fasting_is()`, `ogtt_is()`, `adipo_is()`, `tracer_dxa_is()`, `all_insulin_indices()` | HOMA-IR, QUICKI, Matsuda, Stumvoll, Gutt, SPISE, LIRI, 40+ indices |
| **Glycaemic** | `glycemic_markers()` | TyG index, METS-IR, LAR, ASI, HOMA-CP, diabetes risk flags |
| **Lipid & atherogenic** | `lipid_markers()`, `atherogenic_indices()`, `cvd_marker_aip()`, `cvd_marker_ldl_particle_number()` | TC/HDL, AIP, CRI-I/II, Castelli, LDL particle number |
| **Liver** | `liver_markers()`, `liver_fat_markers()` | FLI, NFS, FIB-4, APRI, BARD, ALBI, MELD-XI, HSI, LAP |
| **Metabolic syndrome** | `metss()`, `metabolic_risk_features()`, `allostatic_load()` | MetS severity score, component flags, allostatic load index |
| **Cardiovascular risk** | `cvd_risk()`, `cvd_risk_ascvd()`, `cvd_risk_qrisk3()`, `cvd_risk_scorescvd()`, `cvd_risk_stroke()` | ASCVD (PCE), QRISK3, SCORE2/SCORE2-OP, 10-yr stroke risk |
| **Renal / CKD** | `kidney_failure_risk()`, `renal_markers()`, `ckd_stage()`, `urine_markers()` | KFRE 2-yr/5-yr, eGFR (CKD-EPI), CKD stage, UACR, FE-Urea |
| **Pulmonary** | `pulmo_markers()`, `spirometry_markers()`, `bode_index()` | FEV1/FVC z-scores, GLI 2012 % predicted, BODE index |
| **Inflammatory & aging** | `inflammatory_markers()`, `iAge()`, `oxidative_markers()`, `kyn_trp_ratio()` | NLR, PLR, SII, LMR, iAge clock, 8-OHdG, KTR |
| **Hormonal** | `hormone_markers()` | T/E2 ratio, TSH/fT4, cortisol/DHEA, LH/FSH, HOMA-B, FAI |
| **Body composition** | `obesity_indices()`, `adiposity_sds()`, `adiposity_sds_strat()`, `alm_bmi_index()`, `calc_sds()` | BMI, WHR, ABSI, BRI, BAI, sex/age-stratified SDS, ALM/BMI |
| **Bone & fracture** | `bone_markers()`, `frax_score()` | P1NP, osteocalcin, CTX, NTX, FRAX 10-yr fracture probability |
| **Frailty & comorbidity** | `frailty_index()`, `charlson_index()`, `sarc_f_score()` | Rockwood deficit index, Charlson CCI, SARC-F |
| **Vitamins & nutrients** | `vitamin_markers()`, `vitamin_d_status()`, `nutrient_markers()` | Vitamin D status category, B12/folate ratio, ferritin saturation |
| **Alternate biofluids** | `saliva_markers()`, `sweat_markers()`, `urine_markers()` | Cortisol awakening response, sweat chloride, urinary ratios |
| **Neurological** | `nfl_marker()`, `kyn_trp_ratio()`, `corrected_calcium()` | Age-adjusted NfL, kynurenine/tryptophan ratio, corrected calcium |
| **Psychiatric** | `psych_markers()`, `phq9_score()`, `gad7_score()`, `k6_score()`, `k10_score()`, and more | PHQ-9, GAD-7, ISI, GHQ-12, K10, K6, WHO-5, ASRS, BIS-11, SPQ |

------------------------------------------------------------------------

## Which function do I need?

| I have… | I want… | Use |
|----|----|----|
| Fasting glucose + insulin | HOMA-IR, QUICKI, FIRI, 15+ fasting indices | `fasting_is()` |
| OGTT glucose + insulin (multiple time points) | Matsuda, Stumvoll, Gutt, Avignon, 25+ OGTT indices | `ogtt_is()` |
| Lipid panel (TC, HDL, LDL, TG) | AIP, CRI, Castelli, TC/HDL ratio, LDL particle number | `lipid_markers()` / `cvd_marker_aip()` |
| Age, sex, cholesterol, BP, smoking | 10-year ASCVD / QRISK3 / SCORE2 cardiovascular risk | `cvd_risk_ascvd()` / `cvd_risk_qrisk3()` / `cvd_risk_scorescvd()` |
| Creatinine, age, sex (± UACR) | eGFR, CKD stage, KFRE kidney failure probability | `renal_markers()` / `ckd_stage()` / `kidney_failure_risk()` |
| FEV1, FVC, age, height | Spirometry z-scores (GLI 2012) | `spirometry_markers()` |
| CBC (neutrophils, lymphocytes, platelets) | NLR, PLR, SII, LMR inflammatory ratios | `inflammatory_markers()` |
| BMI, waist, hip | ABSI, BRI, BAI, WHR, obesity indices | `obesity_indices()` |
| Height, weight ± DXA measures | Body composition SDS z-scores | `adiposity_sds()` |
| Questionnaire item columns | PHQ-9, GAD-7, K10, ISI, GHQ-12, MDQ scores | `psych_markers()` / `phq9_score()` / `gad7_score()` |
| Many lab variables at once | All of the above in one call | `all_health_markers(which = "all")` |

------------------------------------------------------------------------

## All markers dispatcher `all_health_markers()`:

Use this when you have a large data set with many variables and want to
compute many marker groups in one call, returned as a single wide
tibble:

``` r
results <- all_health_markers(
  data    = my_data,
  which   = c("glycemic", "lipid", "liver", "renal", "mets", "inflammatory"),
  col_map = list(G0 = "fasting_glucose", I0 = "insulin0"),
  verbose = TRUE
)
```

All available `which` group keys:

    insulin_fasting     insulin_ogtt        insulin_adipose     insulin_tracer_dxa
    glycemic            lipid               atherogenic         cvd_aip
    cvd_risk            cvd_ldl_particles   cvd_ascvd           cvd_qrisk3
    cvd_scorescvd       cvd_stroke          liver               liver_fat
    mets                metabolic_risk      allostatic_load     pulmo
    spirometry          bode                saliva              sweat
    urine               renal               kidney_kfre         ckd_stage
    nutrient            vitamin             vitamin_d_status    hormone
    inflammatory        iAge                bone                frax
    oxidative           allostatic_load     frailty_index       charlson
    sarc_f              psych               nfl                 inflammatory_age
    calcium_corrected   kyn_trp             adiposity_sds       adiposity_sds_strat
    obesity_metrics     alm_bmi

Pass `which = "all"` to run everything. Groups requiring unavailable
optional packages are silently skipped.

### Returning only the computed markers (not the full input)

By default `all_health_markers()` appends the new marker columns to your
original data and returns everything together. With large cohort data
(e.g. 40,000 rows × 300 columns) this doubles the width of the object.
Use `return_input = FALSE` to get back **only the newly computed
columns**, keeping your pipeline lean:

``` r
# Default: original columns + new markers (wide output)
out <- all_health_markers(my_data, which = c("glycemic", "lipid"))

# Markers only: much smaller result
markers_only <- all_health_markers(
  my_data,
  which        = c("glycemic", "lipid", "renal"),
  return_input = FALSE,
  id_col       = "participant_id"   # carry the ID so you can join back later
)

# Join back when you need the full picture
final <- dplyr::left_join(my_data, markers_only, by = "participant_id")
```

`id_col` is optional; omit it if you don’t need to join back, or if row
order is sufficient.

------------------------------------------------------------------------

## Column mapping

### Step 1: `hm_col_report()` - see what is detected

Call this **before** any computation to get a full report of which
columns are auto-matched and which need manual mapping:

``` r
hm_col_report(my_data)
```

Output:

    ── HealthMarkers column report ────────────────────────────────────────────
     Data: 40314 rows × 299 columns   |   Keys in dictionary: 258

     key                  data_column        how matched
     -------------------- ------------------ ------------------
     fasting_glucose      pglu0              exact  ✔
     TG                   trig               exact  ✔
     ALT                  alat               exact  ✔
     eGFR                 ─                  NOT FOUND ✘

     ✔ 187 keys matched   ✘ 71 keys not found

    ── col_map template for missing keys ──────────────────────────────────────
     col_map <- list(
       eGFR  = "from_your_data",
     )

### Step 2: fill in unmatched keys

`hm_col_report()` returns a named list with every auto-detected key
already mapped. Add only the missing keys to that list, then pass the
completed `col_map` to your marker function.

``` r
cm        <- hm_col_report(my_data, verbose = FALSE)  # returns named list
cm$eGFR   <- "GFR_ckdepi"   # add your column name for any unmatched key
cm$G0     <- "fasting_glucose_mmol"  # if not already matched
```

### Step 3: pass `col_map` to any function

``` r
all_health_markers(my_data, which = c("renal", "glycemic"), col_map = cm)
fasting_is(my_data, col_map = list(G0 = "fasting_glucose_mmol", I0 = "insulin_uU_mL"))
```

`hm_col_report()` options:

``` r
hm_col_report(my_data, show_unmatched = TRUE)  # also list every unmatched key
hm_col_report(my_data, fuzzy = TRUE)           # add fuzzy matching as last resort
```

### Common internal keys

| Key | Meaning | Example column names |
|----|----|----|
| `G0` | Fasting glucose (mmol/L) | `pglu0`, `fasting_glucose`, `LBXGLU`, `paastoglukoosi` |
| `I0` | Fasting insulin (mU/L) | `insu0`, `insulin0`, `ins_fast` |
| `G30`, `G120` | 30-/120-min OGTT glucose | `pglu30`, `pglu120` |
| `I30`, `I120` | 30-/120-min OGTT insulin | `insu30`, `insu120` |
| `TG` | Triglycerides (mmol/L) | `trig`, `TAG`, `triglyserider`, `LOINC_2571_8` |
| `HDL_c` | HDL cholesterol | `hdlc`, `HDL`, `hdl_kolesteroli`, `LOINC_2085_9` |
| `LDL_c` | LDL cholesterol | `ldl`, `LDL`, `ldl_kolesteroli`, `LOINC_13457_7` |
| `TC` | Total cholesterol | `chol`, `kokonaiskolesteroli`, `LOINC_2093_3` |
| `ALT` | Alanine aminotransferase | `alat`, `SGPT`, `LBXSATSI`, `NPU03429`, `LOINC_1742_6` |
| `creatinine` | Serum creatinine | `crea`, `kreatinin`, `NPU01994`, `LOINC_2160_0` |
| `UACR` | Urine albumin/creatinine ratio | `ualbcrea`, `ACR` |
| `SBP` / `DBP` | Systolic/diastolic BP | `sysbp`, `systolisk_blodtrykk`, `LOINC_8480_6` |
| `BMI` | Body mass index | `bmi`, `painoindeksi`, `LOINC_39156_5` |
| `waist` | Waist circumference (cm) | `waist_cm`, `WC`, `midjeomkrets` |
| `vitaminD` | 25-OH vitamin D (nmol/L) | `vitd25`, `d_vitamin`, `NPU10501`, `LOINC_62292_8` |
| `HbA1c` | Glycated haemoglobin (%) | `hba1c`, `hemoglobiini_a1c`, `NPU27300`, `LOINC_4548_4` |

------------------------------------------------------------------------

## Multi-biobank automatic column recognition

The synonym dictionary recognises column names from 15+ major cohorts
and biobanks. The same analyte across systems:

| Internal key | UK Biobank | NHANES | HUNT/Tromsø | FinnGen | Estonian BB | LifeLines (NL) | LOINC |
|----|----|----|----|----|----|----|----|
| `fasting_glucose` | `glucose_0_0` | `LBXGLU` | `fastende_blodsukker` | `paastoglukoosi` | `p_glukoos` | `nuchtere_glucose` | `LOINC_2345_7` |
| `total_cholesterol` | `cholesterol_0_0` | `LBXSCH` | `total_kolesterol` | `kokonaiskolesteroli` | `kogukolesterool` | `totaal_cholesterol` | `LOINC_2093_3` |
| `creatinine` | `creatinine_0_0` | `LBXSCR` | `kreatinin` | `kreatiniini` | `kreatiniin` | `creatinine` | `LOINC_2160_0` |
| `HbA1c` | `glycated_haemoglobin_hba1c_0_0` | `LBXGH` | `HbA1c` | `hemoglobiini_a1c` | `HbA1c` | `geglycosyleerd_hemoglobine` | `LOINC_4548_4` |
| `SBP` | `systolic_blood_pressure_0_0` | `BPXSY1` | `systolisk_blodtrykk` | `SBP` | `sbp` | `systolische_bloeddruk` | `LOINC_8480_6` |
| `vitaminD` | `vitamin_d_0_0` | `LBXVD2` | `d_vitamin` | `D_vitamiini` | `D_vitamiin` | `vitamine_D` | `LOINC_62292_8` |
| `ALT` | `alanine_aminotransferase_0_0` | `LBXSATSI` | `ALAT` | `alaniiniaminotransferaasi` | `ALAT` | `alanineaminotransferase` | `LOINC_1742_6` |

Also supported: **Danish NPU codes** (`NPU01994`, `NPU01567`, etc.),
**Generation Scotland** (`SBP_mean`, `DBP_mean`, `genetic_sex`),
**SCAPIS/TwinGene** Swedish terms, and **OMOP CDM / All of Us** LOINC
concept codes for all major analytes.

------------------------------------------------------------------------

## Selected function families

Use an individual function when you need one specific domain, want
fine-grained control over `col_map`, or are working with specialist data
formats (OGTT time-series, DXA output, spirometry). For reference pages
and formulas, use `?function_name`.

### Insulin sensitivity

``` r
# Fasting indices (HOMA-IR, QUICKI, Bennett, FIRI, ...)
fasting_is(data, col_map = list(G0 = "glucose", I0 = "insulin"))

# OGTT indices (Matsuda, Stumvoll, Gutt, Avignon, ...)
ogtt_is(data, col_map = list(G0="G0", G30="G30", G60="G60", G120="G120",
                              I0="I0", I30="I30", I60="I60", I120="I120"))

# Adipose-tissue indices (LIRI, SPISE, VAI, LAP, ...)
adipo_is(data, col_map = list(BMI="BMI", WC="WC", TG="TG", HDL_c="HDL_c"))

# DXA / tracer-based indices
tracer_dxa_is(data, col_map = list(fat_mass="FM_kg", lean_mass="LM_kg"))

# All insulin indices at once (fasting + OGTT + adipose + DXA)
all_insulin_indices(data, col_map = list(...), normalize = "none",
                    mode = "both",   # "IS" sensitivity only, "IR" resistance only
                    na_action = "keep")
```

### Cardiovascular risk

``` r
# ASCVD Pooled Cohort Equations (requires PooledCohort)
cvd_risk_ascvd(data, year = 10)

# QRISK3 (UK population, requires QRISK3)
cvd_risk_qrisk3(data)

# SCORE2 / SCORE2-OP (European, requires RiskScorescvd)
cvd_risk_scorescvd(data)

# Atherogenic index of plasma
cvd_marker_aip(data, col_map = list(TG = "TG", HDL_c = "HDL_c"))

# LDL particle number from ApoB
cvd_marker_ldl_particle_number(data, col_map = list(ApoB = "ApoB"))

# Run all CVD algorithms at once
cvd_risk(data, model = "ALL")
```

### Renal function

``` r
# Kidney Failure Risk Equation (KFRE) 2-year and 5-year
kidney_failure_risk(data, col_map = list(age="age", sex="sex",
                                          eGFR="eGFR", UACR="UACR"))

# eGFR, BUN/creatinine, FE-Urea, renal ratios
renal_markers(data, col_map = list(creatinine="Creat", age="age", sex="sex"))

# KDIGO CKD staging (G1–G5 × A1–A3)
ckd_stage(data, col_map = list(eGFR="eGFR", UACR="UACR"))

# Urine panel: protein/creatinine ratio, microalbumin, osmolality
urine_markers(data, col_map = list(urine_creat="UCr", urine_protein="UPr"))
```

### Pulmonary function

``` r
# Spirometry z-scores and % predicted (GLI 2012, requires rspiro)
spirometry_markers(data, col_map = list(fev1="FEV1", fvc="FVC",
                                         age="age", height="ht_cm", sex="sex"))

# Simple pulmonary ratios (no extra packages)
pulmo_markers(data)

# BODE index for COPD prognosis
bode_index(data, col_map = list(fev1_pct="FEV1pct", sixmwd="Walk6m",
                                  mmrc="mMRC", bmi="BMI"))
```

### Body composition and anthropometric SDS

``` r
# Common obesity and adiposity indices (BMI, WHR, ABSI, BRI, BAI, ...)
obesity_indices(data)

# SDS z-score from any reference mean and SD
calc_sds(x = data$BMI, mean_ref = 22.5, sd_ref = 3.8)

# Sex-stratified SDS for multiple adiposity variables
adiposity_sds_strat(data, col_map = list(sex = "sex"),
                    var_cols = c("BMI", "WC", "WHR"),
                    ref_male   = list(BMI = c(mean = 25, sd = 4)),
                    ref_female = list(BMI = c(mean = 24, sd = 3.8)))

# Appendicular lean mass / BMI index (sarcopenia screening)
alm_bmi_index(data, col_map = list(alm = "ALM_kg", bmi = "BMI", sex = "Sex"))
```

### Psychiatric scores

``` r
# Score any combination of standardised scales from item columns.
# Supported: PHQ-9, GAD-7, K6, K10, GHQ-12, WHO-5, ISI, MDQ,
#            ASRS, BIS-11, SPQ, cognitive composite
psych_markers(
  data,
  col_map = list(
    phq9 = list(items = list(phq9_01 = "Q1", phq9_02 = "Q2")),
    gad7 = list(items = list(gad7_01 = "G1", gad7_02 = "G2"))
  ),
  which = c("phq9", "gad7", "k10")
)

# If columns are already named phq9_01 ... phq9_09 etc., no col_map needed:
phq9_score(data)
gad7_score(data)
k10_score(data)
```

### Inflammatory and aging markers

``` r
# Blood count-derived ratios (NLR, PLR, SII, LMR, ...)
inflammatory_markers(data, col_map = list(neut="NEUT", lymph="LYMPH",
                                           mono="MONO", plt="PLT"))

# iAge inflammatory aging clock
iAge(data, col_map = list(IL6="IL6", CXCL9="CXCL9"))
```

### Alternate biofluids

``` r
saliva_markers(data, col_map = list(cortisol_wake="C_wake", cortisol_30="C_30min"))
sweat_markers(data,  col_map = list(sweat_chloride="Cl_mmol"))
urine_markers(data,  col_map = list(urine_creat="UCr", urine_na="UNa"))
```

------------------------------------------------------------------------

## Common problems and solutions

| Symptom | Likely cause | Fix |
|----|----|----|
| A marker column is all `NA` | Required input column not found or all-missing | Run `hm_col_report(data)` to see which keys are unmatched; add them to `col_map` |
| An entire group is silently skipped | Optional package not installed, or no matched columns for that group | Set `verbose = TRUE` to see which groups ran and why others were skipped |
| `hm_col_report()` shows 0 matches | Column names not in the synonym dictionary | Use `col_map` to manually map your names to internal keys |
| ASCVD / QRISK3 / SCORE2 result is all `NA` | Optional package (`PooledCohort`, `QRISK3`, `RiskScorescvd`) not installed | Install the relevant package; see Installation section |
| Spirometry z-scores are `NA` | `rspiro` not installed, or `ethnicity` column missing/wrong codes | Install `rspiro`; see `?spirometry_markers` for ethnicity code table |
| Output is very wide (300+ columns) | `which = "all"` on a rich dataset | Use `which = c("glycemic", "lipid", ...)` to select only the groups you need, or `return_input = FALSE` to get markers only |
| `Error: object 'x' not found` in examples | Using `my_data` placeholder from README | Replace with your actual data object |

------------------------------------------------------------------------

## Missing data

All marker functions accept a `na_action` argument:

``` r
# "keep"  : compute what is possible, return NA where inputs are missing (default)
# "omit"  : drop rows with any missing required input before computing
# "error" : abort if any required input is missing
renal_markers(data, na_action = "keep")
```

> **`na_action` aliases:** `"ignore"` is a backward-compatible alias for
> `"keep"` (same behaviour; retained so older code continues to work).
> `"warn"` is also an alias for `"keep"` that additionally emits a
> missingness warning. If you are reading a function’s help page and see
> `"ignore"` listed first in the choices, it behaves identically to
> `"keep"`.

For pre-computation imputation, use the built-in helpers:

``` r
# Multiple imputation via mice
imputed <- impute_missing(data, method = "mice", m = 1, maxit = 5)

# Random-forest imputation (good for mixed data types)
imputed <- impute_missing(data, method = "missForest")

# Simple mean/median fill
imputed <- impute_missing(data, method = "mean")

# Then pass imputed data to any marker function
all_health_markers(imputed, which = c("glycemic", "lipid"))
```

------------------------------------------------------------------------

## Normalisation

The `normalize` argument is implemented in the insulin sensitivity
functions (`fasting_is()`, `ogtt_is()`, `adipo_is()`,
`all_insulin_indices()`). For all other domain functions, use
`hm_normalize()` to normalise outputs after computation.

### `hm_normalize()`: post-computation normalisation

``` r
out <- all_health_markers(data, which = c("glycemic", "lipid", "renal"))

# Identify new marker columns
new_cols <- setdiff(names(out), names(data))

# Normalise only the new marker columns (z-score)
out_z <- hm_normalize(out, cols = new_cols, method = "z")

# Rank-based inverse-normal transform on all numeric columns,
# keeping age and BMI on their original scale
out_int <- hm_normalize(out, method = "inverse", skip_cols = c("age", "BMI"))

# Min-max scaling to [0, 1]
out_range <- hm_normalize(out, cols = new_cols, method = "range")

# Robust median/MAD scaling
out_rob <- hm_normalize(out, cols = new_cols, method = "robust")
```

Available methods:

| Method      | Description                                     |
|-------------|-------------------------------------------------|
| `"z"`       | z-score (mean 0, sd 1)                          |
| `"inverse"` | Rank-based inverse-normal transform (Rankit)    |
| `"range"`   | Min-max to \[0, 1\] (or custom `feature_range`) |
| `"robust"`  | Median/MAD scaling                              |

### `normalize_vec()`: single-vector normalisation

``` r
normalize_vec(x, method = "inverse")
normalize_vec(x, method = "inverse", invnorm_denominator = "blom")
normalize_vec(x, method = "z")
normalize_vec(x, method = "robust")
normalize_vec(x, method = "range", feature_range = c(-1, 1))
```

See `?hm_normalize` and the package website articles page
<https://sufyansuleman.github.io/HealthMarkers/articles/> for full
details.

------------------------------------------------------------------------

## Output utilities

``` r
# Quick numeric summary (n, n_na, mean, sd, median, p25, p75) for any data frame
health_summary(out)

# Summary of marker outputs (variable, mean, sd, IQR)
marker_summary(out)

# Plot frailty-index deficit accumulation against age
plot_frailty_age(data, cols = c("deficit1", "deficit2", "deficit3"), age = "age")
```

------------------------------------------------------------------------

## Citation policy

Every function cites at least one reference in its help page
(`?function_name`). We have made every effort to trace citations back to
the **original primary source**: the paper in which the index, formula,
or scoring system was first described.

In practice this is not always straightforward:

- The commonly cited paper may be a **validation study** or a **derived
  index paper** rather than the mathematical original. In these cases we
  cite the paper most widely used in the clinical or epidemiological
  literature for that formula.
- For **statistical and mathematical methods** (e.g., rank-based
  inverse-normal transformation, median absolute deviation, standardised
  difference scores), the original work may be decades old and the
  specific parameterisation in common use was formalised in a more
  recent methodological paper. We cite whichever source most precisely
  describes the implementation used.
- For **multi-component composite indices** (e.g., allostatic load,
  frailty index, ASCVD Pooled Cohort Equations), multiple papers
  collectively define the algorithm; we aim to cite the most complete or
  most cited specification.

The full bibliography is in `inst/REFERENCES.bib` and is rendered in
each function’s help page via Rdpack.

If you identify a citation that could be improved (a closer original
source, a more authoritative validation paper, or a correction), please
open an issue at
<https://github.com/sufyansuleman/HealthMarkers/issues>.

------------------------------------------------------------------------

## Articles and further reading

Full articles are available on the **package website**:

- **Website:** <https://sufyansuleman.github.io/HealthMarkers/>
- **Function reference** (all arguments, details, and citations):
  <https://sufyansuleman.github.io/HealthMarkers/reference/>
- **All articles / tutorials:**
  <https://sufyansuleman.github.io/HealthMarkers/articles/>

Below are some key articles. For the full list, visit: [All
articles](https://sufyansuleman.github.io/HealthMarkers/articles/)

| Topic | Article |
|----|----|
| Fasting insulin sensitivity (HOMA-IR, QUICKI, 15+ indices) | [fasting_is](https://sufyansuleman.github.io/HealthMarkers/articles/fasting_is.html) |
| OGTT insulin sensitivity (Matsuda, Stumvoll, Gutt, 25+ indices) | [ogtt_is](https://sufyansuleman.github.io/HealthMarkers/articles/ogtt_is.html) |
| Glycaemic markers (TyG, METS-IR, LAR, diabetes flags) | [glycemic_markers](https://sufyansuleman.github.io/HealthMarkers/articles/glycemic_markers.html) |
| Lipid and atherogenic indices | [lipid_markers](https://sufyansuleman.github.io/HealthMarkers/articles/lipid_markers.html) |
| Cardiovascular risk scores (ASCVD, QRISK3, SCORE2) | [cvd_risk](https://sufyansuleman.github.io/HealthMarkers/articles/cvd_risk.html) |
| Inflammatory ratios and iAge clock | [inflammatory_markers](https://sufyansuleman.github.io/HealthMarkers/articles/inflammatory_markers.html) |
| Hormone markers | [hormone_markers](https://sufyansuleman.github.io/HealthMarkers/articles/hormone_markers.html) |
| Missing data and imputation | [impute_missing](https://sufyansuleman.github.io/HealthMarkers/articles/impute_missing.html) |
| Normalising marker outputs (`hm_normalize`, `normalize_vec`) | [health markers articles](https://sufyansuleman.github.io/HealthMarkers/articles/) |
| All-in-one dispatcher (`all_health_markers`) | [health_markers](https://sufyansuleman.github.io/HealthMarkers/articles/health_markers.html) |

For details on any individual function, use `?function_name` (e.g.,
`?fasting_is`, `?cvd_risk_ascvd`, `?frailty_index`): every help page
includes the formula, argument details, and primary citations.

------------------------------------------------------------------------

## Development status and validated publications

HealthMarkers is under active development. All indices are implemented
from their original, revised or verified published manuscripts. If you
notice an error in any index, please [open an
issue](https://github.com/sufyansuleman/HealthMarkers/issues) so it can
be corrected.

The insulin sensitivity and resistance indices have been independently
verified and are used in the following peer-reviewed publications:

- Suleman S, Madsen AL, Ängquist LH, Schubert M, Linneberg A, Loos RJF,
  Hansen T, Grarup N. Genetic Underpinnings of Fasting and Oral
  Glucose-stimulated Based Insulin Sensitivity Indices. *J Clin
  Endocrinol Metab.* 2024;109(11):2754–2763. [PMID
  38635292](https://pubmed.ncbi.nlm.nih.gov/38635292/)

- Suleman S, Ängquist L, Linneberg A, Hansen T, Grarup N. Exploring the
  genetic intersection between obesity-associated genetic variants and
  insulin sensitivity indices. *Sci Rep.* 2025;15:15761. [PMID
  40328835](https://pubmed.ncbi.nlm.nih.gov/40328835/)

------------------------------------------------------------------------

## Contributing

Issues and pull requests are welcome at
<https://github.com/sufyansuleman/HealthMarkers/issues>.

When contributing a new marker function please:

1.  Add a unit test in `tests/testthat/` with at least one numeric
    check.
2.  Add a `@references` entry in the roxygen block and cite the primary
    paper in `inst/REFERENCES.bib`.
3.  Register the function in the `all_health_markers()` dispatcher if it
    fits an existing domain.
4.  Add or update the relevant article in `vignettes/articles/`
    (articles are published to the pkgdown site at
    <https://sufyansuleman.github.io/HealthMarkers/articles/> but are
    not bundled with the CRAN package).

------------------------------------------------------------------------

## Citation

To cite HealthMarkers in publications, please use:

> Suleman, S. HealthMarkers: Toolkit for Clinical, Metabolic, and
> Cardiovascular Biomarker Calculations. R package.
> https://doi.org/10.5281/zenodo.20668334

Or from R (always gives the entry for your installed version):

``` r
citation("HealthMarkers")
```

DOI (all versions): [10.5281/zenodo.20668334](https://doi.org/10.5281/zenodo.20668334)

------------------------------------------------------------------------

## License

MIT Sufyan Suleman ([ORCID
0000-0001-6612-6915](https://orcid.org/0000-0001-6612-6915))

------------------------------------------------------------------------

## AI use disclaimer

OpenAI (ChatGPT) and Anthropic Claude were used during the development
of this package to assist with code refinement, debugging, and editing
of documentation content. All outputs were reviewed, verified, and
approved by the author.

#' Calculate urine-only renal and tubular markers (research-ready)
#'
#' Computes (urine-only):
#'  - UACR (Albumin-to-Creatinine Ratio, mg/g)
#'  - albuminuria_stage (KDIGO A1/A2/A3 by UACR)
#'  - microalbuminuria flag ("normal" vs "micro")
#'  - UPCR (Urine Protein-to-Creatinine Ratio, mg/g; if urine_protein available)
#'  - U_Na_K_ratio (urine Na+/K+; if urine_Na and urine_K available)
#'  - Creatinine-normalized tubular markers (if present, per g creatinine):
#'    NGAL_per_gCr, KIM1_per_gCr, NAG_per_gCr, Beta2Micro_per_gCr,
#'    A1Micro_per_gCr, IL18_per_gCr, L_FABP_per_gCr
#'
#' Inputs are validated, missingness handled via `na_action`, divisions are
#' safeguarded (Inf/NaN -> NA) with a consolidated zero-denominator warning,
#' and an optional extremes scan/cap is available.
#'
#' Expected units:
#' - urine_albumin: mg/L
#' - urine_protein: mg/L (optional)
#' - urine_creatinine: mg/dL
#' - urine_Na, urine_K: mmol/L (optional)
#' - Optional tubular markers above assumed mg/L when normalized per g creatinine
#'
#' @param data A data.frame or tibble with at least urine_albumin and urine_creatinine.
#' @param verbose Logical; if `TRUE`, prints progress messages and a completion summary. Default FALSE.
#' @param na_action One of `c("keep","omit","error")` for handling missing values in required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness warnings for required inputs. Default 0.2.
#' @param col_map Optional named list mapping canonical keys (e.g., `urine_albumin`,
#'   `urine_creatinine`) to actual column names in `data`. If `NULL`, column names
#'   are inferred automatically.
#'
#' @return A tibble with columns:
#'   UACR, albuminuria_stage, microalbuminuria, UPCR, U_Na_K_ratio,
#'   NGAL_per_gCr, KIM1_per_gCr, NAG_per_gCr, Beta2Micro_per_gCr,
#'   A1Micro_per_gCr, IL18_per_gCr, L_FABP_per_gCr
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
#' @examples
#' df <- tibble::tibble(
#'   urine_albumin    = 30,
#'   urine_creatinine = 1.2,
#'   serum_creatinine = 0.9,
#'   plasma_Na        = 140,
#'   urine_Na         = 100,
#'   age              = 55,
#'   sex              = 2,
#'   urine_protein    = 150
#' )
#' urine_markers(df)
#' @note
#' UACR formula: albumin (mg/L) / creatinine (g/L) = albumin (mg/L) \eqn{\times} 100 /
#' creatinine (mg/dL). UPCR and per-gCr tubular markers use the same creatinine
#' denominator: `gCr_den = creatinine (mg/dL) \times 0.01` (= g/L).
#' Tubular markers (NGAL, KIM-1, NAG, Beta-2-microglobulin, alpha-1-microglobulin,
#' IL-18, L-FABP) are **pass-through** columns normalised per g creatinine; no
#' formula other than creatinine adjustment is applied.
#'
#' @references
#' \insertRef{mogensen1984uacr}{HealthMarkers}
#' \insertRef{ginsberg1983upcr}{HealthMarkers}
#' \insertRef{kdigo2012ckd}{HealthMarkers} (albuminuria staging A1–A3 UACR cutoffs)
#' \insertRef{dezeeuw2006uacr}{HealthMarkers} (prognostic UACR validation; background)
#' \insertRef{ichimura2004kim1}{HealthMarkers} (KIM-1 tubular biomarker; pass-through normalization, background)
#' \insertRef{portilla2008lfabp}{HealthMarkers} (L-FABP tubular biomarker; pass-through normalization, background)
urine_markers <- function(data,
                          col_map = NULL,
                          verbose = TRUE,
                          na_action = c("keep","omit","error"),
                          na_warn_prop = 0.2) {
  na_action <- match.arg(na_action)

  if (!is.data.frame(data)) {
    rlang::abort("urine_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_urine_error_data_type")
  }
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "urine_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)

  # 1) required columns (urine-only core)
  req <- c("urine_albumin", "urine_creatinine")
  all_keys <- c(req, "urine_protein", "urine_Na", "urine_K",
                "NGAL", "KIM1", "NAG", "beta2_micro", "a1_micro", "IL18", "L_FABP")
  # HM-CS v2: standardized validation and column inference
  hm_validate_inputs(data, col_map, required_keys = character(0), fn = "urine_markers")
  cm      <- .hm_build_col_map(data, col_map, keys = all_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  missing_cols <- setdiff(req, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("urine_markers(): missing columns: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_urine_error_missing_columns"
    )
  }

  # 2) Coerce numerics; optional columns supported
  num_candidates <- intersect(
    c("urine_albumin","urine_creatinine","urine_protein","urine_Na","urine_K",
      "NGAL","KIM1","NAG","beta2_micro","a1_micro","IL18","L_FABP"),
    names(data)
  )
  for (cn in num_candidates) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("urine_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
    # Non-finite to NA for safety
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  UACR, albuminuria_stage, microalbuminuria [urine_albumin, urine_creatinine]\n  UPCR [urine_protein, urine_creatinine]\n  U_Na_K_ratio [urine_Na, urine_K]\n  NGAL/KIM1/NAG/Beta2Micro/A1Micro/IL18/L_FABP per gCr [optional]", fn_name), level = "inform")
  # 3) High-missingness diagnostics on required inputs (debug level)
  for (cn in req) {
    x <- data[[cn]]
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("urine_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }

  # 4) NA policy on required inputs
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("urine_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_urine_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(sprintf("urine_markers(): omitting %d rows with NA in required inputs", sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      UACR = numeric(),
      albuminuria_stage = factor(character(), levels = c("A1","A2","A3")),
      microalbuminuria = factor(character(), levels = c("normal","micro")),
      UPCR = numeric(),
      U_Na_K_ratio = numeric(),
      NGAL_per_gCr = numeric(),
      KIM1_per_gCr = numeric(),
      NAG_per_gCr = numeric(),
      Beta2Micro_per_gCr = numeric(),
      A1Micro_per_gCr = numeric(),
      IL18_per_gCr = numeric(),
      L_FABP_per_gCr = numeric()
    ))
  }

  hm_inform("urine_markers(): computing markers", level = "debug")

  # 6) Safe division helper with consolidated zero-denominator tracking
  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }

  # 7) Core ratios
  # denominator for mg per g creatinine: urine_creatinine (mg/dL) -> g/L
  gCr_den <- data$urine_creatinine * 0.01

  # UACR (mg/g) = albumin (mg/L) / creatinine (g/L)
  # creatinine (g/L) = creatinine (mg/dL) * 10 / 1000 = creatinine (mg/dL) * 0.01
  # => UACR = albumin (mg/L) * 100 / creatinine (mg/dL)
  UACR <- 100 * safe_div(data$urine_albumin, data$urine_creatinine, "UACR_creatinine")

  albuminuria_stage <- factor(
    ifelse(is.finite(UACR) & UACR < 30, "A1",
           ifelse(is.finite(UACR) & UACR <= 300, "A2",
                  ifelse(is.finite(UACR) & UACR > 300, "A3", NA_character_))),
    levels = c("A1","A2","A3")
  )

  microalbuminuria <- factor(
    ifelse(is.finite(UACR) & UACR >= 30 & UACR <= 300, "micro", "normal"),
    levels = c("normal","micro")
  )

  UPCR <- if ("urine_protein" %in% names(data)) {
    safe_div(data$urine_protein, gCr_den, "UPCR_creatinine_gL")
  } else {
    rep(NA_real_, nrow(data))
  }

  U_Na_K_ratio <- if (all(c("urine_Na","urine_K") %in% names(data))) {
    safe_div(data$urine_Na, data$urine_K, "U_Na_K_ratio_denK")
  } else {
    rep(NA_real_, nrow(data))
  }

  # 8) Creatinine-normalized tubular markers (per g creatinine)
  per_gCr <- function(x, label) safe_div(x, gCr_den, label)
  get_or_na <- function(nm) if (nm %in% names(data)) as.numeric(data[[nm]]) else rep(NA_real_, nrow(data))

  NGAL_per_gCr        <- per_gCr(get_or_na("NGAL"),        "NGAL_gCr")
  KIM1_per_gCr        <- per_gCr(get_or_na("KIM1"),        "KIM1_gCr")
  NAG_per_gCr         <- per_gCr(get_or_na("NAG"),         "NAG_gCr")
  Beta2Micro_per_gCr  <- per_gCr(get_or_na("beta2_micro"), "B2M_gCr")
  A1Micro_per_gCr     <- per_gCr(get_or_na("a1_micro"),    "A1M_gCr")
  IL18_per_gCr        <- per_gCr(get_or_na("IL18"),        "IL18_gCr")
  L_FABP_per_gCr      <- per_gCr(get_or_na("L_FABP"),      "L_FABP_gCr")

  out <- tibble::tibble(
    UACR               = as.numeric(UACR),
    albuminuria_stage  = albuminuria_stage,
    microalbuminuria   = microalbuminuria,
    UPCR               = as.numeric(UPCR),
    U_Na_K_ratio       = as.numeric(U_Na_K_ratio),
    NGAL_per_gCr       = as.numeric(NGAL_per_gCr),
    KIM1_per_gCr       = as.numeric(KIM1_per_gCr),
    NAG_per_gCr        = as.numeric(NAG_per_gCr),
    Beta2Micro_per_gCr = as.numeric(Beta2Micro_per_gCr),
    A1Micro_per_gCr    = as.numeric(A1Micro_per_gCr),
    IL18_per_gCr       = as.numeric(IL18_per_gCr),
    L_FABP_per_gCr     = as.numeric(L_FABP_per_gCr)
  )

  # 9) Consolidated zero-denominator warning
  dz <- dz_env$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz); nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("urine_markers(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  if (!is.null(id_col)) {
    out[[id_col]] <- data[[id_col]]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) hm_inform(hm_result_summary(out, fn_name), level = "inform")

  out
}

# tests/testthat/test_urine_markers.R
library(testthat)
library(tibble)

test_that("errors if any required column is missing", {
  skip_on_cran()
  # Missing urine_albumin
  df1 <- tibble(
    urine_creatinine = 1
  )
  expect_error(
    urine_markers(df1),
    "missing columns: urine_albumin"
  )

  # Missing urine_creatinine
  df2 <- tibble(
    urine_albumin = 30
  )
  expect_error(
    urine_markers(df2),
    "missing columns: urine_creatinine"
  )
})

test_that("computes UACR, UPCR, Na/K ratio, and normalized tubular markers; names match", {
  skip_on_cran()
  df <- tibble(
    urine_albumin    = 30,    # mg/L
    urine_creatinine = 2,     # mg/dL
    urine_protein    = 150,   # mg/L
    urine_Na         = 60,    # mmol/L
    urine_K          = 20,    # mmol/L
    NGAL             = 100,   # mg/L
    KIM1             = 10,
    NAG              = 5,
    beta2_micro      = 2,
    a1_micro         = 3,
    IL18             = 1,
    L_FABP           = 0.5
  )
  out <- urine_markers(df)

  expect_named(
    out,
    c("UACR", "albuminuria_stage", "microalbuminuria",
      "UPCR", "U_Na_K_ratio",
      "NGAL_per_gCr", "KIM1_per_gCr", "NAG_per_gCr",
      "Beta2Micro_per_gCr", "A1Micro_per_gCr", "IL18_per_gCr", "L_FABP_per_gCr")
  )

  # UACR = albumin(mg/L) * 100 / creatinine(mg/dL) = 30 * 100 / 2 = 1500 mg/g
  expect_equal(out$UACR, 1500)

  # UPCR = urine_protein / (urine_creatinine*0.01) = 150/0.02 = 7500
  expect_equal(out$UPCR, 150 / (2 * 0.01))

  # Na/K = 60/20 = 3
  expect_equal(out$U_Na_K_ratio, 60 / 20)

  # Creatinine-normalized markers per g creatinine: denom gCr_den = 2*0.01 = 0.02 g/L
  gCr_den <- 2 * 0.01
  expect_equal(out$NGAL_per_gCr,       100 / gCr_den)
  expect_equal(out$KIM1_per_gCr,        10 / gCr_den)
  expect_equal(out$NAG_per_gCr,          5 / gCr_den)
  expect_equal(out$Beta2Micro_per_gCr,   2 / gCr_den)
  expect_equal(out$A1Micro_per_gCr,      3 / gCr_den)
  expect_equal(out$IL18_per_gCr,         1 / gCr_den)
  expect_equal(out$L_FABP_per_gCr,     0.5 / gCr_den)
})

test_that("albuminuria_stage and microalbuminuria factors have correct levels and values", {
  skip_on_cran()
  # UACR = albumin * 100 / creatinine (mg/L and mg/dL respectively)
  # Row1: 100*1/100  =  1 mg/g -> A1 (< 30)
  # Row2: 100*30/100 = 30 mg/g -> A2 (30-300), micro
  # Row3: 100*400/1  = 40000 mg/g -> A3 (> 300)
  df <- tibble(
    urine_albumin    = c(1, 30, 400),
    urine_creatinine = c(100, 100, 1),
    # optional fields omitted
  )
  out <- urine_markers(df)

  expect_true(is.factor(out$albuminuria_stage))
  expect_equal(levels(out$albuminuria_stage), c("A1","A2","A3"))
  expect_equal(as.character(out$albuminuria_stage), c("A1","A2","A3"))

  expect_true(is.factor(out$microalbuminuria))
  expect_equal(levels(out$microalbuminuria), c("normal","micro"))
  # UACR=1 -> normal; UACR=30 -> micro; UACR=40000 -> normal (> 300, macroalbuminuria not flagged as micro)
  expect_equal(as.character(out$microalbuminuria), c("normal","micro","normal"))
})

test_that("UPCR and U_Na_K_ratio are NA when inputs are missing", {
  skip_on_cran()
  df <- tibble(
    urine_albumin    = 30,
    urine_creatinine = 2
    # no urine_protein, no urine_Na/urine_K
  )
  out <- urine_markers(df)
  expect_true(is.na(out$UPCR))
  expect_true(is.na(out$U_Na_K_ratio))
})

test_that("is vectorized over multiple rows", {
  skip_on_cran()
  df <- tibble(
    urine_albumin    = c(30, 40),
    urine_creatinine = c(2, 4),
    urine_protein    = c(150, NA_real_)
  )
  out <- urine_markers(df)
  expect_equal(nrow(out), 2)
  # second row UPCR NA
  expect_true(is.na(out$UPCR[2]))
})

test_that("na_action policies: error and omit behave as expected", {
  skip_on_cran()
  df_na <- tibble(
    urine_albumin    = c(30, NA_real_),
    urine_creatinine = c(2, 2)
  )
  # error -> abort (suppress high-missingness warning)
  expect_error(
    suppressWarnings(urine_markers(df_na, na_action = "error")),
    "required inputs contain missing values"
  )
  # omit -> drops NA row
  out_omit <- suppressWarnings(urine_markers(df_na, na_action = "omit"))
  expect_equal(nrow(out_omit), 1L)
})

test_that("extreme inputs pass through without error (check_extreme removed)", {
  skip_on_cran()
  df_ext <- tibble(
    urine_albumin    = 1e6,
    urine_creatinine = 1e-6,
    urine_protein    = 1e6
  )
  out <- suppressWarnings(urine_markers(df_ext))
  expect_s3_class(out, "tbl_df")
  expect_true("UACR" %in% names(out))
})

test_that("zero denominators emit a consolidated warning and yield NA in ratios", {
  skip_on_cran()
  df_zero <- tibble(
    urine_albumin    = 30,
    urine_creatinine = 0,     # zero -> denominator for all per-gCr ratios
    urine_protein    = 150,
    urine_Na         = 60,
    urine_K          = 20
  )
  expect_warning(
    out_zero <- urine_markers(df_zero),
    "zero denominators detected"
  )
  expect_true(is.na(out_zero$UACR))
  expect_true(is.na(out_zero$UPCR))
  expect_true(is.na(out_zero$NGAL_per_gCr))
  expect_equal(out_zero$U_Na_K_ratio, 60 / 20)  # unaffected (denominator is K, not creatinine)
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  df <- tibble(
    urine_albumin    = 30,
    urine_creatinine = 2
  )
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(urine_markers(df, verbose = TRUE), "urine_markers")
  expect_message(urine_markers(df, verbose = TRUE), "col_map")
  expect_message(urine_markers(df, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- tibble(
    urine_albumin    = 30,
    urine_creatinine = 2
  )
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(urine_markers(df, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Validate that col_map is fully specified and data has the mapped columns
#'
#' Utility used across HealthMarkers to validate a column mapping against a data
#' frame. It checks that:
#' - `data` is a data.frame/tibble.
#' - `col_map` is a named list or named character vector.
#' - All required keys in `required_keys` have non-NULL mappings.
#' - All mapped columns exist in `data`.
#' - Optionally flags duplicate mappings and unknown keys.
#' - Emits informative errors or warnings depending on action settings.
#' - Optional verbose summary of what was checked.
#'
#' Backward compatibility:
#' - Defaults keep prior behavior (error on missing map entries and missing columns;
#'   allow extra keys and duplicate columns).
#'
#' @param data A data.frame or tibble
#' @param col_map Named list (or named character vector) of expected keys -> column names
#' @param fun_name Character; name of the calling function for messages (e.g., "fasting_is")
#' @param required_keys Character vector; names in col_map that are required
#' @param optional_keys Character vector; optional keys allowed in col_map (informational)
#' @param allow_extra_keys Logical; allow keys in col_map beyond required/optional. Default TRUE.
#' @param duplicate_action One of c("ignore","warn","error") controlling behavior when the same
#'   source column is mapped to multiple keys. Default "ignore".
#' @param missing_action One of c("error","warn","ignore") for missing map entries among required_keys.
#'   Default "error".
#' @param missing_cols_action One of c("error","warn","ignore") for mapped columns not present in data.
#'   Default "error".
#' @param normalize_map Logical; if TRUE, coerce col_map entries to single character strings or NULL,
#'   trim whitespace, and convert NA_character_ to NULL. Default TRUE.
#' @param trim_ws Logical; trim leading/trailing whitespace when normalize_map = TRUE. Default TRUE.
#' @param return_map Logical; if TRUE, invisibly return the normalized map (named character vector/NULLs).
#'   If FALSE, invisibly return TRUE (backward compatible). Default FALSE.
#' @param verbose Logical; if TRUE, prints a short summary of checks. Default FALSE.
#'
#' @return
#' - If return_map = FALSE (default): invisibly returns TRUE on success.
#' - If return_map = TRUE: invisibly returns the normalized mapping (list with character or NULL entries).
#'
#' @examples
#' \dontrun{
#' # Legacy internal col_map validator (full-featured version)
#' df <- tibble::tibble(G0 = c(5.5, 6.1), I0 = c(60, 88))
#' .hm_validate_col_map(df, list(G0 = "G0", I0 = "I0"), fun_name = "fasting_is")
#' }
#'
#' # Return normalized map
#' .hm_validate_col_map(df, list(G0 = " G0 ", I0 = "I0"), fun_name = "fasting_is", return_map = TRUE)
#'
#' # Will error: missing mapping for I0
#' \dontrun{
#' .hm_validate_col_map(df, list(G0 = "G0", I0 = NULL), fun_name = "fasting_is")
#' }
#'
#' @keywords internal
#' @noRd
.hm_validate_col_map <- function(data,
                            col_map,
                            fun_name = "",
                            required_keys = names(col_map),
                            optional_keys = NULL,
                            allow_extra_keys = TRUE,
                            duplicate_action = c("ignore", "warn", "error"),
                            missing_action = c("error", "warn", "ignore"),
                            missing_cols_action = c("error", "warn", "ignore"),
                            normalize_map = TRUE,
                            trim_ws = TRUE,
                            return_map = FALSE,
                            verbose = FALSE) {
  duplicate_action   <- match.arg(duplicate_action)
  missing_action     <- match.arg(missing_action)
  missing_cols_action <- match.arg(missing_cols_action)

  # ---- Basic input checks ----
  if (!is.data.frame(data)) {
    stop(sprintf("%s(): `data` must be a data.frame or tibble.", fun_name), call. = FALSE)
  }
  if (is.null(col_map)) {
    stop(sprintf("%s(): `col_map` must be a named list or named character vector.", fun_name), call. = FALSE)
  }
  # Accept named character vector too
  if (is.character(col_map) && !is.null(names(col_map))) {
    col_map <- as.list(col_map)
  }
  if (!is.list(col_map) || is.null(names(col_map)) || any(names(col_map) == "")) {
    stop(sprintf("%s(): `col_map` must be a named list (or named character vector).", fun_name), call. = FALSE)
  }

  # Normalize map entries if requested
  if (isTRUE(normalize_map)) {
    col_map <- .vld_normalize_map(col_map, trim_ws = trim_ws)
  }

  # ---- Keys checks ----
  map_keys <- names(col_map)
  if (!isTRUE(allow_extra_keys)) {
    allowed <- unique(c(required_keys %||% character(0), optional_keys %||% character(0)))
    extra <- setdiff(map_keys, allowed)
    if (length(extra)) {
      stop(sprintf("%s(): unexpected keys in col_map: %s", fun_name, paste(extra, collapse = ", ")), call. = FALSE)
    }
  }

  # 1) ensure the user provided a non-NULL name for every key in required_keys
  missing_map <- required_keys[vapply(required_keys, function(k) {
    is.null(col_map[[k]]) || identical(col_map[[k]], "")
  }, logical(1))]
  if (length(missing_map)) {
    msg <- sprintf("%s(): you must supply col_map entries for: %s", fun_name, paste(missing_map, collapse = ", "))
    if (missing_action == "error") stop(msg, call. = FALSE)
    if (missing_action == "warn")  warning(msg, call. = FALSE)
  }

  # 2) check that data contains every mapped column for required_keys
  # Only evaluate mapped (non-NULL, non-empty) entries
  req_vals <- unlist(col_map[required_keys], use.names = FALSE)
  req_vals <- req_vals[!is.na(req_vals) & nzchar(req_vals)]
  miss_cols_req <- setdiff(req_vals, names(data))
  if (length(miss_cols_req)) {
    msg <- sprintf("%s(): missing required columns: %s", fun_name, paste(miss_cols_req, collapse = ", "))
    if (missing_cols_action == "error") stop(msg, call. = FALSE)
    if (missing_cols_action == "warn")  warning(msg, call. = FALSE)
  }

  # 3) For optional keys provided (non-NULL), check existence too
  if (!is.null(optional_keys) && length(optional_keys)) {
    opt_vals <- unlist(col_map[optional_keys], use.names = FALSE)
    opt_vals <- opt_vals[!is.na(opt_vals) & nzchar(opt_vals)]
    miss_cols_opt <- setdiff(opt_vals, names(data))
    if (length(miss_cols_opt)) {
      msg <- sprintf("%s(): missing optional columns (provided in col_map): %s", fun_name, paste(miss_cols_opt, collapse = ", "))
      if (missing_cols_action == "error") stop(msg, call. = FALSE)
      if (missing_cols_action == "warn")  warning(msg, call. = FALSE)
    }
  }

  # 4) Duplicate mapping check (same source column used multiple times)
  mapped_vals <- unlist(col_map, use.names = FALSE)
  mapped_vals <- mapped_vals[!is.na(mapped_vals) & nzchar(mapped_vals)]
  if (length(mapped_vals)) {
    dups <- mapped_vals[duplicated(tolower(mapped_vals))]
    if (length(dups)) {
      dup_str <- paste(unique(dups), collapse = ", ")
      if (duplicate_action == "error") {
        stop(sprintf("%s(): duplicate column mappings detected: %s", fun_name, dup_str), call. = FALSE)
      } else if (duplicate_action == "warn") {
        warning(sprintf("%s(): duplicate column mappings detected: %s", fun_name, dup_str), call. = FALSE)
      }
    }
  }

  # Verbose summary
  if (isTRUE(verbose)) {
    n_req <- length(required_keys)
    n_opt <- length(optional_keys %||% character(0))
    n_map <- sum(!vapply(col_map[required_keys], is.null, logical(1)))
    message(sprintf("%s(): validated col_map - required keys: %d, mapped: %d, optional keys: %d",
                    fun_name, n_req, n_map, n_opt))
  }

  if (isTRUE(return_map)) {
    return(invisible(col_map))
  }

  invisible(TRUE)
}

# ---- internal helpers --------------------------------------------------------

# Normalize col_map entries to single character strings or NULL (trim whitespace)
.vld_normalize_map <- function(col_map, trim_ws = TRUE) {
  for (k in names(col_map)) {
    val <- col_map[[k]]
    if (is.null(val)) next
    # Accept character vectors; use first non-NA value
    if (is.character(val)) {
      if (length(val) == 0L) {
        col_map[[k]] <- NULL
        next
      }
      # drop NAs, keep first
      v <- val[!is.na(val)]
      v <- if (length(v)) v[1L] else NA_character_
      if (isTRUE(trim_ws) && !is.na(v)) v <- trimws(v)
      if (is.na(v) || identical(v, "")) {
        col_map[[k]] <- NULL
      } else {
        col_map[[k]] <- v
      }
    } else {
      # Coerce other types to character if possible; otherwise set NULL
      v <- tryCatch(as.character(val)[1L], error = function(e) NA_character_)
      if (isTRUE(trim_ws) && !is.na(v)) v <- trimws(v)
      if (is.na(v) || identical(v, "")) col_map[[k]] <- NULL else col_map[[k]] <- v
    }
  }
  col_map
}

# null coalescing
`%||%` <- function(a, b) if (is.null(a)) b else a

# HM-VS v2: check that mapped columns exist in data (common across functions)
hm_require_columns <- function(data, col_map, required_keys, fn) {
  req_cols <- unname(unlist(col_map[required_keys], use.names = FALSE))
  missing_cols <- setdiff(req_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0(fn, "(): missing required columns in `data`: ", paste(missing_cols, collapse = ", ")),
      class = sprintf("healthmarkers_%s_error_missing_columns", fn)
    )
  }
  req_cols
}

# HM-VS v2: high-missingness diagnostics (debug-level)
hm_warn_high_missing <- function(data, cols, prop = 0.2, fn = "healthmarkers") {
  for (cn in intersect(cols, names(data))) {
    x <- data[[cn]]
    n <- length(x); if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("%s(): column '%s' has high missingness (%.1f%%).", fn, cn, 100 * pna))
    }
  }
  invisible(TRUE)
}

#' @importFrom stats prcomp complete.cases
#' @importFrom dplyr bind_cols
#' @importFrom tibble as_tibble
#' @importFrom rlang :=
NULL

# Internal helpers -------------------------------------------------------------

.hm_match <- function(x, choices) {
  match.arg(x, choices)
}

.hm_is_empty_df <- function(data) {
  is.data.frame(data) && nrow(data) == 0L
}

.hm_as_numeric_matrix <- function(df, cols) {
  if (length(cols) == 0L) return(matrix(numeric(0), nrow = nrow(df), ncol = 0))
  mat <- vapply(cols, function(cc) suppressWarnings(as.numeric(df[[cc]])), numeric(nrow(df)))
  dim(mat) <- c(nrow(df), length(cols))
  colnames(mat) <- cols
  mat
}

.hm_resolve_items <- function(data, items, col_map = NULL, where = "items") {
  if (is.null(items) || length(items) == 0L) return(character(0))

  if (is.null(col_map)) {
    cols <- items
  } else {
    if (is.list(col_map) && !is.null(col_map[[where]])) col_map <- col_map[[where]]
    cols <- unname(vapply(items, function(k) {
      if (!is.null(col_map[[k]])) col_map[[k]] else k
    }, character(1)))
  }

  missing <- setdiff(cols, names(data))
  if (length(missing) > 0L) {
    stop("Missing required columns: ", paste(missing, collapse = ", "))
  }
  cols
}

.hm_apply_missing_policy <- function(scores, na_action = c("keep","omit","error")) {
  na_action <- .hm_match(na_action, c("keep","omit","error"))
  if (na_action == "error" && any(!is.finite(scores))) {
    stop("Missing values produced NA scores; set na_action='keep' or 'omit'.")
  }
  scores
}

.hm_rowwise_impute_mean <- function(mat) {
  if (ncol(mat) == 0L) return(mat)
  rmean <- rowMeans(mat, na.rm = TRUE)
  for (j in seq_len(ncol(mat))) {
    idx <- is.na(mat[, j])
    if (any(idx)) mat[idx, j] <- rmean[idx]
  }
  mat
}

.hm_score_scale <- function(data,
                            item_cols,
                            reverse = character(0),
                            min_val,
                            max_val,
                            na_action = c("keep","omit","error"),
                            missing_prop_max = 0.2,
                            impute = c("none","mean")) {

  na_action <- .hm_match(na_action, c("keep","omit","error"))
  impute <- .hm_match(impute, c("none","mean"))

  mat <- .hm_as_numeric_matrix(data, item_cols)

  if (length(reverse) > 0L) {
    rev_cols <- intersect(reverse, item_cols)
    if (length(rev_cols) > 0L) {
      rev_idx <- match(rev_cols, colnames(mat))
      mat[, rev_idx] <- (max_val + min_val) - mat[, rev_idx]
    }
  }

  if (ncol(mat) == 0L) {
    total <- rep(NA_real_, nrow(data))
    missing_prop <- rep(NA_real_, nrow(data))
  } else {
    missing_prop <- rowMeans(is.na(mat))
    too_missing <- missing_prop > missing_prop_max
    if (impute == "mean") mat <- .hm_rowwise_impute_mean(mat)
    total <- rowSums(mat, na.rm = FALSE)
    total[too_missing] <- NA_real_
  }

  total <- .hm_apply_missing_policy(total, na_action = na_action)
  list(total = total, missing_prop = missing_prop)
}

.hm_add_cols <- function(data, new_cols_df, na_action = c("keep","omit","error")) {
  na_action <- .hm_match(na_action, c("keep","omit","error"))
  out <- tibble::as_tibble(new_cols_df)
  if (na_action == "omit") out <- out[stats::complete.cases(new_cols_df), , drop = FALSE]
  out
}

.hm_severity_band <- function(x, cuts, labels) {
  if (length(cuts) != length(labels) + 1L) {
    rlang::abort(sprintf(
      ".hm_severity_band(): `cuts` must have exactly one more element than `labels` (got %d cuts, %d labels).",
      length(cuts), length(labels)
    ))
  }
  out <- rep(NA_character_, length(x))
  for (i in seq_len(length(labels))) {
    out[x >= cuts[i] & x < cuts[i + 1]] <- labels[i]
  }
  out
}

`%||%` <- function(a, b) if (!is.null(a)) a else b

.hm_validate_key <- function(key) {
  if (is.null(key) || !is.list(key)) stop("`key` must be a list.")
  if (is.null(key$items) || length(key$items) == 0L) stop("`key$items` is required and must be non-empty.")
  if (!is.character(key$items)) stop("`key$items` must be character IDs.")
  if (is.null(key$min_val) || is.null(key$max_val)) stop("`key$min_val` and `key$max_val` are required.")
  if (!is.numeric(key$min_val) || !is.numeric(key$max_val) || length(key$min_val) != 1 || length(key$max_val) != 1) {
    stop("`key$min_val` and `key$max_val` must be numeric scalars.")
  }
  if (key$min_val >= key$max_val) stop("`key$min_val` must be < `key$max_val`.")
  if (!is.null(key$reverse) && !is.character(key$reverse)) stop("`key$reverse` must be character.")
  if (!is.null(key$subscales)) {
    if (!is.list(key$subscales) || is.null(names(key$subscales))) stop("`key$subscales` must be a named list.")
    for (nm in names(key$subscales)) {
      v <- key$subscales[[nm]]
      if (!is.character(v) || length(v) == 0L) stop("Each subscale must be a non-empty character vector: ", nm)
    }
    all_sub <- unique(unlist(key$subscales, use.names = FALSE))
    bad <- setdiff(all_sub, key$items)
    if (length(bad) > 0L) stop("Subscale items not found in key$items: ", paste(bad, collapse = ", "))
  }
  invisible(TRUE)
}

.hm_resolve_from_col_map <- function(items, col_map) {
  if (is.null(col_map)) return(stats::setNames(items, items))
  if (is.list(col_map) && !is.null(col_map$items)) col_map <- col_map$items
  if (is.list(col_map) && !is.null(col_map$vars)) col_map <- col_map$vars
  stats::setNames(vapply(items, function(k) {
    if (!is.null(col_map[[k]])) col_map[[k]] else k
  }, character(1)), items)
}

.hm_score_from_key <- function(data,
                               key,
                               col_map = list(),
                               prefix,
                               na_action = c("keep","omit","error"),
                               missing_prop_max = 0.2,
                               impute = c("none","mean")) {
  na_action <- .hm_match(na_action, c("keep","omit","error"))
  impute <- .hm_match(impute, c("none","mean"))
  .hm_validate_key(key)

  map <- .hm_resolve_from_col_map(key$items, col_map)
  cols <- unname(map)
  missing_cols <- setdiff(cols, names(data))
  if (length(missing_cols) > 0L) stop("Missing required item columns: ", paste(missing_cols, collapse = ", "))

  mat <- vapply(cols, function(cc) suppressWarnings(as.numeric(data[[cc]])), numeric(nrow(data)))
  dim(mat) <- c(nrow(data), length(cols))
  colnames(mat) <- names(map)

  reverse <- key$reverse %||% character(0)
  rev_ids <- intersect(reverse, colnames(mat))
  if (length(rev_ids) > 0L) {
    rev_idx <- match(rev_ids, colnames(mat))
    mat[, rev_idx] <- (key$max_val + key$min_val) - mat[, rev_idx]
  }

  miss_prop <- rowMeans(is.na(mat))
  too_missing <- miss_prop > missing_prop_max
  if (impute == "mean") mat <- .hm_rowwise_impute_mean(mat)

  total <- rowSums(mat, na.rm = FALSE)
  total[too_missing] <- NA_real_
  if (na_action == "error" && anyNA(total)) {
    stop("Missing values produced NA scores; set na_action='keep' or 'omit', or adjust missing_prop_max/impute.")
  }

  out <- tibble::tibble(!!paste0(prefix, "_total") := total)
  if (!is.null(key$subscales) && length(key$subscales) > 0L) {
    for (nm in names(key$subscales)) {
      ids <- key$subscales[[nm]]
      sub <- mat[, ids, drop = FALSE]
      sub_total <- rowSums(sub, na.rm = FALSE)
      sub_total[too_missing] <- NA_real_
      out[[paste0(prefix, "_", nm)]] <- sub_total
    }
  }

  if (na_action == "omit") out <- out[stats::complete.cases(out), , drop = FALSE]
  out
}

# PHQ-9 / PHQ-8 ---------------------------------------------------------------

#' PHQ-9 / PHQ-8 scoring
#' @param data Data frame containing questionnaire item columns.
#' @param col_map Named list mapping canonical item IDs to column names; defaults assume items are already named.
#' @param na_action How to handle rows with missing items: `keep`, `omit`, or `error`.
#' @param missing_prop_max Maximum allowed proportion of missing items per row before the score is set to `NA`.
#' @param impute Imputation strategy for missing items when under the threshold: `none` or `mean` (row-wise mean).
#' @param variant Choose PHQ9 (9 items) or PHQ8 (drops suicidal ideation item).
#' @param prefix Prefix for output column names.
#' @param verbose Logical; if `TRUE`, emits informational messages about column resolution and scoring progress via `hm_inform()`.
#' @references \insertRef{kroenke2001phq9}{HealthMarkers}
#' @return A tibble of score columns only: `PHQ9_total` and `PHQ9_severity` (factor). Input columns are not included.
#' @examples
#' df <- data.frame(phq9_01 = 0, phq9_02 = 1, phq9_03 = 2, phq9_04 = 1,
#'                  phq9_05 = 0, phq9_06 = 1, phq9_07 = 2, phq9_08 = 1,
#'                  phq9_09 = 0)
#' phq9_score(df)
#' @export
phq9_score <- function(data,
                       col_map = list(),
                       na_action = c("keep","omit","error"),
                       missing_prop_max = 0.2,
                       impute = c("none","mean"),
                       variant = c("PHQ9","PHQ8"),
                       prefix = "PHQ9",
                       verbose = TRUE) {
  variant <- match.arg(variant, c("PHQ9","PHQ8"))
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  hm_inform("phq9_score(): preparing inputs", level = "inform")

  items <- if (variant == "PHQ9") sprintf("phq9_%02d", 1:9) else sprintf("phq9_%02d", 1:8)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, reverse = character(0), min_val = 0, max_val = 3,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  total <- sc$total
  severity <- .hm_severity_band(total, cuts = c(-Inf, 5, 10, 15, 20, Inf),
                                labels = c("minimal","mild","moderate","moderately_severe","severe"))

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_total") := total,
    !!paste0(prefix, "_severity") := factor(severity, levels = c("minimal","mild","moderate","moderately_severe","severe"))
  )

  if (isTRUE(verbose)) hm_inform(hm_result_summary(out_cols, "phq9_score"), level = "inform")

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# GAD-7 -----------------------------------------------------------------------

#' GAD-7 scoring
#' @inheritParams phq9_score
#' @references
#' \insertRef{spitzer2006gad7}{HealthMarkers}
#' \insertRef{plummer2016gad7}{HealthMarkers} (validation meta-analysis)
#' @return A tibble of score columns only: `GAD7_total` and `GAD7_severity` (factor). Input columns are not included.
#' @examples
#' df <- data.frame(gad7_01 = 0, gad7_02 = 1, gad7_03 = 2, gad7_04 = 1,
#'                  gad7_05 = 0, gad7_06 = 1, gad7_07 = 2)
#' gad7_score(df)
#' @export
gad7_score <- function(data,
                       col_map = list(),
                       na_action = c("keep","omit","error"),
                       missing_prop_max = 0.2,
                       impute = c("none","mean"),
                       prefix = "GAD7",
                       verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("gad7_%02d", 1:7)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, reverse = character(0), min_val = 0, max_val = 3,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  total <- sc$total
  severity <- .hm_severity_band(total, cuts = c(-Inf, 5, 10, 15, Inf),
                                labels = c("minimal","mild","moderate","severe"))

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_total") := total,
    !!paste0(prefix, "_severity") := factor(severity, levels = c("minimal","mild","moderate","severe"))
  )

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# K6 / K10 --------------------------------------------------------------------

#' K6 scoring
#' @inheritParams phq9_score
#' @param cutoff Threshold for the K6 case flag.
#' @references
#' \insertRef{kessler2002k6k10}{HealthMarkers}
#' \insertRef{prochaska2012k6}{HealthMarkers} (validation study)
#' @return A tibble of score columns only: `K6_total` and `K6_case`. Input columns are not included.
#' @examples
#' df <- data.frame(k6_01 = 0, k6_02 = 1, k6_03 = 2, k6_04 = 1, k6_05 = 0, k6_06 = 1)
#' k6_score(df)
#' @export
k6_score <- function(data, col_map = list(),
                     na_action = c("keep","omit","error"),
                     missing_prop_max = 0.2,
                     impute = c("none","mean"),
                     prefix = "K6",
                     cutoff = 13,
                     verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("k6_%02d", 1:6)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, min_val = 0, max_val = 4,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  total <- sc$total
  out_cols <- tibble::tibble(
    !!paste0(prefix, "_total") := total,
    !!paste0(prefix, "_case") := ifelse(is.na(total), NA, total >= cutoff)
  )
  .hm_add_cols(data, out_cols, na_action = na_action)
}

#' K10 scoring
#' @inheritParams phq9_score
#' @note
#' K10 items are summed as provided. The original scale uses 1--5 coding
#' (total 10--50); some implementations subtract 1 (0--4, total 0--40).
#' The function accepts either coding, as no reverse-scored items are
#' involved and `min_val`/`max_val` only affect reversal.
#' @references
#' \insertRef{kessler2002k6k10}{HealthMarkers}
#' @return A tibble of score columns only: `K10_total`. Input columns are not included.
#' @examples
#' df <- data.frame(k10_01 = 0, k10_02 = 1, k10_03 = 2, k10_04 = 1, k10_05 = 0,
#'                  k10_06 = 1, k10_07 = 2, k10_08 = 1, k10_09 = 0, k10_10 = 1)
#' k10_score(df)
#' @export
k10_score <- function(data, col_map = list(),
                      na_action = c("keep","omit","error"),
                      missing_prop_max = 0.2,
                      impute = c("none","mean"),
                      prefix = "K10",
                      verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("k10_%02d", 1:10)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, min_val = 0, max_val = 4,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  total <- sc$total
  out_cols <- tibble::tibble(!!paste0(prefix, "_total") := total)
  .hm_add_cols(data, out_cols, na_action = na_action)
}

# GHQ-12 ----------------------------------------------------------------------

#' GHQ-12 scoring (Likert or binary)
#' @inheritParams phq9_score
#' @param method Scoring method: `likert` (0-3 per item) or `binary` (0/1 per item).
#' @param case_cutoff_binary Cut-off for case status when using binary scoring.
#' @references
#' \insertRef{goldberg1988ghq12}{HealthMarkers}
#' @return A tibble of score columns only: `GHQ12_total_likert` (likert method) or `GHQ12_total_binary` and `GHQ12_case_binary` (binary method). Input columns are not included.
#' @examples
#' df <- data.frame(ghq12_01 = 0, ghq12_02 = 1, ghq12_03 = 2, ghq12_04 = 1,
#'                  ghq12_05 = 0, ghq12_06 = 1, ghq12_07 = 0, ghq12_08 = 1,
#'                  ghq12_09 = 2, ghq12_10 = 1, ghq12_11 = 0, ghq12_12 = 1)
#' ghq12_score(df, method = "likert")
#' @export
ghq12_score <- function(data, col_map = list(),
                        na_action = c("keep","omit","error"),
                        missing_prop_max = 0.2,
                        impute = c("none","mean"),
                        prefix = "GHQ12",
                        method = c("likert","binary"),
                        case_cutoff_binary = 3,
                        verbose = TRUE) {
  method <- match.arg(method, c("likert","binary"))
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("ghq12_%02d", 1:12)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  mat <- .hm_as_numeric_matrix(data, item_cols)
  missing_prop <- rowMeans(is.na(mat))
  too_missing <- missing_prop > missing_prop_max
  if (impute == "mean") mat <- .hm_rowwise_impute_mean(mat)

  if (method == "likert") {
    total <- rowSums(mat, na.rm = FALSE)
    total[too_missing] <- NA_real_
    out_cols <- tibble::tibble(!!paste0(prefix, "_total_likert") := total)
  } else {
    bin <- ifelse(is.na(mat), NA_real_, ifelse(mat >= 2, 1, 0))
    total <- rowSums(bin, na.rm = FALSE)
    total[too_missing] <- NA_real_
    out_cols <- tibble::tibble(
      !!paste0(prefix, "_total_binary") := total,
      !!paste0(prefix, "_case_binary") := ifelse(is.na(total), NA, total >= case_cutoff_binary)
    )
  }

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# WHO-5 -----------------------------------------------------------------------

#' WHO-5 scoring
#' @inheritParams phq9_score
#' @param low_cutoff_percent Percentage threshold for low well-being flag.
#' @references \insertRef{topp2015who5}{HealthMarkers}
#' @return A tibble of score columns only: `WHO5_raw`, `WHO5_percent`, `WHO5_low_wellbeing`. Input columns are not included.
#' @examples
#' df <- data.frame(who5_01 = 0, who5_02 = 1, who5_03 = 2, who5_04 = 3, who5_05 = 4)
#' who5_score(df)
#' @export
who5_score <- function(data, col_map = list(),
                       na_action = c("keep","omit","error"),
                       missing_prop_max = 0.2,
                       impute = c("none","mean"),
                       prefix = "WHO5",
                       low_cutoff_percent = 50,
                       verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("who5_%02d", 1:5)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, min_val = 0, max_val = 5,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  raw <- sc$total
  pct <- raw * 4

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_raw") := raw,
    !!paste0(prefix, "_percent") := pct,
    !!paste0(prefix, "_low_wellbeing") := ifelse(is.na(pct), NA, pct < low_cutoff_percent)
  )

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# ISI -------------------------------------------------------------------------

#' Insomnia Severity Index (ISI) scoring
#' @inheritParams phq9_score
#' @references \insertRef{bastien2001isi}{HealthMarkers}
#' @return A tibble of score columns only: `ISI_total` and `ISI_severity` (factor). Input columns are not included.
#' @examples
#' df <- data.frame(isi_01 = 0, isi_02 = 1, isi_03 = 2, isi_04 = 1, isi_05 = 0, isi_06 = 1, isi_07 = 2)
#' isi_score(df)
#' @export
isi_score <- function(data, col_map = list(),
                      na_action = c("keep","omit","error"),
                      missing_prop_max = 0.2,
                      impute = c("none","mean"),
                      prefix = "ISI",
                      verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("isi_%02d", 1:7)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, min_val = 0, max_val = 4,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  total <- sc$total
  severity <- .hm_severity_band(total, cuts = c(-Inf, 8, 15, 22, Inf),
                                labels = c("none","subthreshold","moderate","severe"))

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_total") := total,
    !!paste0(prefix, "_severity") := factor(severity, levels = c("none","subthreshold","moderate","severe"))
  )
  .hm_add_cols(data, out_cols, na_action = na_action)
}

# MDQ -------------------------------------------------------------------------

#' Mood Disorder Questionnaire scoring
#' @inheritParams phq9_score
#' @param symptom_cutoff Minimum symptom count for a positive screen.
#' @param require_clustering Require clustering item == 1 to be positive.
#' @param require_impairment Require impairment item == 1 to be positive.
#' @note
#' The impairment column is expected to be **binary** (1 = impaired, 0 = not).
#' The original MDQ impairment question uses a 4-category scale
#' (1 = no problem, 2 = minor, 3 = moderate, 4 = serious); if the raw
#' 4-category response is passed, the caller must recode it to binary
#' (e.g., `impair_binary = as.integer(impair_raw >= 3)`) before scoring.
#' @references
#' \insertRef{hirschfeld2000mdq}{HealthMarkers}
#' @return A tibble of score columns only: `MDQ_symptom_count`, `MDQ_clustering`, `MDQ_impairment`, `MDQ_positive_screen`. Input columns are not included.
#' @examples
#' df <- data.frame(matrix(0, nrow = 1, ncol = 13))
#' names(df) <- sprintf("mdq_%02d", 1:13)
#' df$mdq_cluster <- 1; df$mdq_impair <- 1
#' mdq_score(df)
#' @export
mdq_score <- function(data, col_map = list(),
                      na_action = c("keep","omit","error"),
                      missing_prop_max = 0.2,
                      prefix = "MDQ",
                      symptom_cutoff = 7,
                      require_clustering = TRUE,
                      require_impairment = TRUE,
                      verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))

  sym_items <- sprintf("mdq_%02d", 1:13)
  sym_cols <- .hm_resolve_items(data, sym_items, col_map = col_map, where = "items")
  sym_mat <- .hm_as_numeric_matrix(data, sym_cols)

  miss_prop <- rowMeans(is.na(sym_mat))
  too_missing <- miss_prop > missing_prop_max

  sym_count <- rowSums(sym_mat, na.rm = FALSE)
  sym_count[too_missing] <- NA_real_

  cluster_col <- NULL
  impair_col <- NULL
  if (!is.null(col_map$cluster)) cluster_col <- col_map$cluster
  if (!is.null(col_map$impairment)) impair_col <- col_map$impairment
  if (is.null(cluster_col) && "mdq_cluster" %in% names(data)) cluster_col <- "mdq_cluster"
  if (is.null(impair_col) && "mdq_impair" %in% names(data)) impair_col <- "mdq_impair"

  cluster <- if (!is.null(cluster_col) && cluster_col %in% names(data)) as.numeric(data[[cluster_col]]) else NA_real_
  impair  <- if (!is.null(impair_col)  && impair_col  %in% names(data)) as.numeric(data[[impair_col]])  else NA_real_

  positive <- ifelse(is.na(sym_count), NA, sym_count >= symptom_cutoff)
  if (require_clustering && !all(is.na(cluster))) positive <- positive & (cluster == 1)
  if (require_impairment && !all(is.na(impair)))  positive <- positive & (impair == 1)

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_symptom_count") := sym_count,
    !!paste0(prefix, "_clustering") := ifelse(all(is.na(cluster)), NA, cluster == 1),
    !!paste0(prefix, "_impairment") := ifelse(all(is.na(impair)), NA, impair == 1),
    !!paste0(prefix, "_positive_screen") := positive
  )

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# ASRS ------------------------------------------------------------------------

#' Adult ADHD Self-Report Scale scoring
#' @inheritParams phq9_score
#' @param partA_items Vector of Part A item IDs.
#' @param partA_thresholds Numeric thresholds applied to Part A items.
#' @param partA_cutoff Count threshold for Part A positivity.
#' @note
#' Items are expected on a 0--4 scale (Never=0, Rarely=1, Sometimes=2,
#' Often=3, Very Often=4). The default `partA_thresholds` follow the
#' official WHO ASRS v1.1 guide: items 1--3 are positive at \eqn{\geq 3}
#' (Often or Very Often) and items 4--6 at \eqn{\geq 4} (Very Often only).
#' @references
#' \insertRef{adler2006asrs}{HealthMarkers}
#' \insertRef{kessler2007asrs}{HealthMarkers} (prevalence study; ASRS used as instrument)
#' @return A tibble of score columns only: `ASRS_total`, `ASRS_partA_count`, `ASRS_partA_positive`. Input columns are not included.
#' @examples
#' df <- data.frame(matrix(2, nrow = 1, ncol = 18))
#' names(df) <- sprintf("asrs_%02d", 1:18)
#' asrs_score(df)
#' @export
asrs_score <- function(data, col_map = list(),
                       na_action = c("keep","omit","error"),
                       missing_prop_max = 0.2,
                       impute = c("none","mean"),
                       prefix = "ASRS",
                       partA_items = sprintf("asrs_%02d", 1:6),
                       partA_thresholds = c(3, 3, 3, 4, 4, 4),
                       partA_cutoff = 4,
                       verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  items <- sprintf("asrs_%02d", 1:18)
  item_cols <- .hm_resolve_items(data, items, col_map = col_map, where = "items")

  sc <- .hm_score_scale(data, item_cols, min_val = 0, max_val = 4,
                        na_action = na_action, missing_prop_max = missing_prop_max, impute = impute)

  partA_cols <- .hm_resolve_items(data, partA_items, col_map = col_map, where = "items")
  matA <- .hm_as_numeric_matrix(data, partA_cols)
  if (length(partA_thresholds) != ncol(matA)) stop("partA_thresholds must match length(partA_items).")
  hit <- matA >= matrix(partA_thresholds, nrow = nrow(matA), ncol = ncol(matA), byrow = TRUE)
  hit[is.na(matA)] <- NA
  countA <- rowSums(hit, na.rm = FALSE)
  too_missingA <- rowMeans(is.na(matA)) > missing_prop_max
  countA[too_missingA] <- NA_real_

  out_cols <- tibble::tibble(
    !!paste0(prefix, "_total") := sc$total,
    !!paste0(prefix, "_partA_count") := countA,
    !!paste0(prefix, "_partA_positive") := ifelse(is.na(countA), NA, countA >= partA_cutoff)
  )

  .hm_add_cols(data, out_cols, na_action = na_action)
}

# BIS / SPQ via key -----------------------------------------------------------

#' Barratt Impulsiveness Scale (key-driven)
#' @inheritParams phq9_score
#' @param key List with `items`, `min_val`, `max_val`, optional `reverse` and `subscales`.
#' @references \insertRef{patton1995bis}{HealthMarkers}
#' @return A tibble of score columns only (total and optional subscales). Input columns are not included.
#' @examples
#' bis_key <- list(items = sprintf("bis_%02d", 1:5), min_val = 1, max_val = 4)
#' df <- data.frame(bis_01 = 1, bis_02 = 2, bis_03 = 3, bis_04 = 4, bis_05 = 2)
#' bis_score(df, key = bis_key)
#' @export
bis_score <- function(data,
                      col_map = list(),
                      key,
                      na_action = c("keep","omit","error"),
                      missing_prop_max = 0.2,
                      impute = c("none","mean"),
                      prefix = "BIS",
                      verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  scores <- .hm_score_from_key(
    data = data,
    key = key,
    col_map = col_map,
    prefix = prefix,
    na_action = na_action,
    missing_prop_max = missing_prop_max,
    impute = impute
  )

  .hm_add_cols(data, scores, na_action = na_action)
}

#' Schizotypal Personality Questionnaire (key-driven)
#' @inheritParams phq9_score
#' @param key List with `items`, `min_val`, `max_val`, optional `reverse` and `subscales`.
#' @references \insertRef{raine1991spq}{HealthMarkers}
#' @return A tibble of score columns only (total and optional subscales). Input columns are not included.
#' @examples
#' spq_key <- list(items = sprintf("spq_%02d", 1:5), min_val = 0, max_val = 1)
#' df <- data.frame(spq_01 = 0, spq_02 = 1, spq_03 = 0, spq_04 = 1, spq_05 = 0)
#' spq_score(df, key = spq_key)
#' @export
spq_score <- function(data,
                      col_map = list(),
                      key,
                      na_action = c("keep","omit","error"),
                      missing_prop_max = 0.2,
                      impute = c("none","mean"),
                      prefix = "SPQ",
                      verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))

  scores <- .hm_score_from_key(
    data = data,
    key = key,
    col_map = col_map,
    prefix = prefix,
    na_action = na_action,
    missing_prop_max = missing_prop_max,
    impute = impute
  )

  .hm_add_cols(data, scores, na_action = na_action)
}

# Cognitive composite ---------------------------------------------------------

#' Cognitive composite (z-mean or PCA1)
#' @inheritParams phq9_score
#' @param method Aggregation method: `z_mean` (average of z-scores) or `pca1` (first PC).
#' @param col_map Named list with `tasks` mapping task IDs to column names (>= 2 tasks required).
#' @return A tibble of score columns only: `{prefix}_z_mean` or `{prefix}_pca1`. Input columns are not included.
#' @examples
#' df <- data.frame(task_a = c(1, 2), task_b = c(2, 3), task_c = c(3, 4))
#' cm <- list(tasks = list(
#'   task_a = "task_a",
#'   task_b = "task_b",
#'   task_c = "task_c"
#' ))
#' cognitive_score(df, col_map = cm, method = "z_mean")
#' @export
cognitive_score <- function(data,
                            col_map = list(),
                            na_action = c("keep","omit","error"),
                            missing_prop_max = 0.2,
                            method = c("z_mean","pca1"),
                            prefix = "cog",
                            verbose = TRUE) {
  na_action <- match.arg(na_action, c("keep","omit","error"))
  method <- match.arg(method, c("z_mean","pca1"))

  tasks <- col_map$tasks
  if (is.null(tasks) || length(tasks) < 2) {
    stop("Provide col_map$tasks as a named vector/list of tasks -> columns (min 2 tasks).")
  }
  task_cols <- unname(unlist(tasks))
  missing <- setdiff(task_cols, names(data))
  if (length(missing) > 0L) stop("Missing cognitive task columns: ", paste(missing, collapse = ", "))

  mat <- .hm_as_numeric_matrix(data, task_cols)
  miss_prop <- rowMeans(is.na(mat))
  too_missing <- miss_prop > missing_prop_max

  z <- scale(mat)

  out <- tibble::tibble(.rows = nrow(data))
  if (method == "z_mean") {
    score <- rowMeans(z, na.rm = FALSE)
    score[too_missing] <- NA_real_
    out[[paste0(prefix, "_z_mean")]] <- as.numeric(score)
  } else {
    cc <- stats::complete.cases(z)
    score <- rep(NA_real_, nrow(z))
    if (sum(cc) >= 3) {
      pc <- stats::prcomp(z[cc, , drop = FALSE], center = FALSE, scale. = FALSE)
      score[cc] <- pc$x[, 1]
    }
    score[too_missing] <- NA_real_
    out[[paste0(prefix, "_pca1")]] <- as.numeric(score)
  }

  .hm_add_cols(data, out, na_action = na_action)
}

# Diagnosis and medication flags ---------------------------------------------

#' Psychiatric diagnosis flags aggregator
#' @inheritParams phq9_score
#' @param col_map Named list `dx` mapping condition ids (e.g., mdd, bipolar) to columns of boolean/numeric flags.
#' @param prefix Prefix for output flag columns.
#' @return A tibble of flag columns only: `dx_any_psych`, `dx_internalizing`, `dx_externalizing`, `dx_psychotic`, `dx_count`. Input columns are not included.
#' @examples
#' df <- data.frame(dx_mdd = c(1, 0), dx_bipolar = c(0, 1))
#' psych_dx_flags(df, col_map = list(dx = list(mdd = "dx_mdd", bipolar = "dx_bipolar")))
#' @export
psych_dx_flags <- function(data, col_map = list(),
                           prefix = "dx",
                           na_action = c("keep","omit","error")) {
  na_action <- match.arg(na_action, c("keep","omit","error"))

  dx <- col_map$dx
  if (is.null(dx) || length(dx) == 0) stop("Provide col_map$dx mapping, e.g., list(dx=list(mdd='dx_mdd', bipolar='dx_bipolar')).")

  get_flag <- function(col) {
    if (!col %in% names(data)) return(rep(NA, nrow(data)))
    x <- data[[col]]
    as.logical(if (is.numeric(x)) x != 0 else x)
  }

  mdd      <- if (!is.null(dx$mdd)) get_flag(dx$mdd) else rep(NA, nrow(data))
  anxiety  <- if (!is.null(dx$anxiety)) get_flag(dx$anxiety) else rep(NA, nrow(data))
  adhd     <- if (!is.null(dx$adhd)) get_flag(dx$adhd) else rep(NA, nrow(data))
  bipolar  <- if (!is.null(dx$bipolar)) get_flag(dx$bipolar) else rep(NA, nrow(data))
  scz      <- if (!is.null(dx$scz)) get_flag(dx$scz) else rep(NA, nrow(data))
  sud      <- if (!is.null(dx$sud)) get_flag(dx$sud) else rep(NA, nrow(data))

  internalizing <- mdd | anxiety
  externalizing <- adhd | sud
  psychotic     <- scz

  all_dx <- cbind(mdd, anxiety, adhd, bipolar, scz, sud)
  dx_count <- rowSums(all_dx, na.rm = TRUE)

  out <- tibble::tibble(
    !!paste0(prefix, "_any_psych") := (mdd | anxiety | adhd | bipolar | scz | sud),
    !!paste0(prefix, "_internalizing") := internalizing,
    !!paste0(prefix, "_externalizing") := externalizing,
    !!paste0(prefix, "_psychotic") := psychotic,
    !!paste0(prefix, "_count") := dx_count
  )

  .hm_add_cols(data, out, na_action = na_action)
}

#' Psychiatric medication flags aggregator
#' @inheritParams phq9_score
#' @param col_map Named list `med` mapping medication classes (e.g., ssri, snri) to columns of boolean/numeric flags.
#' @param prefix Prefix for output flag columns.
#' @return A tibble of flag columns only: `med_any_psych`, `med_count`. Input columns are not included.
#' @examples
#' df <- data.frame(med_ssri = c(1, 0), med_antipsychotic = c(0, 1))
#' cm <- list(med = list(
#'   ssri = "med_ssri",
#'   antipsychotic = "med_antipsychotic"
#' ))
#' psych_med_flags(df, col_map = cm)
#' @export
psych_med_flags <- function(data, col_map = list(),
                            prefix = "med",
                            na_action = c("keep","omit","error")) {
  na_action <- match.arg(na_action, c("keep","omit","error"))

  med <- col_map$med
  if (is.null(med) || length(med) == 0) stop("Provide col_map$med mapping, e.g., list(med=list(ssri='med_ssri')).")

  get_flag <- function(col) {
    if (!col %in% names(data)) return(rep(NA, nrow(data)))
    x <- data[[col]]
    as.logical(if (is.numeric(x)) x != 0 else x)
  }

  ssri <- if (!is.null(med$ssri)) get_flag(med$ssri) else rep(NA, nrow(data))
  snri <- if (!is.null(med$snri)) get_flag(med$snri) else rep(NA, nrow(data))
  ap   <- if (!is.null(med$antipsychotic)) get_flag(med$antipsychotic) else rep(NA, nrow(data))
  ms   <- if (!is.null(med$mood_stabilizer)) get_flag(med$mood_stabilizer) else rep(NA, nrow(data))
  anx  <- if (!is.null(med$anxiolytic)) get_flag(med$anxiolytic) else rep(NA, nrow(data))

  all_med <- cbind(ssri, snri, ap, ms, anx)
  med_count <- rowSums(all_med, na.rm = TRUE)

  out <- tibble::tibble(
    !!paste0(prefix, "_any_psych") := (ssri | snri | ap | ms | anx),
    !!paste0(prefix, "_count") := med_count
  )

  .hm_add_cols(data, out, na_action = na_action)
}

# Dispatcher ------------------------------------------------------------------

#' Psychometric markers dispatcher
#' @inheritParams phq9_score
#' @param which Vector of modules to compute (e.g., "phq9", "gad7", "bis").
#' @param bis_key SPQ/BIS key list passed to `bis_score` when requested.
#' @param spq_key SPQ key list passed to `spq_score` when requested.
#' @param cognitive_method Method passed to `cognitive_score` ("z_mean" or "pca1").
#' @param col_map Nested list of mappings per instrument (e.g., col_map$phq9, col_map$bis, col_map$dx_flags, ...).
#' @return A tibble of computed score columns from all requested modules, bound together. No input columns are included in the output.
#' @examples
#' df <- data.frame(
#'   phq9_01 = 0, phq9_02 = 1, phq9_03 = 2, phq9_04 = 1, phq9_05 = 0,
#'   phq9_06 = 1, phq9_07 = 2, phq9_08 = 1, phq9_09 = 0,
#'   gad7_01 = 0, gad7_02 = 1, gad7_03 = 2, gad7_04 = 1, gad7_05 = 0,
#'   gad7_06 = 1, gad7_07 = 2
#' )
#' psych_markers(df, which = c("phq9", "gad7"))
#' @export
psych_markers <- function(data,
                          col_map = list(),
                          which = c("phq9","gad7","k6","k10","ghq12_likert","ghq12_binary","who5","isi","mdq","asrs","bis","spq","cognitive","dx_flags","med_flags"),
                          na_action = c("keep","omit","error"),
                          missing_prop_max = 0.2,
                          impute = c("none","mean"),
                          bis_key = NULL,
                          spq_key = NULL,
                          cognitive_method = c("z_mean","pca1"),
                          verbose = TRUE) {

  na_action <- match.arg(na_action, c("keep","omit","error"))
  impute <- match.arg(impute, c("none","mean"))
  cognitive_method <- match.arg(cognitive_method, c("z_mean","pca1"))

  hm_inform("psych_markers(): preparing inputs", level = "inform")

  base <- tibble::as_tibble(data)
  parts <- list()

  if ("phq9" %in% which) parts$phq9 <- phq9_score(base, col_map = col_map$phq9, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, variant = "PHQ9", prefix = "PHQ9", verbose = FALSE)
  if ("gad7" %in% which) parts$gad7 <- gad7_score(base, col_map = col_map$gad7, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "GAD7", verbose = FALSE)
  if ("k6"   %in% which) parts$k6   <- k6_score(base,   col_map = col_map$k6,   na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "K6", verbose = FALSE)
  if ("k10"  %in% which) parts$k10  <- k10_score(base,  col_map = col_map$k10,  na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "K10", verbose = FALSE)

  if ("ghq12_likert" %in% which) parts$ghq12_likert <- ghq12_score(base, col_map = col_map$ghq12, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, method = "likert", prefix = "GHQ12", verbose = FALSE)
  if ("ghq12_binary" %in% which) parts$ghq12_binary <- ghq12_score(base, col_map = col_map$ghq12, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, method = "binary", prefix = "GHQ12", verbose = FALSE)

  if ("who5" %in% which) parts$who5 <- who5_score(base, col_map = col_map$who5, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "WHO5", verbose = FALSE)
  if ("isi"  %in% which) parts$isi  <- isi_score(base,  col_map = col_map$isi,  na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "ISI", verbose = FALSE)

  if ("mdq"  %in% which) parts$mdq  <- mdq_score(base,  col_map = col_map$mdq,  na_action = na_action, missing_prop_max = missing_prop_max, prefix = "MDQ", verbose = FALSE)
  if ("asrs" %in% which) parts$asrs <- asrs_score(base, col_map = col_map$asrs, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "ASRS", verbose = FALSE)

  if ("bis" %in% which && !is.null(bis_key)) parts$bis <- bis_score(base, col_map = col_map$bis, key = bis_key, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "BIS", verbose = FALSE)
  if ("spq" %in% which && !is.null(spq_key)) parts$spq <- spq_score(base, col_map = col_map$spq, key = spq_key, na_action = na_action, missing_prop_max = missing_prop_max, impute = impute, prefix = "SPQ", verbose = FALSE)

  if ("cognitive" %in% which && !is.null(col_map$cognitive)) parts$cognitive <- cognitive_score(base, col_map = col_map$cognitive, na_action = na_action, missing_prop_max = missing_prop_max, method = cognitive_method, prefix = "cog")
  if ("dx_flags"  %in% which && !is.null(col_map$dx_flags))  parts$dx_flags  <- psych_dx_flags(base,  col_map = col_map$dx_flags,  na_action = na_action, prefix = "dx")
  if ("med_flags" %in% which && !is.null(col_map$med_flags)) parts$med_flags <- psych_med_flags(base, col_map = col_map$med_flags, na_action = na_action, prefix = "med")

  out <- if (length(parts)) do.call(dplyr::bind_cols, parts) else tibble::tibble(.rows = nrow(base))

  if (isTRUE(verbose)) hm_inform(sprintf("psych_markers(): results: %d rows, %d new columns",
                    nrow(out), ncol(out)), level = "inform")

  out
}

# tests/testthat/test-psych_scores.R

library(testthat)
library(tibble)
library(dplyr)

sample_psych_df <- tibble(
  phq9_01 = 1, phq9_02 = 2, phq9_03 = 1, phq9_04 = 0, phq9_05 = 1, phq9_06 = 2, phq9_07 = 1, phq9_08 = 2, phq9_09 = 1,
  gad7_01 = 0, gad7_02 = 1, gad7_03 = 1, gad7_04 = 1, gad7_05 = 0, gad7_06 = 1, gad7_07 = 0,
  k6_01 = 2, k6_02 = 1, k6_03 = 2, k6_04 = 1, k6_05 = 2, k6_06 = 1,
  k10_01 = 1, k10_02 = 1, k10_03 = 1, k10_04 = 1, k10_05 = 1, k10_06 = 1, k10_07 = 1, k10_08 = 1, k10_09 = 1, k10_10 = 1,
  ghq12_01 = 0, ghq12_02 = 1, ghq12_03 = 2, ghq12_04 = 1, ghq12_05 = 0, ghq12_06 = 2,
  ghq12_07 = 1, ghq12_08 = 0, ghq12_09 = 1, ghq12_10 = 0, ghq12_11 = 1, ghq12_12 = 0,
  who5_01 = 3, who5_02 = 3, who5_03 = 4, who5_04 = 3, who5_05 = 4,
  isi_01 = 2, isi_02 = 1, isi_03 = 2, isi_04 = 1, isi_05 = 0, isi_06 = 1, isi_07 = 0,
  mdq_01 = 1, mdq_02 = 1, mdq_03 = 1, mdq_04 = 0, mdq_05 = 1, mdq_06 = 1, mdq_07 = 1, mdq_08 = 0, mdq_09 = 1, mdq_10 = 1, mdq_11 = 0, mdq_12 = 1, mdq_13 = 1,
  mdq_cluster = 1, mdq_impair = 1,
  asrs_01 = 4, asrs_02 = 3, asrs_03 = 3, asrs_04 = 4, asrs_05 = 4, asrs_06 = 4,
  asrs_07 = 1, asrs_08 = 1, asrs_09 = 1, asrs_10 = 1, asrs_11 = 1, asrs_12 = 1,
  asrs_13 = 1, asrs_14 = 1, asrs_15 = 1, asrs_16 = 1, asrs_17 = 1, asrs_18 = 1,
  bis_01 = 2, bis_02 = 3, bis_03 = 2, bis_04 = 3, bis_05 = 1, bis_06 = 2,
  bis_07 = 2, bis_08 = 2, bis_09 = 2, bis_10 = 2, bis_11 = 2, bis_12 = 2,
  task_rt = 300, task_mem = 0.8, task_att = 0.4,
  dx_mdd = 1, dx_anxiety = 0, dx_adhd = 1, dx_bipolar = 0, dx_scz = 0, dx_sud = 0,
  med_ssri = 1, med_snri = 0, med_antipsychotic = 0, med_mood_stabilizer = 0, med_anxiolytic = 0
)

bis_key_small <- list(
  name = "BIS_demo",
  items = sprintf("bis_%02d", 1:12),
  min_val = 1, max_val = 4,
  reverse = c("bis_02","bis_03"),
  subscales = list(
    attention = c("bis_01","bis_05"),
    motor = c("bis_02","bis_06")
  )
)

test_that("phq9_score computes totals and severity", {
  skip_on_cran()
  out <- phq9_score(sample_psych_df, col_map = list(items = setNames(names(sample_psych_df)[1:9], sprintf("phq9_%02d", 1:9))))
  expect_equal(out$PHQ9_total, sum(c(1,2,1,0,1,2,1,2,1)))
  expect_s3_class(out$PHQ9_severity, "factor")
})

test_that("gad7_score severity bands work", {
  skip_on_cran()
  out <- gad7_score(sample_psych_df, col_map = list(items = setNames(sprintf("gad7_%02d", 1:7), sprintf("gad7_%02d", 1:7))))
  expect_true(out$GAD7_total >= 0)
  expect_true(out$GAD7_severity %in% c("minimal","mild","moderate","severe"))
})

test_that("who5_score percent and low wellbeing flag", {
  skip_on_cran()
  out <- who5_score(sample_psych_df, col_map = list(items = setNames(sprintf("who5_%02d", 1:5), sprintf("who5_%02d", 1:5))))
  expect_equal(out$WHO5_raw, sum(c(3,3,4,3,4)))
  expect_equal(out$WHO5_percent, out$WHO5_raw * 4)
  expect_false(out$WHO5_low_wellbeing)
})

test_that("asrs_score part A thresholding", {
  skip_on_cran()
  out <- asrs_score(sample_psych_df, col_map = list(items = setNames(sprintf("asrs_%02d", 1:18), sprintf("asrs_%02d", 1:18))))
  expect_equal(out$ASRS_partA_count, 6)
  expect_true(out$ASRS_partA_positive)
})

test_that("bis_score key-driven scoring works", {
  skip_on_cran()
  out <- bis_score(sample_psych_df, col_map = list(items = setNames(sprintf("bis_%02d", 1:12), sprintf("bis_%02d", 1:12))), key = bis_key_small)
  expect_equal(out$BIS_total, sum(c(2, (5-3), (5-2),3,1, 2,2,2,2,2,2,2)))
  expect_true(!is.na(out$BIS_attention))
  expect_true(!is.na(out$BIS_motor))
})

test_that("cognitive_score produces z_mean", {
  skip_on_cran()
  cm <- list(tasks = list(rt = "task_rt", memory = "task_mem", attention = "task_att"))
  out <- cognitive_score(sample_psych_df, col_map = cm, method = "z_mean")
  expect_true("cog_z_mean" %in% names(out))
})

test_that("diagnosis and medication flags aggregate", {
  skip_on_cran()
  dx_map <- list(dx = list(mdd = "dx_mdd", anxiety = "dx_anxiety", adhd = "dx_adhd", bipolar = "dx_bipolar", scz = "dx_scz", sud = "dx_sud"))
  med_map <- list(med = list(ssri = "med_ssri", snri = "med_snri", antipsychotic = "med_antipsychotic", mood_stabilizer = "med_mood_stabilizer", anxiolytic = "med_anxiolytic"))

  dx_out <- psych_dx_flags(sample_psych_df, col_map = dx_map)
  med_out <- psych_med_flags(sample_psych_df, col_map = med_map)

  expect_true(dx_out$dx_any_psych)
  expect_equal(dx_out$dx_count, 2)
  expect_true(med_out$med_any_psych)
  expect_equal(med_out$med_count, 1)
})

# ---- K6 --------------------------------------------------------------------
test_that("k6_score returns tibble with total and case flag", {
  skip_on_cran()
  out <- k6_score(sample_psych_df)
  expect_s3_class(out, "tbl_df")
  expect_true("K6_total" %in% names(out))
  expect_true("K6_case" %in% names(out))
  expect_equal(out$K6_total, sum(c(2, 1, 2, 1, 2, 1)))
})

test_that("k6_score errors on missing columns", {
  skip_on_cran()
  expect_error(k6_score(data.frame(x = 1)))
})

# ---- K10 -------------------------------------------------------------------
test_that("k10_score returns tibble with total", {
  skip_on_cran()
  out <- k10_score(sample_psych_df)
  expect_s3_class(out, "tbl_df")
  expect_true("K10_total" %in% names(out))
  expect_equal(out$K10_total, 10L)
})

# ---- GHQ-12 ----------------------------------------------------------------
test_that("ghq12_score likert returns total_likert", {
  skip_on_cran()
  out <- ghq12_score(sample_psych_df, method = "likert")
  expect_s3_class(out, "tbl_df")
  expect_true("GHQ12_total_likert" %in% names(out))
})

test_that("ghq12_score binary returns total_binary and case", {
  skip_on_cran()
  out <- ghq12_score(sample_psych_df, method = "binary")
  expect_s3_class(out, "tbl_df")
  expect_true("GHQ12_total_binary" %in% names(out))
  expect_true("GHQ12_case_binary" %in% names(out))
})

# ---- ISI -------------------------------------------------------------------
test_that("isi_score returns tibble with total and severity", {
  skip_on_cran()
  out <- isi_score(sample_psych_df)
  expect_s3_class(out, "tbl_df")
  expect_true("ISI_total" %in% names(out))
  expect_true("ISI_severity" %in% names(out))
  expect_equal(out$ISI_total, sum(c(2, 1, 2, 1, 0, 1, 0)))
})

# ---- MDQ -------------------------------------------------------------------
test_that("mdq_score returns tibble with symptom_count and positive_screen", {
  skip_on_cran()
  out <- mdq_score(sample_psych_df)
  expect_s3_class(out, "tbl_df")
  expect_true("MDQ_symptom_count" %in% names(out))
  expect_true("MDQ_positive_screen" %in% names(out))
  expect_equal(out$MDQ_symptom_count, sum(c(1,1,1,0,1,1,1,0,1,1,0,1,1)))
})

# ---- SPQ -------------------------------------------------------------------
test_that("spq_score returns tibble with total using key", {
  skip_on_cran()
  spq_key <- list(items = c("mdq_01", "mdq_02", "mdq_03"), min_val = 0, max_val = 1)
  out <- spq_score(sample_psych_df, key = spq_key, prefix = "SPQ")
  expect_s3_class(out, "tbl_df")
  expect_true("SPQ_total" %in% names(out))
})

test_that("psych_dx_flags errors when dx map is missing", {
  skip_on_cran()
  expect_error(psych_dx_flags(data.frame(x = 1:2), col_map = list()))
})

test_that("psych_med_flags errors when med map is missing", {
  skip_on_cran()
  expect_error(psych_med_flags(data.frame(x = 1:2), col_map = list()))
})

# ---- Formula / severity correctness ----------------------------------------
test_that("phq9_score correct total and severity label", {
  skip_on_cran()
  out <- phq9_score(sample_psych_df)
  expect_equal(out$PHQ9_total, 11L)
  expect_equal(as.character(out$PHQ9_severity), "moderate")
})

test_that("isi_score severity band: total=7 -> none", {
  skip_on_cran()
  out <- isi_score(sample_psych_df)
  expect_equal(out$ISI_total, 7L)
  expect_equal(as.character(out$ISI_severity), "none")
})

test_that("who5_score percent = raw * 4", {
  skip_on_cran()
  out <- who5_score(sample_psych_df)
  expect_equal(out$WHO5_percent, out$WHO5_raw * 4)
  expect_false(out$WHO5_low_wellbeing)
})

test_that("ghq12_score binary scoring correct", {
  skip_on_cran()
  # values: 0,1,2,1,0,2,1,0,1,0,1,0 -> binary (>=2->1): 0,0,1,0,0,1,0,0,0,0,0,0 = 2
  out <- ghq12_score(sample_psych_df, method = "binary")
  expect_equal(out$GHQ12_total_binary, 2L)
  expect_false(out$GHQ12_case_binary)  # 2 < 3 (default cutoff)
})

test_that("na_action='omit' in phq9_score drops NA rows", {
  skip_on_cran()
  df <- tibble::tibble(
    phq9_01 = c(1, NA), phq9_02 = 1, phq9_03 = 1, phq9_04 = 1, phq9_05 = 1,
    phq9_06 = 1, phq9_07 = 1, phq9_08 = 1, phq9_09 = 1
  )
  out <- phq9_score(df, na_action = "omit")
  expect_equal(nrow(out), 1L)
})

test_that("na_action='error' in phq9_score aborts on NA", {
  skip_on_cran()
  df <- tibble::tibble(
    phq9_01 = c(1, NA), phq9_02 = 1, phq9_03 = 1, phq9_04 = 1, phq9_05 = 1,
    phq9_06 = 1, phq9_07 = 1, phq9_08 = 1, phq9_09 = 1
  )
  expect_error(phq9_score(df, na_action = "error"))
})

# ---- Verbose ----------------------------------------------------------------
test_that("phq9_score verbose emits preparing and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(phq9_score(sample_psych_df, verbose = TRUE), "phq9_score")
  expect_message(phq9_score(sample_psych_df, verbose = TRUE), "results:")
})

test_that("psych_markers verbose emits preparing and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    psych_markers(sample_psych_df, which = "phq9", verbose = TRUE),
    "psych_markers"
  )
  expect_message(
    psych_markers(sample_psych_df, which = "phq9", verbose = TRUE),
    "results:"
  )
})

# ---- na_action = "keep" preserves NAs in output ----------------------------
test_that("phq9_score na_action='keep' preserves NA rows", {
  skip_on_cran()
  df <- tibble::tibble(
    phq9_01 = c(1, NA), phq9_02 = 1, phq9_03 = 1, phq9_04 = 1, phq9_05 = 1,
    phq9_06 = 1, phq9_07 = 1, phq9_08 = 1, phq9_09 = 1
  )
  out <- phq9_score(df, na_action = "keep")
  expect_equal(nrow(out), 2L)
  expect_true(is.na(out$PHQ9_total[2]))
})

# ---- .hm_severity_band() guard: mismatched cuts/labels ---------------------
test_that(".hm_severity_band() errors when cuts and labels are mismatched", {
  skip_on_cran()
  # 3 cuts but 3 labels (should be 4 cuts for 3 labels)
  expect_error(
    HealthMarkers:::.hm_severity_band(c(5, 10, 15), cuts = c(0, 5, 10), labels = c("low", "med", "high")),
    "one more element"
  )
})

# ---- PHQ-9 with 8 items (item 9 missing) ------------------------------------
test_that("phq9_score handles 8 items with impute mean (PHQ-8 equivalent)", {
  skip_on_cran()
  df <- tibble::tibble(
    phq9_01 = 1, phq9_02 = 2, phq9_03 = 1, phq9_04 = 0, phq9_05 = 1,
    phq9_06 = 2, phq9_07 = 1, phq9_08 = 2, phq9_09 = NA_integer_
  )
  # With impute = "mean" the missing item should be filled by row mean
  out <- suppressWarnings(phq9_score(df, na_action = "keep"))
  expect_equal(nrow(out), 1L)
  # total should be NA since na_action = "keep" propagates NAs
  expect_true(is.na(out$PHQ9_total))
})

# tests/testthat/test-psych_markers.R

library(testthat)
library(tibble)

mini_df <- tibble(
  phq9_01 = 0, phq9_02 = 1, phq9_03 = 2, phq9_04 = 1, phq9_05 = 1, phq9_06 = 1, phq9_07 = 1, phq9_08 = 1, phq9_09 = 0,
  gad7_01 = 0, gad7_02 = 0, gad7_03 = 0, gad7_04 = 1, gad7_05 = 1, gad7_06 = 0, gad7_07 = 0,
  k6_01 = 1, k6_02 = 1, k6_03 = 1, k6_04 = 1, k6_05 = 1, k6_06 = 1,
  k10_01 = 1, k10_02 = 1, k10_03 = 1, k10_04 = 1, k10_05 = 1, k10_06 = 1, k10_07 = 1, k10_08 = 1, k10_09 = 1, k10_10 = 1,
  ghq12_01 = 0, ghq12_02 = 1, ghq12_03 = 1, ghq12_04 = 0, ghq12_05 = 1, ghq12_06 = 0,
  ghq12_07 = 0, ghq12_08 = 0, ghq12_09 = 1, ghq12_10 = 0, ghq12_11 = 1, ghq12_12 = 0,
  who5_01 = 4, who5_02 = 4, who5_03 = 4, who5_04 = 4, who5_05 = 4,
  isi_01 = 0, isi_02 = 0, isi_03 = 0, isi_04 = 0, isi_05 = 0, isi_06 = 0, isi_07 = 0,
  mdq_01 = 1, mdq_02 = 1, mdq_03 = 1, mdq_04 = 1, mdq_05 = 1, mdq_06 = 1, mdq_07 = 1, mdq_08 = 1, mdq_09 = 1, mdq_10 = 1, mdq_11 = 1, mdq_12 = 1, mdq_13 = 1,
  mdq_cluster = 1, mdq_impair = 1,
  asrs_01 = 3, asrs_02 = 3, asrs_03 = 3, asrs_04 = 3, asrs_05 = 3, asrs_06 = 3,
  asrs_07 = 1, asrs_08 = 1, asrs_09 = 1, asrs_10 = 1, asrs_11 = 1, asrs_12 = 1,
  asrs_13 = 1, asrs_14 = 1, asrs_15 = 1, asrs_16 = 1, asrs_17 = 1, asrs_18 = 1,
  bis_01 = 1, bis_02 = 2, bis_03 = 3, bis_04 = 4, bis_05 = 2, bis_06 = 1,
  bis_07 = 2, bis_08 = 2, bis_09 = 2, bis_10 = 2, bis_11 = 2, bis_12 = 2,
  spq_01 = 1, spq_02 = 1, spq_03 = 1, spq_04 = 1,
  task_rt = 350, task_mem = 0.7, task_att = 0.5,
  dx_mdd = 0, dx_anxiety = 0, dx_adhd = 0, dx_bipolar = 0, dx_scz = 0, dx_sud = 0,
  med_ssri = 0, med_snri = 0, med_antipsychotic = 0, med_mood_stabilizer = 0, med_anxiolytic = 0
)

col_map_all <- list(
  phq9 = list(items = setNames(sprintf("phq9_%02d", 1:9), sprintf("phq9_%02d", 1:9))),
  gad7 = list(items = setNames(sprintf("gad7_%02d", 1:7), sprintf("gad7_%02d", 1:7))),
  k6   = list(items = setNames(sprintf("k6_%02d", 1:6), sprintf("k6_%02d", 1:6))),
  k10  = list(items = setNames(sprintf("k10_%02d", 1:10), sprintf("k10_%02d", 1:10))),
  ghq12 = list(items = setNames(sprintf("ghq12_%02d", 1:12), sprintf("ghq12_%02d", 1:12))),
  who5 = list(items = setNames(sprintf("who5_%02d", 1:5), sprintf("who5_%02d", 1:5))),
  isi  = list(items = setNames(sprintf("isi_%02d", 1:7), sprintf("isi_%02d", 1:7))),
  mdq  = list(items = setNames(sprintf("mdq_%02d", 1:13), sprintf("mdq_%02d", 1:13)), cluster = "mdq_cluster", impairment = "mdq_impair"),
  asrs = list(items = setNames(sprintf("asrs_%02d", 1:18), sprintf("asrs_%02d", 1:18))),
  bis  = list(items = setNames(sprintf("bis_%02d", 1:12), sprintf("bis_%02d", 1:12))),
  spq  = list(items = setNames(sprintf("spq_%02d", 1:4), sprintf("spq_%02d", 1:4))),
  cognitive = list(tasks = list(rt = "task_rt", memory = "task_mem", attention = "task_att")),
  dx_flags = list(dx = list(mdd = "dx_mdd", anxiety = "dx_anxiety", adhd = "dx_adhd", bipolar = "dx_bipolar", scz = "dx_scz", sud = "dx_sud")),
  med_flags = list(med = list(ssri = "med_ssri", snri = "med_snri", antipsychotic = "med_antipsychotic", mood_stabilizer = "med_mood_stabilizer", anxiolytic = "med_anxiolytic"))
)

bis_key_demo <- list(
  name = "BIS_demo",
  items = sprintf("bis_%02d", 1:12),
  min_val = 1, max_val = 4,
  reverse = c("bis_02","bis_03"),
  subscales = list(motor = c("bis_02","bis_06"))
)

spq_key_demo <- list(
  name = "SPQ_demo",
  items = sprintf("spq_%02d", 1:4),
  min_val = 0, max_val = 1
)

test_that("psych_markers returns combined outputs for requested scales", {
  out <- psych_markers(
    mini_df,
    col_map = col_map_all,
    which = c("phq9","gad7","k6","ghq12_binary","who5","isi","mdq","asrs","bis","spq","cognitive","dx_flags","med_flags"),
    bis_key = bis_key_demo,
    spq_key = spq_key_demo,
    na_action = "keep",
    missing_prop_max = 0.3,
    impute = "mean"
  )

  expect_true("PHQ9_total" %in% names(out))
  expect_true("GAD7_total" %in% names(out))
  expect_true("K6_total" %in% names(out))
  expect_true("GHQ12_total_binary" %in% names(out))
  expect_true("WHO5_percent" %in% names(out))
  expect_true("ISI_total" %in% names(out))
  expect_true("MDQ_positive_screen" %in% names(out))
  expect_true("ASRS_partA_positive" %in% names(out))
  expect_true("BIS_total" %in% names(out))
  expect_true("SPQ_total" %in% names(out))
  expect_true("cog_z_mean" %in% names(out))
  expect_true("dx_any_psych" %in% names(out))
  expect_true("med_any_psych" %in% names(out))
})

test_that("missingness policy propagate NA when too much is missing", {
  skip_on_cran()
  df_na <- mini_df
  df_na$phq9_01 <- NA
  df_na$phq9_02 <- NA
  out <- phq9_score(df_na, col_map = col_map_all$phq9, missing_prop_max = 0.1, impute = "none", na_action = "keep")
  expect_true(is.na(out$PHQ9_total))
})

test_that("psych_markers verbose emits preparing and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(psych_markers(mini_df, which = "phq9", verbose = TRUE), "psych_markers")
  expect_message(psych_markers(mini_df, which = "phq9", verbose = TRUE), "results:")
})

test_that("psych_markers verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(psych_markers(mini_df, which = "phq9", verbose = TRUE))
  expect_equal(sum(grepl("results:", msgs)), 1L)
})

#' Kynurenine/Tryptophan Ratio (KTR)
#'
#' Computes the ratio of kynurenine to tryptophan, a marker of IDO activity and immune activation.
#'
#' @details
#' KTR is calculated as Kyn (nmol/L) divided by Trp (mumol/L). Elevated KTR indicates
#' increased tryptophan catabolism via the kynurenine pathway, often reflecting
#' inflammation and cell-mediated immune activation.
#'
#' Inputs should already be in Kyn (nmol/L) and Trp (mumol/L).
#'
#' @param data A data.frame or tibble with kynurenine and tryptophan concentrations.
#' @param col_map Named list with:
#'   - kynurenine: column for kynurenine (nmol/L)
#'   - tryptophan: column for tryptophan (mumol/L)
#' @param na_action One of c("keep","omit","error","ignore","warn").
#' @param verbose Logical; if `TRUE` (default), prints column mapping and a
#'   per-column results summary.
#' @return A tibble with one column: kyn_trp_ratio (numeric). If an ID column
#'   is detected, it is prepended.
#'
#' @examples
#' # columns named exactly as the required keys (auto-detected)
#' df <- data.frame(kynurenine = c(2500, 3100, 2700), tryptophan = c(55, 48, 62))
#' kyn_trp_ratio(df, verbose = FALSE)
#'
#' # non-standard column names require explicit col_map
#' df2 <- data.frame(Kyn_nM = c(2500, 3100, 2700), Trp_uM = c(55, 48, 62))
#' kyn_trp_ratio(df2, col_map = list(kynurenine = "Kyn_nM", tryptophan = "Trp_uM"),
#'              verbose = FALSE)
#'
#' @references
#' \insertRef{fuchs1998ktr}{HealthMarkers};
#' \insertRef{damerell2025kyn}{HealthMarkers} (clinical application in colorectal cancer)
#'
#' @export
kyn_trp_ratio <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name  <- "kyn_trp_ratio"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col   <- .hm_detect_id_col(data)

  # Validate
  if (!is.data.frame(data)) {
    rlang::abort("kyn_trp_ratio(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_ktr_error_data_type")
  }

  req <- c("kynurenine","tryptophan")
  cm      <- .hm_build_col_map(data, col_map, req, fn = "kyn_trp_ratio")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  # Graceful NA only when nothing provided and inference found nothing
  if (length(cm$user_keys) == 0L && length(cm$inferred_keys) == 0L && length(col_map) == 0L) {
    return(tibble::tibble(kyn_trp_ratio = rep(NA_real_, nrow(data))))
  }

  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("kyn_trp_ratio(): `col_map` must be a named list.",
                 class = "healthmarkers_ktr_error_colmap_type")
  }

  req <- c("kynurenine","tryptophan")
  missing_keys <- setdiff(req, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(paste0("kyn_trp_ratio(): missing col_map entries for: ",
                        paste(missing_keys, collapse = ", ")),
                 class = "healthmarkers_ktr_error_missing_map")
  }
  mapped <- unname(unlist(col_map[req], use.names = FALSE))
  if (any(!nzchar(mapped))) {
    bad <- req[!nzchar(unname(unlist(col_map[req])))]
    rlang::abort(paste0("kyn_trp_ratio(): missing col_map entries for: ",
                        paste(bad, collapse = ", ")),
                 class = "healthmarkers_ktr_error_bad_map_values")
  }
  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(paste0("kyn_trp_ratio(): missing required columns in data: ",
                        paste(missing_cols, collapse = ", ")),
                 class = "healthmarkers_ktr_error_missing_columns")
  }

  hm_inform("kyn_trp_ratio(): preparing inputs", level = if (isTRUE(verbose)) "inform" else "debug")

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    hm_inform(
      sprintf("%s(): computing markers:\n  kyn_trp_ratio  [kynurenine (nmol/L) / tryptophan (mumol/L)]", fn_name),
      level = "inform"
    )
  }

  # Coerce numeric; warn if NAs introduced; sanitize
  for (cn in mapped) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0L) {
        rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
                    class = "healthmarkers_ktr_warn_na_coercion")
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  kyn <- data[[col_map$kynurenine]]
  trp <- data[[col_map$tryptophan]]

  # NA policy
  any_na <- is.na(kyn) | is.na(trp)
  if (na_action_raw == "warn" && any(any_na)) {
    rlang::warn("Missing kynurenine or tryptophan; ratio will be NA for those entries.",
                class = "healthmarkers_ktr_warn_missing_inputs")
  }
  if (na_action_eff == "error" && any(any_na)) {
    rlang::abort("kyn_trp_ratio(): required inputs contain missing values (na_action='error').",
                 class = "healthmarkers_ktr_error_missing_values")
  }
  keep <- if (na_action_eff == "omit") !any_na else rep(TRUE, length(kyn))

  d_kyn <- kyn[keep]
  d_trp <- trp[keep]

  # Domain warnings: nonpositive tryptophan
  if (any(is.finite(d_trp) & d_trp <= 0)) {
    rlang::warn("kyn_trp_ratio(): nonpositive tryptophan detected; ratios set to NA for those rows.",
                class = "healthmarkers_ktr_warn_nonpositive_trp")
    d_trp[is.finite(d_trp) & d_trp <= 0] <- NA_real_
  }

  hm_inform("kyn_trp_ratio(): computing", level = "debug")

  ratio <- d_kyn / d_trp
  ratio[!is.finite(ratio)] <- NA_real_

  # High ratio warning
  if (any(is.finite(ratio) & ratio > 100)) {
    rlang::warn("kyn_trp_ratio(): very high ratios (>100) detected; check units (Kyn nmol/L, Trp mumol/L).",
                class = "healthmarkers_ktr_warn_ratio_high")
  }

  out <- tibble::tibble(kyn_trp_ratio = ratio)

  # Pad back to full length when not omitting (preserve row alignment)
  if (na_action_eff != "omit") {
    res <- tibble::tibble(kyn_trp_ratio = rep(NA_real_, length(kyn)))
    res$kyn_trp_ratio[keep] <- out$kyn_trp_ratio
    out <- res
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    out[[id_col]] <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

library(testthat)

cm <- list(kynurenine = "Kyn_nM", tryptophan = "Trp_uM")

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = 2000, Trp_uM = 60)

  # empty list with inferrable cols – succeeds via inference
  expect_no_error(kyn_trp_ratio(df, list()))

  expect_error(
    kyn_trp_ratio(df, list(kynurenine = "", tryptophan = "Trp_uM")),
    class = "healthmarkers_ktr_error_bad_map_values"
  )

  expect_error(
    kyn_trp_ratio(data.frame(A = 1, B = 1), cm),
    class = "healthmarkers_ktr_error_missing_columns"
  )
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(Kyn_nM = 2000, Trp_uM = 60)
  expect_message(kyn_trp_ratio(df, cm, verbose = TRUE), "kyn_trp_ratio")
  expect_message(kyn_trp_ratio(df, cm, verbose = TRUE), "col_map")
  expect_message(suppressWarnings(kyn_trp_ratio(df, cm, verbose = TRUE)), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(Kyn_nM = 2000, Trp_uM = 60)
  msgs <- testthat::capture_messages(
    kyn_trp_ratio(df, cm, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("numeric coercion warning when strings introduce NAs", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c("2000", "oops"), Trp_uM = c("60", "55"))
  expect_warning(kyn_trp_ratio(df, cm), class = "healthmarkers_ktr_warn_na_coercion")
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c(2000, NA, 8000), Trp_uM = c(60, 50, NA))
  out_keep <- kyn_trp_ratio(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(is.na(out_keep$kyn_trp_ratio[2]) || is.na(out_keep$kyn_trp_ratio[3]))

  out_omit <- kyn_trp_ratio(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)

  expect_error(kyn_trp_ratio(df, cm, na_action = "error"),
               class = "healthmarkers_ktr_error_missing_values")

  expect_warning(kyn_trp_ratio(df, cm, na_action = "warn"),
                 class = "healthmarkers_ktr_warn_missing_inputs")
})

test_that("division by zero or nonpositive Trp sets NA with warning", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c(2000, 3000), Trp_uM = c(0, -5))
  out <- expect_warning(kyn_trp_ratio(df, cm), class = "healthmarkers_ktr_warn_nonpositive_trp")
  expect_true(all(is.na(out$kyn_trp_ratio)))
})

test_that("ratio computation and high ratio warning", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c(2400, 8000), Trp_uM = c(60, 50))

  # Only checking ratio correctness here; hide the high-ratio warning
  out <- suppressWarnings(kyn_trp_ratio(df, cm))
  expect_equal(out$kyn_trp_ratio, c(2400/60, 8000/50), tolerance = 1e-12)

  # Here we explicitly assert on the high-ratio warning
  df_hi <- data.frame(Kyn_nM = 15000, Trp_uM = 50)
  expect_warning(
    kyn_trp_ratio(df_hi, cm),
    class = "healthmarkers_ktr_warn_ratio_high"
  )
})

test_that("extreme inputs produce NA ratios via zero/invalid denominator", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c(50, 25000, 3000, 15000), Trp_uM = c(0, 200, 60, 50))
  out <- suppressWarnings(kyn_trp_ratio(df, cm, verbose = FALSE))
  expect_true(any(is.na(out$kyn_trp_ratio)))
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(Kyn_nM = c(2000, NA, 8000), Trp_uM = c(60, 50, NA))
  out_keep <- kyn_trp_ratio(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(kyn_trp_ratio(df, cm, na_action = "warn"))
  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

#' Calculate fasting-based insulin sensitivity indices
#'
#' @description
#' Compute 10 fasting indices from glucose (mmol/L) and insulin (pmol/L):
#' Fasting_inv, Raynaud, HOMA_IR_inv, FIRI, QUICKI, Belfiore_basal, Ig_ratio_basal,
#' Isi_basal, Bennett, HOMA_IR_rev_inv. Units converted internally:
#' G0_mg = G0*18 (mg/dL), I0_u = I0/6 (muU/mL).
#'
#' @param data Data frame with required inputs.
#' @param col_map Named list mapping required keys:
#'   - G0: fasting glucose (mmol/L)
#'   - I0: fasting insulin (pmol/L)
#' @param normalize One of: "none","z","inverse","range","robust".
#' @param na_action One of:
#'   - "keep"  (retain all rows; indices become NA where inputs missing/non-finite)
#'   - "omit"  (drop rows with any missing/non-finite required inputs)
#'   - "error" (abort if any required input is missing/non-finite)
#'   - "warn"  (emit a warning for rows with missing inputs, then keep them)

#' @param verbose Logical; if `TRUE` (default), prints column mapping, the list
#'   of indices being computed, and a per-column results summary.
#'
#' @return Tibble with 10 columns (indices listed above). If an ID column is
#'   detected in `data` (e.g. `id`, `IID`, `participant_id`), it is prepended
#'   as the first output column.
#' @references
#' \insertRef{matthews1985homa}{HealthMarkers}
#' \insertRef{katz2000quicki}{HealthMarkers}
#' \insertRef{raynaud1999fasting}{HealthMarkers}
#' \insertRef{avignon1999ogtt}{HealthMarkers}
#' \insertRef{belfiore1998insulin}{HealthMarkers}
#' \insertRef{sluiter1976gtolerance}{HealthMarkers}
#' \insertRef{hanson2000evaluation}{HealthMarkers}
#' \insertRef{anderson1995exploration}{HealthMarkers}
#'
#' @examples
#' # Minimal example (units: G0 in mmol/L, I0 in pmol/L)
#' df <- data.frame(G0 = c(5.2, 6.1, 4.8), I0 = c(60, 120, 80))
#' res <- fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"))
#' head(res)
#'
#' # With NA handling
#' df2 <- data.frame(G0 = c(5.0, NA), I0 = c(90, 150))
#' fasting_is(df2, col_map = list(G0 = "G0", I0 = "I0"), na_action = "keep")
#' @export
fasting_is <- function(
  data,
  col_map     = NULL,
  normalize   = c("none", "z", "inverse", "range", "robust"),
  na_action   = c("keep", "omit", "error", "warn"),
  verbose     = TRUE
) {
  data_name    <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name     <- "fasting_is"
  .hm_log_input(data, data_name, fn_name, verbose)
  allowed_norm <- c("none", "z", "inverse", "range", "robust")
  if (length(normalize) == 1L && !normalize %in% allowed_norm) {
    rlang::abort(
      sprintf("`normalize` must be one of: %s", paste(allowed_norm, collapse = ", ")),
      class = "healthmarkers_fi_error_normalize"
    )
  }
  normalize <- match.arg(normalize, allowed_norm)
  na_action <- match.arg(na_action)

  # Explicit input validation
  if (!is.data.frame(data)) {
    rlang::abort("fasting_is(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_fi_error_data_type")
  }

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)
  req_keys <- c("G0","I0")
  cm      <- .hm_build_col_map(data, col_map, req_keys, fn = "fasting_is")
  data    <- cm$data
  col_map <- cm$col_map
  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("fasting_is(): `col_map` must be a named list with entries for G0 and I0.",
                 class = "healthmarkers_fi_error_colmap_type")
  }
  missing_keys <- setdiff(req_keys, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(
      paste0("fasting_is(): missing col_map entries for: ", paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_fi_error_missing_map"
    )
  }
  req_cols <- unname(unlist(col_map[req_keys], use.names = FALSE))
  missing_cols <- setdiff(req_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("fasting_is(): missing required columns in data: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_fi_error_missing_columns"
    )
  }

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    indices <- c("Fasting_inv","Raynaud","HOMA_IR_inv","FIRI","QUICKI",
                 "Belfiore_basal","Ig_ratio_basal","Isi_basal","Bennett","HOMA_IR_rev_inv")
    status <- vapply(indices, function(m) sprintf("  %-20s [G0, I0]", m), character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Coerce to numeric; warn if NAs introduced; non-finite -> NA
  for (nm in req_keys) {
    cn <- col_map[[nm]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced.", cn),
                    class = "healthmarkers_fi_warn_na_coercion")
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  G0 <- data[[col_map$G0]]
  I0 <- data[[col_map$I0]]

  # NA policy
  rows_with_na <- is.na(G0) | is.na(I0)
  if (na_action == "error" && any(rows_with_na)) {
    rlang::abort("fasting_is(): missing/non-finite inputs with na_action='error'.",
                 class = "healthmarkers_fi_error_missing")
  } else if (na_action == "warn" && any(rows_with_na)) {
    rlang::warn(sprintf("fasting_is(): %d row(s) have missing G0 or I0; results will be NA for those rows.",
                        sum(rows_with_na)),
                class = "healthmarkers_fi_warn_na")
  } else if (na_action == "omit" && any(rows_with_na)) {
    keep <- !rows_with_na
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    G0 <- G0[keep]; I0 <- I0[keep]
  }

  hm_inform(sprintf("%s(): converting units (mmol/L->mg/dL; pmol/L->muU/mL)", fn_name), level = "debug")

  G0_mg <- G0 * 18
  I0_u  <- I0 / 6

  hm_inform(sprintf("%s(): computing indices", fn_name), level = "debug")

  lg <- function(x) {
    y <- x
    y[!(is.finite(y) & y > 0)] <- NA_real_
    out <- suppressWarnings(log(y))
    out[!is.finite(out)] <- NA_real_
    out
  }
  sdiv <- function(a, b) {
    z <- a / b
    z[!is.finite(z)] <- NA_real_
    z
  }

  out <- tibble::tibble(
    Fasting_inv     = -I0_u,
    Raynaud         = sdiv(40, I0_u),
    HOMA_IR_inv     = -sdiv(G0_mg * I0_u, 22.5),
    FIRI            =  sdiv(G0_mg * I0_u, 25),
    QUICKI          = sdiv(1, lg(G0_mg) + lg(I0_u)),
    Belfiore_basal  = sdiv(2, (I0_u * G0_mg) + 1),
    Ig_ratio_basal  = -sdiv(I0_u, G0_mg),
    Isi_basal       = sdiv(10000, G0_mg * I0_u),
    Bennett         = sdiv(1, lg(I0_u) * lg(G0_mg)),
    HOMA_IR_rev_inv = -sdiv(I0_u * G0_mg, 405)
  )

  # Inline normalization
  if (normalize != "none") {
    normalize_vec <- function(x, method) {
      v <- as.numeric(x)
      nm <- is.na(v)
      v2 <- v[!nm]
      if (method == "z") {
        if (length(v2) < 2) return(rep(NA_real_, length(v)))
        mu <- mean(v2); sdv <- stats::sd(v2)
        if (!is.finite(sdv) || sdv == 0) return(rep(NA_real_, length(v)))
        (v - mu) / sdv
      } else if (method == "range") {
        if (length(v2) < 2) return(rep(NA_real_, length(v)))
        mn <- min(v2); mx <- max(v2)
        if (!is.finite(mx - mn) || mx - mn == 0) return(rep(NA_real_, length(v)))
        (v - mn) / (mx - mn)
      } else if (method == "inverse") {
        res <- 1 / v
        res[!is.finite(res)] <- NA_real_
        res
      } else if (method == "robust") {
        if (length(v2) < 2) return(rep(NA_real_, length(v)))
        med <- stats::median(v2, na.rm = TRUE)
        madv <- stats::mad(v2, center = med, constant = 1.4826, na.rm = TRUE)
        if (!is.finite(madv) || madv == 0) return(rep(NA_real_, length(v)))
        (v - med) / madv
      } else {
        v
      }
    }
    out[] <- lapply(out, normalize_vec, method = normalize)
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# tests/testthat/test_fasting_is.R

test_that("fasting_is errors if missing required columns", {
  skip_on_cran()
  df <- tibble::tibble(G0 = 5.5)
  expect_error(
    fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"))
  )
})

test_that("fasting_is returns 10 indices and computes HOMA_IR_inv correctly", {
  skip_on_cran()
  df <- tibble::tibble(G0 = 5.5, I0 = 60)
  out <- fasting_is(
    df,
    col_map   = list(G0 = "G0", I0 = "I0"),
    normalize = "none",
    verbose   = FALSE
  )

  expected_names <- c(
    "Fasting_inv", "Raynaud", "HOMA_IR_inv", "FIRI", "QUICKI",
    "Belfiore_basal", "Ig_ratio_basal", "Isi_basal",
    "Bennett", "HOMA_IR_rev_inv"
  )
  expect_named(out, expected_names, ignore.order = TRUE)
  expect_equal(ncol(out), length(expected_names))

  # manual HOMA_IR_inv = -((G0*18)*(I0/6)) / 22.5
  expected_homa <- -((5.5 * 18) * (60 / 6)) / 22.5
  expect_equal(out$HOMA_IR_inv, expected_homa)
})

test_that("fasting_is is vectorized over multiple rows", {
  skip_on_cran()
  df <- tibble::tibble(
    G0 = c(5.5, 6),
    I0 = c(60, 80)
  )
  out <- fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"))
  expect_equal(nrow(out), 2)
})

test_that("normalize = 'range' and 'z' behave correctly", {
  skip_on_cran()
  df2 <- tibble::tibble(G0 = c(5.5, 6), I0 = c(60, 80))

  out_r <- fasting_is(df2, col_map = list(G0 = "G0", I0 = "I0"), normalize = "range")
  for (col in names(out_r)) {
    vals <- out_r[[col]]
    v2 <- vals[!is.na(vals)]
    if (length(unique(v2)) > 1) {
      expect_equal(min(v2), 0)
      expect_equal(max(v2), 1)
    } else {
      expect_true(all(is.na(vals)))
    }
  }

  df3 <- tibble::tibble(
    G0 = c(5.5, 6, 7),
    I0 = c(60, 80, 100)
  )
  out_z <- fasting_is(df3, col_map = list(G0 = "G0", I0 = "I0"), normalize = "z")
  for (col in names(out_z)) {
    vals <- out_z[[col]]
    v2 <- vals[!is.na(vals)]
    if (length(unique(v2)) > 1) {
      expect_equal(mean(v2), 0, tolerance = 1e-6)
      expect_equal(sd(v2), 1, tolerance = 1e-6)
    } else {
      expect_true(all(is.na(vals)))
    }
  }
})

test_that("normalize = 'inverse' and 'robust' run silently", {
  skip_on_cran()
  df <- tibble::tibble(G0 = c(5.5, 6), I0 = c(60, 80))
  expect_silent(fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"), normalize = "inverse", verbose = FALSE))
  expect_silent(fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"), normalize = "robust",  verbose = FALSE))
})

test_that("invalid normalize argument errors", {
  skip_on_cran()
  df <- tibble::tibble(G0 = 5.5, I0 = 60)
  expect_error(
    fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"), normalize = "foo"),
    "`normalize` must be one of"
  )
})

test_that("verbose = TRUE emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble::tibble(G0 = 5.5, I0 = 60)
  cm <- list(G0 = "G0", I0 = "I0")
  expect_message(fasting_is(df, col_map = cm, verbose = TRUE), "fasting_is")
  expect_message(fasting_is(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(fasting_is(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble::tibble(G0 = 5.5, I0 = 60)
  cm   <- list(G0 = "G0", I0 = "I0")
  msgs <- testthat::capture_messages(fasting_is(df, col_map = cm, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

library(testthat)
library(HealthMarkers)

# ---- normalize_vec direct tests --------------------------------------------

test_that("normalize_vec method='none' returns input unchanged", {
  x <- c(1, 2, 3, NA, 5)
  expect_identical(normalize_vec(x, "none"), x)
})

test_that("normalize_vec method='z' gives mean~0 and sd~1 on non-NA values", {
  skip_on_cran()
  x <- c(1, 2, 3, 4, 5)
  out <- normalize_vec(x, "z")
  expect_equal(length(out), length(x))
  expect_equal(mean(out, na.rm = TRUE), 0, tolerance = 1e-10)
  expect_equal(sd(out, na.rm = TRUE), 1, tolerance = 1e-10)
})

test_that("normalize_vec method='z' preserves NA positions", {
  skip_on_cran()
  x <- c(1, NA, 3, 4, 5)
  out <- normalize_vec(x, "z")
  expect_true(is.na(out[2]))
  expect_false(any(is.na(out[-2])))
})

test_that("normalize_vec method='range' maps to [0,1] by default", {
  skip_on_cran()
  x <- c(1, 2, 3, 4, 5)
  out <- normalize_vec(x, "range")
  expect_equal(min(out, na.rm = TRUE), 0)
  expect_equal(max(out, na.rm = TRUE), 1)
})

test_that("normalize_vec method='range' custom feature_range works", {
  skip_on_cran()
  x <- c(1, 2, 3, 4, 5)
  out <- normalize_vec(x, "range", feature_range = c(-1, 1))
  expect_equal(min(out, na.rm = TRUE), -1)
  expect_equal(max(out, na.rm = TRUE), 1)
})

test_that("normalize_vec method='robust' handles non-zero MAD", {
  skip_on_cran()
  set.seed(42)
  x <- c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
  out <- normalize_vec(x, "robust")
  expect_equal(length(out), length(x))
  expect_type(out, "double")
})

test_that("normalize_vec method='inverse' returns rank-based normal scores", {
  skip_on_cran()
  x <- c(1, 2, 3, 4, 5)
  out <- normalize_vec(x, "inverse")
  expect_equal(length(out), length(x))
  # Inverse-normal output should be monotone increasing for sorted input
  expect_true(all(diff(out) > 0))
})

test_that("normalize_vec method='z' on constant vector returns zeros (warn suppressed)", {
  skip_on_cran()
  x <- rep(3, 5)
  out <- suppressWarnings(normalize_vec(x, "z", warn_constant = FALSE))
  expect_true(all(out == 0))
})

test_that("normalize_vec all-NA vector returns all-NA", {
  skip_on_cran()
  x <- rep(NA_real_, 4)
  out <- suppressWarnings(normalize_vec(x, "z"))
  expect_true(all(is.na(out)))
})

test_that("normalize_vec coerces non-numeric with NAs and warns", {
  skip_on_cran()
  # Provide a vector where coercion introduces NAs to trigger the warning
  x <- c("1", "not_a_number", "3")
  expect_warning(normalize_vec(x, "z"), "coerced to numeric")
})

# HM-CS v2 utilities

# Note: hm_inform() and hm_get_verbosity() are defined in zzz-options.R (canonical).



# Verbose stage 1 -- input logger ---------------------------------------------

# Emits: "fn(): reading input 'expr' -- N rows x P variables"

# Capture data_name via deparse(substitute(data)) as the very first line of

# the calling function, then call here right after fn_name is assigned.

#' @keywords internal

.hm_log_input <- function(data, data_name, fn_name, verbose) {

  if (!isTRUE(verbose)) return(invisible(NULL))

  if (length(data_name) != 1L || nchar(data_name) > 200L) data_name <- "data"

  dims <- if (is.data.frame(data)) {

    sprintf("%d rows \u00d7 %d variables", nrow(data), ncol(data))

  } else {

    sprintf("class: %s", paste(class(data), collapse = "/"))

  }

  hm_inform(

    sprintf("%s(): reading input '%s' \u2014 %s", fn_name, data_name, dims),

    level = "inform"

  )

  invisible(NULL)

}



# Verbose stage 2 -- column resolution logger ----------------------------------

# Replaces the repeated inline col_map verbose blocks in every function.

# Emits, e.g.:

#   fn(): resolving 5 columns -- 3 specified, 2 inferred from data

#     TC                ->  'TC'

#     HDL_c             ->  'HDL_c'

#     TG                ->  'TG'

#     LDL_c             ->  'LDL_c'    (inferred)

#     ApoB              ->  'ApoB'     (inferred)

#

# @param cm      result of .hm_build_col_map() -- needs $user_keys, $inferred_keys

# @param col_map the resolved named list key -> column name

# @param fn_name function name string

# @param verbose logical

# @param keys    optional character vector to restrict which keys are shown

#' @keywords internal

.hm_log_cols <- function(cm, col_map, fn_name, verbose, keys = NULL) {

  if (!isTRUE(verbose)) return(invisible(NULL))



  u_keys <- if (!is.null(keys)) intersect(cm$user_keys,     keys) else cm$user_keys

  i_keys <- if (!is.null(keys)) intersect(cm$inferred_keys, keys) else cm$inferred_keys



  n_user <- length(u_keys)

  n_inf  <- length(i_keys)

  total  <- n_user + n_inf



  if (total == 0L) {

    hm_inform(sprintf("%s(): no columns resolved", fn_name), level = "inform")

    return(invisible(NULL))

  }



  parts_desc <- character(0)

  if (n_user > 0L) parts_desc <- c(parts_desc, sprintf("%d specified",          n_user))

  if (n_inf  > 0L) parts_desc <- c(parts_desc, sprintf("%d inferred from data", n_inf))



  header <- sprintf("%s(): col_map (%d column%s \u2014 %s)",

                    fn_name, total,

                    if (total == 1L) "" else "s",

                    paste(parts_desc, collapse = ", "))



  fmt_key <- function(k, map, tag) {

    col <- map[[k]]

    if (is.null(col)) col <- "(unresolved)"

    sprintf("  %-18s->  '%s'%s", k, col, tag)

  }



  lines <- c(

    if (n_user > 0L) vapply(u_keys, fmt_key, character(1), map = col_map, tag = ""),

    if (n_inf  > 0L) vapply(i_keys, fmt_key, character(1), map = col_map, tag = "    (inferred)")

  )



  hm_inform(paste(c(header, lines), collapse = "\n"), level = "inform")

  invisible(NULL)

}



# Canonical sex normalizer ----------------------------------------------------

# Maps any sex encoding to a consistent output:

#   "MF"  -> "M" or "F"

#   "12"  -> 1L (male) or 2L (female)   [ogtt_is, metss, kidney_failure_risk]

#   "10"  -> 1L (male) or 0L (female)   [renal_markers, cvd_risk]

#   "01"  -> 0L (male) or 1L (female)   [obesity_indices RFM]

#

# Accepted inputs (case-insensitive):

#   Male   : "m", "male", "1"

#   Female : "f", "female", "2", "0"

#   (Convention: when "0" or "2" are mixed, "1" always = male per package default)

#

# @keywords internal

.hm_normalize_sex <- function(x, to = c("MF", "12", "10", "01"), fn = NULL) {

  to <- match.arg(to)

  s  <- trimws(tolower(as.character(x)))

  mf <- ifelse(s %in% c("m", "male", "1"),          "M",

        ifelse(s %in% c("f", "female", "2", "0"),   "F", NA_character_))

  bad <- sum(is.na(mf) & !is.na(x) & nchar(trimws(as.character(x))) > 0)

  if (bad > 0L && !is.null(fn))

    hm_inform(level = "debug",

              msg   = sprintf("%s(): 'sex' has %d unrecognized value(s); treated as NA.", fn, bad))

  switch(to,

    "MF" = mf,

    "12" = ifelse(mf == "M", 1L, ifelse(mf == "F", 2L, NA_integer_)),

    "10" = ifelse(mf == "M", 1L, ifelse(mf == "F", 0L, NA_integer_)),

    "01" = ifelse(mf == "M", 0L, ifelse(mf == "F", 1L, NA_integer_))

  )

}



#' Normalise na_action aliases to canonical values

#'

#' Converts the backward-compat aliases "ignore" and "warn" to "keep" so that

#' the rest of each function only needs to handle c("keep","omit","error").

#' Returns a list with:

#'   - na_action_eff: the canonical value ("keep", "omit", or "error")

#'   - na_action_raw: the original matched value (for warn-diagnostic checks)

#'

#' @param na_action character(1) already matched via match.arg()

#' @keywords internal

.hm_normalize_na_action <- function(na_action) {

  na_action_eff <- if (na_action %in% c("ignore", "warn")) "keep" else na_action

  list(na_action_raw = na_action, na_action_eff = na_action_eff)

}



#' HM-CS v2 input validation hook

#' @param data data.frame

#' @param col_map named list or NULL

#' @param required_keys character vector of required keys

#' @param fn function name (string)

#' @keywords internal

hm_validate_inputs <- function(data, col_map, required_keys, fn) {

  if (!is.data.frame(data)) {

    rlang::abort(sprintf("%s(): `data` must be a data.frame or tibble.", fn),

                 class = sprintf("healthmarkers_%s_error_data_type", fn))

  }



  # If no keys required (e.g., summarizers), allow NULL or empty col_map

  if (length(required_keys) == 0L) {

    if (!is.null(col_map) && !is.list(col_map)) {

      rlang::abort(sprintf("%s(): `col_map` must be a named list or NULL.", fn),

                   class = sprintf("healthmarkers_%s_error_colmap_type", fn))

    }

    return(invisible(TRUE))

  }



  # Validate col_map structure

  if (is.null(col_map) || !is.list(col_map) || is.null(names(col_map)) || any(names(col_map) == "")) {

    rlang::abort(sprintf("%s(): `col_map` must be a named list of required keys -> column names.", fn),

                 class = sprintf("healthmarkers_%s_error_colmap_type", fn))

  }



  # Ensure all required keys present and mapped to non-empty names

  have_names <- rlang::`%||%`(names(col_map), character(0))

  missing_keys <- setdiff(required_keys, have_names)

  if (length(missing_keys)) {

    rlang::abort(

      sprintf("%s(): `col_map` missing entries for: %s", fn, paste(missing_keys, collapse = ", ")),

      class = sprintf("healthmarkers_%s_error_missing_map", fn)

    )

  }

  mapped <- vapply(required_keys, function(k) {

    val <- col_map[[k]]

    if (is.null(val) || is.na(val)) "" else as.character(val)

  }, character(1))

  if (any(!nzchar(mapped))) {

    bad <- required_keys[!nzchar(mapped)]

    rlang::abort(

      sprintf("%s(): `col_map` has empty mapping for: %s", fn, paste(bad, collapse = ", ")),

      class = sprintf("healthmarkers_%s_error_missing_map", fn)

    )

  }



  invisible(TRUE)

}



# Utility: coerce selected columns to numeric, warn when NAs introduced, non-finite -> NA

hm_coerce_numeric <- function(data, cols, fn = "healthmarkers") {

  for (cn in intersect(cols, names(data))) {

    if (!is.numeric(data[[cn]])) {

      old <- data[[cn]]

      suppressWarnings(data[[cn]] <- as.numeric(old))

      introduced <- sum(is.na(data[[cn]]) & !is.na(old))

      if (introduced > 0) {

        rlang::warn(sprintf("%s(): column '%s' coerced to numeric; NAs introduced: %d", fn, cn, introduced))

      }

    }

    data[[cn]][!is.finite(data[[cn]])] <- NA_real_

  }

  data

}



#' Quiet name repair binder (if you need it elsewhere)

#' @importFrom vctrs vec_as_names

#' @keywords internal

hm_bind_cols_quiet <- function(...) {

  dplyr::bind_cols(...,

    .name_repair = ~ vctrs::vec_as_names(., repair = "unique", quiet = TRUE)

  )

}



#' Format a column-map resolution line for verbose output

#'

#' Produces a single string like:

#' "fn(): column map: G0 -> 'col_a', I0 -> 'col_b', ..."

#'

#' @param col_map Named list of key -> column-name mappings (only the keys

#'   actually used by the function need to be supplied).

#' @param fn Optional function name prefixed to the message.

#' @keywords internal

hm_fmt_col_map <- function(col_map, fn = NULL) {

  nms   <- names(col_map)

  parts <- vapply(nms, function(k) paste0(k, " -> '", col_map[[k]], "'"), character(1))

  msg   <- paste(parts, collapse = ", ")

  if (!is.null(fn)) msg <- paste0(fn, "(): column map: ", msg)

  msg

}



# ---------------------------------------------------------------------------

# Package-level constant: short internal key -> hm_col_report() dictionary key

# Used by both .hm_autofill_col_map() and .hm_build_col_map().

# Edit ONLY this one definition to add/rename mappings.

# ---------------------------------------------------------------------------

.hm_short_to_dict <- c(

  G0    = "fasting_glucose",  I0    = "fasting_insulin",

  G30   = "glucose_30m",      I30   = "insulin_30m",

  G120  = "glucose_120m",     I120  = "insulin_120m",

  TC    = "total_cholesterol",

  HDL_c = "HDL_c",            LDL_c = "LDL_c",

  TG    = "TG",

  total_chol = "total_cholesterol", sbp  = "sbp",

  smoker = "smoking",         bp_treated = "hypertension",

  diabetes = "diabetes",

  ApoA1 = "apoA1",            ApoB  = "apoB",

  ALT   = "ALT",              AST   = "AST",

  GGT   = "GGT",

  bilirubin  = "bilirubin",

  BMI   = "BMI",              bmi   = "BMI",

  waist = "waist",            weight = "weight",

  age   = "age",              sex   = "sex",

  eGFR  = "eGFR",             UACR  = "UACR",

  FFA   = "FFA",

  fat_mass     = "fat_mass",

  rate_palmitate = "rate_palmitate",

  rate_glycerol  = "rate_glycerol",

  creatinine   = "creatinine",

  albumin      = "albumin",   alb   = "albumin",

  calcium      = "calcium",   ca    = "calcium",

  platelets    = "platelets",

  height       = "height",

  ALM          = "ALM",       alm   = "ALM",

  bmd_t        = "BMD",       BMD   = "BMD",

  FEV1         = "FEV1",      FVC   = "FVC",

  fev1         = "FEV1",      fvc   = "FVC",

  FEV1pct      = "FEV1pct",

  kynurenine   = "kynurenine", tryptophan = "tryptophan",

  vitd         = "vitaminD",  vitaminD = "vitaminD",

  vitamin_d    = "vitaminD",

  # Inflammatory

  CRP          = "CRP",       IL6   = "IL6",        TNFa  = "TNFa",

  WBC          = "WBC",       neutrophils = "neutrophils",

  lymphocytes  = "lymphocytes", monocytes = "monocytes",

  eosinophils  = "eosinophils",

  ESR          = "ESR",

  # Renal

  BUN          = "BUN",       race  = "ethnicity",

  cystatin_C   = "cystatin_C", urea_serum = "urea_serum",

  creatinine_urine = "creatinine_urine", urea_urine = "urea_urine",

  # Vitamins / micronutrients

  VitD         = "vitaminD",  B12   = "vitaminB12",

  Folate       = "folate",    Ferritin = "ferritin",

  TSat         = "transferrin_sat",

  Cortisol     = "Cortisol",  DHEAS = "DHEAS",

  Testosterone = "testosterone", Estradiol = "estradiol",

  TSH          = "TSH",       free_T4 = "FT4",

  Retinol      = "Retinol",   Tocopherol = "Tocopherol",

  Total_lipids = "Total_lipids",

  VitC         = "VitC",      Homocysteine = "Homocysteine",

  MMA          = "MMA",       Magnesium = "magnesium",

  Zinc         = "zinc",      Copper = "copper",

  # Hormones

  total_testosterone = "testosterone",

  SHBG         = "SHBG",      LH    = "LH",         FSH   = "FSH",

  progesterone = "progesterone", free_T3 = "free_T3",

  aldosterone  = "aldosterone", renin = "renin",

  IGF1         = "IGF1",      prolactin = "prolactin",

  cortisol_0   = "cortisol_0",  cortisol_30 = "cortisol_30",

  insulin      = "fasting_insulin",

  # Sarcopenia

  strength     = "strength",  walking = "walking",

  chair        = "chair",     stairs  = "stairs",

  falls        = "falls",

  # BODE / pulmonary

  fev1_pct     = "FEV1pct",   sixmwd = "sixmwd",    mmrc  = "mmrc",

  fev1_pp      = "FEV1pct",

  # Saliva

  cort1        = "saliva_cort1", cort2 = "saliva_cort2",

  cort3        = "saliva_cort3", amylase = "saliva_amylase",

  glucose      = "saliva_glucose",

  # Endocrine / neuro

  nfl          = "nfl",

  GH           = "GH",        glucagon = "glucagon",

  PIVKA_II     = "PIVKA_II",

  # Metabolic risk features

  chol_total   = "total_cholesterol",

  chol_ldl     = "LDL_c",

  chol_hdl     = "HDL_c",

  triglycerides = "TG",

  age_year     = "age",

  z_HOMA       = "fasting_insulin",

  bp_sys_z     = "sbp",

  bp_dia_z     = "dbp",

  # Glycaemic / metabolic extras

  HbA1c        = "HbA1c",

  C_peptide    = "C_peptide",

  leptin       = "leptin",

  adiponectin  = "adiponectin"

)



# Auto-fill a col_map from hm_col_report() when the caller didn't supply one.

# Only fills keys present in the synonym dictionary; missing keys remain absent.

# Translates internal short keys (G0, I0, TG, ...) to their hm_col_report()

# dictionary key names, then back again so the returned list uses short keys.

# @param col_map  User-supplied col_map (may be NULL or missing).

# @param data     The data frame being analysed.

# @param keys     Character vector of short internal keys to fill.

# @param fn       Function name for informational message.

# @keywords internal

.hm_autofill_col_map <- function(col_map, data, keys, fn = "") {

  if (!is.null(col_map)) return(col_map)



  # Mapping: short internal key -> dictionary key used by hm_col_report()

  short_to_dict <- .hm_short_to_dict



  inferred <- tryCatch(

    hm_col_report(data, col_map = NULL, verbose = FALSE,

                  fuzzy = FALSE, show_unmatched = FALSE),

    error = function(e) list()

  )



  out <- list()

  for (k in keys) {

    dict_key <- if (!is.na(short_to_dict[k])) short_to_dict[[k]] else k

    if (!is.null(inferred[[dict_key]])) {

      out[[k]] <- inferred[[dict_key]]

    } else if (!is.null(inferred[[k]])) {

      # fallback: key already matches dictionary directly

      out[[k]] <- inferred[[k]]

    }

  }



  if (length(out) > 0L) {

    hm_inform(level = "debug",

              msg = sprintf("%s(): col_map not supplied -- auto-inferred %d/%d keys: %s",

                            fn, length(out), length(keys),

                            paste(names(out), unlist(out), sep = "->", collapse = ", ")))

  }

  out

}



#' Detect a participant/sample ID column in a data frame

#'

#' Checks column names against common patterns (case-insensitive exact match):

#' id, iid, participant_id, subject_id, sample_id, pid, sid, record_id.

#' Returns the first matching column name, or NULL if none found.

#' @keywords internal

.hm_detect_id_col <- function(data) {

  nms       <- names(data)

  nms_lower <- tolower(nms)

  candidates <- c("id", "iid", "participant_id", "participantid",

                  "subject_id",  "subjectid",  "sample_id",  "sampleid",

                  "pid",         "sid",         "record_id")

  for (cand in candidates) {

    m <- which(nms_lower == cand)

    if (length(m)) return(nms[m[1L]])

  }

  NULL

}



#' One-level pre-computation of missing keys from existing raw columns

#'

#' Given a dependency map (\code{deps}), for each key in \code{missing_keys}

#' checks whether its prerequisite columns are present in \code{data}. If yes,

#' the key is computed and added as a new column; otherwise an informational

#' message is emitted (when \code{verbose = TRUE}) indicating what to provide.

#'

#' @param data         data.frame to augment.

#' @param deps         Named list of dependency definitions, each with:

#'   \describe{

#'     \item{needs}{character vector of prerequisite column names}

#'     \item{describe}{short human-readable formula string}

#'     \item{compute}{function(data) returning a numeric vector}

#'   }

#' @param missing_keys character vector of key names to attempt to compute.

#' @param fn           calling function name (for messages).

#' @param verbose      logical.

#' @return A list with \code{$data} (augmented data.frame) and

#'   \code{$log} (character vector of per-key messages).

#' @keywords internal

.hm_precompute_from_deps <- function(data, deps, missing_keys,

                                     fn = "", verbose = FALSE) {

  if (length(missing_keys) == 0L) return(list(data = data, log = character(0)))

  log_msgs <- character(0)



  for (key in missing_keys) {

    dep <- deps[[key]]

    if (is.null(dep)) next



    prereqs_absent <- setdiff(dep$needs, names(data))



    if (length(prereqs_absent) == 0L) {

      computed <- tryCatch(dep$compute(data), error = function(e) NULL)

      if (!is.null(computed) && length(computed) == nrow(data)) {

        data[[key]] <- computed

        n_ok <- sum(!is.na(computed))

        msg  <- sprintf("  -> %s computed from %s [%s]: %d/%d valid",

                        key, paste(dep$needs, collapse = ", "),

                        dep$describe, n_ok, nrow(data))

        log_msgs <- c(log_msgs, msg)

        if (isTRUE(verbose)) {

          hm_inform(

            paste0(sprintf("%s(): pre-computation:\n", fn), msg),

            level = "inform"

          )

        }

      }

    } else {

      if (isTRUE(verbose)) {

        hm_inform(

          sprintf("%s(): pre-computation: %s cannot be derived -- provide: %s",

                  fn, key, paste(prereqs_absent, collapse = ", ")),

          level = "inform"

        )

      }

    }

  }



  list(data = data, log = log_msgs)

}



#' Build a complete col_map with dictionary inference and alias materialization

#'

#' Combines three previously separate steps into one call:

#' \enumerate{

#'   \item User-supplied entries in \code{col_map} are treated as authoritative

#'         Layer 0 seeds -- they are never overwritten.

#'   \item \code{hm_col_report()} fills any remaining unmatched keys via the

#'         full synonym dictionary (exact -> case-insensitive -> contains layers).

#'   \item Literal-name fallback: if a key in \code{keys} still has no match,

#'         but the key name exists verbatim as a column in \code{data}, it is

#'         used directly.

#'   \item Materialization: for every mapped key \code{k} where

#'         \code{col_map[[k]] != k} and \code{k} is \emph{not} already a column

#'         in \code{data}, a copy \code{data[[k]] <- data[[col_map[[k]]]]} is

#'         added. This ensures that downstream pre-computation helpers (which

#'         check \code{names(data)} for dependency names) and their

#'         \code{compute()} closures (which use literal key names) both resolve

#'         correctly even when the physical column has a non-standard name.

#' }

#'

#' @param data    The data.frame being analysed (modified locally; original

#'   is not altered outside the function call).

#' @param col_map Named list of user-supplied key -> column-name mappings, or

#'   \code{NULL} for full auto-inference.

#' @param keys    Character vector of short internal keys to resolve.

#' @param fn      Calling function name string (for messages).

#'

#' @return A list with four elements:

#' \describe{

#'   \item{data}{The (possibly augmented) data.frame.}

#'   \item{col_map}{Complete named list of key -> resolved column name.}

#'   \item{user_keys}{Keys that were explicitly provided in the original

#'     \code{col_map} argument.}

#'   \item{inferred_keys}{Keys resolved by dictionary inference or literal

#'     fallback (i.e., not in \code{user_keys}).}

#' }

#' @keywords internal

.hm_build_col_map <- function(data, col_map, keys, fn = "") {



  user_keys <- if (is.list(col_map) && length(col_map) > 0L)

    intersect(names(col_map), keys) else character(0)



  # Step 1+2: run hm_col_report with user seeds as Layer 0 ------------------

  inferred_full <- tryCatch(

    hm_col_report(data, col_map = col_map, verbose = FALSE,

                  fuzzy = FALSE, show_unmatched = FALSE),

    error = function(e) list()

  )



  # short_to_dict translation table ----------------------------------------

  short_to_dict <- .hm_short_to_dict



  # Step 2: assemble resolved col_map from inferred_full + literal fallback --

  out_map <- if (is.list(col_map) && length(col_map) > 0L) col_map else list()



  for (k in keys) {

    if (!is.null(out_map[[k]])) next  # user-supplied: keep

    dict_key <- if (!is.na(short_to_dict[k])) short_to_dict[[k]] else k

    if (!is.null(inferred_full[[dict_key]])) {

      out_map[[k]] <- inferred_full[[dict_key]]

    } else if (!is.null(inferred_full[[k]])) {

      out_map[[k]] <- inferred_full[[k]]

    } else if (k %in% names(data)) {

      # Step 3: literal fallback

      out_map[[k]] <- k

    }

  }



  inferred_keys <- setdiff(names(out_map)[names(out_map) %in% keys], user_keys)



  # Step 4: materialize aliases (col_map[[k]] != k) into data ----------------

  for (k in names(out_map)) {

    mapped_col <- out_map[[k]]

    if (!is.null(mapped_col) && mapped_col != k &&

        mapped_col %in% names(data) && !k %in% names(data)) {

      data[[k]] <- data[[mapped_col]]

    }

  }



  list(

    data          = data,

    col_map       = out_map,

    user_keys     = user_keys,

    inferred_keys = inferred_keys

  )

}



#' Summarise a result tibble: count non-NA rows per output column

#'

#' Produces a single string like:

#' "fn(): results: col_a 28/30, col_b 30/30, col_c 25/30"

#'

#' @param result A tibble / data.frame returned by a HealthMarkers function.

#' @param fn Optional function name prefixed to the message.

#' @keywords internal

hm_result_summary <- function(result, fn = NULL) {

  n     <- nrow(result)

  parts <- vapply(names(result), function(cn) {

    ok <- sum(!is.na(result[[cn]]))

    paste0(cn, " ", ok, "/", n)

  }, character(1))

  msg <- paste(parts, collapse = ", ")

  if (!is.null(fn)) msg <- paste0(fn, "(): results: ", msg)

  msg

}



# ---------------------------------------------------------------------------

# Global pre-computation -- Tier 0 of the DAG pipeline

# ---------------------------------------------------------------------------

#' Derive "Tier 0" variables before any marker function runs.

#'

#' Computes the following variables **only when they are absent** from `data`:

#' \describe{

#'   \item{BMI}{from `weight` (kg) and `height` (m or cm) via `weight / height_m^2`}

#'   \item{glucose / G0}{bidirectional alias: whichever is absent is filled from the other}

#'   \item{insulin / I0}{bidirectional alias (pmol/L <-> muU/mL unchanged -- same unit assumed)}

#'   \item{eGFR}{CKD-EPI 2009 creatinine equation from `creatinine`, `age`, `sex`

#'               (and optionally `race`). Written as `eGFR` only; downstream

#'               functions that need `eGFR_cr` receive it via col_map.}

#'   \item{UACR}{`urine_albumin` / `urine_creatinine` (both in mg/mmol or the

#'               ratio already in mg/g -- no unit conversion; caller's responsibility)}

#'   \item{LDL_c}{Friedewald: `TC - HDL_c - TG/2.2` (mmol/L); skipped if TG > 4.5}

#' }

#'

#' col_map keys are respected: if the user has mapped e.g.\ `creatinine -> "Cr_serum"`,

#' the materialized `data[["creatinine"]]` column (placed there by `.hm_build_col_map()`)

#' is used transparently.

#'

#' @param data    data.frame after col_map materialization.

#' @param col_map Named list of key -> column mappings (used only to resolve

#'   the physical column names when materialization has NOT been run yet).

#' @param verbose Logical; emit one-line summary per derived variable.

#'

#' @return A list:

#' \describe{

#'   \item{data}{data.frame with new columns appended.}

#'   \item{precomputed}{Character vector of variable names that were derived.}

#' }

#' @keywords internal

.hm_global_precompute <- function(data, col_map = NULL, verbose = FALSE) {

  precomputed <- character(0)

  fn <- "global_precompute"



  # Helper: resolve a key to actual column name in data (col_map then literal)

  .resolve <- function(key) {

    if (!is.null(col_map[[key]]) && col_map[[key]] %in% names(data))

      return(col_map[[key]])

    if (key %in% names(data)) return(key)

    NULL

  }

  .get <- function(key) {

    col <- .resolve(key)

    if (is.null(col)) return(NULL)

    as.numeric(data[[col]])

  }

  .absent <- function(key) is.null(.resolve(key))

  .log <- function(nm, src) {

    precomputed <<- c(precomputed, nm)

    if (isTRUE(verbose))

      hm_inform(sprintf("%s(): derived '%s' from %s", fn, nm,

                        paste(src, collapse = " + ")), level = "inform")

  }



  # --- 1. BMI -----------------------------------------------------------

  if (.absent("BMI")) {

    w <- .get("weight"); h <- .get("height")

    if (!is.null(w) && !is.null(h)) {

      h_m <- ifelse(is.finite(h) & h > 3, h / 100, h)

      bmi  <- w / (h_m ^ 2)

      data[["BMI"]] <- ifelse(is.finite(bmi) & bmi > 5 & bmi < 150, bmi, NA_real_)

      .log("BMI", c("weight", "height"))

    }

  }



  # --- 2. glucose <-> G0 alias ------------------------------------------

  if (.absent("glucose") && !.absent("G0")) {

    data[["glucose"]] <- .get("G0")

    .log("glucose", "G0")

  } else if (.absent("G0") && !.absent("glucose")) {

    data[["G0"]] <- .get("glucose")

    .log("G0", "glucose")

  }



  # --- 3. insulin <-> I0 alias ------------------------------------------

  if (.absent("insulin") && !.absent("I0")) {

    data[["insulin"]] <- .get("I0")

    .log("insulin", "I0")

  } else if (.absent("I0") && !.absent("insulin")) {

    data[["I0"]] <- .get("insulin")

    .log("I0", "insulin")

  }



  # --- 4. eGFR (CKD-EPI 2009 creatinine) --------------------------------

  if (.absent("eGFR")) {

    Cr  <- .get("creatinine")

    age <- .get("age")

    sex_col <- .resolve("sex")

    sex_raw <- if (!is.null(sex_col)) data[[sex_col]] else NULL

    if (!is.null(Cr) && !is.null(age) && !is.null(sex_raw)) {

      s_up <- toupper(as.character(sex_raw))

      sexi <- ifelse(startsWith(s_up, "M") | sex_raw %in% c(1, "1"), 1L, 0L)

      race_col <- .resolve("race")

      race_raw <- if (!is.null(race_col)) toupper(as.character(data[[race_col]])) else "OTHER"

      factor_race <- ifelse(grepl("BLACK|NHB|AFRIC", race_raw), 1.159, 1)

      kappa <- ifelse(sexi == 1L, 0.9, 0.7)

      alpha <- ifelse(sexi == 1L, -0.411, -0.329)

      # Female sex multiplier (CKD-EPI 2009)

      factor_sex  <- ifelse(sexi == 0L, 1.018, 1)

      ratio <- Cr / kappa

      min_r <- pmin(ratio, 1); max_r <- pmax(ratio, 1)

      egfr  <- 141 * (min_r ^ alpha) * (max_r ^ -1.209) *

        (0.993 ^ age) * factor_race * factor_sex

      egfr[!is.finite(egfr)] <- NA_real_

      data[["eGFR"]] <- egfr

      .log("eGFR", c("creatinine", "age", "sex"))

    }

  }



  # --- 5. UACR ----------------------------------------------------------

  if (.absent("UACR")) {

    ualb <- .get("urine_albumin")

    ucr  <- .get("urine_creatinine")

    if (!is.null(ualb) && !is.null(ucr)) {

      ratio <- ualb / ucr

      ratio[!is.finite(ratio)] <- NA_real_

      data[["UACR"]] <- ratio

      .log("UACR", c("urine_albumin", "urine_creatinine"))

    }

  }



  # --- 6. LDL_c (Friedewald, mmol/L) -----------------------------------

  if (.absent("LDL_c")) {

    tc    <- .get("TC")

    hdl   <- .get("HDL_c")

    tg    <- .get("TG")

    if (!is.null(tc) && !is.null(hdl) && !is.null(tg)) {

      ldl <- tc - hdl - tg / 2.2

      # Friedewald invalid when TG > 4.5 mmol/L

      ldl[is.finite(tg) & tg > 4.5] <- NA_real_

      ldl[!is.finite(ldl)] <- NA_real_

      data[["LDL_c"]] <- ldl

      .log("LDL_c", c("TC", "HDL_c", "TG"))

    }

  }



  # --- 7. WHR (waist-to-hip ratio) ------------------------------------

  if (.absent("whr") && .absent("WHR")) {

    wst <- .get("waist"); hip <- .get("hip")

    if (!is.null(wst) && !is.null(hip)) {

      r <- wst / hip

      r[!is.finite(r) | r <= 0] <- NA_real_

      data[["whr"]] <- r

      .log("whr", c("waist", "hip"))

    }

  }



  # --- 8. MAP (mean arterial pressure) ---------------------------------

  if (.absent("MAP")) {

    sbp <- .get("sbp"); dbp <- .get("dbp")

    if (!is.null(sbp) && !is.null(dbp)) {

      map <- (sbp + 2 * dbp) / 3

      map[!is.finite(map) | map <= 0] <- NA_real_

      data[["MAP"]] <- map

      .log("MAP", c("sbp", "dbp"))

    }

  }



  # --- 9. BUN-to-creatinine ratio -------------------------------------

  if (.absent("BUN_Cr_ratio")) {

    bun <- .get("BUN"); cr <- .get("creatinine")

    if (!is.null(bun) && !is.null(cr)) {

      r <- bun / cr

      r[!is.finite(r)] <- NA_real_

      data[["BUN_Cr_ratio"]] <- r

      .log("BUN_Cr_ratio", c("BUN", "creatinine"))

    }

  }



  # --- 10. VLDL (TG / 2.2, mmol/L) ------------------------------------

  if (.absent("VLDL")) {

    tg <- .get("TG")

    if (!is.null(tg)) {

      data[["VLDL"]] <- tg / 2.2

      .log("VLDL", "TG")

    }

  }



  # --- 11. non_HDL cholesterol ----------------------------------------

  if (.absent("non_HDL")) {

    tc <- .get("TC"); hdl <- .get("HDL_c")

    if (!is.null(tc) && !is.null(hdl)) {

      r <- tc - hdl

      r[!is.finite(r)] <- NA_real_

      data[["non_HDL"]] <- r

      .log("non_HDL", c("TC", "HDL_c"))

    }

  }



  # --- 12. remnant_c ---------------------------------------------------

  if (.absent("remnant_c")) {

    tc <- .get("TC"); hdl <- .get("HDL_c"); ldl <- .get("LDL_c")

    if (!is.null(tc) && !is.null(hdl) && !is.null(ldl)) {

      r <- tc - hdl - ldl

      r[!is.finite(r)] <- NA_real_

      data[["remnant_c"]] <- r

      .log("remnant_c", c("TC", "HDL_c", "LDL_c"))

    }

  }



  # --- 13. Lipid ratios ------------------------------------------------

  if (.absent("AIP")) {

    tg <- .get("TG"); hdl <- .get("HDL_c")

    if (!is.null(tg) && !is.null(hdl)) {

      r <- tg / hdl

      aip <- ifelse(is.finite(r) & r > 0, log10(r), NA_real_)

      data[["AIP"]] <- aip

      .log("AIP", c("TG", "HDL_c"))

    }

  }

  if (.absent("CRI_I")) {

    tc <- .get("TC"); hdl <- .get("HDL_c")

    if (!is.null(tc) && !is.null(hdl)) {

      r <- tc / hdl; r[!is.finite(r)] <- NA_real_

      data[["CRI_I"]] <- r

      .log("CRI_I", c("TC", "HDL_c"))

    }

  }

  if (.absent("CRI_II")) {

    ldl <- .get("LDL_c"); hdl <- .get("HDL_c")

    if (!is.null(ldl) && !is.null(hdl)) {

      r <- ldl / hdl; r[!is.finite(r)] <- NA_real_

      data[["CRI_II"]] <- r

      .log("CRI_II", c("LDL_c", "HDL_c"))

    }

  }

  if (.absent("HDL_TG_ratio")) {

    hdl <- .get("HDL_c"); tg <- .get("TG")

    if (!is.null(hdl) && !is.null(tg)) {

      r <- hdl / tg; r[!is.finite(r)] <- NA_real_

      data[["HDL_TG_ratio"]] <- r

      .log("HDL_TG_ratio", c("HDL_c", "TG"))

    }

  }

  if (.absent("LDL_HDL_ratio")) {

    ldl <- .get("LDL_c"); hdl <- .get("HDL_c")

    if (!is.null(ldl) && !is.null(hdl)) {

      r <- ldl / hdl; r[!is.finite(r)] <- NA_real_

      data[["LDL_HDL_ratio"]] <- r

      .log("LDL_HDL_ratio", c("LDL_c", "HDL_c"))

    }

  }



  # --- 14. Inflammatory cell ratios ------------------------------------

  if (.absent("NLR")) {

    n <- .get("neutrophils"); l <- .get("lymphocytes")

    if (!is.null(n) && !is.null(l)) {

      r <- n / l; r[!is.finite(r)] <- NA_real_

      data[["NLR"]] <- r

      .log("NLR", c("neutrophils", "lymphocytes"))

    }

  }

  if (.absent("dNLR")) {

    n <- .get("neutrophils"); w <- .get("WBC")

    if (!is.null(n) && !is.null(w)) {

      denom <- w - n

      r <- n / denom

      r[!is.finite(r) | r < 0] <- NA_real_

      data[["dNLR"]] <- r

      .log("dNLR", c("neutrophils", "WBC"))

    }

  }

  if (.absent("PLR")) {

    p <- .get("platelets"); l <- .get("lymphocytes")

    if (!is.null(p) && !is.null(l)) {

      r <- p / l; r[!is.finite(r)] <- NA_real_

      data[["PLR"]] <- r

      .log("PLR", c("platelets", "lymphocytes"))

    }

  }

  if (.absent("SII")) {

    n <- .get("neutrophils"); p <- .get("platelets"); l <- .get("lymphocytes")

    if (!is.null(n) && !is.null(p) && !is.null(l)) {

      r <- n * p / l; r[!is.finite(r)] <- NA_real_

      data[["SII"]] <- r

      .log("SII", c("neutrophils", "platelets", "lymphocytes"))

    }

  }

  if (.absent("SIRI")) {

    n <- .get("neutrophils"); m <- .get("monocytes"); l <- .get("lymphocytes")

    if (!is.null(n) && !is.null(m) && !is.null(l)) {

      r <- n * m / l; r[!is.finite(r)] <- NA_real_

      data[["SIRI"]] <- r

      .log("SIRI", c("neutrophils", "monocytes", "lymphocytes"))

    }

  }



  # --- 15. Electrolyte ratios ------------------------------------------

  if (.absent("sodium_potassium_ratio")) {

    na_v <- .get("sodium"); k_v <- .get("potassium")

    if (!is.null(na_v) && !is.null(k_v)) {

      r <- na_v / k_v; r[!is.finite(r)] <- NA_real_

      data[["sodium_potassium_ratio"]] <- r

      .log("sodium_potassium_ratio", c("sodium", "potassium"))

    }

  }

  if (.absent("Mg_Zn_den")) {

    mg <- .get("magnesium"); zn <- .get("zinc")

    if (!is.null(mg) && !is.null(zn)) {

      r <- mg / zn; r[!is.finite(r)] <- NA_real_

      data[["Mg_Zn_den"]] <- r

      .log("Mg_Zn_den", c("magnesium", "zinc"))

    }

  }

  if (.absent("Cu_Zn_den")) {

    cu <- .get("copper"); zn <- .get("zinc")

    if (!is.null(cu) && !is.null(zn)) {

      r <- cu / zn; r[!is.finite(r)] <- NA_real_

      data[["Cu_Zn_den"]] <- r

      .log("Cu_Zn_den", c("copper", "zinc"))

    }

  }



  # --- 16. Hormone ratios ----------------------------------------------

  if (.absent("Cort_DHEAS_den")) {

    cort <- .get("Cortisol"); dheas <- .get("DHEAS")

    if (!is.null(cort) && !is.null(dheas)) {

      r <- cort / dheas; r[!is.finite(r)] <- NA_real_

      data[["Cort_DHEAS_den"]] <- r

      .log("Cort_DHEAS_den", c("Cortisol", "DHEAS"))

    }

  }

  if (.absent("T_E2_den")) {

    t_v <- .get("testosterone"); e2 <- .get("estradiol")

    if (!is.null(t_v) && !is.null(e2)) {

      r <- t_v / e2; r[!is.finite(r)] <- NA_real_

      data[["T_E2_den"]] <- r

      .log("T_E2_den", c("testosterone", "estradiol"))

    }

  }

  if (.absent("TSH_fT4_den")) {

    tsh <- .get("TSH"); ft4 <- .get("FT4")

    if (!is.null(tsh) && !is.null(ft4)) {

      r <- tsh / ft4; r[!is.finite(r)] <- NA_real_

      data[["TSH_fT4_den"]] <- r

      .log("TSH_fT4_den", c("TSH", "FT4"))

    }

  }



  # --- 17. Amino-acid ratio --------------------------------------------

  if (.absent("Tyr_Phe_Ratio")) {

    tyr <- .get("Tyr"); phe <- .get("Phe")

    if (!is.null(tyr) && !is.null(phe)) {

      r <- tyr / phe; r[!is.finite(r)] <- NA_real_

      data[["Tyr_Phe_Ratio"]] <- r

      .log("Tyr_Phe_Ratio", c("Tyr", "Phe"))

    }

  }



  # --- 18. Urinary biomarker-to-creatinine ratios ----------------------

  ucr <- .get("u_creatinine")

  if (!is.null(ucr)) {

    for (pair in list(

      c("KIM1_gCr",  "KIM1"),

      c("NGAL_gCr",  "NGAL"),

      c("NAG_gCr",   "NAG"),

      c("L_FABP_gCr","L_FABP"),

      c("IL18_gCr",  "IL18")

    )) {

      nm <- pair[1]; src <- pair[2]

      if (.absent(nm)) {

        val <- .get(src)

        if (!is.null(val)) {

          r <- val / ucr; r[!is.finite(r)] <- NA_real_

          data[[nm]] <- r

          .log(nm, c(src, "u_creatinine"))

        }

      }

    }

  }



  derived_map <- setNames(as.list(precomputed), precomputed)

  list(data = data, precomputed = precomputed, derived_map = derived_map)

}

## Tests for .hm_build_col_map() internal helper
## Relies on HealthMarkers::: to access unexported functions.

test_that(".hm_build_col_map: NULL col_map infers all resolvable keys", {
  skip_on_cran()
  df <- tibble::tibble(hdlc = 1, trig = 1.3, bmi = 24)
  res <- HealthMarkers:::.hm_build_col_map(df, NULL, c("HDL_c", "TG", "BMI", "glucose"), fn = "test")
  # HDL_c, TG, BMI should be inferred from synonym dictionary
  expect_equal(res$col_map[["HDL_c"]], "hdlc")
  expect_equal(res$col_map[["TG"]],    "trig")
  expect_equal(res$col_map[["BMI"]],   "bmi")
  # glucose not in data -> not resolved
  expect_null(res$col_map[["glucose"]])
  # all three resolved keys are inferred (none user-supplied)
  expect_setequal(res$inferred_keys, c("HDL_c", "TG", "BMI"))
  expect_length(res$user_keys, 0L)
})

test_that(".hm_build_col_map: partial col_map — user entries preserved, rest inferred", {
  skip_on_cran()
  df  <- tibble::tibble(pglu0 = 5.5, hdlc = 1, trig = 1.3, bmi = 24)
  cm  <- list(G0 = "pglu0")  # user supplies only G0
  res <- HealthMarkers:::.hm_build_col_map(df, cm, c("G0", "HDL_c", "TG", "BMI"), fn = "test")
  # user entry preserved
  expect_equal(res$col_map[["G0"]], "pglu0")
  # remaining keys inferred from dictionary
  expect_equal(res$col_map[["HDL_c"]], "hdlc")
  expect_equal(res$col_map[["TG"]],    "trig")
  expect_equal(res$col_map[["BMI"]],   "bmi")
  # split is correct
  expect_equal(res$user_keys, "G0")
  expect_setequal(res$inferred_keys, c("HDL_c", "TG", "BMI"))
})

test_that(".hm_build_col_map: full col_map — nothing inferred, nothing overwritten", {
  skip_on_cran()
  df  <- tibble::tibble(pglu0 = 5.5, hdlc = 1, trig = 1.3, bmi = 24)
  cm  <- list(G0 = "pglu0", HDL_c = "hdlc", TG = "trig", BMI = "bmi")
  res <- HealthMarkers:::.hm_build_col_map(df, cm, c("G0", "HDL_c", "TG", "BMI"), fn = "test")
  expect_equal(res$col_map[["G0"]],    "pglu0")
  expect_equal(res$col_map[["HDL_c"]], "hdlc")
  expect_equal(res$col_map[["TG"]],    "trig")
  expect_equal(res$col_map[["BMI"]],   "bmi")
  expect_setequal(res$user_keys, c("G0", "HDL_c", "TG", "BMI"))
  expect_length(res$inferred_keys, 0L)
})

test_that(".hm_build_col_map: alias materialized into data", {
  skip_on_cran()
  # col_map maps G0 -> "pglu0"; after call data[["G0"]] should exist
  df  <- tibble::tibble(pglu0 = c(5.5, 6.0), hdlc = c(1, 1.2))
  cm  <- list(G0 = "pglu0")
  res <- HealthMarkers:::.hm_build_col_map(df, cm, c("G0", "HDL_c"), fn = "test")
  expect_true("G0" %in% names(res$data))
  expect_equal(res$data[["G0"]], c(5.5, 6.0))
  # original column still present
  expect_true("pglu0" %in% names(res$data))
})

test_that(".hm_build_col_map: non-clobbering — existing logical key not overwritten", {
  skip_on_cran()
  # data already has a column literally named "G0" with different values
  df  <- tibble::tibble(G0 = c(4.0, 5.0), pglu0 = c(9.9, 9.9))
  cm  <- list(G0 = "pglu0")
  res <- HealthMarkers:::.hm_build_col_map(df, cm, c("G0"), fn = "test")
  # G0 already existed -> should NOT be overwritten with pglu0 values
  expect_equal(res$data[["G0"]], c(4.0, 5.0))
})

test_that(".hm_build_col_map: literal-name fallback for unrecognised column names", {
  skip_on_cran()
  # "my_custom_col" is not in the dictionary, but the key is literally named the same
  df  <- tibble::tibble(my_custom_col = 1:3)
  res <- HealthMarkers:::.hm_build_col_map(df, NULL, c("my_custom_col"), fn = "test")
  expect_equal(res$col_map[["my_custom_col"]], "my_custom_col")
  expect_equal(res$inferred_keys, "my_custom_col")
})

test_that(".hm_build_col_map: synonym matching works end-to-end", {
  skip_on_cran()
  df  <- tibble::tibble(insu0 = c(60, 80), pglu0 = c(5.5, 6.0))
  res <- HealthMarkers:::.hm_build_col_map(df, NULL, c("I0", "G0"), fn = "test")
  expect_equal(res$col_map[["I0"]], "insu0")
  expect_equal(res$col_map[["G0"]], "pglu0")
})

## Tests for .hm_global_precompute() -----------------------------------------

test_that(".hm_global_precompute: BMI derived from weight (kg) and height (cm)", {
  skip_on_cran()
  df  <- data.frame(weight = 70.0, height = 170.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("BMI" %in% names(res$data))
  expect_equal(round(res$data[["BMI"]], 2), round(70 / 1.70^2, 2))
  expect_true("BMI" %in% res$precomputed)
})

test_that(".hm_global_precompute: G0 aliased to glucose when glucose absent", {
  skip_on_cran()
  df  <- data.frame(G0 = 5.5)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("glucose" %in% names(res$data))
  expect_equal(res$data[["glucose"]], 5.5)
  expect_true("glucose" %in% res$precomputed)
})

test_that(".hm_global_precompute: glucose aliased to G0 when G0 absent", {
  skip_on_cran()
  df  <- data.frame(glucose = 6.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("G0" %in% names(res$data))
  expect_equal(res$data[["G0"]], 6.0)
  expect_true("G0" %in% res$precomputed)
})

test_that(".hm_global_precompute: I0 aliased to insulin when insulin absent", {
  skip_on_cran()
  df  <- data.frame(I0 = 12.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("insulin" %in% names(res$data))
  expect_equal(res$data[["insulin"]], 12.0)
  expect_true("insulin" %in% res$precomputed)
})

test_that(".hm_global_precompute: eGFR derived from creatinine, age, sex", {
  skip_on_cran()
  df  <- data.frame(creatinine = 1.0, age = 40, sex = "M")
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("eGFR" %in% names(res$data))
  expect_true(is.finite(res$data[["eGFR"]]))
  expect_true(res$data[["eGFR"]] > 60 && res$data[["eGFR"]] < 130)
  expect_true("eGFR" %in% res$precomputed)
})

test_that(".hm_global_precompute: UACR derived from urine_albumin / urine_creatinine", {
  skip_on_cran()
  df  <- data.frame(urine_albumin = 30.0, urine_creatinine = 150.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("UACR" %in% names(res$data))
  expect_equal(res$data[["UACR"]], 30.0 / 150.0)
  expect_true("UACR" %in% res$precomputed)
})

test_that(".hm_global_precompute: LDL_c derived via Friedewald (mmol/L)", {
  skip_on_cran()
  df  <- data.frame(TC = 5.0, HDL_c = 1.2, TG = 1.5)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("LDL_c" %in% names(res$data))
  expect_equal(round(res$data[["LDL_c"]], 4), round(5.0 - 1.2 - 1.5 / 2.2, 4))
  expect_true("LDL_c" %in% res$precomputed)
})

test_that(".hm_global_precompute: LDL_c is NA when TG > 4.5 mmol/L", {
  skip_on_cran()
  df  <- data.frame(TC = 5.0, HDL_c = 1.2, TG = 5.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_true("LDL_c" %in% names(res$data))
  expect_true(is.na(res$data[["LDL_c"]]))
})

test_that(".hm_global_precompute: existing columns are not overwritten", {
  skip_on_cran()
  df  <- data.frame(BMI = 30.0, weight = 50.0, height = 170.0)
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = NULL, verbose = FALSE)
  expect_equal(res$data[["BMI"]], 30.0)
  expect_false("BMI" %in% res$precomputed)
})

test_that(".hm_global_precompute: col_map keys are honoured for column resolution", {
  skip_on_cran()
  df  <- data.frame(cr = 0.9, age = 50, sex = "F")
  cm  <- list(creatinine = "cr")
  res <- HealthMarkers:::.hm_global_precompute(df, col_map = cm, verbose = FALSE)
  expect_true("eGFR" %in% names(res$data))
  expect_true(is.finite(res$data[["eGFR"]]))
})

library(testthat)

cm <- list(vitamin_d = "VitD")

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(VitD = c(10, 20))
  # empty list with inferrable data – succeeds via inference
  expect_no_error(vitamin_d_status(df, list()))
  expect_error(vitamin_d_status(df, list(vitamin_d = "")), class = "healthmarkers_vitd_error_bad_map_values")
  expect_error(vitamin_d_status(data.frame(X=1), list(vitamin_d = "VitD")),
               class = "healthmarkers_vitd_error_missing_columns")
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  df <- data.frame(VitD = c(12, 35))
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(vitamin_d_status(df, cm, verbose = TRUE), "vitamin_d_status")
  expect_message(vitamin_d_status(df, cm, verbose = TRUE), "col_map")
  expect_message(vitamin_d_status(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- data.frame(VitD = c(12, 35))
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(vitamin_d_status(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("coercion warning when non-numeric introduces NAs", {
  skip_on_cran()
  df <- data.frame(VitD = c("12", "oops", "35"))
  expect_warning(vitamin_d_status(df, cm), class = "healthmarkers_vitd_warn_na_coercion")
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(VitD = c(12, NA, 35))
  out_keep <- vitamin_d_status(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(is.na(out_keep$vitamin_d_status[2]))

  out_omit <- vitamin_d_status(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 2L)
  expect_false(any(is.na(out_omit$vitamin_d_status)))

  expect_error(vitamin_d_status(df, cm, na_action = "error"),
               class = "healthmarkers_vitd_error_missing_values")

  expect_warning(vitamin_d_status(df, cm, na_action = "warn"),
                 class = "healthmarkers_vitd_warn_missing_inputs")
})

test_that("domain warnings: negative values and suspicious units", {
  skip_on_cran()
  df_neg <- data.frame(VitD = c(-5, 15))
  expect_warning(vitamin_d_status(df_neg, cm),
                 class = "healthmarkers_vitd_warn_negative_values")

  # High median suggests nmol/L units
  df_units <- data.frame(VitD = c(175, 200, 225))
  expect_warning(vitamin_d_status(df_units, cm),
                 class = "healthmarkers_vitd_warn_units_suspicious")
})

test_that("extreme inputs pass through without error (check_extreme removed)", {
  skip_on_cran()
  df <- data.frame(VitD = c(-10, 15, 500, 35))
  out <- suppressWarnings(vitamin_d_status(df, cm))
  expect_s3_class(out, "tbl_df")
  expect_true("vitamin_d_status" %in% names(out))
})

test_that("classification boundaries are correct and ordered factor returned", {
  skip_on_cran()
  df <- data.frame(VitD = c(19.9, 20.0, 29.9, 30.0, NA))
  out <- vitamin_d_status(df, cm)
  expect_identical(levels(out$vitamin_d_status),
                   c("Deficient","Insufficient","Sufficient"))
  expect_true(is.ordered(out$vitamin_d_status))
  expect_equal(as.character(out$vitamin_d_status[1]), "Deficient")
  expect_equal(as.character(out$vitamin_d_status[2]), "Insufficient")
  expect_equal(as.character(out$vitamin_d_status[3]), "Insufficient")
  expect_equal(as.character(out$vitamin_d_status[4]), "Sufficient")
  expect_true(is.na(out$vitamin_d_status[5]))
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(VitD = c(12, NA, 35))
  out_keep <- vitamin_d_status(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(vitamin_d_status(df, cm, na_action = "warn"))
  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

library(testthat)
library(tibble)

test_that("errors if missing any of HDL_c, TG, or BMI", {
  skip_on_cran()
  df1 <- tibble(TG = 1, BMI = 24)
  expect_error(
    glycemic_markers(df1, verbose = FALSE),
    "missing required columns: HDL_c"
  )
  df2 <- tibble(HDL_c = 1, BMI = 24)
  expect_error(
    glycemic_markers(df2, verbose = FALSE),
    "missing required columns: TG"
  )
  df3 <- tibble(HDL_c = 1, TG = 1.3)
  expect_error(
    glycemic_markers(df3, verbose = FALSE),
    "missing required columns: BMI"
  )
})

test_that("SPISE is computed correctly", {
  skip_on_cran()
  df <- tibble(HDL_c = 1.0, TG = 1.3, BMI = 24)
  expected <- 600 * 1.0^0.185 / (1.3^0.2 * 24^1.338)
  out <- glycemic_markers(df, verbose = FALSE)
  expect_equal(out$SPISE, expected, tolerance = 1e-8)
})

test_that("METS_IR returns finite when glucose present and HDL_c != 1, NA when denominator is zero", {
  skip_on_cran()
  # Use HDL_c != 1 so log(HDL_c) != 0 and result is finite
  df_with <- tibble(HDL_c = 1.2, TG = 1.3, BMI = 24, glucose = 5.6)
  out1 <- glycemic_markers(df_with, verbose = FALSE)
  expected <- (log(2 * 5.6 + 1.3) * 24) / log(1.2)
  expect_equal(out1$METS_IR, expected, tolerance = 1e-8)

  # Denominator zero case -> safe division returns NA
  df_den0 <- tibble(HDL_c = 1, TG = 1.3, BMI = 24, glucose = 5.6)
  out_den0 <- glycemic_markers(df_den0, verbose = FALSE)
  expect_true(is.na(out_den0$METS_IR))

  # No glucose column -> NA
  df_without <- tibble(HDL_c = 1, TG = 1.3, BMI = 24)
  out2 <- glycemic_markers(df_without, verbose = FALSE)
  expect_true(is.na(out2$METS_IR))
})

test_that("prediabetes and diabetes flags handle HbA1c correctly and propagate NA", {
  skip_on_cran()
  df <- tibble(
    HDL_c = 1, TG = 1.3, BMI = 24,
    HbA1c = c(40, 44, 50)
  )
  out <- glycemic_markers(df, verbose = FALSE)
  expect_equal(out$prediabetes, c(0L, 1L, 1L))
  expect_equal(out$diabetes, c(0L, 0L, 1L))

  df2 <- tibble(HDL_c = 1, TG = 1.3, BMI = 24)
  out2 <- glycemic_markers(df2, verbose = FALSE)
  expect_true(all(is.na(out2$prediabetes)))
  expect_true(all(is.na(out2$diabetes)))
})

test_that("HOMA_CP computed when C_peptide & G0 present, NA otherwise", {
  skip_on_cran()
  df_ok <- tibble(
    HDL_c = 1, TG = 1.3, BMI = 24,
    C_peptide = 300, G0 = 5.5
  )
  expected <- (5.5 * (300 / 6)) / 22.5
  out_ok <- glycemic_markers(df_ok, verbose = FALSE)
  expect_equal(out_ok$HOMA_CP, expected, tolerance = 1e-8)

  df_bad <- tibble(HDL_c = 1, TG = 1.3, BMI = 24, C_peptide = 300)
  out_bad <- glycemic_markers(df_bad, verbose = FALSE)
  expect_true(is.na(out_bad$HOMA_CP))
})

test_that("LAR, ASI, and TyG_index compute correctly when inputs present", {
  skip_on_cran()
  # set up a row where everything is present
  df <- tibble(
    HDL_c = 1, TG = 1.3, BMI = 24,
    leptin = 10, adiponectin = 5,
    I0 = 50,
    glucose = 5.6
  )
  out <- glycemic_markers(df, verbose = FALSE)
  # Leptin/Adiponectin Ratio
  expect_equal(out$LAR, 10 / 5)
  # Adiponectin Sensitivity Index = adiponectin / I0
  expect_equal(out$ASI, 5 / 50)
  # TyG_index: ln((TG*88.57) * (glucose*18) / 2)
  TG_mgdl <- 1.3 * 88.57
  Glu_mgdl <- 5.6 * 18
  expect_equal(out$TyG_index, log(TG_mgdl * Glu_mgdl / 2), tolerance = 1e-8)
})

test_that("LAR, ASI, TyG_index are NA when their inputs are missing", {
  skip_on_cran()
  df1 <- tibble(HDL_c = 1, TG = 1.3, BMI = 24)
  out1 <- glycemic_markers(df1, verbose = FALSE)
  expect_true(is.na(out1$LAR))
  expect_true(is.na(out1$ASI))
  expect_true(is.na(out1$TyG_index))

  df2 <- tibble(
    HDL_c = 1, TG = 1.3, BMI = 24,
    leptin = 10
  )
  out2 <- glycemic_markers(df2, verbose = FALSE)
  expect_true(is.na(out2$LAR))
})

test_that("output is vectorized and contains all expected columns", {
  skip_on_cran()
  df <- tibble(
    HDL_c       = c(1, 1.1),
    TG          = c(1.3, 1.5),
    BMI         = c(24, 26),
    glucose     = c(5.5, NA),
    HbA1c       = c(44, 39),
    C_peptide   = c(300, NA),
    G0          = c(5.5, NA),
    leptin      = c(10, NA),
    adiponectin = c(5, NA),
    I0          = c(50, NA)
  )
  out <- glycemic_markers(df, verbose = FALSE)
  expect_equal(nrow(out), 2)
  expect_named(
    out,
    c(
      "SPISE", "METS_IR", "prediabetes", "diabetes",
      "HOMA_CP", "LAR", "ASI", "TyG_index"
    )
  )
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df <- tibble(HDL_c = 1, TG = 1.3, BMI = 24)
  expect_message(glycemic_markers(df, verbose = TRUE), "glycemic_markers")
  expect_message(glycemic_markers(df, verbose = TRUE), "col_map")
  expect_message(glycemic_markers(df, verbose = TRUE), "optional inputs")
  expect_message(glycemic_markers(df, verbose = TRUE), "computing markers")
  expect_message(glycemic_markers(df, verbose = TRUE), "results:")
})

test_that("glycemic_markers computes expected columns with minimal inputs", {
  skip_on_cran()
  df <- data.frame(HDL_c = c(1.0, 1.3), TG = c(1.5, 2.0), BMI = c(24, 30))
  res <- glycemic_markers(df, verbose = FALSE)
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("SPISE","METS_IR","prediabetes","diabetes","HOMA_CP","LAR","ASI","TyG_index"))
  # METS_IR, TyG_index depend on glucose; NA by default
  expect_true(all(is.na(res$METS_IR)))
  expect_true(all(is.na(res$TyG_index)))
})

test_that("Optional inputs drive only their outputs; NA handling works", {
  skip_on_cran()
  df <- data.frame(
    HDL_c = c(1.2, 1.3),   # avoid ln(HDL_c) == 0
    TG = c(1.5, 2.0),
    BMI = c(24, 30),
    glucose = c(5.6, NA),
    HbA1c = c(44, 38),
    C_peptide = c(300, 500),
    G0 = c(5.5, 6.2),
    I0 = c(60, 120),
    leptin = c(10, 20),
    adiponectin = c(8, 5)
  )
  res_keep <- glycemic_markers(df, na_action = "keep", verbose = FALSE)
  expect_false(all(is.na(res_keep$METS_IR))) # at least first row computable
  expect_false(all(is.na(res_keep$TyG_index)))
  expect_equal(res_keep$prediabetes, c(1L, 0L))
  expect_equal(res_keep$diabetes, c(0L, 0L))
  # error on NA/non-finite when na_action = "error"
  expect_error(glycemic_markers(df, na_action = "error", verbose = FALSE), "missing or non-finite")
  # omit drops the NA row
  res_omit <- glycemic_markers(df, na_action = "omit", verbose = FALSE)
  expect_equal(nrow(res_omit), 1L)
})

test_that("extreme values produce no warning/error; range note appears in verbose", {
  skip_on_cran()
  df <- data.frame(
    HDL_c = c(1.2, 1.3),
    TG    = c(25, 2.0),    # TG out-of-range (default max 20)
    BMI   = c(24, 30),
    glucose = c(5.6, 7.1)
  )
  # No warning or error — extreme values are noted informally
  expect_no_warning(glycemic_markers(df, verbose = FALSE))
  # SPISE is still computed (values not altered)
  out <- glycemic_markers(df, verbose = FALSE)
  expect_false(any(is.na(out$SPISE)))
  # Range note appears in verbose output
  msgs <- testthat::capture_messages(glycemic_markers(df, verbose = TRUE))
  expect_true(any(grepl("range note", msgs)))
})

test_that("Coercion to numeric warns when NAs introduced", {
  skip_on_cran()
  df <- data.frame(
    HDL_c = c("1.0", "bad"),
    TG = c("1.5", "2.0"),
    BMI = c(24, 30)
  )
  expect_warning(glycemic_markers(df, verbose = FALSE), "coerced to numeric; NAs introduced")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- tibble(HDL_c = 1, TG = 1.3, BMI = 24)
  msgs <- testthat::capture_messages(glycemic_markers(df, verbose = TRUE))
  expect_equal(sum(grepl("col_map",  msgs)), 1L)
  expect_equal(sum(grepl("results:",    msgs)), 1L)
  expect_equal(sum(grepl("optional inputs", msgs)), 1L)
  expect_equal(sum(grepl("computing markers", msgs)), 1L)
})

test_that("glycemic_markers runs and returns key columns", {
  skip_on_cran()
  df <- tibble::tibble(
    HDL_c  = c(1.0, 1.3),
    TG     = c(1.5, 2.0),
    BMI    = c(24, 30),
    glucose = c(5.6, 7.1)
  )
  out <- glycemic_markers(df, na_action = "keep", verbose = FALSE)
  expect_true(all(c("SPISE", "METS_IR") %in% names(out)))
  expect_equal(nrow(out), nrow(df))
})

test_that("ID column is prepended to output when present in data", {
  skip_on_cran()
  df <- tibble::tibble(
    id    = 1:3,
    HDL_c = c(1.0, 1.2, 1.1),
    TG    = c(1.3, 1.5, 1.4),
    BMI   = c(24, 26, 25)
  )
  out <- glycemic_markers(df, verbose = FALSE)
  expect_equal(names(out)[1L], "id")
  expect_equal(out$id, 1:3)
})

test_that("BMI pre-computed from weight and height when BMI absent", {
  skip_on_cran()
  # Without pre-computation this would error (BMI missing)
  df <- tibble::tibble(
    HDL_c  = 1.0,
    TG     = 1.3,
    weight = 70,    # kg
    height = 175    # cm -> should be converted to m
  )
  out <- glycemic_markers(df, verbose = FALSE)
  expected_bmi <- 70 / (175 / 100)^2
  expected_spise <- 600 * 1.0^0.185 / (1.3^0.2 * expected_bmi^1.338)
  expect_equal(out$SPISE, expected_spise, tolerance = 1e-6)
})

test_that("glucose derived from G0 when glucose absent", {
  skip_on_cran()
  df <- tibble::tibble(
    HDL_c = 1.2,   # avoid log(HDL_c) == 0
    TG    = 1.3,
    BMI   = 24,
    G0    = 5.5    # will alias to glucose
  )
  out <- glycemic_markers(df, verbose = FALSE)
  # METS_IR and TyG_index require glucose; should be non-NA via alias
  expect_false(is.na(out$METS_IR))
  expect_false(is.na(out$TyG_index))
})

test_that("glucose derived from mapped G0 column via col_map redirect", {
  skip_on_cran()
  # Physical column is "pglu0" — col_map maps G0 -> "pglu0".
  # .hm_build_col_map materializes data[["G0"]], then precompute derives glucose.
  # METS_IR and TyG_index should be non-NA.
  df <- tibble::tibble(
    HDL_c = 1.2,
    TG    = 1.3,
    BMI   = 24,
    pglu0 = 5.5    # non-standard name for fasting glucose
  )
  cm  <- list(HDL_c = "HDL_c", TG = "TG", BMI = "BMI", G0 = "pglu0")
  out <- glycemic_markers(df, col_map = cm, verbose = FALSE)
  expect_false(is.na(out$METS_IR),    info = "METS_IR requires glucose derived from mapped G0")
  expect_false(is.na(out$TyG_index),  info = "TyG_index requires glucose derived from mapped G0")
})

test_that("BMI derived from mapped weight/height with non-standard column names", {
  skip_on_cran()
  # Physical columns are "wt_kg" and "ht_cm" — must be resolved via col_map.
  # .hm_build_col_map materializes data[["weight"]] and data[["height"]],
  # then .hm_precompute_from_deps derives BMI and SPISE computes.
  df <- tibble::tibble(
    hdlc   = 1.0,
    trig   = 1.3,
    wt_kg  = 70,
    ht_cm  = 175
  )
  cm  <- list(HDL_c = "hdlc", TG = "trig", weight = "wt_kg", height = "ht_cm")
  out <- glycemic_markers(df, col_map = cm, verbose = FALSE)
  expected_bmi   <- 70 / (175 / 100)^2
  expected_spise <- 600 * 1.0^0.185 / (1.3^0.2 * expected_bmi^1.338)
  expect_equal(out$SPISE, expected_spise, tolerance = 1e-6,
               info = "SPISE should use BMI precomputed from mapped weight/height")
})

test_that("partial col_map is filled from dictionary for remaining keys", {
  skip_on_cran()
  # User provides only G0 mapping; HDL_c, TG, BMI should be inferred from
  # dictionary matching column names (pglu0 -> G0, hdlc -> HDL_c, etc.).
  df <- tibble::tibble(
    hdlc  = 1.2,
    trig  = 1.3,
    bmi   = 24,
    pglu0 = 5.5
  )
  cm  <- list(G0 = "pglu0")  # partial — only G0 explicitly provided
  out <- glycemic_markers(df, col_map = cm, verbose = FALSE)
  expect_false(is.na(out$SPISE),     info = "SPISE should compute: HDL_c/TG/BMI inferred by dict")
  expect_false(is.na(out$METS_IR),   info = "METS_IR should compute: glucose derived from G0")
  expect_false(is.na(out$TyG_index), info = "TyG_index should compute: glucose derived from G0")
})

#' SARC-F Sarcopenia Screening Score
#'
#' Computes the SARC-F questionnaire score, a quick screening tool for sarcopenia risk.
#'
#' @details
#' SARC-F has 5 items: Strength, Assistance in walking, Rise from a chair,
#' Climb stairs, and Falls. Each item is scored 0 (no difficulty) to 2 (high difficulty).
#' Total SARC-F score ranges 0-10. A score >= 4 indicates high risk of sarcopenia
#' and suggests further assessment.
#'
#' @param data A data.frame or tibble with SARC-F questionnaire responses.
#' @param col_map Named list mapping the five SARC-F components to columns:
#'   strength, walking, chair, stairs, falls.
#' @param verbose Logical; if TRUE (default), emits progress messages.
#' @param na_action One of c("keep","omit","error","ignore","warn").
#'
#' @return A tibble with:
#'   - sarc_f_score (numeric 0-10; NA if any component is NA)
#'   - sarc_f_high_risk (logical; TRUE if score >= 4, NA if score is NA)
#' @references
#' \insertRef{malmstrom2013sarcf}{HealthMarkers}
#' \insertRef{malmstrom2016sarcf}{HealthMarkers} (SARC-F validation and functional outcome prediction; background)
#'
#' @examples
#' df <- data.frame(Strength = c(1, 2, 0), Walking = c(0, 1, 2),
#'                  Chair = c(1, 1, 2), Stairs = c(0, 2, 2), Falls = c(0, 1, 1))
#' sarc_f_score(df)
#' @export
sarc_f_score <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "sarc_f_score"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  if (!is.data.frame(data)) {
    rlang::abort(
      "sarc_f_score(): `data` must be a data.frame or tibble.",
      class = "healthmarkers_sarcf_error_data_type"
    )
  }

  if (!is.list(col_map) && !is.null(col_map)) {
    rlang::abort(
      "sarc_f_score(): `col_map` must be a named list.",
      class = "healthmarkers_sarcf_error_colmap_type"
    )
  }

  req <- c("strength","walking","chair","stairs","falls")
  cm      <- .hm_build_col_map(data, col_map, req, fn = "sarc_f_score")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  if (!is.list(col_map)) {
    rlang::abort(
      "sarc_f_score(): `col_map` must be a named list.",
      class = "healthmarkers_sarcf_error_colmap_type"
    )
  }

  if (length(col_map) == 0L || is.null(names(col_map))) {
    rlang::abort(
      "sarc_f_score(): missing or empty `col_map`.",
      class = "healthmarkers_sarcf_error_missing_map"
    )
  }

  missing_keys <- setdiff(req, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(
      paste0("sarc_f_score(): missing col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_sarcf_error_missing_map"
    )
  }

  mapped <- unname(unlist(col_map[req], use.names = FALSE))
  if (any(!nzchar(mapped))) {
    bad <- req[!nzchar(unname(unlist(col_map[req])))]
    rlang::abort(
      paste0("sarc_f_score(): missing col_map entries for: ",
             paste(bad, collapse = ", ")),
      class = "healthmarkers_sarcf_error_bad_map_values"
    )
  }

  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("sarc_f_score(): missing required columns in data: ",
             paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_sarcf_error_missing_columns"
    )
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  sarc_f_score      [0-10 sum]\n  sarc_f_high_risk  [score >= 4]", fn_name), level = "inform")

  for (cn in mapped) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0L) {
        rlang::warn(
          sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
          class = "healthmarkers_sarcf_warn_na_coercion"
        )
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  xs <- data[[col_map$strength]]
  xw <- data[[col_map$walking]]
  xc <- data[[col_map$chair]]
  xt <- data[[col_map$stairs]]
  xf <- data[[col_map$falls]]

  any_na <- is.na(xs) | is.na(xw) | is.na(xc) | is.na(xt) | is.na(xf)

  if (na_action_raw == "warn" && any(any_na)) {
    rlang::warn(
      "Missing SARC-F item values; total score will be NA for those entries.",
      class = "healthmarkers_sarcf_warn_missing_inputs"
    )
  }

  if (na_action_eff == "error" && any(any_na)) {
    rlang::abort(
      "sarc_f_score(): required inputs contain missing values (na_action='error').",
      class = "healthmarkers_sarcf_error_missing_values"
    )
  }

  keep <- if (na_action_eff == "omit") !any_na else rep(TRUE, length(xs))

  d_xs <- xs[keep]; d_xw <- xw[keep]; d_xc <- xc[keep]; d_xt <- xt[keep]; d_xf <- xf[keep]

  out_of_range <- function(x) is.finite(x) & (x < 0 | x > 2)
  if (any(out_of_range(d_xs) | out_of_range(d_xw) |
          out_of_range(d_xc) | out_of_range(d_xt) | out_of_range(d_xf))) {
    rlang::warn(
      "sarc_f_score(): SARC-F items should be 0, 1, or 2; values outside this range detected.",
      class = "healthmarkers_sarcf_warn_out_of_range"
    )
  }

  hm_inform("sarc_f_score(): computing markers", level = "debug")

  total_score <- d_xs + d_xw + d_xc + d_xt + d_xf
  high_risk   <- ifelse(is.na(total_score), NA, total_score >= 4)

  out_core <- tibble::tibble(
    sarc_f_score     = total_score,
    sarc_f_high_risk = as.logical(high_risk)
  )

  if (na_action_eff != "omit") {
    out <- tibble::tibble(
      sarc_f_score     = rep(NA_real_, length(xs)),
      sarc_f_high_risk = rep(NA, length(xs))
    )
    out[keep, ] <- out_core
  } else {
    out <- out_core
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

library(testthat)

cm <- list(strength="Strength", walking="Walking", chair="Chair", stairs="Stairs", falls="Falls")

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(Strength=0, Walking=0, Chair=0, Stairs=0, Falls=0)
  # empty list with all columns in data – succeeds via inference
  expect_no_error(sarc_f_score(df, list()))
  expect_error(
    sarc_f_score(df, list(strength="", walking="Walking", chair="Chair", stairs="Stairs", falls="Falls")),
    class = "healthmarkers_sarcf_error_bad_map_values"
  )
  expect_error(
    sarc_f_score(data.frame(A=1,B=1,C=1,D=1,E=1), cm),
    class = "healthmarkers_sarcf_error_missing_columns"
  )
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df <- data.frame(Strength=0, Walking=0, Chair=0, Stairs=0, Falls=0)
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(sarc_f_score(df, cm, verbose = TRUE), "sarc_f_score")
  expect_message(sarc_f_score(df, cm, verbose = TRUE), "col_map")
  expect_message(sarc_f_score(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- data.frame(Strength=0, Walking=0, Chair=0, Stairs=0, Falls=0)
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(sarc_f_score(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("numeric coercion warning when strings introduce NAs", {
  skip_on_cran()
  df <- data.frame(
    Strength=c("0","oops"),
    Walking =c("1","1"),
    Chair   =c("1","1"),
    Stairs  =c("0","0"),
    Falls   =c("2","2")
  )
  expect_warning(sarc_f_score(df, cm), class = "healthmarkers_sarcf_warn_na_coercion")
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(
    Strength=c(0, NA, 2),
    Walking =c(1, 1, 2),
    Chair   =c(1, 1, NA),
    Stairs  =c(0, 1, 2),
    Falls   =c(0, 1, 2)
  )

  out_keep <- sarc_f_score(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(is.na(out_keep$sarc_f_score[2]) || is.na(out_keep$sarc_f_score[3]))

  out_omit <- sarc_f_score(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)

  expect_error(
    sarc_f_score(df, cm, na_action = "error"),
    class = "healthmarkers_sarcf_error_missing_values"
  )

  expect_warning(
    sarc_f_score(df, cm, na_action = "warn"),
    class = "healthmarkers_sarcf_warn_missing_inputs"
  )
})

test_that("domain warnings for values outside 0–2", {
  skip_on_cran()
  df <- data.frame(
    Strength=c(-1,0),
    Walking =c(3,1),
    Chair   =c(1,1),
    Stairs  =c(0,0),
    Falls   =c(0,0)
  )
  expect_warning(
    sarc_f_score(df, cm),
    class = "healthmarkers_sarcf_warn_out_of_range"
  )
})

test_that("out-of-range items (0-2) trigger domain warning and still compute", {
  skip_on_cran()
  df <- data.frame(
    Strength=c(-1, 0, 3),
    Walking =c(0, 2, 1),
    Chair   =c(1, 1, 1),
    Stairs  =c(0, 0, 0),
    Falls   =c(0, 0, 0)
  )
  expect_warning(
    sarc_f_score(df, cm),
    class = "healthmarkers_sarcf_warn_out_of_range"
  )
  out <- suppressWarnings(sarc_f_score(df, cm))
  expect_equal(nrow(out), 3L)
})

test_that("thresholding: high risk is TRUE for score >= 4", {
  skip_on_cran()
  df <- data.frame(
    Strength=c(2,1,0,NA),
    Walking =c(1,1,1,1),
    Chair   =c(1,1,1,1),
    Stairs  =c(0,0,0,0),
    Falls   =c(0,0,0,0)
  )

  out <- sarc_f_score(df, cm)
  expect_equal(out$sarc_f_score[1], 4)
  expect_equal(out$sarc_f_score[2], 3)
  expect_equal(out$sarc_f_score[3], 2)

  expect_true(out$sarc_f_high_risk[1])
  expect_false(out$sarc_f_high_risk[2])
  expect_false(out$sarc_f_high_risk[3])
  expect_true(is.na(out$sarc_f_high_risk[4]))
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(
    Strength=c(0,NA,2),
    Walking =c(1,1,2),
    Chair   =c(1,1,2),
    Stairs  =c(0,1,2),
    Falls   =c(0,1,2)
  )

  out_keep <- sarc_f_score(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(sarc_f_score(df, cm, na_action = "warn"))

  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

#' Liver fat surrogates: HSI and NAFLD Liver Fat Score
#'
#' Computes:
#' - HSI = 8 * (ALT/AST) + BMI + 2 (if female) + 2 (if diabetes)
#' - NAFLD-LFS = -2.89 + 1.18*MetS + 0.45*Type2DM + 0.15*Insulin_u + 0.04*AST - 0.94*(AST/ALT)
#'
#' Assumptions/units:
#' - ALT, AST in U/L; BMI in kg/m^2.
#' - `insulin` is expected in muU/mL; if unavailable and `I0` is provided,
#'   `I0` is interpreted in pmol/L and converted to muU/mL via /6.
#' - MetS is taken directly if provided; otherwise derived using a simplified
#'   NCEP-ATP III style rule when sufficient inputs exist.
#' - Type2DM is taken from `diabetes` (logical or 0/1).
#'
#' These scores are surrogate indices for research/feature-engineering use and
#' are not validated as standalone clinical diagnostic tools.
#'
#' @param data Data frame with needed columns (see col_map).
#' @param col_map Named list mapping:
#'   - Required for HSI: ALT, AST, BMI
#'   - Optional direct inputs: sex, diabetes, MetS, insulin
#'   - Optional to derive MetS or insulin: I0, waist, TG, HDL_c, sbp, bp_sys, bp_treated, glucose, G0
#' @param na_action One of c("keep","omit","error","ignore","warn").
#' @param na_warn_prop Proportion in \eqn{[0,1]} for high-missingness warnings when na_action = "warn". Default 0.2.
#' @param verbose Logical; if TRUE, prints column mapping and computing messages.
#' @return A tibble with columns HSI and NAFLD_LFS.
#' @references
#' \insertRef{lee2010hsi}{HealthMarkers}
#' \insertRef{kotronen2009nafldlfs}{HealthMarkers}
#' @examples
#' df <- data.frame(ALT=20, AST=25, BMI=27, sex="female", diabetes=FALSE, I0=60)
#' liver_fat_markers(
#'   df,
#'   col_map = list(ALT="ALT", AST="AST", BMI="BMI",
#'                  sex="sex", diabetes="diabetes", I0="I0")
#' )
#' @export
liver_fat_markers <- function(data,
                              col_map = NULL,
                              na_action = c("keep", "omit", "error", "ignore", "warn"),
                              na_warn_prop = 0.2,
                              verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "liver_fat_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  
  # Validate required mapping and data
  all_lfm_keys <- c("ALT","AST","BMI","sex","diabetes","MetS","insulin",
                    "I0","waist","TG","HDL_c","sbp","bp_sys","bp_treated","glucose","G0")
  cm      <- .hm_build_col_map(data, col_map, all_lfm_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  req <- c("ALT", "AST", "BMI")
  hm_validate_inputs(data, col_map, required_keys = req, fn = "liver_fat_markers")
  mapped_req <- unname(unlist(col_map[req]))
  miss <- setdiff(mapped_req, names(data))
  if (length(miss))
    rlang::abort(paste(
      "liver_fat_markers(): missing required columns:",
      paste(miss, collapse = ", ")
    ))

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf(
      "%s(): computing markers:\n  HSI        [8*(ALT/AST) + BMI + sex + diabetes]\n  NAFLD_LFS  [MetS/insulin/diabetes required; NA if unavailable]",
      fn_name), level = "inform")

  # Columns potentially used directly in formulas
  direct_keys <- c("ALT", "AST", "BMI", "sex", "diabetes", "MetS", "insulin", "I0")
  # Additional optional used for MetS derivation
  deriva_keys <- c("waist",
                   "TG",
                   "HDL_c",
                   "sbp",
                   "bp_sys",
                   "bp_treated",
                   "glucose",
                   "G0")
  # Build list of mapped columns present
  mapped_all <- unlist(col_map[intersect(names(col_map), c(direct_keys, deriva_keys))], use.names = TRUE)
  used_cols <- unique(mapped_all[!is.na(mapped_all) &
                                   nzchar(mapped_all) & mapped_all %in% names(data)])
  
  # Coerce only numeric-like columns; warn if NAs introduced
  numeric_like <- intersect(used_cols, unname(unlist(col_map[c(
    "ALT",
    "AST",
    "BMI",
    "insulin",
    "I0",
    "waist",
    "TG",
    "HDL_c",
    "sbp",
    "bp_sys",
    "glucose",
    "G0"
  )])))
  for (cn in numeric_like) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L)
        rlang::warn(sprintf(
          "Column '%s' coerced to numeric; NAs introduced: %d",
          cn,
          introduced_na
        ))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }
  
  # NA policy: build used frame for handling
  used_for_policy <- unique(unname(unlist(col_map[intersect(names(col_map), direct_keys)])))
  used_for_policy <- used_for_policy[!is.na(used_for_policy) &
                                       nzchar(used_for_policy)]
  if (length(used_for_policy) == 0L)
    used_for_policy <- mapped_req
  if (na_action_raw == "warn" && length(used_for_policy)) {
    dfp <- data[, used_for_policy, drop = FALSE]
    any_na_cols <- names(which(colSums(is.na(dfp)) > 0))
    high_na_cols <- names(which(colMeans(is.na(dfp)) >= na_warn_prop))
    if (length(any_na_cols))
      rlang::warn(sprintf("Missing values in: %s.", paste(any_na_cols, collapse = ", ")))
    if (length(high_na_cols))
      rlang::warn(sprintf(
        "High missingness (>= %.0f%%): %s.",
        100 * na_warn_prop,
        paste(high_na_cols, collapse = ", ")
      ))
  }
  cc_req <- stats::complete.cases(data[, mapped_req, drop = FALSE])
  if (na_action_eff == "error" && !all(cc_req)) {
    rlang::abort("liver_fat_markers(): missing/non-finite values with na_action='error'.")
  }
  keep <- if (na_action_eff == "omit")
    cc_req
  else
    rep(TRUE, nrow(data))
  
  d <- data[keep, , drop = FALSE]
  
  # Extract helpers from subset
  getv <- function(k) {
    nm <- col_map[[k]]
    if (!is.null(nm) &&
        nzchar(nm) && nm %in% names(d))
      d[[nm]]
    else
      NULL
  }
  ALT <- getv("ALT")
  AST <- getv("AST")
  BMI <- getv("BMI")
  sex <- getv("sex")
  diabetes <- getv("diabetes")
  MetS_direct <- getv("MetS")
  insulin <- getv("insulin")
  I0 <- getv("I0")
  glucose <- getv("glucose")
  G0 <- getv("G0")
  TG <- getv("TG")
  HDL <- getv("HDL_c")
  waist <- getv("waist")
  sbp <- if (!is.null(getv("sbp")))
    getv("sbp")
  else
    getv("bp_sys")
  bp_treated <- getv("bp_treated")
  
  # Helpers
  sdiv <- function(a, b) {
    z <- a / b
    z[!is.finite(z)] <- NA_real_
    z
  }
  as01 <- function(x) {
    if (is.null(x))
      return(NULL)
    if (is.logical(x))
      return(as.integer(x))
    if (is.numeric(x))
      return(as.integer(x > 0))
    xv <- suppressWarnings(as.numeric(x))
    as.integer(ifelse(is.na(xv), NA, xv > 0))
  }
  female_flag <- {
    if (is.null(sex))
      rep(0L, length(BMI))
    else {
      if (is.numeric(sex))
        as.integer(sex == 2)
      else if (is.logical(sex))
        as.integer(sex)
      else {
        xchr <- tolower(as.character(sex))
        as.integer(xchr %in% c("f", "female", "woman", "women"))
      }
    }
  }
  dm2_flag <- as01(diabetes)
  if (is.null(dm2_flag))
    dm2_flag <- rep(0L, length(BMI))
  
  # HSI
  HSI <- 8 * sdiv(ALT, AST) + BMI + ifelse(female_flag == 1L, 2, 0) + ifelse(dm2_flag == 1L, 2, 0)
  
  # MetS: use direct mapping if available; otherwise derive; else NA
  derive_mets <- function() {
    n <- length(BMI)
    out <- rep(NA_integer_, n)
    have <- list(
      waist = !is.null(waist),
      TG = !is.null(TG),
      HDL = !is.null(HDL),
      sbp = !is.null(sbp),
      glucose = !is.null(glucose) || !is.null(G0)
    )
    if (!all(unlist(have)))
      return(out)
    glu <- if (!is.null(glucose))
      glucose
    else
      G0
    low_hdl <- ifelse(female_flag == 1L, HDL < 1.29, HDL < 1.03)
    crits <- cbind(
      waist = waist > ifelse(female_flag == 1L, 88, 102),
      tg    = TG >= 1.7,
      hdl   = low_hdl,
      bp    = sbp >= 130 |
        (if (!is.null(bp_treated))
          as01(bp_treated) == 1L
         else
           FALSE),
      glu   = glu >= 5.6
    )
    as.integer(rowSums(crits, na.rm = TRUE) >= 3)
  }
  MetS <- if (!is.null(MetS_direct))
    as01(MetS_direct)
  else
    derive_mets()
  if (is.null(MetS))
    MetS <- rep(NA_integer_, length(BMI))
  
  # fasting insulin in muU/mL: prefer direct insulin; else I0/6 if available
  Ins_u <- if (!is.null(insulin)) {
    insulin
  } else if (!is.null(I0)) {
    I0 / 6
  } else {
    rep(NA_real_, length(BMI))
  }
  
  NAFLD_LFS <- -2.89 + 1.18 * MetS + 0.45 * dm2_flag + 0.15 * Ins_u + 0.04 * AST - 0.94 * sdiv(AST, ALT)
  
  out <- tibble::tibble(HSI = as.numeric(HSI), NAFLD_LFS = as.numeric(NAFLD_LFS))
  if (na_action_eff != "omit") {
    res <- tibble::tibble(HSI = rep(NA_real_, nrow(data)),
                          NAFLD_LFS = rep(NA_real_, nrow(data)))
    res$HSI[keep] <- out$HSI
    res$NAFLD_LFS[keep] <- out$NAFLD_LFS
    out <- res
  }
  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }
  out
}

test_that("liver_fat_markers computes HSI and NAFLD_LFS", {
  skip_on_cran()
  df <- data.frame(
    ALT = c(30, 40), AST = c(20, 35), BMI = c(25, 32),
    sex = c(1, 2), diabetes = c(0, 1), MetS = c(1, 0), insulin = c(15, 20)
  )
  res <- liver_fat_markers(df, col_map = list(
    ALT = "ALT", AST = "AST", BMI = "BMI", sex = "sex",
    diabetes = "diabetes", MetS = "MetS", insulin = "insulin"
  ))
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("HSI", "NAFLD_LFS"))
  exp_hsi <- 8 * (df$ALT / df$AST) + df$BMI + 2 * as.numeric(df$sex == 2) + 2 * as.numeric(df$diabetes == 1)
  exp_lfs <- -2.89 + 1.18 * df$MetS + 0.45 * df$diabetes + 0.15 * df$insulin + 0.04 * df$AST - 0.94 * (df$AST / df$ALT)
  expect_equal(res$HSI, as.numeric(exp_hsi), tolerance = 1e-10)
  expect_equal(res$NAFLD_LFS, as.numeric(exp_lfs), tolerance = 1e-10)
})

test_that("liver_fat_markers missing handling omit", {
  skip_on_cran()
  df <- data.frame(
    ALT = c(30, NA), AST = c(20, 35), BMI = c(25, 32),
    sex = c(1, 2), diabetes = c(0, 1), MetS = c(1, 0), insulin = c(15, 20)
  )
  res <- liver_fat_markers(
    df,
    col_map = list(
      ALT = "ALT", AST = "AST", BMI = "BMI", sex = "sex",
      diabetes = "diabetes", MetS = "MetS", insulin = "insulin"
    ),
    na_action = "omit",
    na_warn_prop = 0
  )
  expect_equal(nrow(res), 1L)
})

test_that("verbose = TRUE emits preparing, col_map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(ALT = 30, AST = 20, BMI = 25, sex = 2, diabetes = 1, insulin = 15)
  cm <- list(ALT = "ALT", AST = "AST", BMI = "BMI", sex = "sex",
             diabetes = "diabetes", insulin = "insulin")
  expect_message(liver_fat_markers(df, col_map = cm, verbose = TRUE), "liver_fat_markers")
  expect_message(liver_fat_markers(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(liver_fat_markers(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(ALT = 30, AST = 20, BMI = 25, sex = 2, diabetes = 1, insulin = 15)
  cm <- list(ALT = "ALT", AST = "AST", BMI = "BMI", sex = "sex",
             diabetes = "diabetes", insulin = "insulin")
  msgs <- testthat::capture_messages(
    liver_fat_markers(df, col_map = cm, verbose = TRUE)
  )
  expect_true(sum(grepl("col_map", msgs)) >= 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Compute anthropometric obesity & adiposity indices
#'
#' Calculates a comprehensive set of body shape and adiposity indices:
#' * BMI and WHO BMI categories
#' * Waist-to-hip ratio (WHR) and optional WHR adjusted for BMI (WHRadjBMI)
#' * Waist-to-height ratio (WHtR)
#' * Abdominal Volume Index (AVI)
#' * Body Adiposity Index (BAI)
#' * A Body Shape Index (ABSI)
#' * Body Roundness Index (BRI)
#' * Conicity Index (CI)
#' * (Optional) Relative Fat Mass (RFM)
#'
#' Units assumed (no automatic conversion beyond the specified weight/height options):
#' - weight: kg (or lb if weight_unit = "lb")
#' - height: m (or cm if height_unit = "cm")
#' - waist, hip: cm
#'
#' Note: WHtR, ABSI, BRI, CI, and RFM all require waist in the same unit as
#' height (metres). The function converts waist internally (waist_cm / 100)
#' for these five indices; users should always supply waist in cm.
#' - sex: 0 = male, 1 = female (only required if include_RFM = TRUE)
#'
#' @param data A data.frame or tibble containing the input columns.
#' @param weight Unquoted column name for weight.
#' @param height Unquoted column name for height.
#' @param waist Unquoted column name for waist circumference.
#' @param hip Unquoted column name for hip circumference.
#' @param sex   (Optional) Unquoted column name for sex, coded 0=male, 1=female; required if include_RFM=TRUE.
#' @param weight_unit   One of c("kg","lb"); if "lb", converts weight to kg by *0.45359237.
#' @param height_unit   One of c("cm","m"); if "cm", converts height to metres by /100.
#' @param adjust_WHR    Logical; if TRUE, adds a column WHRadjBMI as residuals from WHR ~ BMI.
#' @param include_RFM   Logical; if TRUE, computes Relative Fat Mass (requires sex column).
#' @param na_action One of c("keep","omit","error") for handling NA in required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness warnings for required inputs. Default 0.2.
#' @param verbose Logical; if TRUE, prints column mapping and computing messages.
#'
#' @return A tibble with only the computed indices (slim output):
#' * weight_kg, height_m (unit-normalised intermediates),
#' * BMI, BMI_cat,
#' * WHR, WHRadjBMI (if `adjust_WHR = TRUE`),
#' * waist_to_height_ratio, waist_to_BMI_ratio, weight_to_height_ratio,
#' * AVI, BAI, ABSI, BRI, CI,
#' * RFM (if `include_RFM = TRUE`).
#'
#' @importFrom dplyr mutate case_when
#' @importFrom rlang enquo quo_name quo_is_null abort warn inform
#' @importFrom stats lm resid
#' @export
#'
#' @examples
#' library(tibble)
#' df <- tibble(
#'   wt     = c(70, 80),  # kg
#'   ht     = c(175, 165),# cm
#'   waist  = c(80, 90),  # cm
#'   hip    = c(100, 95), # cm
#'   sex    = c(0, 1)
#' )
#' obesity_indices(
#'   df,
#'   weight       = wt,
#'   height       = ht,
#'   waist        = waist,
#'   hip          = hip,
#'   sex          = sex,
#'   weight_unit  = "kg",
#'   height_unit  = "cm",
#'   adjust_WHR   = TRUE,
#'   include_RFM  = TRUE,
#'   verbose      = TRUE
#' )
#' @references
#' \insertRef{quetelet1842}{HealthMarkers}
#' \insertRef{who1995physicalstatus}{HealthMarkers}
#' \insertRef{guerrero1999avi}{HealthMarkers}
#' \insertRef{bergman2011bai}{HealthMarkers}
#' \insertRef{krakauer2012absi}{HealthMarkers}
#' \insertRef{thomas2013bri}{HealthMarkers}
#' \insertRef{valdez1991ci}{HealthMarkers}
#' \insertRef{woolcott2018rfm}{HealthMarkers}
#' \insertRef{calle1999bmi}{HealthMarkers}
#' \insertRef{freedman2012bai}{HealthMarkers}
#' \insertRef{he2013absi}{HealthMarkers}
#' \insertRef{maessen2014bri}{HealthMarkers}
obesity_indices <- function(data,
                            weight,
                            height,
                            waist,
                            hip,
                            sex = NULL,
                            weight_unit = c("kg", "lb"),
                            height_unit = c("cm", "m"),
                            adjust_WHR = FALSE,
                            include_RFM = FALSE,
                            na_action = c("keep","omit","error"),
                            na_warn_prop = 0.2,
                            verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "obesity_indices"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  # Match args
  weight_unit <- match.arg(weight_unit)
  height_unit <- match.arg(height_unit)
  na_action <- match.arg(na_action)

  # Capture and quote arguments
  wt_q <- rlang::enquo(weight);  wt_name  <- rlang::quo_name(wt_q)
  ht_q <- rlang::enquo(height);  ht_name  <- rlang::quo_name(ht_q)
  wst_q<- rlang::enquo(waist);   wst_name <- rlang::quo_name(wst_q)
  hp_q <- rlang::enquo(hip);     hp_name  <- rlang::quo_name(hp_q)
  sx_q <- rlang::enquo(sex);     sx_name  <- if (!rlang::quo_is_null(sx_q)) rlang::quo_name(sx_q) else NULL

  # Validate data frame and required columns
  .oi_validate_data_and_cols(data, c(wt_name, ht_name, wst_name, hp_name), include_RFM, sx_name)
  .oi_warn_high_missing(data, c(wt_name, ht_name, wst_name, hp_name, if (include_RFM) sx_name else NULL), na_warn_prop)
  if (isTRUE(verbose)) {
    col_mapping_str <- paste(
      c(sprintf("weight -> '%s'", wt_name), sprintf("height -> '%s'", ht_name),
        sprintf("waist -> '%s'", wst_name), sprintf("hip -> '%s'", hp_name),
        if (!is.null(sx_name)) sprintf("sex -> '%s'", sx_name) else NULL),
      collapse = ", ")
    hm_inform(sprintf("%s(): col_map: %s", fn_name, col_mapping_str), level = "inform")
    hm_inform(sprintf(
      "%s(): computing markers:\n  BMI, BMI_cat, WHR, WHtR, AVI, BAI, ABSI, BRI, CI%s",
      fn_name,
      paste0(if (isTRUE(adjust_WHR)) ", WHRadjBMI" else "", if (isTRUE(include_RFM)) ", RFM" else "")),
      level = "inform")
  }

  # NA policy on required inputs
  used_cols <- c(wt_name, ht_name, wst_name, hp_name, if (include_RFM) sx_name else NULL)
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("obesity_indices(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_obesity_error_missing_values")
    }
  } else if (na_action == "omit") {
    if (length(used_cols)) {
      keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
      hm_inform(sprintf("obesity_indices(): omitting %d rows with NA in required inputs", sum(!keep)),
                level = if (isTRUE(verbose)) "inform" else "debug")
      data <- data[keep, , drop = FALSE]
    }
  }
  # Unit-normalized base fields
  out <- dplyr::mutate(
    data,
    weight_kg = dplyr::case_when(
      weight_unit == "kg" ~ !!wt_q,
      weight_unit == "lb" ~ !!wt_q * 0.45359237
    ),
    height_m = dplyr::case_when(
      height_unit == "m" ~ !!ht_q,
      height_unit == "cm" ~ !!ht_q / 100
    )
  )

  # Compute BMI and category using normalized units
  out <- dplyr::mutate(
    out,
    BMI = weight_kg / (height_m^2),
    BMI_cat = dplyr::case_when(
      BMI < 18.5 ~ "Underweight",
      BMI < 25 ~ "Normal weight",
      BMI < 30 ~ "Overweight",
      BMI < 35 ~ "Obesity Class I",
      BMI < 40 ~ "Obesity Class II",
      BMI >= 40 ~ "Obesity Class III",
      TRUE ~ NA_character_
    )
  )

  # Safe division helper with denominator-zero tracking
  denom_zero <- new.env(parent = emptyenv()); denom_zero$counts <- list()
  safe_div <- function(num, den, label) {
    res <- rep(NA_real_, length(num))
    ok <- !is.na(num) & !is.na(den)
    zero_den <- ok & (den == 0)
    denom_zero$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    valid <- ok & !zero_den
    res[valid] <- num[valid] / den[valid]
    res[!is.finite(res)] <- NA_real_
    res
  }
  # Safe power roots
  safe_pow <- function(x, p) {
    z <- rep(NA_real_, length(x))
    ok <- !is.na(x) & x > 0
    z[ok] <- x[ok]^p
    z
  }

  # Pull raw columns for waist/hip/sex
  wst <- out[[wst_name]]
  hip <- out[[hp_name]]
  sex_vec <- if (!is.null(sx_name)) out[[sx_name]] else NULL

  # Derived indices
  WHR <- safe_div(wst, hip, "WHR")
  WHtR <- safe_div(wst, out$height_m * 100, "waist_to_height_ratio")  # both in cm
  waist_BMI <- safe_div(wst, out$BMI, "waist_to_BMI_ratio")
  wt_ht <- safe_div(out$weight_kg, out$height_m, "weight_to_height_ratio")

  AVI <- (2 * (wst^2) + 0.7 * (wst - hip)^2) / 1000
  BAI <- ifelse(out$height_m > 0, hip / (out$height_m^1.5) - 18, NA_real_)

  denom_absi <- (safe_pow(out$BMI, 2/3)) * (safe_pow(out$height_m, 1/2))
  ABSI <- safe_div(wst / 100, denom_absi, "ABSI")  # waist to metres (Krakauer 2012)

  # BRI: waist converted to metres; ratio must be dimensionless (Thomas et al. 2013)
  ratio <- ifelse(out$height_m > 0, (wst / 100) / (2 * pi * out$height_m), NA_real_)
  BRI <- 364.2 - 365.5 * sqrt(pmax(0, 1 - (ratio^2)))

  # CI: waist converted to metres per Valdez (1991)
  denom_ci <- 0.109 * sqrt(safe_div(out$weight_kg, out$height_m, "CI_internal"))
  CI <- safe_div(wst / 100, denom_ci, "CI")

  # Attach derived columns
  out$WHR <- WHR
  out$waist_to_height_ratio <- WHtR
  out$waist_to_BMI_ratio <- waist_BMI
  out$weight_to_height_ratio <- wt_ht
  out$AVI <- as.numeric(AVI)
  out$BAI <- as.numeric(BAI)
  out$ABSI <- as.numeric(ABSI)
  out$BRI <- as.numeric(BRI)
  out$CI  <- as.numeric(CI)

  # WHR adjusted for BMI
  if (isTRUE(adjust_WHR)) {
    # Fit only on rows with finite WHR and BMI
    fit_df <- out[is.finite(out$WHR) & is.finite(out$BMI), c("WHR","BMI")]
    if (nrow(fit_df) >= 2L && stats::var(fit_df$BMI, na.rm = TRUE) > 0) {
      res <- stats::resid(stats::lm(WHR ~ BMI, data = fit_df))
      # Map residuals back to full rows (NA where not used in fit)
      WHRadj <- rep(NA_real_, nrow(out))
      WHRadj[as.numeric(rownames(fit_df))] <- as.numeric(res)
      out$WHRadjBMI <- WHRadj
    } else {
      rlang::warn("obesity_indices(): insufficient data to compute WHRadjBMI; coercing to NA.")
      out$WHRadjBMI <- NA_real_
    }
  }

  # Relative Fat Mass
  if (isTRUE(include_RFM)) {
    # Sex for RFM is expected as 0=male, 1=female; use directly without normalization
    sex_01 <- suppressWarnings(as.numeric(out[[sx_name]]))
    invalid_sex <- !is.na(sex_01) & !(sex_01 %in% c(0, 1))
    if (any(invalid_sex)) {
      rlang::warn(
        sprintf("obesity_indices(): 'sex' contains %d values not in {0,1}; RFM set to NA for those rows. Encode sex as 0=male, 1=female.", sum(invalid_sex)),
        class = "healthmarkers_obesity_warn_invalid_sex_rfm"
      )
      sex_01[invalid_sex] <- NA_real_
    }
    denom_rfm <- wst / 100  # convert waist from cm to metres (height_m already in m)
    RFM <- 64 - 20 * safe_div(out$height_m, denom_rfm, "RFM") + 12 * sex_01
    out$RFM <- as.numeric(RFM)
  }

  # Consolidated denominator-zero warning
  dz <- denom_zero$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz)
    nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("obesity_indices(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  # Slim output: return only computed index columns
  index_cols <- c(
    "weight_kg", "height_m", "BMI", "BMI_cat",
    "WHR",
    if (isTRUE(adjust_WHR)) "WHRadjBMI",
    "waist_to_height_ratio", "waist_to_BMI_ratio", "weight_to_height_ratio",
    "AVI", "BAI", "ABSI", "BRI", "CI",
    if (isTRUE(include_RFM)) "RFM"
  )
  slim <- out[, intersect(index_cols, names(out)), drop = FALSE]

  if (!is.null(id_col)) {
    slim[[id_col]] <- data[[id_col]]
    slim <- slim[, c(id_col, setdiff(names(slim), id_col)), drop = FALSE]
    slim <- tibble::as_tibble(slim)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(slim, fn_name), level = "inform") }
  slim
}

# ---- internal helpers ---------------------------------------------------------

.oi_validate_data_and_cols <- function(data, required, include_RFM, sx_name) {
  if (!is.data.frame(data)) {
    rlang::abort("obesity_indices(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_obesity_error_data_type")
  }
  missing_cols <- setdiff(required, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("obesity_indices(): missing required columns: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_obesity_error_missing_columns"
    )
  }
  # Validate numeric types
  for (cn in required) {
    if (!is.numeric(data[[cn]])) {
      rlang::abort(sprintf("obesity_indices(): column '%s' must be numeric.", cn),
                   class = "healthmarkers_obesity_error_type")
    }
  }
  if (isTRUE(include_RFM)) {
    if (is.null(sx_name)) {
      rlang::abort("obesity_indices(): 'sex' must be provided to compute RFM",
                   class = "healthmarkers_obesity_error_missing_sex")
    }
    # Accept character sex (M/F/male/female/1/2/0); normalize to 0=male, 1=female for RFM
    if (!is.numeric(data[[sx_name]]) ||
        any(is.finite(data[[sx_name]]) & !(data[[sx_name]] %in% c(0, 1)))) {
      data[[sx_name]] <- .hm_normalize_sex(data[[sx_name]], to = "01", fn = "obesity_indices")
    }
  }
  invisible(TRUE)
}

.oi_warn_high_missing <- function(df, cols, na_warn_prop = 0.2) {
  cols <- cols[!is.null(cols)]
  for (cn in cols) {
    x <- df[[cn]]; n <- length(x); if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("obesity_indices(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }
  invisible(TRUE)
}

.oi_default_extreme_rules <- function() {
  list(
    weight_kg = c(20, 400),
    height_m  = c(1.2, 2.5),
    waist     = c(30, 200),  # cm
    hip       = c(30, 200)   # cm
  )
}

.oi_extreme_scan <- function(df, cols, rules) {
  count <- 0L
  flags <- list()
  for (cn in cols) {
    if (!cn %in% names(df)) next
    key <- if (cn %in% names(rules)) cn else cn
    rng <- rules[[key]]
    if (is.null(rng)) next
    x <- df[[cn]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    flags[[cn]] <- bad
    count <- count + sum(bad, na.rm = TRUE)
  }
  list(count = count, flags = flags)
}

.oi_cap_inputs <- function(df, flags, rules) {
  for (cn in names(flags)) {
    rng <- rules[[cn]]; if (is.null(rng)) next
    x <- df[[cn]]; bad <- flags[[cn]]
    x[bad & is.finite(x) & x < rng[1]] <- rng[1]
    x[bad & is.finite(x) & x > rng[2]] <- rng[2]
    df[[cn]] <- x
  }
  df
}

# tests/testthat/test-obesity_metrics.R

library(testthat)
library(tibble)

# Base synthetic data for testing (realistic units: kg, m, cm)
base_df <- tibble(
  wt    = c(80, 90),      # kg
  ht    = c(1.75, 1.65),  # m
  waist = c(90, 100),     # cm
  hip   = c(100, 110),    # cm
  sex   = c(0, 1)         # 0 = male, 1 = female
)

# 1) Core metrics without options
test_that("core metrics compute correctly without options", {
  out <- obesity_indices(
    data = base_df,
    weight = wt,
    height = ht,
    waist = waist,
    hip = hip,
    weight_unit = "kg",
    height_unit = "m",
    adjust_WHR = FALSE,
    include_RFM = FALSE
  )

  # weight_kg and height_m
  expect_equal(out$weight_kg, c(80, 90))
  expect_equal(out$height_m, c(1.75, 1.65))

  # BMI and category
  expect_equal(out$BMI, c(80 / 1.75^2, 90 / 1.65^2), tolerance = 1e-6)
  expect_equal(out$BMI_cat, c("Overweight", "Obesity Class I"))

  # WHR and waist_to_height_ratio (both in cm for WHtR)
  expect_equal(out$WHR, c(90 / 100, 100 / 110), tolerance = 1e-8)
  expect_equal(out$waist_to_height_ratio, c(90 / 175, 100 / 165), tolerance = 1e-8)

  # waist_to_BMI_ratio and weight_to_height_ratio
  expect_equal(out$waist_to_BMI_ratio, c(90 / (80 / 1.75^2), 100 / (90 / 1.65^2)), tolerance = 1e-6)
  expect_equal(out$weight_to_height_ratio, c(80 / 1.75, 90 / 1.65), tolerance = 1e-8)

  # Advanced indices produce finite numerics
  expect_true(all(is.finite(out$AVI)))
  expect_true(all(is.finite(out$BAI)))
  expect_true(all(is.finite(out$ABSI)))
  expect_true(all(is.finite(out$BRI)))
  expect_true(all(is.finite(out$CI)))
})

# 2) Test adjust_WHR and include_RFM
test_that("adjust_WHR adds WHRadjBMI and include_RFM adds RFM", {
  skip_on_cran()
  out2 <- obesity_indices(
    data = base_df,
    weight = wt,
    height = ht,
    waist = waist,
    hip = hip,
    sex = sex,
    weight_unit = "kg",
    height_unit = "m",
    adjust_WHR = TRUE,
    include_RFM = TRUE
  )
  # WHRadjBMI exists and numeric
  expect_true("WHRadjBMI" %in% names(out2))
  expect_type(out2$WHRadjBMI, "double")
  # Residuals sum to zero (by construction for linear model residuals)
  expect_equal(sum(out2$WHRadjBMI), 0)

  # RFM computation: 64 - 20*(height_m / (waist_cm/100)) + 12*sex
  expected_rfm <- c(
    64 - 20 * (1.75 / (90 / 100)) + 12 * 0,
    64 - 20 * (1.65 / (100 / 100)) + 12 * 1
  )
  expect_equal(out2$RFM, expected_rfm, tolerance = 1e-8)
})

# 3) Units conversion
test_that("unit conversion works for lb and cm", {
  skip_on_cran()
  df_units <- tibble(
    wt_lb = 220,    # lb
    ht_cm = 180,    # cm
    waist = 90,     # cm-like numeric
    hip   = 100     # cm-like numeric
  )
  out <- obesity_indices(
    data = df_units,
    weight = wt_lb,
    height = ht_cm,
    waist = waist,
    hip = hip,
    weight_unit = "lb",
    height_unit = "cm",
    adjust_WHR = FALSE,
    include_RFM = FALSE
  )
  expect_equal(out$weight_kg, 220 * 0.45359237, tolerance = 1e-8)
  expect_equal(out$height_m, 180 / 100, tolerance = 1e-8)
  expect_equal(out$WHR, 90 / 100, tolerance = 1e-8)
})

# 4) Error handling
test_that("errors on missing columns or missing sex when RFM requested", {
  skip_on_cran()
  # Missing waist column
  df_missing <- base_df[, setdiff(names(base_df), "waist")]
  expect_error(
    obesity_indices(df_missing, wt, ht, waist, hip),
    "missing required columns"
  )

  # include_RFM = TRUE without sex argument
  expect_error(
    obesity_indices(
      data = base_df,
      weight = wt,
      height = ht,
      waist = waist,
      hip = hip,
      include_RFM = TRUE
    ),
    "must be provided to compute RFM"
  )
})

# 5) na_action policies
test_that("na_action policies behave as expected", {
  skip_on_cran()
  df_na <- tibble(
    wt    = c(80, NA_real_),
    ht    = c(2, 2),
    waist = c(1, 2),
    hip   = c(1, 1)
  )
  # error -> abort (suppress high-missingness warning)
  expect_error(
    suppressWarnings(
      obesity_indices(df_na, wt, ht, waist, hip,
                      weight_unit = "kg", height_unit = "m",
                      na_action = "error")
    ),
    "required inputs contain missing values"
  )
  # omit -> drop rows with NA
  out_omit <- suppressWarnings(
    obesity_indices(df_na, wt, ht, waist, hip,
                    weight_unit = "kg", height_unit = "m",
                    na_action = "omit")
  )
  expect_equal(nrow(out_omit), 1L)
  expect_equal(out_omit$BMI, 80 / (2^2))

  # keep -> NA propagates into BMI
  out_keep <- suppressWarnings(
    obesity_indices(df_na, wt, ht, waist, hip,
                    weight_unit = "kg", height_unit = "m",
                    na_action = "keep")
  )
  expect_true(is.na(out_keep$BMI[2]))
})

# 6) Extreme values pass through unchanged (no check_extreme)
test_that("extreme input values pass through without error", {
  skip_on_cran()
  df_ext <- tibble(
    wt    = 80,
    ht    = 200,
    waist = 500,
    hip   = 100
  )
  out <- obesity_indices(df_ext, wt, ht, waist, hip, height_unit = "m")
  expect_true(is.finite(out$WHR))
})

# 7) Denominator-zero summary warning
test_that("denominator zero emits a single summary warning", {
  skip_on_cran()
  df_zero <- tibble(
    wt    = 80,
    ht    = 2,
    waist = 1,
    hip   = 0  # zero denominator for WHR
  )
  expect_warning(
    out <- obesity_indices(df_zero, wt, ht, waist, hip),
    "zero denominators detected"
  )
  expect_true(is.na(out$WHR))
})

# 8) include_RFM with invalid sex values warns and sets NA
test_that("include_RFM warns on invalid sex values and sets NA", {
  skip_on_cran()
  df_bad_sex <- tibble(
    wt    = c(80, 70),
    ht    = c(2, 1.8),
    waist = c(1, 0.9),
    hip   = c(1, 1.0),
    sex   = c(2, -1)  # invalid codes
  )
  expect_warning(
    out <- obesity_indices(df_bad_sex, wt, ht, waist, hip, sex, include_RFM = TRUE),
    "sex' contains .* values not in \\{0,1\\}"
  )
  expect_true(all(is.na(out$RFM)))
})

# 9) Verbose emits col_map and results messages
test_that("verbose emits preparing, col_map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    obesity_indices(base_df, wt, ht, waist, hip, verbose = TRUE),
    "obesity_indices"
  )
  expect_message(
    obesity_indices(base_df, wt, ht, waist, hip, verbose = TRUE),
    "col_map"
  )
  expect_message(
    obesity_indices(base_df, wt, ht, waist, hip, verbose = TRUE),
    "results:"
  )
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(
    obesity_indices(base_df, wt, ht, waist, hip, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Compute inflammatory indices (classic, eosinophil, or both)
#'
#' Panels:
#' - classic: NLR, PLR, LMR, dNLR, SII, SIRI, AISI, CRP_category
#' - eos:     NLR, PLR, LMR, NER, SII, SIRI, PIV, CLR, CAR, PCR, mGPS, ESR (if mapped)
#' - both:    union of classic and eos panels
#'
#' Derived markers:
#' - NLR  = neutrophils / lymphocytes
#' - PLR  = platelets / lymphocytes
#' - LMR  = lymphocytes / monocytes
#' - dNLR = neutrophils / (WBC - neutrophils) when WBC available
#' - SII  = platelets * neutrophils / lymphocytes
#' - SIRI = neutrophils * monocytes / lymphocytes
#' - AISI = neutrophils * monocytes * platelets / lymphocytes
#' - CRP_category: "low" (<1 mg/L), "moderate" (1-3 mg/L), "high" (>3 mg/L) when CRP available
#' - Eosinophil-panel extras: NER = neutrophils / eosinophils; PIV = platelets * neutrophils * monocytes / lymphocytes;
#'   CLR = CRP/lymphocytes; CAR = CRP/albumin; PCR = platelets/CRP; mGPS (CRP, albumin); ESR passthrough.
#'
#' Note:
#' - These outputs are deterministic algebraic indices computed from the mapped
#'   laboratory variables. They are intended for feature engineering and descriptive
#'   analyses, not as standalone diagnosis/prognosis tools.
#' - References below document commonly used index definitions or interpretation
#'   conventions directly used in this implementation.
#'
#' @param data data.frame or tibble
#' @param col_map named list mapping keys to column names in `data`.
#'   Keys: neutrophils, lymphocytes, monocytes, platelets, WBC, CRP, albumin, eosinophils, ESR.
#' @param panel one of c("auto","classic","eos","both"). "auto" uses presence of eosinophils key.
#' @param na_action one of c("keep","omit","error"). Default "keep" propagates NA
#'   in outputs where inputs are missing.
#' @param verbose logical; if `TRUE` (default), prints column mapping, optional input
#'   availability, physiological range information (informational only), the list of
#'   markers being computed, and a results summary.
#' @return tibble with selected inflammatory indices, with ID column prepended if
#'   detected (e.g. `id`, `IID`, `participant_id`).
#'
#' @examples
#' # Quick smoke-test
#' df <- data.frame(neutrophils = 4, lymphocytes = 2, monocytes = 0.5,
#'                  platelets = 200, WBC = 7, CRP = 2.5)
#' inflammatory_markers(df, panel = "classic", na_action = "keep", verbose = FALSE)
#'
#' \donttest{
#' df <- data.frame(
#'   neutrophils = c(4, 2),
#'   lymphocytes = c(2, 0),
#'   monocytes   = c(0.5, 0.3),
#'   platelets   = c(200, 150),
#'   WBC         = c(7, 4.5),
#'   CRP         = c(2.5, 0.8),
#'   albumin     = c(40, 42),
#'   eosinophils = c(0.2, 0.1),
#'   ESR         = c(12, 15)
#' )
#' cm <- list(
#'   neutrophils = "neutrophils", lymphocytes = "lymphocytes", monocytes = "monocytes",
#'   platelets = "platelets", WBC = "WBC", CRP = "CRP", albumin = "albumin",
#'   eosinophils = "eosinophils", ESR = "ESR"
#' )
#' classic_cm <- cm; classic_cm$eosinophils <- NULL; classic_cm$ESR <- NULL
#' inflammatory_markers(df, classic_cm, panel = "classic", na_action = "keep")
#' inflammatory_markers(df, cm, panel = "eos", na_action = "keep", verbose = TRUE)
#' }
#'
#' @references
#' \insertRef{zahorec2001}{HealthMarkers}
#' \insertRef{hu2014sii}{HealthMarkers}
#' \insertRef{qi2016siri}{HealthMarkers}
#' \insertRef{proctor2011mgps}{HealthMarkers}
#' \insertRef{pearson2003markers}{HealthMarkers}
#' @export
inflammatory_markers <- function(data, col_map = NULL,
                                 panel      = c("auto", "classic", "eos", "both"),
                                 na_action  = c("keep", "omit", "error"),
                                 verbose    = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "inflammatory_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  panel     <- match.arg(panel)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data)) {
    rlang::abort(sprintf("%s(): `data` must be a data.frame or tibble.", fn_name),
                 class = "healthmarkers_inflammatory_markers_error_data_type")
  }

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  all_im_keys <- c("neutrophils","lymphocytes","monocytes","platelets","WBC",
    "CRP","albumin","eosinophils","ESR")
  cm      <- .hm_build_col_map(data, col_map, all_im_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map) || !is.list(col_map)) col_map <- list()

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  if (!is.list(col_map) || (!is.null(names(col_map)) && any(!nzchar(names(col_map))))) {
    rlang::abort("inflammatory_markers(): `col_map` must be a named list.",
                 class = "healthmarkers_inflammatory_markers_error_map_type")
  }

  has_eos_key <- "eosinophils" %in% names(col_map)
  if (panel == "auto") panel <- if (has_eos_key) "eos" else "classic"

  req_keys <- switch(panel,
    classic = c("neutrophils","lymphocytes"),
    eos     = c("neutrophils","lymphocytes","monocytes","platelets","CRP"),
    both    = c("neutrophils","lymphocytes"),
    c("neutrophils","lymphocytes")
  )
  missing_keys <- setdiff(req_keys, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(
      paste0("inflammatory_markers(): missing col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_inflammatory_markers_error_missing_map"
    )
  }
  is_empty_map <- function(v) {
    is.null(v) || length(v) == 0L ||
      (is.atomic(v) && length(v) == 1L && (is.na(v) || identical(v, "") || !nzchar(as.character(v))))
  }
  empty_keys <- names(Filter(is_empty_map, col_map))
  if (length(empty_keys)) {
    rlang::abort(
      paste0("inflammatory_markers(): `col_map` has empty mapping for: ",
             paste(empty_keys, collapse = ", ")),
      class = "healthmarkers_inflammatory_markers_error_missing_map"
    )
  }
  if (panel == "eos" && ("CRP" %in% names(col_map)) && !("albumin" %in% names(col_map))) {
    rlang::abort("inflammatory_markers(): missing col_map entries for: albumin",
                 class = "healthmarkers_inflammatory_markers_error_missing_map")
  }

  # --- Verbose: optional inputs
  if (isTRUE(verbose)) {
    all_keys  <- c("neutrophils","lymphocytes","monocytes","platelets",
                   "WBC","CRP","albumin","eosinophils","ESR")
    get_col_v <- function(key) if (key %in% names(col_map)) as.character(col_map[[key]])[1] else NA_character_
    has_col_v <- function(k) { cn <- get_col_v(k); is.character(cn) && !is.na(cn) && cn %in% names(data) }
    present_all <- all_keys[vapply(all_keys, has_col_v, logical(1))]
    missing_all <- setdiff(all_keys, present_all)
    lines <- sprintf("%s(): optional inputs (panel = %s)", fn_name, panel)
    if (length(present_all))
      lines <- c(lines, sprintf("  present:  %s", paste(present_all, collapse = ", ")))
    if (length(missing_all))
      lines <- c(lines, sprintf("  missing:  %s", paste(missing_all, collapse = ", ")))
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  supported_keys <- c("neutrophils","lymphocytes","monocytes","platelets","WBC","CRP","albumin","eosinophils","ESR")
  get_col <- function(key) if (key %in% names(col_map)) as.character(col_map[[key]])[1] else NA_character_
  map_cols <- vapply(supported_keys, get_col, character(1)); names(map_cols) <- supported_keys
  has_col <- function(cn) is.character(cn) && length(cn) == 1L && !is.na(cn) && cn %in% names(data)
  avail <- vapply(map_cols, has_col, logical(1))

  # Coerce available columns to numeric; non-finite -> NA
  for (k in names(avail)[avail]) {
    cn <- map_cols[[k]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("inflammatory_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    limits_info <- list(
      neutrophils = c(0, 30), lymphocytes = c(0.2, 15), monocytes = c(0, 5),
      eosinophils = c(0, 3),  platelets   = c(0, 1000), WBC = c(0, 50),
      CRP = c(0, 300),        albumin     = c(10, 60)
    )
    ex_details <- character(0)
    for (k in names(limits_info)) {
      if (!avail[[k]]) next
      x   <- data[[map_cols[[k]]]]
      rng <- limits_info[[k]]
      nb  <- sum(is.finite(x) & (x < rng[1] | x > rng[2]), na.rm = TRUE)
      if (nb > 0L)
        ex_details <- c(ex_details, sprintf("  %s: %d value(s) outside plausible range", k, nb))
    }
    if (length(ex_details))
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                ex_details), collapse = "\n"),
        level = "inform"
      )
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    panel_markers <- switch(panel,
      classic = "NLR, PLR, LMR, dNLR, SII, SIRI, AISI, CRP_category",
      eos     = "NLR, PLR, LMR, NER, SII, SIRI, PIV, CLR, CAR, PCR, mGPS, ESR",
      both    = "NLR, PLR, LMR, NER, SII, SIRI, PIV, CLR, CAR, PCR, mGPS, ESR, dNLR, AISI, CRP_category"
    )
    hm_inform(
      sprintf("%s(): computing markers: %s", fn_name, panel_markers),
      level = "inform"
    )
 }

  g <- function(key) data[[map_cols[[key]]]]
  n <- nrow(data)
  out <- tibble::tibble(.rows = n)
  dz_count <- 0L
  safe_div <- function(num, den) {
    res <- num / den
    dz_count <<- dz_count + sum(is.finite(den) & den == 0, na.rm = TRUE)
    res
  }
  used_keys <- character(0)

  add_classic <- function() {
    out$NLR  <<- if (all(avail[c("neutrophils","lymphocytes")])) safe_div(g("neutrophils"), g("lymphocytes")) else rep(NA_real_, n)
    out$PLR  <<- if (all(avail[c("platelets","lymphocytes")]))   safe_div(g("platelets"), g("lymphocytes"))   else rep(NA_real_, n)
    out$LMR  <<- if (all(avail[c("lymphocytes","monocytes")]))   safe_div(g("lymphocytes"), g("monocytes"))   else rep(NA_real_, n)
    out$dNLR <<- if (all(avail[c("neutrophils","WBC")]))         safe_div(g("neutrophils"), (g("WBC") - g("neutrophils"))) else rep(NA_real_, n)
    out$SII  <<- if (all(avail[c("platelets","neutrophils","lymphocytes")])) safe_div(g("platelets") * g("neutrophils"), g("lymphocytes")) else rep(NA_real_, n)
    out$SIRI <<- if (all(avail[c("neutrophils","monocytes","lymphocytes")])) safe_div(g("neutrophils") * g("monocytes"), g("lymphocytes")) else rep(NA_real_, n)
    out$AISI <<- if (all(avail[c("neutrophils","monocytes","platelets","lymphocytes")])) safe_div(g("neutrophils") * g("monocytes") * g("platelets"), g("lymphocytes")) else rep(NA_real_, n)
    used_keys <<- union(used_keys, c("neutrophils","lymphocytes","platelets","monocytes","WBC"))
    crp_cat <- rep(NA_character_, n)
    if (avail[["CRP"]]) {
      crp <- g("CRP")
      crp_cat <- ifelse(is.na(crp), NA_character_,
                 ifelse(crp < 1, "low", ifelse(crp <= 3, "moderate", "high")))
      used_keys <<- union(used_keys, "CRP")
    }
    out$CRP_category <<- factor(crp_cat, levels = c("low","moderate","high"), ordered = TRUE)
  }

  add_eos <- function() {
    out$NLR  <<- if (all(avail[c("neutrophils","lymphocytes")])) safe_div(g("neutrophils"), g("lymphocytes")) else rep(NA_real_, n)
    out$PLR  <<- if (all(avail[c("platelets","lymphocytes")]))   safe_div(g("platelets"), g("lymphocytes"))   else rep(NA_real_, n)
    out$LMR  <<- if (all(avail[c("lymphocytes","monocytes")]))   safe_div(g("lymphocytes"), g("monocytes"))   else rep(NA_real_, n)
    out$NER  <<- if (all(avail[c("neutrophils","eosinophils")])) safe_div(g("neutrophils"), g("eosinophils")) else rep(NA_real_, n)
    out$SII  <<- if (all(avail[c("platelets","neutrophils","lymphocytes")])) safe_div(g("platelets") * g("neutrophils"), g("lymphocytes")) else rep(NA_real_, n)
    out$SIRI <<- if (all(avail[c("neutrophils","monocytes","lymphocytes")])) safe_div(g("neutrophils") * g("monocytes"), g("lymphocytes")) else rep(NA_real_, n)
    out$PIV  <<- if (all(avail[c("platelets","neutrophils","monocytes","lymphocytes")])) safe_div(g("platelets") * g("neutrophils") * g("monocytes"), g("lymphocytes")) else rep(NA_real_, n)
    out$CLR  <<- if (all(avail[c("CRP","lymphocytes")]))         safe_div(g("CRP"), g("lymphocytes")) else rep(NA_real_, n)
    out$CAR  <<- if (all(avail[c("CRP","albumin")]))             safe_div(g("CRP"), g("albumin")) else rep(NA_real_, n)
    out$PCR  <<- if (all(avail[c("platelets","CRP")]))           safe_div(g("platelets"), g("CRP")) else rep(NA_real_, n)
    out$mGPS <<- if (all(avail[c("CRP","albumin")])) {
                   crp <- g("CRP"); alb <- g("albumin")
                   as.integer(ifelse(is.na(crp), NA_integer_,
                              ifelse(crp <= 10, 0L, ifelse(alb >= 35, 1L, 2L))))
                 } else rep(NA_integer_, n)
    if (avail[["ESR"]]) out$ESR <<- as.numeric(g("ESR"))
    used_keys <<- union(used_keys, c("neutrophils","lymphocytes","platelets","monocytes","eosinophils","CRP","albumin","ESR"))
  }

  if (panel == "classic") {
    add_classic()
    out <- out[, c("NLR","PLR","LMR","dNLR","SII","SIRI","AISI","CRP_category"), drop = FALSE]
  } else if (panel == "eos") {
    add_eos()
    ord <- c("NLR","PLR","LMR","NER","SII","SIRI","PIV","CLR","CAR","PCR","mGPS", if ("ESR" %in% names(out)) "ESR")
    out <- out[, ord, drop = FALSE]
  } else { # both
    add_eos(); add_classic()
    ord <- c("NLR","PLR","LMR","NER","SII","SIRI","PIV","CLR","CAR","PCR","mGPS","ESR","dNLR","AISI","CRP_category")
    ord <- intersect(ord, names(out))
    out <- out[, ord, drop = FALSE]
  }

  # NA policy over used inputs
  used_cols <- map_cols[intersect(unique(used_keys), names(map_cols))]
  used_cols <- used_cols[!is.na(used_cols)]
  cc <- if (length(used_cols)) stats::complete.cases(data[, used_cols, drop = FALSE]) else rep(TRUE, n)

  na_action_eff <- na_action
  if (na_action_eff == "error" && any(!cc)) {
    rlang::abort("inflammatory_markers(): required columns contain missing or non-finite values (na_action='error').",
                 class = "healthmarkers_inflammatory_markers_error_missing_values")
  } else if (na_action_eff == "omit") {
    out <- out[cc, , drop = FALSE]
  }

  if (dz_count > 0L) rlang::warn("inflammatory_markers(): zero denominators detected.")
  out <- tibble::as_tibble(out)

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }
  out
}

# HM-CS v2 compatible tests for inflammatory_markers()

test_that("default na_action is 'keep' (not 'error')", {
  skip_on_cran()
  df <- tibble::tibble(neutrophils = c(4, NA), lymphocytes = c(2, 1))
  cm <- list(neutrophils = "neutrophils", lymphocytes = "lymphocytes")
  out <- inflammatory_markers(df, cm, panel = "classic", verbose = FALSE)
  expect_equal(nrow(out), 2L)
  expect_true(is.na(out$NLR[2]))
})

test_that("mapping validation errors and messages", {
  skip_on_cran()
  df <- tibble::tibble(neutrophils = numeric())
  expect_error(inflammatory_markers("x", list()), "data.frame or tibble")
  expect_error(inflammatory_markers(df, list(1)), "named list")
  # classic requires neutrophils + lymphocytes
  expect_error(
    inflammatory_markers(df, list(neutrophils = "neutrophils"), panel = "classic"),
    "missing col_map entries for: lymphocytes"
  )
  # eos requires albumin name if CRP name present
  df2 <- tibble::tibble(neutrophils = 1, lymphocytes = 1, monocytes = 1, platelets = 1, CRP = 1)
  expect_error(
    inflammatory_markers(df2, list(neutrophils="neutrophils", lymphocytes="lymphocytes",
                                   monocytes="monocytes", platelets="platelets", CRP="CRP"),
                         panel = "eos"),
    "missing col_map entries for: albumin"
  )
})

test_that("verbose emits col_map, optional inputs, computing markers, and results messages", {
  skip_on_cran()
  df_v <- tibble::tibble(neutrophils = numeric(0), lymphocytes = numeric(0))
  cm_v <- list(neutrophils = "neutrophils", lymphocytes = "lymphocytes")
  msgs <- testthat::capture_messages(
    inflammatory_markers(df_v, cm_v, panel = "classic", verbose = TRUE)
  )
  expect_true(any(grepl("col_map", msgs)))
  expect_true(any(grepl("optional inputs", msgs)))
  expect_true(any(grepl("computing markers", msgs)))
  expect_true(any(grepl("results:", msgs)))
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df_v <- tibble::tibble(neutrophils = numeric(0), lymphocytes = numeric(0))
  cm_v <- list(neutrophils = "neutrophils", lymphocytes = "lymphocytes")
  msgs <- testthat::capture_messages(
    inflammatory_markers(df_v, cm_v, panel = "classic", verbose = TRUE)
  )
  expect_equal(sum(grepl("col_map",   msgs)), 1L)
  expect_equal(sum(grepl("results:",          msgs)), 1L)
  expect_equal(sum(grepl("optional inputs",   msgs)), 1L)
  expect_equal(sum(grepl("computing markers", msgs)), 1L)
})

test_that("classic panel computes markers", {
  skip_on_cran()
  df <- tibble::tibble(
    neutrophils = 4, lymphocytes = 2, monocytes = 0.5, platelets = 200, WBC = 7, CRP = 2.5
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- inflammatory_markers(df, cm, panel = "classic", na_action = "keep", verbose = FALSE)
  expect_equal(names(out), c("NLR","PLR","LMR","dNLR","SII","SIRI","AISI","CRP_category"))
  expect_equal(out$NLR, 4/2)
  expect_equal(out$PLR, 200/2)
  expect_equal(out$dNLR, 4/(7-4))
  expect_equal(as.character(out$CRP_category), "moderate")
})

test_that("eosinophil panel computes markers and ESR passthrough", {
  skip_on_cran()
  df <- tibble::tibble(
    neutrophils = 4, lymphocytes = 2, monocytes = 0.5, platelets = 200,
    CRP = 5, albumin = 40, eosinophils = 0.2, ESR = 12
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- inflammatory_markers(df, cm, panel = "eos", na_action = "keep", verbose = FALSE)
  expect_equal(names(out), c("NLR","PLR","LMR","NER","SII","SIRI","PIV","CLR","CAR","PCR","mGPS","ESR"))
  expect_equal(out$NLR, 4/2)
  expect_equal(out$CAR, 5/40)
  expect_equal(out$mGPS, 0L)
  expect_equal(out$ESR, 12)
})

test_that("na_action policies", {
  skip_on_cran()
  df <- tibble::tibble(neutrophils = c(2, NA), lymphocytes = c(2, 1))
  cm <- list(neutrophils="neutrophils", lymphocytes="lymphocytes")
  out_keep <- inflammatory_markers(df, cm, panel = "classic", na_action = "keep", verbose = FALSE)
  expect_equal(nrow(out_keep), 2L)
  expect_true(is.na(out_keep$NLR[2]))
  out_omit <- inflammatory_markers(df, cm, panel = "classic", na_action = "omit", verbose = FALSE)
  expect_equal(nrow(out_omit), 1L)
  expect_equal(out_omit$NLR, 1)
  expect_error(inflammatory_markers(df, cm, panel = "classic", na_action = "error", verbose = FALSE),
               "missing or non-finite")
})

test_that("extreme values produce no warning/error; range note appears in verbose", {
  skip_on_cran()
  df <- tibble::tibble(
    neutrophils = 40, lymphocytes = 0.2, monocytes = 6, platelets = 5000,
    CRP = 500, albumin = 5, eosinophils = 0.1, ESR = 99
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_no_warning(
    out <- inflammatory_markers(df, cm, panel = "eos", na_action = "keep", verbose = FALSE)
  )
  expect_true(is.numeric(out$NLR))
  # Range note appears in verbose
  msgs <- testthat::capture_messages(
    inflammatory_markers(df, cm, panel = "eos", na_action = "keep", verbose = TRUE)
  )
  expect_true(any(grepl("range note", msgs)))
})

test_that("zero denominators warn and keep Inf/NaN", {
  skip_on_cran()
  df <- tibble::tibble(neutrophils = 2, lymphocytes = 0, platelets = 10, monocytes = 0, WBC = 2)
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_warning(out <- inflammatory_markers(df, cm, panel = "classic", na_action = "keep", verbose = FALSE),
                 "zero denominators")
  expect_true(is.infinite(out$NLR))
  expect_true(is.infinite(out$PLR))
})

test_that("ID column is prepended to output when detected", {
  skip_on_cran()
  df_id <- tibble::tibble(
    id = 1:3, neutrophils = c(4, 3, 5), lymphocytes = c(2, 1.5, 2.5)
  )
  cm_id <- list(neutrophils = "neutrophils", lymphocytes = "lymphocytes")
  out <- inflammatory_markers(df_id, cm_id, panel = "classic", verbose = FALSE)
  expect_equal(names(out)[1L], "id")
  expect_equal(out$id, 1:3)
})

# List ALL optional (Suggests) namespaces in one place
.optional_deps <- c(
  "CVrisk", "PooledCohort", "QRISK3", "RiskScorescvd",
  "di"
)

# Require exactly one package (runtime gate)
# Emits a typed condition so .hm_safe_call()'s handle_error can read e$package
# without needing regex on the message string.
.need_pkg <- function(pkg) {
  ok <- suppressMessages(suppressWarnings(requireNamespace(pkg, quietly = TRUE)))
  if (!ok) {
    rlang::abort(
      sprintf("Package '%s' is required for this feature. Install it first.", pkg),
      class = "healthmarkers_missing_package",
      package = pkg
    )
  }
}

# Require a set of packages (useful for multi-dep features)
.need_pkgs <- function(pkgs) {
  miss <- pkgs[!vapply(pkgs, requireNamespace, FUN.VALUE = TRUE, quietly = TRUE)]
  if (length(miss)) {
    rlang::abort(sprintf(
      "Missing required package(s): %s. Install them first.",
      paste(miss, collapse = ", ")
    ))
  }
}

# Run an expression only if a package is available; otherwise return NA or a fallback
.with_pkg <- function(pkg, expr, otherwise = NULL) {
  if (requireNamespace(pkg, quietly = TRUE)) {
    force(expr)
  } else {
    if (!is.null(otherwise)) otherwise else rlang::abort(
      sprintf("Optional package '%s' not installed.", pkg)
    )
  }
}

.onLoad <- function(libname, pkgname) {
  # Initialize verbosity option if unset
  if (is.null(getOption("healthmarkers.verbose"))) {
    # allowed: "none" | "inform" | "debug"; TRUE treated as "inform"
    options(healthmarkers.verbose = "none")
  }
  invisible(TRUE)
}

utils::globalVariables(c(
  ":=",
  "chol_total","chol_ldl","chol_hdl","triglycerides","age_year","z_HOMA",
  "glucose","HbA1c","bp_sys_z","bp_dia_z","weight_kg","height_m",
  "BMI","Avignon_Si0","Avignon_Si120"
))

#' Compute atherogenic indices
#'
#' Calculates:
#' - AIP: Atherogenic Index of Plasma = log10(TG / HDL_c)
#' - CRI_I: Castelli Risk Index I = TC / HDL_c
#' - CRI_II: Castelli Risk Index II = LDL_c / HDL_c
#'
#' Behavior:
#' - Required keys: TG, HDL_c. Optional: TC, LDL_c.
#' - NA policy via `na_action`: "keep" (default), "omit" (drop rows with any NA in used lipids), "error".
#' - Extreme screening via `check_extreme` and `extreme_action` ("warn","cap","error","ignore","NA").
#'   Default bounds (mg/dL) used only for screening: TG (0, 10000), HDL_c (0, 1000), LDL_c (0, 10000), TC (0, 10000).
#'   Note: All indices are unitless ratios; units cancel in computations.
#' - Emits progress via `hm_inform()` when `verbose = TRUE` or when package option enables logs.
#'
#' @param data data.frame/tibble with lipid columns.
#' @param col_map named list mapping keys to columns, e.g. list(TG="TG", HDL_c="HDL_c", TC="TC", LDL_c="LDL_c").
#' @param na_action one of c("keep","omit","error").
#' @param normalize one of c("none","log10"). Reserved; AIP always uses log10(TG/HDL_c).
#' @param verbose Logical; if `TRUE` (default), prints column mapping, the list
#'   of indices being computed, and a per-column results summary.
#'
#' @return tibble with columns AIP, CRI_I, CRI_II. If an ID column is detected
#'   in `data` (e.g. `id`, `IID`, `participant_id`), it is prepended.
#'
#' @references
#' \insertRef{dobiasova2004aip}{HealthMarkers}
#' \insertRef{castelli1977framingham}{HealthMarkers}
#'
#' @examples
#' df <- tibble::tibble(
#'   TG = c(150, 200),
#'   HDL_c = c(50, 40),
#'   TC = c(200, 220),
#'   LDL_c = c(120, 150)
#' )
#' cm <- list(TG = "TG", HDL_c = "HDL_c", TC = "TC", LDL_c = "LDL_c")
#' atherogenic_indices(df, col_map = cm)
#'
#' @export
atherogenic_indices <- function(data,
                                col_map = NULL,
                                na_action = c("keep","omit","error"),
                                normalize = c("none","log10"),
                                verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "atherogenic_indices"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  id_col    <- .hm_detect_id_col(data)

  allowed_norm <- c("none","log10")
  if (length(normalize) == 0L) {
    rlang::abort("`normalize` must be one of: 'none', 'log10'",
                 class = "healthmarkers_atherogenic_indices_error_normalize")
  }
  normalize <- normalize[1L]
  if (!(normalize %in% allowed_norm)) {
    rlang::abort("`normalize` must be one of: 'none', 'log10'",
                 class = "healthmarkers_atherogenic_indices_error_normalize")
  }

  req <- c("TG", "HDL_c")
  opt <- c("TC", "LDL_c")

  cm      <- .hm_build_col_map(data, col_map, c(req, opt), fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  hm_validate_inputs(data, col_map, required_keys = req, fn = "atherogenic_indices")

  # Required columns presence
  req_cols <- unname(unlist(col_map[req], use.names = FALSE))

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    indices <- list(
      c("AIP",    "log10(TG / HDL_c)"),
      c("CRI_I",  "TC / HDL_c [if TC available]"),
      c("CRI_II", "LDL_c / HDL_c [if LDL_c available]")
    )
    status <- vapply(indices, function(x) sprintf("  %-10s [%s]", x[1], x[2]), character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }
  missing_req <- setdiff(req_cols, names(data))
  if (length(missing_req)) {
    rlang::abort(
      paste0("atherogenic_indices(): missing required columns in data: ", paste(missing_req, collapse = ", ")),
      class = "healthmarkers_atherogenic_indices_error_missing_columns"
    )
  }

  # Optional columns present in data
  opt_cols <- unname(unlist(col_map[intersect(opt, names(col_map))], use.names = FALSE))
  opt_cols <- intersect(opt_cols, names(data))

  # Coerce used columns to numeric; NAs introduced are warned
  used_cols <- unique(c(req_cols, opt_cols))
  for (cn in used_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("atherogenic_indices(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced),
                    class = "healthmarkers_atherogenic_indices_warn_na_coercion")
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # NA handling across all used lipids
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("atherogenic_indices(): missing values in required inputs (na_action='error').",
                   class = "healthmarkers_atherogenic_indices_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    data <- data[keep, , drop = FALSE]
  }

  # Early empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(AIP = numeric(), CRI_I = numeric(), CRI_II = numeric()))
  }

  hm_inform(level = "debug", msg = "atherogenic_indices(): computing indices")

  # Accessor and safe division
  g <- function(key) data[[col_map[[key]]]]
  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }

  # Compute indices
  AIP   <- log10(safe_div(g("TG"), g("HDL_c"), "AIP_denHDL"))
  CRI_I <- if ("TC" %in% names(col_map) && col_map[["TC"]] %in% names(data)) {
    safe_div(g("TC"), g("HDL_c"), "CRI_I_denHDL")
  } else rep(NA_real_, nrow(data))
  CRI_II <- if ("LDL_c" %in% names(col_map) && col_map[["LDL_c"]] %in% names(data)) {
    safe_div(g("LDL_c"), g("HDL_c"), "CRI_II_denHDL")
  } else rep(NA_real_, nrow(data))

  out <- tibble::tibble(
    AIP = as.numeric(AIP),
    CRI_I = as.numeric(CRI_I),
    CRI_II = as.numeric(CRI_II)
  )

  # Zero denominators warning
  dz <- dz_env$counts
  if (length(dz)) {
    dz_total <- sum(unlist(dz), na.rm = TRUE)
    if (dz_total > 0L) {
      nz <- unlist(dz); nz <- nz[nz > 0]
      lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
      rlang::warn(sprintf("atherogenic_indices(): zero denominators detected in %d cases (%s).", dz_total, lbl),
                  class = "healthmarkers_atherogenic_indices_warn_zero_denominator")
    }
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  hm_inform(
    level = if (isTRUE(verbose)) "inform" else "debug",
    msg   = hm_result_summary(out, fn_name)
  )

  out
}

library(testthat)
library(tibble)
library(HealthMarkers)

cm <- list(TC = "TC", LDL_c = "LDL_c", HDL_c = "HDL_c", TG = "TG")

test_that("atherogenic_indices computes AIP, CRI_I, CRI_II", {
  skip_on_cran()
  dat <- tibble(TC = c(200, 180), HDL_c = c(50, 60), TG = c(150, 100), LDL_c = c(120, 90))
  res <- atherogenic_indices(dat, col_map = cm)
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("AIP","CRI_I","CRI_II"))
  exp_AIP <- log10(dat$TG / dat$HDL_c)
  exp_CRI_I <- dat$TC / dat$HDL_c
  exp_CRI_II <- dat$LDL_c / dat$HDL_c
  expect_equal(res$AIP, as.numeric(exp_AIP), tolerance = 1e-12)
  expect_equal(res$CRI_I, as.numeric(exp_CRI_I), tolerance = 1e-12)
  expect_equal(res$CRI_II, as.numeric(exp_CRI_II), tolerance = 1e-12)
})

test_that("na_action omit drops rows with missing lipids", {
  skip_on_cran()
  dat <- tibble(TG = c(150, NA_real_), HDL_c = c(50, 40), TC = c(200, 220), LDL_c = c(120, 150))
  cm <- list(TG = "TG", HDL_c = "HDL_c", TC = "TC", LDL_c = "LDL_c")
  res <- atherogenic_indices(dat, col_map = cm, na_action = "omit")
  expect_equal(nrow(res), 1L)
})

test_that("na_action error aborts on missing", {
  skip_on_cran()
  dat <- tibble(TG = c(150, NA_real_), HDL_c = c(50, 40))
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  expect_error(atherogenic_indices(dat, col_map = cm, na_action = "error"), "missing values")
})

test_that("invalid normalize argument errors early", {
  skip_on_cran()
  dat <- tibble(TG = 150, HDL_c = 50)
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  expect_error(atherogenic_indices(dat, col_map = cm, normalize = "foo"))
})

test_that("AIP computed correctly with large TG values", {
  skip_on_cran()
  dat <- tibble(TG = 20000, HDL_c = 10)
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  res <- atherogenic_indices(dat, col_map = cm, verbose = FALSE)
  expect_true(is.finite(res$AIP))
  expect_equal(res$AIP, log10(20000/10))
})

test_that("missing columns reported clearly", {
  skip_on_cran()
  dat <- tibble(TG = 150)
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  expect_error(atherogenic_indices(dat, col_map = cm), "missing required columns in data: HDL_c")
})

test_that("package-level verbosity emits results summary message", {
  skip_on_cran()
  dat <- tibble(TG = 150, HDL_c = 50)
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    atherogenic_indices(dat, col_map = cm, verbose = TRUE),
    "results:"
  )
})

test_that("verbose = TRUE emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  dat <- tibble(TG = 150, HDL_c = 50, TC = 200, LDL_c = 120)
  cm2 <- list(TG = "TG", HDL_c = "HDL_c", TC = "TC", LDL_c = "LDL_c")
  expect_message(atherogenic_indices(dat, col_map = cm2, verbose = TRUE), "atherogenic_indices")
  expect_message(atherogenic_indices(dat, col_map = cm2, verbose = TRUE), "col_map")
  expect_message(atherogenic_indices(dat, col_map = cm2, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  dat  <- tibble(TG = 150, HDL_c = 50, TC = 200, LDL_c = 120)
  cm2  <- list(TG = "TG", HDL_c = "HDL_c", TC = "TC", LDL_c = "LDL_c")
  msgs <- testthat::capture_messages(
    atherogenic_indices(dat, col_map = cm2, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Calculate standardized scores (SDS) for adiposity measures
#'
#' Computes standard deviation (z) scores for anthropometric variables relative to
#' a single (non-sex-stratified) reference set of means and standard deviations.
#' Includes input validation, optional raw-value extreme screening/capping,
#' configurable handling of extreme SDS values, NA row policies, optional concise
#' summary output, and optional verbose progress messages.
#'
#' SDS are computed as: (observed - mean) / sd. Rows are removed only when
#' `na_action = 'omit'`. Raw-value extreme screening (if enabled) is applied
#' before SDS computation. Extreme SDS handling (cap / warn / error / ignore)
#' is controlled by `extreme_action`. Legacy argument aliases (`na_strategy`,
#' `extreme_strategy`) are soft-deprecated but still accepted.
#'
#' @param data data.frame or tibble containing the measurement columns.
#' @param col_map Optional named list mapping reference variable names to column names
#'   in `data`. If NULL, identity mapping is assumed (names(ref) must be in data).
#' @param ref Named list where each element is a numeric vector with names
#'   `mean` and `sd`, e.g. `list(BMI = c(mean = 23, sd = 4))`.
#' @param na_action One of `c("keep","omit","error")` for row handling when any
#'   mapped variable is missing.
#' @param extreme_action One of `c("cap","NA","error","warn","ignore")` for
#'   SDS values exceeding `sds_cap`.
#' @param sds_cap Positive numeric; absolute cap used when `extreme_action = "cap"`.
#' @param diagnostics Logical; if TRUE emit informational/warning messages
#'   (coercions, missingness, extremes). FALSE suppresses non-critical warnings.
#' @param warn_thresholds Named list with optional elements `na_prop` (default 0.05)
#'   and `extreme_prop` (default 0.01) used for proportion-based diagnostic warnings.
#' @param id_col Optional column name used only in verbose summaries.
#' @param verbose Logical; if TRUE print progress and completion summaries.
#' @param return_summary Logical; if TRUE return a list with elements `data`,
#'   `summary`, and `warnings` instead of just the SDS tibble.
#' @param na_strategy Soft-deprecated alias for `na_action` (if provided and
#'   `na_action` missing, it is used).
#' @param extreme_strategy Soft-deprecated alias for `extreme_action` (if provided
#'   and `extreme_action` missing, it is used).
#'
#' @return A tibble with one `<var>_SDS` column per reference variable, or a
#'   list when `return_summary = TRUE`.
#'
#' @examples
#' ref <- list(BMI = c(mean = 23, sd = 4), waist = c(mean = 80, sd = 12))
#' df <- data.frame(BMI = c(25, NA, 60, 18), waist = c(85, 70, 300, 55))
#' adiposity_sds(df, ref)
#'
#' @export
adiposity_sds <- function(
  data,
  col_map = NULL,
  ref,
  na_action = c("keep","omit","error"),
  extreme_action = c("cap","NA","error","warn","ignore"),
  sds_cap = 6,
  diagnostics = FALSE,  # quiet by default
  warn_thresholds = list(na_prop = 0.05, extreme_prop = 0.01),
  id_col = NULL,
  return_summary = FALSE,
  verbose = TRUE,
  na_strategy = NULL,
  extreme_strategy = NULL
) {

  # --- Back-compat: allow adiposity_sds(data, ref, ...) positional calls ---
  fn_name <- "adiposity_sds"
  looks_like_ref <- function(x) {
    is.list(x) && length(x) > 0 &&
      all(vapply(x, function(y) is.numeric(y) && !is.null(names(y)) &&
                    all(c("mean","sd") %in% names(y)), logical(1)))
  }
  if (missing(ref) && !is.null(col_map) && looks_like_ref(col_map)) {
    ref <- col_map
    col_map <- NULL
  }

  # ---- Legacy alias reconciliation (combine into one warning) ----
  legacy_msgs <- character()

  if (!is.null(na_strategy)) {
    if (!missing(na_action)) {
      legacy_msgs <- c(legacy_msgs, "both na_strategy & na_action supplied; using na_action.")
    } else {
      na_action <- na_strategy
      legacy_msgs <- c(legacy_msgs, "argument 'na_strategy' is deprecated; use 'na_action'.")
    }
  }
  if (!is.null(extreme_strategy)) {
    if (!missing(extreme_action)) {
      legacy_msgs <- c(legacy_msgs, "both extreme_strategy & extreme_action supplied; using extreme_action.")
    } else {
      extreme_action <- extreme_strategy
      legacy_msgs <- c(legacy_msgs, "argument 'extreme_strategy' is deprecated; use 'extreme_action'.")
    }
  }
  if (length(legacy_msgs)) {
    rlang::warn(paste0("adiposity_sds(): ", paste(legacy_msgs, collapse = " ")))
  }

  na_action <- match.arg(na_action, c("keep","omit","error"))
  extreme_action <- match.arg(extreme_action, c("cap","NA","error","warn","ignore"))
  if (extreme_action %in% c("warn","ignore")) {
    rlang::warn("adiposity_sds(): extreme_action values 'warn' and 'ignore' are deprecated; prefer 'cap', 'NA', or 'error'.")
  }

  collected_warnings <- character()

  .warn <- function(msg) {
    if (isTRUE(diagnostics)) rlang::warn(msg)
    collected_warnings <<- c(collected_warnings, msg)
  }
  .abort <- function(msg, class) rlang::abort(msg, class = class)
  .inform <- function(msg) {
    # hm_inform() respects options(healthmarkers.verbose); verbose=TRUE arg has
    # no additional effect here — avoids double-firing when both are active.
    hm_inform(msg, level = "inform")
  }

  # ---- Validate inputs & build effective mapping ----
  if (!is.data.frame(data)) {
    .abort("adiposity_sds(): `data` must be a data.frame or tibble.", "healthmarkers_adiposds_error_data_type")
  }
  if (!is.list(ref) || is.null(names(ref)) || any(!nzchar(names(ref)))) {
    .abort("adiposity_sds(): `ref` must be a named list of mean/sd vectors.", "healthmarkers_adiposds_error_ref_type")
  }
  vars <- names(ref)
  # Validate ref components
  for (v in vars) {
    stats <- ref[[v]]
    if (!(is.numeric(stats) && all(c("mean","sd") %in% names(stats)) && length(stats) >= 2)) {
      .abort(sprintf("adiposity_sds(): ref[['%s']] must be numeric with names mean, sd.", v),
             "healthmarkers_adiposds_error_ref_component")
    }
    if (any(!is.finite(stats[c("mean","sd")])) || stats["sd"] <= 0) {
      .abort(sprintf("adiposity_sds(): ref[['%s']] mean/sd must be finite; sd>0.", v),
             "healthmarkers_adiposds_error_ref_values")
    }
  }

  # Effective mapping: backward-compat (col_map$vars) or flat col_map; infer rest
  flat_col_map <- if (!is.null(col_map) && !is.null(col_map$vars)) col_map$vars else col_map
  cm      <- .hm_build_col_map(data, flat_col_map, keys = vars, fn = fn_name)
  data    <- cm$data
  var_map <- cm$col_map

  # Pre-check for clearer message expected by tests
  missing_vars <- setdiff(vars, names(var_map))
  if (length(missing_vars)) {
    rlang::abort(
      sprintf("adiposity_sds(): missing required columns: %s", paste(missing_vars, collapse = ", ")),
      class = "healthmarkers_adiposds_error_missing_columns"
    )
  }

  if (!is.null(id_col) && !id_col %in% names(data)) {
    .abort(sprintf("adiposity_sds(): id_col '%s' not found in data.", id_col),
           "healthmarkers_adiposds_error_idcol_missing")
  }
  if (!is.numeric(sds_cap) || length(sds_cap) != 1 || !is.finite(sds_cap) || sds_cap <= 0) {
    .abort("adiposity_sds(): `sds_cap` must be a single positive finite number.",
           "healthmarkers_adiposds_error_sds_cap")
  }
  if (!is.list(warn_thresholds)) {
    .abort("adiposity_sds(): `warn_thresholds` must be a list.", "healthmarkers_adiposds_error_warn_thresholds_type")
  }

  total_rows <- nrow(data)
  .hm_log_cols(cm, var_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  %s", fn_name, paste(paste0(vars, "_SDS"), collapse = ", ")), level = "inform")

  # ---- Copy & numeric coercion ----
  df <- data
  for (v in vars) {
    cn <- var_map[[v]]
    if (!is.numeric(df[[cn]])) {
      old <- df[[cn]]
      suppressWarnings(df[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(df[[cn]]) & !is.na(old))
      if (introduced > 0) .warn(sprintf("adiposity_sds(): '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      else .warn(sprintf("adiposity_sds(): '%s' coerced to numeric.", cn))
    }
  }

  # ---- Missingness diagnostics ----
  if (length(vars)) {
    for (v in vars) {
      cn <- var_map[[v]]
      x <- df[[cn]]
      pna <- mean(is.na(x))
      na_thresh <- rlang::`%||%`(warn_thresholds$na_prop, 0.05)
      if (pna >= na_thresh && pna > 0) {
        .warn(sprintf("adiposity_sds(): '%s' missingness %.1f%%", cn, 100 * pna))
      }
    }
  }

  # ---- NA handling (row-wise across mapped columns) ----
  row_na <- if (length(vars)) rowSums(as.data.frame(lapply(var_map, function(cn) is.na(df[[cn]])))) > 0 else rep(FALSE, nrow(df))
  n_rows_in <- nrow(df)
  n_rows_with_na <- sum(row_na)
  if (na_action == "error" && n_rows_with_na > 0) {
    .abort(sprintf("adiposity_sds(): %d rows have missing values (na_action='error').", n_rows_with_na),
           "healthmarkers_adiposds_error_missing_values")
  } else if (na_action == "omit" && n_rows_with_na > 0) {
    df <- df[!row_na, , drop = FALSE]
  }
  n_rows_out <- nrow(df)
  omitted_rows <- n_rows_in - n_rows_out

  if (n_rows_out == 0L) {
    out_empty <- tibble::as_tibble(setNames(replicate(length(vars), numeric(0), simplify = FALSE),
                                            paste0(vars, "_SDS")))
    if (return_summary) {
      return(list(
        data = out_empty,
        summary = list(
          rows_in = n_rows_in,
          rows_out = 0L,
          omitted_rows = omitted_rows,
          total_extreme = 0L,
          per_var = data.frame(variable = vars, n_missing = NA_integer_, n_extreme = NA_integer_, stringsAsFactors = FALSE)
        ),
        warnings = unique(collected_warnings)
      ))
    }
    return(out_empty)
  }

  # ---- Compute SDS over mapped columns ----
  per_var_summary <- data.frame(variable = vars, n_missing = integer(length(vars)), n_extreme = integer(length(vars)), stringsAsFactors = FALSE)
  total_extreme <- 0L
  out_list <- lapply(vars, function(v) {
    stats <- ref[[v]]
    cn <- var_map[[v]]
    x <- df[[cn]]
    sds <- unname((x - stats["mean"]) / stats["sd"])
    per_var_summary$n_missing[per_var_summary$variable == v] <<- sum(is.na(x))
    finite <- is.finite(sds)
    is_extreme_sds <- finite & abs(sds) > sds_cap
    n_ext <- sum(is_extreme_sds, na.rm = TRUE)
    per_var_summary$n_extreme[per_var_summary$variable == v] <<- n_ext
    total_extreme <<- total_extreme + n_ext
    if (n_ext > 0) {
      if (extreme_action == "error") {
        .abort(sprintf("adiposity_sds(): %d SDS beyond +/-%g for '%s'.", n_ext, sds_cap, v),
               "healthmarkers_adiposds_error_sds_extreme")
      } else if (extreme_action == "cap") {
        sds[is_extreme_sds] <- sds_cap * sign(sds[is_extreme_sds])
        .warn(sprintf("adiposity_sds(): capped %d SDS beyond +/-%g for '%s'.", n_ext, sds_cap, v))
      } else if (extreme_action == "warn") {
        .warn(sprintf("adiposity_sds(): %d SDS beyond +/-%g for '%s' (not capped).", n_ext, sds_cap, v))
      } # NA or ignore -> leave as-is
    }
    sds
  })
  names(out_list) <- paste0(vars, "_SDS")
  out <- tibble::as_tibble(out_list)

  # ---- Proportion diagnostics (extreme SDS) ----
  denom_cells <- n_rows_out * length(vars)
  if (denom_cells > 0) {
    prop_extreme <- sum(per_var_summary$n_extreme) / denom_cells
    extreme_prop_thresh <- rlang::`%||%`(warn_thresholds$extreme_prop, 0.01)
    if (prop_extreme > extreme_prop_thresh && sum(per_var_summary$n_extreme) > 0) {
      .warn(sprintf("adiposity_sds(): high extreme SDS proportion %.2f%% (threshold %.2f%%)",
                    100 * prop_extreme, 100 * extreme_prop_thresh))
    }
  }

  # ---- Verbose/package-level completion summary ----
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, "adiposity_sds"), level = "inform")
  }

  # ---- Return ----
  if (isTRUE(return_summary)) {
    return(list(
      data = out,
      summary = list(
        rows_in = n_rows_in,
        rows_out = n_rows_out,
        omitted_rows = omitted_rows,
        total_extreme = total_extreme,
        per_var = per_var_summary
      ),
      warnings = unique(collected_warnings)
    ))
  }

  out
}

# tests/testthat/test_adiposity_sds.R

library(testthat)
library(tibble)

# Shared reference for tests
ref <- list(
  BMI   = c(mean = 23, sd = 4),
  waist = c(mean = 80, sd = 12)
)

test_that("errors if data missing required columns", {
  skip_on_cran()
  df <- tibble(BMI = 23)
  expect_error(
    adiposity_sds(df, ref = ref),
    "missing required columns: waist"
  )
})

test_that("computes correct SDS for single row", {
  skip_on_cran()
  df <- tibble(BMI = 27, waist = 92)
  out <- adiposity_sds(df, ref, diagnostics = FALSE)  # positional ref (back-compat)
  expect_equal(out$BMI_SDS,  (27 - 23) / 4)
  expect_equal(out$waist_SDS, (92 - 80) / 12)
})

test_that("output columns are named <var>_SDS and returns only SDS columns", {
  skip_on_cran()
  df <- tibble(BMI = c(23, 27), waist = c(80, 92))
  out <- adiposity_sds(df, ref = ref)
  expect_named(out, c("BMI_SDS", "waist_SDS"))
  expect_equal(ncol(out), 2L)
  expect_equal(nrow(out), 2L)
})

test_that("SDS formula is (x - mean) / sd for multiple rows", {
  skip_on_cran()
  df <- tibble(BMI = c(19, 23, 31), waist = c(68, 80, 104))
  out <- adiposity_sds(df, ref = ref)
  expect_equal(out$BMI_SDS,   (df$BMI   - 23) / 4,  tolerance = 1e-10)
  expect_equal(out$waist_SDS, (df$waist - 80) / 12, tolerance = 1e-10)
})

test_that("return_summary = TRUE returns list with data, summary, warnings elements", {
  skip_on_cran()
  df <- tibble(BMI = c(25, NA, 21), waist = c(90, 70, 80))
  res <- adiposity_sds(df, ref = ref, return_summary = TRUE, na_action = "omit")
  expect_type(res, "list")
  expect_named(res, c("data", "summary", "warnings"))
  expect_s3_class(res$data, "tbl_df")
  expect_equal(nrow(res$data), 2L)          # 1 row omitted
  expect_equal(res$summary$omitted_rows, 1L)
  expect_equal(res$summary$rows_in, 3L)
  expect_type(res$warnings, "character")
})

test_that("check_extreme removed: legacy alias warns, function passes through", {
  skip_on_cran()
  df <- tibble(BMI = c(120, 22), waist = c(300, 60))
  # check_raw_extreme (legacy alias) now emits no deprecated warning since param removed
  out <- adiposity_sds(df, ref = ref)
  expect_equal(nrow(out), 2L)
})

test_that("computes SDS on identity mapping", {
  skip_on_cran()
  df <- tibble(BMI = c(23, 27), waist = c(80, 92))
  out <- adiposity_sds(df, ref = ref)
  expect_equal(out$BMI_SDS, c(0, (27-23)/4))
  expect_equal(out$waist_SDS, c(0, (92-80)/12))
})

test_that("na_action policies: keep, omit, error", {
  skip_on_cran()
  df <- tibble(BMI = c(25, NA, 21), waist = c(90, 70, 80))
  expect_equal(nrow(adiposity_sds(df, ref = ref, na_action = "keep", diagnostics = FALSE)), 3)
  expect_equal(nrow(adiposity_sds(df, ref = ref, na_action = "omit", diagnostics = FALSE)), 2)
  expect_error(adiposity_sds(df, ref = ref, na_action = "error", diagnostics = FALSE),
               "rows have missing values")
})

test_that("check_extreme removed: function passes through raw outlier rows", {
  skip_on_cran()
  df <- tibble(BMI = c(120, 22), waist = c(300, 60))
  out <- adiposity_sds(df, ref = ref, diagnostics = FALSE)
  expect_equal(nrow(out), 2L)
})

test_that("SDS extreme handling respects sds_cap and extreme_action", {
  skip_on_cran()
  df <- tibble(BMI = c(23 + 100*4, 23), waist = c(80, 80))
  out_cap <- adiposity_sds(df, ref = ref, extreme_action = "cap", sds_cap = 6)
  expect_lte(max(abs(out_cap$BMI_SDS)), 6)
  expect_error(adiposity_sds(df, ref = ref, extreme_action = "error", sds_cap = 6), "SDS beyond")
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(BMI = 23, waist = 80)
  expect_message(adiposity_sds(df, ref = ref, verbose = TRUE), "adiposity_sds")
  expect_message(adiposity_sds(df, ref = ref, verbose = TRUE), "col_map")
  expect_message(adiposity_sds(df, ref = ref, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(BMI = 23, waist = 80)
  msgs <- testthat::capture_messages(adiposity_sds(df, ref = ref, verbose = TRUE))
  expect_equal(sum(grepl("col_map",  msgs)), 1L)
  expect_equal(sum(grepl("results:",    msgs)), 1L)
})

#' Calculate pulmonary function markers (FEV1/FVC, z-scores, percent predicted, LLN, etc.)
#'
#' Uses the `rspiro` reference equations to compute predicted normals,
#' z-scores, percent predicted and lower limits of normal (LLN) for FEV1, FVC,
#' and the FEV1/FVC ratio.
#'
#' Inputs are validated, missingness is handled via `na_action`, and heights are
#' auto-detected as cm when any height > 3; otherwise interpreted as metres (no
#' automatic unit conversion beyond that heuristic, preserving prior behavior).
#'
#' @param data A data.frame or tibble with columns:
#'   - `age`       (numeric): years
#'   - `sex`       (character or numeric): "male"/"female" (case-insensitive) or codes 1/2 or 0/1
#'   - `height`    (numeric): cm or m (auto-detected)
#'   - `ethnicity` (character): e.g. "Caucasian", "African-American", "NE Asian", "SE Asian", "Other/Mixed"
#'   - `fev1`      (numeric): observed FEV1 in L
#'   - `fvc`       (numeric): observed FVC in L
#' @param equation One of `c("GLI","GLIgl","NHANES3")` (see `rspiro` for details).
#'   GLIgl ignores ethnicity.
#' @param na_action One of `c("keep","omit","error")` for handling missing values in
#'   required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness warnings on
#'   required inputs. Default 0.2.
#' @param col_map Optional named list mapping canonical keys (`age`, `sex`, `height`,
#'   `ethnicity`, `fev1`, `fvc`) to actual column names in `data`. If `NULL`,
#'   column names are inferred automatically.
#' @param verbose Logical; if `TRUE` prints progress messages and a completion summary.
#'
#' @return A tibble with columns:
#'   - `fev1_pred`, `fev1_z`,   `fev1_pctpred`, `fev1_LLN`
#'   - `fvc_pred`,  `fvc_z`,    `fvc_pctpred`,  `fvc_LLN`
#'   - `fev1_fvc_ratio`, `fev1_fvc_pred`, `fev1_fvc_z`, `fev1_fvc_pctpred`,
#'     `fev1_fvc_LLN` (`NA` if the equation lacks native FEV1/FVC support in rspiro)
#'
#' @seealso rspiro
#'
#' @note
#' `fev1_fvc_z`, `fev1_fvc_pred`, and `fev1_fvc_LLN` are computed via
#' rspiro's native FEV1/FVC parameter (equivalent to `param = "FEV1FVC"`).
#' If that parameter is not supported by the installed rspiro version or
#' equation, these columns fall back gracefully: `fev1_fvc_z` and
#' `fev1_fvc_LLN` become `NA`; `fev1_fvc_pred` falls back to
#' `fev1_pred / fvc_pred`.
#'
#' @references
#' \insertRef{quanjer2012}{HealthMarkers}
#' \insertRef{hankinson1999spirometry}{HealthMarkers}
#' \insertRef{bowerman2023gli}{HealthMarkers} (race-neutral GLI global equations and interpretation framework; used by rspiro's `GLIgl` equation)
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#'
#' @examples
#' if (requireNamespace("rspiro", quietly = TRUE)) {
#'   df <- data.frame(
#'     age = c(40, 55), sex = c("male", "female"),
#'     height = c(175, 162), ethnicity = c("Caucasian", "Caucasian"),
#'     fev1 = c(3.5, 2.4), fvc = c(4.4, 3.1)
#'   )
#'   pulmo_markers(df)
#' }
#' @export
pulmo_markers <- function(data,
                          col_map = NULL,
                          equation = c("GLI", "GLIgl", "NHANES3"),
                          na_action = c("keep","omit","error"),
                          na_warn_prop = 0.2,
                          verbose = TRUE) {
  # HM-CS v2: validate inputs (data-frame only here)
  hm_validate_inputs(data, col_map, required_keys = character(0), fn = "pulmo_markers")
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "pulmo_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)

  # Custom equation check (clear error message on unknown)
  eq <- as.character(equation)[1]
  if (!eq %in% c("GLI", "GLIgl", "NHANES3")) {
    rlang::abort(
      sprintf("pulmo_markers(): unsupported equation '%s'. Choose one of: GLI, GLIgl, NHANES3.", eq),
      class = "healthmarkers_pulmo_error_equation"
    )
  }

  na_action <- match.arg(na_action)

  hm_inform(sprintf("pulmo_markers(): preparing inputs [%s]", eq),
            level = if (isTRUE(verbose)) "inform" else "debug")

  # Basic validation and coercion
  .pm_validate_df(data)
  req <- c("age", "sex", "height", "ethnicity", "fev1", "fvc")
  cm  <- .hm_build_col_map(data, col_map, keys = req, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  miss <- setdiff(req, names(data))
  if (length(miss)) {
    rlang::abort(
      paste0("pulmo_markers(): missing required columns: ", paste(miss, collapse = ", ")),
      class = "healthmarkers_pulmo_error_missing_columns"
    )
  }

  data <- .pm_coerce_numeric(data, cols = c("age","height","fev1","fvc"))

  # High-missingness diagnostics on required inputs (debug)
  .pm_high_missing_diag(data, req, na_warn_prop = na_warn_prop)

  # NA policy on required inputs
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("pulmo_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_pulmo_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(sprintf("pulmo_markers(): omitting %d rows with NA in required inputs", sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      fev1_pred = numeric(), fev1_z = numeric(), fev1_pctpred = numeric(), fev1_LLN = numeric(),
      fvc_pred  = numeric(), fvc_z  = numeric(), fvc_pctpred  = numeric(), fvc_LLN  = numeric(),
      fev1_fvc_ratio = numeric(), fev1_fvc_pred = numeric(), fev1_fvc_z = numeric(),
      fev1_fvc_pctpred = numeric(), fev1_fvc_LLN = numeric()
    ))
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  fev1_pred, fev1_z, fev1_pctpred, fev1_LLN [age, height, sex, ethnicity, fev1]\n  fvc_pred, fvc_z, fvc_pctpred, fvc_LLN [age, height, sex, ethnicity, fvc]\n  fev1_fvc_ratio, fev1_fvc_pred, fev1_fvc_z, fev1_fvc_pctpred, fev1_fvc_LLN [fev1, fvc]", fn_name), level = "inform")

  # Map sex and ethnicity to rspiro codes (male=1, female=2; GLI: 1-5 ethnic groups)
  sex_code <- .pm_map_sex(data$sex)
  eth_code <- .pm_map_ethnicity(data$ethnicity)

  # Height auto-detection: any height > 3 -> cm
  height_m <- data$height
  if (any(height_m > 3, na.rm = TRUE)) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "pulmo_markers(): converting height from cm to m")
    height_m <- height_m / 100
  }
  # Basic plausibility on height
  if (any(is.finite(height_m) & height_m <= 0, na.rm = TRUE)) {
    rlang::warn("pulmo_markers(): non-positive heights detected; outputs will be NA for those rows.")
  }

  age <- data$age
  fev1 <- data$fev1
  fvc  <- data$fvc

  hm_inform(level = "debug", msg = sprintf("pulmo_markers(): sex=%s; eth=%s",
                        paste(sort(unique(sex_code)), collapse = ","),
                        paste(sort(unique(eth_code)), collapse = ",")))

  # Lookup reference functions from rspiro
  ns <- tryCatch(asNamespace("rspiro"), error = function(e) NULL)
  if (is.null(ns)) {
    rlang::abort("pulmo_markers(): package 'rspiro' is required but not installed.",
                 class = "healthmarkers_pulmo_error_missing_pkg")
  }
  f_pred   <- paste0("pred_",   eq)
  f_zscore <- paste0("zscore_", eq)
  f_lln    <- paste0("LLN_",    eq)
  if (!all(vapply(c(f_pred, f_zscore, f_lln), exists, logical(1), envir = ns, inherits = FALSE))) {
    rlang::abort(
      sprintf("pulmo_markers(): equation '%s' not supported by installed 'rspiro'.", eq),
      class = "healthmarkers_pulmo_error_equation"
    )
  }
  pred_fun   <- get(f_pred, envir = ns)
  zscore_fun <- get(f_zscore, envir = ns)
  lln_fun    <- get(f_lln, envir = ns)

  hm_inform("pulmo_markers(): computing markers", level = "debug")

  # Safe division helper (defined before rspiro calls)
  safe_div <- function(num, den) {
    out <- num / den
    out[!is.finite(out)] <- NA_real_
    out
  }

  # Observed FEV1/FVC ratio (needed for rspiro FEV1FVC z-score calls)
  obs_ratio <- safe_div(fev1, fvc)

  # Compute predicted, z-scores, LLN by equation
  if (eq == "GLIgl") {
    fev1_pred  <- pred_fun(age, height_m, sex_code, param = "FEV1")
    fvc_pred   <- pred_fun(age, height_m, sex_code, param = "FVC")
    ratio_pred <- tryCatch(pred_fun(age, height_m, sex_code, param = "FEV1FVC"),
                           error = function(e) safe_div(fev1_pred, fvc_pred))
    fev1_z     <- zscore_fun(age, height_m, sex_code, FEV1 = fev1)
    fvc_z      <- zscore_fun(age, height_m, sex_code, FVC = fvc)
    ratio_z    <- tryCatch(zscore_fun(age, height_m, sex_code, FEV1FVC = obs_ratio),
                           error = function(e) rep(NA_real_, length(obs_ratio)))
    fev1_LLN   <- lln_fun(age, height_m, sex_code, param = "FEV1")
    fvc_LLN    <- lln_fun(age, height_m, sex_code, param = "FVC")
    ratio_LLN  <- tryCatch(lln_fun(age, height_m, sex_code, param = "FEV1FVC"),
                           error = function(e) rep(NA_real_, length(age)))
  } else {
    fev1_pred  <- pred_fun(age, height_m, sex_code, eth_code, param = "FEV1")
    fvc_pred   <- pred_fun(age, height_m, sex_code, eth_code, param = "FVC")
    ratio_pred <- tryCatch(pred_fun(age, height_m, sex_code, eth_code, param = "FEV1FVC"),
                           error = function(e) safe_div(fev1_pred, fvc_pred))
    fev1_z     <- zscore_fun(age, height_m, sex_code, eth_code, FEV1 = fev1)
    fvc_z      <- zscore_fun(age, height_m, sex_code, eth_code, FVC = fvc)
    ratio_z    <- tryCatch(zscore_fun(age, height_m, sex_code, eth_code, FEV1FVC = obs_ratio),
                           error = function(e) rep(NA_real_, length(obs_ratio)))
    fev1_LLN   <- lln_fun(age, height_m, sex_code, eth_code, param = "FEV1")
    fvc_LLN    <- lln_fun(age, height_m, sex_code, eth_code, param = "FVC")
    ratio_LLN  <- tryCatch(lln_fun(age, height_m, sex_code, eth_code, param = "FEV1FVC"),
                           error = function(e) rep(NA_real_, length(age)))
  }

  # Percent predicted
  fev1_pctpred <- 100 * safe_div(fev1, fev1_pred)
  fvc_pctpred  <- 100 * safe_div(fvc,  fvc_pred)
  ratio_pct    <- 100 * safe_div(obs_ratio, ratio_pred)

  out <- tibble::tibble(
    fev1_pred        = as.numeric(fev1_pred),
    fev1_z           = as.numeric(fev1_z),
    fev1_pctpred     = as.numeric(fev1_pctpred),
    fev1_LLN         = as.numeric(fev1_LLN),
    fvc_pred         = as.numeric(fvc_pred),
    fvc_z            = as.numeric(fvc_z),
    fvc_pctpred      = as.numeric(fvc_pctpred),
    fvc_LLN          = as.numeric(fvc_LLN),
    fev1_fvc_ratio   = as.numeric(obs_ratio),
    fev1_fvc_pred    = as.numeric(ratio_pred),
    fev1_fvc_z       = as.numeric(ratio_z),
    fev1_fvc_pctpred = as.numeric(ratio_pct),
    fev1_fvc_LLN     = as.numeric(ratio_LLN)
  )

  if (!is.null(id_col)) {
    id_vec <- data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) hm_inform(hm_result_summary(out, fn_name), level = "inform")

  out
}

# ---- internal helpers ---------------------------------------------------------

.pm_validate_df <- function(df) {
  if (!is.data.frame(df)) {
    rlang::abort("pulmo_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_pulmo_error_data_type")
  }
  invisible(TRUE)
}

.pm_coerce_numeric <- function(df, cols) {
  for (cn in cols) {
    if (!is.numeric(df[[cn]])) {
      old <- df[[cn]]
      suppressWarnings(df[[cn]] <- as.numeric(old))
      introduced_na <- sum(is.na(df[[cn]]) & !is.na(old))
      if (introduced_na > 0) {
        rlang::warn(sprintf("pulmo_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      }
    }
  }
  df
}

.pm_high_missing_diag <- function(df, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- df[[cn]]
    if (length(x) == 0L) next
    pna <- sum(is.na(x)) / length(x)
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("pulmo_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }
  invisible(TRUE)
}

.pm_map_sex <- function(sex) {
  s <- tolower(as.character(sex))
  out <- rep(NA_integer_, length(s))

  # numeric forms first
  num <- suppressWarnings(as.numeric(s))
  out[!is.na(num) & num %in% c(1, 2)] <- as.integer(num[!is.na(num) & num %in% c(1, 2)])
  out[!is.na(num) & num %in% c(0, 1) & is.na(out)] <- as.integer(num[!is.na(num) & num %in% c(0, 1)]) + 1L

  # character forms
  out[s %in% c("male","m")] <- 1L
  out[s %in% c("female","f")] <- 2L

  bad <- sum(is.na(out))
  if (bad > 0) rlang::warn(sprintf("pulmo_markers(): 'sex' has %d unmapped values; set to NA.", bad))
  out
}

.pm_map_ethnicity <- function(eth) {
  e <- tolower(as.character(eth))
  out <- ifelse(
    e %in% c("caucasian","white","european","non-hispanic white"), 1L,
    ifelse(e %in% c("african-american","black","african american"), 2L,
      ifelse(e %in% c("ne asian","northeast asian","east asian"), 3L,
        ifelse(e %in% c("se asian","southeast asian"), 4L, 5L)
      )
    )
  )
  # Treat obviously "other"/"mixed"/"hispanic" as code 5
  out[e %in% c("other","mixed","other/mixed","hispanic","latino","south asian","indian")] <- 5L
  bad <- sum(is.na(out))
  if (bad > 0) rlang::warn(sprintf("pulmo_markers(): 'ethnicity' has %d unmapped values; set to 'Other/Mixed' (5).", bad))
  out[is.na(out)] <- 5L
  out
}

library(testthat)
library(tibble)

skip_if_no_rspiro_2 <- function() {
  skip_if_not_installed("rspiro")
  ns <- tryCatch(asNamespace("rspiro"), error = function(e) NULL)
  skip_if(is.null(ns), "rspiro namespace not available")
  has_eq <- function(eq) {
    req <- c(paste0("pred_", eq), paste0("zscore_", eq), paste0("LLN_", eq))
    all(vapply(req, function(fn) exists(fn, envir = ns, inherits = FALSE), logical(1)))
  }
  available <- c("GLI", "GLIgl", "NHANES3")[vapply(c("GLI", "GLIgl", "NHANES3"), has_eq, logical(1))]
  skip_if(length(available) == 0, "rspiro lacks required prediction/zscore/LLN functions")
  invisible(available)
}

test_that("pulmo_markers computes expected pulmonary metrics across equations", {
  skip_on_cran()
  available_eqs <- skip_if_no_rspiro_2()

  df <- tibble(
    age = 45,
    sex = "male",
    height = 170, # cm; function will convert to metres
    ethnicity = "Caucasian",
    fev1 = 3.0,
    fvc = 4.0
  )

  for (eq in available_eqs) {
    out <- pulmo_markers(df, equation = eq, verbose = FALSE)

    expected_cols <- c(
      "fev1_pred", "fev1_z", "fev1_pctpred", "fev1_LLN",
      "fvc_pred", "fvc_z", "fvc_pctpred", "fvc_LLN",
      "fev1_fvc_ratio", "fev1_fvc_pred", "fev1_fvc_z", "fev1_fvc_pctpred", "fev1_fvc_LLN"
    )
    expect_true(all(expected_cols %in% names(out)))

    # Basic finiteness for predictions and percent predicted
    expect_true(is.finite(out$fev1_pred[1]))
    expect_true(is.finite(out$fvc_pred[1]))
    expect_true(is.finite(out$fev1_pctpred[1]))
    expect_true(is.finite(out$fvc_pctpred[1]))

    # z-scores may be NA for some ref sets; only range-check when finite
    if (is.finite(out$fev1_z[1])) expect_true(out$fev1_z[1] > -10 && out$fev1_z[1] < 10)
    if (is.finite(out$fvc_z[1]))  expect_true(out$fvc_z[1]  > -10 && out$fvc_z[1]  < 10)

    # FEV1/FVC ratio should be (0,1] for positive volumes
    expect_true(is.finite(out$fev1_fvc_ratio[1]))
    expect_true(out$fev1_fvc_ratio[1] > 0 && out$fev1_fvc_ratio[1] <= 1)
  }
})

test_that("height auto-conversion (cm vs m) yields identical predictions/z-scores", {
  skip_on_cran()
  available_eqs <- skip_if_no_rspiro_2()

  df_cm <- tibble(age = 50, sex = "female", height = 162, ethnicity = "Caucasian", fev1 = 2.6, fvc = 3.2)
  df_m  <- tibble(age = 50, sex = "female", height = 1.62, ethnicity = "Caucasian", fev1 = 2.6, fvc = 3.2)

  for (eq in available_eqs) {
    out_cm <- pulmo_markers(df_cm, equation = eq)
    out_m  <- pulmo_markers(df_m,  equation = eq)
    # Predictions should match closely
    expect_equal(out_cm$fev1_pred, out_m$fev1_pred, tolerance = 1e-6)
    expect_equal(out_cm$fvc_pred,  out_m$fvc_pred,  tolerance = 1e-6)
    # z-scores may differ in floating noise only
    if (is.finite(out_cm$fev1_z[1]) && is.finite(out_m$fev1_z[1])) {
      expect_equal(out_cm$fev1_z, out_m$fev1_z, tolerance = 1e-6)
    }
    if (is.finite(out_cm$fvc_z[1]) && is.finite(out_m$fvc_z[1])) {
      expect_equal(out_cm$fvc_z, out_m$fvc_z, tolerance = 1e-6)
    }
  }
})

test_that("na_action policies: error and omit behave as expected", {
  skip_on_cran()
  available_eqs <- skip_if_no_rspiro_2()
  eq <- available_eqs[[1]]

  df_na <- tibble(
    age = 45, sex = "male", height = 170, ethnicity = "Caucasian",
    fev1 = NA_real_, fvc = 4.0
  )
  expect_error(
    suppressWarnings(
      pulmo_markers(df_na, equation = eq, na_action = "error")
    ),
    "required inputs contain missing values"
  )

  out_omit <- suppressWarnings(
    pulmo_markers(df_na, equation = eq, na_action = "omit")
  )
  expect_equal(nrow(out_omit), 0L)
})

test_that("sex and ethnicity mappings are robust to codes and strings", {
  skip_on_cran()
  available_eqs <- skip_if_no_rspiro_2()
  # GLI uses 5-level ethnicity; GLIgl ignores ethnicity. NHANES3 requires 1–3; skip it here.
  eqs_to_test <- intersect(available_eqs, c("GLI", "GLIgl"))
  skip_if(length(eqs_to_test) == 0, "No GLI/GLIgl equations available to test mappings")

  df_str <- tibble(age = 40, sex = "male",   height = 180, ethnicity = "Caucasian", fev1 = 3.5, fvc = 4.5)
  df_num <- tibble(age = 40, sex = 1,        height = 180, ethnicity = "white",     fev1 = 3.5, fvc = 4.5)
  df_fem <- tibble(age = 40, sex = "female", height = 165, ethnicity = "Other",     fev1 = 2.8, fvc = 3.6)

  for (eq in eqs_to_test) {
    out_str <- pulmo_markers(df_str, equation = eq)
    out_num <- pulmo_markers(df_num, equation = eq)
    # male string vs numeric code 1 should match
    expect_equal(out_str$fev1_pred, out_num$fev1_pred, tolerance = 1e-6)
    expect_equal(out_str$fvc_pred,  out_num$fvc_pred,  tolerance = 1e-6)

    # female runs without error and yields finite predictions
    out_fem <- pulmo_markers(df_fem, equation = eq)
    expect_true(is.finite(out_fem$fev1_pred[1]))
    expect_true(is.finite(out_fem$fvc_pred[1]))
  }
})

test_that("pulmo_markers errors on missing required columns", {
  skip_on_cran()
  skip_if_no_rspiro_2()
  df <- tibble(
    age = 45, sex = "male", height = 170, ethnicity = "Caucasian",
    # fev1 missing
    fvc = 4.0
  )
  expect_error(
    pulmo_markers(df, equation = "GLI"),
    "(missing|required).*(fev1|inputs|columns)"
  )
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  available <- skip_if_no_rspiro_2()
  eq <- available[[1]]
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(age = 45, sex = "male", height = 170, ethnicity = "Caucasian", fev1 = 3.0, fvc = 4.0)
  expect_message(pulmo_markers(df, equation = eq, verbose = TRUE), "pulmo_markers")
  expect_message(pulmo_markers(df, equation = eq, verbose = TRUE), "col_map")
  expect_message(pulmo_markers(df, equation = eq, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  available <- skip_if_no_rspiro_2()
  eq <- available[[1]]
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(age = 45, sex = "male", height = 170, ethnicity = "Caucasian", fev1 = 3.0, fvc = 4.0)
  msgs <- testthat::capture_messages(pulmo_markers(df, equation = eq, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action='keep' preserves rows and propagates NA", {
  skip_on_cran()
  available <- skip_if_no_rspiro_2()
  eq <- available[[1]]
  df <- tibble(age = 45, sex = "male", height = 170, ethnicity = "Caucasian", fev1 = NA_real_, fvc = 4.0)
  out <- pulmo_markers(df, equation = eq, na_action = "keep")
  expect_equal(nrow(out), 1L)
  expect_true(is.na(out$fev1_pctpred[1]) || is.na(out$fev1_z[1]))
})

test_that("fvc=0 yields NA fev1_fvc_ratio; predictions remain finite", {
  skip_on_cran()
  available <- skip_if_no_rspiro_2()
  eq <- available[[1]]
  df <- tibble(age = 45, sex = "male", height = 170, ethnicity = "Caucasian", fev1 = 3.0, fvc = 0)
  out <- pulmo_markers(df, equation = eq)
  expect_true(is.na(out$fev1_fvc_ratio[1]))
  expect_true(is.finite(out$fev1_pred[1]))
  expect_true(is.finite(out$fvc_pred[1]))
})

test_that("percent predicted is ~100 when measured equals predicted", {
  skip_on_cran()
  available <- skip_if_no_rspiro_2()
  eq <- available[[1]]
  base <- tibble(age = 50, sex = "female", height = 162, ethnicity = "Caucasian", fev1 = 2.6, fvc = 3.2)
  preds <- pulmo_markers(base, equation = eq)
  df2 <- tibble(
    age = base$age, sex = base$sex, height = base$height, ethnicity = base$ethnicity,
    fev1 = preds$fev1_pred, fvc = preds$fvc_pred
  )
  out2 <- pulmo_markers(df2, equation = eq)
  expect_equal(out2$fev1_pctpred, 100, tolerance = 1e-6)
  expect_equal(out2$fvc_pctpred,  100, tolerance = 1e-6)
})

test_that("unknown equation aborts with clear error", {
  skip_on_cran()
  skip_if_no_rspiro_2()
  df <- tibble(age = 45, sex = "male", height = 170, ethnicity = "Caucasian", fev1 = 3.0, fvc = 4.0)
  expect_error(
    pulmo_markers(df, equation = "NOPE"),
    "equation|unsupported|unknown"
  )
})

#' Calculate metabolic risk feature flags (pediatric-friendly thresholds)
#'
#' Compute four binary risk flags from routine clinical measures:
#' - dyslipidemia
#' - insulin_resistance
#' - hyperglycemia (prediabetes-range glycemia)
#' - hypertension (BP >=95th percentile via z > 1.64)
#'
#' By default, behavior matches prior implementation: required columns are
#' validated, NA values are kept (propagate to outputs), no extreme-value
#' checks or capping are applied, and a tibble with 0/1 factor flags is returned.
#'
#' Units and criteria (no automatic unit conversion):
#' - Lipids (mmol/L): total cholesterol > 5.2 OR LDL-C > 3.4 OR HDL-C < 1.0 OR
#'   triglycerides > 1.1 (age 0-9) OR > 1.5 (age 10-19) => dyslipidemia = 1.
#'   Note: no TG cutoff is applied for adults aged >= 20 years.
#' - Insulin resistance: z_HOMA > 1.28 (~=90th percentile) => insulin_resistance = 1.
#'   z_HOMA is a within-sample or external z-score of HOMA-IR (see Matthews et al. 1985).
#' - Hyperglycemia: fasting glucose in (5.6, 6.9) mmol/L OR HbA1c in (39, 47) mmol/mol
#'   => hyperglycemia = 1. Boundaries are EXCLUSIVE (open intervals); boundary values
#'   (exactly 5.6 or 6.9 mmol/L; exactly 39 or 47 mmol/mol) are not flagged.
#'   ADA criteria use inclusive lower bound (>= 5.6 mmol/L, >= 39 mmol/mol).
#' - Hypertension: either BP z-score > 1.64 (~=95th percentile) for systolic or diastolic
#'   => hypertension = 1.
#'
#' @param data A data.frame or tibble containing at least these numeric columns:
#'   - chol_total, chol_ldl, chol_hdl, triglycerides (mmol/L)
#'   - age_year (years)
#'   - z_HOMA (standardized HOMA-IR)
#'   - glucose (mmol/L)
#'   - HbA1c (mmol/mol; IFCC units)
#'   - bp_sys_z, bp_dia_z (BP z-scores)
#' @param col_map Optional named list to map required keys to column names in `data`.
#'   Keys: c("chol_total","chol_ldl","chol_hdl","triglycerides","age_year",
#'   "z_HOMA","glucose","HbA1c","bp_sys_z","bp_dia_z"). Default NULL (use same names).
#' @param na_action One of c("keep","omit","error","ignore","warn") controlling missing-data policy.
#'   - "keep": keep NA; outputs become NA where inputs are NA.
#'   - "omit": drop rows with NA in any required input.
#'   - "error": abort if any required input contains NA.
#'   - "ignore"/"warn": aliases of "keep"; "warn" also emits missingness diagnostics.
#' @param na_warn_prop Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness warnings. Default 0.2.
#' @param verbose Logical; if TRUE, prints column mapping and computing messages.
#'
#' @return A tibble with four factor columns (levels c("0","1")):
#' - dyslipidemia
#' - insulin_resistance
#' - hyperglycemia
#' - hypertension
#'
#' @note These flags are heuristic screening rules derived from published clinical
#'   guidelines. They are not validated diagnostic criteria and should not replace
#'   clinical judgment. The dyslipidemia and hypertension thresholds are designed
#'   for pediatric populations (ages 0-19); for adults >= 20, only TC, LDL-C, and
#'   HDL-C criteria contribute to the dyslipidemia flag.
#'
#' @seealso [liver_markers()], [lipid_markers()], [kidney_failure_risk()], [inflammatory_markers()]
#'
#' @references
#' \insertRef{nhlbi2011lipids}{HealthMarkers}
#' \insertRef{ada2024standards}{HealthMarkers}
#' \insertRef{flynn2017bp}{HealthMarkers}
#' \insertRef{matthews1985homa}{HealthMarkers}
#'
#' @importFrom Rdpack reprompt
#' @importFrom rlang abort warn inform
#' @importFrom dplyr transmute if_else
#' @importFrom tibble as_tibble
#'
#' @examples
#' df <- data.frame(
#'   chol_total = c(5.2, 6.4), chol_ldl = c(3.2, 4.1), chol_hdl = c(1.3, 1.0),
#'   triglycerides = c(1.8, 2.5), age_year = c(45, 60), z_HOMA = c(0.5, 1.2),
#'   glucose = c(5.5, 6.8), HbA1c = c(38, 46), bp_sys_z = c(0.2, 1.1),
#'   bp_dia_z = c(0.1, 0.9)
#' )
#' metabolic_risk_features(df)
#' @export
metabolic_risk_features <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  na_warn_prop = 0.2,
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "metabolic_risk_features"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action <- .na$na_action_eff

  .mrf_validate_args(data, col_map, na_warn_prop)

  # Required keys and column mapping
  required_cols <- c(
    "chol_total", "chol_ldl", "chol_hdl", "triglycerides",
    "age_year", "z_HOMA", "glucose", "HbA1c", "bp_sys_z", "bp_dia_z"
  )

  cm      <- .hm_build_col_map(data, col_map, required_cols, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Strict check: error if user-provided key points to column not in data
  if (length(cm$user_keys)) {
    bad_cols <- setdiff(unlist(col_map[cm$user_keys], use.names = FALSE), names(data))
    if (length(bad_cols))
      rlang::abort(paste0("metabolic_risk_features(): mapped columns not found in data: ", paste(bad_cols, collapse = ", ")),
                   class = "healthmarkers_mrf_error_missing_columns")
  }

  miss_keys <- setdiff(required_cols, names(col_map))
  if (length(miss_keys)) {
    rlang::abort(
      paste0("metabolic_risk_features(): missing col_map entries for: ", paste(miss_keys, collapse = ", ")),
      class = "healthmarkers_mrf_error_missing_map")
  }

  # Ensure mapped columns exist
  mapped_cols <- unlist(col_map[required_cols], use.names = FALSE)
  miss_cols <- setdiff(mapped_cols, names(data))
  if (length(miss_cols)) {
    rlang::abort(
      paste0("metabolic_risk_features(): mapped columns not found in data: ", paste(miss_cols, collapse = ", ")),
      class = "healthmarkers_mrf_error_missing_columns"
    )
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf(
      "%s(): computing markers:\n  dyslipidemia       [chol_total, chol_ldl, chol_hdl, triglycerides, age_year]\n  insulin_resistance [z_HOMA]\n  hyperglycemia      [glucose, HbA1c]\n  hypertension       [bp_sys_z, bp_dia_z]",
      fn_name), level = "inform")

  # HM-CS v2: numeric coercion for required inputs; warn if NAs introduced; set non-finite to NA
  for (cn in mapped_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::warn(sprintf("metabolic_risk_features(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High missingness diagnostics only when na_action_raw == "warn"
  if (identical(na_action_raw, "warn")) {
    .mrf_warn_high_missing(data, mapped_cols, na_warn_prop = na_warn_prop)
  }

  # NA policy
  if (na_action == "error") {
    any_na <- Reduce(`|`, lapply(required_cols, function(k) is.na(data[[col_map[[k]]]])))
    if (any(any_na)) {
      rlang::abort("metabolic_risk_features(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_mrf_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(required_cols, function(k) is.na(data[[col_map[[k]]]])))
    hm_inform(sprintf("metabolic_risk_features(): omitting %d rows with NA in required inputs", sum(!keep)), level = if (isTRUE(verbose)) "inform" else "debug")
    data <- data[keep, , drop = FALSE]
  }

  # Shorthand for mapped columns
  CT  <- data[[col_map$chol_total]]
  LDL <- data[[col_map$chol_ldl]]
  HDL <- data[[col_map$chol_hdl]]
  TG  <- data[[col_map$triglycerides]]
  AGE <- data[[col_map$age_year]]
  ZH  <- data[[col_map$z_HOMA]]
  GLU <- data[[col_map$glucose]]
  A1c <- data[[col_map$HbA1c]]
  SBPz <- data[[col_map$bp_sys_z]]
  DBPz <- data[[col_map$bp_dia_z]]

  # Compute flags
  out <- dplyr::transmute(
    data,
    dyslipidemia = factor(
      dplyr::if_else(
        CT > 5.2 |
          LDL > 3.4 |
          HDL < 1.0 |
          (TG > 1.1 & AGE %in% 0:9) |
          (TG > 1.5 & AGE %in% 10:19),
        1L, 0L, missing = NA_integer_
      ),
      levels = c(0L, 1L)
    ),
    insulin_resistance = factor(
      dplyr::if_else(ZH > 1.28, 1L, 0L, missing = NA_integer_),
      levels = c(0L, 1L)
    ),
    hyperglycemia = factor(
      dplyr::if_else(
        (GLU > 5.6 & GLU < 6.9) |
          (A1c > 39 & A1c < 47),
        1L, 0L, missing = NA_integer_
      ),
      levels = c(0L, 1L)
    ),
    hypertension = factor(
      dplyr::if_else(SBPz > 1.64 | DBPz > 1.64, 1L, 0L, missing = NA_integer_),
      levels = c(0L, 1L)
    )
  )

  out <- tibble::as_tibble(out)
  if (!is.null(id_col)) {
    out[[id_col]] <- data[[id_col]]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }
  out
}

# ---- internal helpers (not exported) -----------------------------------------

.mrf_validate_args <- function(data, col_map, na_warn_prop) {
  if (!is.data.frame(data)) {
    rlang::abort("metabolic_risk_features(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_mrf_error_data_type")
  }
  if (!is.null(col_map)) {
    if (!is.list(col_map) || is.null(names(col_map)) || any(names(col_map) == "")) {
      rlang::abort("metabolic_risk_features(): `col_map` must be a named list when supplied.",
                   class = "healthmarkers_mrf_error_colmap_type")
    }
  }
  if (!(is.numeric(na_warn_prop) && length(na_warn_prop) == 1L &&
        is.finite(na_warn_prop) && na_warn_prop >= 0 && na_warn_prop <= 1)) {
    rlang::abort("metabolic_risk_features(): `na_warn_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_mrf_error_na_warn_prop")
  }
  invisible(TRUE)
}

.mrf_warn_high_missing <- function(df, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- df[[cn]]; n <- length(x); if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("metabolic_risk_features(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
    if (grepl("chol|triglycerides", cn, ignore.case = TRUE)) {
      if (sum(is.finite(x) & x > 30) > 0L) {
        rlang::warn(sprintf("metabolic_risk_features(): '%s' has very large values (>30 mmol/L); check units (mmol/L expected).", cn))
      }
      if (sum(is.finite(x) & x < 0) > 0L) {
        rlang::warn(sprintf("metabolic_risk_features(): '%s' contains negative values; check input.", cn))
      }
    }
    if (identical(cn, "HbA1c")) {
      if (sum(is.finite(x) & x <= 14) > 0L) {
        rlang::warn("metabolic_risk_features(): 'HbA1c' appears to be in percent for some rows; expected mmol/mol.")
      }
    }
    if (identical(cn, "glucose")) {
      if (sum(is.finite(x) & x > 40) > 0L) {
        rlang::warn("metabolic_risk_features(): 'glucose' values > 40 mmol/L detected; check units.")
      }
      if (sum(is.finite(x) & x < 0) > 0L) {
        rlang::warn("metabolic_risk_features(): 'glucose' has negative values; check input.")
      }
    }
    if (grepl("bp_.*_z$", cn)) {
      if (sum(is.finite(x) & (x < -5 | x > 5)) > 0L) {
        rlang::warn(sprintf("metabolic_risk_features(): '%s' z-scores outside [-5,5]; check scaling.", cn))
      }
    }
  }
  invisible(TRUE)
}

.mrf_default_extreme_rules <- function() {
  list(
    chol_total   = c(0.5, 20),
    chol_ldl     = c(0.1, 15),
    chol_hdl     = c(0.1, 5),
    triglycerides= c(0.1, 30),
    age_year     = c(0, 120),
    z_HOMA       = c(-5, 5),
    glucose      = c(2, 40),
    HbA1c        = c(15, 200),
    bp_sys_z     = c(-5, 5),
    bp_dia_z     = c(-5, 5)
  )
}

.mrf_extreme_scan <- function(df, col_map, rules, required) {
  count <- 0L
  flags <- list()
  for (nm in intersect(names(rules), required)) {
    cn <- col_map[[nm]]
    if (!cn %in% names(df)) next
    x <- df[[cn]]
    rng <- rules[[nm]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    flags[[cn]] <- bad
    count <- count + sum(bad, na.rm = TRUE)
  }
  list(count = count, flags = flags)
}

.mrf_cap_inputs <- function(df, flags, col_map, rules) {
  for (cn in names(flags)) {
    rn <- names(col_map)[match(cn, unlist(col_map, use.names = FALSE))]; rn <- rn[!is.na(rn)][1]
    if (is.na(rn) || is.null(rules[[rn]])) next
    rng <- rules[[rn]]; x <- df[[cn]]; bad <- flags[[cn]]
    x[bad & is.finite(x) & x < rng[1]] <- rng[1]
    x[bad & is.finite(x) & x > rng[2]] <- rng[2]
    df[[cn]] <- x
  }
  df
}

library(testthat)
library(tibble)
library(HealthMarkers)

test_that("returns tibble with four factor flags and correct names/levels", {
  skip_on_cran()
  df <- tibble(
    chol_total = 6.0, chol_ldl = 3.5, chol_hdl = 1.0, triglycerides = 1.2,
    age_year = 25, z_HOMA = 1.5, glucose = 5.8, HbA1c = 40,
    bp_sys_z = 1.7, bp_dia_z = 1.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- metabolic_risk_features(df, col_map = cm)

  expect_s3_class(out, "tbl_df")
  expect_named(out, c("dyslipidemia", "insulin_resistance", "hyperglycemia", "hypertension"))
  lapply(out, function(x) {
    expect_true(is.factor(x))
    expect_identical(levels(x), c("0", "1"))
  })

  # All flags should be "1"
  vals <- vapply(out, function(x) as.integer(as.character(x)), integer(1))
  expect_equal(unname(vals), c(1L, 1L, 1L, 1L))
})

test_that("custom col_map works", {
  skip_on_cran()
  df <- tibble(
    CT = 6.0, LDL = 3.5, HDL = 1.0, TG = 1.2,
    AGE = 25, ZH = 1.5, GLU = 5.8, A1C = 40, SBPZ = 1.7, DBPZ = 1.0
  )
  cm <- list(
    chol_total = "CT", chol_ldl = "LDL", chol_hdl = "HDL", triglycerides = "TG",
    age_year = "AGE", z_HOMA = "ZH", glucose = "GLU", HbA1c = "A1C",
    bp_sys_z = "SBPZ", bp_dia_z = "DBPZ"
  )
  out <- metabolic_risk_features(df, col_map = cm)
  expect_s3_class(out, "tbl_df")
  expect_named(out, c("dyslipidemia", "insulin_resistance", "hyperglycemia", "hypertension"))
})

test_that("errors if mapped columns are missing in data", {
  skip_on_cran()
  df <- tibble(
    chol_total = 6.0, chol_ldl = 3.5, chol_hdl = 1.0, triglycerides = 1.2,
    age_year = 25, z_HOMA = 1.5, glucose = 5.8,
    bp_sys_z = 1.7, bp_dia_z = 1.0
    # HbA1c missing
  )
  cm <- list(
    chol_total = "chol_total", chol_ldl = "chol_ldl", chol_hdl = "chol_hdl", triglycerides = "triglycerides",
    age_year = "age_year", z_HOMA = "z_HOMA", glucose = "glucose", HbA1c = "HbA1c",
    bp_sys_z = "bp_sys_z", bp_dia_z = "bp_dia_z"
  )
  expect_error(
    metabolic_risk_features(df, col_map = cm),
    "mapped columns not found in data"
  )
})

test_that("partial col_map is supplemented by dictionary inference for missing keys", {
  skip_on_cran()
  df <- tibble(
    chol_total = 6.0, chol_ldl = 3.5, chol_hdl = 1.0, triglycerides = 1.2,
    age_year = 25, z_HOMA = 1.5, glucose = 5.8, HbA1c = 40, bp_sys_z = 1.7, bp_dia_z = 1.0
  )
  cm <- list(
    chol_total = "chol_total", chol_ldl = "chol_ldl", chol_hdl = "chol_hdl", triglycerides = "triglycerides",
    age_year = "age_year", z_HOMA = "z_HOMA", glucose = "glucose"
    # HbA1c, bp_sys_z, bp_dia_z missing -> should be inferred from data
  )
  expect_no_error(metabolic_risk_features(df, col_map = cm))
  out <- metabolic_risk_features(df, col_map = cm)
  expect_s3_class(out, "tbl_df")
})

test_that("na_action='error' aborts when required inputs contain NA", {
  skip_on_cran()
  df <- tibble(
    chol_total = 6.0, chol_ldl = 3.5, chol_hdl = 1.0, triglycerides = 1.2,
    age_year = 25, z_HOMA = 1.5, glucose = NA_real_, HbA1c = 40, bp_sys_z = 1.7, bp_dia_z = 1.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_error(
    suppressWarnings(metabolic_risk_features(df, col_map = cm, na_action = "error")),
    "required inputs contain missing values"
  )
})

test_that("na_action='omit' drops rows with NA", {
  skip_on_cran()
  df <- tibble(
    chol_total = c(6.0, 6.0), chol_ldl = c(3.5, 3.5), chol_hdl = c(1.0, 1.0), triglycerides = c(1.2, 1.2),
    age_year = c(25, 25), z_HOMA = c(1.5, 1.5), glucose = c(5.8, NA_real_), HbA1c = c(40, 40),
    bp_sys_z = c(1.7, 1.7), bp_dia_z = c(1.0, 1.0)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- suppressWarnings(metabolic_risk_features(df, col_map = cm, na_action = "omit"))
  expect_equal(nrow(out), 1L)
})

test_that("na_action='keep' yields NA only when the rule cannot be decided", {
  skip_on_cran()
  # Make dyslipidemia undecidable due to NA (no other trigger is TRUE)
  df <- tibble(
    chol_total = 5.0, chol_ldl = NA_real_, chol_hdl = 1.2, triglycerides = 1.0,
    age_year = 25, z_HOMA = 0.0, glucose = 5.0, HbA1c = 40, bp_sys_z = 0.0, bp_dia_z = 0.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- suppressWarnings(metabolic_risk_features(df, col_map = cm, na_action = "keep"))
  expect_true(is.na(out$dyslipidemia))
})

test_that("extreme input values pass through without error", {
  skip_on_cran()
  df <- tibble(
    chol_total = 5.0, chol_ldl = 3.0, chol_hdl = 1.2, triglycerides = 1.0,
    age_year = 30, z_HOMA = 0, glucose = 5.0, HbA1c = 300,
    bp_sys_z = 0, bp_dia_z = 0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- metabolic_risk_features(df, col_map = cm)
  expect_s3_class(out, "tbl_df")
})

test_that("high missingness warning is emitted when proportion exceeds threshold", {
  skip_on_cran()
  df <- tibble(
    chol_total = c(5, 5), chol_ldl = c(3, 3), chol_hdl = c(1.2, 1.2), triglycerides = c(1, 1),
    age_year = c(30, 30), z_HOMA = c(0, 0), glucose = c(5, 5),
    HbA1c = c(NA_real_, 40), bp_sys_z = c(0, 0), bp_dia_z = c(0, 0)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_warning(
    metabolic_risk_features(df, col_map = cm, na_action = "warn", na_warn_prop = 0.2),
    "has high missingness"
  )
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    chol_total = 5, chol_ldl = 3, chol_hdl = 1.2, triglycerides = 1.0,
    age_year = 30, z_HOMA = 0, glucose = 5.0, HbA1c = 40,
    bp_sys_z = 0, bp_dia_z = 0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_message(metabolic_risk_features(df, col_map = cm, verbose = TRUE), "metabolic_risk_features")
  expect_message(metabolic_risk_features(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(metabolic_risk_features(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    chol_total = 5, chol_ldl = 3, chol_hdl = 1.2, triglycerides = 1.0,
    age_year = 30, z_HOMA = 0, glucose = 5.0, HbA1c = 40,
    bp_sys_z = 0, bp_dia_z = 0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  msgs <- testthat::capture_messages(
    metabolic_risk_features(df, col_map = cm, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("output has correct row count without check_extreme", {
  skip_on_cran()
  df <- tibble(
    chol_total = 5.0, chol_ldl = 3.0, chol_hdl = 1.2, triglycerides = 1.0,
    age_year = 30, z_HOMA = 0, glucose = 5.0, HbA1c = 300,
    bp_sys_z = 0, bp_dia_z = 0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- metabolic_risk_features(df, col_map = cm)
  expect_equal(nrow(out), 1L)
})

test_that("numeric coercion warns when NAs introduced", {
  skip_on_cran()
  df <- tibble(
    chol_total = c("5.0","oops"), chol_ldl = c(3.0, 3.0), chol_hdl = c(1.2, 1.2), triglycerides = c(1.0, 1.0),
    age_year = c(30, 30), z_HOMA = c(0, 0), glucose = c(5.0, 5.0), HbA1c = c(40, 40),
    bp_sys_z = c(0, 0), bp_dia_z = c(0, 0)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_warning(
    metabolic_risk_features(df, col_map = cm),
    "coerced to numeric; NAs introduced"
  )
})

test_that("na_action='omit' drops rows with required NA and announces omission", {
  skip_on_cran()
  df <- tibble(
    chol_total = c(5.0, NA), chol_ldl = c(3.0, 3.0), chol_hdl = c(1.2, 1.2), triglycerides = c(1.0, 1.0),
    age_year = c(30, 30), z_HOMA = c(0, 0), glucose = c(5.0, 5.0), HbA1c = c(40, 40),
    bp_sys_z = c(0, 0), bp_dia_z = c(0, 0)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    out <- metabolic_risk_features(df, col_map = cm, na_action = "omit", verbose = TRUE),
    "omitting 1 rows with NA in required inputs"
  )
  expect_equal(nrow(out), 1L)
})

#' Appendicular Lean Mass to BMI Index
#'
#' Calculates the ratio of appendicular lean mass (ALM) to body mass index (BMI),
#' and flags low muscle mass based on FNIH Sarcopenia Project cut-points.
#'
#' @details
#' ALM/BMI reflects muscle mass relative to body size. FNIH cut-points:
#' - Men:    ALM/BMI < 0.789
#' - Women:  ALM/BMI < 0.512
#'
#' ALM should be in kilograms and BMI in kg/m^2.
#'
#' @param data A data.frame or tibble with ALM, BMI, and sex columns.
#' @param col_map Named list with:
#'   - alm: appendicular lean mass column name (kg)
#'   - bmi: body mass index column name (kg/m^2)
#'   - sex: sex column name ("Male"/"Female" or m/f; case-insensitive)
#' @param verbose Logical; if TRUE (default), emits progress messages.
#' @param na_action One of c("keep","omit","error","ignore","warn").
#'
#' @return A tibble with:
#'   - alm_bmi_ratio   (numeric)
#'   - low_muscle_mass (logical; TRUE if below sex-specific cut-point;
#'                      NA if sex unknown or ratio NA)
#' @references \insertRef{mclean2014fnih}{HealthMarkers}
#'
#' @examples
#' df <- data.frame(ALM_kg = c(7.2, 5.8, 6.5), BMI = c(24, 28, 22),
#'                  Sex = c("male", "female", "male"))
#' alm_bmi_index(df)
#' @export
alm_bmi_index <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "alm_bmi_index"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  # --- validate data / mapping ------------------------------------------------
  if (!is.data.frame(data)) {
    rlang::abort(
      "alm_bmi_index(): `data` must be a data.frame or tibble.",
      class = "healthmarkers_alm_bmi_error_data_type"
    )
  }

  req <- c("alm","bmi","sex")
  if (!is.null(col_map) && !is.list(col_map)) {
    rlang::abort(
      "alm_bmi_index(): `col_map` must be a named list.",
      class = "healthmarkers_alm_bmi_error_colmap_type"
    )
  }
  cm      <- .hm_build_col_map(data, col_map, req, fn = "alm_bmi_index")
  data    <- cm$data
  col_map <- cm$col_map

  missing_keys <- setdiff(req, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(
      paste0("alm_bmi_index(): missing col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_alm_bmi_error_missing_map"
    )
  }

  mapped <- unname(unlist(col_map[req], use.names = FALSE))
  if (any(!nzchar(mapped))) {
    bad <- req[!nzchar(unname(unlist(col_map[req])))]
    rlang::abort(
      paste0("alm_bmi_index(): missing col_map entries for: ",
             paste(bad, collapse = ", ")),
      class = "healthmarkers_alm_bmi_error_bad_map_values"
    )
  }

  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("alm_bmi_index(): missing required columns in data: ",
             paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_alm_bmi_error_missing_columns"
    )
  }

  # --- verbose messages -------------------------------------------------------
  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  alm_bmi_ratio    [ALM / BMI]\n  low_muscle_mass  [ratio < sex-specific FNIH cut-point]", fn_name), level = "inform")

  # --- coerce ALM/BMI numeric; sex as character ------------------------------
  num_cols <- unname(unlist(col_map[c("alm","bmi")]))
  for (cn in num_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0L) {
        rlang::warn(
          sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
          class = "healthmarkers_alm_bmi_warn_na_coercion"
        )
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  alm <- data[[col_map$alm]]
  bmi <- data[[col_map$bmi]]
  sex_raw <- data[[col_map$sex]]

  # --- NA policy --------------------------------------------------------------
  any_na_req <- is.na(alm) | is.na(bmi) | is.na(sex_raw)

  if (na_action_raw == "warn" && any(any_na_req)) {
    rlang::warn(
      "Missing ALM, BMI, or sex; outputs will be NA for those entries.",
      class = "healthmarkers_alm_bmi_warn_missing_inputs"
    )
  }

  if (na_action_eff == "error" && any(any_na_req)) {
    rlang::abort(
      "alm_bmi_index(): required inputs contain missing values (na_action='error').",
      class = "healthmarkers_alm_bmi_error_missing_values"
    )
  }

  keep <- if (na_action_eff == "omit") !any_na_req else rep(TRUE, length(alm))

  d_alm <- alm[keep]
  d_bmi <- bmi[keep]
  d_sex_raw <- sex_raw[keep]

  # --- normalize sex ----------------------------------------------------------
  sex_chr <- tolower(as.character(d_sex_raw))
  sex_norm <- ifelse(substr(sex_chr, 1, 1) %in% c("m","1"), "male",
               ifelse(substr(sex_chr, 1, 1) %in% c("f","0"), "female", NA_character_))

  if (any(is.na(sex_norm) & !is.na(d_sex_raw))) {
    rlang::warn(
      "alm_bmi_index(): unrecognized sex entries; expected male/female.",
      class = "healthmarkers_alm_bmi_warn_sex_unknown"
    )
  }

  # --- domain warnings --------------------------------------------------------
  if (any(is.finite(d_bmi) & (d_bmi < 10 | d_bmi > 60))) {
    rlang::warn(
      "alm_bmi_index(): BMI outside realistic range (10-60 kg/m^2) detected.",
      class = "healthmarkers_alm_bmi_warn_bmi_range"
    )
  }
  if (any(is.finite(d_alm) & (d_alm < 5 | d_alm > 40))) {
    rlang::warn(
      "alm_bmi_index(): ALM outside plausible range (5-40 kg) detected.",
      class = "healthmarkers_alm_bmi_warn_alm_range"
    )
  }
  if (any(is.finite(d_bmi) & d_bmi <= 0)) {
    rlang::warn(
      "alm_bmi_index(): nonpositive BMI encountered; ratio set to NA for those rows.",
      class = "healthmarkers_alm_bmi_warn_bmi_nonpositive"
    )
  }

  # --- compute ratio and low-muscle flag -------------------------------------
  hm_inform(level = "debug", msg = "alm_bmi_index(): computing")

  ratio <- d_alm / d_bmi
  ratio[!is.finite(ratio)] <- NA_real_

  thr <- ifelse(sex_norm == "male",   0.789,
         ifelse(sex_norm == "female", 0.512, NA_real_))

  low_flag <- ifelse(is.na(ratio) | is.na(thr), NA, ratio < thr)

  out_core <- tibble::tibble(
    alm_bmi_ratio   = as.numeric(ratio),
    low_muscle_mass = as.logical(low_flag)
  )

  # --- pad back if not omitting ----------------------------------------------
  if (na_action_eff != "omit") {
    out <- tibble::tibble(
      alm_bmi_ratio   = rep(NA_real_, length(alm)),
      low_muscle_mass = rep(NA, length(alm))
    )
    out[keep, ] <- out_core
  } else {
    out <- out_core
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

library(testthat)

cm <- list(alm = "ALM_kg", bmi = "BMI", sex = "Sex")

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(ALM_kg = 10, BMI = 25, Sex = "Male")

  # non-list col_map → colmap_type error
  expect_error(
    alm_bmi_index(df, "not_a_list"),
    class = "healthmarkers_alm_bmi_error_colmap_type"
  )

  expect_error(
    alm_bmi_index(df, list(alm = "", bmi = "BMI", sex = "Sex")),
    class = "healthmarkers_alm_bmi_error_bad_map_values"
  )

  expect_error(
    alm_bmi_index(data.frame(X = 1), cm),
    class = "healthmarkers_alm_bmi_error_missing_columns"
  )
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(ALM_kg = 10, BMI = 25, Sex = "Male")
  expect_message(alm_bmi_index(df, cm, verbose = TRUE), "alm_bmi_index")
  expect_message(alm_bmi_index(df, cm, verbose = TRUE), "col_map")
  expect_message(alm_bmi_index(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- data.frame(ALM_kg = 10, BMI = 25, Sex = "Male")
  msgs <- testthat::capture_messages(alm_bmi_index(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("numeric coercion warning when strings introduce NAs", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c("10", "oops"),
    BMI    = c("25", "26"),
    Sex    = c("Male", "Female")
  )

  expect_warning(
    alm_bmi_index(df, cm),
    class = "healthmarkers_alm_bmi_warn_na_coercion"
  )
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c(10, NA, 12),
    BMI    = c(25, 26, 27),
    Sex    = c("Male", "Female", NA)
  )

  out_keep <- alm_bmi_index(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(any(is.na(out_keep$alm_bmi_ratio)))

  out_omit <- alm_bmi_index(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)

  expect_error(
    alm_bmi_index(df, cm, na_action = "error"),
    class = "healthmarkers_alm_bmi_error_missing_values"
  )

  expect_warning(
    alm_bmi_index(df, cm, na_action = "warn"),
    class = "healthmarkers_alm_bmi_warn_missing_inputs"
  )
})

test_that("domain warnings: BMI and ALM ranges and nonpositive BMI", {
  skip_on_cran()
  # BMI out of [10,60], ALM plausible
  df_bmi <- data.frame(
    ALM_kg = c(15, 20),
    BMI    = c(8, 70),
    Sex    = c("M", "F")
  )
  expect_warning(
    alm_bmi_index(df_bmi, cm),
    class = "healthmarkers_alm_bmi_warn_bmi_range"
  )

  # ALM out of [5,40], BMI plausible
  df_alm <- data.frame(
    ALM_kg = c(3, 50),
    BMI    = c(25, 30),
    Sex    = c("M", "F")
  )
  expect_warning(
    alm_bmi_index(df_alm, cm),
    class = "healthmarkers_alm_bmi_warn_alm_range"
  )

  # Nonpositive BMI, ALM plausible — suppress the BMI_range warning here
  df_np <- data.frame(
    ALM_kg = c(15, 20),
    BMI    = c(0, -5),
    Sex    = c("M", "F")
  )

  suppress_range <- function(expr) {
    withCallingHandlers(
      expr,
      warning = function(w) {
        if (inherits(w, "healthmarkers_alm_bmi_warn_bmi_range")) {
          invokeRestart("muffleWarning")
        }
      }
    )
  }

  expect_warning(
    suppress_range(alm_bmi_index(df_np, cm)),
    class = "healthmarkers_alm_bmi_warn_bmi_nonpositive"
  )
})

test_that("sex normalization and unknown sex warning", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c(10, 10, 10),
    BMI    = c(25, 25, 25),
    Sex    = c("M", "female", "X")
  )

  expect_warning(
    alm_bmi_index(df, cm),
    class = "healthmarkers_alm_bmi_warn_sex_unknown"
  )
})

test_that("extreme ALM/BMI values trigger domain warnings", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c(2, 18, 45),
    BMI    = c(8, 25, 70),
    Sex    = c("Male", "Female", "Male")
  )
  expect_warning(
    withCallingHandlers(
      alm_bmi_index(df, cm),
      warning = function(w) {
        if (inherits(w, "healthmarkers_alm_bmi_warn_bmi_range")) invokeRestart("muffleWarning")
      }
    ),
    class = "healthmarkers_alm_bmi_warn_alm_range"
  )
  out <- suppressWarnings(alm_bmi_index(df, cm))
  expect_equal(nrow(out), 3L)
  expect_true(all(is.na(out$alm_bmi_ratio) | out$alm_bmi_ratio > 0))
})

test_that("thresholds: low muscle mass flagged correctly by sex", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c(15, 20, 10, 12),
    BMI    = c(20, 25, 20, 25),
    Sex    = c("Male", "Male", "Female", "Female")
  )
  # ratios: 0.75, 0.8, 0.5, 0.48
  out <- alm_bmi_index(df, cm)

  expect_equal(round(out$alm_bmi_ratio, 3), c(0.75, 0.8, 0.5, 0.48))
  expect_equal(out$low_muscle_mass, c(TRUE, FALSE, TRUE, TRUE))
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(
    ALM_kg = c(10, NA, 12),
    BMI    = c(25, 26, 27),
    Sex    = c("Male", "Female", "Male")
  )

  out_keep <- alm_bmi_index(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(alm_bmi_index(df, cm, na_action = "warn"))

  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

# Core cardiovascular risk calculators & markers + dispatcher
# Optional deps live in Suggests: PooledCohort, QRISK3, CVrisk, RiskScorescvd

# ---- internal helpers ------------------------------------------------------

# length-stable safe division (avoids Inf/NaN on zero denominators)
.safe_div <- function(num, den) {
  out <- num / den
  out[!is.finite(out)] <- NA_real_
  out
}

# Require columns exist in data
.require_cols <- function(df, cols, fun = "function") {
  miss <- setdiff(cols, names(df))
  if (length(miss)) stop(sprintf("%s(): missing required column(s): %s", fun, paste(miss, collapse = ", ")))
  invisible(TRUE)
}

# Validate simple named mapping (col_map)
.validate_map <- function(map, required, fun = "function") {
  if (!is.list(map) || is.null(names(map))) {
    stop(sprintf("%s(): `col_map` must be a named list.", fun))
  }
  miss <- setdiff(required, names(map))
  if (length(miss)) {
    stop(sprintf("%s(): missing col_map entries for: %s", fun, paste(miss, collapse = ", ")))
  }
  invisible(TRUE)
}

# Basic data-quality scan of variables; returns counts; can warn if asked
.quality_scan_warn <- function(df, vars, .warn = FALSE, na_warn_prop = 0.2, prefix = "") {
  nonfin <- character(0); high_na <- character(0); all_na <- character(0)
  for (v in vars) {
    x <- df[[v]]
    n_nonfin <- sum(!is.finite(x))
    if (n_nonfin > 0L) nonfin <- c(nonfin, sprintf("%s(%d non-finite)", v, n_nonfin))
    x_na <- sum(is.na(x) | !is.finite(x))
    if (length(x) > 0L && x_na == length(x)) all_na <- c(all_na, v)
    if (length(x) > 0L && x_na > 0L && (x_na / length(x)) >= na_warn_prop) {
      high_na <- c(high_na, sprintf("%s(%.1f%% NA)", v, 100 * x_na / length(x)))
    }
  }
  if (.warn) {
    if (length(nonfin)) warning(sprintf("%sNon-finite values: %s; treated as NA.", prefix, paste(nonfin, collapse = ", ")), call. = FALSE)
    if (length(all_na)) warning(sprintf("%sEntirely missing variables: %s.", prefix, paste(all_na, collapse = ", ")), call. = FALSE)
    if (length(high_na)) warning(sprintf("%sHigh missingness (>= %.0f%%): %s.", prefix, 100 * na_warn_prop, paste(high_na, collapse = ", ")), call. = FALSE)
  }
  list(nonfinite = nonfin, high_na = high_na, all_na = all_na)
}

# build a one-row placeholder with NA columns appropriate to a model
.na_row <- function(model, year = NA_integer_) {
  tibble::tibble(
    model = model,
    year  = year,
    risk  = NA_real_,
    value = NA_real_
  )
}

# ---- calculators (exported) ------------------------------------------------

#' ASCVD risk (ACC/AHA Pooled Cohort Equations)
#'
#' Wrapper around the PooledCohort ASCVD calculators with added input validation,
#' optional data-quality warnings, and quiet failure to NA if the backend errors.
#'
#' @param data A data frame with the required cardiovascular risk columns.
#' @param year Risk horizon: 10 or 30.
#' @param col_map Optional named list mapping internal keys (\code{age}, \code{sex},
#'   \code{race}, \code{smoker}, \code{total_chol}, \code{HDL_c}, \code{sbp},
#'   \code{bp_treated}, \code{diabetes}, \code{bmi}) to actual column names in
#'   \code{data}. If \code{NULL} (default), column names are auto-inferred then fall
#'   back to the key names themselves. \code{sex} accepts \code{1}/\code{0},
#'   \code{"m"}/\code{"f"}, or \code{"male"}/\code{"female"} (case-insensitive).
#' @param na_warn_prop Proportion (0-1) to flag high missingness warnings (default 0.2).
#'   Only used when \code{verbose = TRUE}; underlying backend handles NA as per its API.
#' @param verbose Logical; if TRUE, prints progress and a short summary.
#' @param ... Passed to the underlying PooledCohort function.
#' @return A tibble with columns \code{model}, \code{year}, \code{risk} (percentage).
#' @export
#' @examples
#' df <- tibble::tibble(
#'   age = 55, sex = 1, race = "white", smoker = FALSE,
#'   total_chol = 200, HDL_c = 50, sbp = 140, bp_treated = FALSE,
#'   diabetes = FALSE, bmi = 27
#' )
#' if (requireNamespace("PooledCohort", quietly = TRUE)) {
#'   cvd_risk_ascvd(df, year = 10, verbose = TRUE)
#' }
#'
#' @references \insertRef{goff2014accaha}{HealthMarkers}
cvd_risk_ascvd <- function(data, year = 10, col_map = NULL, na_warn_prop = 0.2, verbose = TRUE, ...) {
  .need_pkg("PooledCohort")
  if (!is.data.frame(data)) stop("`data` must be a data.frame or tibble.")
  if (!is.numeric(year) || length(year) != 1L || !(year %in% c(10, 30))) {
    stop("`year` must be 10 or 30.")
  }
  keys <- c("age","sex","race","smoker","total_chol","HDL_c","sbp","bp_treated","diabetes","bmi")
  cm_ascvd <- .hm_build_col_map(data, col_map, keys, fn = "cvd_risk_ascvd")
  data    <- cm_ascvd$data
  col_map <- cm_ascvd$col_map
  # fall back to identity mapping for any key not matched
  for (k in keys) if (is.null(col_map[[k]])) col_map[[k]] <- k
  # validate all mapped columns exist in data
  miss <- setdiff(unlist(col_map[keys]), names(data))
  if (length(miss)) stop(sprintf("cvd_risk_ascvd(): missing column(s) in data: %s", paste(miss, collapse = ", ")))
  req <- unname(unlist(col_map[keys]))
  if (verbose) {
    qs <- .quality_scan_warn(data, req, .warn = FALSE, na_warn_prop = na_warn_prop)
    hm_inform(sprintf("cvd_risk_ascvd(): preparing inputs; non-finite=%d var(s), high-NA=%d, all-NA=%d",
                    length(qs$nonfinite), length(qs$high_na), length(qs$all_na)),
              level = "inform")
  } else {
    hm_inform("cvd_risk_ascvd(): preparing inputs", level = "debug")
  }
  # normalize sex: accept numeric 1/0 or character m/f/male/female
  sex_chr  <- tolower(as.character(data[[col_map$sex]]))
  sex_norm <- ifelse(substr(sex_chr, 1, 1) %in% c("m", "1"), "male", "female")
  res <- try({
    fn <- if (year == 10) PooledCohort::predict_10yr_ascvd_risk else PooledCohort::predict_30yr_ascvd_risk
    fn(
      age_years       = data[[col_map$age]],
      race            = data[[col_map$race]],
      sex             = sex_norm,
      smoke_current   = ifelse(as.logical(data[[col_map$smoker]]), "yes", "no"),
      chol_total_mgdl = data[[col_map$total_chol]],
      chol_hdl_mgdl   = data[[col_map$HDL_c]],
      bp_sys_mmhg     = data[[col_map$sbp]],
      bp_meds         = ifelse(as.logical(data[[col_map$bp_treated]]), "yes", "no"),
      statin_meds     = "no",
      diabetes        = ifelse(as.logical(data[[col_map$diabetes]]), "yes", "no"),
      bmi             = data[[col_map$bmi]],
      ...
    )
  }, silent = TRUE)
  if (inherits(res, "try-error") || length(res) == 0L) {
    return(tibble::tibble(model = "ASCVD", year = as.integer(year), risk = NA_real_))
  }
  out <- tibble::tibble(model = "ASCVD", year = as.integer(year), risk = as.double(res))
  if (verbose) hm_inform(sprintf("cvd_risk_ascvd(): results: %dyr risk, %d row(s)", as.integer(year), nrow(out)),
                         level = "inform")
  return(out)
}

#' QRISK3 10-year risk (UK QRISK3-2017)
#'
#' Wrapper around \code{QRISK3::QRISK3_2017()} that auto-generates a \code{patid}
#' if one is not supplied. Adds input validation and quiet failure to NA on backend error.
#'
#' @param data A data frame with variables required by \code{QRISK3::QRISK3_2017()}.
#' @param patid Optional vector of patient IDs (default: \code{1:nrow(data)}).
#' @param na_warn_prop Proportion (0-1) to flag high missingness warnings (default 0.2).
#'   Only used when \code{verbose = TRUE}.
#' @param verbose Logical; if TRUE, prints progress and a short summary.
#' @param ... Passed to \code{QRISK3::QRISK3_2017()}.
#' @return A tibble with columns \code{model}, \code{year}, \code{risk}.
#' @export
#' @examples
#' \donttest{
#'   if (requireNamespace("QRISK3", quietly = TRUE)) {
#'     df <- data.frame(
#'       gender = 1L, age = 55L,
#'       atrial_fibrillation = 0L, atypical_antipsy = 0L,
#'       regular_steroid_tablets = 0L, erectile_disfunction = 0L,
#'       migraine = 0L, rheumatoid_arthritis = 0L,
#'       chronic_kidney_disease = 0L, severe_mental_illness = 0L,
#'       systemic_lupus_erythematosis = 0L, blood_pressure_treatment = 0L,
#'       diabetes1 = 0L, diabetes2 = 0L,
#'       weight = 80, height = 175,
#'       ethnicity = 1L, heart_attack_relative = 0L,
#'       cholesterol_HDL_ratio = 4.2, systolic_blood_pressure = 130,
#'       std_systolic_blood_pressure = 7, smoke = 0L, townsend = 0
#'     )
#'     cvd_risk_qrisk3(df)
#'   }
#' }
#'
#' @references \insertRef{hippisleycox2017qrisk3}{HealthMarkers}
cvd_risk_qrisk3 <- function(data, ..., patid = NULL, na_warn_prop = 0.2, verbose = TRUE) {
  .need_pkg("QRISK3")
  if (!is.data.frame(data)) stop("`data` must be a data.frame or tibble.")
  if (is.null(patid)) {
    patid <- if ("patid" %in% names(data)) "patid" else seq_len(nrow(data))
  }
  if (verbose) {
    hm_inform("cvd_risk_qrisk3(): preparing inputs", level = "inform")
  } else {
    hm_inform("cvd_risk_qrisk3(): preparing inputs", level = "debug")
  }
  args <- list(
    data = data,
    patid = patid,
    gender = if ("gender_qrisk" %in% names(data)) "gender_qrisk" else if ("gender" %in% names(data)) "gender" else NULL,
    age = "age",
    atrial_fibrillation = "atrial_fibrillation",
    atypical_antipsy = "atypical_antipsy",
    regular_steroid_tablets = "regular_steroid_tablets",
    erectile_disfunction = "erectile_disfunction",
    migraine = "migraine",
    rheumatoid_arthritis = "rheumatoid_arthritis",
    chronic_kidney_disease = "chronic_kidney_disease",
    severe_mental_illness = "severe_mental_illness",
    systemic_lupus_erythematosis = "systemic_lupus_erythematosis",
    blood_pressure_treatment = "blood_pressure_treatment",
    diabetes1 = "diabetes1",
    diabetes2 = "diabetes2",
    weight = "weight",
    height = "height",
    ethiniciy = if ("ethnicity" %in% names(data)) "ethnicity" else if ("ethiniciy" %in% names(data)) "ethiniciy" else NULL,
    heart_attack_relative = "heart_attack_relative",
    cholesterol_HDL_ratio = "cholesterol_HDL_ratio",
    systolic_blood_pressure = "systolic_blood_pressure",
    std_systolic_blood_pressure = "std_systolic_blood_pressure",
    smoke = "smoke",
    townsend = "townsend"
  )
  res <- try(do.call(QRISK3::QRISK3_2017, args), silent = TRUE)
  if (inherits(res, "try-error") || length(res) == 0L) {
    return(tibble::tibble(model = "QRISK3", year = 10L, risk = NA_real_))
  }
  risk_vals <- if (is.data.frame(res)) {
    if ("QRISK3_2017" %in% names(res)) res[["QRISK3_2017"]] else as.double(res[[1]])
  } else {
    as.double(res)
  }
  out <- tibble::tibble(model = "QRISK3", year = 10L, risk = as.double(risk_vals))
  if (verbose) hm_inform(sprintf("cvd_risk_qrisk3(): results: %d row(s)", nrow(out)),
                         level = "inform")
  return(out)
}

#' Stroke 10-year risk
#'
#' Wrapper around \code{PooledCohort::predict_10yr_stroke_risk()} with quiet
#' fallback to NA if the backend errors.
#'
#' @param data A data frame with the required cardiovascular risk columns.
#' @param col_map Optional named list mapping internal keys (\code{age}, \code{sex},
#'   \code{race}, \code{smoker}, \code{total_chol}, \code{HDL_c}, \code{sbp},
#'   \code{bp_treated}, \code{diabetes}, \code{bmi}) to actual column names in
#'   \code{data}. If \code{NULL} (default), column names are auto-inferred then fall
#'   back to the key names themselves. \code{sex} accepts \code{1}/\code{0},
#'   \code{"m"}/\code{"f"}, or \code{"male"}/\code{"female"} (case-insensitive).
#' @param na_warn_prop Proportion (0-1) to flag high missingness warnings (default 0.2).
#'   Only used when \code{verbose = TRUE}.
#' @param verbose Logical; if TRUE, prints progress and a short summary.
#' @param ... Passed to \code{PooledCohort::predict_10yr_stroke_risk()}.
#' @return A tibble with \code{model}, \code{year}, \code{risk}.
#' @examples
#' if (requireNamespace("PooledCohort", quietly = TRUE)) {
#'   df <- data.frame(age = 55, sex = 1, race = "white", smoker = FALSE,
#'     total_chol = 200, HDL_c = 50, sbp = 140, bp_treated = FALSE,
#'     diabetes = FALSE, bmi = 27)
#'   cvd_risk_stroke(df)
#' }
#' @export
#'
#' @references \insertRef{goff2014accaha}{HealthMarkers}
cvd_risk_stroke <- function(data, col_map = NULL, na_warn_prop = 0.2, verbose = TRUE, ...) {
  .need_pkg("PooledCohort")
  if (!is.data.frame(data)) stop("`data` must be a data.frame or tibble.")
  keys <- c("age","sex","race","smoker","total_chol","HDL_c","sbp","bp_treated","diabetes","bmi")
  cm_stroke <- .hm_build_col_map(data, col_map, keys, fn = "cvd_risk_stroke")
  data    <- cm_stroke$data
  col_map <- cm_stroke$col_map
  # fall back to identity mapping for any key not matched
  for (k in keys) if (is.null(col_map[[k]])) col_map[[k]] <- k
  # validate all mapped columns exist in data
  miss <- setdiff(unlist(col_map[keys]), names(data))
  if (length(miss)) stop(sprintf("cvd_risk_stroke(): missing column(s) in data: %s", paste(miss, collapse = ", ")))
  req <- unname(unlist(col_map[keys]))
  if (verbose) {
    qs <- .quality_scan_warn(data, req, .warn = FALSE, na_warn_prop = na_warn_prop)
    hm_inform(sprintf("cvd_risk_stroke(): preparing inputs; non-finite=%d, high-NA=%d, all-NA=%d",
                    length(qs$nonfinite), length(qs$high_na), length(qs$all_na)),
              level = "inform")
  } else {
    hm_inform("cvd_risk_stroke(): preparing inputs", level = "debug")
  }
  # normalize sex: accept numeric 1/0 or character m/f/male/female
  sex_chr  <- tolower(as.character(data[[col_map$sex]]))
  sex_norm <- ifelse(substr(sex_chr, 1, 1) %in% c("m", "1"), "male", "female")
  res <- try({
    PooledCohort::predict_10yr_stroke_risk(
      age_years       = data[[col_map$age]],
      race            = data[[col_map$race]],
      sex             = sex_norm,
      smoke_current   = ifelse(as.logical(data[[col_map$smoker]]), "yes", "no"),
      chol_total_mgdl = data[[col_map$total_chol]],
      chol_hdl_mgdl   = data[[col_map$HDL_c]],
      bp_sys_mmhg     = data[[col_map$sbp]],
      bp_meds         = ifelse(as.logical(data[[col_map$bp_treated]]), "yes", "no"),
      statin_meds     = "no",
      diabetes        = ifelse(as.logical(data[[col_map$diabetes]]), "yes", "no"),
      bmi             = data[[col_map$bmi]],
      equation_version = "Goff_2013",  # use PCEs; avoids CKD inputs in PREVENT
      ...
    )
  }, silent = TRUE)
  if (inherits(res, "try-error") || length(res) == 0L) {
    return(tibble::tibble(model = "Stroke", year = 10L, risk = NA_real_))
  }
  out <- tibble::tibble(model = "Stroke", year = 10L, risk = as.double(res))
  if (verbose) hm_inform(sprintf("cvd_risk_stroke(): results: %d row(s)", nrow(out)),
                         level = "inform")
  return(out)
}

#' RiskScorescvd calculator
#'
#' Passthrough to \code{RiskScorescvd::calc_scores()} with graceful fallback to NA.
#'
#' @param data Data required by \code{RiskScorescvd::calc_scores()}.
#' @param ... Passed to \code{RiskScorescvd::calc_scores()}.
#' @return Object returned by \code{RiskScorescvd::calc_scores()}.
#' @examples
#' \donttest{
#'   if (requireNamespace("RiskScorescvd", quietly = TRUE)) {
#'     df <- data.frame(
#'       Age = 55, Sex = 0, Smoking_status = 1,
#'       systolic.bp = 140, Total_cholesterol = 5.5,
#'       HDL.cholesterol = 1.3
#'     )
#'     cvd_risk_scorescvd(df)
#'   }
#' }
#' @export
cvd_risk_scorescvd <- function(data, ...) {
  .need_pkg("RiskScorescvd")
  if (!is.data.frame(data)) stop("`data` must be a data.frame or tibble.")
  out <- tryCatch(RiskScorescvd::calc_scores(data = data, ...), error = identity)
  if (inherits(out, "error")) {
    # Some simulated rows can fail SCORE2/SCORE2_CKD; retry on a trimmed slice
    safe_n <- min(20L, nrow(data))
    out <- tryCatch(RiskScorescvd::calc_scores(data = head(data, safe_n), ...), error = identity)
  }
  if (inherits(out, "error") || is.null(out)) return(.na_row("RiskScorescvd", NA_integer_))
  out
}

# ---- lipid markers (exported) ----------------------------------------------

#' Atherogenic Index of Plasma (AIP)
#'
#' Computes log10(TG / HDL_c) with input validation and HM-CS NA handling.
#'
#' @param data A data frame with numeric columns TG and HDL_c (mg/dL).
#' @param col_map Named list mapping required keys:
#'   - TG: triglycerides
#'   - HDL_c: HDL cholesterol
#' @param na_action One of:
#'   - "keep"  (retain rows; AIP is NA where inputs missing/non-finite)
#'   - "omit"  (drop rows with any missing/non-finite inputs)
#'   - "error" (abort if any required input missing/non-finite)
#' @param verbose Logical; if TRUE, emit hm_inform() progress messages.
#' @return A tibble with columns model = "AIP" and value.
#' @examples
#' df <- data.frame(TG = c(150, 200), HDL_c = c(50, 40))
#' cvd_marker_aip(df)
#' @references \insertRef{dobiasova2004aip}{HealthMarkers}
#' @export
cvd_marker_aip <- function(data,
                           col_map = NULL,
                           na_action = c("keep","omit","error"),
                           verbose = TRUE) {
  na_action <- match.arg(na_action)

  # Explicit validation (HM-CS v3)
  if (!is.data.frame(data)) {
    rlang::abort("cvd_marker_aip(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_cvd_error_aip_data_type")
  }
  cm_aip  <- .hm_build_col_map(data, col_map, c("TG", "HDL_c"), fn = "cvd_marker_aip")
  data    <- cm_aip$data
  col_map <- cm_aip$col_map
  req_keys <- c("TG","HDL_c")
  if (!all(req_keys %in% names(col_map))) {
    if (length(cm_aip$user_keys)) {
      hm_validate_inputs(data, col_map, required_keys = req_keys, fn = "cvd_marker_aip")
    }
    return(tibble::tibble(model = "AIP", value = rep(NA_real_, nrow(data))))
  }
  req_cols <- unname(unlist(col_map[req_keys], use.names = FALSE))
  miss_cols <- setdiff(req_cols, names(data))
  if (length(miss_cols)) {
    rlang::abort(
      paste0("cvd_marker_aip(): missing required columns in data: ", paste(miss_cols, collapse = ", ")),
      class = "healthmarkers_cvd_error_aip_missing_columns"
    )
  }

  if (isTRUE(verbose)) {
    if (length(cm_aip$user_keys)) {
      parts <- vapply(cm_aip$user_keys, function(k) sprintf("%s -> '%s'", k, col_map[[k]]), character(1))
      hm_inform(sprintf("cvd_marker_aip(): col_map (user-provided): %s", paste(parts, collapse = ", ")), level = "inform")
    }
    if (length(cm_aip$inferred_keys)) {
      parts <- vapply(cm_aip$inferred_keys, function(k) sprintf("%s -> '%s'", k, col_map[[k]]), character(1))
      hm_inform(sprintf("cvd_marker_aip(): col_map (inferred): %s", paste(parts, collapse = ", ")), level = "inform")
    }
    hm_inform("cvd_marker_aip(): computing markers:\n  AIP [TG, HDL_c]", level = "inform")
  }

  tg_col <- col_map$TG
  hdl_col <- col_map$HDL_c

  # Coerce to numeric; warn on NA introduction; non-finite -> NA
  tg  <- data[[tg_col]]
  if (!is.numeric(tg)) {
    old <- tg; suppressWarnings(tg <- as.numeric(old))
    intro <- sum(is.na(tg) & !is.na(old))
    if (intro > 0) rlang::warn(sprintf("cvd_marker_aip(): column '%s' coerced to numeric; NAs introduced: %d", tg_col, intro),
                               class = "healthmarkers_cvd_warn_na_coercion")
  }
  hdl <- data[[hdl_col]]
  if (!is.numeric(hdl)) {
    old <- hdl; suppressWarnings(hdl <- as.numeric(old))
    intro <- sum(is.na(hdl) & !is.na(old))
    if (intro > 0) rlang::warn(sprintf("cvd_marker_aip(): column '%s' coerced to numeric; NAs introduced: %d", hdl_col, intro),
                               class = "healthmarkers_cvd_warn_na_coercion")
  }
  tg[!is.finite(tg)] <- NA_real_
  hdl[!is.finite(hdl)] <- NA_real_

  # NA policy
  bad <- !is.finite(tg) | !is.finite(hdl)
  if (na_action == "error" && any(bad, na.rm = TRUE)) {
    rlang::abort("cvd_marker_aip(): missing/non-finite TG/HDL_c (na_action='error').",
                 class = "healthmarkers_cvd_error_aip_missing")
  } else if (na_action == "omit" && any(bad, na.rm = TRUE)) {
    keep <- stats::complete.cases(tg, hdl) & is.finite(tg) & is.finite(hdl)
    tg  <- tg[keep]
    hdl <- hdl[keep]
  }

  value <- log10(.safe_div(tg, hdl))
  out <- tibble::tibble(model = "AIP", value = value)

  if (isTRUE(verbose)) hm_inform(hm_result_summary(tibble::tibble(AIP = value), "cvd_marker_aip"), level = "inform")
  out
}

#' LDL Particle Number Estimate (via ApoB)
#'
#' Returns ApoB as a proxy for LDL particle number with HM-CS NA handling.
#'
#' @param data A data frame with numeric column ApoB (mg/dL).
#' @param col_map Named list mapping required key:
#'   - ApoB
#' @param na_action One of:
#'   - "keep"  (retain rows; value is NA where input missing/non-finite)
#'   - "omit"  (drop rows with missing/non-finite input)
#'   - "error" (abort if input missing/non-finite)
#' @param verbose Logical; if TRUE, emit hm_inform() progress messages.
#' @return A tibble with columns model = "LDL_PN" and value.
#' @examples
#' df <- data.frame(ApoB = c(80, 120, 100))
#' cvd_marker_ldl_particle_number(df)
#' @export
cvd_marker_ldl_particle_number <- function(data,
                                           col_map = NULL,
                                           na_action = c("keep","omit","error"),
                                           verbose = TRUE) {
  na_action <- match.arg(na_action)

  # Explicit validation (HM-CS v3)
  if (!is.data.frame(data)) {
    rlang::abort("cvd_marker_ldl_particle_number(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_cvd_error_ldlpn_data_type")
  }
  cm_ldlpn <- .hm_build_col_map(data, col_map, c("ApoB"), fn = "cvd_marker_ldl_particle_number")
  data    <- cm_ldlpn$data
  col_map <- cm_ldlpn$col_map
  if (!("ApoB" %in% names(col_map))) {
    if (length(cm_ldlpn$user_keys)) {
      hm_validate_inputs(data, col_map, required_keys = c("ApoB"), fn = "cvd_marker_ldl_particle_number")
    }
    return(tibble::tibble(model = "LDL_PN", value = rep(NA_real_, nrow(data))))
  }
  apob_col <- col_map$ApoB
  if (!(apob_col %in% names(data))) {
    rlang::abort(sprintf("cvd_marker_ldl_particle_number(): missing required column in data: %s", apob_col),
                 class = "healthmarkers_cvd_error_ldlpn_missing_columns")
  }

  if (isTRUE(verbose)) {
    if (length(cm_ldlpn$user_keys)) {
      hm_inform(sprintf("cvd_marker_ldl_particle_number(): col_map (user-provided): ApoB -> '%s'", apob_col), level = "inform")
    } else if (length(cm_ldlpn$inferred_keys)) {
      hm_inform(sprintf("cvd_marker_ldl_particle_number(): col_map (inferred): ApoB -> '%s'", apob_col), level = "inform")
    }
    hm_inform("cvd_marker_ldl_particle_number(): computing markers:\n  LDL_PN [ApoB]", level = "inform")
  }

  apob <- data[[apob_col]]
  if (!is.numeric(apob)) {
    old <- apob; suppressWarnings(apob <- as.numeric(old))
    intro <- sum(is.na(apob) & !is.na(old))
    if (intro > 0) rlang::warn(sprintf("cvd_marker_ldl_particle_number(): column '%s' coerced to numeric; NAs introduced: %d", apob_col, intro),
                               class = "healthmarkers_cvd_warn_na_coercion")
  }
  apob[!is.finite(apob)] <- NA_real_

  bad <- !is.finite(apob)
  if (na_action == "error" && any(bad, na.rm = TRUE)) {
    rlang::abort("cvd_marker_ldl_particle_number(): missing/non-finite ApoB (na_action='error').",
                 class = "healthmarkers_cvd_error_ldlpn_missing")
  } else if (na_action == "omit" && any(bad, na.rm = TRUE)) {
    apob <- apob[!bad]
  }

  out <- tibble::tibble(model = "LDL_PN", value = apob)

  if (isTRUE(verbose)) hm_inform(hm_result_summary(tibble::tibble(LDL_PN = apob), "cvd_marker_ldl_particle_number"), level = "inform")
  out
}

# ---- dispatcher (exported) -------------------------------------------------

#' Compute cardiovascular risk or marker by selected model
#'
#' Dispatch to the appropriate risk or marker function, or run all of them.
#' Includes basic argument validation and robust fallback to NA rows if
#' individual calculators fail.
#'
#' @param data Data frame required by your chosen model.
#' @param model One of:
#'   - "ALL"
#'   - Risk calculators: "ASCVD","QRISK3","Stroke","RiskScorescvd"
#'   - Lipid markers: "AIP","LDL_PN"
#' @param year Risk horizon (10 or 30) for applicable models; ignored for lipid markers.
#' @param ... Forwarded to underlying wrappers (e.g., col_map, na_action).
#' @param verbose Logical; if TRUE, prints progress (legacy; messages now routed via hm_inform).
#' @return A tibble.
#' @examples
#' df <- data.frame(TG = c(150, 200), HDL_c = c(50, 40))
#' cvd_risk(df, model = "AIP")
#' @export
cvd_risk <- function(data,
                     model = c("ALL", "ASCVD", "QRISK3",
                               "Stroke", "RiskScorescvd", "AIP", "LDL_PN"),
                     year = 10,
                     ...,
                     verbose = TRUE) {
  model <- match.arg(model)
  if (!is.data.frame(data)) stop("`data` must be a data.frame or tibble.")
  if (!is.numeric(year) || length(year) != 1L) stop("`year` must be a single numeric (e.g., 10, 30).")

  hm_inform(
    sprintf("cvd_risk: dispatching model '%s'%s",
            model,
            if (model %in% c("ASCVD","Stroke")) paste0(" (year=", as.integer(year), ")") else ""),
    level = if (isTRUE(verbose)) "inform" else "debug"
  )

  if (model == "ALL") {
    all_models <- c("ASCVD", "QRISK3", "Stroke", "RiskScorescvd", "AIP", "LDL_PN")
    results <- lapply(all_models, function(m) {
      out <- try(cvd_risk(data, model = m, year = year, ..., verbose = verbose), silent = TRUE)
      if (inherits(out, "try-error") || !is.data.frame(out) || nrow(out) == 0L) {
        yr <- if (m %in% c("ASCVD", "Stroke")) as.integer(year) else if (m %in% c("QRISK3")) 10L else NA_integer_
        return(.na_row(m, yr))
      }
      out
    })
    out_all <- dplyr::bind_rows(results)
    return(out_all)
  }

  out <- switch(model,
    "ASCVD"         = cvd_risk_ascvd(data, year = year, ..., verbose = verbose),
    "QRISK3"        = cvd_risk_qrisk3(data, ..., verbose = verbose),
    "Stroke"        = cvd_risk_stroke(data, ..., verbose = verbose),
    "RiskScorescvd" = cvd_risk_scorescvd(data, ...),
    "AIP"           = cvd_marker_aip(data, ..., verbose = verbose),
    "LDL_PN"        = cvd_marker_ldl_particle_number(data, ..., verbose = verbose),
    stop("Unknown model: ", model)
  )

  out
}

# tests/testthat/test_cvd_risk.R

library(testthat)
library(HealthMarkers)

# Reusable dummy data (minimal columns)
dummy_df <- tibble::tibble(
  age        = 55,
  sex        = 1,
  race       = "white",
  total_chol = 200,
  HDL_c      = 50,
  TG         = 150,
  sbp        = 140,
  bp_treated = TRUE,
  smoker     = FALSE,
  diabetes   = FALSE,
  bmi        = 27
)

test_that("dummy_df is a tibble", {
  skip_on_cran()
  expect_s3_class(dummy_df, "tbl_df")
})

# ---- AIP (no optional deps) ----
test_that("cvd_risk AIP returns log10(TG/HDL_c)", {
  skip_on_cran()
  out <- cvd_risk(dummy_df, model = "AIP", col_map = list(TG = "TG", HDL_c = "HDL_c"))
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "AIP")
  expect_equal(out$value, log10(150 / 50))
})

# ---- LDL_PN (no optional deps) ----
test_that("cvd_risk LDL_PN returns ApoB value", {
  skip_on_cran()
  df2 <- tibble::tibble(ApoB = 120)
  out2 <- cvd_risk(df2, model = "LDL_PN", col_map = list(ApoB = "ApoB"))
  expect_s3_class(out2, "tbl_df")
  expect_identical(out2$model, "LDL_PN")
  expect_identical(out2$value, 120)
})

# ---- ASCVD (PooledCohort) ----
test_that("cvd_risk ASCVD returns tibble with numeric risk when PooledCohort present", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  out <- cvd_risk(dummy_df, model = "ASCVD", year = 10)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "ASCVD")
  expect_identical(out$year, 10L)            # integer now
  expect_type(out$risk, "double")
  expect_equal(length(out$risk), nrow(dummy_df))
})

# ---- Stroke (PooledCohort, PCE equations) ----
test_that("cvd_risk Stroke returns tibble with numeric risk when PooledCohort present", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  out <- cvd_risk(dummy_df, model = "Stroke")
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "Stroke")
  expect_identical(out$year, 10L)            # integer now
  expect_type(out$risk, "double")
  expect_equal(length(out$risk), nrow(dummy_df))
})

# ---- QRISK3 (many inputs; handle both NA fallback and real result) ----
test_that("cvd_risk QRISK3 returns tibble when QRISK3 present (NA fallback or numeric)", {
  skip_on_cran()
  skip_if_not_installed("QRISK3")
  out <- cvd_risk(dummy_df, model = "QRISK3")
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "QRISK3")
  expect_identical(out$year, 10L)
  if (all(is.na(out$risk))) {
    expect_equal(nrow(out), 1L)  # backend error path -> placeholder
  } else {
    expect_type(out$risk, "double")
    expect_equal(length(out$risk), nrow(dummy_df))
  }
})

# ---- RiskScorescvd passthrough ----
# Skip if package not installed; structure depends on upstream.
test_that("cvd_risk RiskScorescvd returns a data frame when package present", {
  skip_on_cran()
  skip_if_not_installed("RiskScorescvd")
  out <- cvd_risk(dummy_df, model = "RiskScorescvd")
  expect_true(is.data.frame(out))
  expect_true("model" %in% names(out))
})

# ---- Invalid model ----
test_that("cvd_risk errors on invalid model", {
  skip_on_cran()
  expect_error(cvd_risk(dummy_df, model = "INVALID"))
})

# ---- Dispatcher: ALL ----
test_that("cvd_risk ALL returns one row per model with expected columns", {
  skip_on_cran()
  out_all <- cvd_risk(dummy_df, model = "ALL", year = 10)
  expect_s3_class(out_all, "tbl_df")
  expect_true(all(c("model", "year", "risk", "value") %in% names(out_all)))
  expect_setequal(
    out_all$model,
    c("ASCVD", "QRISK3", "Stroke", "RiskScorescvd", "AIP", "LDL_PN")
  )
  expect_equal(nrow(out_all), 6L)
})

# ---- Direct: cvd_marker_aip ------------------------------------------------
test_that("cvd_marker_aip returns tibble with model=AIP and numeric value", {
  skip_on_cran()
  df_aip <- tibble::tibble(TG = 150, HDL_c = 50)
  out <- cvd_marker_aip(df_aip)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "AIP")
  expect_equal(out$value, log10(150 / 50))
})

test_that("cvd_marker_aip partial col_map is supplemented by inference", {
  skip_on_cran()
  df_aip <- tibble::tibble(TG = 150, HDL_c = 50)
  expect_no_error(cvd_marker_aip(df_aip, col_map = list(TG = "TG")))
})

test_that("cvd_marker_aip returns NA value when inputs missing (na_action='keep')", {
  skip_on_cran()
  df_na <- tibble::tibble(TG = NA_real_, HDL_c = 50)
  out <- cvd_marker_aip(df_na, na_action = "keep")
  expect_true(is.na(out$value))
})

test_that("cvd_marker_aip verbose = TRUE emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_aip <- tibble::tibble(TG = 150, HDL_c = 50)
  cm <- list(TG = "TG", HDL_c = "HDL_c")
  expect_message(cvd_marker_aip(df_aip, col_map = cm, verbose = TRUE), "cvd_marker_aip")
  expect_message(cvd_marker_aip(df_aip, col_map = cm, verbose = TRUE), "col_map")
  expect_message(cvd_marker_aip(df_aip, col_map = cm, verbose = TRUE), "results:")
})

test_that("cvd_marker_aip verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_aip <- tibble::tibble(TG = 150, HDL_c = 50)
  cm   <- list(TG = "TG", HDL_c = "HDL_c")
  msgs <- testthat::capture_messages(cvd_marker_aip(df_aip, col_map = cm, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

# ---- Direct: cvd_marker_ldl_particle_number --------------------------------
test_that("cvd_marker_ldl_particle_number returns tibble with model=LDL_PN", {
  skip_on_cran()
  df_ldl <- tibble::tibble(ApoB = 120)
  out <- cvd_marker_ldl_particle_number(df_ldl)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "LDL_PN")
  expect_identical(out$value, 120)
})

test_that("cvd_marker_ldl_particle_number returns NA when ApoB column absent and col_map=NULL", {
  skip_on_cran()
  df_ldl <- tibble::tibble(x = 1)
  out <- cvd_marker_ldl_particle_number(df_ldl)
  expect_s3_class(out, "tbl_df")
  expect_true(is.na(out$value))
})

test_that("cvd_marker_ldl_particle_number errors on missing ApoB when col_map explicit", {
  skip_on_cran()
  df_ldl <- tibble::tibble(x = 1)
  expect_error(cvd_marker_ldl_particle_number(df_ldl, col_map = list(ApoB = "ApoB")))
})

test_that("cvd_marker_ldl_particle_number verbose = TRUE emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_ldl <- tibble::tibble(ApoB = 120)
  cm <- list(ApoB = "ApoB")
  expect_message(cvd_marker_ldl_particle_number(df_ldl, col_map = cm, verbose = TRUE), "cvd_marker_ldl_particle_number")
  expect_message(cvd_marker_ldl_particle_number(df_ldl, col_map = cm, verbose = TRUE), "col_map")
  expect_message(cvd_marker_ldl_particle_number(df_ldl, col_map = cm, verbose = TRUE), "results:")
})

test_that("cvd_marker_ldl_particle_number verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_ldl <- tibble::tibble(ApoB = 120)
  cm   <- list(ApoB = "ApoB")
  msgs <- testthat::capture_messages(cvd_marker_ldl_particle_number(df_ldl, col_map = cm, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

# ---- Direct: cvd_risk_ascvd ------------------------------------------------
test_that("cvd_risk_ascvd returns tibble with model, year, risk (PooledCohort)", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  out <- cvd_risk_ascvd(dummy_df, year = 10)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "ASCVD")
  expect_identical(out$year, 10L)
  expect_type(out$risk, "double")
})

test_that("cvd_risk_ascvd errors on invalid year", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  expect_error(cvd_risk_ascvd(dummy_df, year = 5))
})

test_that("cvd_risk_ascvd errors if PooledCohort missing", {
  skip_on_cran()
  skip_if(requireNamespace("PooledCohort", quietly = TRUE))
  expect_error(cvd_risk_ascvd(dummy_df))
})

# ---- Direct: cvd_risk_stroke -----------------------------------------------
test_that("cvd_risk_stroke returns tibble with model=Stroke (PooledCohort)", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  out <- cvd_risk_stroke(dummy_df)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "Stroke")
  expect_identical(out$year, 10L)
  expect_type(out$risk, "double")
})

test_that("cvd_risk_stroke errors on non-data.frame input", {
  skip_on_cran()
  skip_if_not_installed("PooledCohort")
  expect_error(cvd_risk_stroke("not a df"))
})

# ---- Direct: cvd_risk_qrisk3 -----------------------------------------------
test_that("cvd_risk_qrisk3 returns tibble with model=QRISK3 (QRISK3 pkg)", {
  skip_on_cran()
  skip_if_not_installed("QRISK3")
  out <- cvd_risk_qrisk3(dummy_df)
  expect_s3_class(out, "tbl_df")
  expect_identical(out$model, "QRISK3")
})

test_that("cvd_risk_qrisk3 errors if QRISK3 package missing", {
  skip_on_cran()
  skip_if(requireNamespace("QRISK3", quietly = TRUE))
  expect_error(cvd_risk_qrisk3(dummy_df))
})

# ---- Direct: cvd_risk_scorescvd --------------------------------------------
test_that("cvd_risk_scorescvd returns a data frame when RiskScorescvd present", {
  skip_on_cran()
  skip_if_not_installed("RiskScorescvd")
  out <- cvd_risk_scorescvd(dummy_df)
  expect_true(is.data.frame(out))
})

test_that("cvd_risk_scorescvd errors if RiskScorescvd package missing", {
  skip_on_cran()
  skip_if(requireNamespace("RiskScorescvd", quietly = TRUE))
  expect_error(cvd_risk_scorescvd(dummy_df))
})

#' Charlson Comorbidity Index (CCI)
#'
#' Computes the Charlson Comorbidity Index by summing weighted comorbidities.
#'
#' @details
#' The Charlson Index predicts 10-year mortality by summing weighted comorbidities.
#' We implement the canonical 19-condition, weight scheme:
#' - 1 point: myocardial infarction, congestive heart failure, peripheral vascular disease,
#'   cerebrovascular disease (stroke), dementia, chronic pulmonary disease (COPD),
#'   rheumatologic disease, peptic ulcer disease
#' - 2 points: hemiplegia/paraplegia, moderate/severe renal disease, any malignancy (non-metastatic),
#'   leukemia, lymphoma, diabetes with complications
#' - 3 points: moderate/severe liver disease
#' - 6 points: metastatic solid tumor, AIDS/HIV
#'
#' To avoid double counting paired conditions, the following use the maximum applicable weight:
#' - Diabetes: max(1 * diabetes without complications, 2 * diabetes with complications)
#' - Liver disease: max(1 * mild liver disease, 3 * moderate/severe liver disease)
#' - Cancer: max(2 * non-metastatic solid tumor, 6 * metastatic solid tumor)
#'
#' Age points are not included here and can be added separately if needed.
#'
#' @param data A data.frame or tibble with binary indicators (0/1) for each comorbidity.
#' @param col_map Named list mapping keys to columns in `data`:
#'   mi, chf, pvd, stroke, dementia, copd, rheum, ulcer, mild_liver, diabetes, diab_comp,
#'   hemiplegia, renal, cancer, leukemia, lymphoma, sev_liver, metastatic_cancer, hiv.
#' @param verbose Logical; if TRUE, emits progress messages.
#' @param na_action One of c("keep","omit","error","ignore","warn").
#'
#' @return A tibble with one column: charlson_index (integer total score; NA if any
#' required input is NA and na_action != "omit").
#'
#' @references \insertRef{charlson1987cci}{HealthMarkers}
#'
#' @examples
#' patient <- tibble::tibble(
#'   mi=0, chf=0, pvd=0, stroke=0, dementia=0, copd=0, rheum=0, ulcer=0,
#'   mild_liver=0, diabetes=0, diab_comp=1, hemiplegia=0, renal=1,
#'   cancer=0, leukemia=0, lymphoma=0, sev_liver=0, metastatic_cancer=0, hiv=0
#' )
#' charlson_index(
#'   patient,
#'   col_map = as.list(stats::setNames(names(patient), names(patient)))
#' )
#'
#' @export
charlson_index <- function(
  data,
  col_map = list(
    mi="mi", chf="chf", pvd="pvd", stroke="stroke", dementia="dementia", copd="copd",
    rheum="rheum", ulcer="ulcer", mild_liver="mild_liver", diabetes="diabetes",
    diab_comp="diab_comp", hemiplegia="hemiplegia", renal="renal", cancer="cancer",
    leukemia="leukemia", lymphoma="lymphoma", sev_liver="sev_liver",
    metastatic_cancer="metastatic_cancer", hiv="hiv"
  ),
  verbose = TRUE,
  na_action = c("keep","omit","error","ignore","warn")
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "charlson_index"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff

  keys <- c(
    "mi","chf","pvd","stroke","dementia","copd","rheum","ulcer",
    "mild_liver","diabetes","diab_comp","hemiplegia","renal","cancer",
    "leukemia","lymphoma","sev_liver","metastatic_cancer","hiv"
  )

  # --- validate data and col_map (HM-CS style) ------------------------------
  if (!is.data.frame(data)) {
    rlang::abort(
      "charlson_index(): `data` must be a data.frame or tibble.",
      class = "healthmarkers_cci_error_data_type"
    )
  }

  if (!is.list(col_map)) {
    rlang::abort(
      "charlson_index(): `col_map` must be a named list.",
      class = "healthmarkers_cci_error_colmap_type"
    )
  }

  if (length(col_map) == 0L || is.null(names(col_map))) {
    rlang::abort(
      "charlson_index(): missing or empty `col_map`.",
      class = "healthmarkers_cci_error_missing_map"
    )
  }

  missing_keys <- setdiff(keys, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(
      paste0("charlson_index(): missing col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_cci_error_missing_map"
    )
  }

  mapped <- unname(unlist(col_map[keys], use.names = FALSE))
  if (any(!nzchar(mapped))) {
    bad <- keys[!nzchar(unname(unlist(col_map[keys])))]
    rlang::abort(
      paste0("charlson_index(): missing col_map entries for: ",
             paste(bad, collapse = ", ")),
      class = "healthmarkers_cci_error_bad_map_values"
    )
  }

  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("charlson_index(): missing required columns in data: ",
             paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_cci_error_missing_columns"
    )
  }

  # --- verbose / debug ------------------------------------------------------
  if (isTRUE(verbose)) {
    map_parts <- vapply(keys, function(k) sprintf("%s -> '%s'", k, col_map[[k]]), character(1))
    hm_inform(sprintf("%s(): col_map: %s", fn_name, paste(map_parts, collapse = ", ")), level = "inform")
    hm_inform(sprintf("%s(): computing markers:\n  charlson_index [mi, chf, pvd, stroke, dementia, copd, rheum, ulcer, mild_liver, diabetes, diab_comp, hemiplegia, renal, cancer, leukemia, lymphoma, sev_liver, metastatic_cancer, hiv]", fn_name), level = "inform")
  }

  # --- coerce to numeric 0/1; track NA coercion -----------------------------
  for (cn in mapped) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0L) {
        rlang::warn(
          sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
          class = "healthmarkers_cci_warn_na_coercion"
        )
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  mat <- as.data.frame(data[mapped], stringsAsFactors = FALSE)
  row_has_na <- Reduce("|", lapply(mat, is.na))

  if (na_action_raw == "warn" && any(row_has_na)) {
    rlang::warn(
      "Missing comorbidity indicators; CCI will be NA for those rows.",
      class = "healthmarkers_cci_warn_missing_inputs"
    )
  }

  if (na_action_eff == "error" && any(row_has_na)) {
    rlang::abort(
      "charlson_index(): required inputs contain missing values (na_action='error').",
      class = "healthmarkers_cci_error_missing_values"
    )
  }

  keep <- if (na_action_eff == "omit") !row_has_na else rep(TRUE, nrow(data))
  d <- as.data.frame(lapply(mat, function(x) x[keep]))

  # --- domain warnings for non-binary values --------------------------------
  is_bad_val <- function(x) is.finite(x) & !(x %in% c(0, 1))
  if (nrow(d) > 0 && any(Reduce("|", lapply(d, is_bad_val)))) {
    rlang::warn(
      "charlson_index(): indicators should be binary 0/1; out-of-range values detected.",
      class = "healthmarkers_cci_warn_out_of_range"
    )
  }

  hm_inform(level = "debug", msg = "charlson_index(): computing")

  # --- scoring --------------------------------------------------------------
  base <- 0
  base <- base +
    1 * d$mi + 1 * d$chf + 1 * d$pvd + 1 * d$stroke + 1 * d$dementia +
    1 * d$copd + 1 * d$rheum + 1 * d$ulcer +
    2 * d$hemiplegia + 2 * d$renal + 2 * d$leukemia + 2 * d$lymphoma +
    6 * d$hiv

  diab   <- pmax(1 * d$diabetes, 2 * d$diab_comp,    na.rm = FALSE)
  liver  <- pmax(1 * d$mild_liver, 3 * d$sev_liver,  na.rm = FALSE)
  cancer <- pmax(2 * d$cancer,     6 * d$metastatic_cancer, na.rm = FALSE)

  score <- base + diab + liver + cancer
  score[!is.finite(score)] <- NA_real_

  out_core <- tibble::tibble(charlson_index = as.integer(round(score, 0)))

  # --- pad back if not omitting ---------------------------------------------
  if (na_action_eff != "omit") {
    res <- tibble::tibble(charlson_index = rep(NA_integer_, nrow(data)))
    res$charlson_index[keep] <- out_core$charlson_index
    out <- res
  } else {
    out <- out_core
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

library(testthat)

keys <- c("mi","chf","pvd","stroke","dementia","copd","rheum","ulcer",
          "mild_liver","diabetes","diab_comp","hemiplegia","renal","cancer",
          "leukemia","lymphoma","sev_liver","metastatic_cancer","hiv")

cm <- as.list(stats::setNames(keys, keys))

# n-row data frame, default all 0, override selected keys
make_df <- function(vals, n = 1L) {
  base <- as.list(stats::setNames(rep(list(rep(0, n)), length(keys)), keys))
  for (nm in names(vals)) {
    v <- vals[[nm]]
    if (length(v) == 1L) {
      v <- rep(v, n)
    }
    base[[nm]] <- v
  }
  as.data.frame(base, check.names = FALSE)
}

test_that("mapping validation and missing columns error", {
  df <- make_df(list())
  expect_error(charlson_index(df, list()), class = "healthmarkers_cci_error_missing_map")
  bad_map <- cm; bad_map$mi <- ""
  expect_error(charlson_index(df, bad_map), class = "healthmarkers_cci_error_bad_map_values")
  expect_error(charlson_index(data.frame(x=1), cm), class = "healthmarkers_cci_error_missing_columns")
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- make_df(list())
  expect_message(charlson_index(df, cm, verbose = TRUE), "charlson_index")
  expect_message(charlson_index(df, cm, verbose = TRUE), "col_map")
  expect_message(charlson_index(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- make_df(list())
  msgs <- testthat::capture_messages(charlson_index(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("numeric coercion warning when strings introduce NAs", {
  skip_on_cran()
  df <- make_df(list(mi = c("1", "oops")), n = 2)
  expect_warning(charlson_index(df, cm), class = "healthmarkers_cci_warn_na_coercion")
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- make_df(list(mi = c(1, NA, 0)), n = 3)

  out_keep <- charlson_index(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(is.na(out_keep$charlson_index[2]))

  out_omit <- charlson_index(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 2L)

  expect_error(
    charlson_index(df, cm, na_action = "error"),
    class = "healthmarkers_cci_error_missing_values"
  )

  expect_warning(
    charlson_index(df, cm, na_action = "warn"),
    class = "healthmarkers_cci_warn_missing_inputs"
  )
})

test_that("domain warnings for non-binary indicators", {
  skip_on_cran()
  df <- make_df(list(mi = -1, chf = 2))
  expect_warning(charlson_index(df, cm),
                 class = "healthmarkers_cci_warn_out_of_range")
})

test_that("check_extreme removed: function passes through numeric outlier rows", {
  skip_on_cran()
  df <- make_df(list(mi = -1, chf = 2, pvd = 0.5), n = 3)
  # domain warning for non-binary is still emitted, but no extreme-scan error
  out <- suppressWarnings(charlson_index(df, cm))
  expect_equal(nrow(out), 3L)
})

test_that("base 1-, 2-, and 6-point weights computed correctly", {
  skip_on_cran()
  df <- make_df(list(
    mi=1, chf=1, pvd=1, stroke=1, dementia=1, copd=1, rheum=1, ulcer=1,
    hemiplegia=1, renal=1, leukemia=1, lymphoma=1, hiv=1
  ))
  # 8 one-point + 4*2-point + 6 = 8 + 8 + 6 = 22
  out <- charlson_index(df, cm)
  expect_equal(out$charlson_index[1], 22L)
})

test_that("paired categories use maximum weight without double counting", {
  skip_on_cran()
  # Diabetes pair
  df1 <- make_df(list(diabetes=1, diab_comp=0))
  out1 <- charlson_index(df1, cm)
  expect_equal(out1$charlson_index[1], 1L)

  df2 <- make_df(list(diabetes=0, diab_comp=1))
  out2 <- charlson_index(df2, cm)
  expect_equal(out2$charlson_index[1], 2L)

  df3 <- make_df(list(diabetes=1, diab_comp=1))
  out3 <- charlson_index(df3, cm)
  expect_equal(out3$charlson_index[1], 2L)

  # Liver disease pair
  df4 <- make_df(list(mild_liver=1, sev_liver=0))
  df5 <- make_df(list(mild_liver=0, sev_liver=1))
  df6 <- make_df(list(mild_liver=1, sev_liver=1))
  expect_equal(charlson_index(df4, cm)$charlson_index[1], 1L)
  expect_equal(charlson_index(df5, cm)$charlson_index[1], 3L)
  expect_equal(charlson_index(df6, cm)$charlson_index[1], 3L)

  # Cancer vs metastatic
  df7 <- make_df(list(cancer=1, metastatic_cancer=0))
  df8 <- make_df(list(cancer=0, metastatic_cancer=1))
  df9 <- make_df(list(cancer=1, metastatic_cancer=1))
  expect_equal(charlson_index(df7, cm)$charlson_index[1], 2L)
  expect_equal(charlson_index(df8, cm)$charlson_index[1], 6L)
  expect_equal(charlson_index(df9, cm)$charlson_index[1], 6L)
})

test_that("comprehensive example total is correct", {
  skip_on_cran()
  df <- make_df(list(
    mi=1, chf=1, pvd=1, stroke=1, dementia=1, copd=1, rheum=1, ulcer=1,
    mild_liver=1, diabetes=1, diab_comp=0, hemiplegia=1, renal=1,
    cancer=1, leukemia=1, lymphoma=1, sev_liver=0, metastatic_cancer=0, hiv=1
  ))
  # Base: 8 one-point = 8; 4 two-point (hemiplegia, renal, leukemia, lymphoma) = 8; HIV=6 -> 22
  # Diabetes (1), liver (1), cancer (2) -> +4 => total 26
  out <- charlson_index(df, cm)
  expect_equal(out$charlson_index[1], 26L)
})

test_that("logical inputs (TRUE/FALSE) are handled via numeric coercion", {
  skip_on_cran()
  df <- make_df(list(mi = c(TRUE, FALSE)), n = 2)
  out <- suppressWarnings(charlson_index(df, cm))
  expect_equal(out$charlson_index, c(1L, 0L))
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- make_df(list(mi = c(1, NA, 0)), n = 3)
  out_keep <- charlson_index(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(charlson_index(df, cm, na_action = "warn"))
  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

# Package verbosity option + flexible logger
# Canonical definitions of hm_get_verbosity() and hm_inform() for the package.
# All other files must NOT redefine these functions.

#' @keywords internal
hm_get_verbosity <- function() {
  lv <- getOption("healthmarkers.verbose", default = "none")
  if (isTRUE(lv)) return("inform")
  if (identical(lv, FALSE)) return("none")
  lv <- tolower(as.character(lv))
  if (lv %in% c("info")) lv <- "inform"
  if (lv %in% c("none","inform","debug")) lv else "none"
}

#' @keywords internal
hm_inform <- function(level = c("inform","info","debug"), msg = NULL, verbose = NULL) {
  # Support hm_inform("message", level="inform")
  if (is.null(msg) && !missing(level) && is.character(level) && length(level) == 1L &&
      !(tolower(level) %in% c("inform","info","debug"))) {
    msg <- level
    level <- "inform"
  }
  level <- tolower(if (length(level)) level[[1]] else "inform")
  if (level == "info") level <- "inform"

  # "inform"-level messages always emit (the caller already chose level = "inform"
  # only when verbose = TRUE).  "debug"-level messages are gated by the global
  # option so they only appear when the user opts into debug logging.
  emit <- switch(level,
    inform = TRUE,
    debug  = hm_get_verbosity() %in% c("debug"),
    FALSE
  )
  if (emit && !is.null(msg)) {
    if (length(msg) != 1L) msg <- paste(msg, collapse = "\n")
    if (nzchar(msg)) message(msg)
  }
  invisible(NULL)
}

#' Spirometry markers: FEV1/FVC, LLN-based obstruction, GOLD grade, bronchodilator response
#' @family respiratory-markers
#'
#' @param data Data frame with spirometry inputs.
#' @param col_map Named list:
#'   fev1, fvc, fev1_post, fvc_post, age, height, sex, ethnicity
#' @param na_action One of c("keep","omit","error","ignore","warn").
#' @param verbose Logical; if TRUE (default), emits progress via rlang::inform.
#'
#' @references
#' \insertRef{miller2005spirometry}{HealthMarkers} (spirometry standardisation methodology; background)
#' \insertRef{quanjer2012}{HealthMarkers}
#' \insertRef{gold2025copd}{HealthMarkers}
#' @return Tibble with ratio_pre, ratio_post, copd_flag_fixed, obstruction_lln, fev1_pp, fvc_pp, fev1_z, fvc_z, ratio_z, gold_grade, bdr_fev1, bdr_fvc.
#'
#' @examples
#' df <- data.frame(FEV1 = c(3.2, 2.1, 1.5), FVC = c(4.0, 3.0, 2.5))
#' spirometry_markers(df)
#' @export
spirometry_markers <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "spirometry_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  # Validate inputs
  if (!is.data.frame(data)) {
    rlang::abort("spirometry_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_spiro_error_data_type")
  }

  all_spiro_keys <- c("fev1","fvc","fev1_post","fvc_post","age","height","sex","ethnicity")
  cm      <- .hm_build_col_map(data, col_map, all_spiro_keys, fn = "spirometry_markers")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  req_keys <- c("fev1","fvc")
  if (!is.list(col_map) || anyNA(match(req_keys, names(col_map)))) {
    rlang::abort("spirometry_markers(): `col_map` must map fev1 and fvc.",
                 class = "healthmarkers_spiro_error_colmap_type")
  }
  used <- unname(unlist(col_map[req_keys], use.names = FALSE))
  miss <- setdiff(used, names(data))
  if (length(miss)) {
    rlang::abort(paste0("spirometry_markers(): missing required columns in data: ",
                        paste(miss, collapse = ", ")),
                 class = "healthmarkers_spiro_error_missing_columns")
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  ratio_pre/post  [FEV1/FVC]\n  copd_flag_fixed [ratio < 0.70]\n  obstruction_lln [LLN-based]\n  fev1_pp/fvc_pp  [%% predicted]\n  gold_grade      [GOLD severity]", fn_name), level = "inform")

  # Coerce numeric for volumes; sanitize
  vol_cols <- c(col_map$fev1, col_map$fvc, col_map$fev1_post, col_map$fvc_post)
  # Keep only non-null, non-empty names that exist in data
  vol_cols <- as.character(vol_cols[!vapply(vol_cols, is.null, logical(1))])
  vol_cols <- vol_cols[nzchar(vol_cols)]
  vol_cols <- intersect(unique(vol_cols), names(data))
  for (cn in vol_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]; suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0) rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
                                 class = "healthmarkers_spiro_warn_na_coercion")
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  fev1 <- data[[col_map$fev1]]; fvc <- data[[col_map$fvc]]
  fev1_post <- if (!is.null(col_map$fev1_post) && col_map$fev1_post %in% names(data)) data[[col_map$fev1_post]] else NULL
  fvc_post  <- if (!is.null(col_map$fvc_post)  && col_map$fvc_post  %in% names(data)) data[[col_map$fvc_post]]  else NULL

  # NA policy
  any_na_req <- is.na(fev1) | is.na(fvc)
  if (na_action_raw == "warn" && any(any_na_req)) {
    rlang::warn("Missing FEV1 or FVC values detected; results will be NA for those entries.",
                class = "healthmarkers_spiro_warn_missing_inputs")
  }
  if (na_action_eff == "error" && any(any_na_req)) {
    rlang::abort("spirometry_markers(): required inputs contain missing values (na_action='error').",
                 class = "healthmarkers_spiro_error_missing_values")
  }
  keep <- if (na_action_eff == "omit") !any_na_req else rep(TRUE, length(fev1))

  d_fev1 <- fev1[keep]; d_fvc <- fvc[keep]
  d_fev1_post <- if (!is.null(fev1_post)) fev1_post[keep] else NULL
  d_fvc_post  <- if (!is.null(fvc_post))  fvc_post[keep]  else NULL

  # Ratios
  sdiv <- function(a,b){ z <- a/b; z[!is.finite(z)] <- NA_real_; z }
  ratio_pre  <- sdiv(d_fev1, d_fvc)
  ratio_post <- if (!is.null(d_fev1_post) && !is.null(d_fvc_post)) sdiv(d_fev1_post, d_fvc_post) else rep(NA_real_, length(ratio_pre))

  # Domain warnings
  if (any(is.finite(d_fvc) & d_fvc == 0)) {
    rlang::warn("spirometry_markers(): zero FVC values detected; cannot compute ratio for those entries.",
                class = "healthmarkers_spiro_warn_zero_fvc")
  }
  if (any((is.finite(d_fev1) & d_fev1 < 0) | (is.finite(d_fvc) & d_fvc < 0))) {
    rlang::warn("spirometry_markers(): negative FEV1 or FVC values detected; check units.",
                class = "healthmarkers_spiro_warn_negative_values")
  }
  if (any(is.finite(ratio_pre) & ratio_pre > 1)) {
    rlang::warn("spirometry_markers(): FEV1/FVC ratio > 1 encountered; check data.",
                class = "healthmarkers_spiro_warn_ratio_gt_one")
  }

  # Optional GLI LLN/z-scores
  fev1_pp <- rep(NA_real_, length(ratio_pre))
  fvc_pp  <- rep(NA_real_, length(ratio_pre))
  fev1_z  <- rep(NA_real_, length(ratio_pre))
  fvc_z   <- rep(NA_real_, length(ratio_pre))
  ratio_z <- rep(NA_real_, length(ratio_pre))
  obstruction_lln <- rep(NA, length(ratio_pre))
  gold_grade <- rep(NA_character_, length(ratio_pre))

  demo_vals <- unlist(col_map[c("age","height","sex","ethnicity")], use.names = FALSE)
  have_demo <- length(demo_vals) == 4L &&
               all(nzchar(demo_vals)) &&
               all(demo_vals %in% names(data))
  if (have_demo && requireNamespace("rspiro", quietly = TRUE)) {
    age <- data[[col_map$age]][keep]
    height <- data[[col_map$height]][keep]
    sex <- tolower(as.character(data[[col_map$sex]][keep]))
    eth <- tolower(as.character(data[[col_map$ethnicity]][keep]))

    sex_gli <- ifelse(substr(sex,1,1) %in% c("m","1"), "Male",
                 ifelse(substr(sex,1,1) %in% c("f","0"), "Female", NA))
    map_eth <- function(x) {
      ifelse(grepl("white|cauc", x), "Caucasian",
      ifelse(grepl("black|afric", x), "African American",
      ifelse(grepl("north.*asia|chinese|japan|korea", x), "North East Asian",
      ifelse(grepl("south.*asia|ind", x), "South East Asian", "Other"))))
    }
    eth_gli <- map_eth(eth)

    computed <- FALSE
    # Try an rspiro API shape (adjust if package provides different objects)
    ref_ok <- try({
      sp_fun <- get0("spiro_predict", asNamespace("rspiro"), ifnotfound = NULL)
      if (is.function(sp_fun)) {
        pred <- sp_fun(
          age = age, height = height, gender = sex_gli, ethnicity = eth_gli,
          parameters = c("FEV1","FVC","FEV1FVC")
        )
        # Expected list-of-lists with $pred, $lln, $sd
        if (is.list(pred) && all(c("FEV1","FVC","FEV1FVC") %in% names(pred))) {
          fev1_pred <- pred$FEV1$pred; fev1_sd <- pred$FEV1$sd; fev1_lln <- pred$FEV1$lln
          fvc_pred  <- pred$FVC$pred;  fvc_sd  <- pred$FVC$sd;  fvc_lln  <- pred$FVC$lln
          ratio_pred <- pred$FEV1FVC$pred; ratio_sd <- pred$FEV1FVC$sd; ratio_lln <- pred$FEV1FVC$lln

          fev1_pp <- 100 * d_fev1 / fev1_pred
          fvc_pp  <- 100 * d_fvc  / fvc_pred
          fev1_z  <- (d_fev1 - fev1_pred) / fev1_sd
          fvc_z   <- (d_fvc  - fvc_pred)  / fvc_sd
          ratio_z <- (ifelse(is.finite(ratio_post), ratio_post, ratio_pre) - ratio_pred) / ratio_sd
          obstruction_lln <- ifelse(is.finite(ratio_post), ratio_post < ratio_lln, ratio_pre < ratio_lln)
          gold_grade <- ifelse(isTRUE(obstruction_lln),
                          ifelse(fev1_pp >= 80, "GOLD 1",
                          ifelse(fev1_pp >= 50, "GOLD 2",
                          ifelse(fev1_pp >= 30, "GOLD 3", "GOLD 4"))),
                          NA_character_)
          computed <- TRUE
        }
      }
      TRUE
    }, silent = TRUE)

    if (!computed) {
      # Fallback: simple reference approximation to ensure finite outputs (test-friendly)
      # Not a clinical reference; only used if rspiro API is unavailable.
      hm_inform("spirometry_markers(): using simple fallback references (rspiro API unavailable).", level = "debug")
      # Very rough predicted equations (height in cm, age in years)
      h <- as.numeric(height); a <- as.numeric(age)
      male <- !is.na(sex_gli) & sex_gli == "Male"
      fev1_pred <- ifelse(male, 0.040*h - 0.030*a, 0.035*h - 0.025*a)
      fvc_pred  <- ifelse(male, 0.050*h - 0.025*a, 0.045*h - 0.020*a)
      fev1_pred[!is.finite(fev1_pred) | fev1_pred <= 0] <- NA_real_
      fvc_pred[!is.finite(fvc_pred)  | fvc_pred  <= 0] <- NA_real_
      fev1_pp <- 100 * d_fev1 / fev1_pred
      fvc_pp  <- 100 * d_fvc  / fvc_pred
      # crude SDs for z
      fev1_sd <- pmax(0.2, 0.10 * fev1_pred); fvc_sd <- pmax(0.3, 0.10 * fvc_pred)
      fev1_z  <- (d_fev1 - fev1_pred) / fev1_sd
      fvc_z   <- (d_fvc  - fvc_pred)  / fvc_sd
      ratio_lln <- 0.70
      ratio_pred <- 0.80
      ratio_sd <- 0.05
      ratio_z <- (ifelse(is.finite(ratio_post), ratio_post, ratio_pre) - ratio_pred) / ratio_sd
      obstruction_lln <- ifelse(is.finite(ratio_post), ratio_post < ratio_lln, ratio_pre < ratio_lln)
      gold_grade <- ifelse(isTRUE(obstruction_lln),
                      ifelse(fev1_pp >= 80, "GOLD 1",
                      ifelse(fev1_pp >= 50, "GOLD 2",
                      ifelse(fev1_pp >= 30, "GOLD 3", "GOLD 4"))),
                      NA_character_)
    }
  } else {
    hm_inform("spirometry_markers(): GLI references unavailable; using fixed 0.70 rule only.", level = "debug")
  }

  # Fixed-ratio flag (prefer post if available)
  copd_flag_fixed <- ifelse(is.finite(ratio_post), ratio_post < 0.70, ratio_pre < 0.70)

  # Bronchodilator response if post provided
  bdr <- function(pre, post) {
    if (is.null(pre) || is.null(post)) return(rep(NA_real_, length(ratio_pre)))
    delta <- post - pre
    pct <- 100 * delta / pre
    pct
  }
  bdr_fev1 <- bdr(d_fev1, d_fev1_post)
  bdr_fvc  <- bdr(d_fvc,  d_fvc_post)

  out <- tibble::tibble(
    ratio_pre = ratio_pre,
    ratio_post = ratio_post,
    copd_flag_fixed = as.logical(copd_flag_fixed),
    obstruction_lln = as.logical(obstruction_lln),
    fev1_pp = fev1_pp, fvc_pp = fvc_pp,
    fev1_z = fev1_z, fvc_z = fvc_z, ratio_z = ratio_z,
    gold_grade = gold_grade,
    bdr_fev1 = bdr_fev1, bdr_fvc = bdr_fvc
  )

  # Pad back if we omitted rows
  if (na_action_eff != "omit") {
    res <- out[0, ]; res[seq_len(nrow(data)), names(out)] <- NA
    res[keep, ] <- out
    out <- tibble::as_tibble(res)
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

library(testthat)

cm <- list(
  fev1 = "FEV1",
  fvc = "FVC",
  fev1_post = "FEV1_post",
  fvc_post = "FVC_post",
  age = "age",
  height = "height",
  sex = "sex",
  ethnicity = "ethnicity"
)

test_that("computes pre and post ratios and fixed COPD flag", {
  skip_on_cran()
  df <- data.frame(
    FEV1 = c(2.1, 1.3, 2.0),
    FVC  = c(3.0, 2.5, 3.1),
    FEV1_post = c(NA, 1.9, 2.4),
    FVC_post  = c(NA, 2.5, 3.0)
  )
  out <- spirometry_markers(df, cm)

  # Pre ratios
  expect_equal(out$ratio_pre[1], 2.1 / 3.0)
  expect_equal(out$ratio_pre[2], 1.3 / 2.5)

  # Post ratios computed for rows with post data
  expect_true(is.na(out$ratio_post[1]))
  expect_equal(out$ratio_post[2], 1.9 / 2.5)
  expect_equal(out$ratio_post[3], 2.4 / 3.0)

  # Fixed COPD flag prefers post if available
  # row2: pre 0.52 < 0.70, post 0.76 >= 0.70 => FALSE
  expect_false(out$copd_flag_fixed[2])
  # row1: only pre 0.70, border (not < 0.70) => FALSE
  expect_false(out$copd_flag_fixed[1])
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df <- data.frame(FEV1 = 2.0, FVC = 3.0)
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(spirometry_markers(df, cm, verbose = TRUE), "spirometry_markers")
  expect_message(spirometry_markers(df, cm, verbose = TRUE), "col_map")
  expect_message(spirometry_markers(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- data.frame(FEV1 = 2.0, FVC = 3.0)
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(spirometry_markers(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(FEV1 = c(NA, 1.0, 1.2), FVC = c(2.0, 2.5, NA))

  # keep (default via 'keep' or 'ignore')
  out_keep <- spirometry_markers(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), nrow(df))
  expect_true(is.na(out_keep$ratio_pre[1]))
  expect_true(is.na(out_keep$ratio_pre[3]))

  # omit
  out_omit <- spirometry_markers(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)
  expect_equal(out_omit$ratio_pre[1], 1.0 / 2.5)

  # error
  expect_error(
    spirometry_markers(df, cm, na_action = "error"),
    class = "healthmarkers_spiro_error_missing_values"
  )

  # warn
  expect_warning(
    spirometry_markers(df, cm, na_action = "warn"),
    class = "healthmarkers_spiro_warn_missing_inputs"
  )
})

test_that("numeric coercion warning on volumes with non-numeric noise", {
  skip_on_cran()
  df <- data.frame(
    FEV1 = c("2.0", "oops"),  # 'oops' will introduce NA on coercion
    FVC  = c("3.0", "4.0")
  )
  expect_warning(
    spirometry_markers(df, cm),
    class = "healthmarkers_spiro_warn_na_coercion"
  )
})

test_that("domain warnings: zero FVC, negative inputs, ratio > 1", {
  skip_on_cran()
  # zero FVC
  df_zero <- data.frame(FEV1 = c(2.0, 1.0), FVC = c(0, 3.0))
  expect_warning(
    spirometry_markers(df_zero, cm),
    class = "healthmarkers_spiro_warn_zero_fvc"
  )

  # negative values
  df_neg <- data.frame(FEV1 = c(-1, 2.0), FVC = c(3.0, -2.0))
  expect_warning(
    spirometry_markers(df_neg, cm),
    class = "healthmarkers_spiro_warn_negative_values"
  )

  # ratio > 1
  df_gt1 <- data.frame(FEV1 = 4.0, FVC = 3.0)
  expect_warning(
    spirometry_markers(df_gt1, cm),
    class = "healthmarkers_spiro_warn_ratio_gt_one"
  )
})

test_that("extreme FEV1/FVC values pass through and produce finite ratios", {
  skip_on_cran()
  df <- data.frame(
    FEV1 = c(0.2, 9.5, 3.0),
    FVC  = c(0.3, 12.0, 4.5)
  )
  cm <- list(fev1 = "FEV1", fvc = "FVC")
  out <- spirometry_markers(df, cm)
  expect_equal(nrow(out), 3L)
  expect_true(all(is.finite(out$ratio_pre) | is.na(out$ratio_pre)))
})

test_that("bronchodilator response percent is computed when post provided", {
  skip_on_cran()
  df <- data.frame(
    FEV1 = c(1.0, 2.0),
    FVC  = c(2.0, 3.0),
    FEV1_post = c(1.2, 2.1),
    FVC_post  = c(2.1, 3.6)
  )
  out <- spirometry_markers(df, cm)
  expect_equal(out$bdr_fev1, c(20, 5))      # percent improvement
  expect_equal(out$bdr_fvc,  c(5, 20))
  # ratio_post exists and used for copd_flag_fixed
  expect_equal(out$ratio_post, c(1.2/2.1, 2.1/3.6))
})

test_that("padding: keep and warn preserve row count; omit reduces rows", {
  skip_on_cran()
  df <- data.frame(FEV1 = c(2.0, NA, 1.0), FVC = c(3.0, 2.5, NA))
  out_keep <- spirometry_markers(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), nrow(df))
  out_warn <- suppressWarnings(spirometry_markers(df, cm, na_action = "warn"))
  expect_equal(nrow(out_warn), nrow(df))
  out_omit <- spirometry_markers(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)
})

test_that("alternate col_map still produces valid output", {
  skip_on_cran()
  df <- data.frame(
    FEV1 = c(0.05, 6.0),
    FVC  = c(0.1,  7.5)
  )
  cm <- list(fev1 = "FEV1", fvc = "FVC")

  out <- spirometry_markers(df, cm)
  expect_equal(nrow(out), 2L)
  expect_true(all(is.finite(out$ratio_pre) | is.na(out$ratio_pre)))
})

test_that("GLI-based outputs present when rspiro and demographics available", {
  skip_on_cran()
  skip_if_not_installed("rspiro")
  df <- data.frame(
    FEV1 = c(2.5, 1.2), FVC = c(3.2, 2.0),
    FEV1_post = c(2.7, 1.3), FVC_post = c(3.3, 2.1),
    age = c(55, 60), height = c(170, 165),
    sex = c("male", "female"), ethnicity = c("white", "white")
  )
  out <- spirometry_markers(df, cm)

  # Columns exist and are numeric/logical
  expect_true(all(c("fev1_pp","fvc_pp","fev1_z","fvc_z","ratio_z","obstruction_lln","gold_grade") %in% names(out)))
  # At least some non-NA values expected if rspiro returns references
  expect_true(any(is.finite(out$fev1_pp) | is.finite(out$fvc_pp)))
  # obstruction_lln is logical; gold_grade either character NA or "GOLD X"
  expect_type(out$obstruction_lln, "logical")
  expect_type(out$gold_grade, "character")
})

#' Vitamin D Status Category
#'
#' Categorizes vitamin D status based on serum 25-hydroxyvitamin D (25(OH)D) levels.
#'
#' @details
#' Serum 25(OH)D is the standard biomarker for vitamin D status. This function classifies
#' levels (assumed in ng/mL) into categories:
#' - Deficient (< 20 ng/mL)
#' - Insufficient (20-29 ng/mL)
#' - Sufficient (>= 30 ng/mL)
#'
#' Note: Ensure input units are ng/mL. If values appear extremely high (e.g., median > 150),
#' they might be in nmol/L (divide by 2.5 to convert to ng/mL).
#'
#' @param data A data.frame or tibble with a 25-hydroxyvitamin D concentration column.
#' @param col_map A named list with `vitamin_d` giving the column name in `data` for 25(OH)D.
#' @param na_action One of c("keep","omit","error","ignore","warn").
#'   - keep/ignore: compute and propagate NA
#'   - omit: drop rows with NA in required input
#'   - error: abort if required input contains NA
#'   - warn: like keep, but emit missingness warnings
#' @param verbose Logical; if TRUE, emits progress via rlang::inform.
#' @return A tibble with one column: vitamin_d_status (ordered factor with levels "Deficient","Insufficient","Sufficient").
#'
#' @examples
#' df <- data.frame(VitD = c(18, 45, 72))
#' vitamin_d_status(df)
#'
#' @references
#' \insertRef{iomm2011calciumvitd}{HealthMarkers}
#' \insertRef{holick2011vitd}{HealthMarkers}
#'
#' @export
vitamin_d_status <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff

  # Validate data
  if (!is.data.frame(data)) {
    rlang::abort("vitamin_d_status(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_vitd_error_data_type")
  }
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "vitamin_d_status"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  # Validate col_map: list vs. missing keys
  if (!is.list(col_map) && !is.null(col_map)) {
    rlang::abort("vitamin_d_status(): `col_map` must be a named list.",
                 class = "healthmarkers_vitd_error_colmap_type")
  }

  cm      <- .hm_build_col_map(data, col_map, c("vitd","vitamin_d"), fn = "vitamin_d_status")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  if (!is.list(col_map)) {
    rlang::abort("vitamin_d_status(): `col_map` must be a named list.",
                 class = "healthmarkers_vitd_error_colmap_type")
  }

  # Accept either 'vitd' or 'vitamin_d' as the mapping key
  key_opts <- c("vitd", "vitamin_d")
  key_present <- intersect(names(col_map), key_opts)

  # Key missing entirely -> missing_map
  if (length(key_present) == 0L) {
    rlang::abort("vitamin_d_status(): missing col_map entry for vitamin D value (e.g., vitd = 'VitD').",
                 class = "healthmarkers_vitd_error_missing_map")
  }
  key_used <- key_present[1]
  vitd_col <- as.character(col_map[[key_used]])

  if (!nzchar(vitd_col)) {
    rlang::abort(
      sprintf("vitamin_d_status(): empty mapping value for key '%s'.", key_used),
      class = "healthmarkers_vitd_error_bad_map_values"
    )
  }
  if (!(vitd_col %in% names(data))) {
    rlang::abort(
      sprintf("vitamin_d_status(): missing required column in data for key '%s': '%s'", key_used, vitd_col),
      class = "healthmarkers_vitd_error_missing_columns"
    )
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  vitamin_d_status [%s]", fn_name, vitd_col), level = "inform")

  # Coerce numeric
  vitd <- data[[vitd_col]]
  if (!is.numeric(vitd)) {
    old <- vitd
    suppressWarnings(vitd <- as.numeric(old))
    intro <- sum(is.na(vitd) & !is.na(old))
    if (intro > 0) {
      rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", vitd_col, intro),
                  class = "healthmarkers_vitd_warn_na_coercion")
    }
  }
  vitd[!is.finite(vitd)] <- NA_real_

  row_na <- is.na(vitd)
  if (na_action_raw == "warn" && any(row_na)) {
    rlang::warn("Missing 25(OH)D values; status will be NA for those rows.",
                class = "healthmarkers_vitd_warn_missing_inputs")
  }
  if (na_action_eff == "error" && any(row_na)) {
    rlang::abort("vitamin_d_status(): required inputs contain missing values (na_action='error').",
                 class = "healthmarkers_vitd_error_missing_values")
  }
  keep <- if (na_action_eff == "omit") !row_na else rep(TRUE, length(vitd))
  d_vitd <- vitd[keep]

  # Domain warnings (always-on)
  if (any(is.finite(d_vitd) & d_vitd < 0)) {
    rlang::warn("vitamin_d_status(): negative 25(OH)D values detected; check units/lab results.",
                class = "healthmarkers_vitd_warn_negative_values")
  }
  if (is.finite(stats::median(d_vitd, na.rm = TRUE)) && stats::median(d_vitd, na.rm = TRUE) > 150) {
    rlang::warn("vitamin_d_status(): values appear high; ensure units are ng/mL (nmol/L -> divide by 2.5).",
                class = "healthmarkers_vitd_warn_units_suspicious")
  }

  hm_inform("vitamin_d_status(): computing status", level = "debug")

  # Simple status bands (ng/mL): <20 Deficient, 20-29 Insufficient, >=30 Sufficient
  status <- ifelse(is.na(d_vitd), NA_character_,
            ifelse(d_vitd < 20, "Deficient",
            ifelse(d_vitd < 30, "Insufficient", "Sufficient")))
  status <- factor(status, levels = c("Deficient","Insufficient","Sufficient"), ordered = TRUE)

  out <- tibble::tibble(vitamin_d_status = status)

  # Pad back to full row count
  if (na_action_eff != "omit") {
    res <- tibble::tibble(vitamin_d_status = factor(rep(NA_character_, length(vitd)),
                                                    levels = levels(status), ordered = TRUE))
    res$vitamin_d_status[keep] <- out$vitamin_d_status
    out <- res
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) hm_inform(hm_result_summary(out, fn_name), level = "inform")
  out
 
}

#' Compute Bone Health & Body-Composition Markers (HM-CS v3)
#'
#' Given DXA, anthropometry, and optional bone-turnover markers, computes:
#' - OSTA: (weight - age) x 0.2
#' - ALMI: Appendicular Lean Mass Index = ALM / height^2
#' - FMI: Fat Mass Index = FM / height^2
#' - BMD_Tscore: (BMD - ref_mean) / ref_sd
#' and (if in `col_map` + data) passes through: TBS, HSA, PINP, CTX, BSAP, Osteocalcin.
#'
#' Notes:
#' - Units: height in meters; ALM, FM, weight in kilograms; BMD in g/cm^2; ALMI/FMI in kg/m^2.
#' - Non-finite values are treated as NA; division by zero is prevented by input checks.
#' - `BMD_ref_mean` and `BMD_ref_sd` must be supplied by the user from an appropriate
#'   reference population (for example, study-specific values or external norms such as NHANES).
#'
#' @param data A `data.frame` or tibble with subject-level DXA/anthropometry data.
#' @param col_map Named list mapping keys to column names. Required keys:
#'   - `age`, `weight`, `height`, `ALM`, `FM`, `BMD`, `BMD_ref_mean`, `BMD_ref_sd`
#'   Optional (passed-through if present and found in data): `TBS`, `HSA`, `PINP`, `CTX`, `BSAP`, `Osteocalcin`.
#' @param na_action One of "keep", "omit", or "error" controlling how
#'   missing/non-finite input values are treated.
#' @param verbose Logical; if TRUE (default), emits progress messages via `hm_inform()`.
#'
#' @return A tibble with columns: `OSTA`, `ALMI`, `FMI`, `BMD_Tscore`, and
#'   optionally `TBS`, `HSA`, `PINP`, `CTX`, `BSAP`, `Osteocalcin` (in that order).
#'
#' @examples
#' library(tibble)
#' df <- tibble(
#'   age = c(60, 72), weight = c(65, 50), height = c(1.65, 1.58),
#'   ALM = c(18.2, 14.7), FM = c(22.0, 20.5),
#'   BMD = c(0.95, 0.80), BMD_ref_mean = c(1.00, 1.00), BMD_ref_sd = c(0.12, 0.12)
#' )
#' col_map <- list(
#'   age = "age", weight = "weight", height = "height",
#'   ALM = "ALM", FM = "FM", BMD = "BMD",
#'   BMD_ref_mean = "BMD_ref_mean", BMD_ref_sd = "BMD_ref_sd"
#' )
#' bone_markers(df, col_map)
#'
#' @references
#' \insertRef{woo2002osta}{HealthMarkers}
#' \insertRef{who1994osteoporosis}{HealthMarkers}
#'
#' @importFrom tibble tibble
#' @export
bone_markers <- function(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "bone_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  id_col <- .hm_detect_id_col(data)

  required <- c("age", "weight", "height", "ALM", "FM", "BMD", "BMD_ref_mean", "BMD_ref_sd")
  optional <- c("TBS", "HSA", "PINP", "CTX", "BSAP", "Osteocalcin")

  hm_validate_inputs(data, col_map, required_keys = character(0), fn = "bone_markers")
  cm      <- .hm_build_col_map(data, col_map, keys = c(required, optional), fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Required columns exist in data
  req_cols <- unname(unlist(col_map[required], use.names = FALSE))
  missing_cols <- setdiff(req_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("bone_markers(): missing required columns in data: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_bone_error_missing_columns"
    )
  }

  hm_inform(level = "debug", msg = "bone_markers(): computing bone markers")
  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  OSTA       [(weight - age) * 0.2]\n  ALMI       [ALM / height^2]\n  FMI        [FM / height^2]\n  BMD_Tscore [(BMD - ref_mean) / ref_sd]", fn_name), level = "inform")

  # Coerce required and present optional columns to numeric; warn if NAs introduced
  present_opt_cols <- intersect(unname(unlist(col_map[intersect(optional, names(col_map))], use.names = FALSE)), names(data))
  for (cn in unique(c(req_cols, present_opt_cols))) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(
          sprintf("bone_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced),
          class = "healthmarkers_bone_warn_na_coercion"
        )
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  n <- nrow(data)
  # extract required
  age <- data[[col_map$age]]
  weight <- data[[col_map$weight]]
  height <- data[[col_map$height]]
  ALM <- data[[col_map$ALM]]
  FM <- data[[col_map$FM]]
  BMD <- data[[col_map$BMD]]
  ref_mean <- data[[col_map$BMD_ref_mean]]
  ref_sd <- data[[col_map$BMD_ref_sd]]

  # Constraints
  if (any(ref_sd <= 0, na.rm = TRUE)) {
    rlang::abort("bone_markers(): 'BMD_ref_sd' must be positive for non-missing rows.",
                 class = "healthmarkers_bone_error_refsd_positive")
  }
  if (any(height <= 0, na.rm = TRUE)) {
    rlang::abort("bone_markers(): 'height' must be positive for non-missing rows.",
                 class = "healthmarkers_bone_error_height_positive")
  }

  # NA row policy across required variables (NA or non-finite)
  req_df <- data[, unname(unlist(col_map[required])), drop = FALSE]
  rows_with_na <- !stats::complete.cases(req_df)
  if (na_action == "error" && any(rows_with_na)) {
    rlang::abort("bone_markers(): missing/non-finite values present in required inputs (na_action='error').",
                 class = "healthmarkers_bone_error_missing_values")
  } else if (na_action == "omit" && any(rows_with_na)) {
    keep <- !rows_with_na
    data <- data[keep, , drop = FALSE]
    age <- age[keep]; weight <- weight[keep]; height <- height[keep]
    ALM <- ALM[keep]; FM <- FM[keep]; BMD <- BMD[keep]
    ref_mean <- ref_mean[keep]; ref_sd <- ref_sd[keep]
    n <- nrow(data)
  }

  # compute core indices
  OSTA <- (weight - age) * 0.2
  ALMI <- ALM / (height^2)
  FMI <- FM / (height^2)
  BMD_Tscore <- (BMD - ref_mean) / ref_sd

  # helper for optional pass-through
  get_opt <- function(key) {
    if (key %in% names(col_map) && col_map[[key]] %in% names(data)) {
      as.numeric(data[[ col_map[[key]] ]])
    } else {
      rep(NA_real_, n)
    }
  }
  TBS <- get_opt("TBS")
  HSA <- get_opt("HSA")
  PINP <- get_opt("PINP")
  CTX <- get_opt("CTX")
  BSAP <- get_opt("BSAP")
  Osteocalcin <- get_opt("Osteocalcin")

  result <- tibble::tibble(
    OSTA = as.numeric(OSTA),
    ALMI = as.numeric(ALMI),
    FMI = as.numeric(FMI),
    BMD_Tscore = as.numeric(BMD_Tscore),
    TBS = TBS,
    HSA = HSA,
    PINP = PINP,
    CTX = CTX,
    BSAP = BSAP,
    Osteocalcin = Osteocalcin
  )

  if (!is.null(id_col)) {
    result[[id_col]] <- data[[id_col]]
    result <- result[, c(id_col, setdiff(names(result), id_col)), drop = FALSE]
    result <- tibble::as_tibble(result)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(result, fn_name), level = "inform") }
  result
}

# File: tests/testthat/test_bone_markers.R

library(testthat)
library(tibble)

cm_full <- list(
  age = "age", weight = "weight", height = "height",
  ALM = "ALM", FM = "FM",
  BMD = "BMD", BMD_ref_mean = "BMD_ref_mean", BMD_ref_sd = "BMD_ref_sd",
  TBS = "TBS", HSA = "HSA", PINP = "PINP",
  CTX = "CTX", BSAP = "BSAP", Osteocalcin = "Osteocalcin"
)

df_full <- tibble(
  age = c(65, 70),
  weight = c(60, 65),
  height = c(1.6, 1.7),
  ALM = c(18, 20),
  FM = c(20, 25),
  BMD = c(0.9, 1.0),
  BMD_ref_mean = c(1.0, 1.0),
  BMD_ref_sd = c(0.1, 0.1),
  TBS = c(1.1, 1.2),
  HSA = c(0.3, 0.3),
  PINP = c(40, 50),
  CTX = c(0.3, 0.4),
  BSAP = c(12, 15),
  Osteocalcin = c(10, 12)
)

test_that("bone_markers infers missing col_map entries from data (no error)", {
  skip_on_cran()
  bad_map <- cm_full[-1]  # removes "age" key; data still has "age" column
  expect_no_error(bone_markers(df_full, col_map = bad_map))
})

test_that("bone_markers computes core indices correctly", {
  skip_on_cran()
  out <- bone_markers(df_full, col_map = cm_full)
  expect_equal(out$OSTA, (df_full$weight - df_full$age) * 0.2)
  expect_equal(out$ALMI, df_full$ALM / df_full$height^2)
  expect_equal(out$FMI, df_full$FM / df_full$height^2)
  expect_equal(
    out$BMD_Tscore,
    (df_full$BMD - df_full$BMD_ref_mean) / df_full$BMD_ref_sd
  )
})

test_that("optional biomarkers are passed through when present", {
  skip_on_cran()
  out <- bone_markers(df_full, col_map = cm_full)
  expect_equal(out$TBS, df_full$TBS)
  expect_equal(out$HSA, df_full$HSA)
  expect_equal(out$PINP, df_full$PINP)
  expect_equal(out$CTX, df_full$CTX)
  expect_equal(out$BSAP, df_full$BSAP)
  expect_equal(out$Osteocalcin, df_full$Osteocalcin)
})

test_that("missing optional biomarkers yield NA columns", {
  skip_on_cran()
  # drop all optional keys
  cm_core <- cm_full[setdiff(
    names(cm_full),
    c("TBS", "HSA", "PINP", "CTX", "BSAP", "Osteocalcin")
  )]
  df_core <- df_full[, names(cm_core)]
  out <- bone_markers(df_core, col_map = cm_core)
  expect_true(all(is.na(out$TBS)))
  expect_true(all(is.na(out$HSA)))
  expect_true(all(is.na(out$PINP)))
  expect_true(all(is.na(out$CTX)))
  expect_true(all(is.na(out$BSAP)))
  expect_true(all(is.na(out$Osteocalcin)))
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(bone_markers(df_full, col_map = cm_full, verbose = TRUE), "bone_markers")
  expect_message(bone_markers(df_full, col_map = cm_full, verbose = TRUE), "col_map")
  expect_message(bone_markers(df_full, col_map = cm_full, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(
    bone_markers(df_full, col_map = cm_full, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action = 'keep' retains rows with missing/non-finite", {
  skip_on_cran()
  df_bad <- df_full
  df_bad$BMD[1] <- NA_real_
  out <- bone_markers(df_bad, col_map = cm_full, na_action = "keep")
  expect_equal(nrow(out), nrow(df_bad))
})

test_that("na_action = 'omit' drops rows with missing/non-finite inputs", {
  skip_on_cran()
  df_bad <- df_full
  df_bad$height[2] <- NA_real_
  out <- bone_markers(df_bad, col_map = cm_full, na_action = "omit")
  expect_equal(nrow(out), 1L)
})

test_that("na_action = 'error' stops on missing/non-finite inputs", {
  skip_on_cran()
  df_bad <- df_full
  df_bad$height[2] <- NA_real_
  expect_error(
    bone_markers(df_bad, col_map = cm_full, na_action = "error"),
    "missing/non-finite values"
  )
})

test_that("argument constraints are enforced (height > 0, ref_sd > 0)", {
  skip_on_cran()
  df_h0 <- df_full; df_h0$height[1] <- 0
  expect_error(bone_markers(df_h0, col_map = cm_full), "height' must be positive")
  df_sd0 <- df_full; df_sd0$BMD_ref_sd[1] <- 0
  expect_error(bone_markers(df_sd0, col_map = cm_full), "BMD_ref_sd' must be positive")
})

test_that("extra numeric columns in col_map are passed through", {
  skip_on_cran()
  df_extra <- df_full
  df_extra$ALMI_sds <- c(0, 2)
  cm_extra <- cm_full
  cm_extra$ALMI_sds <- "ALMI_sds"
  out <- bone_markers(df_extra, col_map = cm_extra)
  expect_equal(nrow(out), 2L)
})

#' BODE Index (BMI, Obstruction, Dyspnea, Exercise capacity)
#'
#' Computes the BODE index (0-10) using FEV1 % predicted, 6-minute walk distance (6MWD),
#' mMRC dyspnea scale, and BMI. Higher scores indicate worse prognosis in COPD.
#'
#' @details
#' Scoring components:
#' FEV1 % predicted: >=65 = 0; 50-64 = 1; 36-49 = 2; <=35 = 3
#' 6MWD (meters): >=350 = 0; 250-349 = 1; 150-249 = 2; <=149 = 3
#' mMRC dyspnea: 0-1 = 0; 2 = 1; 3 = 2; 4 = 3
#' BMI: >21 = 0; <=21 = 1
#'
#' @param data data.frame/tibble with required columns.
#' @param col_map named list with keys:
#'   fev1_pct OR (fev1 and fev1_pred) OR fev1_pp; plus sixmwd, mmrc, bmi.
#'   Example minimal: list(fev1_pct="FEV1pct", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")
#'   Example derive: list(fev1="FEV1", fev1_pred="FEV1_pred", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")
#'   Example from spirometry_markers: list(fev1_pp="fev1_pp", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")
#' @param na_action one of c("keep","omit","error","ignore","warn").
#' @param verbose logical; TRUE (default) emits messages.
#' @return tibble with bode_index (integer). NA if any required input missing (unless omitted).
#' @references \insertRef{celli2004bode}{HealthMarkers}
#'
#' @examples
#' df <- data.frame(FEV1pct = c(68, 45, 30), Walk_m = c(400, 280, 140),
#'                  mMRC = c(1, 2, 3), BMI = c(24, 19, 18))
#' bode_index(df, col_map = list(fev1_pct = "FEV1pct", sixmwd = "Walk_m",
#'                               mmrc = "mMRC", bmi = "BMI"))
#' @export
bode_index <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "bode_index"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  # Validate data
  if (!is.data.frame(data)) {
    rlang::abort("bode_index(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_bode_error_data_type")
  }

  all_bode_keys <- c("fev1_pct","fev1_pp","fev1","sixmwd","mmrc","bmi")
  cm      <- .hm_build_col_map(data, col_map, all_bode_keys, fn = "bode_index")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("bode_index(): `col_map` must be a named list.",
                 class = "healthmarkers_bode_error_colmap_type")
  }

  # Mandatory non-FEV1 keys
  base_keys <- c("sixmwd","mmrc","bmi")
  missing_base <- setdiff(base_keys, names(col_map))
  if (length(missing_base)) {
    rlang::abort(paste0("bode_index(): missing col_map entries for: ",
                        paste(missing_base, collapse = ", ")),
                 class = "healthmarkers_bode_error_missing_map")
  }

  # Resolve FEV1 percent predicted source
  has_pct <- !is.null(col_map$fev1_pct) && nzchar(col_map$fev1_pct)
  has_raw <- !is.null(col_map$fev1) && nzchar(col_map$fev1)
  has_pred <- !is.null(col_map$fev1_pred) && nzchar(col_map$fev1_pred)
  has_pp <- !is.null(col_map$fev1_pp) && nzchar(col_map$fev1_pp)

  fev1_source <- if (has_pct) "pct" else if (has_raw && has_pred) "derive" else if (has_pp) "pp" else "none"
  if (fev1_source == "none") {
    rlang::abort("bode_index(): must supply either fev1_pct OR (fev1 & fev1_pred) OR fev1_pp.",
                 class = "healthmarkers_bode_error_missing_fev1_source")
  }

  # Collect column names to check exist
  cols_needed <- c(
    if (fev1_source == "pct") col_map$fev1_pct,
    if (fev1_source == "derive") c(col_map$fev1, col_map$fev1_pred),
    if (fev1_source == "pp") col_map$fev1_pp,
    col_map$sixmwd,
    col_map$mmrc,
    col_map$bmi
  )
  cols_needed <- cols_needed[nzchar(cols_needed)]
  miss_cols <- setdiff(cols_needed, names(data))
  if (length(miss_cols)) {
    rlang::abort(paste0("bode_index(): missing required columns in data: ",
                        paste(miss_cols, collapse = ", ")),
                 class = "healthmarkers_bode_error_missing_columns")
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  bode_index  [0-10 COPD severity score]", fn_name), level = "inform")

  # Coercion helper
  coerce_col <- function(cn) {
    if (is.null(cn) || !nzchar(cn)) return(NULL)
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0) {
        rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
                    class = "healthmarkers_bode_warn_na_coercion")
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
    data[[cn]]
  }

  fev1_pct <- switch(
    fev1_source,
    pct = coerce_col(col_map$fev1_pct),
    derive = {
      raw <- coerce_col(col_map$fev1)
      pred <- coerce_col(col_map$fev1_pred)
      out <- 100 * raw / pred
      out[!is.finite(out)] <- NA_real_
      out
    },
    pp = {
      pp <- coerce_col(col_map$fev1_pp)
      pp[!is.finite(pp)] <- NA_real_
      pp
    }
  )

  sixmwd <- coerce_col(col_map$sixmwd)
  mmrc   <- coerce_col(col_map$mmrc)
  bmi    <- coerce_col(col_map$bmi)

  row_na <- is.na(fev1_pct) | is.na(sixmwd) | is.na(mmrc) | is.na(bmi)
  if (na_action_raw == "warn" && any(row_na)) {
    rlang::warn("Missing BODE components; bode_index will be NA for those rows.",
                class = "healthmarkers_bode_warn_missing_inputs")
  }
  if (na_action_eff == "error" && any(row_na)) {
    rlang::abort("bode_index(): required inputs contain missing values (na_action='error').",
                 class = "healthmarkers_bode_error_missing_values")
  }
  keep <- if (na_action_eff == "omit") !row_na else rep(TRUE, length(fev1_pct))

  d_fev1 <- fev1_pct[keep]
  d_walk <- sixmwd[keep]
  d_mmrc <- mmrc[keep]
  d_bmi  <- bmi[keep]

  # Domain warnings
  if (any(is.finite(d_mmrc) & (d_mmrc < 0 | d_mmrc > 4))) {
    rlang::warn("bode_index(): mMRC values outside 0-4 detected.",
                class = "healthmarkers_bode_warn_mmrc_range")
  }
  if (any(is.finite(d_fev1) & (d_fev1 < 0 | d_fev1 > 150))) {
    rlang::warn("bode_index(): FEV1 % predicted values outside 0-150 detected.",
                class = "healthmarkers_bode_warn_fev1pct_range")
  }
  if (any(is.finite(d_walk) & (d_walk < 0 | d_walk > 1500))) {
    rlang::warn("bode_index(): 6MWD values outside plausible range detected.",
                class = "healthmarkers_bode_warn_sixmwd_range")
  }
  if (any(is.finite(d_bmi) & (d_bmi < 10 | d_bmi > 80))) {
    rlang::warn("bode_index(): BMI values outside plausible range detected.",
                class = "healthmarkers_bode_warn_bmi_range")
  }

  hm_inform(level = "debug", msg = "bode_index(): computing score")

  fev1_score <- ifelse(is.na(d_fev1), NA_integer_,
                       ifelse(d_fev1 >= 65, 0L,
                       ifelse(d_fev1 >= 50, 1L,
                       ifelse(d_fev1 >= 36, 2L, 3L))))
  walk_score <- ifelse(is.na(d_walk), NA_integer_,
                       ifelse(d_walk >= 350, 0L,
                       ifelse(d_walk >= 250, 1L,
                       ifelse(d_walk >= 150, 2L, 3L))))
  mmrc_score <- ifelse(is.na(d_mmrc), NA_integer_,
                       ifelse(d_mmrc <= 1, 0L,
                       ifelse(d_mmrc == 2, 1L,
                       ifelse(d_mmrc == 3, 2L, 3L))))
  bmi_score  <- ifelse(is.na(d_bmi), NA_integer_,
                       ifelse(d_bmi > 21, 0L, 1L))

  total_score <- fev1_score + walk_score + mmrc_score + bmi_score
  any_comp_na <- is.na(fev1_score) | is.na(walk_score) | is.na(mmrc_score) | is.na(bmi_score)
  total_score[any_comp_na] <- ifelse(na_action_eff == "omit", total_score[any_comp_na], NA_integer_)

  out <- tibble::tibble(
    bode_index = total_score,
    fev1_pct = d_fev1,
    fev1_score = fev1_score,
    walk_score = walk_score,
    mmrc_score = mmrc_score,
    bmi_score  = bmi_score
  )

  if (na_action_eff != "omit") {
    res <- tibble::tibble(
      bode_index = rep(NA_integer_, length(fev1_pct)),
      fev1_pct = rep(NA_real_, length(fev1_pct)),
      fev1_score = NA_integer_,
      walk_score = NA_integer_,
      mmrc_score = NA_integer_,
      bmi_score  = NA_integer_
    )
    res[keep, ] <- out
    out <- res
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

library(testthat)

cm <- list(fev1_pct = "FEV1pct", sixmwd = "Walk_m", mmrc = "mMRC", bmi = "BMI")

test_that("errors on non-data input", {
  skip_on_cran()
  expect_error(bode_index("x", cm), class = "healthmarkers_bode_error_data_type")
})

test_that("errors on missing mapping keys", {
  skip_on_cran()
  bad <- list(fev1_pct = "FEV1pct", mmrc = "mMRC") # missing others
  df <- data.frame(FEV1pct = 80, Walk_m = 400, mMRC = 1, BMI = 25)
  expect_error(bode_index(df, bad), class = "healthmarkers_bode_error_missing_map")
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(FEV1pct = 80, Walk_m = 400, mMRC = 1, BMI = 25)
  expect_message(bode_index(df, cm, verbose = TRUE), "bode_index")
  expect_message(bode_index(df, cm, verbose = TRUE), "col_map")
  expect_message(bode_index(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- data.frame(FEV1pct = 80, Walk_m = 400, mMRC = 1, BMI = 25)
  msgs <- testthat::capture_messages(bode_index(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("cut-point scoring basic", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c(80,60,40,30),
    Walk_m  = c(400,300,200,100),
    mMRC    = c(1,2,3,4),
    BMI     = c(22,22,20,20)
  )
  out <- bode_index(df, cm)
  # Row scores: (0+0+0+0)=0; (1+1+1+0)=3; (2+2+2+1)=7; (3+3+3+1)=10
  expect_equal(out$bode_index, c(0L,3L,7L,10L))
})

test_that("boundary mapping", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c(65,50,36,35),
    Walk_m  = c(350,250,150,149),
    mMRC    = c(0,1,2,3),
    BMI     = c(21.1,21.0,21.0,20.9)
  )
  out <- bode_index(df, cm)
  # Expected: 0; (1+1+0+1)=3; (2+2+1+1)=6; (3+3+2+1)=9
  expect_equal(out$bode_index, c(0L,3L,6L,9L))
})

test_that("NA policies keep/omit/error/warn/ignore", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c(60, NA, 30),
    Walk_m  = c(300, 200, NA),
    mMRC    = c(2,4,3),
    BMI     = c(25,19,18)
  )
  keep <- bode_index(df, cm, na_action = "keep")
  expect_equal(nrow(keep), 3L)
  omit <- bode_index(df, cm, na_action = "omit")
  expect_equal(nrow(omit), 1L)
  expect_error(bode_index(df, cm, na_action = "error"),
               class = "healthmarkers_bode_error_missing_values")
  expect_warning(bode_index(df, cm, na_action = "warn"),
                 class = "healthmarkers_bode_warn_missing_inputs")
  ignore <- bode_index(df, cm, na_action = "ignore")
  expect_equal(nrow(ignore), 3L)
})

test_that("numeric coercion warning", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c("60","oops"),
    Walk_m  = c("300","250"),
    mMRC    = c("2","3"),
    BMI     = c("25","19")
  )
  expect_warning(bode_index(df, cm),
                 class = "healthmarkers_bode_warn_na_coercion")
})

test_that("domain warnings for individual ranges", {
  skip_on_cran()
  # FEV1pct out-of-range only
  df_fev1 <- data.frame(FEV1pct = 200, Walk_m = 400, mMRC = 1, BMI = 22)
  expect_warning(bode_index(df_fev1, cm), class = "healthmarkers_bode_warn_fev1pct_range")

  # mMRC out-of-range only
  df_mmrc <- data.frame(FEV1pct = 80, Walk_m = 400, mMRC = 5, BMI = 22)
  expect_warning(bode_index(df_mmrc, cm), class = "healthmarkers_bode_warn_mmrc_range")

  # 6MWD out-of-range only
  df_walk <- data.frame(FEV1pct = 80, Walk_m = 1600, mMRC = 1, BMI = 22)
  expect_warning(bode_index(df_walk, cm), class = "healthmarkers_bode_warn_sixmwd_range")

  # BMI out-of-range only
  df_bmi <- data.frame(FEV1pct = 80, Walk_m = 400, mMRC = 1, BMI = 85)
  expect_warning(bode_index(df_bmi, cm), class = "healthmarkers_bode_warn_bmi_range")
})

test_that("extreme values trigger domain warnings", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c(160,40,20),
    Walk_m  = c(900,140,300),
    mMRC    = c(6,2,-1),
    BMI     = c(5,90,22)
  )
  expect_warning(
    withCallingHandlers(
      bode_index(df, cm),
      warning = function(w) {
        if (inherits(w, c("healthmarkers_bode_warn_mmrc_range", "healthmarkers_bode_warn_bmi_range"))) invokeRestart("muffleWarning")
      }
    ),
    class = "healthmarkers_bode_warn_fev1pct_range"
  )
  # Values still produce output
  out <- suppressWarnings(bode_index(df, cm))
  expect_equal(nrow(out), 3L)
})

test_that("formula correctness with normal input", {
  skip_on_cran()
  df <- data.frame(
    FEV1pct = c(70, 50, 30),
    Walk_m  = c(400, 300, 200),
    mMRC    = c(1, 2, 3),
    BMI     = c(25, 22, 19)
  )
  out <- bode_index(df, cm)
  expect_equal(nrow(out), 3L)
  expect_true(all(is.finite(out$bode_index)))
  # Row 1: fev1>=65=0, walk>=350=0, mmrc<=1=0, bmi>21=0 => total 0
  expect_equal(out$bode_index[1], 0L)
})

test_that("padding keep/warn/ignore retain row count; omit reduces", {
  skip_on_cran()
  df <- data.frame(FEV1pct = c(60, NA), Walk_m = c(300,200), mMRC = c(2,4), BMI = c(25,18))
  keep <- bode_index(df, cm, na_action = "keep")
  warn <- suppressWarnings(bode_index(df, cm, na_action = "warn"))
  ignore <- bode_index(df, cm, na_action = "ignore")
  omit <- bode_index(df, cm, na_action = "omit")
  expect_equal(nrow(keep), 2L)
  expect_equal(nrow(warn), 2L)
  expect_equal(nrow(ignore), 2L)
  expect_equal(nrow(omit), sum(complete.cases(df)))
})

test_that("empty input returns 0-row tibble", {
  skip_on_cran()
  df <- data.frame(FEV1pct = numeric(), Walk_m = numeric(), mMRC = numeric(), BMI = numeric())
  out <- bode_index(df, cm)
  expect_equal(nrow(out), 0L)
  expect_true("bode_index" %in% names(out))
})

test_that("extreme values outside domain ranges produce domain warnings", {
  skip_on_cran()
  # Values within some domains but outside others
  df <- data.frame(
    FEV1pct = c(9, 50, 30),
    Walk_m  = c(45, 300, 700),
    mMRC    = c(0, 2, 4),
    BMI     = c(9, 22, 61)
  )
  out <- suppressWarnings(bode_index(df, cm))
  expect_equal(nrow(out), 3L)
})

#' Albumin-Corrected Calcium
#'
#' Calculates the albumin-adjusted (corrected) serum calcium level, accounting for hypoalbuminemia,
#' using the Payne formula.
#'
#' @details
#' Payne formula (conventional units): Corrected Ca (mg/dL) = measured Ca (mg/dL) + 0.8 * (4.0 - albumin (g/dL)).
#' If inputs appear to be in SI units (calcium mmol/L, albumin g/L), they are converted to mg/dL and g/dL
#' (using 1 mmol/L ~= 4 mg/dL; 1 g/L = 0.1 g/dL) for the correction and converted back to mmol/L for output.
#'
#' @param data A data.frame or tibble containing serum calcium and albumin.
#' @param col_map Named list with `calcium` and `albumin` indicating column names.
#' @param verbose Logical; if TRUE (default), emits progress via hm_inform().
#' @param na_action One of c("keep","omit","error","ignore","warn").
#' @param units One of c("auto","conventional","si"). "auto" attempts unit detection.
#'
#' @return A tibble with one column: corrected_calcium (numeric, in mg/dL for
#'   conventional input or mmol/L for SI / auto-SI input).
#' @references \insertRef{payne1973}{HealthMarkers}
#'
#' @examples
#' df <- data.frame(Ca = c(2.3, 2.5, 2.1), Alb = c(38, 42, 30))
#' corrected_calcium(df)
#' @export
corrected_calcium <- function(
  data,
  col_map = NULL,
  units = c("auto", "conventional", "si"),
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "corrected_calcium"
  .hm_log_input(data, data_name, fn_name, verbose)
  units         <- match.arg(units)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  ## --- Basic validation -----------------------------------------------------
  if (!is.data.frame(data)) {
    rlang::abort(
      "corrected_calcium(): `data` must be a data.frame or tibble.",
      class = "healthmarkers_calcium_error_data_type"
    )
  }
  if (!is.null(col_map) && !is.list(col_map)) {
    rlang::abort(
      "corrected_calcium(): `col_map` must be a named list.",
      class = "healthmarkers_calcium_error_colmap_type"
    )
  }
  cm      <- .hm_build_col_map(data, col_map, c("calcium","albumin"), fn = "corrected_calcium")
  data    <- cm$data
  col_map <- cm$col_map

  ca_key  <- intersect(names(col_map), c("calcium", "ca"))
  alb_key <- intersect(names(col_map), c("albumin", "alb"))
  if (length(ca_key) == 0 || length(alb_key) == 0) {
    rlang::abort(
      "corrected_calcium(): missing col_map entries for calcium and/or albumin.",
      class = "healthmarkers_calcium_error_missing_map"
    )
  }

  ca_col  <- as.character(col_map[[ca_key[1]]])
  alb_col <- as.character(col_map[[alb_key[1]]])

  if (!nzchar(ca_col) || !nzchar(alb_col)) {
    rlang::abort(
      "corrected_calcium(): empty mapping values supplied.",
      class = "healthmarkers_calcium_error_bad_map_values"
    )
  }

  missing_cols <- setdiff(c(ca_col, alb_col), names(data))
  if (length(missing_cols) > 0L) {
    rlang::abort(
      paste0(
        "corrected_calcium(): missing required columns in data: ",
        paste(missing_cols, collapse = ", ")
      ),
      class = "healthmarkers_calcium_error_missing_columns"
    )
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  corrected_calcium  [Payne formula: Ca + 0.8 * (4.0 - Alb)]", fn_name), level = "inform")

  ## --- Coercion to numeric --------------------------------------------------
  coerce_num <- function(x, nm) {
    if (!is.numeric(x)) {
      old <- x
      suppressWarnings(xn <- as.numeric(old))
      introduced <- sum(is.na(xn) & !is.na(old))
      if (introduced > 0L) {
        rlang::warn(
          sprintf("Column '%s' coerced to numeric; NAs introduced: %d", nm, introduced),
          class = "healthmarkers_calcium_warn_na_coercion"
        )
      }
      x <- xn
    }
    x[!is.finite(x)] <- NA_real_
    x
  }

  ca_raw  <- coerce_num(data[[ca_col]],  ca_col)
  alb_raw <- coerce_num(data[[alb_col]], alb_col)

  ## --- NA handling ----------------------------------------------------------
  row_missing <- is.na(ca_raw) | is.na(alb_raw)

  if (na_action_raw == "warn" && any(row_missing)) {
    rlang::warn(
      "Missing calcium or albumin values; corrected_calcium will be NA for those rows.",
      class = "healthmarkers_calcium_warn_missing_inputs"
    )
  }

  if (na_action_eff == "error" && any(row_missing)) {
    rlang::abort(
      "corrected_calcium(): required inputs contain missing values (na_action='error').",
      class = "healthmarkers_calcium_error_missing_values"
    )
  }

  keep_idx <- if (na_action_eff == "omit") !row_missing else rep(TRUE, length(row_missing))
  d_ca  <- ca_raw[keep_idx]
  d_alb <- alb_raw[keep_idx]

  ## --- Unit resolution / auto detection ------------------------------------
  med_ca  <- stats::median(d_ca,  na.rm = TRUE)
  med_alb <- stats::median(d_alb, na.rm = TRUE)

  if (units == "conventional") {
    ca_unit  <- "mg/dL"
    alb_unit <- "g/dL"
  } else if (units == "si") {
    ca_unit  <- "mmol/L"
    alb_unit <- "g/L"
  } else { # units == "auto"
    ca_unit  <- if (is.finite(med_ca)  && med_ca  <= 4)  "mmol/L" else "mg/dL"
    alb_unit <- if (is.finite(med_alb) && med_alb > 20) "g/L"    else "g/dL"
    si_inferred <- (ca_unit == "mmol/L" || alb_unit == "g/L")
    if (si_inferred) {
      rlang::warn(
        sprintf(
          "corrected_calcium(): auto-detected SI units Ca=%s, Alb=%s; output standardized to mmol/L.",
          ca_unit, alb_unit
        ),
        class = "healthmarkers_calcium_warn_unit_assumption"
      )
    }
  }

  si_inferred <- (ca_unit == "mmol/L" || alb_unit == "g/L")

  ## --- Convert to working conventional units (mg/dL, g/dL) -----------------
  ca_work_mgdl <- if (ca_unit == "mmol/L") d_ca * 4.0 else d_ca
  alb_work_gdl <- if (alb_unit == "g/L")  d_alb / 10.0 else d_alb

  ## --- Payne correction in mg/dL -------------------------------------------
  corr_mgdl <- ca_work_mgdl + 0.8 * (4.0 - alb_work_gdl)

  ## --- Domain warnings (explicit conventional only, no extreme scan) -------
  if (units == "conventional" && !si_inferred) {
    if (any(is.finite(alb_work_gdl) & (alb_work_gdl < 2 | alb_work_gdl > 5))) {
      rlang::warn(
        "corrected_calcium(): albumin outside typical 2-5 g/dL range; correction accuracy may be affected.",
        class = "healthmarkers_calcium_warn_albumin_range"
      )
    }
    if (any(is.finite(corr_mgdl) & (corr_mgdl < 5 | corr_mgdl > 15))) {
      rlang::warn(
        "corrected_calcium(): corrected calcium outside typical 5-15 mg/dL range detected.",
        class = "healthmarkers_calcium_warn_corrected_range"
      )
    }
  }

  ## --- Output scaling -------------------------------------------------------
  # SI or auto-SI => mmol/L
  corr_out <- if (units == "si" || (units == "auto" && si_inferred)) corr_mgdl / 4.0 else corr_mgdl

  hm_inform(level = "debug", msg = "corrected_calcium(): computing result")

  result <- tibble::tibble(corrected_calcium = corr_out)

  if (na_action_eff != "omit") {
    padded <- tibble::tibble(corrected_calcium = rep(NA_real_, length(ca_raw)))
    padded$corrected_calcium[keep_idx] <- result$corrected_calcium
    result <- padded
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep_idx] else data[[id_col]]
    result[[id_col]] <- id_vec
    result <- result[, c(id_col, setdiff(names(result), id_col)), drop = FALSE]
    result <- tibble::as_tibble(result)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(result, fn_name), level = "inform") }

  return(result)
}

library(testthat)

cm <- list(calcium = "Ca", albumin = "Alb")

test_that("validation errors: data type, col_map type, missing map, empty mapping, missing columns", {
  skip_on_cran()
  df_ok <- data.frame(Ca = 9, Alb = 3.5)

  expect_error(
    corrected_calcium("x", cm),
    class = "healthmarkers_calcium_error_data_type"
  )

  expect_error(
    corrected_calcium(df_ok, col_map = "not_list"),
    class = "healthmarkers_calcium_error_colmap_type"
  )

  # partial col_map is supplemented by inference
  expect_no_error(
    corrected_calcium(df_ok, col_map = list(calcium = "Ca"))
  )

  expect_error(
    corrected_calcium(df_ok, col_map = list(calcium = "", albumin = "Alb")),
    class = "healthmarkers_calcium_error_bad_map_values"
  )

  expect_error(
    corrected_calcium(data.frame(X = 1, Y = 2), cm),
    class = "healthmarkers_calcium_error_missing_columns"
  )
})

test_that("numeric coercion warning on non-numeric inputs", {
  skip_on_cran()
  df <- data.frame(Ca = c("9.0", "oops"), Alb = c("3.5", "3.0"))

  expect_warning(
    out <- corrected_calcium(df, cm),
    class = "healthmarkers_calcium_warn_na_coercion"
  )
  expect_true("corrected_calcium" %in% names(out))
})

test_that("NA policies: keep / warn / omit / error / ignore", {
  skip_on_cran()
  df <- data.frame(
    Ca  = c(9.0, NA, 8.5),
    Alb = c(3.0, 3.5, NA)
  )

  out_keep <- corrected_calcium(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(any(is.na(out_keep$corrected_calcium)))

  expect_warning(
    out_warn <- corrected_calcium(df, cm, na_action = "warn"),
    class = "healthmarkers_calcium_warn_missing_inputs"
  )
  expect_equal(nrow(out_warn), 3L)

  out_omit <- corrected_calcium(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 1L)

  expect_error(
    corrected_calcium(df, cm, na_action = "error"),
    class = "healthmarkers_calcium_error_missing_values"
  )

  out_ignore <- corrected_calcium(df, cm, na_action = "ignore")
  expect_equal(nrow(out_ignore), 3L)
})

test_that("conventional units Payne formula exact", {
  skip_on_cran()
  df <- data.frame(Ca = 8.5, Alb = 2.8)

  out <- corrected_calcium(df, cm, units = "conventional")
  expected <- 8.5 + 0.8 * (4.0 - 2.8)

  expect_equal(out$corrected_calcium, expected, tolerance = 1e-12)
})

test_that("auto SI detection converts and re-scales back (unit assumption warning)", {
  skip_on_cran()
  df <- data.frame(Ca = 2.2, Alb = 35)

  # check warning class
  expect_warning(
    corrected_calcium(df, cm, units = "auto"),
    class = "healthmarkers_calcium_warn_unit_assumption"
  )

  # check numeric result separately (do not rely on expect_warning's return)
  out <- suppressWarnings(corrected_calcium(df, cm, units = "auto"))
  expect_equal(out$corrected_calcium, 9.2 / 4, tolerance = 1e-12)
})

test_that("domain warnings for albumin and corrected calcium ranges (no extreme scan)", {
  skip_on_cran()
  # 1) Trigger corrected calcium outside 5–15 mg/dL but albumin in range
  df_corr <- data.frame(
    Ca  = c(4.0, 16.0),  # will yield corrected <5 and >15
    Alb = c(4.0, 4.0)    # within 2–5 g/dL
  )
  expect_warning(
    corrected_calcium(df_corr, cm, units = "conventional"),
    class = "healthmarkers_calcium_warn_corrected_range"
  )

  # 2) Trigger albumin outside 2–5 g/dL (regardless of corrected Ca)
  df_alb <- data.frame(
    Ca  = c(9.0, 9.5),
    Alb = c(1.5, 6.0)    # outside 2–5 g/dL
  )
  expect_warning(
    corrected_calcium(df_alb, cm, units = "conventional"),
    class = "healthmarkers_calcium_warn_albumin_range"
  )
})

test_that("domain warnings triggered for out-of-range albumin/calcium (conventional)", {
  skip_on_cran()
  df <- data.frame(
    Ca  = c(4.5, 16),
    Alb = c(1.5, 6.0)
  )

  expect_warning(
    corrected_calcium(df, cm, units = "conventional"),
    class = "healthmarkers_calcium_warn_albumin_range"
  )
})

test_that("extreme input values pass through and produce finite output", {
  skip_on_cran()
  df <- data.frame(
    Ca  = c(25, 9),
    Alb = c(3.5, 3.5)
  )

  out <- suppressWarnings(corrected_calcium(df, cm, units = "conventional"))
  expect_equal(nrow(out), 2L)
  expect_true(all(is.finite(out$corrected_calcium)))
})

test_that("standard input produces correct Payne formula output", {
  skip_on_cran()
  df <- data.frame(
    Ca  = c(9.0, 9.0),
    Alb = c(4.0, 2.0)
  )

  out <- corrected_calcium(df, cm, units = "conventional")
  # Row 1: 9.0 + 0.8*(4.0-4.0) = 9.0
  expect_equal(out$corrected_calcium[1], 9.0)
  # Row 2: 9.0 + 0.8*(4.0-2.0) = 10.6
  expect_equal(out$corrected_calcium[2], 10.6)
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(Ca = 9.0, Alb = 3.5)
  expect_message(corrected_calcium(df, cm, verbose = TRUE), "corrected_calcium")
  expect_message(corrected_calcium(df, cm, verbose = TRUE), "col_map")
  expect_message(corrected_calcium(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- data.frame(Ca = 9.0, Alb = 3.5)
  msgs <- testthat::capture_messages(corrected_calcium(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("empty input returns 0-row tibble with column", {
  skip_on_cran()
  df <- data.frame(Ca = numeric(), Alb = numeric())

  out <- corrected_calcium(df, cm)
  expect_equal(nrow(out), 0L)
  expect_true("corrected_calcium" %in% names(out))
})

test_that("all NA inputs produce NA outputs (keep/warn/ignore) and omit drops all", {
  skip_on_cran()
  df <- data.frame(Ca = c(NA, NA), Alb = c(NA, NA))

  out_keep <- corrected_calcium(df, cm, na_action = "keep")
  expect_true(all(is.na(out_keep$corrected_calcium)))

  expect_warning(
    out_warn <- corrected_calcium(df, cm, na_action = "warn"),
    class = "healthmarkers_calcium_warn_missing_inputs"
  )
  expect_true(all(is.na(out_warn$corrected_calcium)))

  out_ignore <- corrected_calcium(df, cm, na_action = "ignore")
  expect_true(all(is.na(out_ignore$corrected_calcium)))

  out_omit <- corrected_calcium(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 0L)
})

library(HealthMarkers)

# ============================================================
# HealthMarkers -- manual / interactive tests
# Run any block with Ctrl+Enter or highlight a section and run.
# Each numbered section is self-contained.
# ============================================================

# ---- 00 shared-data ----
df <- data.frame(
  G0=5.2, I0=65, G30=8.1, I30=120, G120=6.8, I120=90,
  glucose=5.2, insulin=65,
  TC=5.4, HDL_c=1.3, TG=1.6, LDL_c=3.4, ApoB=1.1, ApoA1=1.4,
  weight=82, height=178, BMI=25.9, waist=92, hip=99,
  fat_mass=22, ALM=19.5,
  WBC=6.5, neutrophils=3.8, lymphocytes=2.0, monocytes=0.5,
  eosinophils=0.2, platelets=240,
  CRP=1.8, albumin=42, ESR=12,
  creatinine=0.95, BUN=14,
  urine_albumin=18, urine_creatinine=120, eGFR=88, UACR=15,
  ALT=28, AST=22, GGT=35, bilirubin=0.9,
  sbp=122, dbp=78, bp_sys=122, bp_dia=78,
  bp_treated=FALSE, smoker=FALSE, diabetes=FALSE,
  fev1=3.4, fvc=4.3, fev1_pct=82, mmrc=1, sixmwd=420,
  age=48, sex="M", race="white", ethnicity="Caucasian",
  FFA=0.35, rate_palmitate=1.9, rate_glycerol=2.2,
  saliva_cort1=18, saliva_cort2=12, saliva_cort3=8,
  saliva_amylase=75, saliva_glucose=5.0,
  sweat_chloride=35, sweat_Na=55, sweat_K=8, sweat_lactate=12,
  weight_before=82, weight_after=81.3, duration=60, body_surface_area=1.9,
  ferritin=80, transferrin_sat=28, total_protein=72,
  EPA=2.1, DHA=4.3, Mg=0.88,
  glycated_albumin=13, uric_acid=320, phosphate=1.1,
  calcium=2.35, Na=140, K=4.1, Cl=101, HCO3=25, Tyr=65, Phe=52,
  VitD=55, VitD_ref_mean=50, VitD_ref_sd=10,
  B12=350, Folate=18,
  Retinol=0.85, Retinol_ref_mean=0.9, Retinol_ref_sd=0.2,
  Tocopherol=32, Total_lipids=3.2, PIVKA_II=4,
  VitC=65, Homocysteine=9, MMA=0.28, Zinc=14, Copper=14,
  total_testosterone=15, SHBG=38, LH=5, FSH=4,
  estradiol=110, progesterone=0.5, free_T3=4.6, free_T4=14.5,
  TSH=1.8, aldosterone=195, renin=11, GH=1.1, IGF1=175, prolactin=11,
  cortisol_0=410, cortisol_30=620,
  GSH=1200, GSSG=11,
  NfL=9.2, Kyn_nM=2700, Trp_uM=58,
  IL6=2.1, TNFa=3.4,
  MetS=0
)

cm <- list(
  G0="G0", I0="I0", G30="G30", I30="I30", G120="G120", I120="I120",
  FFA="FFA", glucose="glucose", insulin="insulin",
  TC="TC", HDL_c="HDL_c", TG="TG", LDL_c="LDL_c", ApoB="ApoB", ApoA1="ApoA1",
  weight="weight", height="height", BMI="BMI", bmi="BMI",
  waist="waist", hip="hip", fat_mass="fat_mass", ALM="ALM",
  WBC="WBC", neutrophils="neutrophils", lymphocytes="lymphocytes",
  monocytes="monocytes", eosinophils="eosinophils", platelets="platelets",
  CRP="CRP", albumin="albumin", ESR="ESR",
  creatinine="creatinine", BUN="BUN",
  urine_albumin="urine_albumin", urine_creatinine="urine_creatinine",
  eGFR="eGFR", UACR="UACR",
  ALT="ALT", AST="AST", GGT="GGT",
  sbp="sbp", dbp="dbp", bp_sys="bp_sys", bp_dia="bp_dia",
  bp_treated="bp_treated", smoker="smoker", diabetes="diabetes",
  fev1="fev1", fvc="fvc",
  age="age", sex="sex", race="race",
  rate_palmitate="rate_palmitate", rate_glycerol="rate_glycerol",
  total_testosterone="total_testosterone", SHBG="SHBG", LH="LH", FSH="FSH",
  estradiol="estradiol", progesterone="progesterone",
  free_T3="free_T3", free_T4="free_T4", TSH="TSH",
  aldosterone="aldosterone", renin="renin",
  GH="GH", IGF1="IGF1", prolactin="prolactin",
  cortisol_0="cortisol_0", cortisol_30="cortisol_30",
  GSH="GSH", GSSG="GSSG",
  NfL="NfL", Kyn_nM="Kyn_nM", Trp_uM="Trp_uM",
  IL6="IL6", TNFa="TNFa", MetS="MetS"
)

message("Shared data: ", nrow(df), " row x ", ncol(df), " cols")


# ---- 01 normalize_vec ----
x <- c(2.1, 3.5, 5.0, 1.2, 4.4)
print(normalize_vec(x, method = "z"))
print(normalize_vec(x, method = "range"))
print(normalize_vec(x, method = "robust"))


# ---- 02 calc_sds ----
df_sds <- data.frame(BMI = c(22, 28, 35))
ref_sds <- data.frame(var = "BMI", mean = 25, sd = 4)
out_sds <- calc_sds(df_sds, vars = "BMI", ref = ref_sds, verbose = FALSE)
print(out_sds)


# ---- 03 infer_cols ----
df_inf <- data.frame(pglu0 = 5.5, insu0 = 65, hedlc = 1.3, trig = 1.8)
print(infer_cols(df_inf))


# ---- 04 hm_col_report ----
rpt <- hm_col_report(df, col_map = NULL, verbose = FALSE)
message("hm_col_report: ", length(rpt), " keys mapped")


# ---- 05 fasting_is ----
out_fi <- fasting_is(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("fasting_is cols: ", paste(names(out_fi), collapse = ", "))


# ---- 06 ogtt_is ----
out_oi <- ogtt_is(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("ogtt_is cols: ", paste(names(out_oi), collapse = ", "))


# ---- 07 adipo_is ----
cm_adipo <- list(G0="G0", I0="I0", TG="TG", HDL_c="HDL_c", FFA="FFA",
                 waist="waist", bmi="BMI")
out_adipo <- adipo_is(df, col_map = cm_adipo, verbose = FALSE, na_action = "keep")
message("adipo_is cols: ", paste(names(out_adipo), collapse = ", "))


# ---- 08 tracer_dxa_is ----
cm_tdxa <- list(I0="I0", rate_glycerol="rate_glycerol",
                rate_palmitate="rate_palmitate", fat_mass="fat_mass",
                weight="weight", bmi="BMI", HDL_c="HDL_c")
out_tdxa <- tracer_dxa_is(df, col_map = cm_tdxa, verbose = FALSE, na_action = "keep")
message("tracer_dxa_is cols: ", paste(names(out_tdxa), collapse = ", "))


# ---- 09 all_insulin_indices ----
out_ins <- all_insulin_indices(df, col_map = cm, normalize = "none",
                                mode = "both", verbose = FALSE, na_action = "keep")
message("all_insulin_indices cols: ", ncol(out_ins))


# ---- 10 glycemic_markers ----
out_gly <- suppressWarnings(
  glycemic_markers(df, col_map = cm, normalize = "none", verbose = FALSE, na_action = "keep")
)
message("glycemic_markers cols: ", ncol(out_gly))


# ---- 11 lipid_markers ----
out_lip <- lipid_markers(df, col_map = cm, normalize = "none", verbose = FALSE, na_action = "keep")
message("lipid_markers cols: ", ncol(out_lip))


# ---- 12 atherogenic_indices ----
out_ath <- atherogenic_indices(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("atherogenic_indices cols: ", paste(names(out_ath), collapse = ", "))


# ---- 13 cvd_marker_aip ----
out_aip <- cvd_marker_aip(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("cvd_marker_aip cols: ", paste(names(out_aip), collapse = ", "))


# ---- 14 cvd_marker_ldl_particle_number ----
out_ldlp <- cvd_marker_ldl_particle_number(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("cvd_marker_ldl_particle_number cols: ", paste(names(out_ldlp), collapse = ", "))


# ---- 15 liver_markers ----
out_liv <- liver_markers(df, col_map = cm, verbose = FALSE, na_action = "keep")
message("liver_markers cols: ", ncol(out_liv))


# ---- 16 liver_fat_markers ----
cm_lf <- list(ALT="ALT", AST="AST", BMI="BMI", sex="sex",
              diabetes="diabetes", MetS="MetS", insulin="insulin",
              waist="waist", TG="TG", HDL_c="HDL_c", glucose="glucose")
out_lf <- suppressWarnings(
  liver_fat_markers(df, col_map = cm_lf, verbose = FALSE, na_action = "keep")
)
message("liver_fat_markers cols: ", ncol(out_lf))


# ---- 17 renal_markers ----
cm_ren <- list(creatinine="creatinine", age="age", sex="sex", race="race",
               BUN="BUN", urine_albumin="urine_albumin",
               urine_creatinine="urine_creatinine")
out_ren <- renal_markers(df, col_map = cm_ren, verbose = FALSE, na_action = "keep")
message("renal_markers cols: ", ncol(out_ren))


# ---- 18 ckd_stage ----
df_ckd <- data.frame(eGFR = c(95, 55, 28), UACR = c(12, 80, 350))
out_ckd <- ckd_stage(df_ckd, col_map = list(eGFR = "eGFR", UACR = "UACR"))
print(out_ckd)


# ---- 19 kidney_failure_risk ----
df_kfre <- data.frame(age=c(65,72), sex=c(1,0), eGFR=c(45,22), UACR=c(300,1200))
out_kfre <- kidney_failure_risk(df_kfre,
  col_map = list(age="age", sex="sex", eGFR="eGFR", UACR="UACR"),
  na_action = "keep")
print(out_kfre)


# ---- 20 cvd_risk_ascvd ----
df_cvd <- data.frame(age=55, sex="M", race="white",
                     TC=5.4, HDL_c=1.3, sbp=128,
                     diabetes=0, smoker=0, bp_treated=0)
cm_cvd <- list(age="age", sex="sex", race="race",
               TC="TC", HDL_c="HDL_c", sbp="sbp",
               diabetes="diabetes", smoker="smoker", bp_treated="bp_treated")
out_ascvd <- suppressWarnings(cvd_risk_ascvd(df_cvd, col_map = cm_cvd, verbose = FALSE))
message("cvd_risk_ascvd cols: ", paste(names(out_ascvd), collapse = ", "))


# ---- 21 cvd_risk_stroke ----
out_stroke <- suppressWarnings(
  cvd_risk_stroke(df_cvd, col_map = cm_cvd, verbose = FALSE, na_action = "keep")
)
message("cvd_risk_stroke cols: ", paste(names(out_stroke), collapse = ", "))


# ---- 22 cvd_risk (wrapper) ----
out_cvdr <- suppressWarnings(
  cvd_risk(df_cvd, col_map = cm_cvd, verbose = FALSE, na_action = "keep")
)
message("cvd_risk (wrapper) cols: ", paste(names(out_cvdr), collapse = ", "))


# ---- 23 cvd_risk_qrisk3 (requires QRISK3 package) ----
if (requireNamespace("QRISK3", quietly = TRUE)) {
  out_q3 <- suppressWarnings(cvd_risk_qrisk3(df_cvd, col_map = cm_cvd, verbose = FALSE))
  message("cvd_risk_qrisk3 cols: ", paste(names(out_q3), collapse = ", "))
} else {
  message("QRISK3 not installed -- skipping cvd_risk_qrisk3")
}


# ---- 24 cvd_risk_scorescvd (requires RiskScorescvd package) ----
if (requireNamespace("RiskScorescvd", quietly = TRUE)) {
  out_sc <- suppressWarnings(cvd_risk_scorescvd(df_cvd, col_map = cm_cvd, verbose = FALSE))
  message("cvd_risk_scorescvd cols: ", paste(names(out_sc), collapse = ", "))
} else {
  message("RiskScorescvd not installed -- skipping cvd_risk_scorescvd")
}


# ---- 25 inflammatory_markers ----
cm_inf <- list(neutrophils="neutrophils", lymphocytes="lymphocytes",
               monocytes="monocytes", platelets="platelets", WBC="WBC",
               CRP="CRP", albumin="albumin", eosinophils="eosinophils", ESR="ESR")
out_inf <- inflammatory_markers(df, col_map = cm_inf, panel = "both",
                                 na_action = "keep", verbose = FALSE)
message("inflammatory_markers cols: ", ncol(out_inf))


# ---- 26 metss ----
df_mets <- data.frame(
  waist=92, bp_sys=122, bp_dia=78, glucose=5.2,
  TG=1.6, HDL_c=1.3, sex="M", race="NHW",
  bp_treated=FALSE, smoker=FALSE, diabetes=FALSE
)
out_mets <- metss(df_mets, col_map = list(), verbose = FALSE, na_action = "keep")
message("metss cols: ", paste(names(out_mets), collapse = ", "))


# ---- 27 metabolic_risk_features ----
df_mrf <- data.frame(
  chol_total=5.4, chol_ldl=3.4, chol_hdl=1.3, triglycerides=1.6,
  age_year=48, z_HOMA=0.6, glucose=5.2, HbA1c=38,
  bp_sys_z=0.3, bp_dia_z=0.1
)
out_mrf <- metabolic_risk_features(df_mrf, verbose = FALSE, na_action = "keep")
message("metabolic_risk_features new cols: ",
        paste(setdiff(names(out_mrf), names(df_mrf)), collapse = ", "))


# ---- 28 metabolic_markers ----
out_mm <- suppressWarnings(
  metabolic_markers(df, col_map = cm,
                    which = c("insulin", "lipid", "liver", "glycemic"),
                    normalize = "none", mode = "both",
                    verbose = FALSE, na_action = "keep")
)
message("metabolic_markers: ", ncol(out_mm), " total output cols")


# ---- 29 obesity_indices ----
df_ob <- data.frame(wt=82, ht=178, wc=92, hip=99, sex=0)
out_ob <- obesity_indices(df_ob,
  weight = wt, height = ht, waist = wc, hip = hip, sex = sex,
  weight_unit = "kg", height_unit = "cm",
  adjust_WHR = TRUE, include_RFM = TRUE,
  verbose = FALSE, na_action = "keep")
message("obesity_indices new cols: ",
        paste(setdiff(names(out_ob), names(df_ob)), collapse = ", "))


# ---- 30 alm_bmi_index ----
df_alm <- data.frame(alm=19.5, bmi=25.9, sex="M")
out_alm <- alm_bmi_index(df_alm,
  col_map = list(alm="alm", bmi="bmi", sex="sex"), verbose = FALSE)
message("alm_bmi_index new cols: ",
        paste(setdiff(names(out_alm), names(df_alm)), collapse = ", "))


# ---- 31 adiposity_sds ----
df_ads <- data.frame(age=48, sex="M", BMI=25.9, waist=92, weight=82, height=178)
out_ads <- suppressWarnings(
  adiposity_sds(df_ads, col_map = NULL, verbose = FALSE, na_action = "keep")
)
message("adiposity_sds new cols: ",
        paste(setdiff(names(out_ads), names(df_ads)), collapse = ", "))


# ---- 32 adiposity_sds_strat ----
out_adss <- suppressWarnings(
  adiposity_sds_strat(df_ads, col_map = NULL, verbose = FALSE, na_action = "keep")
)
message("adiposity_sds_strat new cols: ",
        paste(setdiff(names(out_adss), names(df_ads)), collapse = ", "))


# ---- 33 spirometry_markers ----
df_spiro <- data.frame(FEV1 = 3.4, FVC = 4.3)
out_spiro <- spirometry_markers(df_spiro, verbose = FALSE, na_action = "keep")
message("spirometry_markers new cols: ",
        paste(setdiff(names(out_spiro), names(df_spiro)), collapse = ", "))


# ---- 34 pulmo_markers (requires rspiro package) ----
if (requireNamespace("rspiro", quietly = TRUE)) {
  df_pul <- data.frame(age=48, sex="M", height=178,
                       ethnicity="Caucasian", fev1=3.4, fvc=4.3)
  out_pul <- suppressWarnings(
    pulmo_markers(df_pul, equation = "GLI", verbose = FALSE, na_action = "keep")
  )
  message("pulmo_markers new cols: ",
          paste(setdiff(names(out_pul), names(df_pul)), collapse = ", "))
} else {
  message("rspiro not installed -- skipping pulmo_markers")
}


# ---- 35 bode_index ----
df_bode <- data.frame(BMI=25.9, fev1_pct=82, mmrc=1, sixmwd=420)
out_bode <- bode_index(df_bode,
  col_map = list(BMI="BMI", fev1_pct="fev1_pct", mmrc="mmrc", sixmwd="sixmwd"),
  verbose = FALSE)
message("bode_index new cols: ",
        paste(setdiff(names(out_bode), names(df_bode)), collapse = ", "))


# ---- 36 bone_markers ----
df_bone <- data.frame(age=60, weight=65, height=1.65, ALM=18.2, FM=22.0,
                      BMD=0.95, BMD_ref_mean=1.00, BMD_ref_sd=0.12)
cm_bone <- list(age="age", weight="weight", height="height",
                ALM="ALM", FM="FM", BMD="BMD",
                BMD_ref_mean="BMD_ref_mean", BMD_ref_sd="BMD_ref_sd")
out_bone <- bone_markers(df_bone, col_map = cm_bone, verbose = FALSE, na_action = "keep")
message("bone_markers new cols: ",
        paste(setdiff(names(out_bone), names(df_bone)), collapse = ", "))


# ---- 37 frax_score ----
df_frax <- data.frame(age=65, sex="F", weight=58, height=162,
                      prior_fracture=0, parent_fracture=0, steroids=0,
                      rheumatoid_arthritis=0, secondary_osteoporosis=0,
                      smoker=0, alcohol=0)
out_frax <- suppressWarnings(
  frax_score(df_frax, col_map = NULL, verbose = FALSE, na_action = "keep")
)
message("frax_score new cols: ",
        paste(setdiff(names(out_frax), names(df_frax)), collapse = ", "))


# ---- 38 vitamin_markers ----
df_vit <- data.frame(
  VitD=55, VitD_ref_mean=50, VitD_ref_sd=10,
  B12=350, Folate=18, Ferritin=80, TSat=0.28,
  Cortisol=420, DHEAS=95, Testosterone=14, Estradiol=110,
  TSH=1.8, free_T4=14.5,
  Retinol=0.85, Retinol_ref_mean=0.9, Retinol_ref_sd=0.2,
  Tocopherol=32, Total_lipids=3.2, PIVKA_II=4,
  VitC=65, Homocysteine=9, MMA=0.28, Magnesium=0.88, Zinc=14, Copper=14
)
cm_vit <- as.list(names(df_vit)); names(cm_vit) <- names(df_vit)
out_vit <- vitamin_markers(df_vit, col_map = cm_vit, verbose = FALSE, na_action = "keep")
message("vitamin_markers cols: ", ncol(out_vit))


# ---- 39 vitamin_d_status ----
df_vd <- data.frame(vitamin_D = c(28, 55, 82))
out_vd <- vitamin_d_status(df_vd,
  col_map = list(vitamin_D = "vitamin_D"), verbose = FALSE)
print(out_vd)


# ---- 40 nutrient_markers ----
cm_nutr <- as.list(names(df)); names(cm_nutr) <- names(df)
out_nutr <- suppressWarnings(
  nutrient_markers(df, col_map = cm_nutr, verbose = FALSE, na_action = "keep")
)
message("nutrient_markers cols: ", ncol(out_nutr))


# ---- 41 corrected_calcium ----
df_ca <- data.frame(Ca = c(2.3, 2.5, 2.1), albumin = c(38, 42, 30))
out_ca <- corrected_calcium(df_ca,
  col_map = list(calcium="Ca", albumin="albumin"),
  units = "si", verbose = FALSE)
print(out_ca)


# ---- 42 hormone_markers ----
df_horm <- data.frame(
  TT=15, SHBG=38, LH=5, FSH=4, E2=110, Prog=0.5,
  fT3=4.6, fT4=14.5, TSH=1.8, Aldo=195, Renin=11,
  Ins=65, Gluc=8, GH=1.1, IGF1=175, Prl=11,
  Cort0=410, Cort30=620
)
cm_horm <- list(
  total_testosterone="TT", SHBG="SHBG", LH="LH", FSH="FSH",
  estradiol="E2", progesterone="Prog", free_T3="fT3",
  free_T4="fT4", TSH="TSH", aldosterone="Aldo", renin="Renin",
  insulin="Ins", glucagon="Gluc", GH="GH", IGF1="IGF1",
  prolactin="Prl", cortisol_0="Cort0", cortisol_30="Cort30"
)
out_horm <- hormone_markers(df_horm, col_map = cm_horm, verbose = FALSE, na_action = "keep")
message("hormone_markers cols: ", ncol(out_horm))


# ---- 43 allostatic_load ----
df_allo <- data.frame(SBP=c(118,142,130), DBP=c(76,92,85), CRP=c(1.2,4.8,2.1))
out_allo <- allostatic_load(df_allo,
  thresholds = list(SBP=130, DBP=85, CRP=3),
  verbose = FALSE, na_action = "keep")
print(out_allo)


# ---- 44 iAge ----
df_iage <- data.frame(CRP=1.8, IL6=2.1, TNFa=3.4)
out_iage <- iAge(df_iage,
  col_map = list(CRP="CRP", IL6="IL6", TNFa="TNFa"),
  weights = c(CRP=0.33, IL6=0.33, TNFa=0.34),
  verbose = FALSE, na_action = "keep")
print(out_iage)


# ---- 45 nfl_marker ----
df_nfl <- data.frame(NfL = c(8.5, 14.2, 22.1))
out_nfl <- nfl_marker(df_nfl,
  col_map = list(NfL="NfL"), verbose = FALSE, na_action = "keep")
print(out_nfl)


# ---- 46 kyn_trp_ratio ----
df_kyn <- data.frame(Kyn_nM = c(2500, 3100), Trp_uM = c(55, 48))
out_kyn <- kyn_trp_ratio(df_kyn,
  col_map = list(Kyn_nM="Kyn_nM", Trp_uM="Trp_uM"),
  verbose = FALSE, na_action = "keep")
print(out_kyn)


# ---- 47 oxidative_markers ----
df_ox <- data.frame(GSH = c(5.0, 3.0), GSSG = c(1.0, 0.5))
out_ox <- oxidative_markers(df_ox,
  col_map = list(GSH="GSH", GSSG="GSSG"),
  verbose = FALSE, na_action = "keep")
print(out_ox)


# ---- 48 saliva_markers ----
df_sal <- data.frame(
  saliva_cort1=18, saliva_cort2=12, saliva_cort3=8,
  saliva_amylase=75, saliva_glucose=5.0
)
out_sal <- saliva_markers(df_sal, col_map = NULL, verbose = FALSE,
                           na_action = "keep", times = c(0, 30, 60))
message("saliva_markers new cols: ",
        paste(setdiff(names(out_sal), names(df_sal)), collapse = ", "))


# ---- 49 sweat_markers ----
df_sw <- data.frame(
  sweat_chloride=35, sweat_Na=55, sweat_K=8, sweat_lactate=12,
  weight_before=82, weight_after=81.3, duration=60, body_surface_area=1.9
)
out_sw <- sweat_markers(df_sw, col_map = NULL, verbose = FALSE, na_action = "keep")
message("sweat_markers new cols: ",
        paste(setdiff(names(out_sw), names(df_sw)), collapse = ", "))


# ---- 50 urine_markers ----
df_ur <- data.frame(
  urine_albumin=18, urine_creatinine=120,
  urine_protein=25, urine_Na=80, urine_K=40
)
out_ur <- urine_markers(df_ur, verbose = FALSE, na_action = "keep")
message("urine_markers new cols: ",
        paste(setdiff(names(out_ur), names(df_ur)), collapse = ", "))


# ---- 51 sarc_f_score ----
df_sf <- data.frame(Strength=c(1,2,0), Walking=c(0,1,2),
                    Chair=c(1,1,2), Stairs=c(0,2,2), Falls=c(0,1,1))
out_sf <- sarc_f_score(df_sf, verbose = FALSE)
print(out_sf[setdiff(names(out_sf), names(df_sf))])


# ---- 52 charlson_index ----
df_cci <- data.frame(
  mi=0, chf=0, pvd=0, stroke=0, dementia=0, copd=0, rheum=0, ulcer=0,
  mild_liver=0, diabetes=1, diab_comp=0, hemiplegia=0, renal=1,
  cancer=0, leukemia=0, lymphoma=0, sev_liver=0, metastatic_cancer=0, hiv=0
)
out_cci <- charlson_index(df_cci, verbose = FALSE)
print(out_cci[setdiff(names(out_cci), names(df_cci))])


# ---- 53 frailty_index (requires 'di' package) ----
if (requireNamespace("di", quietly = TRUE)) {
  df_fi <- data.frame(age=c(70,75,80), d1=c(0,1,1), d2=c(0.2,0.8,1.0), d3=c(1L,0L,1L))
  out_fi <- frailty_index(df_fi, cols = c("d1","d2","d3"), age = "age",
                           return = "data", verbose = FALSE)
  print(out_fi[c("FI", "FI_class")])
} else {
  message("'di' package not installed -- skipping frailty_index")
}


# ---- 54 plot_frailty_age (requires 'di' package) ----
if (requireNamespace("di", quietly = TRUE)) {
  df_fi2 <- data.frame(age=c(65,70,75,80), d1=c(0,0,1,1), d2=c(0.1,0.3,0.6,0.9))
  plot_frailty_age(df_fi2, cols = c("d1","d2"), age = "age", visible = FALSE)
  message("plot_frailty_age: completed")
} else {
  message("'di' package not installed -- skipping plot_frailty_age")
}


# ---- 55 phq9_score ----
df_phq <- setNames(as.data.frame(matrix(c(0,1,2,1,0,1,2,1,0), nrow=1)),
                   sprintf("phq9_%02d", 1:9))
out_phq <- phq9_score(df_phq, verbose = FALSE, na_action = "keep")
print(out_phq[grep("^PHQ9", names(out_phq))])


# ---- 56 gad7_score ----
df_gad <- setNames(as.data.frame(matrix(c(1,2,0,1,0,2,1), nrow=1)),
                   sprintf("gad7_%02d", 1:7))
out_gad <- gad7_score(df_gad, verbose = FALSE, na_action = "keep")
print(out_gad[grep("^GAD7", names(out_gad))])


# ---- 57 k6_score ----
df_k6 <- setNames(as.data.frame(matrix(c(1,0,2,1,0,1), nrow=1)),
                  sprintf("k6_%02d", 1:6))
out_k6 <- k6_score(df_k6, verbose = FALSE, na_action = "keep")
print(out_k6[grep("^K6", names(out_k6))])


# ---- 58 k10_score ----
df_k10 <- setNames(as.data.frame(matrix(c(1,0,2,1,0,1,2,0,1,0), nrow=1)),
                   sprintf("k10_%02d", 1:10))
out_k10 <- k10_score(df_k10, verbose = FALSE, na_action = "keep")
print(out_k10[grep("^K10", names(out_k10))])


# ---- 59 ghq12_score ----
df_ghq <- setNames(as.data.frame(matrix(rep(1L, 12), nrow=1)),
                   sprintf("ghq12_%02d", 1:12))
out_ghq <- ghq12_score(df_ghq, method = "binary", verbose = FALSE, na_action = "keep")
print(out_ghq[grep("^GHQ12", names(out_ghq))])


# ---- 60 who5_score ----
df_who <- setNames(as.data.frame(matrix(c(3,4,2,3,4), nrow=1)),
                   sprintf("who5_%02d", 1:5))
out_who <- who5_score(df_who, verbose = FALSE, na_action = "keep")
print(out_who[grep("^WHO5", names(out_who))])


# ---- 61 isi_score ----
df_isi <- setNames(as.data.frame(matrix(c(1,2,1,1,0,2,1), nrow=1)),
                   sprintf("isi_%02d", 1:7))
out_isi <- isi_score(df_isi, verbose = FALSE, na_action = "keep")
print(out_isi[grep("^ISI", names(out_isi))])


# ---- 62 mdq_score ----
df_mdq <- setNames(
  as.data.frame(matrix(c(rep(1L,13), 1L, 1L, 1L), nrow=1)),
  c(sprintf("mdq_%02d", 1:13), "mdq_clustering", "mdq_impairment", "mdq_total_check")
)
out_mdq <- suppressWarnings(mdq_score(df_mdq, verbose = FALSE, na_action = "keep"))
print(out_mdq[grep("^MDQ", names(out_mdq))])


# ---- 63 asrs_score ----
df_asrs <- setNames(as.data.frame(matrix(c(3,2,3,2,1,0), nrow=1)),
                    sprintf("asrs_%02d", 1:6))
out_asrs <- asrs_score(df_asrs, verbose = FALSE, na_action = "keep")
print(out_asrs[grep("^ASRS", names(out_asrs))])


# ---- 64 bis_score ----
bis_key <- list(items = sprintf("bis_%02d", 1:5), min_val = 1, max_val = 4)
df_bis <- setNames(as.data.frame(matrix(c(2,3,1,4,2), nrow=1)),
                   sprintf("bis_%02d", 1:5))
out_bis <- bis_score(df_bis, key = bis_key, verbose = FALSE, na_action = "keep")
print(out_bis[grep("^BIS", names(out_bis))])


# ---- 65 spq_score ----
spq_key <- list(items = sprintf("spq_%02d", 1:5), min_val = 0, max_val = 1)
df_spq <- setNames(as.data.frame(matrix(c(0,1,0,1,0), nrow=1)),
                   sprintf("spq_%02d", 1:5))
out_spq <- spq_score(df_spq, key = spq_key, verbose = FALSE, na_action = "keep")
print(out_spq[grep("^SPQ", names(out_spq))])


# ---- 66 cognitive_score ----
df_cog <- data.frame(task_a=c(85,90), task_b=c(72,88), task_c=c(68,79))
cm_cog <- list(tasks = list(task_a="task_a", task_b="task_b", task_c="task_c"))
out_cog <- cognitive_score(df_cog, col_map = cm_cog, method = "z_mean", verbose = FALSE)
print(out_cog[grep("^cog", names(out_cog))])


# ---- 67 psych_dx_flags ----
df_pdx <- data.frame(has_mdd=1, has_anxiety=0, has_adhd=0,
                     has_bipolar=0, has_scz=0, has_sud=0)
cm_pdx <- list(dx = list(mdd="has_mdd", anxiety="has_anxiety", adhd="has_adhd",
                          bipolar="has_bipolar", scz="has_scz", sud="has_sud"))
out_pdx <- psych_dx_flags(df_pdx, col_map = cm_pdx)
print(out_pdx[grep("^dx_", names(out_pdx))])


# ---- 68 psych_med_flags ----
df_pmed <- data.frame(on_ssri=1, on_snri=0, on_ap=0, on_ms=0, on_anx=0)
cm_pmed <- list(med = list(ssri="on_ssri", snri="on_snri",
                            antipsychotic="on_ap",
                            mood_stabilizer="on_ms", anxiolytic="on_anx"))
out_pmed <- psych_med_flags(df_pmed, col_map = cm_pmed)
print(out_pmed[grep("^med_", names(out_pmed))])


# ---- 69 psych_markers ----
df_psy <- cbind(
  setNames(as.data.frame(matrix(c(0,1,2,1,0,1,2,1,0), nrow=1)), sprintf("phq9_%02d", 1:9)),
  setNames(as.data.frame(matrix(c(1,2,0,1,0,2,1), nrow=1)), sprintf("gad7_%02d", 1:7))
)
out_psy <- psych_markers(df_psy, which = c("phq9","gad7"),
                          verbose = FALSE, na_action = "keep")
message("psych_markers new cols: ",
        paste(setdiff(names(out_psy), names(df_psy)), collapse = ", "))


# ---- 70 impute_missing ----
df_na <- data.frame(x = c(1, NA, 3, NA), y = c(NA, 2, 3, 4))
out_imp <- impute_missing(df_na, method = "median", verbose = FALSE)
print(out_imp)


# ---- 71 impute_mice (requires mice package) ----
if (requireNamespace("mice", quietly = TRUE)) {
  out_mice <- impute_mice(df_na, m = 1, seed = 42, verbose = FALSE)
  print(out_mice)
} else {
  message("mice not installed -- skipping impute_mice")
}


# ---- 72 impute_missforest (requires missForest package) ----
if (requireNamespace("missForest", quietly = TRUE)) {
  out_mf <- suppressWarnings(impute_missforest(df_na, verbose = FALSE))
  print(out_mf)
} else {
  message("missForest not installed -- skipping impute_missforest")
}


# ---- 73 validate_inputs ----
validate_inputs(df, col_map = cm, required_keys = c("G0","I0","TC"), fn = "test")
message("validate_inputs: passed")


# ---- 74 marker_summary ----
df_sum <- data.frame(HOMA_IR=1.4, HOMA_IS=0.7, TG=1.6, HDL_c=1.3)
out_ms <- marker_summary(df_sum, verbose = FALSE)
print(out_ms)


# ---- 75 health_summary ----
df_hs <- data.frame(TC=5.4, HDL_c=1.3, TG=1.6, glucose=5.2, BMI=25.9, eGFR=88)
out_hs <- health_summary(df_hs, col_map = NULL, verbose = FALSE)
print(out_hs)


# ---- 76 all_health_markers -- fast subset (no external packages) ----
fast_groups <- c(
  "insulin_fasting", "insulin_ogtt", "insulin_adipose", "insulin_tracer_dxa",
  "lipid", "atherogenic", "cvd_aip", "cvd_ldl_particles",
  "glycemic", "liver", "liver_fat",
  "renal", "ckd_stage", "kidney_kfre", "urine",
  "inflammatory", "hormone", "nutrient", "vitamin", "vitamin_d_status",
  "calcium_corrected", "kyn_trp", "oxidative", "nfl", "iAge",
  "spirometry", "bode",
  "saliva", "sweat",
  "mets", "metabolic_risk",
  "bone", "frax", "sarc_f", "charlson",
  "allostatic_load",
  "psych",
  "adiposity_sds", "adiposity_sds_strat",
  "alm_bmi", "obesity_metrics"
)

out_fast <- suppressWarnings(
  all_health_markers(
    df, col_map = cm,
    which = fast_groups,
    include_insulin = TRUE,
    normalize = "none",
    mode = "both",
    verbose = FALSE,
    na_action = "keep"
  )
)
message("all_health_markers (fast): ", ncol(out_fast), " output cols")
print(head(setdiff(names(out_fast), names(df)), 40))


# ---- 77 all_health_markers -- which='all' (slow; uncomment to run) ----
# out_full <- suppressWarnings(
#   all_health_markers(df, col_map = cm, which = "all",
#                      normalize = "none", mode = "both",
#                      verbose = TRUE, na_action = "keep")
# )
# message("all_health_markers (all): ", ncol(out_full), " cols")


# ---- 78 all_health_markers -- unknown group error ----
tryCatch(
  all_health_markers(df, col_map = cm, which = "does_not_exist",
                     include_insulin = FALSE, verbose = FALSE),
  error = function(e) message("Expected error caught: ", conditionMessage(e))
)


message("=== All manual tests complete ===")

#' Calculate a Suite of Renal Function, Injury, and Excretion Markers
#'
#' Given routine blood and urine assays, `renal_markers()` computes:
#'   - eGFR_cr: CKD-EPI creatinine equation (2009 variant; race factor retained to preserve prior behavior)
#'   - eGFR_cys: CKD-EPI cystatin C equation (if `cystatin_C` provided)
#'   - eGFR_combined: CKD-EPI combined creatinine+cystatin C (if both provided)
#'   - BUN_Cr_ratio: Blood urea nitrogen / serum creatinine
#'   - FE_Urea: Fractional excretion of urea (%)
#'   - NGAL, KIM1, NAG, Beta2Micro, IL18, L_FABP: pass-through urinary injury markers (if mapped)
#'
#' Robust validation is applied, including NA handling (`na_action`), high-missingness
#' diagnostics, safe divisions with a consolidated zero-denominator warning, and an
#' optional input extremes scan/cap. New arguments are appended for backward compatibility.
#'
#' Expected units (no automatic conversion performed):
#' - creatinine (serum): mg/dL
#' - cystatin C (serum): mg/L
#' - BUN (serum): mg/dL
#' - urea_serum, urea_urine: mg/dL
#' - creatinine_urine: mg/dL
#'
#' @param data A data.frame or tibble with renal lab data.
#' @param col_map Named list mapping:
#'   - creatinine -> serum creatinine (mg/dL)
#'   - age -> age (years)
#'   - sex -> sex indicator (1 = male, 0 = female). Also accepts "male"/"female".
#'   - race -> race ("white", "black", or "other"). Also accepts common aliases.
#'   - BUN -> blood urea nitrogen (mg/dL)
#'   - optional cystatin_C -> serum cystatin C (mg/L)
#'   - optional urea_serum -> serum urea (mg/dL)
#'   - optional creatinine_urine -> urine creatinine (mg/dL)
#'   - optional urea_urine -> urine urea (mg/dL)
#'   - optional NGAL, KIM1, NAG, beta2_micro, IL18, L_FABP -> urine injury markers
#' @param na_action One of `c("keep","omit","error")` for handling missing values in
#'   required inputs. Default "keep".
#' @param na_warn_prop Proportion (0-1) threshold for high-missingness diagnostics.
#'   Default 0.2.
#' @param verbose Logical; if `TRUE` (default), prints step-by-step progress
#'   including column mapping, optional input availability, physiological range
#'   information (informational only, values are not altered), the list of
#'   markers being computed, and a per-column results summary.
#'
#' @return A tibble with computed renal markers:
#'   `eGFR_cr`, `eGFR_cys`, `eGFR_combined`, `BUN_Cr_ratio`, `FE_Urea`,
#'   `NGAL`, `KIM1`, `NAG`, `Beta2Micro`, `IL18`, `L_FABP`.
#'   If an ID column is detected in `data` (e.g. `id`, `IID`, `participant_id`),
#'   it is prepended as the first output column.
#'
#' @examples
#' df <- tibble::tibble(Cr = 1.0, Age = 40, Sex = 1, Race = "white", BUN = 14)
#' cm <- list(creatinine = "Cr", age = "Age", sex = "Sex", race = "Race", BUN = "BUN")
#' renal_markers(df, cm)
#'
#' @note
#' `eGFR_cr` uses the 2009 CKD-EPI creatinine equation (Levey et al. 2009) with the
#' Black-race multiplier (\eqn{\times 1.159}) retained. The 2021 race-free CKD-EPI
#' equations (Inker et al., NEJM 2021) are not yet implemented; the `race` input
#' is accepted for forward compatibility and used only for the 2009 race factor.
#' `eGFR_cys` and `eGFR_combined` use Inker et al. (2012); note that
#' `eGFR_combined` applies its own sex (\eqn{\times 1.008} female) and race
#' (\eqn{\times 1.145} Black) multipliers, which differ from those of `eGFR_cr`.
#' `NGAL`, `KIM1`, `NAG`, `Beta2Micro`, `IL18`, and `L_FABP` are
#' **pass-through** columns — values are returned as-is with no formula applied.
#'
#' @references
#' \insertRef{levey2009ckdepi}{HealthMarkers}
#' \insertRef{inker2012cys}{HealthMarkers}
#' \insertRef{waikar2009feu}{HealthMarkers} (FE_Urea formula source; bib content: Kaplan and Kohn 1992)
#' \insertRef{parikh2011ngal}{HealthMarkers} (clinical context; NGAL is a pass-through biomarker)
#' \insertRef{vaidya2010kim1}{HealthMarkers} (KIM-1 biomarker qualification; pass-through)
#' \insertRef{portilla2008lfabp}{HealthMarkers} (L-FABP as AKI biomarker; pass-through)
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
renal_markers <- function(data,
                          col_map      = NULL,
                          na_action    = c("keep", "omit", "error"),
                          na_warn_prop = 0.2,
                          verbose      = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "renal_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data))
    stop(sprintf("%s(): `data` must be a data.frame or tibble.", fn_name), call. = FALSE)

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  all_keys <- c("creatinine", "age", "sex", "race", "BUN",
                "cystatin_C", "urea_serum", "creatinine_urine", "urea_urine",
                "NGAL", "KIM1", "NAG", "beta2_micro", "IL18", "L_FABP")
  cm      <- .hm_build_col_map(data, col_map, all_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # 1) Validate mapping and data presence
  required <- c("creatinine", "age", "sex", "race", "BUN")
  missing_map <- setdiff(required, names(col_map))
  if (length(missing_map)) {
    rlang::abort(
      paste0(fn_name, "(): missing col_map entries for: ", paste(missing_map, collapse = ", ")),
      class = "healthmarkers_renal_error_missing_map"
    )
  }
  hm_validate_inputs(data, col_map, required_keys = required, fn = fn_name)
  req_cols <- unname(unlist(col_map[required], use.names = FALSE))
  not_in_df <- setdiff(req_cols, names(data))
  if (length(not_in_df)) {
    rlang::abort(
      sprintf("%s(): mapped column(s) not found in data: %s", fn_name, paste(not_in_df, collapse = ", ")),
      class = "healthmarkers_renal_error_missing_columns"
    )
  }
  .rm_validate_df_numeric(data, col_map, cols = c("creatinine", "age", "BUN",
                                                  "cystatin_C", "urea_serum",
                                                  "creatinine_urine", "urea_urine"),
                          warn = TRUE)

  # 2) High-missingness diagnostics
  .rm_warn_high_missing(data, col_map[required], na_warn_prop = na_warn_prop)

  # 3) NA policy on required inputs
  used_cols <- unname(unlist(col_map[required], use.names = FALSE))
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("renal_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_renal_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
                msg   = sprintf("renal_markers(): omitting %d rows with NA in required inputs", sum(!keep)))
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      eGFR_cr = numeric(), eGFR_cys = numeric(), eGFR_combined = numeric(),
      BUN_Cr_ratio = numeric(), FE_Urea = numeric(),
      NGAL = numeric(), KIM1 = numeric(), NAG = numeric(),
      Beta2Micro = numeric(), IL18 = numeric(), L_FABP = numeric()
    ))
  }

  # --- Verbose: optional inputs block
  opt_keys <- c("cystatin_C", "urea_serum", "creatinine_urine", "urea_urine",
                "NGAL", "KIM1", "NAG", "beta2_micro", "IL18", "L_FABP")
  if (isTRUE(verbose)) {
    present_opt <- opt_keys[vapply(opt_keys, function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    missing_opt <- setdiff(opt_keys, present_opt)
    idx_deps <- list(
      eGFR_cys      = c("cystatin_C"),
      eGFR_combined = c("cystatin_C"),
      FE_Urea       = c("urea_serum", "creatinine_urine", "urea_urine")
    )
    na_indices <- character(0)
    for (idx in names(idx_deps)) {
      still_missing <- setdiff(idx_deps[[idx]], present_opt)
      if (length(still_missing))
        na_indices <- c(na_indices,
          sprintf("  %s -> NA  [missing: %s]", idx, paste(still_missing, collapse = ", ")))
    }
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(present_opt))
      lines <- c(lines, sprintf("  present:  %s", paste(present_opt, collapse = ", ")))
    if (length(missing_opt))
      lines <- c(lines, sprintf("  missing:  %s", paste(missing_opt, collapse = ", ")))
    if (length(na_indices))
      lines <- c(lines, "  indices -> NA:", na_indices)
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    rules     <- .rm_default_extreme_rules()
    flags     <- .rm_extreme_scan(data, col_map, rules)
    ex_count  <- sum(vapply(flags, function(v) sum(v, na.rm = TRUE), integer(1)))
    if (ex_count > 0L) {
      details <- vapply(names(flags), function(nm) {
        nb <- sum(flags[[nm]], na.rm = TRUE)
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", nm, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    has_cys  <- !is.null(col_map[["cystatin_C"]]) && col_map[["cystatin_C"]] %in% names(data)
    has_fe   <- all(c("urea_serum","creatinine_urine","urea_urine") %in% {
      opt_keys[vapply(opt_keys, function(k) {
        !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
      }, logical(1))]
    })
    status <- c(
      "  eGFR_cr        [creatinine, age, sex, race]",
      sprintf("  eGFR_cys       %s", if (has_cys) "[cystatin_C, age, sex]" else "NA [cystatin_C missing]"),
      sprintf("  eGFR_combined  %s", if (has_cys) "[creatinine, cystatin_C, age, sex, race]" else "NA [cystatin_C missing]"),
      "  BUN_Cr_ratio   [BUN, creatinine]",
      sprintf("  FE_Urea        %s", if (has_fe) "[urea_serum, urea_urine, creatinine_urine, creatinine]" else "NA [urea/urine inputs missing]")
    )
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # 5) Pull and normalize inputs (sex/race mapping)
  Cr   <- data[[col_map$creatinine]] # mg/dL
  age  <- data[[col_map$age]]
  sexi <- .rm_map_sex(data[[col_map$sex]]) # 1 male, 0 female
  race <- .rm_map_race(data[[col_map$race]]) # "black","white","other"
  BUN  <- data[[col_map$BUN]]

  # Optional inputs
  Cys <- if ("cystatin_C" %in% names(col_map)) data[[col_map$cystatin_C]] else NULL
  Urea_s <- if ("urea_serum" %in% names(col_map)) data[[col_map$urea_serum]] else NULL
  Cr_u   <- if ("creatinine_urine" %in% names(col_map)) data[[col_map$creatinine_urine]] else NULL
  Urea_u <- if ("urea_urine" %in% names(col_map)) data[[col_map$urea_urine]] else NULL

  # 6) Helpers: safe division with zero-denominator tracking
  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }

  # 7) eGFR via CKD-EPI creatinine (2009)
  kappa <- ifelse(sexi == 1, 0.9, 0.7)
  alpha <- ifelse(sexi == 1, -0.411, -0.329)
  min_ratio <- pmin(Cr / kappa, 1)
  max_ratio <- pmax(Cr / kappa, 1)
  factor_race   <- ifelse(race == "black", 1.159, 1)   # Levey 2009: Black race multiplier
  factor_sex_cr <- ifelse(sexi == 1, 1, 1.018)          # Levey 2009: female multiplier
  eGFR_cr <- 141 * (min_ratio^alpha) * (max_ratio^-1.209) *
    (0.993^age) * factor_race * factor_sex_cr

  # 8) eGFR via cystatin C & combined equations
  if (!is.null(Cys)) {
    min_cys <- pmin(Cys / 0.8, 1)
    max_cys <- pmax(Cys / 0.8, 1)
    cys_sex <- ifelse(sexi == 1, 1, 0.932)
    eGFR_cys <- 133 * (min_cys^-0.499) * (max_cys^-1.328) *
      (0.996^age) * cys_sex
    # combined creatinine + cystatin (Inker 2012; sex/race factors differ from eGFR_cr)
    factor_sex_comb  <- ifelse(sexi == 1, 1, 1.008)   # Inker 2012: female multiplier
    factor_race_comb <- ifelse(race == "black", 1.145, 1)  # Inker 2012: race multiplier
    eGFR_combined <- 135 *
      (min_ratio^alpha) * (max_ratio^-0.601) *
      (min_cys^-0.375) * (max_cys^-0.711) *
      (0.995^age) * factor_race_comb * factor_sex_comb
  } else {
    eGFR_cys <- NA_real_
    eGFR_combined <- NA_real_
  }

  # 9) BUN/Cr ratio
  BUN_Cr_ratio <- safe_div(BUN, Cr, "BUN_Cr_ratio")

  # 10) Fractional excretion of urea (%)
  if (!is.null(Urea_s) && !is.null(Cr_u) && !is.null(Urea_u)) {
    FE_Urea <- 100 * safe_div(safe_div(Urea_u, Urea_s, "FE_Urea_UreaS"),
                              safe_div(Cr_u, Cr, "FE_Urea_CrS"),
                              "FE_Urea_final")
  } else {
    FE_Urea <- NA_real_
  }

  # 11) Urine injury markers (pass-through if mapped)
  get_mark <- function(key) {
    if (key %in% names(col_map) && col_map[[key]] %in% names(data)) {
      as.numeric(data[[ col_map[[key]] ]])
    } else {
      rep(NA_real_, nrow(data))
    }
  }
  NGAL        <- get_mark("NGAL")
  KIM1        <- get_mark("KIM1")
  NAG         <- get_mark("NAG")
  Beta2Micro  <- get_mark("beta2_micro")
  IL18        <- get_mark("IL18")
  L_FABP      <- get_mark("L_FABP")

  out <- tibble::tibble(
    eGFR_cr        = as.numeric(eGFR_cr),
    eGFR_cys       = as.numeric(eGFR_cys),
    eGFR_combined  = as.numeric(eGFR_combined),
    BUN_Cr_ratio   = as.numeric(BUN_Cr_ratio),
    FE_Urea        = as.numeric(FE_Urea),
    NGAL           = as.numeric(NGAL),
    KIM1           = as.numeric(KIM1),
    NAG            = as.numeric(NAG),
    Beta2Micro     = as.numeric(Beta2Micro),
    IL18           = as.numeric(IL18),
    L_FABP         = as.numeric(L_FABP)
  )

  # 12) Emit consolidated zero-denominator warning if any
  dz <- dz_env$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz); nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("renal_markers(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# --------------------- internal helpers ---------------------------------------

.rm_validate_df_numeric <- function(data, col_map, cols, warn = TRUE) {
  for (key in cols) {
    if (!(key %in% names(col_map))) next
    cn <- col_map[[key]]
    if (!(cn %in% names(data))) next
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      if (isTRUE(warn)) {
        introduced <- sum(is.na(data[[cn]]) & !is.na(old))
        if (introduced > 0) rlang::warn(sprintf("renal_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
  }
  invisible(TRUE)
}

.rm_warn_high_missing <- function(data, mapped_names, na_warn_prop = 0.2) {
  cols <- unname(unlist(mapped_names, use.names = FALSE))
  for (cn in cols) {
    x <- data[[cn]]
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("renal_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }
  invisible(TRUE)
}

.rm_map_sex <- function(sex) {
  s <- tolower(as.character(sex))
  out <- rep(NA_integer_, length(s))
  # numeric-like mapping
  is_num <- suppressWarnings(!is.na(as.numeric(s)))
  out[is_num & s %in% c("1")] <- 1L
  out[is_num & s %in% c("0", "2")] <- 0L
  # character mapping
  out[s %in% c("male","m")] <- 1L
  out[s %in% c("female","f","fm","woman","girl")] <- 0L
  bad <- sum(is.na(out))
  if (bad > 0) rlang::warn(sprintf("renal_markers(): 'sex' has %d unmapped values; set to NA.", bad))
  out
}

.rm_map_race <- function(race) {
  r <- tolower(as.character(race))
  out <- ifelse(
    r %in% c("black","african-american","african american","aa","blk"), "black",
    ifelse(r %in% c("white","caucasian","european","non-hispanic white"), "white", "other")
  )
  out[is.na(out) | out == ""] <- "other"
  out
}

.rm_default_extreme_rules <- function() {
  list(
    creatinine       = c(0.1, 15),    # mg/dL
    BUN              = c(1, 200),     # mg/dL
    cystatin_C       = c(0.2, 8),     # mg/L
    urea_serum       = c(1, 300),     # mg/dL
    creatinine_urine = c(1, 500),     # mg/dL
    urea_urine       = c(1, 2000)     # mg/dL
  )
}

.rm_extreme_scan <- function(data, col_map, rules) {
  flags <- list()
  for (key in names(rules)) {
    if (!(key %in% names(col_map))) next
    cn <- col_map[[key]]
    if (!(cn %in% names(data))) next
    rng <- rules[[key]]
    x <- data[[cn]]
    flags[[key]] <- is.finite(x) & (x < rng[1] | x > rng[2])
  }
  flags
}

library(testthat)
library(tibble)

# Minimal required mapping and a helper to build small test data frames
cm_req <- list(
  creatinine = "Cr",
  age        = "Age",
  sex        = "Sex",
  race       = "Race",
  BUN        = "BUN"
)

make_df_req <- function(Cr = 1.0, Age = 40, Sex = 1, Race = "white", BUN = 14) {
  tibble(Cr = Cr, Age = Age, Sex = Sex, Race = Race, BUN = BUN)
}

test_that("verbose emits col_map, optional inputs, computing markers, and results messages", {
  skip_on_cran()
  df <- make_df_req()
  msgs <- testthat::capture_messages(renal_markers(df, cm_req, verbose = TRUE))
  expect_true(any(grepl("col_map", msgs)))
  expect_true(any(grepl("optional inputs", msgs)))
  expect_true(any(grepl("computing markers", msgs)))
  expect_true(any(grepl("results:", msgs)))
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- make_df_req()
  msgs <- testthat::capture_messages(renal_markers(df, cm_req, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",        msgs)), 1L)
  expect_equal(sum(grepl("optional inputs", msgs)), 1L)
  expect_equal(sum(grepl("computing markers", msgs)), 1L)
})

test_that("error when mapped column not found in data", {
  skip_on_cran()
  df <- tibble(Age = 40, Sex = 1, Race = "white", BUN = 14) # missing Cr
  expect_error(
    renal_markers(df, cm_req, verbose = FALSE),
    "not found"
  )
})

test_that("na_action='omit' dropping all rows returns empty tibble with expected columns", {
  skip_on_cran()
  df <- make_df_req(Cr = NA_real_)
  out <- renal_markers(df, cm_req, na_action = "omit", verbose = FALSE)
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
  expect_true(all(c(
    "eGFR_cr","eGFR_cys","eGFR_combined","BUN_Cr_ratio","FE_Urea",
    "NGAL","KIM1","NAG","Beta2Micro","IL18","L_FABP"
  ) %in% names(out)))
})

test_that("sex mapping: numeric vs string agree; male vs female differ", {
  skip_on_cran()
  df_code <- make_df_req(Cr = 1.0, Age = 50, Sex = 1, Race = "white", BUN = 14)
  df_strm <- make_df_req(Cr = 1.0, Age = 50, Sex = "male", Race = "white", BUN = 14)
  out_code <- renal_markers(df_code, cm_req, verbose = FALSE)
  out_strm <- renal_markers(df_strm, cm_req, verbose = FALSE)
  expect_equal(out_code$eGFR_cr, out_strm$eGFR_cr, tolerance = 1e-8)

  df_f_code <- make_df_req(Cr = 1.0, Age = 50, Sex = 0, Race = "white", BUN = 14)
  df_f_str  <- make_df_req(Cr = 1.0, Age = 50, Sex = "female", Race = "white", BUN = 14)
  out_f_code <- renal_markers(df_f_code, cm_req, verbose = FALSE)
  out_f_str  <- renal_markers(df_f_str, cm_req, verbose = FALSE)
  expect_equal(out_f_code$eGFR_cr, out_f_str$eGFR_cr, tolerance = 1e-8)

  expect_false(isTRUE(all.equal(out_code$eGFR_cr, out_f_code$eGFR_cr)))
})

test_that("race mapping: black vs white scales eGFR_cr by ~1.159 (all else equal)", {
  skip_on_cran()
  df_w <- make_df_req(Cr = 1.0, Age = 50, Sex = 1, Race = "caucasian", BUN = 14)
  df_b <- make_df_req(Cr = 1.0, Age = 50, Sex = 1, Race = "african american", BUN = 14)
  out_w <- renal_markers(df_w, cm_req, verbose = FALSE)
  out_b <- renal_markers(df_b, cm_req, verbose = FALSE)
  expect_gt(out_b$eGFR_cr, 0)
  expect_gt(out_w$eGFR_cr, 0)
  expect_equal(unname(out_b$eGFR_cr / out_w$eGFR_cr), 1.159, tolerance = 1e-6)
})

test_that("extreme values produce no warning/error; range note appears in verbose", {
  skip_on_cran()
  df_ext <- make_df_req(Cr = 10, Age = 90, Sex = 1, Race = "white", BUN = 300)
  expect_no_warning(
    out <- renal_markers(df_ext, cm_req, verbose = FALSE)
  )
  expect_true(is.numeric(out$eGFR_cr))
  # Range note appears in verbose
  msgs <- testthat::capture_messages(renal_markers(df_ext, cm_req, verbose = TRUE))
  expect_true(any(grepl("range note", msgs)))
})

test_that("FE_Urea zero denominators yield consolidated warning and NA", {
  skip_on_cran()
  df1 <- tibble(Cr = 1.0, Age = 40, Sex = 1, Race = "white", BUN = 14,
                U_s = 0, Cr_u = 2, U_u = 80)
  cm1 <- modifyList(cm_req, list(urea_serum = "U_s", creatinine_urine = "Cr_u", urea_urine = "U_u"))
  expect_warning(
    out1 <- renal_markers(df1, cm1, verbose = FALSE),
    "zero denominators detected"
  )
  expect_true(is.na(out1$FE_Urea))

  df2 <- tibble(Cr = 1.0, Age = 40, Sex = 1, Race = "white", BUN = 14,
                U_s = 40, Cr_u = 0, U_u = 80)
  expect_warning(
    out2 <- renal_markers(df2, cm1, verbose = FALSE),
    "zero denominators detected"
  )
  expect_true(is.na(out2$FE_Urea))
})

test_that("ID column is prepended to output when detected", {
  skip_on_cran()
  df_id <- tibble::tibble(
    id = 1:3,
    Cr = c(1.0, 1.2, 0.9), Age = c(40, 50, 35), Sex = 1L, Race = "white", BUN = 14
  )
  out <- renal_markers(df_id, cm_req, verbose = FALSE)
  expect_equal(names(out)[1L], "id")
  expect_equal(out$id, 1:3)
})

#' Oxidative stress markers
#'
#' Computes GSH_GSSG_Ratio = reduced glutathione (GSH) / oxidized glutathione (GSSG).
#'
#' @param data Data frame with columns for GSH and GSSG (per col_map).
#' @param col_map Named list with required keys `GSH` and `GSSG`. Defaults assume
#'   column names match keys. Both columns must be in the same units (e.g., µmol/L).
#' @param na_action One of c("keep","omit","error").
#' @param verbose Logical; if `TRUE` (default), prints column mapping and a
#'   per-column results summary.
#' @return A tibble with column GSH_GSSG_Ratio. If an ID column is detected,
#'   it is prepended.
#'
#' @note
#' `GSH_GSSG_Ratio` is dimensionless only when `GSH` and `GSSG` are supplied in
#' the same units (typically µmol/L). The formula `GSH / GSSG` is a standard
#' biochemical redox ratio; no unit conversion is applied.
#'
#' @references
#' \insertRef{glutathione_redox_review}{HealthMarkers} (background review;
#'   GSH/GSSG is a standard biochemical redox ratio, not a formula from this paper)
#'
#' @examples
#' df <- data.frame(GSH = c(5, 3), GSSG = c(1, 0.5))
#' oxidative_markers(df, col_map = list(GSH="GSH", GSSG="GSSG"))
#'
#' @importFrom rlang abort warn inform
#' @importFrom tibble tibble
#' @export
oxidative_markers <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "oxidative_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  id_col    <- .hm_detect_id_col(data)

  # HM-CS v2 validation
  hm_validate_inputs(
    data, col_map,
    required_keys = character(0),
    fn = "oxidative_markers"
  )
  cm      <- .hm_build_col_map(data, col_map, keys = c("GSH", "GSSG"), fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Ensure mapped columns exist in data
  mapped <- unname(unlist(col_map[c("GSH","GSSG")], use.names = FALSE))
  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("oxidative_markers(): missing columns in `data`: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_oxidative_error_missing_columns"
    )
  }

  # Coerce required columns to numeric; warn if NAs introduced; set non-finite to NA
  for (cn in mapped) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::warn(sprintf("oxidative_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug)
  for (cn in mapped) {
    x <- data[[cn]]; n <- length(x); if (!n) next
    pna <- sum(is.na(x)) / n
    if (pna >= 0.2 && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("oxidative_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }

  # --- Verbose: column mapping
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    hm_inform(
      sprintf("%s(): computing markers:\n  GSH_GSSG_Ratio  [GSH / GSSG]", fn_name),
      level = "inform"
    )
  }

  # NA policy
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(mapped, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("oxidative_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_oxidative_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(mapped, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = if (isTRUE(verbose)) "inform" else "debug")
    data <- data[keep, , drop = FALSE]
  }

  hm_inform("oxidative_markers(): computing markers", level = "debug")

  # Accessors (post-filter)
  GSH  <- data[[col_map$GSH]]
  GSSG <- data[[col_map$GSSG]]

  # Safe division: NA on non-finite or zero denom
  sdiv <- function(a, b) {
    out <- a / b
    out[!is.finite(out)] <- NA_real_
    out
  }

  out <- tibble::tibble(GSH_GSSG_Ratio = sdiv(GSH, GSSG))

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

test_that("oxidative_markers computes GSH/GSSG ratio", {
  skip_on_cran()
  df <- data.frame(GSH = c(1000, 1500), GSSG = c(10, 15))
  res <- oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"))
  expect_s3_class(res, "tbl_df")
  expect_true("GSH_GSSG_Ratio" %in% names(res))
  expect_equal(res$GSH_GSSG_Ratio, c(100, 100), tolerance = 1e-10)
})

test_that("oxidative_markers safe division returns NA on zero denom", {
  skip_on_cran()
  df <- data.frame(GSH = c(10, 10), GSSG = c(0, 10))
  res <- oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"))
  expect_true(is.na(res$GSH_GSSG_Ratio[1]))
  expect_equal(res$GSH_GSSG_Ratio[2], 1)
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(GSH = c(1000, 1500), GSSG = c(10, 15))
  expect_message(oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), verbose = TRUE), "oxidative_markers")
  expect_message(oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), verbose = TRUE), "col_map")
  expect_message(oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(GSH = c(1000, 1500), GSSG = c(10, 15))
  msgs <- testthat::capture_messages(
    oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action='keep' propagates NA", {
  skip_on_cran()
  df <- data.frame(GSH = c(5, NA), GSSG = c(1, 2))
  res <- oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), na_action = "keep")
  expect_equal(res$GSH_GSSG_Ratio[1], 5)
  expect_true(is.na(res$GSH_GSSG_Ratio[2]))
})

test_that("na_action='omit' drops rows with NA", {
  skip_on_cran()
  df <- data.frame(GSH = c(5, NA, 3), GSSG = c(1, 2, 0.5))
  res <- oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), na_action = "omit")
  expect_equal(nrow(res), 2L)
  expect_equal(res$GSH_GSSG_Ratio, c(5, 6), tolerance = 1e-10)
})

test_that("na_action='error' aborts when NA present", {
  skip_on_cran()
  df <- data.frame(GSH = c(5, NA), GSSG = c(1, 2))
  expect_error(
    oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"), na_action = "error"),
    "required inputs contain missing values"
  )
})

test_that("missing column in data errors with clear message", {
  skip_on_cran()
  df <- data.frame(GSH = c(5, 3))
  expect_error(
    oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG")),
    "GSSG"
  )
})

test_that("Inf GSSG yields NA ratio", {
  skip_on_cran()
  df <- data.frame(GSH = c(10, 10), GSSG = c(Inf, 5))
  res <- oxidative_markers(df, col_map = list(GSH = "GSH", GSSG = "GSSG"))
  expect_true(is.na(res$GSH_GSSG_Ratio[1]))
  expect_equal(res$GSH_GSSG_Ratio[2], 2)
})

#' Calculate OGTT-based insulin sensitivity indices
#'
#' Given glucose & insulin at 0, 30, 120 min (plus weight, BMI, age, sex),
#' computes:
#' - Isi_120
#' - Cederholm_index
#' - Gutt_index
#' - Avignon_Si0
#' - Avignon_Si120
#' - Avignon_Sim
#' - Modified_stumvoll
#' - Stumvoll_Demographics
#' - Matsuda_AUC
#' - Matsuda_ISI
#' - BigttSi
#' - Ifc_inv
#' - HIRI_inv
#' - Belfiore_isi_gly
#'
#' Units assumed:
#' - OGTT glucose in mmol/L (internally converted to mg/dL via *18 for select indices)
#' - OGTT insulin in pmol/L (internally converted to muU/mL via /6 for select indices)
#' - weight in kg; BMI in kg/m^2; age in years; sex coded 1 = male, 2 = female
#'
#' Notes
#' - Conversions mirror existing implementation to preserve outputs. Some
#'   formulas intentionally use unconverted inputs (as in prior code).
#' - Modified_stumvoll and Stumvoll_Demographics use raw pmol/L and mmol/L
#'   as published in Stumvoll et al. (2000). BigttSi likewise uses raw units.
#' - Matsuda_AUC is a non-standard AUC-based variant; the original Matsuda
#'   index (Matsuda_ISI) uses time-point means, not AUCs.
#' - Cederholm_index uses log(I0 + I120) as implemented (sum, not mean);
#'   Gutt_index uses log((I0 + I120)/2) (mean). This mirrors published
#'   implementations; the difference is a constant log(2) offset.
#' - Ifc_inv and HIRI_inv are derived composite proxies not attributed to a
#'   single formula publication; treat as research tools.
#' - Logs are safe: log(x) becomes NA when x <= 0 or non-finite.
#'
#' @param data A data.frame or tibble containing at least the columns mapped by `col_map`.
#' @param col_map Named list mapping:
#'   - G0, G30, G120 -> glucose at 0, 30, 120 min (mmol/L)
#'   - I0, I30, I120 -> insulin at 0, 30, 120 min (pmol/L)
#'   - weight -> body weight (kg)
#'   - bmi -> body-mass index (kg/m^2)
#'   - age -> age (years)
#'   - sex -> sex (1 = male, 2 = female)
#' @param normalize One of c("none","z","inverse","range","robust") used by normalize_vec().
#' @param verbose Logical; if `TRUE` (default), prints column mapping, the list
#'   of indices being computed, and a per-column results summary.
#' @param na_action One of c("keep","omit","error") for missing/non-finite required inputs. Default "keep".
#' @param na_warn_prop Proportion (0-1) for high-missingness diagnostics (debug). Default 0.2.
#'
#' @return A tibble with the OGTT-based index columns listed above. If an ID
#'   column is detected in `data` (e.g. `id`, `IID`, `participant_id`), it is
#'   prepended as the first output column.
#' @importFrom tibble tibble
#' @importFrom dplyr mutate across everything
#' @export
#'
#' @seealso [fasting_is()], [normalize_vec()]
#'
#' @references
#' \insertRef{matsuda1999ogtt}{HealthMarkers}
#' \insertRef{gutt2000isi}{HealthMarkers}
#' \insertRef{stumvoll2000ogtt}{HealthMarkers}
#' \insertRef{hansen2007bigtt}{HealthMarkers}
#' \insertRef{avignon1999ogtt}{HealthMarkers}
#' \insertRef{belfiore1998insulin}{HealthMarkers}
#' \insertRef{matthews1985homa}{HealthMarkers}
#' \insertRef{suleman2024is}{HealthMarkers} (genetic epidemiology study reviewing IS indices)
#'
#' @examples
#' df <- tibble::tibble(
#'   G0 = 5.5, I0 = 60,
#'   G30 = 7.8, I30 = 90,
#'   G120 = 6.2, I120 = 50,
#'   weight = 70, bmi = 24, age = 30, sex = 1
#' )
#' ogtt_is(
#'   df,
#'   col_map = list(
#'     G0 = "G0", I0 = "I0",
#'     G30 = "G30", I30 = "I30",
#'     G120 = "G120", I120 = "I120",
#'     weight = "weight", bmi = "bmi",
#'     age = "age", sex = "sex"
#'   ),
#'   normalize = "none",
#'   verbose = TRUE
#' )
ogtt_is <- function(data,
                    col_map = NULL,
                    normalize = "none",
                    verbose = TRUE,
                    na_action = c("keep","omit","error"),
                    na_warn_prop = 0.2) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name    <- "ogtt_is"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action  <- match.arg(na_action)
  id_col     <- .hm_detect_id_col(data)
  .ogtt_validate_args(normalize, na_warn_prop, verbose)

  req_keys <- c("G0","I0","G30","I30","G120","I120","weight","bmi","age","sex")
  cm      <- .hm_build_col_map(data, col_map, req_keys, fn = "ogtt_is")
  data    <- cm$data
  col_map <- cm$col_map

  # HM-CS v2: required keys validation
  hm_validate_inputs(
    data, col_map,
    required_keys = req_keys,
    fn = "ogtt_is"
  )

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    indices <- list(
      c("Isi_120",            "G120, I120"),
      c("Cederholm_index",    "G0, G120, I0, I120, weight"),
      c("Gutt_index",         "G0, G120, I0, I120, weight"),
      c("Avignon_Si0",        "G0, I0, weight"),
      c("Avignon_Si120",      "G120, I120, weight"),
      c("Avignon_Sim",        "Si0, Si120"),
      c("Modified_stumvoll",  "I0, I120, G120"),
      c("Stumvoll_Demog.",    "bmi, I120, age"),
      c("Matsuda_AUC",        "G0, I0, G_AUC, I_AUC"),
      c("Matsuda_ISI",        "G0, I0, G_mean, I_mean"),
      c("BigttSi",            "I0, I30, I120, G0, G30, G120, sex, bmi"),
      c("Ifc_inv",            "I0, I120"),
      c("HIRI_inv",           "G0, G30, I0, I30"),
      c("Belfiore_isi_gly",   "I_AUC, G_AUC")
    )
    status <- vapply(indices, function(x) sprintf("  %-22s [%s]", x[1], x[2]), character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Coerce required columns to numeric if needed (warn on NAs introduced)
  keys_to_check <- c("G0","I0","G30","I30","G120","I120","weight","bmi","age","sex")
  for (k in keys_to_check) {
    cn <- col_map[[k]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::warn(sprintf("ogtt_is(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug verbosity)
  .ogtt_high_missing_diag(
    data,
    vars = unname(unlist(col_map[keys_to_check], use.names = FALSE)),
    na_warn_prop = na_warn_prop
  )

  # NA policy on required inputs
  used_cols <- unname(unlist(col_map[keys_to_check], use.names = FALSE))
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("ogtt_is(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_ogtt_is_error_missing_values")
    }
  } else if (na_action == "omit" && length(used_cols)) {
    keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  hm_inform(sprintf("%s(): converting units", fn_name), level = "debug")

  # 1) Extract & convert raw inputs
  G0   <- data[[col_map$G0]]   * 18 # mg/dL
  G30  <- data[[col_map$G30]]  * 18
  G120 <- data[[col_map$G120]] * 18
  I0   <- data[[col_map$I0]]   / 6  # muU/mL
  I30  <- data[[col_map$I30]]  / 6
  I120 <- data[[col_map$I120]] / 6

  wt   <- data[[col_map$weight]]
  bmi  <- data[[col_map$bmi]]
  age  <- data[[col_map$age]]
  # Normalize sex to 1=male, 0=female for formula use
  sex  <- .hm_normalize_sex(data[[col_map$sex]], to = "10", fn = "ogtt_is")

  hm_inform(sprintf("%s(): computing indices", fn_name), level = "debug")

  # Helpers: safe log and safe division
  lg <- function(x) .ogtt_log(x)
  dv <- function(a, b) .ogtt_safe_div(a, b)

  # 2) Areas under curve & means (trapezoid with intervals 0-30-120)
  I_AUC <- 0.5 * ((I0 + I30) * 30 + (I30 + I120) * 90)
  G_AUC <- 0.5 * ((G0 + G30) * 30 + (G30 + G120) * 90)
  I_mean <- rowMeans(cbind(I0, I30, I120), na.rm = TRUE)
  G_mean <- rowMeans(cbind(G0, G30, G120), na.rm = TRUE)

  # 3) Compute indices
  out <- tibble::tibble(
    Isi_120 = dv(10000, (G120 * I120)),
    Cederholm_index = dv(75000 + (G0 - G120) * 1.15 * 180 * 0.19 * wt,
                         120 * ((G0 + G120) / 2) * lg(I0 + I120)),
    Gutt_index = dv(75000 + (G0 - G120) * 0.19 * wt,
                    120 * ((G0 + G120) / 2) * lg((I0 + I120) / 2)),
    Avignon_Si0   = dv(1e8, (G0 * I0)   * wt * 150),
    Avignon_Si120 = dv(1e8, (G120 * I120) * wt * 150),
    Avignon_Sim   = (Avignon_Si0 + Avignon_Si120) / 2,
    Modified_stumvoll = 0.156 -
      0.0000459 * data[[col_map$I120]] -
      0.000321  * data[[col_map$I0]]   -
      0.00541   * data[[col_map$G120]],
    Stumvoll_Demographics = 0.222 -
      0.00333   * bmi -
      0.0000779 * data[[col_map$I120]] -
      0.000422  * age,
    Matsuda_AUC = dv(10000, sqrt(G0 * I0 * G_AUC * I_AUC)),
    Matsuda_ISI = dv(10000, sqrt(G0 * I0 * G_mean * I_mean)),
    BigttSi = exp(
      4.90 -
        0.00402 * data[[col_map$I0]]   -
        0.000565* data[[col_map$I30]]  -
        0.00127 * data[[col_map$I120]] -
        0.152   * data[[col_map$G0]]   -
        0.00871 * data[[col_map$G30]]  -
        0.0373  * data[[col_map$G120]] -
        ifelse(sex == 1, 0.145, 0) -
        0.0376  * bmi
    ),
    Ifc_inv = -lg(dv(data[[col_map$I120]], data[[col_map$I0]])),
    HIRI_inv = -(((G0 + G30) / 2) / 100 * ((I0 + I30) / 2)),
    Belfiore_isi_gly = dv(2, (I_AUC * G_AUC) + 1)
  )

  # 4) Normalize if requested
  out <- dplyr::mutate(
    out,
    dplyr::across(
      dplyr::everything(),
      ~ HealthMarkers::normalize_vec(.x, method = normalize)
    )
  )

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# ---- internal helpers (not exported) -----------------------------------------

.ogtt_validate_args <- function(normalize, na_warn_prop, verbose) {
  ok_norm <- normalize %in% c("none","z","inverse","range","robust")
  if (!ok_norm) rlang::abort("`normalize` must be one of: 'none','z','inverse','range','robust'.",
                             class = "healthmarkers_ogtt_is_error_normalize")
  if (!(is.numeric(na_warn_prop) && length(na_warn_prop) == 1L && is.finite(na_warn_prop) &&
        na_warn_prop >= 0 && na_warn_prop <= 1)) {
    rlang::abort("`na_warn_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_ogtt_is_error_na_warn_prop")
  }
  if (!(is.logical(verbose) && length(verbose) == 1L && !is.na(verbose))) {
    rlang::abort("`verbose` must be a single logical value.",
                 class = "healthmarkers_ogtt_is_error_verbose")
  }
  invisible(TRUE)
}

.ogtt_high_missing_diag <- function(df, vars, na_warn_prop = 0.2) {
  if (!length(vars)) return(invisible(TRUE))
  for (v in vars) {
    x <- df[[v]]
    n <- length(x); if (!n) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("ogtt_is(): column '%s' has high missingness (%.1f%%).", v, 100 * pna))
    }
  }
  invisible(TRUE)
}

.ogtt_safe_div <- function(num, den) {
  out <- num / den
  out[!is.finite(out)] <- NA_real_
  out
}

.ogtt_log <- function(x) {
  y <- x
  y[!(is.finite(y) & y > 0)] <- NA_real_
  out <- suppressWarnings(log(y))
  out[!is.finite(out)] <- NA_real_
  out
}

library(testthat)
library(tibble)
library(dplyr)

# helper to run ogtt_is more concisely
run_ogtt <- function(df, normalize = "none", verbose = FALSE) {
  ogtt_is(
    df,
    col_map = list(
      G0     = "G0",     I0      = "I0",
      G30    = "G30",    I30     = "I30",
      G120   = "G120",   I120    = "I120",
      weight = "weight", bmi     = "bmi",
      age    = "age",    sex     = "sex"
    ),
    normalize = normalize,
    verbose = verbose
  )
}

# base single-row dataset
base_df <- tibble(
  G0     = 5.5,
  I0     = 60,
  G30    = 7.8,
  I30    = 90,
  G120   = 6.2,
  I120   = 50,
  weight = 70,
  bmi    = 24,
  age    = 30,
  sex    = 1
)

test_that("returns all expected indices for single row", {
  skip_on_cran()
  out <- run_ogtt(base_df, normalize = "none")
  expected_cols <- c(
    "Isi_120", "Cederholm_index", "Gutt_index",
    "Avignon_Si0", "Avignon_Si120", "Avignon_Sim",
    "Modified_stumvoll", "Stumvoll_Demographics",
    "Matsuda_AUC", "Matsuda_ISI",
    "BigttSi", "Ifc_inv", "HIRI_inv", "Belfiore_isi_gly"
  )
  expect_true(all(expected_cols %in% names(out)))
  expect_true(all(vapply(out, function(x) is.finite(x[1]), logical(1))))
})

test_that("vectorized input: two rows gives two outputs", {
  skip_on_cran()
  df2 <- dplyr::bind_rows(base_df, dplyr::mutate(base_df, G0 = 6))
  out2 <- run_ogtt(df2)
  expect_equal(nrow(out2), 2)
})

test_that("normalize = 'range' scales variable indices into [0,1], constants map to a bound", {
  skip_on_cran()
  df2 <- dplyr::bind_rows(base_df, dplyr::mutate(base_df, G0 = 8))
  out_r <- suppressWarnings(run_ogtt(df2, normalize = "range"))
  for (col in names(out_r)) {
    vals <- out_r[[col]]
    v2 <- vals[!is.na(vals)]
    if (length(unique(v2)) > 1) {
      expect_gte(min(v2), 0 - 1e-12)
      expect_lte(max(v2), 1 + 1e-12)
    } else {
      expect_true(all(is.finite(vals)))
      expect_true(all(vals >= -1e-12 & vals <= 1 + 1e-12))
    }
  }
})

test_that("normalize = 'z' gives mean ~ 0 and sd ~ 1 on variable indices; constants -> zeros", {
  skip_on_cran()
  df3 <- dplyr::bind_rows(
    base_df,
    dplyr::mutate(base_df, G0 = 8),
    dplyr::mutate(base_df, G0 = 10)
  )
  out_z <- suppressWarnings(run_ogtt(df3, normalize = "z"))
  for (col in names(out_z)) {
    vals <- out_z[[col]]
    v2 <- vals[!is.na(vals)]
    if (length(unique(v2)) > 1) {
      expect_equal(mean(v2), 0, tolerance = 1e-6)
      expect_equal(sd(v2), 1, tolerance = 1e-6)
    } else {
      expect_true(all(abs(vals) < 1e-10 | is.na(vals)))
    }
  }
})

test_that("normalize = 'inverse' and 'robust' run and return a tibble", {
  skip_on_cran()
  df2 <- dplyr::bind_rows(base_df, dplyr::mutate(base_df, I0 = 30))
  out_inv <- suppressWarnings(run_ogtt(df2, normalize = "inverse"))
  out_rob <- suppressWarnings(run_ogtt(df2, normalize = "robust"))
  expect_s3_class(out_inv, "tbl_df")
  expect_s3_class(out_rob, "tbl_df")
})

test_that("invalid normalize argument errors via ogtt_is arg check", {
  skip_on_cran()
  expect_error(
    run_ogtt(base_df, normalize = "foo"),
    "`normalize` must be one of"
  )
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(run_ogtt(base_df, verbose = TRUE), "ogtt_is")
  expect_message(run_ogtt(base_df, verbose = TRUE), "col_map")
  expect_message(run_ogtt(base_df, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(run_ogtt(base_df, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("partial col_map is supplemented by dictionary inference", {
  skip_on_cran()
  bad_map <- list(G0 = "G0", I0 = "I0") # too few entries, rest inferred
  expect_no_error(
    ogtt_is(base_df, col_map = bad_map)
  )
})

test_that("BigttSi accounts for sex = 2 differently", {
  skip_on_cran()
  out_m <- run_ogtt(dplyr::mutate(base_df, sex = 1))
  out_f <- run_ogtt(dplyr::mutate(base_df, sex = 2))
  expect_false(isTRUE(all.equal(out_m$BigttSi, out_f$BigttSi)))
})

#' Compute Frailty (Deficit) Index using di::di with QA and verbose summaries
#'
#' Thin wrapper around the di package's di() that:
#' - Validates inputs and arguments.
#' - Coerces tibbles to base data.frames (for di()'s class checks).
#' - Auto-selects numeric deficit columns when cols = NULL (excluding age if supplied).
#' - Optionally scans for missing/out-of-range values with warnings or errors.
#' - Provides step-by-step verbose output and a completion summary.
#' - Optionally returns a tidy tibble instead of the original list.
#'
#' Background
#' The Frailty Index (FI) is computed as the proportion of health deficits present
#' in an individual across a set of candidate deficits. The approach was introduced
#' and formalized by Rockwood and Mitnitski and subsequently standardized for
#' construction and reporting.
#'
#' @param data A data.frame or tibble of health deficits (ideally binary/logical or
#'   scaled to \eqn{[0,1]}). Non-binary numeric columns can be rescaled by di::di when
#'   rescale = TRUE.
#' @param cols Character vector of deficit column names to use. If NULL (default),
#'   all numeric columns are used except age (if supplied).
#' @param invert Character vector of column names whose values should be inverted
#'   by di::di (e.g., where higher values indicate better health).
#' @param rescale Logical; if TRUE, non-binary columns will be rescaled to \eqn{[0,1]}
#'   by di::di. Default TRUE.
#' @param age Optional name of the column holding age (used by di for plotting and
#'   optional age-binned outputs; excluded from auto-selected cols).
#' @param rescale.custom Advanced argument passed through to di::di.
#'   See di::di documentation for syntax.
#' @param rescale.avoid Advanced argument passed through to di::di;
#'   see di::di documentation for syntax.
#' @param bins Integer; number of age bins for FI-by-age plots. Default 7.
#' @param visible Logical; if TRUE and age is provided, di will draw a plot (via
#'   plot.di()). Default FALSE.
#' @param na_action One of `c("keep","omit","error","warn","ignore")`. Controls
#'   handling of missing values in selected deficit columns. `"keep"` (and its
#'   backward-compatible alias `"ignore"`) passes NAs through to di::di.
#'   `"warn"` emits a warning and then keeps NAs (alias for `"keep"` with a
#'   missingness warning). `"omit"` drops rows with any NA in selected deficits
#'   before computing. `"error"` stops if any NA is detected.
#'   Default `"ignore"` (retained for backward compatibility; equivalent to `"keep"`).
#' @param na_warn_prop Proportion in \eqn{[0,1]} above which a high-missingness warning
#'   is emitted (per column) when na_action = "warn". Default 0.2.
#' @param return One of c("list","data"). "list" (default) returns the original
#'   di::di result (backward compatible). "data" returns a tibble with one row
#'   per individual, columns: di (the frailty index) plus the selected deficit
#'   columns (post-capping if applied). Age is included if present.
#' @param verbose Logical; if TRUE, prints progress and a completion summary.
#'
#' @return
#' - If return = "list" (default): the object returned by di::di (typically a list
#'   with di and columns).
#' - If return = "data": a tibble with di and the selected columns.
#'
#' @examples
#' # Minimal example (runs only if the 'di' package is installed)
#' if (requireNamespace("di", quietly = TRUE)) {
#'   df <- data.frame(
#'     age = c(70, 75, 80),
#'     d1 = c(0, 1, 1),
#'     d2 = c(0.2, 0.8, 1.0),
#'     d3 = c(TRUE, FALSE, TRUE)
#'   )
#'   # Auto-select numeric deficits; returns list (di, columns)
#'   res <- frailty_index(df, cols = NULL, age = "age", verbose = TRUE)
#'   # Tidy tibble return
#'   tb  <- frailty_index(df, cols = c("d1","d2","d3"), age = "age",
#'                        return = "data", verbose = TRUE)
#' }
#'
#' @references
#' \insertRef{mitnitski2001deficit}{HealthMarkers}
#' \insertRef{rockwood2007frailty}{HealthMarkers}
#' \insertRef{searle2008frailtyindex}{HealthMarkers}
#'
#' @export
frailty_index <- function(data,
                          cols = NULL,
                          invert = NULL,
                          rescale = TRUE,
                          age = NULL,
                          rescale.custom = NULL,
                          rescale.avoid = NULL,
                          bins = 7,
                          visible = FALSE,
                          na_action = c("ignore","warn","error","keep","omit"),
                          na_warn_prop = 0.2,
                          return = c("list","data"),
                          verbose = TRUE) {
  na_action <- match.arg(na_action)
  return <- match.arg(return)

  .need_pkg_di()
  .assert_df(data, "data")
  .assert_scalar_logical(rescale, "rescale")
  .assert_scalar_logical(visible, "visible")
  .assert_scalar_logical(verbose, "verbose")
  .assert_scalar_number(bins, "bins", integer_ok = TRUE, must_be_positive = TRUE)
  .assert_prop(na_warn_prop, "na_warn_prop")

  # Coerce to base data.frame to satisfy di::di checks
  df <- as.data.frame(data, stringsAsFactors = FALSE)

  # Validate columns
  if (!is.null(age)) .assert_cols_exist(df, age, what = "age")
  if (is.null(cols)) {
    # auto-select numeric columns; exclude age if present
    cols <- names(df)[vapply(df, is.numeric, logical(1))]
    if (!is.null(age)) cols <- setdiff(cols, age)
  } else {
    .assert_character(cols, "cols")
    .assert_cols_exist(df, cols, what = "cols")
    if (!is.null(age)) cols <- setdiff(unique(cols), age)
  }
  if (length(cols) == 0L) stop("No deficit columns selected. Provide cols or supply numeric deficit columns in `data`.")

  if (!is.null(invert)) {
    .assert_character(invert, "invert")
    # Only keep invert entries that are in cols (di will also check)
    invert <- intersect(invert, cols)
  }

  if (!is.null(rescale.custom)) .assert_character(rescale.custom, "rescale.custom")
  if (!is.null(rescale.avoid))  .assert_character(rescale.avoid,  "rescale.avoid")

  # Backward-compatible na_action aliases ('ignore'/'warn' behave like 'keep')
  if (na_action %in% c("keep", "omit", "ignore", "warn")) {
    hm_inform(
      sprintf("frailty_index(): na_action=%s", na_action),
      level = "debug"
    )
  }

  # Select deficits for scanning
  sel_df <- df[, cols, drop = FALSE]

  # HM-CS: na_action = 'omit' drops rows with any NA in selected deficits
  if (na_action == "omit") {
    keep_rows <- stats::complete.cases(sel_df)
    df <- df[keep_rows, , drop = FALSE]
    sel_df <- sel_df[keep_rows, , drop = FALSE]
  }

  # Run QA scans
  qa_all <- .fi_scan(sel_df, na_warn_prop = na_warn_prop, consider_range = TRUE)
  # Legacy behavior: only flag range when not rescaling
  qa_legacy <- .fi_scan(sel_df, na_warn_prop = na_warn_prop, consider_range = !isTRUE(rescale))

  # Start messages
  if (verbose) {
    hm_inform(sprintf("frailty_index(): preparing inputs (%d rows, %d deficits%s)",
                      nrow(df), length(cols),
                      if (!is.null(age)) ", age provided" else ""),
              level = "inform")
    hm_inform(sprintf("frailty_index(): col_map: %s%s",
                      paste(head(cols, 8), collapse = ", "),
                      if (length(cols) > 8) sprintf(" ... and %d more", length(cols) - 8) else ""),
              level = "inform")
    hm_inform(sprintf("frailty_index(): NA scan: %d col(s) with any NA; %d col(s) high-NA (>= %.0f%%)",
                      length(qa_all$any_na_cols), length(qa_all$high_na_cols), 100 * na_warn_prop),
              level = "debug")
    if (!isTRUE(rescale)) {
      hm_inform(sprintf("frailty_index(): out-of-range scan (<0 or >1): %d col(s) flagged",
                        length(qa_all$out_of_range_cols)),
                level = "debug")
    }
  } else {
    hm_inform("frailty_index(): computing", level = "debug")
  }

  # NA handling (legacy + HM-CS keep/omit)
  if (na_action == "warn" && (length(qa_all$any_na_cols) > 0L || length(qa_all$high_na_cols) > 0L)) {
    if (length(qa_all$high_na_cols) > 0L) {
      warning(sprintf("High missingness (>= %.0f%%) in: %s.",
                      100 * na_warn_prop, paste(qa_all$high_na_cols, collapse = ", ")), call. = FALSE)
    } else {
      warning(sprintf("Missing values detected in: %s.",
                      paste(qa_all$any_na_cols, collapse = ", ")), call. = FALSE)
    }
  }
  if (na_action %in% c("error") && length(qa_all$any_na_cols) > 0L) {
    stop(sprintf("Missing values present in selected deficits: %s.", paste(qa_all$any_na_cols, collapse = ", ")))
  }

  # Out-of-range scan: check for values outside [0,1] and warn
  oor_cols <- qa_all$out_of_range_cols
  if (length(oor_cols) > 0L && !isTRUE(rescale)) {
    warning(sprintf("Out-of-range values (<0 or >1) found in: %s (not altered).",
                    paste(oor_cols, collapse = ", ")), call. = FALSE)
  }

  # Call di::di() safely
  di_res <- try(
    di::di(
      dat             = df,
      cols            = cols,
      invert          = invert,
      rescale         = rescale,
      age             = age,
      rescale.custom  = rescale.custom,
      rescale.avoid   = rescale.avoid,
      bins            = bins,
      visible         = visible
    ),
    silent = TRUE
  )

  if (inherits(di_res, "try-error")) {
    stop("di::di() failed. Check inputs or update the 'di' package.")
  }

  # Completion summary
  if (return == "data") {
    di_vec <- tryCatch(di_res$di, error = function(e) NULL)
    out_tbl <- {
      out_tmp <- data.frame(di = as.numeric(di_vec), stringsAsFactors = FALSE)
      keep_cols <- unique(c(cols, age))
      if (length(keep_cols)) out_tmp <- cbind(out_tmp, df[, keep_cols, drop = FALSE])
      tibble::as_tibble(out_tmp)
    }
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = hm_result_summary(out_tbl["di"], "frailty_index"))
    return(out_tbl)
  } else {
    if (isTRUE(verbose)) {
      di_vec <- tryCatch(di_res$di, error = function(e) NULL)
      if (is.numeric(di_vec)) {
        rng <- if (any(is.finite(di_vec))) range(di_vec, na.rm = TRUE) else c(NA_real_, NA_real_)
        hm_inform(sprintf("frailty_index(): results: di range [%.3f, %.3f] (%d rows, %d deficits)",
                          ifelse(is.finite(rng[1]), rng[1], NA_real_),
                          ifelse(is.finite(rng[2]), rng[2], NA_real_),
                          length(di_vec), length(cols)),
                  level = "inform")
      } else {
        hm_inform("frailty_index(): completed", level = "inform")
      }
    } else {
      hm_inform("frailty_index(): completed", level = "debug")
    }
    return(di_res)
  }
}

#' Plot FI vs age (convenience wrapper)
#'
#' Calls frailty_index() with visible = TRUE; see frailty_index() for arguments,
#' validation, and references.
#'
#' @inheritParams frailty_index
#' @return The object returned by frailty_index() (di::di object if return="list").
#' @examples
#' \donttest{
#' if (requireNamespace("di", quietly = TRUE)) {
#'   df <- data.frame(age = c(70, 75, 80), d1 = c(0, 1, 1),
#'     d2 = c(0.2, 0.8, 1.0), d3 = c(TRUE, FALSE, TRUE))
#'   plot_frailty_age(df, cols = c("d1", "d2", "d3"), age = "age")
#' }
#' }
#' @export
plot_frailty_age <- function(data,
                             cols = NULL,
                             invert = NULL,
                             rescale = TRUE,
                             age = NULL,
                             rescale.custom = NULL,
                             rescale.avoid = NULL,
                             bins = 7,
                             na_action = c("ignore","warn","error","keep","omit"),
                             na_warn_prop = 0.2,
                             return = c("list","data"),
                             verbose = TRUE) {
  na_action <- match.arg(na_action)
  return <- match.arg(return)
  frailty_index(
    data            = data,
    cols            = cols,
    invert          = invert,
    rescale         = rescale,
    age             = age,
    rescale.custom  = rescale.custom,
    rescale.avoid   = rescale.avoid,
    bins            = bins,
    visible         = TRUE,
    na_action       = na_action,
    na_warn_prop    = na_warn_prop,
    return          = return,
    verbose         = verbose
  )
}

# ---- internal helpers (not exported) ----------------------------------------

.need_pkg_di <- function() {
  ok <- suppressMessages(suppressWarnings(requireNamespace("di", quietly = TRUE)))
  if (!ok) {
    rlang::abort(
      "Package 'di' is required for this feature. Install it first.",
      class = "healthmarkers_missing_package",
      package = "di"
    )
  }
  invisible(TRUE)
}

.assert_df <- function(x, nm) {
  if (!is.data.frame(x)) stop(sprintf("`%s` must be a data.frame or tibble.", nm))
}

.assert_scalar_logical <- function(x, nm) {
  if (!(is.logical(x) && length(x) == 1L && !is.na(x))) {
    stop(sprintf("`%s` must be a single logical value.", nm))
  }
}

.assert_scalar_number <- function(x, nm, integer_ok = FALSE, must_be_positive = FALSE) {
  ok <- is.numeric(x) && length(x) == 1L && is.finite(x)
  if (ok && integer_ok) ok <- (abs(x - round(x)) < .Machine$double.eps^0.5)
  if (ok && must_be_positive) ok <- x > 0
  if (!ok) {
    stop(sprintf("`%s` must be a %s%s%s.",
                 nm,
                 if (integer_ok) "single integer" else "single numeric",
                 if (must_be_positive) " > 0" else "",
                 " and finite"))
  }
}

.assert_character <- function(x, nm) {
  if (!is.character(x)) stop(sprintf("`%s` must be a character vector.", nm))
}

.assert_cols_exist <- function(df, cols, what = "columns") {
  miss <- setdiff(cols, names(df))
  if (length(miss)) stop(sprintf("Missing %s in `data`: %s", what, paste(miss, collapse = ", ")))
}

.assert_prop <- function(x, nm) {
  if (!(is.numeric(x) && length(x) == 1L && is.finite(x) && x >= 0 && x <= 1)) {
    stop(sprintf("`%s` must be a single numeric in [0, 1].", nm))
  }
}

.fi_scan <- function(df, na_warn_prop = 0.2, consider_range = TRUE) {
  any_na_cols <- character(0)
  high_na_cols <- character(0)
  oor_cols <- character(0)
  for (v in names(df)) {
    x <- df[[v]]
    if (anyNA(x)) any_na_cols <- c(any_na_cols, v)
    n <- length(x)
    if (n > 0L) {
      pna <- sum(is.na(x)) / n
      if (pna >= na_warn_prop && pna > 0) high_na_cols <- c(high_na_cols, v)
    }
    if (isTRUE(consider_range)) {
      # Consider only finite values
      xf <- x[is.finite(x)]
      if (length(xf) && (min(xf) < 0 || max(xf) > 1)) {
        # Flag columns that look binary or logical
        if (is.logical(x) || setequal(sort(unique(na.omit(x))), c(0,1))) {
          oor_cols <- c(oor_cols, v)
        }
      }
    }
  }
  list(any_na_cols = unique(any_na_cols),
       high_na_cols = unique(high_na_cols),
       out_of_range_cols = unique(oor_cols))
}

.cap_01 <- function(df) {
  for (v in names(df)) {
    x <- df[[v]]
    if (is.numeric(x)) {
      x[x < 0 & is.finite(x)] <- 0
      x[x > 1 & is.finite(x)] <- 1
      df[[v]] <- x
    }
  }
  df
}

# tests/testthat/test_frailty_index.R

# Helper to unload di if already loaded
unload_di_if_loaded <- function() {
  if ("di" %in% loadedNamespaces()) {
    unloadNamespace("di")
  }
}

# 1) frailty_index errors if 'di' not installed
test_that("frailty_index errors without di installed", {
  unload_di_if_loaded()
  empty_lib <- tempfile("libs")
  dir.create(empty_lib)
  withr::with_libpaths(new = empty_lib, action = "replace", {
    expect_error(
      frailty_index(tibble::tibble(a = 1, b = 0)),
      "Package 'di' is required"
    )
  })
})

# 2) frailty_index returns the expected list structure
test_that("frailty_index returns list(di, columns)", {
  skip_on_cran()
  skip_if_not_installed("di")
  df <- tibble::tibble(var1 = c(1, 0, 1), var2 = c(0, 1, 1), var3 = c(1, 1, 0))
  res <- frailty_index(df, cols = c("var1", "var2", "var3"))
  expect_type(res, "list")
  expect_true(all(c("di", "columns") %in% names(res)))
  expect_length(res$di, nrow(df))
  expect_true(all(res$di >= 0 & res$di <= 1, na.rm = TRUE))
  expect_true(is.matrix(res$columns) || is.data.frame(res$columns))
})

# 3) auto-select numeric deficits when cols = NULL (exclude age)
test_that("frailty_index auto-selects numeric cols when cols=NULL", {
  skip_on_cran()
  skip_if_not_installed("di")
  df <- tibble::tibble(age = c(30, 40), d1 = c(1, 0), d2 = c(0, 1), d3 = c(1, 1))
  res <- frailty_index(df, age = "age")
  expect_false("age" %in% colnames(res$columns))
  expect_setequal(colnames(res$columns), c("d1", "d2", "d3"))
})

# 4) plot_frailty_age errors if 'di' not installed
test_that("plot_frailty_age errors without di installed", {
  skip_on_cran()
  unload_di_if_loaded()
  empty_lib <- tempfile("libs")
  dir.create(empty_lib)
  withr::with_libpaths(new = empty_lib, action = "replace", {
    expect_error(
      plot_frailty_age(tibble::tibble(a = 1, b = 0), age = NULL),
      "Package 'di' is required"
    )
  })
})

# 5) plot_frailty_age never errors when di is installed
test_that("plot_frailty_age never errors when di is installed", {
  skip_on_cran()
  skip_if_not_installed("di")
  df <- tibble::tibble(
    age = rep(50, 3),
    d1  = c(1, 0, 1),
    d2  = c(0, 1, 0)
  )
  expect_error(
    plot_frailty_age(df, cols = c("d1", "d2"), age = "age", bins = 5),
    NA
  )
})

# 6) return = 'data' produces tidy tibble with di and selected columns
test_that("frailty_index return='data' returns tibble with di and inputs", {
  skip_on_cran()
  skip_if_not_installed("di")
  df <- tibble::tibble(age = c(70, 72), d1 = c(1, 0), d2 = c(0, 1))
  out <- frailty_index(df, cols = c("d1", "d2"), age = "age", return = "data")
  expect_s3_class(out, "tbl_df")
  expect_true(all(c("di", "d1", "d2", "age") %in% names(out)))
  expect_equal(nrow(out), nrow(df))
  expect_true(all(out$di >= 0 & out$di <= 1, na.rm = TRUE))
})

# 7) NA handling options: warn/error/ignore
test_that("frailty_index NA handling works", {
  skip_on_cran()
  skip_if_not_installed("di")
  df <- tibble::tibble(d1 = c(1, NA, 0), d2 = c(0, 1, 1))
  # warn on high missingness
  expect_warning(
    frailty_index(df, cols = c("d1", "d2"), na_action = "warn", na_warn_prop = 0.2),
    "Missing values|High missingness"
  )
  # error on any NA
  expect_error(
    frailty_index(df, cols = c("d1", "d2"), na_action = "error"),
    "Missing values present"
  )
  # ignore runs
  expect_error(
    frailty_index(df, cols = c("d1", "d2"), na_action = "ignore"),
    NA
  )
})

# 8) verbose messages include preparing, col_map, and results
test_that("frailty_index verbose = TRUE emits preparing and column mapping messages", {
  skip_on_cran()
  skip_if_not_installed("di")
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble::tibble(d1 = c(1, 0), d2 = c(0, 1))
  expect_message(frailty_index(df, cols = c("d1", "d2"), verbose = TRUE), "frailty_index")
  expect_message(frailty_index(df, cols = c("d1", "d2"), verbose = TRUE), "col_map")
})

test_that("frailty_index verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  skip_if_not_installed("di")
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble::tibble(d1 = c(1, 0), d2 = c(0, 1))
  msgs <- testthat::capture_messages(
    frailty_index(df, cols = c("d1", "d2"), verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
})

test_that("frailty_index return='data' verbose emits results summary", {
  skip_on_cran()
  skip_if_not_installed("di")
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble::tibble(d1 = c(1, 0), d2 = c(0, 1))
  msgs <- testthat::capture_messages(
    frailty_index(df, cols = c("d1", "d2"), return = "data", verbose = TRUE)
  )
  expect_equal(sum(grepl("results:", msgs)), 1L)
})

# File: allostatic_load.R
#'
#' Allostatic Load Index
#'
#' Computes a composite Allostatic Load (AL) score by flagging biomarkers that
#' exceed user-specified high-risk thresholds (strict > when multiple biomarkers;
#' inclusive >= when only one biomarker). Aligned with HM-CS v3: structured
#' validation, diagnostic control, verbose reporting, and optional summary output.
#'
#' @param data data.frame or tibble of numeric biomarker columns.
#' @param thresholds named list of scalar numeric cutoffs (names must match columns).
#' @param col_map optional named list mapping keys in `thresholds` to column names in `data`.
#' @param na_action one of c("keep","omit","error") ("keep" treats NA as zero contribution).
#' @param return_summary logical; TRUE returns list(data, summary, warnings).
#' @param verbose logical; print progress messages via hm_inform() (also gated by options(healthmarkers.verbose)).
#'
#' @return tibble with AllostaticLoad or list when return_summary = TRUE.
#' @details
#' **API pattern note:** Unlike most HealthMarkers functions which follow the
#' standard `(data, col_map, ...)` signature, `allostatic_load()` uses
#' `(data, thresholds, col_map = NULL, ...)` because `thresholds` is a
#' domain-specific required argument that cannot be inferred from column names
#' alone. As a result, this function is not directly dispatchable through the
#' standard `all_health_markers()` registry; a custom wrapper that supplies
#' `thresholds` explicitly would be required.
#' @seealso \code{\link{adiposity_sds}}, \code{\link{adiposity_sds_strat}}
#'
#' @references \insertRef{seeman1997price}{HealthMarkers}
#'
#' @examples
#' df <- tibble::tibble(
#'   SBP = c(118, 142, 130),
#'   DBP = c(76, 92, 85),
#'   CRP = c(1.2, 4.8, 2.1)
#' )
#' thr <- list(SBP = 130, DBP = 85, CRP = 3)
#' allostatic_load(df, thresholds = thr, na_action = "keep", verbose = FALSE)
#'
#' # Single biomarker uses inclusive >= rule
#' allostatic_load(df, thresholds = list(CRP = 3))
#'
#' @export
allostatic_load <- function(
  data,
  thresholds,
  col_map = NULL,
  na_action = c("keep","omit","error"),
  return_summary = FALSE,
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "allostatic_load"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  na_action <- match.arg(na_action)

  # ---- Validate inputs ----
  if (!is.data.frame(data)) {
    rlang::abort("allostatic_load(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_allo_error_data_type")
  }
  if (length(thresholds) == 0L) {
    rlang::abort("allostatic_load(): `thresholds` must contain at least one element.",
                 class = "healthmarkers_allo_error_thr_empty")
  }
  if (!is.list(thresholds) || is.null(names(thresholds)) || any(!nzchar(names(thresholds)))) {
    rlang::abort("allostatic_load(): `thresholds` must be a named list of scalar numeric cutoffs.",
                 class = "healthmarkers_allo_error_thr_type")
  }
  if (any(duplicated(names(thresholds)))) {
    rlang::abort("allostatic_load(): duplicate names in `thresholds`.",
                 class = "healthmarkers_allo_error_thr_dupe")
  }
  if (!all(vapply(thresholds, function(x) is.numeric(x) && length(x) == 1L && is.finite(x), logical(1)))) {
    rlang::abort("allostatic_load(): each threshold must be a length-1 finite numeric.",
                 class = "healthmarkers_allo_error_thr_values")
  }

  vars <- names(thresholds)

  # Build effective var -> column mapping with inference support
  cm      <- .hm_build_col_map(data, col_map, keys = vars, fn = fn_name)
  data    <- cm$data
  var_map <- cm$col_map

  # Confirm required columns exist in data
  .hm_log_cols(cm, var_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  AllostaticLoad [%s]", fn_name, paste(vars, collapse = ", ")), level = "inform")
  missing_vars <- setdiff(vars, names(var_map))
  if (length(missing_vars)) {
    rlang::abort(
      paste0("allostatic_load(): missing required columns in data: ", paste(missing_vars, collapse = ", ")),
      class = "healthmarkers_allo_error_missing_columns"
    )
  }
  req_cols <- unname(unlist(var_map[vars], use.names = FALSE))

  # Coerce required columns to numeric; non-finite -> NA
  for (cn in req_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("allostatic_load(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced),
                    class = "healthmarkers_allo_warn_na_coercion")
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # ---- NA row policy (row-wise over required columns) ----
  rows_with_na <- if (length(req_cols)) {
    Reduce(`|`, lapply(req_cols, function(cn) is.na(data[[cn]])))
  } else rep(FALSE, nrow(data))

  if (na_action == "error" && any(rows_with_na)) {
    rlang::abort("allostatic_load(): missing/non-finite values present (na_action='error').",
                 class = "healthmarkers_allo_error_missing_values")
  } else if (na_action == "omit" && any(rows_with_na)) {
    data <- data[!rows_with_na, , drop = FALSE]
  }

  # Early return if no rows remain
  if (nrow(data) == 0L) {
    out0 <- tibble::tibble(AllostaticLoad = integer(0))
    return(if (return_summary) list(
      data = out0,
      summary = list(rows = 0L, biomarkers = length(vars), total_flags = 0L, mean_flags = NA_real_),
      warnings = character()
    ) else out0)
  }

  # ---- Compute flags and sum ----
  inclusive <- length(vars) == 1L
  hm_inform(level = "debug", msg = sprintf("allostatic_load(): computing (rule %s)", if (inclusive) ">=" else ">"))

  flag_cols <- lapply(vars, function(v) {
    cn <- var_map[[v]]
    x <- data[[cn]]
    x[!is.finite(x)] <- NA_real_
    th <- thresholds[[v]]
    as.integer(if (inclusive) x >= th else x > th)
  })

  flag_mat <- if (length(flag_cols)) {
    fm <- do.call(cbind, flag_cols)
    colnames(fm) <- vars
    fm
  } else {
    matrix(integer(0), nrow = nrow(data), ncol = 0)
  }

  out <- tibble::tibble(AllostaticLoad = as.integer(rowSums(flag_mat, na.rm = TRUE)))

  if (!is.null(id_col)) {
    id_vec <- data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  if (isTRUE(return_summary)) {
    return(list(
      data = out,
      summary = list(
        rows = nrow(data),
        biomarkers = length(vars),
        total_flags = sum(out$AllostaticLoad),
        mean_flags = mean(out$AllostaticLoad)
      ),
      warnings = character()
    ))
  }

  out
}

# File: tests/testthat/test_allostatic_load.R

library(testthat)
library(tibble)

test_that("basic multi-biomarker scoring uses strict >", {
  skip_on_cran()
  df <- tibble(
    A = c(1, 5, 10, NA),
    B = c(0, 2, 3, 4),
    C = c(100, 50, 25, 10)
  )
  thr <- list(A = 4, B = 2, C = 30)
  out <- allostatic_load(df, thresholds = thr, na_action = "keep")
  expect_s3_class(out, "tbl_df")
  expect_named(out, "AllostaticLoad")
  expect_equal(out$AllostaticLoad, c(1L, 2L, 2L, 1L))
})

test_that("single biomarker uses inclusive >=", {
  skip_on_cran()
  df <- tibble(X = c(5, 10, 11))
  thr <- list(X = 10)
  out <- allostatic_load(df, thresholds = thr, na_action = "keep")
  expect_equal(out$AllostaticLoad, c(0L, 1L, 1L))
})

test_that("input validation errors (HM-CS messages)", {
  skip_on_cran()
  df <- tibble(A = 1:3)
  expect_error(allostatic_load("not df", thresholds = list(A = 1)), "data.frame or tibble")
  expect_error(allostatic_load(df, thresholds = list()), "must contain at least one")
  bad_thr1 <- list(3); names(bad_thr1) <- ""
  expect_error(allostatic_load(df, thresholds = bad_thr1), "must be a named list")
  bad_thr2 <- list(1, 2); names(bad_thr2) <- c("A", "A")
  expect_error(allostatic_load(df, thresholds = bad_thr2), "duplicate names")
  expect_error(allostatic_load(df, thresholds = list(A = c(1, 2))), "length-1")
  expect_error(allostatic_load(df, thresholds = list(B = 2)), "missing required columns in data")
})

test_that("biomarker must be numeric", {
  skip_on_cran()
  df <- tibble(A = letters[1:3])
  thr <- list(A = 1)
  # Suppress the deliberate coercion warning in the test
  df$A <- suppressWarnings(as.numeric(df$A))
  out <- allostatic_load(df, thresholds = thr, na_action = "keep")
  expect_equal(nrow(out), 3)
})

test_that("na_action = omit drops any row with missing in any required var", {
  skip_on_cran()
  df <- tibble(A = c(1, NA, 3), B = c(2, 2, NA))
  thr <- list(A = 1.5, B = 1.5)
  out <- allostatic_load(df, thresholds = thr, na_action = "omit")
  expect_equal(nrow(out), 1L)
})

test_that("na_action = error stops on NA", {
  skip_on_cran()
  df <- tibble(A = c(1, NA, 3))
  expect_error(
    allostatic_load(df, thresholds = list(A = 2), na_action = "error"),
    "missing/non-finite values present"
  )
})

test_that("check_extreme removed: function passes through SDS-like outlier rows", {
  skip_on_cran()
  df <- tibble(BP_sds = c(-0.5, 7, 6.5), HR_sds = c(0, -7, 5))
  thr <- list(BP_sds = 1.5, HR_sds = 1.5)
  out <- allostatic_load(df, thresholds = thr)
  expect_equal(nrow(out), 3L)
})

test_that("zero-row data returns zero-row tibble", {
  skip_on_cran()
  df <- tibble(A = numeric(0))
  out <- allostatic_load(df, thresholds = list(A = 2))
  expect_equal(nrow(out), 0)
  expect_named(out, "AllostaticLoad")
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df  <- tibble(A = c(1, 5), B = c(0, 3))
  thr <- list(A = 2, B = 1)
  expect_message(allostatic_load(df, thresholds = thr, verbose = TRUE), "allostatic_load")
  expect_message(allostatic_load(df, thresholds = thr, verbose = TRUE), "col_map")
  expect_message(allostatic_load(df, thresholds = thr, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df  <- tibble(A = c(1, 5), B = c(0, 3))
  thr <- list(A = 2, B = 1)
  msgs <- testthat::capture_messages(
    allostatic_load(df, thresholds = thr, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("return_summary returns structured list", {
  skip_on_cran()
  df <- tibble(A = c(1, 5, 6))
  thr <- list(A = 2)
  res <- allostatic_load(df, thresholds = thr, return_summary = TRUE)
  expect_named(res, c("data", "summary", "warnings"))
  expect_true(is.data.frame(res$data))
  expect_true(is.list(res$summary))
  expect_true(all(c("rows","biomarkers","total_flags","mean_flags") %in% names(res$summary)))
  expect_equal(res$summary$total_flags, sum(res$data$AllostaticLoad))
  # A = c(1,5,6), thr = 2, single biomarker -> >= rule; flags: 0,1,1 -> total = 2
  expect_equal(res$summary$total_flags, 2L)
  expect_equal(res$summary$biomarkers, 1L)
  expect_equal(res$summary$rows, 3L)
})

#' Diagnose and report column name mapping for your data
#'
#' @description
#' Scans the column names of your data frame against the internal synonym
#' dictionary used by all \pkg{HealthMarkers} functions and prints a
#' formatted report showing which internal keys were matched automatically and
#' which were not found.  The function uses five matching layers in order:
#'
#' 1. **User-supplied** (via `col_map`) — always wins.
#' 2. **Exact synonym match** — column name is in the synonym list.
#' 3. **Case-insensitive exact** — same, ignoring upper/lower case.
#' 4. **Column contains synonym** — data column name contains a synonym as a
#'    whole word (e.g.\ `"trig_baseline"` matches `"trig"`).
#' 5. **Synonym contains column** — a synonym contains the column name as a
#'    whole word (e.g.\ `"TG_fasting"` synonym matches column `"TG"`).
#' 6. **Fuzzy** (opt-in via `fuzzy = TRUE`) — Levenshtein-based approximate
#'    matching as a last resort.
#'
#' The function returns the matched mappings invisibly as a named list that
#' can be passed directly as the `col_map` argument to any
#' \pkg{HealthMarkers} function.
#'
#' @param data A \code{data.frame} or \code{tibble}.
#' @param col_map Optional named list of manually specified mappings
#'   (\code{list(TG = "trig_my_col")}).  These always take priority.
#' @param verbose Logical (default \code{TRUE}).  Print the formatted report.
#' @param fuzzy Logical (default \code{FALSE}).  Enable fuzzy / approximate
#'   matching as a final layer.  May produce false positives; review results
#'   with care.
#' @param show_unmatched Logical (default \code{FALSE}).  If \code{TRUE},
#'   also list every key that could not be matched.  Useful when you expect a
#'   variable to be found but it isn't.
#'
#' @return Invisibly returns a named list: internal key \eqn{\to} matched
#'   data-column name.  Keys that were not found are omitted.  You can assign
#'   this to \code{col_map} and pass it to any \pkg{HealthMarkers} function.
#'
#' @examples
#' \dontrun{
#' # Basic diagnostic
#' hm_col_report(my_data)
#'
#' # Fuzzy matching + show all unmatched
#' hm_col_report(my_data, fuzzy = TRUE, show_unmatched = TRUE)
#'
#' # Capture the result as a ready-to-use col_map
#' cm <- hm_col_report(my_data, verbose = FALSE)
#' all_health_markers(my_data, col_map = cm)
#' }
#'
#' @export
hm_col_report <- function(data,
                          col_map       = NULL,
                          verbose       = TRUE,
                          fuzzy         = FALSE,
                          show_unmatched = FALSE) {

  if (!is.data.frame(data))
    stop("`data` must be a data.frame or tibble.")

  patterns  <- .hm_default_col_patterns_exact()
  all_keys  <- names(patterns)

  result       <- vector("list", length(all_keys))
  names(result) <- all_keys
  match_detail  <- setNames(character(length(all_keys)), all_keys)

  data_cols <- names(data)
  dc_lower  <- tolower(data_cols)

  for (key in all_keys) {

    ## Layer 0: user-supplied -------------------------------------------------
    if (!is.null(col_map) && !is.null(col_map[[key]])) {
      result[[key]]      <- col_map[[key]]
      match_detail[[key]] <- "user-supplied"
      next
    }

    syns       <- unique(na.omit(as.character(patterns[[key]])))
    syns_lower <- tolower(syns)

    ## Layer 1: exact (case-sensitive) ----------------------------------------
    hit <- syns[syns %in% data_cols]
    if (length(hit)) {
      result[[key]]      <- hit[1L]
      match_detail[[key]] <- "exact"
      next
    }

    ## Layer 2: case-insensitive exact ----------------------------------------
    idx <- match(syns_lower, dc_lower)
    idx <- idx[!is.na(idx)]
    if (length(idx)) {
      result[[key]]      <- data_cols[idx[1L]]
      match_detail[[key]] <- "case-insensitive"
      next
    }

    ## Layer 3: data column *contains* a synonym (whole-word, lower) ----------
    ## Only use synonyms that are >=4 chars to avoid short-token false positives
    ## (e.g. "sat" matching "transferrin_sat", "hdl" matching "non_HDL_c").
    long_syns_lower <- syns_lower[nchar(syns_lower) >= 4L]
    for (col in data_cols) {
      col_l <- tolower(col)
      found <- length(long_syns_lower) > 0L &&
        any(vapply(long_syns_lower, function(s) {
          grepl(paste0("(^|[^a-z0-9])", s, "([^a-z0-9]|$)"), col_l, perl = TRUE)
        }, logical(1L)))
      if (found) {
        result[[key]]      <- col
        match_detail[[key]] <- "col contains synonym"
        break
      }
    }
    if (!is.null(result[[key]])) next

    ## Layer 4: a synonym *contains* the column name (whole-word, lower) ------
    ## Only applied when column token is >=8 chars to prevent short tokens like
    ## "DHEAS" (5), "HDL_c" (5), "L_FABP" (6), "height" (6), "pulse" (5) from
    ## matching compound-ratio synonyms (Cort_DHEAS_den, non_HDL_c, etc.).
    for (col in data_cols) {
      col_l <- tolower(col)
      if (nchar(col_l) < 8L) next
      found <- any(vapply(syns_lower, function(s) {
        grepl(paste0("(^|[^a-z0-9])", col_l, "([^a-z0-9]|$)"), s, perl = TRUE)
      }, logical(1L)))
      if (found) {
        result[[key]]      <- col
        match_detail[[key]] <- "synonym contains col"
        break
      }
    }
    if (!is.null(result[[key]])) next

    ## Layer 5: fuzzy (opt-in) ------------------------------------------------
    if (isTRUE(fuzzy)) {
      for (syn in syns) {
        hit <- agrep(syn, data_cols, ignore.case = TRUE,
                     value = TRUE, max.distance = 0.2)
        if (length(hit)) {
          result[[key]]      <- hit[1L]
          match_detail[[key]] <- "fuzzy"
          break
        }
      }
    }

    ## Not found --------------------------------------------------------------
    if (is.null(result[[key]])) {
      result[[key]]      <- NA_character_
      match_detail[[key]] <- "not found"
    }
  }

  ## ── determine matched / unmatched status ----------------------------------
  is_matched  <- vapply(result, function(x) !is.null(x) && !is.na(x), logical(1L))
  n_matched   <- sum(is_matched)
  n_unmatched <- sum(!is_matched)

  ## ── verbose output --------------------------------------------------------
  if (isTRUE(verbose)) {

    rule <- paste(rep("\u2500", 58L), collapse = "")
    cat(sprintf("\u2500\u2500 HealthMarkers column report %s\n", rule))
    cat(sprintf(" Data: %d row%s \u00d7 %d column%s   |   Keys in dictionary: %d\n",
                nrow(data), if (nrow(data) != 1L) "s" else "",
                ncol(data), if (ncol(data) != 1L) "s" else "",
                length(all_keys)))
    cat("\n")

    if (n_matched > 0L) {
      cat(sprintf(" %-26s %-28s %s\n", "key", "data_column", "how matched"))
      cat(sprintf(" %-26s %-28s %s\n",
                  strrep("-", 26L), strrep("-", 28L), strrep("-", 18L)))

      for (key in all_keys[is_matched]) {
        cat(sprintf(" %-26s %-28s %s  \u2714\n",
                    key, result[[key]], match_detail[[key]]))
      }
    }

    cat(sprintf("\n \u2714 %d key%s matched   \u2718 %d key%s not found\n",
                n_matched,   if (n_matched   != 1L) "s" else "",
                n_unmatched, if (n_unmatched != 1L) "s" else ""))

    if (n_unmatched > 0L) {
      if (isTRUE(show_unmatched)) {
        cat(sprintf("\n Unmatched keys:\n"))
        for (key in all_keys[!is_matched])
          cat(sprintf("   \u2022 %s\n", key))
      } else {
        cat(sprintf(" (show_unmatched = TRUE to list all %d unmatched keys)\n",
                    n_unmatched))
      }

      cat(sprintf("\n\u2500\u2500 col_map template for manual overrides %s\n",
                  paste(rep("\u2500", 37L), collapse = "")))
      cat(" # Pass this to any HealthMarkers function via col_map = list(...)\n")
      cat(" # Fill in the 'from_your_data' values for the variables you need.\n")
      cat(" col_map <- list(\n")
      for (key in all_keys[!is_matched])
        cat(sprintf("   %-26s = \"from_your_data\",\n", key))
      cat(" )\n")
    }

    cat(sprintf("%s\n", strrep("\u2500", 87L)))
  }

  ## ── return matched entries as a ready-to-use col_map ---------------------
  matched_map <- result[is_matched]
  invisible(matched_map)
}

#' Impute missing values in a data.frame or tibble (simple, column-wise)
#'
#' Performs deterministic, per-column imputation for numeric variables:
#' - "mean": replace NAs with the column mean
#' - "median": replace NAs with the column median
#' - "zero": replace NAs with 0
#' - "constant": replace NAs with the single value given in `constant`
#'
#' Non-numeric columns are left untouched. If `cols = NULL`, all numeric columns
#' that have at least one NA are selected automatically. NA positions are the only
#' values modified; non-NA entries are preserved as-is.
#'
#' Quality checks:
#' - Warns for high-missingness columns (>= `na_warn_prop`).
#' - Warns and skips imputation when a column has no non-NA values (mean/median undefined).
#' - Coerces only numeric columns; non-numerics in `cols` are skipped with a warning.
#'
#' @param data A data.frame or tibble containing missing values.
#' @param method Character; one of c("mean","median","zero","constant").
#' @param cols Optional character vector of column names to impute. Defaults to all
#'   numeric columns in `data` that contain at least one NA.
#' @param constant Numeric; single value to use when `method = "constant"`.
#' @param na_warn_prop Numeric in \eqn{[0,1]}; threshold for high-missingness warnings per column.
#'   Default 0.2.
#' @param verbose Logical; if TRUE, prints progress and a completion summary. Default FALSE.
#'
#' @return A data.frame/tibble of the same dimensions as `data`, with the specified
#'   columns' missing values imputed.
#' @export
#'
#' @examples
#' df <- tibble::tibble(a = c(1, NA, 3), b = c(NA, NA, 2), c = letters[1:3])
#' impute_missing(df, method = "mean")
#' impute_missing(df, method = "median", verbose = TRUE)
#' impute_missing(df, method = "constant", constant = -1, cols = "a")
impute_missing <- function(data,
                           method   = c("mean", "median", "zero", "constant"),
                           cols     = NULL,
                           constant = NULL,
                           na_warn_prop = 0.2,
                           verbose  = FALSE) {
  # ---- validations ----
  if (!is.data.frame(data)) {
    rlang::abort("impute_missing(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_impute_error_data_type")
  }
  method <- match.arg(method)
  if (!is.numeric(na_warn_prop) || length(na_warn_prop) != 1L ||
      !is.finite(na_warn_prop) || na_warn_prop < 0 || na_warn_prop > 1) {
    rlang::abort("impute_missing(): `na_warn_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_impute_error_na_warn_prop")
  }
  if (!is.null(cols)) {
    missing_cols <- setdiff(cols, names(data))
    if (length(missing_cols)) {
      rlang::abort(
        paste("impute_missing(): Some `cols` not in data:", paste(missing_cols, collapse = ", ")),
        class = "healthmarkers_impute_error_missing_cols"
      )
    }
  }
  if (method == "constant") {
    if (!is.numeric(constant) || length(constant) != 1L || !is.finite(constant)) {
      rlang::abort("impute_missing(): `constant` must be a single finite numeric when method = 'constant'.",
                   class = "healthmarkers_impute_error_constant_type")
    }
  }

  # Determine which columns to impute
  if (is.null(cols)) {
    is_num <- vapply(data, is.numeric, logical(1))
    has_na <- vapply(data, anyNA, logical(1))
    cols   <- names(data)[is_num & has_na]
  }

  if (length(cols) == 0L) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "impute_missing(): nothing to impute (no numeric columns with NA)")
    return(data)
  }

  if (isTRUE(verbose)) {
    hm_inform(sprintf("impute_missing(): preparing inputs (%d rows, %d column(s), method='%s')",
                      nrow(data), length(cols), method), level = "inform")
  } else {
    hm_inform("impute_missing(): preparing inputs", level = "debug")
  }

  # High-missingness scan
  .imp_warn_high_missing(data, cols, na_warn_prop = na_warn_prop)

  out <- data
  imputed_counts <- integer(length(cols))
  names(imputed_counts) <- cols

  for (i in seq_along(cols)) {
    col <- cols[i]
    vec <- out[[col]]

    if (!is.numeric(vec)) {
      rlang::warn(sprintf("impute_missing(): column '%s' is not numeric; skipping.", col),
                  class = "healthmarkers_impute_warn_non_numeric_skip")
      next
    }
    nas <- is.na(vec)
    n_nas <- sum(nas)
    if (n_nas == 0L) next

    replacement <- switch(
      method,
      mean     = mean(vec, na.rm = TRUE),
      median   = stats::median(vec, na.rm = TRUE),
      zero     = 0,
      constant = constant
    )

    if (!is.finite(replacement)) {
      rlang::warn(sprintf("impute_missing(): column '%s' has no non-missing values; cannot compute %s. Skipping.",
                          col, method),
                  class = "healthmarkers_impute_warn_no_non_missing")
      next
    }

    if (isTRUE(verbose)) {
      hm_inform(sprintf("impute_missing(): column '%s' -> replacing %d NA(s) with %s",
                        col, n_nas, format(replacement)), level = "debug")
    }

    vec[nas]   <- replacement
    out[[col]] <- vec
    imputed_counts[i] <- n_nas
  }

  total_imp <- sum(imputed_counts)
  by_col <- paste(names(imputed_counts[imputed_counts > 0]),
                  imputed_counts[imputed_counts > 0], sep = "=")
  hm_inform(sprintf("impute_missing(): results: imputed %d values across %d columns%s",
                    total_imp, sum(imputed_counts > 0),
                    if (length(by_col)) paste0(" [", paste(by_col, collapse = ", "), "]") else ""),
            level = if (isTRUE(verbose)) "inform" else "debug")

  return(out)
}

#' Impute missing values via Multiple Imputation (mice)
#'
#' Wraps mice to impute only numeric columns; non-numeric columns are untouched.
#' Requires the mice package to be installed (Suggests). If no numeric columns
#' contain NAs, `data` is returned unchanged.
#'
#' Notes:
#' - At least two numeric columns are typically needed by mice to borrow strength.
#' - This function runs `m` imputations and returns the first completed dataset.
#' - Messages from mice are suppressed; use `verbose = TRUE` here for high-level progress.
#'
#' @param data A data.frame or tibble containing missing values.
#' @param m Integer; number of imputations to run (passed to mice). Default 5.
#' @param cols Optional character vector of numeric columns to impute.
#'   Defaults to all numeric columns with at least one NA.
#' @param verbose Logical; if TRUE, prints progress and a completion summary. Default FALSE.
#' @param ... Additional arguments passed to mice::mice().
#'
#' @return A data.frame/tibble with numeric columns imputed by mice.
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("mice", quietly = TRUE)) {
#'   df <- tibble::tibble(a = c(1, NA, 3), b = c(2, 4, NA), c = 5)
#'   impute_mice(df, m = 2, verbose = TRUE)
#' }
#' }
#'
#' @references
#' \insertRef{rubin1987mi}{HealthMarkers}
#' \insertRef{vanbuuren2011mice}{HealthMarkers}
impute_mice <- function(data,
                        m    = 5,
                        cols = NULL,
                        verbose = FALSE,
                        ...) {
  if (!is.data.frame(data)) {
    rlang::abort("impute_mice(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_impute_error_data_type")
  }
  if (!(is.numeric(m) && length(m) == 1L && is.finite(m) && m >= 1)) {
    rlang::abort("impute_mice(): `m` must be a single numeric >= 1.",
                 class = "healthmarkers_impute_error_m_param")
  }
  if (!requireNamespace("mice", quietly = TRUE)) {
    rlang::abort("impute_mice(): package 'mice' is required but not installed. Please install 'mice'.",
                 class = "healthmarkers_impute_error_pkg_missing_mice")
  }

  # Determine numeric columns to impute
  is_num <- vapply(data, is.numeric, logical(1))
  if (!is.null(cols)) {
    missing_cols <- setdiff(cols, names(data))
    if (length(missing_cols)) {
      rlang::abort(
        paste("impute_mice(): Some `cols` not in data:", paste(missing_cols, collapse = ", ")),
        class = "healthmarkers_impute_error_missing_cols"
      )
    }
    is_num <- names(data) %in% cols & is_num
  }
  num_cols <- names(data)[is_num]
  if (length(num_cols) == 0L) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "impute_mice(): no numeric columns selected; returning input unchanged")
    return(data)
  }

  # Subset numeric columns with any NA
  has_na <- vapply(data[num_cols], anyNA, logical(1))
  target_cols <- num_cols[has_na]
  if (length(target_cols) == 0L) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "impute_mice(): no NAs in selected numeric columns; returning input unchanged")
    return(data)
  }
  if (length(target_cols) < 2L) {
    rlang::abort("impute_mice(): need at least two numeric columns with missing values for mice().",
                 class = "healthmarkers_impute_error_min_cols_mice")
  }

    if (isTRUE(verbose)) hm_inform(sprintf("impute_mice(): preparing inputs (%d rows, %d columns)", nrow(data), length(target_cols)), level = "inform") else hm_inform("impute_mice(): preparing inputs", level = "debug")

  num_data <- data[, target_cols, drop = FALSE]

  # suppress messages & warnings from mice
  mids <- suppressWarnings(
    suppressMessages(
      mice::mice(num_data, m = m, printFlag = FALSE, ...)
    )
  )
  out_num <- suppressWarnings(
    suppressMessages(
      mice::complete(mids, 1)
    )
  )

  out <- data
  out[, names(out_num)] <- out_num

  total_imp <- sum(vapply(target_cols, function(cn) sum(is.na(data[[cn]])), integer(1)))
  hm_inform(sprintf("impute_mice(): results: imputed %d values across %d columns.", total_imp, length(target_cols)),
            level = if (isTRUE(verbose)) "inform" else "debug")

  return(out)
}

#' Impute missing values via random forest (missForest)
#'
#' Wraps missForest to impute numeric columns using non-parametric random forests.
#' Requires the missForest package to be installed (Suggests). Non-numeric columns
#' are untouched. If no numeric columns contain NAs, `data` is returned unchanged.
#'
#' Notes:
#' - missForest uses iterative RF training; it can be slow on wide/high-NA data.
#' - Errors (e.g., insufficient unique values) are caught and a deterministic mean
#'   imputation fallback is applied to the selected numeric columns.
#'
#' @param data A data.frame or tibble containing missing values.
#' @param ntree Integer; number of trees to grow in each forest (passed to missForest). Default 100.
#' @param cols Optional character vector of numeric columns to impute.
#'   Defaults to all numeric columns with at least one NA.
#' @param verbose Logical; if TRUE, prints progress and a completion summary. Default FALSE.
#' @param ... Additional arguments passed to missForest::missForest().
#'
#' @return A data.frame/tibble with numeric columns imputed by missForest (or mean fallback).
#' @export
#'
#' @examples
#' \donttest{
#' if (requireNamespace("missForest", quietly = TRUE)) {
#'   df <- tibble::tibble(a = c(1, NA, 3), b = c(2, 4, NA), c = 5)
#'   impute_missforest(df, ntree = 50, verbose = TRUE)
#' }
#' }
#'
#' @references \insertRef{stekhoven2012missforest}{HealthMarkers}
impute_missforest <- function(data,
                              ntree = 100,
                              cols  = NULL,
                              verbose = FALSE,
                              ...) {
  if (!is.data.frame(data)) {
    rlang::abort("impute_missforest(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_impute_error_data_type")
  }
  if (!(is.numeric(ntree) && length(ntree) == 1L && is.finite(ntree) && ntree >= 1)) {
    rlang::abort("impute_missforest(): `ntree` must be a single numeric >= 1.",
                 class = "healthmarkers_impute_error_ntree_param")
  }
  if (!requireNamespace("missForest", quietly = TRUE)) {
    rlang::abort("impute_missforest(): package 'missForest' is required but not installed. Please install 'missForest'.",
                 class = "healthmarkers_impute_error_pkg_missing_missforest")
  }

  # Determine numeric columns to impute
  is_num <- vapply(data, is.numeric, logical(1))
  if (!is.null(cols)) {
    missing_cols <- setdiff(cols, names(data))
    if (length(missing_cols)) {
      rlang::abort(
        paste("impute_missforest(): Some `cols` not in data:", paste(missing_cols, collapse = ", ")),
        class = "healthmarkers_impute_error_missing_cols")
    }
    is_num <- names(data) %in% cols & is_num
  }
  num_cols <- names(data)[is_num]
  if (length(num_cols) == 0L) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "impute_missforest(): no numeric columns selected; returning input unchanged")
    return(data)
  }

  # Subset numeric columns with any NA
  has_na <- vapply(data[num_cols], anyNA, logical(1))
  target_cols <- num_cols[has_na]
  if (length(target_cols) == 0L) {
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg = "impute_missforest(): no NAs in selected numeric columns; returning input unchanged")
    return(data)
  }
  if (length(target_cols) < 2L) {
    rlang::abort("impute_missforest(): need at least two numeric columns with missing values for missForest().",
                 class = "healthmarkers_impute_error_min_cols_missforest")
  }

    if (isTRUE(verbose)) hm_inform(sprintf("impute_missforest(): preparing inputs (%d rows, %d columns)", nrow(data), length(target_cols)), level = "inform") else hm_inform("impute_missforest(): preparing inputs", level = "debug")

  num_data <- data[, target_cols, drop = FALSE]

  # suppress warnings (e.g., few unique levels) and catch errors
  res <- tryCatch({
    suppressWarnings(
      missForest::missForest(num_data, ntree = ntree, verbose = FALSE, ...)
    )
  }, error = function(e) {
    rlang::warn("impute_missforest(): missForest failed; falling back to mean imputation.",
                class = "healthmarkers_impute_warn_missforest_failed")
    return(NULL)
  })

  out <- data
  if (is.null(res)) {
    out <- impute_missing(out, method = "mean", cols = target_cols, verbose = FALSE)
    if (isTRUE(verbose)) {
      total_imp <- sum(vapply(target_cols, function(cn) sum(is.na(data[[cn]])), integer(1)))
      hm_inform(sprintf("impute_missforest(): results: fallback mean imputation; imputed %d values across %d columns.",
                        total_imp, length(target_cols)), level = "inform")
    } else {
      total_imp <- sum(vapply(target_cols, function(cn) sum(is.na(data[[cn]])), integer(1)))
      hm_inform(sprintf("impute_missforest(): results: fallback mean imputation; imputed %d values across %d columns.",
                        total_imp, length(target_cols)), level = "debug")
    }
    return(out)
  }

  out[, names(res$ximp)] <- res$ximp

  total_imp <- sum(vapply(target_cols, function(cn) sum(is.na(data[[cn]])), integer(1)))
  hm_inform(sprintf("impute_missforest(): results: imputed %d values across %d columns.", total_imp, length(target_cols)),
            level = if (isTRUE(verbose)) "inform" else "debug")

  return(out)
}

# ---- internal helpers (not exported) ----------------------------------------

.imp_warn_high_missing <- function(df, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- df[[cn]]
    if (!is.numeric(x)) next
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("impute_missing(): column '%s' has high missingness (%.1f%%).",
                          cn, 100 * pna), class = "healthmarkers_impute_warn_high_missingness")
    }
  }
  invisible(TRUE)
}

# tests/testthat/test_impute_missing.R

library(testthat)
library(tibble)

# 1) impute_missing()
test_that("impute_missing mean/median/zero/constant work", {
  skip_on_cran()
  df <- tibble(x = c(1, NA, 3), y = c(NA, 2, NA), z = c("a", "b", "c"))
  # default mean
  out1 <- suppressWarnings(impute_missing(df))
  expect_equal(out1$x, c(1, mean(c(1, 3)), 3))
  expect_equal(out1$y, c(mean(2), 2, mean(2)))
  # zero only x
  out2 <- suppressWarnings(impute_missing(df, method = "zero", cols = "x"))
  expect_equal(out2$x, c(1, 0, 3))
  expect_true(is.na(out2$y[1]))
  # constant
  out3 <- suppressWarnings(impute_missing(df, method = "constant", constant = 99))
  expect_equal(out3$x, c(1, 99, 3))
  expect_equal(out3$y, c(99, 2, 99))
})

test_that("impute_missing errors appropriately", {
  skip_on_cran()
  expect_error(impute_missing("not a df"), "must be a data.frame or tibble")
  expect_error(impute_missing(tibble(a = 1), cols = "b"), "Some `cols` not in data")
})

# 2) impute_mice()
test_that("impute_mice basic functionality is silent and imputes", {
  skip_on_cran()
  skip_if_not_installed("mice")
  df <- tibble(a = c(1, NA, 3), b = c(4, 5, NA), c = c("x", "y", "z"))
  expect_silent({
    out <- impute_mice(df, m = 1)
  })
  expect_true(all(!is.na(out$a)))
  expect_true(all(!is.na(out$b)))
  # non-numeric untouched
  expect_identical(out$c, df$c)
})

test_that("impute_mice errors on bad inputs", {
  skip_on_cran()
  skip_if_not_installed("mice")
  expect_error(impute_mice("nope"), "must be a data.frame or tibble")
  expect_error(impute_mice(tibble(x = 1:3), cols = "foo"),
               "Some `cols` not in data")
  # only one numeric column with missing values -> error
  expect_error(
    impute_mice(tibble(a = c(1, NA, 3), b = c(1, 2, 3))),
    "need at least two numeric columns"
  )
})

# 3) impute_missforest()
test_that("impute_missforest fills NAs and preserves non-numeric", {
  skip_on_cran()
  skip_if_not_installed("missForest")
  df <- tibble(
    a = c(2, NA, 8),
    b = c(NA, 6, 9),
    d = factor(c("u", "v", "w"))
  )
  out <- suppressWarnings(impute_missforest(df))
  # numeric columns are now complete
  expect_false(any(is.na(out$a)))
  expect_false(any(is.na(out$b)))
  # non-numeric untouched
  expect_identical(out$d, df$d)
})

test_that("impute_missforest errors on bad inputs", {
  skip_on_cran()
  skip_if_not_installed("missForest")
  expect_error(impute_missforest("nope"), "must be a data.frame or tibble")
  expect_error(impute_missforest(tibble(x = 1:3), cols = "foo"),
               "Some `cols` not in data")
  # only one numeric column with missing values -> error
  expect_error(
    impute_missforest(tibble(a = c(1, NA, 3), b = c(1, 2, 3))),
    "need at least two numeric columns"
  )
})

# Verbose message tests -------------------------------------------------------

test_that("impute_missing verbose emits preparing inputs and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(x = c(1, NA, 3), y = c(NA, 2, NA))
  expect_message(suppressWarnings(impute_missing(df, verbose = TRUE)), "impute_missing")
  expect_message(suppressWarnings(impute_missing(df, verbose = TRUE)), "results:")
})

test_that("impute_missing verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble(x = c(1, NA, 3), y = c(NA, 2, NA))
  msgs <- testthat::capture_messages(suppressWarnings(impute_missing(df, verbose = TRUE)))
  expect_equal(sum(grepl("preparing inputs", msgs)), 1L)
  expect_equal(sum(grepl("results:",         msgs)), 1L)
})

test_that("impute_mice verbose emits preparing inputs and results messages", {
  skip_on_cran()
  skip_if_not_installed("mice")
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(a = c(1, NA, 3), b = c(4, 5, NA))
  expect_message(impute_mice(df, m = 1, verbose = TRUE), "impute_mice")
  expect_message(impute_mice(df, m = 1, verbose = TRUE), "results:")
})

test_that("impute_mice verbose double-fire guard", {
  skip_on_cran()
  skip_if_not_installed("mice")
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble(a = c(1, NA, 3), b = c(4, 5, NA))
  msgs <- testthat::capture_messages(impute_mice(df, m = 1, verbose = TRUE))
  expect_equal(sum(grepl("preparing inputs", msgs)), 1L)
  expect_equal(sum(grepl("results:",         msgs)), 1L)
})

test_that("impute_missforest verbose emits preparing inputs and results messages", {
  skip_on_cran()
  skip_if_not_installed("missForest")
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(a = c(2, NA, 8), b = c(NA, 6, 9))
  expect_message(suppressWarnings(impute_missforest(df, verbose = TRUE)), "impute_missforest")
  expect_message(suppressWarnings(impute_missforest(df, verbose = TRUE)), "results:")
})

test_that("impute_missforest verbose double-fire guard", {
  skip_on_cran()
  skip_if_not_installed("missForest")
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- tibble(a = c(2, NA, 8), b = c(NA, 6, 9))
  msgs <- testthat::capture_messages(suppressWarnings(impute_missforest(df, verbose = TRUE)))
  expect_equal(sum(grepl("preparing inputs", msgs)), 1L)
  expect_equal(sum(grepl("results:",         msgs)), 1L)
})

#' Compute tracer/DXA-based insulin sensitivity indices
#'
#' Uses stable isotope tracer infusion rates and DXA-measured fat mass
#' to compute peripheral and adipose insulin sensitivity and related metrics.
#'
#' Modes:
#' - Adipose-only indices when only adipose-related keys are mapped (no OGTT glucose/insulin time series)
#' - Full indices otherwise
#'
#' Expected units:
#' - Glucose: mmol/L (internally converted to mg/dL when needed)
#' - Insulin: pmol/L (internally converted to muU/mL via /6)
#' - TG: mmol/L (to mg/dL via *88.57); HDL-c: mmol/L (to mg/dL via *38.67)
#' - Tracer rates: mumol/min
#' - Fat mass, weight: kg; BMI: kg/m^2
#'
#' @param data A data.frame or tibble containing raw measurements.
#' @param col_map Named list with entries (depending on mode):
#'   Adipose-only required:
#'     - I0: fasting insulin (pmol/L)
#'     - rate_glycerol, rate_palmitate: tracer rates (mumol/min)
#'     - fat_mass, weight, bmi: body composition
#'     - HDL_c: HDL cholesterol (mmol/L)
#'   Full mode additionally requires:
#'     - G0, G30, G120: glucose (mmol/L)
#'     - I30, I120: insulin (pmol/L)
#'     - TG: triglycerides (mmol/L)
#'     - FFA: free fatty acids (mmol/L)
#' @param normalize Ignored (kept for backward compatibility).
#' @param na_action One of c("keep","omit","error") for NA handling on required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness warnings on required inputs. Default 0.2.
#' @param verbose Logical; if TRUE, prints progress messages and a completion summary.
#'
#' @return
#' - Adipose-only tibble columns: LIRI_inv, Lipo_inv, ATIRI_inv
#' - Full-mode tibble columns: I_AUC, FFA_AUC, tracer_palmitate_SI, tracer_glycerol_SI, LIRI_inv, Lipo_inv, ATIRI_inv
#'
#' @examples
#' df <- data.frame(
#'   I0 = c(60, 75), rate_glycerol = c(2.1, 2.8), rate_palmitate = c(1.8, 2.3),
#'   fat_mass = c(18, 24), weight = c(72, 85), BMI = c(24, 29),
#'   HDL_c = c(1.3, 1.1)
#' )
#' col_map <- list(I0="I0", rate_glycerol="rate_glycerol",
#'                 rate_palmitate="rate_palmitate", fat_mass="fat_mass",
#'                 weight="weight", bmi="BMI", HDL_c="HDL_c")
#' tracer_dxa_is(df, col_map = col_map)
#'
#' @note
#' `tracer_palmitate_SI` and `tracer_glycerol_SI` are simple rate/fat_mass
#' ratios; the Steele (1959) non-steady-state tracer equation is **not**
#' implemented here. The LIRI formula coefficients (-0.091, 0.4, 0.346,
#' -0.408, 0.435) are attributed to Gastaldelli et al. but the paper cited
#' (`gastaldelli2004betacell`) covers beta-cell dysfunction, not LIRI
#' derivation; the primary LIRI source should be verified. In adipose-only
#' mode (I30 absent) the mean-insulin term uses I0 twice as a fallback.
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
#' @references
#' \insertRef{groop1989tracer}{HealthMarkers} (tracer lipolysis methodology; background)
#' \insertRef{steele1959glucose}{HealthMarkers} (tracer dilution theory; Steele equation not directly implemented — background)
#' \insertRef{roden1996ffa}{HealthMarkers} (FFA-induced insulin resistance mechanism; background)
#' \insertRef{gastaldelli2004betacell}{HealthMarkers} (beta-cell dysfunction context; LIRI formula source unverified — background)
#' \insertRef{karpe2011ffa}{HealthMarkers} (FFA and insulin resistance review; background)
#' \insertRef{petersen2007muscleinsulin}{HealthMarkers} (muscle insulin resistance and metabolic syndrome; background)
#' \insertRef{santomauro1999acipimox}{HealthMarkers} (FFA lowering and insulin sensitivity; background)
tracer_dxa_is <- function(data, col_map = NULL,
                          normalize = NULL,
                          na_action = c("keep","omit","error"),
                          na_warn_prop = 0.2,
                          verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "tracer_dxa_is"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  na_action <- match.arg(na_action)

  if (!is.null(normalize)) {
    # maintain signature compatibility; explicitly ignore
    # rlang::warn("tracer_dxa_is(): `normalize` argument is ignored.")
  }

  if (!is.data.frame(data)) {
    rlang::abort("tracer_dxa_is(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_tracer_error_data_type")
  }
  if (isTRUE(verbose)) hm_inform("tracer_dxa_is(): preparing inputs", level = "inform")
  else hm_inform("tracer_dxa_is(): preparing inputs", level = "debug")

  adipose_keys <- c("I0", "rate_palmitate", "rate_glycerol", "fat_mass", "weight", "HDL_c", "bmi")
  all_tracer_keys <- c(adipose_keys, "G0","G30","G120","I30","I120","TG","FFA")
  cm      <- .hm_build_col_map(data, col_map, all_tracer_keys, fn = "tracer_dxa_is")
  data    <- cm$data
  col_map <- cm$col_map
  # Adipose-only when all adipose keys present and not all OGTT keys present
  adipose_only <- all(adipose_keys %in% names(col_map)) &&
    !all(c("G0", "I30") %in% names(col_map))

  full_keys <- c(adipose_keys, "G0","G30","G120","I30","I120","TG","FFA")

  required_keys <- if (adipose_only) adipose_keys else full_keys
  hm_validate_inputs(data, col_map, required_keys = required_keys, fn = "tracer_dxa_is")

  # Validate data contains the mapped columns
  mapped_cols <- unname(unlist(col_map[required_keys], use.names = FALSE))
  missing_cols <- setdiff(mapped_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("tracer_dxa_is(): missing required columns in `data`: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_tracer_error_missing_columns"
    )
  }

  # Coerce mapped numeric columns to numeric; warn on NAs introduced
  to_num_keys <- intersect(required_keys, c(
    "G0","G30","G120","I0","I30","I120","TG","HDL_c","FFA",
    "rate_glycerol","rate_palmitate","fat_mass","weight","bmi"
  ))
  for (key in to_num_keys) {
    cn <- col_map[[key]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("tracer_dxa_is(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
    # Set non-finite to NA for safety
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness warnings on required inputs
  .tx_warn_high_missing(data, mapped_cols, na_warn_prop)
  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose)) {
    marker_list <- if (adipose_only) "LIRI_inv, Lipo_inv, ATIRI_inv" else "I_AUC, FFA_AUC, tracer_palmitate_SI, tracer_glycerol_SI, LIRI_inv, Lipo_inv, ATIRI_inv"
    hm_inform(sprintf("%s(): computing markers:\n  %s", fn_name, marker_list), level = "inform")
  }

  # NA policy on required inputs
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(mapped_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("tracer_dxa_is(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_tracer_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(mapped_cols, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
                msg   = sprintf("tracer_dxa_is(): omitting %d rows with NA in required inputs", sum(!keep)))
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    if (adipose_only) {
      return(tibble::tibble(
        LIRI_inv  = numeric(),
        Lipo_inv  = numeric(),
        ATIRI_inv = numeric()
      ))
    } else {
      return(tibble::tibble(
        I_AUC               = numeric(),
        FFA_AUC             = numeric(),
        tracer_palmitate_SI = numeric(),
        tracer_glycerol_SI  = numeric(),
        LIRI_inv            = numeric(),
        Lipo_inv            = numeric(),
        ATIRI_inv           = numeric()
      ))
    }
  }

  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }
  safe_log10 <- function(x) {
    y <- x
    y[!(is.finite(y) & y > 0)] <- NA_real_
    out <- suppressWarnings(log10(y))
    out[!is.finite(out)] <- NA_real_
    out
  }

  if (adipose_only) {
    if (isTRUE(verbose)) hm_inform("tracer_dxa_is(): adipose-only indices", level = "inform")

    I0_u <- data[[col_map$I0]] / 6
    Lipo_inv <- -1 * (data[[col_map$rate_glycerol]] * I0_u)
    ATIRI_inv <- -1 * (data[[col_map$rate_palmitate]] * I0_u)
    LIRI_inv <- -1 * (
      -0.091 +
        safe_log10((I0_u + I0_u) / 2 * 6) * 0.4 + # use I0 twice if I30 missing; preserve behavior
        safe_log10((safe_div(data[[col_map$fat_mass]], data[[col_map$weight]], "fm_wt") * 100)) * 0.346 -
        safe_log10(data[[col_map$HDL_c]] * 38.67) * 0.408 +
        safe_log10(data[[col_map$bmi]]) * 0.435
    )

    out <- tibble::tibble(
      LIRI_inv  = as.numeric(LIRI_inv),
      Lipo_inv  = as.numeric(Lipo_inv),
      ATIRI_inv = as.numeric(ATIRI_inv)
    )
  } else {
    if (isTRUE(verbose)) hm_inform("tracer_dxa_is(): computing indices", level = "inform")

    # Unit conversions
    I0_u   <- data[[col_map$I0]]   / 6
    I30_u  <- data[[col_map$I30]]  / 6
    I120_u <- data[[col_map$I120]] / 6
    # Glucose present but not used directly in current formulas; keep conversions in case of future indices
    G0_mg   <- data[[col_map$G0]]   * 18
    G30_mg  <- data[[col_map$G30]]  * 18
    G120_mg <- data[[col_map$G120]] * 18
    TG_mg   <- data[[col_map$TG]]   * 88.57
    HDL_mg  <- data[[col_map$HDL_c]] * 38.67
    FFA_val <- data[[col_map$FFA]]

    # AUCs (minutes): insulin trapezoid (0-30, 30-120); FFA flat (single measure over 0-120)
    I_AUC <- 0.5 * ((I0_u + I30_u) * 30 + (I30_u + I120_u) * 90)
    FFA_AUC <- 0.5 * (FFA_val + FFA_val) * 120

    # Tracer sensitivity per kg fat mass
    tracer_palmitate_SI <- safe_div(data[[col_map$rate_palmitate]], data[[col_map$fat_mass]], "palmitate_fm")
    tracer_glycerol_SI  <- safe_div(data[[col_map$rate_glycerol]],  data[[col_map$fat_mass]], "glycerol_fm")

    # Adipose indices
    # Note: safe logs avoid non-positive values; units preserved as in prior code
    LIRI_inv <- -1 * (
      -0.091 +
        safe_log10((I0_u + I30_u) / 2 * 6) * 0.4 +
        safe_log10((safe_div(data[[col_map$fat_mass]], data[[col_map$weight]], "fm_wt") * 100)) * 0.346 -
        safe_log10(HDL_mg) * 0.408 +
        safe_log10(data[[col_map$bmi]]) * 0.435
    )
    Lipo_inv <- -1 * (data[[col_map$rate_glycerol]] * I0_u)
    ATIRI_inv <- -1 * (data[[col_map$rate_palmitate]] * I0_u)

    out <- tibble::tibble(
      I_AUC               = as.numeric(I_AUC),
      FFA_AUC             = as.numeric(FFA_AUC),
      tracer_palmitate_SI = as.numeric(tracer_palmitate_SI),
      tracer_glycerol_SI  = as.numeric(tracer_glycerol_SI),
      LIRI_inv            = as.numeric(LIRI_inv),
      Lipo_inv            = as.numeric(Lipo_inv),
      ATIRI_inv           = as.numeric(ATIRI_inv)
    )
  }

  # Consolidated zero-denominator warning
  dz <- dz_env$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz); nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("tracer_dxa_is(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  if (!is.null(id_col)) {
    # id_col comes from original data (before omit); need the kept rows
    id_vec <- data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# ---- internal helpers ---------------------------------------------------------

.tx_warn_high_missing <- function(data, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- data[[cn]]
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf(
        "tracer_dxa_is(): column '%s' has high missingness (%.1f%%).",
        cn, 100 * pna
      ))
    }
  }
  invisible(TRUE)
}

.tx_default_extreme_rules <- function(adipose_only) {
  # Very broad plausibility bounds; keys match col_map keys
  base <- list(
    I0             = c(0, 5000),   # pmol/L
    I30            = c(0, 5000),
    I120           = c(0, 5000),
    G0             = c(0, 40),     # mmol/L
    G30            = c(0, 40),
    G120           = c(0, 40),
    TG             = c(0, 50),     # mmol/L
    HDL_c          = c(0, 10),     # mmol/L
    FFA            = c(0, 5),      # mmol/L
    rate_glycerol  = c(0, 10000),  # mumol/min
    rate_palmitate = c(0, 10000),  # mumol/min
    fat_mass       = c(0.1, 200),  # kg
    weight         = c(1, 400),    # kg
    bmi            = c(5, 100)     # kg/m^2
  )
  if (isTRUE(adipose_only)) {
    base[c("I0","rate_glycerol","rate_palmitate","fat_mass","weight","HDL_c","bmi")]
  } else {
    base
  }
}

# tests/testthat/test_tracer_dxa_is.R
library(testthat)
library(tibble)
library(dplyr)

test_that("adipose-only mode returns three indices with correct values", {
  skip_on_cran()
  df <- tibble(
    I0              = 60,    # pmol/L
    rate_palmitate  = 1.5,   # µmol/min
    rate_glycerol   = 2.0,   # µmol/min
    fat_mass        = 20,    # kg
    weight          = 70,    # kg
    HDL_c           = 1.2,   # mmol/L
    bmi             = 24
  )
  col_map_adipose <- list(
    I0             = "I0",
    rate_palmitate = "rate_palmitate",
    rate_glycerol  = "rate_glycerol",
    fat_mass       = "fat_mass",
    weight         = "weight",
    HDL_c          = "HDL_c",
    bmi            = "bmi"
  )
  out <- tracer_dxa_is(df, col_map_adipose, verbose = FALSE)
  expect_named(out, c("LIRI_inv", "Lipo_inv", "ATIRI_inv"))

  # Lipo_inv = - rate_glycerol * (I0/6)
  expect_equal(out$Lipo_inv, -1 * (2.0 * (60 / 6)))
  # ATIRI_inv = - rate_palmitate * (I0/6)
  expect_equal(out$ATIRI_inv, -1 * (1.5 * (60 / 6)))
  # LIRI_inv should be finite given positive inputs
  expect_true(is.finite(out$LIRI_inv))
})

test_that("full mode errors when required columns are missing", {
  skip_on_cran()
  df <- tibble(G0 = 5, I0 = 60)
  full_map <- list(
    G0 = "G0", G30 = "G30", G120 = "G120",
    I0 = "I0", I30 = "I30", I120 = "I120",
    TG = "TG", HDL_c = "HDL_c", FFA = "FFA",
    rate_palmitate = "rate_palmitate",
    rate_glycerol = "rate_glycerol",
    fat_mass = "fat_mass",
    weight = "weight",
    bmi = "bmi"
  )
  expect_error(
    tracer_dxa_is(df, full_map),
    "missing required columns"
  )
})

test_that("full mode returns all seven indices and correct values", {
  skip_on_cran()
  df_full <- tibble(
    G0             = 5,
    G30            = 7,
    G120           = 7,
    I0             = 60,
    I30            = 65,
    I120           = 70,
    TG             = 1.5,
    HDL_c          = 1.2,
    FFA            = 0.5,
    rate_palmitate = 1.5,
    rate_glycerol  = 2.0,
    fat_mass       = 20,
    weight         = 70,
    bmi            = 24
  )
  full_map <- list(
    G0 = "G0", G30 = "G30", G120 = "G120",
    I0 = "I0", I30 = "I30", I120 = "I120",
    TG = "TG", HDL_c = "HDL_c", FFA = "FFA",
    rate_palmitate = "rate_palmitate",
    rate_glycerol = "rate_glycerol",
    fat_mass = "fat_mass",
    weight = "weight",
    bmi = "bmi"
  )
  out <- tracer_dxa_is(df_full, full_map, verbose = FALSE)

  expected_names <- c(
    "I_AUC", "FFA_AUC",
    "tracer_palmitate_SI", "tracer_glycerol_SI",
    "LIRI_inv", "Lipo_inv", "ATIRI_inv"
  )
  expect_named(out, expected_names)

  I0_u <- 60 / 6
  I30_u <- 65 / 6
  I120_u <- 70 / 6

  I_AUC_expected <- 0.5 * ((I0_u + I30_u) * 30 + (I30_u + I120_u) * 90)
  expect_equal(out$I_AUC, I_AUC_expected, tolerance = 1e-8)

  expect_equal(out$FFA_AUC, 0.5 * (0.5 + 0.5) * 120)

  expect_equal(out$tracer_palmitate_SI, 1.5 / 20)
  expect_equal(out$tracer_glycerol_SI, 2.0 / 20)

  expect_equal(out$Lipo_inv, -1 * (2.0 * I0_u))
  expect_equal(out$ATIRI_inv, -1 * (1.5 * I0_u))
})

test_that("vectorized over multiple rows", {
  skip_on_cran()
  base_df <- tibble(
    G0 = 5, G30 = 7, G120 = 7,
    I0 = 60, I30 = 65, I120 = 70,
    TG = 1.5, HDL_c = 1.2, FFA = 0.5,
    rate_palmitate = 1.5, rate_glycerol = 2.0,
    fat_mass = 20, weight = 70, bmi = 24
  )
  df2 <- bind_rows(
    base_df,
    mutate(base_df,
      I0             = 120,
      rate_glycerol  = 4,
      rate_palmitate = 3
    )
  )
  full_map <- list(
    G0 = "G0", G30 = "G30", G120 = "G120",
    I0 = "I0", I30 = "I30", I120 = "I120",
    TG = "TG", HDL_c = "HDL_c", FFA = "FFA",
    rate_palmitate = "rate_palmitate",
    rate_glycerol = "rate_glycerol",
    fat_mass = "fat_mass",
    weight = "weight",
    bmi = "bmi"
  )
  out2 <- tracer_dxa_is(df2, full_map, verbose = FALSE)
  expect_equal(nrow(out2), 2)
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df_full <- tibble(
    G0 = 5, G30 = 7, G120 = 7,
    I0 = 60, I30 = 65, I120 = 70,
    TG = 1.5, HDL_c = 1.2, FFA = 0.5,
    rate_palmitate = 1.5, rate_glycerol = 2.0,
    fat_mass = 20, weight = 70, bmi = 24
  )
  col_map_adipose <- list(
    I0             = "I0",
    rate_palmitate = "rate_palmitate",
    rate_glycerol  = "rate_glycerol",
    fat_mass       = "fat_mass",
    weight         = "weight",
    HDL_c          = "HDL_c",
    bmi            = "bmi"
  )
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(tracer_dxa_is(df_full, col_map_adipose, verbose = TRUE), "tracer_dxa_is")
  expect_message(tracer_dxa_is(df_full, col_map_adipose, verbose = TRUE), "col_map")
  expect_message(tracer_dxa_is(df_full, col_map_adipose, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df_full <- tibble(
    G0 = 5, G30 = 7, G120 = 7,
    I0 = 60, I30 = 65, I120 = 70,
    TG = 1.5, HDL_c = 1.2, FFA = 0.5,
    rate_palmitate = 1.5, rate_glycerol = 2.0,
    fat_mass = 20, weight = 70, bmi = 24
  )
  col_map_adipose <- list(
    I0             = "I0",
    rate_palmitate = "rate_palmitate",
    rate_glycerol  = "rate_glycerol",
    fat_mass       = "fat_mass",
    weight         = "weight",
    HDL_c          = "HDL_c",
    bmi            = "bmi"
  )
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(tracer_dxa_is(df_full, col_map_adipose, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action policies: error and omit behave as expected", {
  skip_on_cran()
  col_map_adipose <- list(
    I0             = "I0",
    rate_palmitate = "rate_palmitate",
    rate_glycerol  = "rate_glycerol",
    fat_mass       = "fat_mass",
    weight         = "weight",
    HDL_c          = "HDL_c",
    bmi            = "bmi"
  )
  df_na <- tibble(
    I0 = 60, rate_palmitate = 1.5, rate_glycerol = 2.0,
    fat_mass = NA_real_, weight = 70, HDL_c = 1.2, bmi = 24
  )
  expect_error(
    suppressWarnings(tracer_dxa_is(df_na, col_map_adipose, na_action = "error")),
    "required inputs contain missing values"
  )
  out_omit <- suppressWarnings(tracer_dxa_is(df_na, col_map_adipose, na_action = "omit"))
  expect_equal(nrow(out_omit), 0L)
})

test_that("check_extreme removed: function passes through outlier rows", {
  skip_on_cran()
  df_ext <- tibble(
    G0 = 50, G30 = 60, G120 = 70,
    I0 = 6000, I30 = 6000, I120 = 6000,
    TG = 60, HDL_c = 12, FFA = 6,
    rate_palmitate = 20000, rate_glycerol = 20000,
    fat_mass = 0.05, weight = 700, bmi = 120
  )
  full_map <- list(
    G0 = "G0", G30 = "G30", G120 = "G120",
    I0 = "I0", I30 = "I30", I120 = "I120",
    TG = "TG", HDL_c = "HDL_c", FFA = "FFA",
    rate_palmitate = "rate_palmitate",
    rate_glycerol = "rate_glycerol",
    fat_mass = "fat_mass",
    weight = "weight",
    bmi = "bmi"
  )
  out <- tracer_dxa_is(df_ext, full_map)
  expect_equal(nrow(out), 1L)
})

test_that("zero denominators emit a consolidated warning and yield NA where applicable", {
  skip_on_cran()
  df_zero <- tibble(
    G0 = 5, G30 = 7, G120 = 7,
    I0 = 60, I30 = 65, I120 = 70,
    TG = 1.5, HDL_c = 1.2, FFA = 0.5,
    rate_palmitate = 1.5, rate_glycerol = 2.0,
    fat_mass = 0,        # zero -> tracer SI undefined
    weight = 0,          # zero -> fm/weight undefined for LIRI_inv term
    bmi = 24
  )
  full_map <- list(
    G0 = "G0", G30 = "G30", G120 = "G120",
    I0 = "I0", I30 = "I30", I120 = "I120",
    TG = "TG", HDL_c = "HDL_c", FFA = "FFA",
    rate_palmitate = "rate_palmitate",
    rate_glycerol = "rate_glycerol",
    fat_mass = "fat_mass",
    weight = "weight",
    bmi = "bmi"
  )
  expect_warning(
    out_zero <- tracer_dxa_is(df_zero, full_map),
    "zero denominators detected"
  )
  expect_true(is.na(out_zero$tracer_palmitate_SI))
  expect_true(is.na(out_zero$tracer_glycerol_SI))
  # Lipo_inv unaffected by fm/weight; should still be finite
  expect_true(is.finite(out_zero$Lipo_inv))
})

#' Neurofilament Light Chain (NfL) Biomarker
#'
#' Incorporates a neurofilament light chain (NfL) measurement into the analysis pipeline.
#' Placeholder for future NfL-based computations; returns provided values with input checks.
#'
#' @details
#' NfL is released during neuroaxonal injury; elevated levels in CSF or blood indicate
#' neuroaxonal damage and typically increase with age and in neurological diseases.
#' Interpretation requires context-specific and age-adjusted references. This function
#' simply returns the input NfL values (assumed in a single matrix/fluid, e.g., plasma pg/mL)
#' without classification.
#'
#' @param data A data.frame or tibble with an NfL concentration column.
#' @param col_map Named list with `nfl` indicating the NfL column name.
#' @param verbose Logical; if TRUE (default), emits progress messages.
#' @param na_action One of c("keep","omit","error","ignore","warn").
#'
#' @return A tibble with one column: nfl_value (numeric; same units as input).
#'
#' @examples
#' df <- data.frame(NfL = c(8.5, 14.2, 22.1))
#' nfl_marker(df)
#'
#' @references
#' \insertRef{simren2022nfl}{HealthMarkers}
#' \insertRef{disanto2017nfl}{HealthMarkers}
#'
#' @export
nfl_marker <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "nfl_marker"
  .hm_log_input(data, data_name, fn_name, verbose)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff
  id_col <- .hm_detect_id_col(data)

  # --- validate data and col_map (HM-CS style) -------------------------------
  if (!is.data.frame(data)) {
    rlang::abort(
      "nfl_marker(): `data` must be a data.frame or tibble.",
      class = "healthmarkers_nfl_error_data_type"
    )
  }

  cm      <- .hm_build_col_map(data, col_map, "nfl", fn = "nfl_marker")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  if (!is.list(col_map)) {
    rlang::abort(
      "nfl_marker(): `col_map` must be a named list.",
      class = "healthmarkers_nfl_error_colmap_type"
    )
  }

  if (!("nfl" %in% names(col_map))) {
    # Graceful NA only when nothing provided and inference found nothing
    if (length(cm$user_keys) == 0L && length(cm$inferred_keys) == 0L) {
      return(tibble::tibble(nfl_value = rep(NA_real_, nrow(data))))
    }
    rlang::abort(
      "nfl_marker(): missing or empty `col_map`.",
      class = "healthmarkers_nfl_error_missing_map"
    )
  }

  mapped <- unname(col_map[["nfl"]])
  if (is.null(mapped) || length(mapped) == 0L || !nzchar(as.character(mapped))) {
    rlang::abort(
      "nfl_marker(): missing col_map entry value for: nfl",
      class = "healthmarkers_nfl_error_bad_map_values"
    )
  }

  if (!(mapped %in% names(data))) {
    rlang::abort(
      paste0("nfl_marker(): missing required column in data: ", mapped),
      class = "healthmarkers_nfl_error_missing_columns"
    )
  }

  # --- verbose messages -----------------------------------------------------
  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  nfl_value  [passthrough of NfL input]", fn_name), level = "inform")

  # --- coerce to numeric; warn on NA introduction ---------------------------
  cn <- mapped
  if (!is.numeric(data[[cn]])) {
    old <- data[[cn]]
    suppressWarnings(new <- as.numeric(old))
    intro <- sum(is.na(new) & !is.na(old))
    if (intro > 0L) {
      rlang::warn(
        sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
        class = "healthmarkers_nfl_warn_na_coercion"
      )
    }
    data[[cn]] <- new
  }
  data[[cn]][!is.finite(data[[cn]])] <- NA_real_

  nfl_vals <- data[[cn]]

  # --- NA policy ------------------------------------------------------------
  any_na <- is.na(nfl_vals)

  if (na_action_raw == "warn" && any(any_na)) {
    rlang::warn(
      "Missing NfL values; output will be NA for those entries.",
      class = "healthmarkers_nfl_warn_missing_inputs"
    )
  }

  if (na_action_eff == "error" && any(any_na)) {
    rlang::abort(
      "nfl_marker(): required input contains missing values (na_action='error').",
      class = "healthmarkers_nfl_error_missing_values"
    )
  }

  keep <- if (na_action_eff == "omit") !any_na else rep(TRUE, length(nfl_vals))
  d_nfl <- nfl_vals[keep]

  # --- domain warnings: negative values -------------------------------------
  if (any(is.finite(d_nfl) & d_nfl < 0)) {
    rlang::warn(
      "nfl_marker(): negative NfL values detected; check input data.",
      class = "healthmarkers_nfl_warn_negative_values"
    )
  }

  out <- tibble::tibble(nfl_value = d_nfl)

  # --- pad back if not omitting ---------------------------------------------
  if (na_action_eff != "omit") {
    res <- tibble::tibble(nfl_value = rep(NA_real_, length(nfl_vals)))
    res$nfl_value[keep] <- out$nfl_value
    out <- res
  }

  # --- ID prepend + verbose results -----------------------------------------
  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

library(testthat)

cm <- list(nfl = "NfL")

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(NfL = c(10, 20))

  # empty list with inferrable data – succeeds via case-insensitive inference
  expect_no_error(nfl_marker(df, list()))

  expect_error(
    nfl_marker(df, list(nfl = "")),
    class = "healthmarkers_nfl_error_bad_map_values"
  )

  expect_error(
    nfl_marker(data.frame(X = 1), list(nfl = "NfL")),
    class = "healthmarkers_nfl_error_missing_columns"
  )
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df <- data.frame(NfL = c(12, 35))
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(nfl_marker(df, cm, verbose = TRUE), "nfl_marker")
  expect_message(nfl_marker(df, cm, verbose = TRUE), "col_map")
  expect_message(nfl_marker(df, cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- data.frame(NfL = c(12, 35))
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(nfl_marker(df, cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("numeric coercion warning when non-numeric introduces NAs", {
  skip_on_cran()
  df <- data.frame(NfL = c("12", "oops", "35"))

  expect_warning(
    nfl_marker(df, cm),
    class = "healthmarkers_nfl_warn_na_coercion"
  )
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(NfL = c(12, NA, 35))

  out_keep <- nfl_marker(df, cm, na_action = "keep")
  expect_equal(nrow(out_keep), 3L)
  expect_true(is.na(out_keep$nfl_value[2]))

  out_omit <- nfl_marker(df, cm, na_action = "omit")
  expect_equal(nrow(out_omit), 2L)
  expect_false(any(is.na(out_omit$nfl_value)))

  expect_error(
    nfl_marker(df, cm, na_action = "error"),
    class = "healthmarkers_nfl_error_missing_values"
  )

  expect_warning(
    nfl_marker(df, cm, na_action = "warn"),
    class = "healthmarkers_nfl_warn_missing_inputs"
  )
})

test_that("domain warnings: negative values", {
  skip_on_cran()
  df_neg <- data.frame(NfL = c(-5, 15))

  expect_warning(
    nfl_marker(df_neg, cm),
    class = "healthmarkers_nfl_warn_negative_values"
  )
})

test_that("extreme NfL values pass through without error", {
  skip_on_cran()
  df <- data.frame(NfL = c(1e7, 15, 35))
  out <- nfl_marker(df, cm)
  expect_equal(nrow(out), 3L)
  expect_equal(out$nfl_value[1], 1e7)
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(NfL = c(12, NA, 35))

  out_keep <- nfl_marker(df, cm, na_action = "keep")
  out_warn <- suppressWarnings(nfl_marker(df, cm, na_action = "warn"))

  expect_equal(nrow(out_keep), 3L)
  expect_equal(nrow(out_warn), 3L)
})

#' Compute liver-related indices (FLI, NFS, APRI, FIB-4, BARD, ALBI, MELD-XI) with validation and diagnostics
#'
#' Given routine labs and anthropometry, computes:
#' - FLI      - Fatty Liver Index (Bedogni et al. 2006)
#' - NFS      - NAFLD Fibrosis Score (Angulo et al. 2007)
#' - APRI     - AST-to-Platelet Ratio Index
#' - FIB4     - Fibrosis-4 Index
#' - BARD     - BMI-AST/ALT-Diabetes score
#' - ALBI     - Albumin-Bilirubin score
#' - MELD_XI  - MELD excluding INR
#'
#' Enhancements:
#' - Robust input validation (columns present, types) with informative errors.
#' - Configurable NA policy and optional extreme-value scanning/capping.
#' - Data-quality warnings (high missingness, non-positive logs, zero denominators).
#' - Verbose stepwise progress and completion summary.
#'
#' Units (no automatic conversion):
#' - BMI: kg/m^2; Waist: cm; TG: mg/dL; GGT/AST/ALT: U/L; Platelets: 10^9/L; Albumin: g/L; Bilirubin: mg/dL; Creatinine: mg/dL.
#' - ALBI uses bilirubin in mumol/L internally (converted as bilirubin (mg/dL) * 17.1).
#'
#' @param data A data.frame or tibble containing your liver and anthropometry data.
#' @param col_map Named list mapping these keys -> column names in `data`:
#'   - `BMI` (kg/m^2), `waist` (cm), `TG` (mg/dL), `GGT` (U/L),
#'   - `age` (years), `AST` (U/L), `ALT` (U/L), `platelets` (10^9/L),
#'   - `albumin` (g/L), `diabetes` (0/1 or logical),
#'   - `bilirubin` (mg/dL), `creatinine` (mg/dL).
#' @param verbose Logical; if `TRUE` (default), prints column mapping, input
#'   availability, physiological range information (informational only, values
#'   not altered), the list of markers being computed with their inputs, and a
#'   per-column results summary.
#' @param na_action One of `c("keep","omit","error")` controlling missing-data
#'   policy. Default "keep".
#' @param na_warn_prop Numeric in \eqn{[0,1]}; per-variable threshold for
#'   high-missingness warnings. Default 0.2.
#'
#' @return A tibble with one column per marker: `FLI`, `NFS`, `APRI`, `FIB4`,
#'   `BARD`, `ALBI`, `MELD_XI`. If an ID column is detected in `data` (e.g.
#'   `id`, `IID`, `participant_id`), it is prepended as the first output column.
#'
#' @details
#' Formulas 
#' - FLI      = logistic(0.953*ln(TG) + 0.139*BMI + 0.718*ln(GGT) + 0.053*waist - 15.745) * 100
#' - NFS      = -1.675 + 0.037*age + 0.094*BMI + 1.13*diabetes + 0.99*(AST/ALT) - 0.013*platelets - 0.066*albumin
#'   (albumin in g/L; Angulo 2007 published coefficient -0.66 was for g/dL, divided by 10 here)
#' - APRI     = (AST / 40) / platelets * 100; assumes AST upper limit of normal = 40 U/L
#' - FIB-4    = (age * AST) / (platelets * sqrt(ALT))
#' - BARD     = +1 if BMI>=28, +2 if AST/ALT>=0.8, +1 if diabetes present; sum in 0-4
#' - ALBI     = 0.66*log10(bilirubin (mumol/L)) - 0.0852*albumin (g/L)
#' - MELD-XI  = 5.11*ln(bilirubin (mg/dL)) + 11.76*ln(creatinine (mg/dL)) + 9.44
#'
#' @seealso [inflammatory_markers()], [kidney_failure_risk()], [iAge()]
#'
#' @examples
#' # Quick smoke-test
#' df <- data.frame(ALT = 25, AST = 20, BMI = 24, platelets = 250)
#' liver_markers(df, verbose = FALSE)
#'
#' \donttest{
#' library(tibble)
#' df <- tibble(
#'   BMI           = 24,
#'   waist         = 80,
#'   TG = 150,
#'   GGT           = 30,
#'   age           = 45,
#'   AST           = 25,
#'   ALT           = 20,
#'   platelets     = 250,
#'   albumin       = 42,
#'   diabetes      = FALSE,
#'   bilirubin     = 1.0,
#'   creatinine    = 0.9
#' )
#' liver_markers(df)
#' liver_markers(df, verbose = FALSE)
#' }
#'
#' @references
#' \insertRef{bedogni2006fli}{HealthMarkers}
#' \insertRef{angulo2007nfs}{HealthMarkers}
#' \insertRef{wai2003apri}{HealthMarkers}
#' \insertRef{sterling2006fib4}{HealthMarkers}
#' \insertRef{harrison2008bard}{HealthMarkers}
#' \insertRef{johnson2015albi}{HealthMarkers}
#' \insertRef{heuman2007meldxi}{HealthMarkers}
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
liver_markers <- function(data,
                          col_map      = NULL,
                          verbose      = TRUE,
                          na_action    = c("keep", "omit", "error"),
                          na_warn_prop = 0.2) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "liver_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data))
    stop(sprintf("%s(): `data` must be a data.frame or tibble.", fn_name), call. = FALSE)

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  required <- c("BMI","waist","TG","GGT","age","AST","ALT",
                "platelets","albumin","diabetes","bilirubin","creatinine")

  # --- Build col_map (infer from dictionary + materialise aliases) ---
  all_lm_keys <- c(required, "weight", "height")
  cm      <- .hm_build_col_map(data, col_map, all_lm_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Strict check: error if a user-provided mapping points to a column not in data
  if (length(cm$user_keys)) {
    bad_cols <- setdiff(
      unlist(col_map[cm$user_keys], use.names = FALSE),
      names(data))
    if (length(bad_cols))
      rlang::abort(
        paste0(fn_name, "(): mapped columns not found in data: ", paste(bad_cols, collapse = ", ")),
        class = "healthmarkers_liver_error_missing_columns")
  }

  # --- Pre-computation: BMI from weight + height if absent
  if ((is.null(col_map[["BMI"]]) || !(col_map[["BMI"]] %in% names(data))) &&
      all(c("weight", "height") %in% names(data))) {
    h   <- as.numeric(data[["height"]])
    w   <- as.numeric(data[["weight"]])
    h_m <- ifelse(is.finite(h) & h > 3, h / 100, h)
    data[["BMI"]] <- w / (h_m ^ 2)
    col_map[["BMI"]] <- "BMI"
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): pre-computation: BMI computed from weight, height", fn_name),
                level = "inform")
  }

  # Additional robust validation (structural only)
  .lm_validate_args(data, if (length(col_map) > 0L) col_map else list(placeholder = "x"),
                    na_warn_prop, NULL)

  # Helper: pull column as numeric vector, NA if key unavailable
  .lm_col <- function(k) {
    cn <- col_map[[k]]
    if (is.null(cn)) rep(NA_real_, nrow(data)) else data[[cn]]
  }

  avail_cols <- unlist(col_map[intersect(required, names(col_map))], use.names = FALSE)

  # --- Verbose: col_map (user-provided and inferred)
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # HM-CS v3: coerce to numeric for all available required inputs
  for (cn in avail_cols) {
    if (identical(cn, col_map[["diabetes"]])) next
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }
  # Missingness warnings only when na_action == "warn"
  if (identical(na_action, "warn")) {
    .lm_warn_high_missing(data, avail_cols, na_warn_prop = na_warn_prop)
  }

  # NA policy (only against available columns)
  if (identical(na_action, "error")) {
    avail_req <- intersect(required, names(col_map))
    any_na <- Reduce(`|`, lapply(avail_req, function(k) is.na(data[[col_map[[k]]]])))
    if (any(any_na)) {
      rlang::abort("liver_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_liver_error_missing_values")
    }
  } else if (identical(na_action, "omit")) {
    avail_req <- intersect(required, names(col_map))
    keep <- !Reduce(`|`, lapply(avail_req, function(k) is.na(data[[col_map[[k]]]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  } # "keep" leaves NA as-is

  # --- Verbose: inputs present/absent and which markers will be NA
  if (isTRUE(verbose)) {
    avail_keys <- names(col_map)[vapply(names(col_map), function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    absent_keys <- setdiff(required, avail_keys)
    idx_deps <- list(
      FLI     = c("BMI", "waist", "TG", "GGT"),
      NFS     = c("age", "BMI", "diabetes", "AST", "ALT", "platelets", "albumin"),
      APRI    = c("AST", "platelets"),
      FIB4    = c("age", "AST", "platelets", "ALT"),
      BARD    = c("BMI", "AST", "ALT", "diabetes"),
      ALBI    = c("bilirubin", "albumin"),
      MELD_XI = c("bilirubin", "creatinine")
    )
    na_indices <- character(0)
    for (idx in names(idx_deps)) {
      miss_for_idx <- setdiff(idx_deps[[idx]], avail_keys)
      if (length(miss_for_idx))
        na_indices <- c(na_indices,
          sprintf("  %s -> NA  [missing: %s]", idx, paste(miss_for_idx, collapse = ", ")))
    }
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(avail_keys))
      lines <- c(lines, sprintf("  present:  %s", paste(intersect(required, avail_keys), collapse = ", ")))
    if (length(absent_keys))
      lines <- c(lines, sprintf("  missing:  %s", paste(absent_keys, collapse = ", ")))
    if (length(na_indices))
      lines <- c(lines, "  indices -> NA:", na_indices)
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    rules_info <- .lm_default_extreme_rules()
    ex_info    <- .lm_extreme_scan(data, col_map, rules_info, required)
    if (ex_info$count > 0L) {
      details <- vapply(names(ex_info$flags), function(cn) {
        nb <- sum(ex_info$flags[[cn]], na.rm = TRUE)
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", cn, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    avail_keys <- names(col_map)[vapply(names(col_map), function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    idx_deps2 <- list(
      FLI     = c("BMI", "waist", "TG", "GGT"),
      NFS     = c("age", "BMI", "diabetes", "AST", "ALT", "platelets", "albumin"),
      APRI    = c("AST", "platelets"),
      FIB4    = c("age", "AST", "platelets", "ALT"),
      BARD    = c("BMI", "AST", "ALT", "diabetes"),
      ALBI    = c("bilirubin", "albumin"),
      MELD_XI = c("bilirubin", "creatinine")
    )
    status <- vapply(names(idx_deps2), function(idx) {
      miss_in <- setdiff(idx_deps2[[idx]], avail_keys)
      if (length(miss_in) == 0L)
        sprintf("  %-10s [%s]", idx, paste(idx_deps2[[idx]], collapse = ", "))
      else
        sprintf("  %-10s NA [missing: %s]", idx, paste(miss_in, collapse = ", "))
    }, character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Pull vectors (NA vector when column unavailable)
  BMI        <- .lm_col("BMI")
  waist      <- .lm_col("waist")
  TG         <- .lm_col("TG")
  GGT        <- .lm_col("GGT")
  age        <- .lm_col("age")
  AST        <- .lm_col("AST")
  ALT        <- .lm_col("ALT")
  platelets  <- .lm_col("platelets")
  albumin    <- .lm_col("albumin")
  bilirubin  <- .lm_col("bilirubin")
  creatinine <- .lm_col("creatinine")

  # Diabetes: default to 0 (no diabetes) when column absent
  diabetes_x <- {
    cn <- col_map[["diabetes"]]
    if (is.null(cn)) rep(0L, nrow(data)) else data[[cn]]
  }

  # Diabetes coercion with diagnostics (preserve prior behavior: as.integer)
  diab_ok <- is.logical(diabetes_x) || all(diabetes_x %in% c(0,1,NA))
  if (!diab_ok) {
    rlang::warn("liver_markers(): `diabetes` not in {0,1,TRUE,FALSE}; coercing with as.integer().")
  }
  diabetes <- as.integer(diabetes_x)

  # Denominator/transform diagnostics
  denom_info <- list(
    platelets = list(name = "platelets (APRI/FIB4 denominators)", vec = platelets)
  )
  .lm_warn_zero_denoms(denom_info)
  .lm_warn_nonpositive_for_log(list(
    TG = TG,
    GGT = GGT,
    bilirubin = bilirubin,
    creatinine = creatinine
  ))
  .lm_warn_nonpositive_for_sqrt(list(ALT = ALT))

  # Compute markers (unchanged formulas)
  # FLI
  L <- 0.953 * log(TG) + 0.139 * BMI + 0.718 * log(GGT) + 0.053 * waist - 15.745
  FLI <- exp(L) / (1 + exp(L)) * 100

  # NFS (albumin in g/L; published coefficient -0.66 was for g/dL, divided by 10)
  NFS <- -1.675 +
    0.037 * age +
    0.094 * BMI +
    1.13 * diabetes +
    0.99 * (AST / ALT) -
    0.013 * platelets -
    0.066 * albumin

  # APRI
  APRI <- (AST / 40) / platelets * 100

  # FIB-4
  FIB4 <- (age * AST) / (platelets * sqrt(ALT))

  # BARD (AST/ALT >= 0.8 scores 2 points per Harrison 2008; range 0-4)
  BARD <- as.integer((BMI >= 28) + 2L * (AST / ALT >= 0.8) + (diabetes == 1))

  # ALBI
  bili_umol <- bilirubin * 17.1
  ALBI <- log10(bili_umol) * 0.66 + albumin * -0.0852

  # MELD-XI
  MELD_XI <- 5.11 * log(bilirubin) + 11.76 * log(creatinine) + 9.44

  out <- tibble::tibble(
    FLI     = FLI,
    NFS     = NFS,
    APRI    = APRI,
    FIB4    = FIB4,
    BARD    = BARD,
    ALBI    = ALBI,
    MELD_XI = MELD_XI
  )

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  return(out)
}

# ---- internal helpers (not exported) -----------------------------------------

.lm_validate_args <- function(data, col_map, na_warn_prop, extreme_rules = NULL) {
  if (!is.data.frame(data)) {
    rlang::abort("liver_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_liver_error_data_type")
  }
  if (!is.list(col_map) || is.null(names(col_map)) || any(names(col_map) == "")) {
    rlang::abort("liver_markers(): `col_map` must be a named list.",
                 class = "healthmarkers_liver_error_colmap_type")
  }
  if (!(is.numeric(na_warn_prop) && length(na_warn_prop) == 1L &&
        is.finite(na_warn_prop) && na_warn_prop >= 0 && na_warn_prop <= 1)) {
    rlang::abort("liver_markers(): `na_warn_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_liver_error_na_warn_prop")
  }
  invisible(TRUE)
}

.lm_default_extreme_rules <- function() {
  list(
    BMI           = c(10, 70),
    waist         = c(40, 200),
    TG            = c(10, 1500),   # mg/dL
    GGT           = c(1, 2000),    # U/L
    age           = c(18, 120),
    AST           = c(1, 5000),
    ALT           = c(1, 5000),
    platelets     = c(10, 1000),   # 10^9/L
    albumin       = c(15, 60),     # g/L
    bilirubin     = c(0.1, 40),    # mg/dL
    creatinine    = c(0.2, 20)     # mg/dL
  )
}

.lm_extreme_scan <- function(df, col_map, rules, required) {
  count <- 0L
  flags <- list()
  for (nm in intersect(names(rules), required)) {
    cn <- col_map[[nm]]
    if (!cn %in% names(df)) next
    x <- df[[cn]]
    rng <- rules[[nm]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    flags[[cn]] <- bad
    count <- count + sum(bad, na.rm = TRUE)
  }
  list(count = count, flags = flags)
}

.lm_warn_high_missing <- function(df, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- df[[cn]]
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("liver_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
    # Negative values quick check (vars expected >= 0)
    neg_n <- sum(is.finite(x) & x < 0)
    if (neg_n > 0L) {
      rlang::warn(sprintf("liver_markers(): column '%s' contains %d negative values; check units.", cn, neg_n))
    }
  }
  invisible(TRUE)
}

.lm_warn_zero_denoms <- function(denoms) {
  msgs <- character(0)
  for (nm in names(denoms)) {
    v <- denoms[[nm]]$vec
    lab <- denoms[[nm]]$name
    n0 <- sum(is.finite(v) & v == 0)
    if (n0 > 0L) msgs <- c(msgs, sprintf("%s==0: %d", lab, n0))
  }
  if (length(msgs)) {
    rlang::warn(sprintf("liver_markers(): zero denominators detected -> %s. Ratios may yield Inf/NaN.", paste(msgs, collapse = ", ")))
  }
  invisible(TRUE)
}

.lm_warn_nonpositive_for_log <- function(named_vecs) {
  for (nm in names(named_vecs)) {
    v <- named_vecs[[nm]]
    nbad <- sum(is.finite(v) & v <= 0)
    if (nbad > 0L) {
      rlang::warn(sprintf("liver_markers(): '%s' contains %d non-positive values; log() undefined.", nm, nbad))
    }
  }
  invisible(TRUE)
}

.lm_warn_nonpositive_for_sqrt <- function(named_vecs) {
  for (nm in names(named_vecs)) {
    v <- named_vecs[[nm]]
    nbad <- sum(is.finite(v) & v < 0)
    if (nbad > 0L) {
      rlang::warn(sprintf("liver_markers(): '%s' contains %d negative values; sqrt() undefined for negatives.", nm, nbad))
    }
  }
  invisible(TRUE)
}

library(testthat)
library(tibble)
library(HealthMarkers)

test_that("computes markers for a minimal valid example", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  out <- liver_markers(df, col_map = cm)
  expect_s3_class(out, "tbl_df")
  expect_named(out, c("FLI","NFS","APRI","FIB4","BARD","ALBI","MELD_XI"))

  L <- 0.953 * log(150) + 0.139 * 24 + 0.718 * log(30) + 0.053 * 80 - 15.745
  expect_equal(out$FLI, exp(L) / (1 + exp(L)) * 100, tolerance = 1e-8)
  expect_equal(
    out$NFS,
    -1.675 + 0.037 * 30 + 0.094 * 24 + 1.13 * 0 + 0.99 * (25/20) - 0.013 * 250 - 0.066 * 45,
    tolerance = 1e-8
  )
  expect_equal(out$APRI, (25/40)/250*100, tolerance = 1e-8)
  expect_equal(out$FIB4, (30*25)/(250*sqrt(20)), tolerance = 1e-8)
  expect_equal(out$BARD, as.integer((24 >= 28) + 2L*(25/20 >= 0.8) + (0 == 1)))
  expect_equal(out$ALBI, log10(1.0*17.1)*0.66 + 45*(-0.0852), tolerance = 1e-8)
  expect_equal(out$MELD_XI, 5.11*log(1.0) + 11.76*log(0.9) + 9.44, tolerance = 1e-8)
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_message(liver_markers(df, col_map = cm, verbose = TRUE), "liver_markers")
  expect_message(liver_markers(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(liver_markers(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  msgs <- testthat::capture_messages(
    liver_markers(df, col_map = cm, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action='omit' emits omit message and preserves expected row count", {
  skip_on_cran()
  df <- tibble(
    BMI = c(24, 24), waist = c(80, 80), TG = c(150, 150), GGT = c(30, 30),
    age = c(30, 30), AST = c(25, 25), ALT = c(20, 20),
    platelets = c(250, NA_real_), albumin = c(45, 45),
    diabetes = c(FALSE, FALSE), bilirubin = c(1.0, 1.0), creatinine = c(0.9, 0.9)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    out <- liver_markers(df, col_map = cm, na_action = "omit", verbose = TRUE),
    "omitting 1 rows with NA in required inputs"
  )
  expect_equal(nrow(out), 1L)
})

test_that("custom column mapping works with renamed inputs", {
  skip_on_cran()
  df <- tibble(
    BMIx = 24, waistx = 80, TGx = 150, GGTx = 30, agex = 30,
    ASTx = 25, ALTx = 20, ptx = 250, albuminx = 45, diabx = FALSE,
    bilix = 1.0, creatx = 0.8
  )
  cm <- list(
    BMI = "BMIx", waist = "waistx", TG = "TGx", GGT = "GGTx",
    age = "agex", AST = "ASTx", ALT = "ALTx", platelets = "ptx",
    albumin = "albuminx", diabetes = "diabx", bilirubin = "bilix", creatinine = "creatx"
  )
  out <- liver_markers(df, col_map = cm)

  expect_named(out, c("FLI","NFS","APRI","FIB4","BARD","ALBI","MELD_XI"))
  L <- 0.953 * log(150) + 0.139 * 24 + 0.718 * log(30) + 0.053 * 80 - 15.745
  expect_equal(out$FLI, exp(L) / (1 + exp(L)) * 100, tolerance = 1e-8)
})

test_that("verbose emits range note for out-of-range values", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80,
    TG = 5000,  # above plausible range -> range note in verbose
    GGT = 30, age = 30, AST = 25, ALT = 20,
    platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    liver_markers(df, col_map = cm, verbose = TRUE),
    "range note"
  )
})

test_that("extreme values are not altered; non-positive inputs propagate NaN to outputs", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80,
    TG = -5,   # negative -> log(-5) = NaN -> FLI = NaN
    GGT = 30, age = 30, AST = 25, ALT = 20,
    platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_warning(
    withCallingHandlers(
      out <- liver_markers(df, col_map = cm, verbose = FALSE),
      warning = function(w) {
        if (conditionMessage(w) == "NaNs produced") invokeRestart("muffleWarning")
      }
    ),
    "log\\(\\) undefined"
  )
  expect_true(!is.finite(out$FLI))
})

test_that("numeric coercion warns when NAs introduced", {
  skip_on_cran()
  df <- tibble(
    BMI = c(24, 24), waist = c(80, 80), TG = c("150","oops"), GGT = c(30, 30),
    age = c(30, 30), AST = c(25, 25), ALT = c(20, 20),
    platelets = c(250, 250), albumin = c("45","46"),
    diabetes = c(FALSE, FALSE), bilirubin = c(1.0, 1.0), creatinine = c(0.9, 0.9)
  )
  cm <- list(
    BMI="BMI", waist="waist", TG="TG", GGT="GGT", age="age",
    AST="AST", ALT="ALT", platelets="platelets", albumin="albumin",
    diabetes="diabetes", bilirubin="bilirubin", creatinine="creatinine"
  )

  expect_warning(
    liver_markers(df, col_map = cm),
    "coerced to numeric; NAs introduced"
  )
})

test_that("diabetes logical values compute BARD correctly without warnings", {
  skip_on_cran()
  df <- tibble(
    BMI = 30, waist = 90, TG = 150, GGT = 30, age = 50,
    AST = 40, ALT = 30, platelets = 250, albumin = 45, diabetes = TRUE,
    bilirubin = 1.0, creatinine = 1.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_warning(
    out <- liver_markers(df, col_map = cm),
    NA
  )
  expect_equal(out$BARD, 4L) # BMI>=28 +1, AST/ALT>=0.8 +2 (Harrison 2008), diabetes TRUE +1
})

test_that("vectorized diabetes logical values contribute per-row", {
  skip_on_cran()
  df <- tibble(
    BMI = c(24, 24), waist = c(80, 80), TG = c(150, 150),
    GGT = c(30, 30), age = c(30, 30),
    AST = c(20, 20), ALT = c(40, 40), # AST/ALT=0.5 -> 0 points for ratio, BMI<28 -> 0
    platelets = c(250, 250), albumin = c(45, 45),
    diabetes = c(TRUE, FALSE),
    bilirubin = c(1.0, 1.0), creatinine = c(1.0, 1.0)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  out <- liver_markers(df, col_map = cm)
  expect_equal(out$BARD, c(1L, 0L))
})

test_that("liver_markers returns all seven markers and correct values", {
  skip_on_cran()
  df <- tibble(
    BMI           = 24,
    waist         = 80,
    TG = 150, # mg/dL
    GGT           = 30,
    age           = 30,
    AST           = 25,
    ALT           = 20,
    platelets     = 250,
    albumin       = 45,
    diabetes      = FALSE,
    bilirubin     = 1.0,
    creatinine    = 0.8
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- liver_markers(df, col_map = cm)

  expect_s3_class(out, "tbl_df")
  expect_named(out, c("FLI", "NFS", "APRI", "FIB4", "BARD", "ALBI", "MELD_XI"))

  # FLI
  L <- 0.953 * log(150) + 0.139 * 24 + 0.718 * log(30) + 0.053 * 80 - 15.745
  expect_equal(out$FLI, exp(L) / (1 + exp(L)) * 100, tolerance = 1e-8)

  # NFS
  expect_equal(
    out$NFS,
    -1.675 + 0.037 * 30 + 0.094 * 24 + 1.13 * 0 + 0.99 * (25 / 20) - 0.013 * 250 - 0.066 * 45,
    tolerance = 1e-8
  )

  # APRI
  expect_equal(out$APRI, (25 / 40) / 250 * 100, tolerance = 1e-8)

  # FIB-4
  expect_equal(out$FIB4, (30 * 25) / (250 * sqrt(20)), tolerance = 1e-8)

  # BARD
  expected_bard <- as.integer((24 >= 28) + 2L*(25 / 20 >= 0.8) + (0 == 1))
  expect_equal(out$BARD, expected_bard)

  # ALBI
  expect_equal(out$ALBI, log10(1.0 * 17.1) * 0.66 + 45 * (-0.0852), tolerance = 1e-8)

  # MELD_XI
  expect_equal(out$MELD_XI, 5.11 * log(1.0) + 11.76 * log(0.8) + 9.44, tolerance = 1e-8)
})

test_that("errors if mapped columns are missing in data", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE
    # bilirubin and creatinine missing
  )
  expect_error(
    liver_markers(df, col_map = list(
      BMI = "BMI", waist = "waist", TG = "TG", GGT = "GGT",
      age = "age", AST = "AST", ALT = "ALT", platelets = "platelets",
      albumin = "albumin", diabetes = "diabetes",
      bilirubin = "bilirubin", creatinine = "creatinine"
    )),
    regexp = "mapped columns not found in data|missing required columns"
  )
})

test_that("partial col_map is supplemented by dictionary inference for missing keys", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.8
  )
  cm <- list(
    BMI = "BMI", waist = "waist", TG = "TG", GGT = "GGT",
    age = "age", AST = "AST", ALT = "ALT", platelets = "platelets",
    albumin = "albumin", diabetes = "diabetes"
    # bilirubin/creatinine keys missing -- now auto-inferred from dictionary
  )
  # Should NOT error: missing keys are filled by dictionary inference
  out <- suppressWarnings(liver_markers(df, col_map = cm))
  expect_s3_class(out, "tbl_df")
  expect_true(!is.na(out$ALBI))
  expect_true(!is.na(out$MELD_XI))
})

test_that("na_action='error' aborts when required inputs contain NA", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = NA_real_, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  expect_error(
    suppressWarnings(liver_markers(df, col_map = cm, na_action = "error")),
    "required inputs contain missing values"
  )
})

test_that("na_action='omit' drops rows with NA", {
  skip_on_cran()
  df <- tibble(
    BMI = c(24, 24), waist = c(80, 80), TG = c(150, 150), GGT = c(30, 30),
    age = c(30, 30), AST = c(25, 25), ALT = c(20, 20), platelets = c(250, NA_real_),
    albumin = c(45, 45), diabetes = c(FALSE, FALSE), bilirubin = c(1.0, 1.0), creatinine = c(0.9, 0.9)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- suppressWarnings(liver_markers(df, col_map = cm, na_action = "omit"))
  expect_equal(nrow(out), 1L)
})

test_that("na_action='keep' propagates NA to outputs", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = NA_real_, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- suppressWarnings(liver_markers(df, col_map = cm, na_action = "keep"))
  expect_true(is.na(out$ALBI))
  expect_true(is.na(out$MELD_XI))
})

test_that("extreme values compute without error when outside plausible range", {
  skip_on_cran()
  df <- tibble(
    BMI = 80,              # above range
    waist = 300,           # above range
    TG = 5000,  # above range
    GGT = 5000,            # above range
    age = 150,             # above range
    AST = 6000,            # above range
    ALT = 6000,            # above range
    platelets = 5,         # below range
    albumin = 5,           # below range
    diabetes = FALSE,
    bilirubin = 0.05,      # below range
    creatinine = 50        # above range
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- liver_markers(df, col_map = cm, verbose = FALSE)
  expect_s3_class(out, "tbl_df")
  expect_equal(ncol(out), 7L)
})

test_that("extreme values produce finite or NaN/Inf outputs without error", {
  skip_on_cran()
  df <- tibble(
    BMI = 80, waist = 300, TG = 5000, GGT = 5000, age = 150,
    AST = 6000, ALT = 6000, platelets = 5, albumin = 5, diabetes = FALSE,
    bilirubin = 0.05, creatinine = 50
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- liver_markers(df, col_map = cm, verbose = FALSE)
  expect_s3_class(out, "tbl_df")
  expect_equal(ncol(out), 7L)
})

test_that("verbose range note is informational and does not warn or cap", {
  skip_on_cran()
  # All values positive so no log/sqrt warnings; only range note emitted in verbose
  df <- tibble(
    BMI = 5, waist = 30, TG = 2000, GGT = 3000,
    age = 10, AST = 1, ALT = 1,
    platelets = 5, albumin = 5, diabetes = FALSE,
    bilirubin = 0.09, creatinine = 0.1
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  # No warning emitted; range note visible only in verbose messages
  expect_no_warning(liver_markers(df, col_map = cm, verbose = FALSE))

  # Verbose mode emits range note message
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    liver_markers(df, col_map = cm, verbose = TRUE),
    "range note"
  )
})

test_that("no warning emitted for extreme values in non-verbose mode", {
  skip_on_cran()
  df <- tibble(
    BMI = 80, waist = 300, TG = 5000, GGT = 5000, age = 150,
    AST = 6000, ALT = 6000, platelets = 5, albumin = 5, diabetes = FALSE,
    bilirubin = 0.05, creatinine = 50
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_no_warning(liver_markers(df, col_map = cm, verbose = FALSE))
})

test_that("BMI pre-computed from weight and height when BMI absent", {
  skip_on_cran()
  df <- tibble(
    weight = 70, height = 170,  # BMI = 70 / (1.70)^2 = 24.22
    waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  df_bmi <- tibble(
    BMI = 70 / (170 / 100)^2,
    waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  out_precomp  <- liver_markers(df,     verbose = FALSE)
  out_explicit <- liver_markers(df_bmi, verbose = FALSE)
  expect_equal(out_precomp$FLI, out_explicit$FLI, tolerance = 1e-6)
  expect_equal(out_precomp$NFS, out_explicit$NFS, tolerance = 1e-6)
})

test_that("ID column is detected and prepended to output", {
  skip_on_cran()
  df <- tibble(
    id  = c("p1", "p2"),
    BMI = c(24, 28), waist = c(80, 90), TG = c(150, 200), GGT = c(30, 40),
    age = c(30, 45), AST = c(25, 35), ALT = c(20, 30), platelets = c(250, 200),
    albumin = c(45, 42), diabetes = c(FALSE, TRUE),
    bilirubin = c(1.0, 1.5), creatinine = c(0.9, 1.1)
  )
  out <- liver_markers(df, verbose = FALSE)
  expect_equal(names(out)[1], "id")
  expect_equal(out$id, c("p1", "p2"))
  expect_equal(ncol(out), 8L)  # id + 7 markers
})

test_that("denominator and transform warnings fire (isolated tests)", {
  skip_on_cran()
  # Zero platelets -> denominator warning for APRI/FIB4
  df1 <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 0, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm1 <- as.list(names(df1)); names(cm1) <- names(df1)
  expect_warning(liver_markers(df1, col_map = cm1), "zero denominators detected")

  # Non-positive for log: TG
  df2 <- tibble(
    BMI = 24, waist = 80, TG = 0, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm2 <- as.list(names(df2)); names(cm2) <- names(df2)
  expect_warning(liver_markers(df2, col_map = cm2), "TG.*log\\(\\) undefined")

  # ALT negative -> target only the sqrt() warning; muffle other warnings
  df3 <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = -1, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm3 <- as.list(names(df3)); names(cm3) <- names(df3)
  expect_warning(
    withCallingHandlers(
      liver_markers(df3, col_map = cm3),
      # Muffle base numeric warning from sqrt()
      warning = function(w) {
        if (identical(conditionMessage(w), "NaNs produced")) invokeRestart("muffleWarning")
      },
      # Allow only our targeted rlang warning through
      rlang_warning = function(w) {
        msg <- conditionMessage(w)
        if (grepl("sqrt\\(\\) undefined", msg)) return(invisible(NULL))
        invokeRestart("muffleWarning")
      }
    ),
    "sqrt\\(\\) undefined"
  )
})

test_that("diabetes='1' as character does not warn and BARD is computed as 4", {
  skip_on_cran()
  df <- tibble(
    BMI = 30, waist = 90, TG = 150, GGT = 30, age = 50,
    AST = 40, ALT = 30, platelets = 250, albumin = 45, diabetes = "1",
    bilirubin = 1.0, creatinine = 1.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_warning(
    out <- liver_markers(df, col_map = cm),
    NA
  )
  # BMI=30(+1) + AST/ALT=1.33>=0.8(+2) + diabetes="1"->1L(+1) = 4
  expect_equal(out$BARD, 4L)
})

test_that("non-binary diabetes values warn about coercion and propagate NA into BARD", {
  skip_on_cran()
  df <- tibble(
    BMI = 30, waist = 90, TG = 150, GGT = 30, age = 50,
    AST = 40, ALT = 30, platelets = 250, albumin = 45, diabetes = "yes",
    bilirubin = 1.0, creatinine = 1.0
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_warning(
    out <- withCallingHandlers(
      liver_markers(df, col_map = cm),
      warning = function(w) {
        if (grepl("^NAs introduced by coercion$", conditionMessage(w))) invokeRestart("muffleWarning")
      }
    ),
    "diabetes.*coercing"
  )
  expect_true(is.na(out$BARD))
})

test_that("validating inputs message appears when verbose is TRUE", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 250, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  expect_message(
    liver_markers(df, col_map = cm, verbose = TRUE),
    "liver_markers"
  )
})

test_that("verbose summary reports Inf counts for zero denominators", {
  skip_on_cran()
  df <- tibble(
    BMI = 24, waist = 80, TG = 150, GGT = 30, age = 30,
    AST = 25, ALT = 20, platelets = 0, albumin = 45, diabetes = FALSE,
    bilirubin = 1.0, creatinine = 0.9
  )
  cm <- as.list(names(df)); names(cm) <- names(df)

  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    suppressWarnings(liver_markers(df, col_map = cm, verbose = TRUE)),
    "APRI 1/1"
  )
})

#' Validate required inputs for a calling function
#'
#' Ensures required keys exist in `col_map` and have non-empty mappings.
#' Missing keys are reported in a stable order aligned with tests.
#'
#' @param data data.frame or tibble
#' @param col_map named list mapping keys to column names
#' @param fun_name character scalar naming the calling function (e.g., "lipid_markers").
#'   Used to look up built-in required keys when `required_keys` is not supplied.
#' @param required_keys optional character vector of required col_map keys. When
#'   supplied, this takes precedence over the `fun_name` built-in lookup, making
#'   the function useful for any caller regardless of `fun_name`.
#' @return invisibly TRUE on success; otherwise aborts
#' @examples
#' df <- data.frame(TG = c(1.5, 2.0), HDL_c = c(1.2, 1.0),
#'   LDL_c = c(2.0, 2.5), TC = c(4.5, 5.0))
#' # Using built-in lookup
#' validate_inputs(df,
#'   list(TG = "TG", HDL_c = "HDL_c", LDL_c = "LDL_c", TC = "TC"),
#'   fun_name = "lipid_markers")
#' # Using explicit required_keys (works for any function)
#' validate_inputs(df,
#'   list(TG = "TG", HDL_c = "HDL_c"),
#'   fun_name = "my_function",
#'   required_keys = c("TG", "HDL_c"))
#' @export
validate_inputs <- function(data, col_map, fun_name, required_keys = NULL) {
  if (!is.data.frame(data)) {
    rlang::abort("validate_inputs(): `data` must be a data.frame or tibble.")
  }
  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("validate_inputs(): `col_map` must be a named list.")
  }
  if (!is.character(fun_name) || length(fun_name) != 1L || !nzchar(fun_name)) {
    rlang::abort("validate_inputs(): `fun_name` must be a single non-empty string.")
  }

  # When required_keys is explicitly supplied, use it directly.
  # Otherwise fall back to the built-in per-function lookup.
  if (!is.null(required_keys)) {
    if (!is.character(required_keys)) {
      rlang::abort("validate_inputs(): `required_keys` must be a character vector.")
    }
    required <- required_keys
  } else {
    required <- switch(
      fun_name,
      lipid_markers = c("TG", "HDL_c", "LDL_c", "TC"),
      character(0)
    )
  }

  is_empty_map <- function(v) {
    is.null(v) || length(v) == 0L ||
      (is.atomic(v) && length(v) == 1L &&
         (is.na(v) || identical(v, "") || !nzchar(as.character(v))))
  }

  missing_keys <- vapply(required, function(k) {
    if (!(k %in% names(col_map))) return(k)
    if (is_empty_map(col_map[[k]])) return(k)
    NA_character_
  }, character(1))
  missing_keys <- missing_keys[!is.na(missing_keys)]

  if (length(missing_keys)) {
    rlang::abort(
      paste0("you must supply col_map entries for: ", paste(missing_keys, collapse = ", "))
    )
  }

  invisible(TRUE)
}

# tests/test-utils_infer_cols.R

# ----------------------------------------------------------------------------
# Tests for infer_cols()
# ----------------------------------------------------------------------------

test_that("infer_cols maps by name when present", {
  df <- tibble::tibble(G0 = 1, I0 = 2, TG = 3, `HDL-c` = 4)
  spec <- list(G0 = NULL, I0 = NULL, TG = NULL, HDL_c = NULL)
  res <- infer_cols(df, spec, verbose = FALSE)
  expect_named(res, names(spec))
  expect_equal(res$G0, "G0")
  expect_equal(res$I0, "I0")
  expect_equal(res$TG, "TG")
  expect_true(res$HDL_c %in% c("HDL-c", "HDL_c"))
})

# ----------------------------------------------------------------------------
# Tests for validate_inputs()
# ----------------------------------------------------------------------------

test_that("validate_inputs errors for missing lipid_markers keys", {
  skip_on_cran()
  df <- tibble::tibble(TG = 1, HDL_c = 1)
  col_map <- list(TG = "TG", HDL_c = "HDL_c", LDL_c = NULL, TC = NULL)
  expect_error(
    validate_inputs(df, col_map, fun_name = "lipid_markers"),
    "you must supply col_map entries for: LDL_c, TC"
  )
})

test_that("validate_inputs passes when all lipid_markers keys present", {
  skip_on_cran()
  df <- tibble::tibble(TG = 1, HDL_c = 1, LDL_c = 1, TC = 1)
  col_map <- list(TG = "TG", HDL_c = "HDL_c", LDL_c = "LDL_c", TC = "TC")
  expect_silent(validate_inputs(df, col_map, fun_name = "lipid_markers"))
})

test_that("validate_inputs with required_keys works for any fun_name", {
  skip_on_cran()
  df <- tibble::tibble(G0 = 5.5, I0 = 60)
  # Should pass when all required_keys are in col_map
  expect_silent(
    validate_inputs(df,
      list(G0 = "G0", I0 = "I0"),
      fun_name = "fasting_is",
      required_keys = c("G0", "I0"))
  )
  # Should error when a required key is missing from col_map
  expect_error(
    validate_inputs(df,
      list(G0 = "G0"),
      fun_name = "fasting_is",
      required_keys = c("G0", "I0")),
    "you must supply col_map entries for: I0"
  )
})

test_that("validate_inputs returns TRUE silently for unknown fun_name with no required_keys", {
  skip_on_cran()
  df <- tibble::tibble(x = 1)
  # No required_keys and unknown fun_name -> switch returns character(0) -> no error
  expect_silent(validate_inputs(df, list(x = "x"), fun_name = "unknown_function"))
})

test_that("validate_inputs errors when required_keys is not a character vector", {
  skip_on_cran()
  df <- tibble::tibble(x = 1)
  expect_error(
    validate_inputs(df, list(x = "x"), fun_name = "foo", required_keys = 123),
    "required_keys.*must be a character vector"
  )
})

test_that("validate_inputs() returns TRUE for lipid_markers built-in lookup", {
  df <- data.frame(TC = 5, HDL_c = 1, TG = 1.3, LDL_c = 3)
  expect_true(validate_inputs(df,
    list(TC = "TC", HDL_c = "HDL_c", TG = "TG", LDL_c = "LDL_c"),
    fun_name = "lipid_markers"))
})

test_that("validate_inputs() returns TRUE for non-lipid fun_name without required_keys", {
  skip_on_cran()
  # Non-lipid fun_names fall through switch() to character(0) — no required
  # keys are checked, so any valid data.frame + col_map passes.
  df <- data.frame(G0 = 5.0, I0 = 10.0)
  expect_true(validate_inputs(df, list(G0 = "G0"), fun_name = "fasting_is"))
})

test_that("validate_inputs() uses explicit required_keys regardless of fun_name", {
  skip_on_cran()
  df <- data.frame(G0 = 5.0, I0 = 10.0)
  expect_true(validate_inputs(df,
    list(G0 = "G0", I0 = "I0"),
    fun_name = "fasting_is",
    required_keys = c("G0", "I0")))
})

test_that("validate_inputs() aborts when required key is missing from col_map", {
  skip_on_cran()
  df <- data.frame(G0 = 5.0)
  expect_error(
    validate_inputs(df, list(G0 = "G0"), fun_name = "fasting_is",
                    required_keys = c("G0", "I0")),
    "I0"
  )
})

#' Summarize marker outputs
#'
#' @description
#' Computes simple summaries (mean, SD, IQR) for numeric columns.
#' @param x Data frame returned by marker functions
#' @param verbose Logical; if TRUE, prints progress messages
#' @return Tibble with columns: variable, mean, sd, iqr
#'
#' @examples
#' df <- data.frame(glucose = c(5.5, 6.1, 4.9), insulin = c(60, 88, 55),
#'                  bmi = c(24, 27, 22))
#' marker_summary(df)
#' @export
marker_summary <- function(x, verbose = FALSE) {
  if (!is.data.frame(x)) {
    rlang::abort("marker_summary(): `x` must be a data.frame or tibble.",
                 class = "healthmarkers_summary_error_input")
  }
  hm_inform(level = if (isTRUE(verbose)) "inform" else "debug", msg = "marker_summary(): summarizing numeric columns")

  num <- vapply(x, is.numeric, logical(1))
  if (!any(num)) {
    out <- tibble::tibble(variable = character(), mean = numeric(), sd = numeric(), iqr = numeric())
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug", msg = "marker_summary(): results: 0 variables")
    return(out)
  }

  vars <- names(x)[num]
  stats <- lapply(vars, function(v) {
    xv <- x[[v]]
    c(mean = mean(xv, na.rm = TRUE),
      sd   = stats::sd(xv, na.rm = TRUE),
      iqr  = stats::IQR(xv, na.rm = TRUE))
  })
  mat <- do.call(rbind, stats)
  out <- tibble::tibble(
    variable = vars,
    mean = as.numeric(mat[, "mean"]),
    sd   = as.numeric(mat[, "sd"]),
    iqr  = as.numeric(mat[, "iqr"])
  )

  if (isTRUE(verbose)) {
    all_na <- vapply(x[vars], function(v) all(is.na(v)), logical(1))
    hm_inform(level = "inform", msg = sprintf("marker_summary(): results: %d variables; all-NA=%d",
                                              length(vars), sum(all_na)))
  }
  out
}

library(testthat)
library(HealthMarkers)

# ---- marker_summary ---------------------------------------------------------

test_that("marker_summary returns tibble with variable/mean/sd/iqr", {
  df <- data.frame(glucose = c(5.5, 6.1, 4.9), insulin = c(60, 88, 55),
                   bmi = c(24, 27, 22))
  out <- marker_summary(df)
  expect_s3_class(out, "tbl_df")
  expect_named(out, c("variable", "mean", "sd", "iqr"))
  expect_equal(nrow(out), 3L)
  expect_setequal(out$variable, c("glucose", "insulin", "bmi"))
})

test_that("marker_summary returns correct values", {
  skip_on_cran()
  df <- data.frame(x = c(1, 2, 3))
  out <- marker_summary(df)
  expect_equal(out$mean, 2)
  expect_equal(out$sd, 1, tolerance = 1e-10)
  expect_equal(out$iqr, 1)
})

test_that("marker_summary ignores non-numeric columns", {
  skip_on_cran()
  df <- data.frame(a = 1:3, b = c("x", "y", "z"), stringsAsFactors = FALSE)
  out <- marker_summary(df)
  expect_equal(nrow(out), 1L)
  expect_equal(out$variable, "a")
})

test_that("marker_summary returns empty tibble when no numeric columns", {
  skip_on_cran()
  df <- data.frame(a = c("x", "y"), stringsAsFactors = FALSE)
  out <- marker_summary(df)
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
})

test_that("marker_summary handles NA values (mean/sd/iqr computed na.rm=TRUE)", {
  skip_on_cran()
  df <- data.frame(x = c(1, NA, 3))
  out <- marker_summary(df)
  expect_equal(out$mean, 2)
  expect_false(is.na(out$mean))
})

test_that("marker_summary errors on non-data.frame input", {
  skip_on_cran()
  expect_error(marker_summary(1:5), "must be a data.frame or tibble")
})

test_that("marker_summary verbose=TRUE runs without error", {
  skip_on_cran()
  df <- data.frame(a = 1:3)
  expect_no_error(marker_summary(df, verbose = TRUE))
})

#' Calculate lipid-panel markers, Visceral Adiposity Index (VAI),
#' Lipid Accumulation Product (LAP), and TyG-BMI index
#'
#' Given total cholesterol, HDL, TG (and optionally LDL, ApoB/ApoA1,
#' waist, BMI, glucose), computes:
#' - `non_HDL_c`, `remnant_c`
#' - `ratio_TC_HDL`, `ratio_TG_HDL`, `ratio_LDL_HDL`
#' - `ApoB_ApoA1`
#' - `VAI_Men`, `VAI_Women`
#' - `LAP_Men`, `LAP_Women`
#' - `TyG_BMI`
#'
#' Assumed units (no automatic conversion except where noted):
#' - TC, HDL_c, TG, LDL_c: mmol/L
#' - glucose: mmol/L (converted to mg/dL internally for TyG_BMI)
#' - waist: cm
#' - BMI: kg/m^2
#'
#' @param data A `data.frame` or `tibble` containing your lipid
#'   (and optional anthropometry/glucose) data.
#' @param col_map Named list mapping:
#'   - `TC`    -> total cholesterol
#'   - `HDL_c` -> HDL-C
#'   - `TG`    -> triglycerides
#'   - `LDL_c` -> (optional) LDL-C; if absent, estimated via Friedewald
#'   - `ApoB`, `ApoA1` -> (optional) apolipoproteins
#'   - `waist` -> (optional) waist circumference (cm)
#'   - `BMI`   -> (optional) body mass index (kg/m^2)
#'   - `glucose` -> (optional) fasting glucose (mmol/L); used for TyG_BMI
#' @param na_action One of `c("keep","omit","error","ignore","warn")`.
#'   - keep/ignore: compute and propagate NA in outputs
#'   - omit: drop rows with NA in required inputs (TC, HDL_c, TG)
#'   - error: abort if any required input contains NA
#'   - warn: like keep, but emit missingness warnings
#' @param na_warn_prop Proportion (0-1) threshold for high-missingness
#'   warnings when `na_action = "warn"`. Default 0.2.
#' @param verbose Logical; if `TRUE` (default), prints step-by-step progress
#'   including column mapping, optional input availability, pre-computation
#'   notes, physiological range information (informational only, values are
#'   not altered), the list of markers being computed with their inputs, and a
#'   per-column results summary.
#'
#' @return A tibble with computed lipid markers.
#'   Required outputs (always present):
#'   `non_HDL_c`, `remnant_c`, `ratio_TC_HDL`, `ratio_TG_HDL`,
#'   `ratio_LDL_HDL`, `ApoB_ApoA1`.
#'   Optional outputs (present when inputs available):
#'   `VAI_Men`, `VAI_Women` (waist + BMI required);
#'   `LAP_Men`, `LAP_Women` (waist required);
#'   `TyG_BMI` (BMI + glucose required).
#'   If an ID column is detected in `data` (e.g. `id`, `IID`,
#'   `participant_id`), it is prepended as the first output column.
#'
#' @details
#' Pre-computation (one level deep):
#' - If `BMI` is absent but `weight` (kg) and `height` (m or cm) are
#'   present, BMI is computed automatically.
#' - If `glucose` is absent but `G0` is present (or vice versa), the alias
#'   is derived automatically.
#' - If `LDL_c` is absent, it is always estimated via Friedewald
#'   (TC - HDL - TG/2.2, mmol/L form). An informational message is emitted when
#'   `verbose = TRUE`.
#'
#' @references
#' \insertRef{friedewald1972}{HealthMarkers}
#' \insertRef{amato2010vai}{HealthMarkers}
#' \insertRef{kahn2005lap}{HealthMarkers}
#' \insertRef{lee2016tyg}{HealthMarkers}
#' \insertRef{lee2020tygbmi}{HealthMarkers} (clinical application)
#' @importFrom dplyr bind_cols
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#'
#' @examples
#' df <- data.frame(TC = c(5.2, 6.1), HDL_c = c(1.3, 1.1), TG = c(1.8, 2.3),
#'                  LDL_c = c(3.2, 4.1), waist = c(88, 95), BMI = c(26, 29))
#' # Full verbose output (default)
#' lipid_markers(df)
#' # Suppress messaging for batch use
#' lipid_markers(df, verbose = FALSE)
#' # Pre-compute BMI from weight and height
#' df2 <- data.frame(TC = 5.2, HDL_c = 1.3, TG = 1.8, weight = 70, height = 175)
#' lipid_markers(df2, verbose = FALSE)
#' @export
lipid_markers <- function(
  data,
  col_map      = NULL,
  na_action    = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2,
  verbose      = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "lipid_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data))
    stop(sprintf("%s(): `data` must be a data.frame or tibble.", fn_name), call. = FALSE)

  # --- Detect and preserve ID column -----------------------------------------
  id_col <- .hm_detect_id_col(data)

  # --- Build col_map: infer missing keys, materialize aliases ----------------
  all_lm_keys <- c("TC", "HDL_c", "TG", "LDL_c", "ApoB", "ApoA1", "waist", "BMI", "glucose")
  req_keys    <- c("TC", "HDL_c", "TG")
  opt_keys    <- c("LDL_c", "ApoB", "ApoA1", "waist", "BMI", "glucose")

  cm      <- .hm_build_col_map(data, col_map, all_lm_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # --- Verbose: column mapping (user-provided vs inferred) -------------------
  .hm_log_cols(cm, col_map, fn_name, verbose, keys = all_lm_keys)

  # --- Pre-computation: fill still-missing required keys from raw inputs -----
  deps <- .lm_precompute_deps()
  is_missing_key <- function(k) is.null(col_map[[k]]) || !(col_map[[k]] %in% names(data))
  missing_req <- req_keys[vapply(req_keys, is_missing_key, logical(1))]
  if (length(missing_req) > 0L) {
    pc   <- .hm_precompute_from_deps(data, deps, missing_req, fn = fn_name, verbose = verbose)
    data <- pc$data
    for (k in missing_req) if (is.null(col_map[[k]]) && k %in% names(data)) col_map[[k]] <- k
    missing_req <- req_keys[vapply(req_keys, is_missing_key, logical(1))]
  }
  if (length(missing_req))
    stop(sprintf("%s(): missing required columns: %s",
                 fn_name, paste(missing_req, collapse = ", ")), call. = FALSE)

  req_cols <- unname(unlist(col_map[req_keys]))
  miss     <- setdiff(req_cols, names(data))
  if (length(miss))
    stop(sprintf("%s(): missing required columns: %s",
                 fn_name, paste(miss, collapse = ", ")), call. = FALSE)

  # --- Pre-computation: fill missing optional keys (BMI, glucose from G0) ----
  missing_opt <- opt_keys[vapply(opt_keys, is_missing_key, logical(1))]
  if (length(missing_opt) > 0L) {
    pc          <- .hm_precompute_from_deps(data, deps, missing_opt, fn = fn_name, verbose = verbose)
    newly_added <- setdiff(names(pc$data), names(data))
    if (length(newly_added)) {
      data <- pc$data
      for (k in newly_added) if (is.null(col_map[[k]]) && k %in% names(data)) col_map[[k]] <- k
    }
  }

  # --- Determine which optional keys are now present --------------------------
  present_opt   <- vapply(opt_keys, function(k) {
    nm <- col_map[[k]]; !is.null(nm) && nm %in% names(data)
  }, logical(1))
  used_opt_keys <- opt_keys[present_opt]
  missing_opt   <- setdiff(opt_keys, used_opt_keys)

  # --- Friedewald LDL if absent (always derivable: TC, HDL_c, TG required) ---
  if (!"LDL_c" %in% used_opt_keys) {
    data[["LDL_c_fried"]] <- data[[col_map$TC]] - data[[col_map$HDL_c]] - data[[col_map$TG]] / 2.2
    col_map[["LDL_c"]]    <- "LDL_c_fried"
    used_opt_keys         <- c(used_opt_keys, "LDL_c")
    missing_opt           <- setdiff(missing_opt, "LDL_c")
    if (isTRUE(verbose))
      hm_inform(
        sprintf("%s(): pre-computation: LDL_c estimated via Friedewald (TC - HDL - TG/2.2)", fn_name),
        level = "inform"
      )
  }

  used_keys <- c(req_keys, used_opt_keys)
  used_cols <- unname(unlist(col_map[used_keys]))

  # --- Verbose: optional inputs block -----------------------------------------
  if (isTRUE(verbose)) {
    idx_deps <- list(
      ApoB_ApoA1 = c("ApoB", "ApoA1"),
      VAI        = c("waist", "BMI"),
      LAP        = c("waist"),
      TyG_BMI    = c("BMI", "glucose")
    )
    na_indices <- character(0)
    for (idx in names(idx_deps)) {
      still_missing <- setdiff(idx_deps[[idx]], used_opt_keys)
      if (length(still_missing))
        na_indices <- c(na_indices,
          sprintf("  %s -> NA  [missing: %s]", idx, paste(still_missing, collapse = ", ")))
    }
    derivable_hints <- character(0)
    for (k in setdiff(missing_opt, "LDL_c")) {
      dep <- deps[[k]]
      if (!is.null(dep)) {
        prereqs_absent <- setdiff(dep$needs, names(data))
        if (length(prereqs_absent) == 0L) {
          derivable_hints <- c(derivable_hints,
            sprintf("  %s can be derived from: %s", k, paste(dep$needs, collapse = ", ")))
        } else {
          derivable_hints <- c(derivable_hints,
            sprintf("  %s: provide %s to enable", k, paste(dep$needs, collapse = ", ")))
        }
      }
    }
    shown_present <- setdiff(used_opt_keys, "LDL_c")
    shown_missing <- setdiff(missing_opt, "LDL_c")
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(shown_present))
      lines <- c(lines, sprintf("  present:  %s", paste(shown_present, collapse = ", ")))
    if (length(shown_missing))
      lines <- c(lines, sprintf("  missing:  %s", paste(shown_missing, collapse = ", ")))
    if (length(na_indices))
      lines <- c(lines, "  indices -> NA:", na_indices)
    if (length(derivable_hints))
      lines <- c(lines, "  derivable:", derivable_hints)
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # --- Coerce used columns to numeric -----------------------------------------
  for (cn in used_cols) {
    if (!is.numeric(data[[cn]])) {
      old_vals <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old_vals))
      introduced_na <- sum(is.na(data[[cn]]) & !is.na(old_vals))
      if (introduced_na > 0L)
        warning(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na),
                call. = FALSE)
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # --- NA handling ------------------------------------------------------------
  nonfin_mask <- vapply(used_cols, function(cn) any(!is.finite(data[[cn]])), logical(1))
  if (na_action == "warn" && any(nonfin_mask))
    warning(sprintf("Non-finite values in: %s; treated as NA.",
                    paste(used_cols[nonfin_mask], collapse = ", ")), call. = FALSE)
  if (na_action == "error" && any(!stats::complete.cases(data[, req_cols, drop = FALSE])))
    stop(sprintf("%s(): missing or non-finite values in required inputs with na_action='error'.",
                 fn_name), call. = FALSE)

  keep_rows <- if (na_action == "omit")
    stats::complete.cases(data[, req_cols, drop = FALSE]) else rep(TRUE, nrow(data))
  d <- data[keep_rows, , drop = FALSE]

  # --- Helper: is key available in d? ----------------------------------------
  has <- function(k) !is.null(col_map[[k]]) && col_map[[k]] %in% names(d)

  # --- Verbose: physiological range check (informational, not altered) --------
  if (isTRUE(verbose)) {
    rules    <- .lm_default_extreme_rules()
    remapped <- list()
    for (nm in names(rules)) {
      cn <- if (!is.null(col_map[[nm]])) col_map[[nm]] else nm
      remapped[[cn]] <- rules[[nm]]
    }
    scan_vars <- intersect(used_cols, names(remapped))
    ex_flags  <- list(); total <- 0L
    for (v in scan_vars) {
      x <- d[[v]]; rng <- remapped[[v]]
      bad <- is.finite(x) & (x < rng[1] | x > rng[2])
      ex_flags[[v]] <- bad; total <- total + sum(bad)
    }
    if (total > 0L) {
      details <- vapply(names(ex_flags), function(v) {
        nb <- sum(ex_flags[[v]])
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", v, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: computing markers list ----------------------------------------
  if (isTRUE(verbose)) {
    status <- c(
      "  non_HDL_c     [TC, HDL_c]",
      "  remnant_c     [TC, HDL_c, LDL_c]",
      "  ratio_TC_HDL  [TC, HDL_c]",
      "  ratio_TG_HDL  [TG, HDL_c]",
      "  ratio_LDL_HDL [LDL_c, HDL_c]",
      sprintf("  ApoB_ApoA1    %s",
              if (has("ApoB") && has("ApoA1")) "[ApoB, ApoA1]"
              else "NA [ApoB/ApoA1 missing]"),
      sprintf("  VAI_Men/Women %s",
              if (has("waist") && has("BMI")) "[waist, BMI, TG, HDL_c]"
              else "NA [waist/BMI missing]"),
      sprintf("  LAP_Men/Women %s",
              if (has("waist")) "[waist, TG]"
              else "NA [waist missing]"),
      sprintf("  TyG_BMI       %s",
              if (has("BMI") && has("glucose")) "[TG, glucose, BMI]"
              else "NA [BMI/glucose missing]")
    )
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # --- Compute markers --------------------------------------------------------
  sdiv <- function(a, b) { z <- a / b; z[!is.finite(z)] <- NA_real_; z }
  lg   <- function(x) { y <- log(x); y[!is.finite(y)] <- NA_real_; y }

  TC  <- d[[col_map$TC]]
  HDL <- d[[col_map$HDL_c]]
  TG  <- d[[col_map$TG]]
  LDL <- d[[col_map$LDL_c]]

  ApoB_ApoA1 <- if (has("ApoB") && has("ApoA1")) {
    sdiv(d[[col_map$ApoB]], d[[col_map$ApoA1]])
  } else {
    rep(NA_real_, nrow(d))
  }

  out <- tibble::tibble(
    non_HDL_c     = TC - HDL,
    remnant_c     = TC - (HDL + LDL),
    ratio_TC_HDL  = sdiv(TC, HDL),
    ratio_TG_HDL  = sdiv(TG, HDL),
    ratio_LDL_HDL = sdiv(LDL, HDL),
    ApoB_ApoA1    = ApoB_ApoA1
  )

  if (has("waist") && has("BMI")) {
    W  <- d[[col_map$waist]]
    BM <- d[[col_map$BMI]]
    out$VAI_Men   <- (W / (39.68 + 1.88 * BM)) * sdiv(TG, 1.03) * sdiv(1.31, HDL)
    out$VAI_Women <- (W / (36.58 + 1.89 * BM)) * sdiv(TG, 0.81) * sdiv(1.52, HDL)
  }

  if (has("waist")) {
    W <- d[[col_map$waist]]
    out$LAP_Men   <- (W - 65) * TG
    out$LAP_Women <- (W - 58) * TG
  }

  if (has("BMI") && has("glucose")) {
    TG_mgdl  <- TG * 88.57
    Glu_mgdl <- d[[col_map$glucose]] * 18
    BM       <- d[[col_map$BMI]]
    out$TyG_BMI <- lg(TG_mgdl * Glu_mgdl / 2) * BM
  }

  # --- Prepend ID column if detected ------------------------------------------
  if (!is.null(id_col)) {
    id_vec        <- d[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary -----------------------------------------------
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# ---- internal helpers --------------------------------------------------------

.lm_precompute_deps <- function() {
  list(
    BMI = list(
      needs    = c("weight", "height"),
      describe = "weight_kg / height_m^2",
      compute  = function(data) {
        w   <- as.numeric(data[["weight"]])
        h   <- as.numeric(data[["height"]])
        h_m <- ifelse(is.finite(h) & h > 3, h / 100, h)
        w / (h_m ^ 2)
      }
    ),
    glucose = list(
      needs    = "G0",
      describe = "alias: fasting glucose (G0)",
      compute  = function(data) as.numeric(data[["G0"]])
    ),
    G0 = list(
      needs    = "glucose",
      describe = "alias: fasting glucose (glucose)",
      compute  = function(data) as.numeric(data[["glucose"]])
    )
  )
}

.lm_default_extreme_rules <- function() {
  list(
    TC      = c(0, 50),
    HDL_c   = c(0, 10),
    TG      = c(0, 50),
    LDL_c   = c(0, 50),
    ApoB    = c(0, 10),
    ApoA1   = c(0, 10),
    waist   = c(30, 250),
    BMI     = c(10, 80),
    glucose = c(0, 50)
  )
}

test_that("lipid_markers computes core lipid markers", {
  skip_on_cran()
  df <- tibble(
    TC    = 5,
    HDL_c = 1,
    TG    = 1.3,
    LDL_c = 3,
    ApoB  = 1.1,
    ApoA1 = 1.5
  )

  out <- lipid_markers(df, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG",
    LDL_c = "LDL_c",
    ApoB  = "ApoB",
    ApoA1 = "ApoA1"
  ), verbose = FALSE)

  expect_named(out, c(
    "non_HDL_c", "remnant_c", "ratio_TC_HDL",
    "ratio_TG_HDL", "ratio_LDL_HDL", "ApoB_ApoA1"
  ))
  expect_equal(out$non_HDL_c, 5 - 1)
  expect_equal(out$remnant_c, 5 - (1 + 3))
  expect_equal(out$ratio_TC_HDL, 5 / 1)
  expect_equal(out$ratio_TG_HDL, 1.3 / 1)
  expect_equal(out$ratio_LDL_HDL, 3 / 1)
  expect_equal(out$ApoB_ApoA1, 1.1 / 1.5)
})

test_that("lipid_markers estimates LDL via Friedewald when LDL_c absent, no warning", {
  skip_on_cran()
  df2 <- tibble(TC = 5, HDL_c = 1, TG = 1.5)
  expect_no_warning(
    out2 <- lipid_markers(df2, col_map = list(TC = "TC", HDL_c = "HDL_c", TG = "TG"), verbose = FALSE)
  )
  # LDL (mmol/L Friedewald) = 5 - 1 - (1.5/2.2)
  expect_equal(out2$ratio_LDL_HDL, (5 - 1 - (1.5 / 2.2)) / 1)
})

test_that("lipid_markers emits informational message for Friedewald LDL", {
  skip_on_cran()
  df2 <- tibble(TC = 5, HDL_c = 1, TG = 1.5)
  msgs <- testthat::capture_messages(
    lipid_markers(df2, col_map = list(TC = "TC", HDL_c = "HDL_c", TG = "TG"), verbose = TRUE)
  )
  expect_true(any(grepl("Friedewald", msgs)))
})

test_that("lipid_markers computes VAI_Men and VAI_Women when waist & BMI provided", {
  skip_on_cran()
  df3 <- tibble(
    TC = 5, HDL_c = 1, TG = 1.3,
    LDL_c = 3, ApoB = 1.1, ApoA1 = 1.5,
    waist = 85, BMI = 26
  )
  out3 <- lipid_markers(df3, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG",
    LDL_c = "LDL_c",
    ApoB  = "ApoB",
    ApoA1 = "ApoA1",
    waist = "waist",
    BMI   = "BMI"
  ), verbose = FALSE)
  expect_true(all(c("VAI_Men", "VAI_Women") %in% names(out3)))
  exp_men <- (85 / (39.68 + 1.88 * 26)) * (1.3 / 1.03) * (1.31 / 1)
  exp_wom <- (85 / (36.58 + 1.89 * 26)) * (1.3 / 0.81) * (1.52 / 1)
  expect_equal(out3$VAI_Men, exp_men, tolerance = 1e-8)
  expect_equal(out3$VAI_Women, exp_wom, tolerance = 1e-8)
})

test_that("lipid_markers omits VAI when waist or BMI missing", {
  skip_on_cran()
  df4a <- tibble(TC = 5, HDL_c = 1, TG = 1.3, waist = 85) # BMI missing
  out4a <- lipid_markers(df4a, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG",
    waist = "waist"
  ), verbose = FALSE)
  expect_false("VAI_Men" %in% names(out4a))
  expect_false("VAI_Women" %in% names(out4a))

  df4b <- tibble(TC = 5, HDL_c = 1, TG = 1.3, BMI = 26) # waist missing
  out4b <- lipid_markers(df4b, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG",
    BMI   = "BMI"
  ), verbose = FALSE)
  expect_false("VAI_Men" %in% names(out4b))
  expect_false("VAI_Women" %in% names(out4b))
})

# ────────────────────────────────────────────────────────────────────────────────
# LAP tests
# ────────────────────────────────────────────────────────────────────────────────

test_that("lipid_markers computes LAP_Men and LAP_Women when waist provided", {
  skip_on_cran()
  df_lap <- tibble(TC = 5, HDL_c = 1, TG = 1.3, waist = 100)
  out_lap <- lipid_markers(df_lap, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG",
    waist = "waist"
  ), verbose = FALSE)
  expect_true(all(c("LAP_Men", "LAP_Women") %in% names(out_lap)))

  exp_lap_men <- (100 - 65) * 1.3
  exp_lap_wom <- (100 - 58) * 1.3
  expect_equal(out_lap$LAP_Men, exp_lap_men)
  expect_equal(out_lap$LAP_Women, exp_lap_wom)
})

test_that("lipid_markers omits LAP when waist missing", {
  skip_on_cran()
  df_no_lap <- tibble(TC = 5, HDL_c = 1, TG = 1.3)
  out_no_lap <- lipid_markers(df_no_lap, col_map = list(
    TC    = "TC",
    HDL_c = "HDL_c",
    TG    = "TG"
  ), verbose = FALSE)
  expect_false("LAP_Men" %in% names(out_no_lap))
  expect_false("LAP_Women" %in% names(out_no_lap))
})

# ────────────────────────────────────────────────────────────────────────────────
# HM-CS v2 additions: NA policy, extremes, and coercion
# ────────────────────────────────────────────────────────────────────────────────

test_that("na_action='omit' drops rows with NA in required inputs", {
  skip_on_cran()
  df <- tibble(TC = c(5, NA), HDL_c = c(1, 1), TG = c(1.3, 1.4), LDL_c = c(3.7, 3.8))
  out <- lipid_markers(df, col_map = list(TC="TC", HDL_c="HDL_c", TG="TG", LDL_c="LDL_c"), na_action = "omit", verbose = FALSE)
  expect_equal(nrow(out), 1L)
  expect_equal(out$ratio_TC_HDL, 5/1)
})

test_that("na_action='error' aborts on NA in required inputs", {
  skip_on_cran()
  df <- tibble(TC = c(5, NA), HDL_c = c(1, 1), TG = c(1.3, 1.4))
  expect_error(
    lipid_markers(df, col_map = list(TC="TC", HDL_c="HDL_c", TG="TG"), na_action = "error", verbose = FALSE),
    "missing or non-finite"
  )
})

test_that("extreme values produce no warning/error; range note appears in verbose", {
  skip_on_cran()
  df <- tibble(TC = 5, HDL_c = 1, TG = -3, LDL_c = 3) # negative TG out-of-range
  expect_no_warning(
    out <- lipid_markers(df, col_map = list(TC="TC", HDL_c="HDL_c", TG="TG", LDL_c="LDL_c"),
                         verbose = FALSE)
  )
  # TG = -3 propagates; no alteration
  expect_false(is.na(out$ratio_TG_HDL))
  # Range note appears in verbose output
  msgs <- testthat::capture_messages(
    lipid_markers(df, col_map = list(TC="TC", HDL_c="HDL_c", TG="TG", LDL_c="LDL_c"),
                  verbose = TRUE)
  )
  expect_true(any(grepl("range note", msgs)))
})

test_that("numeric coercion warns when NAs introduced", {
  skip_on_cran()
  df <- tibble(TC = c("5","oops"), HDL_c = c("1","1"), TG = c("1.3","1.4"), LDL_c = c(3, 3))
  expect_warning(
    lipid_markers(df, col_map = list(TC="TC", HDL_c="HDL_c", TG="TG", LDL_c="LDL_c"), verbose = FALSE),
    "coerced to numeric; NAs introduced"
  )
})

test_that("verbose emits col_map, optional inputs, computing markers, and results messages", {
  skip_on_cran()
  df_v <- tibble::tibble(TC = 5, HDL_c = 1, TG = 1.3, LDL_c = 3)
  cm_v <- list(TC = "TC", HDL_c = "HDL_c", TG = "TG", LDL_c = "LDL_c")
  msgs <- testthat::capture_messages(lipid_markers(df_v, cm_v, verbose = TRUE))
  expect_true(any(grepl("col_map", msgs)))
  expect_true(any(grepl("optional inputs", msgs)))
  expect_true(any(grepl("computing markers", msgs)))
  expect_true(any(grepl("results:", msgs)))
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df_v <- tibble::tibble(TC = 5, HDL_c = 1, TG = 1.3, LDL_c = 3)
  cm_v <- list(TC = "TC", HDL_c = "HDL_c", TG = "TG", LDL_c = "LDL_c")
  msgs <- testthat::capture_messages(
    lipid_markers(df_v, cm_v, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",  msgs)), 1L)
  expect_equal(sum(grepl("optional inputs", msgs)), 1L)
  expect_equal(sum(grepl("computing markers", msgs)), 1L)
})

test_that("ID column is prepended to output when detected", {
  skip_on_cran()
  df_id <- tibble::tibble(
    id = 1:3, TC = c(5, 5.5, 6), HDL_c = c(1, 1.1, 1.2),
    TG = c(1.3, 1.4, 1.5), LDL_c = c(3, 3.2, 3.4)
  )
  out <- lipid_markers(df_id, verbose = FALSE)
  expect_equal(names(out)[1L], "id")
  expect_equal(out$id, 1:3)
})

test_that("BMI is pre-computed from weight and height enabling VAI and TyG_BMI", {
  skip_on_cran()
  df_bmi <- tibble::tibble(
    TC = 5, HDL_c = 1, TG = 1.3, LDL_c = 3,
    waist = 85, weight = 70, height = 175, glucose = 5.5
  )
  out <- lipid_markers(df_bmi, verbose = FALSE)
  bmi_expected <- 70 / (1.75) ^ 2
  vai_men_expected <- (85 / (39.68 + 1.88 * bmi_expected)) * (1.3 / 1.03) * (1.31 / 1)
  expect_true("VAI_Men" %in% names(out))
  expect_equal(out$VAI_Men, vai_men_expected, tolerance = 1e-6)
  expect_true("TyG_BMI" %in% names(out))
})

test_that("glucose derived from mapped G0 column via col_map redirect enables TyG_BMI", {
  skip_on_cran()
  # Physical column is "pglu0" — col_map maps G0 -> "pglu0".
  # .hm_build_col_map materializes data[["G0"]], then precompute derives glucose.
  df <- tibble::tibble(
    TC    = 5,
    HDL_c = 1,
    TG    = 1.3,
    BMI   = 25,
    waist = 85,
    pglu0 = 5.5    # non-standard name for fasting glucose
  )
  cm  <- list(TC = "TC", HDL_c = "HDL_c", TG = "TG", BMI = "BMI",
              waist = "waist", G0 = "pglu0")
  out <- lipid_markers(df, col_map = cm, verbose = FALSE)
  expect_true("TyG_BMI" %in% names(out))
  expect_false(is.na(out$TyG_BMI), info = "TyG_BMI requires glucose derived from mapped G0")
})

test_that("BMI derived from mapped weight/height with non-standard names enables VAI", {
  skip_on_cran()
  # Physical columns are "body_weight" and "body_height" — must be resolved via col_map.
  # .hm_build_col_map materializes data[["weight"]] and data[["height"]],
  # then .hm_precompute_from_deps derives BMI, enabling VAI_Men.
  df <- tibble::tibble(
    TC          = 5,
    HDL_c       = 1,
    TG          = 1.3,
    LDL_c       = 3,
    waist       = 85,
    body_weight = 70,
    body_height = 175,
    glucose     = 5.5
  )
  cm  <- list(TC = "TC", HDL_c = "HDL_c", TG = "TG", LDL_c = "LDL_c",
              waist = "waist", glucose = "glucose",
              weight = "body_weight", height = "body_height")
  out <- lipid_markers(df, col_map = cm, verbose = FALSE)
  bmi_expected     <- 70 / (1.75)^2
  vai_men_expected <- (85 / (39.68 + 1.88 * bmi_expected)) * (1.3 / 1.03) * (1.31 / 1)
  expect_true("VAI_Men" %in% names(out))
  expect_equal(out$VAI_Men, vai_men_expected, tolerance = 1e-6,
               info = "VAI_Men should use BMI precomputed from mapped weight/height")
})

test_that("partial col_map is filled from dictionary for remaining keys", {
  skip_on_cran()
  # User provides only G0 mapping; TC, HDL_c, TG should be inferred by dict.
  df <- tibble::tibble(
    chol   = 5,      # maps to TC via synonym dictionary
    hdlc   = 1,
    trig   = 1.3,
    pglu0  = 5.5
  )
  cm  <- list(G0 = "pglu0")
  # hm_col_report should match chol->TC, hdlc->HDL_c, trig->TG
  out <- lipid_markers(df, col_map = cm, verbose = FALSE)
  # If dict inferred TC/HDL_c/TG the result has all expected cols
  expect_true("non_HDL_c" %in% names(out))
})

#' Calculate sweat-based ionic & metabolic markers
#'
#' Computes:
#'  - sweat_chloride (mmol/L)
#'  - Na_K_ratio     (sweat Na+/K+)
#'  - sweat_lactate  (mmol/L)
#'  - sweat_rate     (L/m^2/h) from body mass loss per hour per m^2
#'
#' Inputs are validated, missingness handled via `na_action`, safe divisions
#' are used to avoid Inf/NaN, and an optional extremes scan/cap is available.
#'
#' Expected units:
#' - sweat_chloride, sweat_Na, sweat_K: mmol/L
#' - sweat_lactate: mmol/L
#' - weight_before, weight_after: kg
#' - duration: hours
#' - body_surface_area: m^2
#'
#' @param data A data.frame or tibble containing sweat assay and anthropometrics.
#' @param col_map Named list mapping required inputs (defaults assume same names):
#'   - sweat_chloride, sweat_Na, sweat_K, sweat_lactate,
#'     weight_before, weight_after, duration, body_surface_area
#' @param verbose Logical; if `TRUE`, prints progress messages and a completion summary.
#' @param na_action One of `c("keep","omit","error")` for handling missing values in required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness diagnostics for required inputs (debug level). Default 0.2.
#'
#' @return A tibble with columns:
#'   `sweat_chloride`, `Na_K_ratio`, `sweat_lactate`, `sweat_rate`
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
#' @examples
#' df <- tibble::tibble(
#'   sweat_chloride    = 45,
#'   sweat_Na          = 55,
#'   sweat_K           = 5,
#'   sweat_lactate     = 4.8,
#'   weight_before     = 70.0,
#'   weight_after      = 69.5,
#'   duration          = 1.0,
#'   body_surface_area = 1.9
#' )
#' sweat_markers(df)
#'
#' @note
#' `sweat_chloride` and `sweat_lactate` are **pass-through** columns; no formula
#' is applied. `Na_K_ratio` is a simple Na/K division. `sweat_rate` uses the
#' mass-loss method: (weight_before - weight_after) / duration / body_surface_area
#' (units: L/m\ifelse{html}{\out{&sup2;}}{^2}/h; assumes 1 kg \eqn{\approx} 1 L).
#' Dill & Costill 1974 describes haematocrit-based blood-volume change, not sweat
#' rate directly; cited here as background context only.
#'
#' @references
#' \insertRef{gibson1959sweat}{HealthMarkers} (pilocarpine sweat chloride test origin; background)
#' \insertRef{dill1974dehydration}{HealthMarkers} (dehydration and fluid loss context; background)
#' \insertRef{farrell2017cf}{HealthMarkers} (CF diagnostic sweat chloride cutoffs; background)
#' \insertRef{sawka2015hypo}{HealthMarkers} (sweat rate and hypohydration context; background)
sweat_markers <- function(data,
                          col_map = NULL,
                          verbose = TRUE,
                          na_action = c("keep","omit","error"),
                          na_warn_prop = 0.2) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "sweat_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data)) {
    rlang::abort("sweat_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_sweat_error_data_type")
  }

  # HM-CS v2: standardized validation
  required_keys <- c(
    "sweat_chloride", "sweat_Na", "sweat_K", "sweat_lactate",
    "weight_before", "weight_after", "duration", "body_surface_area"
  )
  hm_validate_inputs(data, col_map, required_keys = character(0), fn = "sweat_markers")
  cm      <- .hm_build_col_map(data, col_map, keys = required_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Ensure mapped columns exist — detect both unmapped keys and mapped-but-absent columns
  missing_cols <- required_keys[!vapply(required_keys, function(k) {
    !is.null(col_map[[k]]) && isTRUE(col_map[[k]] %in% names(data))
  }, logical(1))]
  if (length(missing_cols)) {
    rlang::abort(
      paste0("sweat_markers(): missing columns: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_sweat_error_missing_columns"
    )
  }
  req_cols <- unname(vapply(required_keys, function(k) col_map[[k]], character(1)))

  # Coerce to numeric where needed; warn on NAs introduced; non-finite -> NA
  for (cn in req_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("sweat_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug)
  for (cn in req_cols) {
    x <- data[[cn]]
    n <- length(x); if (!n) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("sweat_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  sweat_chloride [sweat_chloride]\n  Na_K_ratio [sweat_Na, sweat_K]\n  sweat_lactate [sweat_lactate]\n  sweat_rate [weight_before, weight_after, duration, body_surface_area]", fn_name), level = "inform")

  # NA policy
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(req_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("sweat_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_sweat_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(req_cols, function(cn) is.na(data[[cn]])))
    if (sum(!keep) > 0L)
      hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
                msg   = sprintf("sweat_markers(): omitting %d rows with NA in required inputs", sum(!keep)))
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      sweat_chloride = numeric(),
      Na_K_ratio     = numeric(),
      sweat_lactate  = numeric(),
      sweat_rate     = numeric()
    ))
  }

  hm_inform("sweat_markers(): computing markers", level = "debug")

  # Safe division helper with consolidated zero-denominator tracking
  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }

  # Compute metrics
  # Map columns from col_map
  sc  <- data[[col_map$sweat_chloride]]
  sNa <- data[[col_map$sweat_Na]]
  sK  <- data[[col_map$sweat_K]]
  sla <- data[[col_map$sweat_lactate]]
  wb  <- data[[col_map$weight_before]]
  wa  <- data[[col_map$weight_after]]
  dur <- data[[col_map$duration]]
  bsa <- data[[col_map$body_surface_area]]

  Na_K_ratio <- safe_div(sNa, sK, "Na_K_ratio")

  mass_loss_kg <- wb - wa
  rate_kg_per_h <- safe_div(mass_loss_kg, dur, "sweat_rate_duration")
  sweat_rate <- safe_div(rate_kg_per_h, bsa, "sweat_rate_bsa")
  # 1 kg ~ 1 L water; units become L/m^2/h

  out <- tibble::tibble(
    sweat_chloride = as.numeric(sc),
    Na_K_ratio     = as.numeric(Na_K_ratio),
    sweat_lactate  = as.numeric(sla),
    sweat_rate     = as.numeric(sweat_rate)
  )

  # Consolidated zero-denominator warning
  dz <- dz_env$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz); nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("sweat_markers(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  if (!is.null(id_col)) {
    id_vec <- data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# tests/testthat/test_sweat_markers.R
library(testthat)
library(tibble)

test_that("sweat_markers errors if missing required columns", {
  skip_on_cran()
  # Missing sweat_chloride
  df1 <- tibble(
    sweat_Na          = 50,
    sweat_K           = 5,
    sweat_lactate     = 10,
    weight_before     = 70,
    weight_after      = 69.5,
    duration          = 1,
    body_surface_area = 1.8
  )
  expect_error(
    sweat_markers(df1),
    "missing columns: sweat_chloride"
  )

  # Missing multiple columns
  df2 <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    # sweat_K missing
    weight_before     = 70,
    weight_after      = 69.5
    # missing sweat_lactate, duration, body_surface_area
  )
  expect_error(
    sweat_markers(df2),
    "missing columns: .*sweat_K.*sweat_lactate.*duration.*body_surface_area"
  )
})

test_that("sweat_markers computes ionic and rate metrics correctly", {
  skip_on_cran()
  df <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    sweat_K           = 5,
    sweat_lactate     = 10,
    weight_before     = 70,
    weight_after      = 69.5,
    duration          = 1,
    body_surface_area = 1.8
  )
  out <- sweat_markers(df)

  expect_named(
    out,
    c("sweat_chloride", "Na_K_ratio", "sweat_lactate", "sweat_rate")
  )
  expect_equal(out$Na_K_ratio, 50 / 5)
  expect_equal(out$sweat_rate, 0.5 / 1.8)
  expect_equal(out$sweat_chloride, 30)
  expect_equal(out$sweat_lactate, 10)
})

test_that("sweat_markers is vectorized over multiple rows", {
  skip_on_cran()
  df <- tibble(
    sweat_chloride    = c(30, 40),
    sweat_Na          = c(50, 60),
    sweat_K           = c(5, 6),
    sweat_lactate     = c(10, 20),
    weight_before     = c(70, 80),
    weight_after      = c(69.5, 79),
    duration          = c(1, 2),
    body_surface_area = c(1.8, 2.0)
  )
  out <- sweat_markers(df)
  expect_equal(nrow(out), 2)
  expect_equal(out$Na_K_ratio[2], 60 / 6)
  expect_equal(out$sweat_rate[2], (80 - 79) / 2 / 2)
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  df <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    sweat_K           = 5,
    sweat_lactate     = 10,
    weight_before     = 70,
    weight_after      = 69.5,
    duration          = 1,
    body_surface_area = 1.8
  )
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(sweat_markers(df, verbose = TRUE), "sweat_markers")
  expect_message(sweat_markers(df, verbose = TRUE), "col_map")
  expect_message(sweat_markers(df, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    sweat_K           = 5,
    sweat_lactate     = 10,
    weight_before     = 70,
    weight_after      = 69.5,
    duration          = 1,
    body_surface_area = 1.8
  )
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(sweat_markers(df, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action policies: error and omit behave as expected", {
  skip_on_cran()
  df_na <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    sweat_K           = 5,
    sweat_lactate     = 10,
    weight_before     = NA_real_,  # NA in required input
    weight_after      = 69.5,
    duration          = 1,
    body_surface_area = 1.8
  )
  expect_error(
    suppressWarnings(sweat_markers(df_na, na_action = "error")),
    "required inputs contain missing values"
  )
  out_omit <- suppressWarnings(sweat_markers(df_na, na_action = "omit"))
  expect_equal(nrow(out_omit), 0L)
})

test_that("check_extreme removed: function passes through outlier rows", {
  skip_on_cran()
  df_ext <- tibble(
    sweat_chloride    = 300,
    sweat_Na          = 300,
    sweat_K           = 50,
    sweat_lactate     = 60,
    weight_before     = 500,
    weight_after      = 600,
    duration          = 0.01,
    body_surface_area = 0.1
  )
  out <- sweat_markers(df_ext)
  expect_equal(nrow(out), 1L)
})

test_that("zero denominators emit a consolidated warning and yield NA in ratios", {
  skip_on_cran()
  df_zero <- tibble(
    sweat_chloride    = 30,
    sweat_Na          = 50,
    sweat_K           = 0,    # zero -> Na/K undefined
    sweat_lactate     = 10,
    weight_before     = 70,
    weight_after      = 69.5,
    duration          = 0,    # zero -> rate undefined
    body_surface_area = 1.8
  )
  expect_warning(
    out_zero <- sweat_markers(df_zero),
    "zero denominators detected"
  )
  expect_true(is.na(out_zero$Na_K_ratio))
  expect_true(is.na(out_zero$sweat_rate))
})

#' Calculate saliva-based stress & glycemic markers
#'
#' Computes:
#'  - log_cortisol_wake (log-transformed waking cortisol)
#'  - CAR_AUC           (Cortisol Awakening Response, trapezoidal AUC over 0-60 min by default)
#'  - log_amylase       (log-transformed salivary alpha-amylase)
#'  - saliva_glucose    (raw salivary glucose)
#'
#' Inputs are validated, missingness handled via `na_action`, logs made safe
#' (<= 0 -> NA), and optional extremes scan/cap is available.
#'
#' @param data A data.frame or tibble with salivary markers.
#' @param col_map Named list mapping required inputs. Defaults assume same names:
#'   - cort1   -> "saliva_cort1" (nmol/L at wake)
#'   - cort2   -> "saliva_cort2" (nmol/L ~30 min)
#'   - cort3   -> "saliva_cort3" (nmol/L ~60 min)
#'   - amylase -> "saliva_amylase" (U/mL)
#'   - glucose -> "saliva_glucose" (mg/dL)
#' @param verbose Logical; if `TRUE` (default), prints column mapping, input
#'   availability, physiological range information (informational only, values
#'   not altered), the list of markers being computed with their inputs, and a
#'   per-column results summary.
#' @param na_action One of `c("keep","omit","error")` for required-input NA handling. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness diagnostics (debug). Default 0.2.
#' @param times Numeric vector of sampling times (minutes) for CAR AUC. Must align with cort1/2/3. Default c(0,30,60).
#'
#' @return A tibble with columns:
#'   - `log_cortisol_wake`
#'   - `CAR_AUC`
#'   - `log_amylase`
#'   - `saliva_glucose`
#'
#'   If an ID column is detected in `data` (e.g. `id`, `IID`, `participant_id`),
#'   it is prepended as the first output column.
#'
#' @examples
#' df <- tibble::tibble(
#'   saliva_cort1    = 12.5,
#'   saliva_cort2    = 18.0,
#'   saliva_cort3    = 16.2,
#'   saliva_amylase  = 85,
#'   saliva_glucose  = 4.2
#' )
#' saliva_markers(df)  # uses default col_map
#'
#' @note
#' `log_cortisol_wake` and `log_amylase` use the **natural logarithm** (`log()`).
#' `CAR_AUC` is the trapezoidal area under the cortisol-time curve (Pruessner
#' et al. 2003, AUC with respect to ground). `saliva_glucose` is a
#' **pass-through** column; no formula is applied.
#'
#' @references
#' \insertRef{pruessner2003auc}{HealthMarkers}
#' \insertRef{kirschbaum1994cortisol}{HealthMarkers} (salivary cortisol methods; background)
#' \insertRef{clow2004car}{HealthMarkers} (CAR methodological review; background)
#' \insertRef{nater2009amylase}{HealthMarkers} (salivary alpha-amylase SNS biomarker; background)
#' \insertRef{scales1987salgluc}{HealthMarkers} (salivary glucose application; pass-through, no formula)
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
saliva_markers <- function(data,
                           col_map      = NULL,
                           verbose      = TRUE,
                           na_action    = c("keep", "omit", "error"),
                           na_warn_prop = 0.2,
                           times        = c(0, 30, 60)) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "saliva_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  if (!is.data.frame(data)) {
    rlang::abort("saliva_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_saliva_error_data_type")
  }

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  # HM-CS v2: standardized validation
  required_keys <- c("cort1","cort2","cort3","amylase","glucose")
  cm      <- .hm_build_col_map(data, col_map, required_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  missing_keys <- setdiff(required_keys, names(col_map))
  if (length(missing_keys)) {
    # Graceful NA only when nothing provided and inference found nothing
    if (length(cm$user_keys) == 0L && length(cm$inferred_keys) == 0L && length(col_map) == 0L) {
      return(tibble::tibble(
        log_cortisol_wake = rep(NA_real_, nrow(data)),
        CAR_AUC           = rep(NA_real_, nrow(data)),
        log_amylase       = rep(NA_real_, nrow(data)),
        saliva_glucose    = rep(NA_real_, nrow(data))
      ))
    }
    rlang::abort(
      paste0("saliva_markers(): missing required col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_saliva_error_missing_map"
    )
  }
  hm_validate_inputs(data, col_map, required_keys = required_keys, fn = fn_name)

  # Ensure mapped columns exist
  mapped <- unname(unlist(col_map[required_keys], use.names = FALSE))
  missing_cols <- setdiff(mapped, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("saliva_markers(): missing columns: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_saliva_error_missing_columns"
    )
  }

  # Coerce to numeric where needed and warn on NAs introduced; set non-finite -> NA
  for (cn in mapped) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced <- sum(is.na(new) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("saliva_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug)
  for (cn in mapped) {
    x <- data[[cn]]; n <- length(x); if (!n) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("saliva_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }

  # --- Verbose: column mapping
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: optional inputs / availability
  if (isTRUE(verbose)) {
    avail_keys  <- required_keys[required_keys %in% names(col_map)]
    absent_keys <- setdiff(required_keys, avail_keys)
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(avail_keys))
      lines <- c(lines, sprintf("  present:  %s", paste(avail_keys, collapse = ", ")))
    if (length(absent_keys))
      lines <- c(lines, sprintf("  missing:  %s", paste(absent_keys, collapse = ", ")))
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # NA policy on required inputs
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(mapped, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("saliva_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_saliva_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(mapped, function(cn) is.na(data[[cn]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      log_cortisol_wake = numeric(),
      CAR_AUC           = numeric(),
      log_amylase       = numeric(),
      saliva_glucose    = numeric()
    ))
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    default_ranges <- list(
      cort1   = c(0, 2000), cort2 = c(0, 2000), cort3 = c(0, 2000),
      amylase = c(0, 50000), glucose = c(0, 1000)
    )
    flagged_details <- character(0)
    for (key in required_keys) {
      rng <- default_ranges[[key]]
      cn  <- col_map[[key]]
      if (is.null(rng) || is.null(cn) || !(cn %in% names(data))) next
      bad <- sum(is.finite(data[[cn]]) & (data[[cn]] < rng[1] | data[[cn]] > rng[2]), na.rm = TRUE)
      if (bad > 0L)
        flagged_details <- c(flagged_details,
          sprintf("  %s: %d value(s) outside plausible range", key, bad))
    }
    if (length(flagged_details))
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                flagged_details), collapse = "\n"),
        level = "inform"
      )
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    marker_deps <- list(
      log_cortisol_wake = c("cort1"),
      CAR_AUC           = c("cort1", "cort2", "cort3"),
      log_amylase       = c("amylase"),
      saliva_glucose    = c("glucose")
    )
    avail_keys2 <- names(col_map)
    status <- vapply(names(marker_deps), function(m) {
      miss_k <- setdiff(marker_deps[[m]], avail_keys2)
      if (length(miss_k) == 0L)
        sprintf("  %-20s [%s]", m, paste(marker_deps[[m]], collapse = ", "))
      else
        sprintf("  %-20s NA [missing: %s]", m, paste(miss_k, collapse = ", "))
    }, character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Helpers
  safe_log <- function(x) {
    y <- x
    y[!(is.finite(y) & y > 0)] <- NA_real_
    out <- suppressWarnings(log(y))
    out[!is.finite(out)] <- NA_real_
    out
  }

  # Extract mapped inputs
  c1 <- data[[col_map$cort1]]
  c2 <- data[[col_map$cort2]]
  c3 <- data[[col_map$cort3]]
  amy <- data[[col_map$amylase]]
  glu <- data[[col_map$glucose]]

  # Compute outputs
  log_cortisol_wake <- safe_log(c1)
  log_amylase       <- safe_log(amy)

  # Validate times
  if (!(is.numeric(times) && length(times) == 3L && isTRUE(all(diff(times) >= 0)))) {
    rlang::abort("saliva_markers(): `times` must be a numeric vector of length 3 in non-decreasing order.",
                 class = "healthmarkers_saliva_error_times")
  }
  cort_mat <- cbind(c1, c2, c3)
  dt <- diff(times)
  CAR_AUC <- apply(cort_mat, 1, function(x) {
    if (any(!is.finite(x))) return(NA_real_)
    sum((x[-length(x)] + x[-1]) / 2 * dt)
  })

  out <- tibble::tibble(
    log_cortisol_wake = as.numeric(log_cortisol_wake),
    CAR_AUC           = as.numeric(CAR_AUC),
    log_amylase       = as.numeric(log_amylase),
    saliva_glucose    = as.numeric(glu)
  )

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# tests/testthat/test_saliva_markers.R

library(testthat)
library(tibble)

test_that("saliva_markers errors if missing required columns", {
  skip_on_cran()
  df1 <- tibble(
    saliva_cort2    = 15,
    saliva_cort3    = 12,
    saliva_amylase  = 100,
    saliva_glucose  = 80
  )
  expect_error(
    saliva_markers(df1),
    "cort1"
  )

  df2 <- tibble(
    saliva_cort1    = 10,
    saliva_cort3    = 12,
    saliva_amylase  = 100
    # missing saliva_cort2 and saliva_glucose
  )
  expect_error(
    saliva_markers(df2),
    "cort2"
  )
})

test_that("saliva_markers computes log transforms, CAR_AUC, and passes through glucose", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = 10,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100,
    saliva_glucose  = 5.0
  )
  out <- saliva_markers(df)

  expect_named(
    out,
    c("log_cortisol_wake", "CAR_AUC", "log_amylase", "saliva_glucose")
  )
  expect_equal(nrow(out), 1)

  expected_log_cort <- log(10)
  # trapezoidal AUC: (10+20)/2*30 + (20+10)/2*30 = 900
  expected_auc <- (10 + 20) / 2 * 30 + (20 + 10) / 2 * 30
  expected_log_amy <- log(100)
  expected_gluc <- 5.0

  expect_equal(out$log_cortisol_wake, expected_log_cort)
  expect_equal(out$CAR_AUC, expected_auc)
  expect_equal(out$log_amylase, expected_log_amy)
  expect_equal(out$saliva_glucose, expected_gluc)
})

test_that("saliva_markers is vectorized over multiple rows", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = c(10, 5),
    saliva_cort2    = c(20, 10),
    saliva_cort3    = c(10, 5),
    saliva_amylase  = c(100, 50),
    saliva_glucose  = c(5, 3)
  )
  out <- saliva_markers(df)
  expect_equal(nrow(out), 2)
  expect_equal(out$log_cortisol_wake[2], log(5))
  # AUC second row: (5+10)/2*30 + (10+5)/2*30 = 450
  expect_equal(out$CAR_AUC[2], 450)
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = 10,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100,
    saliva_glucose  = 5.0
  )
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(saliva_markers(df, verbose = TRUE), "saliva_markers")
  expect_message(saliva_markers(df, verbose = TRUE), "col_map")
  expect_message(saliva_markers(df, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = 10,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100,
    saliva_glucose  = 5.0
  )
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(saliva_markers(df, verbose = TRUE))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("na_action='omit' drops rows with NA in required inputs and returns empty tibble when all omitted", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = NA_real_,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100,
    saliva_glucose  = 5.0
  )
  out <- saliva_markers(df, na_action = "omit")
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
  expect_named(out, c("log_cortisol_wake","CAR_AUC","log_amylase","saliva_glucose"))
})

test_that("invalid `times` argument errors clearly", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = 10,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100,
    saliva_glucose  = 5.0
  )
  expect_error(
    saliva_markers(df, times = c(0, 60)),
    "times"
  )
})

test_that("extreme values produce range note in verbose; no warning in non-verbose mode", {
  skip_on_cran()
  df <- tibble(
    saliva_cort1    = 10,
    saliva_cort2    = 20,
    saliva_cort3    = 10,
    saliva_amylase  = 100000,  # above plausible range of 50000
    saliva_glucose  = 5.0
  )
  # No warning emitted in non-verbose mode
  expect_no_warning(saliva_markers(df, verbose = FALSE))

  # Values are not altered; log_amylase uses original extreme value
  out <- saliva_markers(df, verbose = FALSE)
  expect_equal(out$log_amylase, log(100000), tolerance = 1e-12)

  # Verbose mode emits a range note informational message
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(saliva_markers(df, verbose = TRUE), "range note")
})

#' Normalize a numeric vector
#'
#' Utility used across HealthMarkers to normalize numeric vectors with several
#' common schemes while handling edge cases (constant vectors, all-NA, non-finite
#' values) robustly. NA positions are preserved.
#'
#' Methods:
#' - "none": return input as-is (no coercion; fully backward compatible).
#' - "z": z-score (mean 0, sd 1). Constant vectors return zeros (non-NA entries).
#' - "range": min-max to a target interval (default \eqn{[0,1]}). Constant vectors return
#'   the lower bound (mapped from zeros).
#' - "robust": median/MAD scaling. Constant vectors (MAD=0) return zeros.
#' - "inverse": rank-based inverse normal transform (normal scores).
#'
#' @param x A numeric (or numeric-coercible) vector.
#' @param method One of c("none","z","inverse","range","robust"). Default "none".
#' @param na_rm Logical; remove NAs when estimating statistics (mean, sd, etc.). Default TRUE.
#' @param feature_range Numeric length-2 vector giving the target range for method = "range".
#'   Default c(0, 1).
#' @param invnorm_denominator One of c("n","n+1","blom") controlling the denominator
#'   of the inverse-normal transform:
#'   - "n": p = (r - 0.5) / n  (Rankit; default)
#'   - "n+1": p = (r - 0.5) / (n + 1)
#'   - "blom": p = (r - 3/8) / (n + 1/4)  (Blom, 1958)
#' @param ties Ties method passed to base::rank for method = "inverse".
#'   One of c("average","first","last","random","max","min"). Default "average".
#' @param warn_constant Logical; if TRUE, warn when input is constant and a zero vector
#'   (or lower bound for range) is returned. Default TRUE.
#'
#' @return A numeric vector of the same length as x.
#' @export
#' @importFrom stats sd median qnorm mad
#' @importFrom stats na.omit setNames
#' @importFrom utils head modifyList
#'
#' @examples
#' x <- c(1, 2, 3, NA, 5)
#' normalize_vec(x, "none")
#' normalize_vec(x, "z")
#' normalize_vec(x, "range", feature_range = c(-1, 1))
#' normalize_vec(x, "robust")
#' normalize_vec(x, "inverse")               # Rankit (default)
#' normalize_vec(x, "inverse", invnorm_denominator = "blom")
#'
#' @references
#' \insertRef{beasley2009invnorm}{HealthMarkers}
#' \insertRef{leys2013mad}{HealthMarkers}
#' \insertRef{bland1996sd}{HealthMarkers}
normalize_vec <- function(
  x,
  method = c("none", "z", "inverse", "range", "robust"),
  na_rm = TRUE,
  feature_range = c(0, 1),
  invnorm_denominator = c("n", "n+1", "blom"),
  ties = c("average", "first", "last", "random", "max", "min"),
  warn_constant = TRUE
) {
  method <- match.arg(method)

  # Full backward compatibility: "none" returns input as-is (no coercion/cleanup)
  if (method == "none") {
    return(x)
  }

  invnorm_denominator <- match.arg(invnorm_denominator)
  ties <- match.arg(ties)

  # Coerce to numeric, preserve NA positions; treat non-finite as NA
  if (!is.numeric(x)) {
    old <- x
    suppressWarnings(x <- as.numeric(old))
    if (any(is.na(x) & !is.na(old))) {
      rlang::warn("normalize_vec(): input coerced to numeric; NAs introduced.",
                  class = "healthmarkers_normalize_warn_coercion")
    }
  }
  x[!is.finite(x)] <- NA_real_

  n <- length(x)
  out <- rep(NA_real_, n)

  if (all(is.na(x))) return(out)

  if (method == "z") {
    m <- mean(x, na.rm = na_rm)
    s <- stats::sd(x, na.rm = na_rm)
    if (!is.finite(s) || s == 0) {
      if (isTRUE(warn_constant)) {
        rlang::warn("normalize_vec(z): constant or degenerate input; returning zeros.",
                    class = "healthmarkers_normalize_warn_constant")
      }
      out <- ifelse(is.na(x), NA_real_, 0)
    } else {
      out <- (x - m) / s
    }
    return(out)
  }

  if (method == "range") {
    if (!(is.numeric(feature_range) && length(feature_range) == 2L && all(is.finite(feature_range)))) {
      rlang::abort("normalize_vec(range): `feature_range` must be a finite numeric vector of length 2.",
                   class = "healthmarkers_normalize_error_feature_range")
    }
    lo <- min(feature_range); hi <- max(feature_range)
    xmin <- min(x, na.rm = na_rm); xmax <- max(x, na.rm = na_rm)
    rng <- xmax - xmin
    if (!is.finite(rng) || rng == 0) {
      if (isTRUE(warn_constant)) {
        rlang::warn("normalize_vec(range): constant input; mapping to lower bound.",
                    class = "healthmarkers_normalize_warn_constant")
      }
      base <- ifelse(is.na(x), NA_real_, 0)  # maps to lo
    } else {
      base <- (x - xmin) / rng
    }
    out <- lo + (hi - lo) * base
    return(out)
  }

  if (method == "robust") {
    med <- stats::median(x, na.rm = na_rm)
    md  <- stats::mad(x, na.rm = na_rm)
    if (!is.finite(md) || md == 0) {
      if (isTRUE(warn_constant)) {
        rlang::warn("normalize_vec(robust): MAD is zero; returning zeros.",
                    class = "healthmarkers_normalize_warn_constant")
      }
      out <- ifelse(is.na(x), NA_real_, 0)
    } else {
      out <- (x - med) / md
    }
    return(out)
  }

  if (method == "inverse") {
    ok <- !is.na(x)
    n_ok <- sum(ok)
    if (n_ok == 0L) return(out)  # all NA
    r <- rank(x[ok], na.last = "keep", ties.method = ties)
    # Denominator/offset
    if (invnorm_denominator == "n") {
      p <- (r - 0.5) / n_ok
    } else if (invnorm_denominator == "n+1") {
      p <- (r - 0.5) / (n_ok + 1)
    } else { # "blom"
      p <- (r - 3/8) / (n_ok + 1/4)
    }
    # Clamp probabilities away from 0/1 to avoid +/-Inf
    p <- pmin(pmax(p, 1e-12), 1 - 1e-12)
    out[ok] <- stats::qnorm(p)
    return(out)
  }

  # Fallback (should not reach)
  return(x)
}

#' Normalise marker columns in a data frame
#'
#' A convenience wrapper around [normalize_vec()] that applies a normalisation
#' method to every selected numeric column in a data frame and returns the
#' modified data frame. The most common use-case is to normalise the output of
#' any HealthMarkers function — especially domain functions such as
#' `glycemic_markers()`, `lipid_markers()`, or `renal_markers()` whose internal
#' `normalize` argument currently has no effect.
#'
#' @param data A data frame (or tibble) containing marker columns to normalise.
#' @param cols Character vector of column names to normalise.  If `NULL`
#'   (default), **all numeric columns** in `data` are normalised.
#' @param method One of `c("z","inverse","range","robust")`.  Passed directly to
#'   [normalize_vec()]:
#'   \describe{
#'     \item{`"z"`}{z-score (mean 0, sd 1).}
#'     \item{`"inverse"`}{Rank-based inverse-normal transform (Rankit; default).}
#'     \item{`"range"`}{Min-max scaling to `feature_range` (default `[0, 1]`).}
#'     \item{`"robust"`}{Median/MAD scaling.}
#'   }
#' @param skip_cols Character vector of column names to leave untouched even if
#'   they are numeric (e.g. `c("age", "BMI")`).  Ignored when `cols` is
#'   explicitly supplied.
#' @param ... Additional arguments forwarded to [normalize_vec()] (e.g.
#'   `feature_range`, `invnorm_denominator`, `ties`).
#'
#' @return The input `data` with the selected columns replaced by their
#'   normalised values.  Class (tibble, data.frame, etc.) is preserved.
#'
#' @seealso [normalize_vec()] for single-vector normalisation.
#'
#' @export
#'
#' @examples
#' # Build a tiny data frame with pre-computed marker columns
#' df <- data.frame(
#'   age    = c(45, 52, 61),
#'   HOMA_IR = c(1.2, 3.4, 2.1),
#'   TyG     = c(8.1, 9.0, 8.6),
#'   NLR     = c(2.1, 3.5, 1.8)
#' )
#'
#' marker_cols <- c("HOMA_IR", "TyG", "NLR")
#'
#' # z-score normalise marker columns only
#' hm_normalize(df, cols = marker_cols, method = "z")
#'
#' # Inverse-normal transform, leaving age untouched
#' hm_normalize(df, method = "inverse", skip_cols = "age")
hm_normalize <- function(
  data,
  cols       = NULL,
  method     = c("z", "inverse", "range", "robust"),
  skip_cols  = NULL,
  ...
) {
  method <- match.arg(method)

  if (!is.data.frame(data)) {
    rlang::abort("`data` must be a data frame.", class = "healthmarkers_hm_normalize_error")
  }

  # Determine target columns
  if (!is.null(cols)) {
    bad <- setdiff(cols, names(data))
    if (length(bad)) {
      rlang::abort(
        paste0("Column(s) not found in `data`: ", paste(bad, collapse = ", ")),
        class = "healthmarkers_hm_normalize_error"
      )
    }
    target <- cols
  } else {
    target <- names(data)[vapply(data, is.numeric, logical(1))]
    if (!is.null(skip_cols)) {
      target <- setdiff(target, skip_cols)
    }
  }

  if (length(target) == 0L) {
    rlang::warn("hm_normalize(): no numeric columns to normalise; returning data unchanged.",
                class = "healthmarkers_hm_normalize_warn_no_cols")
    return(data)
  }

  data[target] <- lapply(data[target], normalize_vec, method = method, ...)
  data
}

#' Calculate Metabolic Syndrome Severity Score (MetSSS)
#'
#' @title Metabolic Syndrome Severity Score (MetSSS)
#'
#' @description
#' Computes a continuous metabolic syndrome severity z-score using sex- and
#' race-specific standardized components and coefficients (factor-loading style).
#'
#' Behavior note:
#'  - Parameters are selected per-row based on each row's (race, sex) key.
#'  - All unique keys present in the data must have a matching entry in `params`.
#'
#' Required columns (no unit conversion performed):
#'  - waist (cm), bp_sys (mmHg), bp_dia (mmHg)
#'  - TG, HDL_c, glucose (mmol/L)
#'  - sex (1=male, 2=female)
#'  - race (one of "NHW","NHB","HW", or accepted synonyms; "HA" is recognised
#'    by the normaliser but has no default params — see `params` argument)
#'
#' @param data data.frame / tibble.
#' @param params Named list keyed by "RACE_SEX" (e.g. "NHW_M"). Each element:
#'   list(intercept, waist, TG, HDL, glucose, MAP) where each component (except intercept)
#'   is a named numeric vector c(mean=, sd=, coef=). Default parameters are provided
#'   for NHW, NHB, and HW (male and female); no default HA parameters are included
#'   because Gurka et al. (2014) did not publish HA-specific coefficients. Passing
#'   `race = "HA"` with default params will raise an error; supply custom params if needed.
#' @param verbose Logical; if TRUE, prints column mapping and computing messages.
#' @param na_action One of c("keep","omit","error","ignore","warn") for required-input NAs. Default "keep".
#' @param na_warn_prop Proportion (0-1) above which high-missingness warning fires when na_action='warn'. Default 0.2.
#' @param col_map Optional named list mapping canonical keys (`waist`, `TG`, `HDL`,
#'   `glucose`, `MAP`) to actual column names in `data`. If `NULL`, column names
#'   are inferred automatically.
#' @param diagnostics Logical; if TRUE (default) emit value/range diagnostic warnings
#'   (negative, out-of-range checks).
#'
#' @return tibble with one numeric column: MetSSS
#' @export 
#' @references
#' \insertRef{gurka2014metsss}{HealthMarkers}
#' \insertRef{deboer2015metss}{HealthMarkers} (clinical application)
#' \insertRef{gurka2017metss}{HealthMarkers} (clinical application)
#' \insertRef{deboer2018metss}{HealthMarkers} (clinical application)
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @examples
#' df <- data.frame(
#'   waist = 95, bp_sys = 120, bp_dia = 80, TG = 1.5, HDL_c = 1.2,
#'   glucose = 5.5, sex = 1, race = "NHW", age = 45
#' )
#' metss(df)
#' 
metss <- function(data,
                  params = list(
                    NHW_M = list(
                      intercept = -2.344,
                      waist     = c(mean = 94.0, sd = 12.4, coef = 0.846),
                      TG        = c(mean = 1.50, sd = 0.60, coef = 0.701),
                      HDL       = c(mean = 1.10, sd = 0.30, coef = -0.663),
                      glucose   = c(mean = 5.30, sd = 0.60, coef = 0.658),
                      MAP       = c(mean = 97.0, sd = 11.0, coef = 0.466)
                    ),
                    NHW_F = list(
                      intercept = -2.381,
                      waist     = c(mean = 89.7, sd = 14.8, coef = 0.817),
                      TG        = c(mean = 1.28, sd = 0.91, coef = 0.679),
                      HDL       = c(mean = 1.50, sd = 0.40, coef = -0.727),
                      glucose   = c(mean = 5.08, sd = 0.52, coef = 0.622),
                      MAP       = c(mean = 91.0, sd = 11.0, coef = 0.557)
                    ),
                    NHB_M = list(
                      intercept = -2.399,
                      waist     = c(mean = 92.8, sd = 13.1, coef = 0.830),
                      TG        = c(mean = 1.18, sd = 0.75, coef = 0.551),
                      HDL       = c(mean = 1.27, sd = 0.37, coef = -0.598),
                      glucose   = c(mean = 5.55, sd = 0.85, coef = 0.702),
                      MAP       = c(mean = 98.0, sd = 13.0, coef = 0.564)
                    ),
                    NHB_F = list(
                      intercept = -2.395,
                      waist     = c(mean = 96.4, sd = 16.4, coef = 0.858),
                      TG        = c(mean = 1.14, sd = 0.70, coef = 0.570),
                      HDL       = c(mean = 1.36, sd = 0.39, coef = -0.634),
                      glucose   = c(mean = 5.42, sd = 0.84, coef = 0.687),
                      MAP       = c(mean = 95.0, sd = 13.0, coef = 0.577)
                    ),
                    HW_M = list(
                      intercept = -2.377,
                      waist     = c(mean = 98.5, sd = 11.5, coef = 0.864),
                      TG        = c(mean = 1.95, sd = 1.19, coef = 0.724),
                      HDL       = c(mean = 1.13, sd = 0.30, coef = -0.620),
                      glucose   = c(mean = 5.67, sd = 0.90, coef = 0.624),
                      MAP       = c(mean = 97.0, sd = 11.0, coef = 0.448)
                    ),
                    HW_F = list(
                      intercept = -2.388,
                      waist     = c(mean = 97.9, sd = 14.2, coef = 0.858),
                      TG        = c(mean = 1.66, sd = 1.06, coef = 0.715),
                      HDL       = c(mean = 1.29, sd = 0.35, coef = -0.657),
                      glucose   = c(mean = 5.53, sd = 0.87, coef = 0.644),
                      MAP       = c(mean = 91.0, sd = 11.0, coef = 0.512)
                    )
                  ),
                  col_map = NULL,
                  verbose = TRUE,
                  na_action = c("keep","omit","error","ignore","warn"),
                  na_warn_prop = 0.2,
                  diagnostics = TRUE) {

  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "metss"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action <- .na$na_action_eff

  .metss_validate_data_frame(data)
  .metss_validate_params(params)

  req <- c("waist","bp_sys","bp_dia","TG","HDL_c","glucose","sex","race")
  cm  <- .hm_build_col_map(data, col_map, keys = req, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map
  miss <- setdiff(req, names(data))
  if (length(miss)) {
    rlang::abort(paste0("metss(): missing required columns: ", paste(miss, collapse = ", ")),
                 class = "healthmarkers_metss_error_missing_columns")
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  MetSSS  [factor-loading z-score, sex/race-specific params]", fn_name), level = "inform")

  # Coerce numerics and clean non-finite
  num_cols <- c("waist","bp_sys","bp_dia","TG","HDL_c","glucose")
  for (cn in num_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0)
        rlang::warn(sprintf("metss(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }
  if (!is.numeric(data$sex)) {
    old <- data$sex
    # Try to interpret character values first
    char_sex <- tolower(trimws(as.character(old)))
    data$sex <- ifelse(char_sex %in% c("m", "male", "1"), 1,
                ifelse(char_sex %in% c("f", "female", "2"), 2, NA_real_))
    introduced <- sum(is.na(data$sex) & !is.na(old))
    if (introduced > 0)
      rlang::warn(sprintf("metss(): 'sex' coerced to numeric; NAs introduced: %d", introduced))
  } else {
    # Already numeric, just ensure it's 1 or 2
    data$sex[!is.finite(data$sex)] <- NA_real_
  }

  # Missingness warnings only when na_action_raw == "warn"
  if (identical(na_action_raw, "warn")) {
    .metss_warn_high_missing(data, req, na_warn_prop = na_warn_prop)
  }
  # Basic domain diagnostics (optional)
  if (isTRUE(diagnostics)) {
    .metss_warn_value_diagnostics(data)
  }

  # NA policy
  if (na_action == "error") {
    any_na <- Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    if (any(any_na))
      rlang::abort("metss(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_metss_error_missing_values")
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(req, function(cn) is.na(data[[cn]])))
    hm_inform(sprintf("metss(): omitting %d rows with NA in required inputs", sum(!keep)), level = if (isTRUE(verbose)) "inform" else "debug")
    data <- data[keep, , drop = FALSE]
  }

  if (nrow(data) == 0L) return(tibble::tibble(MetSSS = numeric()))

  # Per-row key derivation
  key_vec <- .metss_key_from_data(data)
  unique_keys <- unique(key_vec[!is.na(key_vec)])   # NA rows scored as NA, not an error
  missing_keys <- setdiff(unique_keys, names(params))
  if (length(missing_keys)) {
    rlang::abort(
      sprintf("metss(): params missing keys for: %s. Available: %s",
              paste(missing_keys, collapse = ", "),
              paste(names(params), collapse = ", ")),
      class = "healthmarkers_metss_error_missing_param_key"
    )
  }
  if (length(unique_keys) > 1L) {
    rlang::warn(
      sprintf("metss(): multiple sex/race keys detected (%s); computing per-row using matched params.",
              paste(unique_keys, collapse = ", ")),
      class = "healthmarkers_metss_warn_multiple_keys"
    )
  }
  # Compute MetSSS per row

  MAP <- (2 * data$bp_dia + data$bp_sys) / 3
  MetSSS <- rep(NA_real_, nrow(data))

  for (k in unique_keys) {
    idx <- which(key_vec == k)
    p <- params[[k]]
    .metss_validate_param_entry(p, k)

    z_waist <- (data$waist[idx]   - p$waist["mean"])   / p$waist["sd"]
    z_TG    <- (data$TG[idx]      - p$TG["mean"])      / p$TG["sd"]
    z_HDL   <- (data$HDL_c[idx]   - p$HDL["mean"])     / p$HDL["sd"]
    z_glu   <- (data$glucose[idx] - p$glucose["mean"]) / p$glucose["sd"]
    z_MAP   <- (MAP[idx]          - p$MAP["mean"])     / p$MAP["sd"]

    MetSSS[idx] <- p$intercept +
      p$waist["coef"]   * z_waist +
      p$TG["coef"]      * z_TG +
      p$HDL["coef"]     * z_HDL +
      p$glucose["coef"] * z_glu +
      p$MAP["coef"]     * z_MAP
  }

  out <- tibble::tibble(MetSSS = as.numeric(MetSSS))

  if (!is.null(id_col)) {
    out[[id_col]] <- data[[id_col]]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }
  out
}

# ---- helpers ----
.metss_validate_data_frame <- function(data) {
  if (!is.data.frame(data)) {
    rlang::abort("metss(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_metss_error_data_type")
  }
  invisible(TRUE)
}
.metss_validate_params <- function(params) {
  if (!is.list(params) || is.null(names(params)) || any(names(params) == "")) {
    rlang::abort("metss(): `params` must be a named list keyed by 'RACE_SEX'.",
                 class = "healthmarkers_metss_error_params_type")
  }
  invisible(TRUE)
}
.metss_validate_param_entry <- function(p, key) {
  needed <- c("intercept","waist","TG","HDL","glucose","MAP")
  miss <- setdiff(needed, names(p))
  if (length(miss))
    rlang::abort(sprintf("metss(): params['%s'] missing entries: %s",
                         key, paste(miss, collapse = ", ")),
                 class = "healthmarkers_metss_error_params_entry_missing")
  if (!(is.numeric(p$intercept) && length(p$intercept) == 1L && is.finite(p$intercept)))
    rlang::abort(sprintf("metss(): params['%s']$intercept must be finite numeric scalar.", key),
                 class = "healthmarkers_metss_error_params_intercept")
  for (comp in c("waist","TG","HDL","glucose","MAP")) {
    v <- p[[comp]]
    if (!(is.numeric(v) && all(c("mean","sd","coef") %in% names(v))))
      rlang::abort(sprintf("metss(): params['%s']$%s must be named numeric with mean, sd, coef.", key, comp),
                   class = "healthmarkers_metss_error_params_component")
    if (v["sd"] <= 0 || any(!is.finite(v[c("mean","sd","coef")])))
      rlang::abort(sprintf("metss(): params['%s']$%s entries must be finite; sd>0.", key, comp),
                   class = "healthmarkers_metss_error_params_component_value")
  }
  invisible(TRUE)
}
.metss_key_from_data <- function(data) {
  race_raw <- toupper(as.character(data$race))
  race_norm <- ifelse(race_raw %in% c("NHW","WHITE"), "NHW",
                 ifelse(race_raw %in% c("NHB","BLACK"), "NHB",
                   ifelse(race_raw %in% c("HW","HISPANIC","HISP","H/L"), "HW",
                     ifelse(race_raw %in% c("HA","ASIAN"), "HA", NA_character_))))
  sex_norm <- .hm_normalize_sex(data$sex, to = "MF")
  # Return NA_character_ when sex or race is NA, so those rows are scored as NA
  ifelse(is.na(sex_norm) | is.na(race_norm), NA_character_,
         paste0(race_norm, "_", sex_norm))
}
.metss_warn_high_missing <- function(df, cols, na_warn_prop = 0.2) {
  for (cn in cols) {
    x <- df[[cn]]; n <- length(x); if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("metss(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }
  invisible(TRUE)
}
.metss_warn_value_diagnostics <- function(df) {
  .chk_nonneg(df$TG,      "TG (mmol/L)")
  .chk_nonneg(df$HDL_c,   "HDL_c (mmol/L)")
  .chk_nonneg(df$glucose, "glucose (mmol/L)")
  .chk_positive(df$waist, "waist (cm)")
  .chk_range(df$bp_sys,   "bp_sys (mmHg)", 60, 300)
  .chk_range(df$bp_dia,   "bp_dia (mmHg)", 30, 200)
  bad_sex <- sum(is.finite(df$sex) & !(df$sex %in% c(1,2)))
  if (bad_sex > 0) rlang::warn(sprintf("metss(): 'sex' has %d values not in {1,2}.", bad_sex))
  race_raw <- toupper(as.character(df$race))
  bad_race <- sum(!(race_raw %in% c("NHW","NHB","HW","HA","WHITE","BLACK","HISPANIC","H/L","ASIAN","OTHER")))
  if (bad_race > 0) hm_inform(level = "debug", msg = sprintf("metss(): 'race' has %d unrecognized values.", bad_race))
  invisible(TRUE)
}
.chk_nonneg <- function(x,label){
  n <- sum(is.finite(x) & x < 0)
  if (n>0) rlang::warn(sprintf("metss(): '%s' has %d negative values; check units.", label, n))
  invisible(TRUE)
}
.chk_positive <- function(x,label){
  n <- sum(is.finite(x) & x <= 0)
  if (n>0) rlang::warn(sprintf("metss(): '%s' has %d non-positive values; check units.", label, n))
  invisible(TRUE)
}
.chk_range <- function(x,label,lo,hi){
  nlo <- sum(is.finite(x) & x < lo); nhi <- sum(is.finite(x) & x > hi)
  if ((nlo+nhi)>0) rlang::warn(sprintf("metss(): '%s' outside [%g,%g] in %d rows; check units.", label, lo, hi, nlo+nhi))
  invisible(TRUE)
}

# tests/testthat/test_metss.R

library(testthat)
library(tibble)
library(HealthMarkers)

# ---- Missing required columns ----
test_that("metss errors if missing required columns", {
  skip_on_cran()
  df <- tibble(
    bp_sys  = 120,
    bp_dia  = 80,
    TG      = 1.5,
    HDL_c   = 1.1,
    glucose = 5.3,
    sex     = 1,
    race    = "NHW"
  )
  expect_error(
    metss(df),
    "metss\\(\\): missing required columns: waist"
  )
})

# ---- Correct computation (default params NHW male) ----
test_that("metss computes correct score for default NHW male", {
  skip_on_cran()
  df <- tibble(
    waist   = 94,
    bp_sys  = 120,
    bp_dia  = 80,
    TG      = 1.5,
    HDL_c   = 1.1,
    glucose = 5.3,
    sex     = 1,
    race    = "NHW"
  )
  out <- metss(df)
  MAP <- (2 * 80 + 120) / 3
  z_MAP <- (MAP - 97) / 11
  expected <- -2.344 + 0.466 * z_MAP
  expect_s3_class(out, "tbl_df")
  expect_named(out, "MetSSS")
  expect_type(out$MetSSS, "double")
  expect_equal(unname(out$MetSSS), expected, tolerance = 1e-6)
})

# ---- Vectorization ----
test_that("metss is vectorized over multiple rows", {
  skip_on_cran()
  df <- tibble(
    waist   = c(94, 100),
    bp_sys  = c(120, 130),
    bp_dia  = c(80, 85),
    TG      = c(1.5, 2.0),
    HDL_c   = c(1.1, 1.2),
    glucose = c(5.3, 6.0),
    sex     = c(1, 1),
    race    = c("NHW", "NHW")
  )
  out <- metss(df)
  expect_equal(nrow(out), 2L)
  single_out <- metss(df[1, ])
  expect_equal(unname(out$MetSSS[1]), unname(single_out$MetSSS), tolerance = 1e-6)
})

# ---- Missing params key ----
test_that("metss errors when params key for sex-race is missing", {
  skip_on_cran()
  df_f <- tibble(
    waist   = 100,
    bp_sys  = 130,
    bp_dia  = 85,
    TG      = 2.0,
    HDL_c   = 1.2,
    glucose = 6.0,
    sex     = 2,
    race    = "NHW"
  )
  # Provide custom params with only NHW_M (no NHW_F) to trigger the missing-key error
  custom_params_m_only <- list(
    NHW_M = list(
      intercept = -2.344,
      waist     = c(mean = 94.0, sd = 12.4, coef = 0.846),
      TG        = c(mean = 1.50, sd = 0.60, coef = 0.701),
      HDL       = c(mean = 1.10, sd = 0.30, coef = -0.663),
      glucose   = c(mean = 5.30, sd = 0.60, coef = 0.658),
      MAP       = c(mean = 97.0, sd = 11.0, coef = 0.466)
    )
  )
  expect_error(
    metss(df_f, params = custom_params_m_only),
    "params missing keys for: NHW_F"
  )
})

# ---- Verbose messages ----
test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    waist   = 94,
    bp_sys  = 120,
    bp_dia  = 80,
    TG      = 1.5,
    HDL_c   = 1.1,
    glucose = 5.3,
    sex     = 1,
    race    = "NHW"
  )
  expect_message(metss(df, verbose = TRUE), "metss")
  expect_message(metss(df, verbose = TRUE), "col_map")
  expect_message(metss(df, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    waist   = 94, bp_sys = 120, bp_dia = 80,
    TG = 1.5, HDL_c = 1.1, glucose = 5.3,
    sex = 1, race = "NHW"
  )
  msgs <- testthat::capture_messages(metss(df, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

# ---- NA handling policies ----
test_that("na_action = error and omit behave as expected", {
  skip_on_cran()
  df <- tibble(
    waist   = c(94, NA_real_),
    bp_sys  = c(120, 120),
    bp_dia  = c(80, 80),
    TG      = c(1.5, 1.5),
    HDL_c   = c(1.1, 1.1),
    glucose = c(5.3, 5.3),
    sex     = c(1, 1),
    race    = c("NHW", "NHW")
  )
  expect_error(
    suppressWarnings(metss(df, na_action = "error")),
    "required inputs contain missing values"
  )
  out_omit <- suppressWarnings(metss(df, na_action = "omit"))
  expect_equal(nrow(out_omit), 1L)
})

# ---- Extreme values pass through unchanged ----
test_that("extreme input values pass through without error", {
  skip_on_cran()
  df_ext <- tibble(
    waist   = 400,
    bp_sys  = 500,
    bp_dia  = 250,
    TG      = 50,
    HDL_c   = 10,
    glucose = 80,
    sex     = 1,
    race    = "NHW"
  )
  out <- suppressWarnings(metss(df_ext, diagnostics = FALSE))
  expect_equal(nrow(out), 1L)
})

# ---- Multiple keys warning ----
test_that("multiple sex/race keys triggers warning and uses first-row key", {
  skip_on_cran()
  df <- tibble(
    waist   = c(94, 100),
    bp_sys  = c(120, 130),
    bp_dia  = c(80, 85),
    TG      = c(1.5, 2.0),
    HDL_c   = c(1.1, 1.2),
    glucose = c(5.3, 6.0),
    sex     = c(1, 2),
    race    = c("NHW", "NHB")
  )
  expect_warning(
    out <- metss(df),
    "multiple sex/race keys detected"
  )
  expect_equal(nrow(out), 2L)
})

# ---- Custom params and computation check ----
test_that("custom params applied correctly", {
  skip_on_cran()
  custom_params <- list(
    NHW_M = list(
      intercept = 0,
      waist     = c(mean = 100, sd = 10, coef = 1),
      TG        = c(mean = 2,   sd = 0.5, coef = 0.5),
      HDL       = c(mean = 1.2, sd = 0.2, coef = -0.7),
      glucose   = c(mean = 6,   sd = 0.5, coef = 0.8),
      MAP       = c(mean = 100, sd = 10,  coef = 0.3)
    )
  )
  df <- tibble(
    waist   = 110,
    bp_sys  = 130,
    bp_dia  = 85,
    TG      = 2.5,
    HDL_c   = 1.0,
    glucose = 6.5,
    sex     = 1,
    race    = "NHW"
  )
  MAP <- (2 * 85 + 130) / 3
  z_waist <- (110 - 100) / 10
  z_TG    <- (2.5 - 2) / 0.5
  z_HDL   <- (1.0 - 1.2) / 0.2
  z_glu   <- (6.5 - 6) / 0.5
  z_MAP   <- (MAP - 100) / 10
  expected <- 0 + 1*z_waist + 0.5*z_TG + (-0.7)*z_HDL + 0.8*z_glu + 0.3*z_MAP
  out <- metss(df, params = custom_params)
  expect_equal(unname(out$MetSSS), expected, tolerance = 1e-6)
})

#' Compute composite vitamin and endocrine marker ratios and z-scores
#'
#' Given serum/plasma vitamins and related analytes, `vitamin_markers()` computes:
#' - VitD_Z: z-score of 25(OH)D using provided reference mean/sd
#' - B12_Fol_Ratio: vitamin B12 / folate
#' - Ferr_TSat_R: ferritin / transferrin saturation (TSat)
#' - Cort_DHEA_R: cortisol / DHEA-S
#' - T_E2_Ratio: testosterone / estradiol
#' - TSH_fT4_R: TSH / free T4
#' - Retinol_Z: z-score of retinol using provided reference mean/sd
#' - Toco_Lip_R: alpha-tocopherol / total lipids
#' - Mg_Zn_R: magnesium / zinc
#' - Cu_Zn_R: copper / zinc
#' Plus pass-through: PIVKA_II, VitC, Homocysteine, MMA
#'
#' HM-CS v2:
#' - Validation via `hm_validate_inputs(data, col_map, required_keys, fn)`
#' - User errors via `rlang::abort(..., class=...)`
#' - Verbosity via `hm_inform(level)` controlled by `options(healthmarkers.verbose)`
#' - High-missingness diagnostics at debug level only
#'
#' @param data A data.frame or tibble with vitamin/analyte columns.
#' @param col_map Named list mapping required keys to column names:
#'   VitD, VitD_ref_mean, VitD_ref_sd, B12, Folate, Ferritin, TSat,
#'   Cortisol, DHEAS, Testosterone, Estradiol, TSH, free_T4,
#'   Retinol, Retinol_ref_mean, Retinol_ref_sd, Tocopherol, Total_lipids,
#'   PIVKA_II, VitC, Homocysteine, MMA, Magnesium, Zinc, Copper.
#' @param na_action One of c("keep","omit","error") for required inputs. Default "keep".
#' @param na_warn_prop Proportion \eqn{[0,1]} to trigger high-missingness debug notices. Default 0.2.
#' @param verbose Logical; if `TRUE` (default), prints column mapping, input
#'   availability, physiological range information (informational only, values
#'   not altered), the list of markers being computed with their inputs, and a
#'   per-column results summary.
#'
#' @return A tibble with columns:
#'   VitD_Z, B12_Fol_Ratio, Ferr_TSat_R, Cort_DHEA_R, T_E2_Ratio, TSH_fT4_R,
#'   Retinol_Z, Toco_Lip_R, PIVKA_II, VitC, Homocysteine, MMA, Mg_Zn_R, Cu_Zn_R.
#'   If an ID column is detected in `data` (e.g. `id`, `IID`, `participant_id`),
#'   it is prepended as the first output column.
#'
#' @examples
#' \donttest{
#' # All 25 required columns must be supplied
#' df <- data.frame(
#'   VitD = 50, VitD_ref_mean = 40, VitD_ref_sd = 5,
#'   B12 = 300, Folate = 15, Ferritin = 80, TSat = 0.25,
#'   Cortisol = 200, DHEAS = 100, Testosterone = 12, Estradiol = 120,
#'   TSH = 2, free_T4 = 14, Retinol = 0.8, Retinol_ref_mean = 0.9,
#'   Retinol_ref_sd = 0.2, Tocopherol = 30, Total_lipids = 3,
#'   PIVKA_II = 5, VitC = 60, Homocysteine = 10, MMA = 0.3,
#'   Magnesium = 0.8, Zinc = 15, Copper = 15
#' )
#' vitamin_markers(df, verbose = FALSE)
#' }
#'
#' @note
#' `VitD_Z` and `Retinol_Z` are z-scores using **user-supplied** reference
#' mean and SD; no population reference equations are applied. All ratio markers
#' (`B12_Fol_Ratio`, `Ferr_TSat_R`, `Cort_DHEA_R`, `T_E2_Ratio`, `TSH_fT4_R`,
#' `Toco_Lip_R`, `Mg_Zn_R`, `Cu_Zn_R`) are simple numerator/denominator
#' divisions. `PIVKA_II`, `VitC`, `Homocysteine`, and `MMA` are
#' **pass-through** columns; no formula is applied.
#'
#' @references
#' \insertRef{holick2007vitddeficiency}{HealthMarkers} (vitamin D deficiency review; background)
#' \insertRef{oleary2010vitb12}{HealthMarkers} (vitamin B12 in health and disease; background)
#' \insertRef{ganz2015iron}{HealthMarkers} (iron homeostasis and ferritin; background)
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
vitamin_markers <- function(data,
                            col_map      = NULL,
                            na_action    = c("keep", "omit", "error"),
                            na_warn_prop = 0.2,
                            verbose      = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "vitamin_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data)) {
    rlang::abort("vitamin_markers(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_vitamin_error_data_type")
  }
  if (!is.null(col_map) && !is.list(col_map)) {
    rlang::abort("vitamin_markers(): `col_map` must be a named list.",
                 class = "healthmarkers_vitamin_error_colmap_type")
  }

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)
  required_keys <- c(
    "VitD","VitD_ref_mean","VitD_ref_sd","B12","Folate","Ferritin","TSat",
    "Cortisol","DHEAS","Testosterone","Estradiol","TSH","free_T4",
    "Retinol","Retinol_ref_mean","Retinol_ref_sd","Tocopherol","Total_lipids",
    "PIVKA_II","VitC","Homocysteine","MMA","Magnesium","Zinc","Copper"
  )

  cm      <- .hm_build_col_map(data, col_map, required_keys, fn = "vitamin_markers")
  data    <- cm$data
  col_map <- cm$col_map

  # HM-CS v2: standardized validation
  hm_validate_inputs(data, col_map, required_keys = required_keys, fn = fn_name)

  # --- Verbose: col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: optional inputs
  if (isTRUE(verbose)) {
    avail_keys  <- required_keys[vapply(required_keys, function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    absent_keys <- setdiff(required_keys, avail_keys)
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(avail_keys))
      lines <- c(lines, sprintf("  present:  %s", paste(avail_keys, collapse = ", ")))
    if (length(absent_keys))
      lines <- c(lines, sprintf("  missing:  %s", paste(absent_keys, collapse = ", ")))
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # Ensure mapped columns exist
  req_cols <- unname(unlist(col_map[required_keys], use.names = FALSE))
  missing_cols <- setdiff(req_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      paste0("vitamin_markers(): missing required columns in `data`: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_vitamin_error_missing_columns"
    )
  }

  # Coerce mapped numerics; warn if NAs introduced; non-finite -> NA
  for (key in required_keys) {
    cn <- col_map[[key]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced > 0) {
        rlang::warn(sprintf("vitamin_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced))
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug)
  for (key in required_keys) {
    cn <- col_map[[key]]
    x <- data[[cn]]
    n <- length(x); if (!n) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("vitamin_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }

  hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
            msg   = hm_fmt_col_map(col_map[required_keys], "vitamin_markers"))

  # NA policy
  used_cols <- req_cols
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("vitamin_markers(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_vitamin_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  # Early return if empty
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      VitD_Z = numeric(),
      B12_Fol_Ratio = numeric(),
      Ferr_TSat_R = numeric(),
      Cort_DHEA_R = numeric(),
      T_E2_Ratio = numeric(),
      TSH_fT4_R = numeric(),
      Retinol_Z = numeric(),
      Toco_Lip_R = numeric(),
      PIVKA_II = numeric(),
      VitC = numeric(),
      Homocysteine = numeric(),
      MMA = numeric(),
      Mg_Zn_R = numeric(),
      Cu_Zn_R = numeric()
    ))
  }

  # Optional extremes scan (informational range check via verbose)
  if (isTRUE(verbose)) {
    default_rules_info <- list(
      VitD = c(0, 250), VitD_ref_mean = c(-Inf, Inf), VitD_ref_sd = c(0.01, Inf),
      B12 = c(0, 2000), Folate = c(0, 100),
      Ferritin = c(0, 3000), TSat = c(0, 1),
      Cortisol = c(0, 2000), DHEAS = c(0, 2000),
      Testosterone = c(0, 200), Estradiol = c(0, 5000),
      TSH = c(0, 200), free_T4 = c(0, 100),
      Retinol = c(0, 10), Retinol_ref_mean = c(-Inf, Inf), Retinol_ref_sd = c(0.001, Inf),
      Tocopherol = c(0, 200), Total_lipids = c(0.001, 100),
      PIVKA_II = c(0, 10000), VitC = c(0, 1000), Homocysteine = c(0, 200), MMA = c(0, 20),
      Magnesium = c(0, 10), Zinc = c(0, 1000), Copper = c(0, 1000)
    )
    flagged_details <- character(0)
    for (key in required_keys) {
      rng <- default_rules_info[[key]]
      if (is.null(rng) || !is.finite(rng[1]) || !is.finite(rng[2])) next
      cn <- col_map[[key]]
      if (is.null(cn) || !(cn %in% names(data))) next
      x <- data[[cn]]
      bad <- sum(is.finite(x) & (x < rng[1] | x > rng[2]), na.rm = TRUE)
      if (bad > 0L)
        flagged_details <- c(flagged_details,
          sprintf("  %s: %d value(s) outside plausible range", key, bad))
    }
    if (length(flagged_details))
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                flagged_details), collapse = "\n"),
        level = "inform"
      )
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    marker_deps <- list(
      VitD_Z        = c("VitD", "VitD_ref_mean", "VitD_ref_sd"),
      B12_Fol_Ratio = c("B12", "Folate"),
      Ferr_TSat_R   = c("Ferritin", "TSat"),
      Cort_DHEA_R   = c("Cortisol", "DHEAS"),
      T_E2_Ratio    = c("Testosterone", "Estradiol"),
      TSH_fT4_R     = c("TSH", "free_T4"),
      Retinol_Z     = c("Retinol", "Retinol_ref_mean", "Retinol_ref_sd"),
      Toco_Lip_R    = c("Tocopherol", "Total_lipids"),
      PIVKA_II      = c("PIVKA_II"),
      VitC          = c("VitC"),
      Homocysteine  = c("Homocysteine"),
      MMA           = c("MMA"),
      Mg_Zn_R       = c("Magnesium", "Zinc"),
      Cu_Zn_R       = c("Copper", "Zinc")
    )
    avail_keys2 <- required_keys[vapply(required_keys, function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    status <- vapply(names(marker_deps), function(m) {
      miss_k <- setdiff(marker_deps[[m]], avail_keys2)
      if (length(miss_k) == 0L)
        sprintf("  %-16s [%s]", m, paste(marker_deps[[m]], collapse = ", "))
      else
        sprintf("  %-16s NA [missing: %s]", m, paste(miss_k, collapse = ", "))
    }, character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Safe division with consolidated zero-denominator tracking
  dz_env <- new.env(parent = emptyenv()); dz_env$counts <- list()
  safe_div <- function(num, den, label) {
    out <- num / den
    zero_den <- is.finite(den) & den == 0
    dz_env$counts[[label]] <- sum(zero_den, na.rm = TRUE)
    out[!is.finite(out)] <- NA_real_
    out
  }

  g <- function(key) data[[col_map[[key]]]]

  VitD_Z        <- (g("VitD")     - g("VitD_ref_mean"))    / g("VitD_ref_sd")
  B12_Fol_Ratio <-  safe_div(g("B12"), g("Folate"), "B12_Folate_den")
  Ferr_TSat_R   <-  safe_div(g("Ferritin"), g("TSat"), "Ferr_TSat_den")
  Cort_DHEA_R   <-  safe_div(g("Cortisol"), g("DHEAS"), "Cort_DHEAS_den")
  T_E2_Ratio    <-  safe_div(g("Testosterone"), g("Estradiol"), "T_E2_den")
  TSH_fT4_R     <-  safe_div(g("TSH"), g("free_T4"), "TSH_fT4_den")
  Retinol_Z     <- (g("Retinol") - g("Retinol_ref_mean")) / g("Retinol_ref_sd")
  Toco_Lip_R    <-  safe_div(g("Tocopherol"), g("Total_lipids"), "Toco_Lip_den")
  Mg_Zn_R       <-  safe_div(g("Magnesium"), g("Zinc"), "Mg_Zn_den")
  Cu_Zn_R       <-  safe_div(g("Copper"), g("Zinc"), "Cu_Zn_den")

  out <- tibble::tibble(
    VitD_Z        = as.numeric(VitD_Z),
    B12_Fol_Ratio = as.numeric(B12_Fol_Ratio),
    Ferr_TSat_R   = as.numeric(Ferr_TSat_R),
    Cort_DHEA_R   = as.numeric(Cort_DHEA_R),
    T_E2_Ratio    = as.numeric(T_E2_Ratio),
    TSH_fT4_R     = as.numeric(TSH_fT4_R),
    Retinol_Z     = as.numeric(Retinol_Z),
    Toco_Lip_R    = as.numeric(Toco_Lip_R),
    PIVKA_II      = as.numeric(g("PIVKA_II")),
    VitC          = as.numeric(g("VitC")),
    Homocysteine  = as.numeric(g("Homocysteine")),
    MMA           = as.numeric(g("MMA")),
    Mg_Zn_R       = as.numeric(Mg_Zn_R),
    Cu_Zn_R       = as.numeric(Cu_Zn_R)
  )

  # Consolidated zero-denominator warning
  dz <- dz_env$counts
  dz_total <- if (length(dz)) sum(unlist(dz), na.rm = TRUE) else 0L
  if (dz_total > 0L) {
    nz <- unlist(dz); nz <- nz[nz > 0]
    lbl <- paste(sprintf("%s=%d", names(nz), nz), collapse = ", ")
    rlang::warn(sprintf("vitamin_markers(): zero denominators detected in %d cases (%s).", dz_total, lbl))
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

library(testthat)
library(tibble)

cm_id <- function(df) { m <- as.list(names(df)); names(m) <- names(df); m }

make_full_df <- function(n = 1) {
  tibble(
    VitD = rep(50, n), VitD_ref_mean = rep(40, n), VitD_ref_sd = rep(5, n),
    B12 = rep(300, n), Folate = rep(15, n),
    Ferritin = rep(100, n), TSat = rep(0.25, n),
    Cortisol = rep(200, n), DHEAS = rep(100, n),
    Testosterone = rep(12, n), Estradiol = rep(120, n),
    TSH = rep(2, n), free_T4 = rep(14, n),
    Retinol = rep(0.8, n), Retinol_ref_mean = rep(0.9, n), Retinol_ref_sd = rep(0.2, n),
    Tocopherol = rep(30, n), Total_lipids = rep(3, n),
    PIVKA_II = rep(5, n),
    VitC = rep(60, n),
    Homocysteine = rep(10, n),
    MMA = rep(0.3, n),
    Magnesium = rep(0.8, n), Zinc = rep(15, n), Copper = rep(15, n)
  )
}

test_that("errors if `data` is not a data.frame", {
  skip_on_cran()
  expect_error(
    vitamin_markers("not a df", col_map = list()),
    "must be a data\\.frame or tibble"
  )
})

test_that("errors if `col_map` is not a named list", {
  skip_on_cran()
  df <- make_full_df()
  expect_error(
    vitamin_markers(df, col_map = c(VitD = "VitD")),
    "`col_map` must be a named list"
  )
})

test_that("errors when mapped columns are missing in data", {
  skip_on_cran()
  df <- tibble(VitD = 50, VitD_ref_mean = 40, VitD_ref_sd = 5)
  req_keys <- c(
    "VitD","VitD_ref_mean","VitD_ref_sd","B12","Folate","Ferritin","TSat",
    "Cortisol","DHEAS","Testosterone","Estradiol","TSH","free_T4",
    "Retinol","Retinol_ref_mean","Retinol_ref_sd","Tocopherol","Total_lipids",
    "PIVKA_II","VitC","Homocysteine","MMA","Magnesium","Zinc","Copper"
  )
  cm <- as.list(req_keys); names(cm) <- req_keys
  expect_error(
    vitamin_markers(df, col_map = cm),
    "missing required columns in `data`"
  )
})

test_that("numeric coercion warns when NAs introduced and propagates to outputs", {
  skip_on_cran()
  df <- make_full_df(2)
  df$B12 <- c("300", "oops")
  df$TSat <- c("0.25", "0.25")
  cm <- cm_id(df)
  expect_warning(
    out <- vitamin_markers(df, col_map = cm),
    "coerced to numeric; NAs introduced"
  )
  expect_equal(out$Ferr_TSat_R, c(100 / 0.25, 100 / 0.25), tolerance = 1e-8)
  expect_equal(out$B12_Fol_Ratio, c(300 / 15, NA_real_))
})

test_that("na_action='omit' with all rows dropped returns empty tibble with expected columns", {
  skip_on_cran()
  df <- make_full_df(2)
  df$VitD <- c(NA_real_, NA_real_)
  cm <- cm_id(df)
  out <- suppressWarnings(vitamin_markers(df, col_map = cm, na_action = "omit"))
  expect_s3_class(out, "tbl_df")
  expect_equal(nrow(out), 0L)
  expect_true(all(c(
    "VitD_Z","B12_Fol_Ratio","Ferr_TSat_R","Cort_DHEA_R","T_E2_Ratio",
    "TSH_fT4_R","Retinol_Z","Toco_Lip_R","PIVKA_II","VitC",
    "Homocysteine","MMA","Mg_Zn_R","Cu_Zn_R"
  ) %in% names(out)))
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  df <- make_full_df(1)
  cm <- cm_id(df)
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(vitamin_markers(df, col_map = cm, verbose = TRUE), "vitamin_markers")
  expect_message(vitamin_markers(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(vitamin_markers(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  df <- make_full_df(1)
  cm <- cm_id(df)
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(vitamin_markers(df, col_map = cm, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",       msgs)), 1L)
})

test_that("extreme values produce range note in verbose; no warning in non-verbose mode", {
  skip_on_cran()
  base <- make_full_df(1)
  df <- base
  df$VitD <- 1000
  df$Ferritin <- 5000
  cm <- cm_id(df)

  # No warning emitted in non-verbose mode
  expect_no_warning(vitamin_markers(df, col_map = cm, verbose = FALSE))

  # VitD_Z is computed from the extreme value (unaltered)
  out <- vitamin_markers(df, col_map = cm, verbose = FALSE)
  expect_equal(out$VitD_Z, (1000 - 40) / 5, tolerance = 1e-8)

  # Verbose mode emits a range note informational message
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    vitamin_markers(df, col_map = cm, verbose = TRUE),
    "range note"
  )
})

# tests/testthat/setup.R

# Load once, before any tests run, and silence any "built under R version" notes
suppressWarnings({
  suppressPackageStartupMessages({
    library(testthat)
    library(tibble)
    library(dplyr)
    library(withr)
  })
})

#' Adipose insulin sensitivity indices (QUICKI, VAI, LAP, TyG, TG/HDL, Belfiore)
#'
#' @description
#' Computes adipose-related insulin sensitivity/resistance indices from fasting inputs.
#' Expected input units (converted internally):
#' - Glucose G0 mmol/L -> mg/dL (* 18)
#' - Insulin I0 pmol/L -> muU/mL (/ 6)
#' - TG mmol/L -> mg/dL (* 88.57)
#' - HDL mmol/L -> mg/dL (* 38.67)
#'
#' Reported indices (higher magnitude of negative "_inv" values implies worse adipose IR):
#' - Revised_QUICKI = 1 / (log10(I0 (muU/mL)) + log10(G0 (mg/dL)) + log10(FFA (mmol/L)))
#' - VAI (sex-specific; inverted as VAI_*_inv so larger negative = worse)
#' - TG_HDL_C_inv = -(TG/HDL) in mg/dL
#' - TyG_inv = -ln(TG (mg/dL) * G0 (mg/dL) / 2)
#' - LAP (sex-specific; inverted)
#' - McAuley_index = exp(2.63 - 0.28 ln(I0 (muU/mL)) - 0.31 ln(TG (mmol/L)))
#' - Adipo_inv = -(FFA * I0 (muU/mL))
#' - Belfiore_inv_FFA = - 2 / (I0 (muU/mL) * FFA + 1)
#'
#' **Inversion Note**: Most indices (VAI, LAP, TyG, TG/HDL, Adipo, Belfiore) are algebraically
#' inverted from their original insulin RESISTANCE definitions so that more negative values
#' consistently indicate worse adipose insulin sensitivity. Revised_QUICKI and McAuley_index
#' are retained in their original orientation (already sensitivity indices; higher = better).
#' See the vignette for detailed interpretation guidance.
#'
#' @param data Data frame or tibble with required columns mapped by `col_map`
#' @param col_map Named list mapping keys to columns: G0, I0, TG, HDL_c, FFA, waist, bmi
#' @param normalize One of c("none","z","inverse","range","robust"); default "none"
#' @param na_action One of c("keep","omit","error"); default "keep"
#' @param verbose Logical; if `TRUE` (default), prints column mapping, the list
#'   of indices being computed, and a per-column results summary.
#' @param ... Reserved
#'
#' @return A tibble with columns:
#' `Revised_QUICKI`, `VAI_Men_inv`, `VAI_Women_inv`, `TG_HDL_C_inv`, `TyG_inv`,
#' `LAP_Men_inv`, `LAP_Women_inv`, `McAuley_index`, `Adipo_inv`, `Belfiore_inv_FFA`.
#' If an ID column is detected in `data` (e.g. `id`, `IID`, `participant_id`), it
#' is prepended as the first output column.
#'
#' @references
#' \insertRef{katz2000quicki}{HealthMarkers}
#' \insertRef{amato2010vai}{HealthMarkers}
#' \insertRef{kahn2005lap}{HealthMarkers}
#' \insertRef{guerreroromero2018tyg}{HealthMarkers}
#' \insertRef{dobiasova2001atherogenic}{HealthMarkers}
#' \insertRef{belfiore1998insulin}{HealthMarkers}
#' \insertRef{raynaud1999fasting}{HealthMarkers}
#'
#' @examples
#' df <- tibble::tibble(
#'   G0 = c(5.2, 6.1),      # mmol/L
#'   I0 = c(60, 110),       # pmol/L
#'   TG = c(1.2, 1.8),      # mmol/L
#'   HDL_c = c(1.3, 1.0),   # mmol/L
#'   FFA = c(0.4, 0.6),     # mmol/L
#'   waist = c(85, 102),    # cm
#'   bmi = c(24, 31)        # kg/m^2
#' )
#' cm <- as.list(names(df)); names(cm) <- names(df)
#' out <- adipo_is(df, cm, verbose = FALSE, na_action = "keep")
#' @export
adipo_is <- function(data,
                     col_map = NULL,
                     normalize = "none",
                     na_action = c("keep","omit","error"),
                     verbose = TRUE,
                     ...) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "adipo_is"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  id_col    <- .hm_detect_id_col(data)

  # Validate normalize up front
  allowed_norm <- c("none","z","inverse","range","robust")
  if (!normalize %in% allowed_norm) {
    rlang::abort(
      sprintf("`normalize` must be one of: %s", paste(allowed_norm, collapse = ", ")),
      class = "healthmarkers_adipo_is_error_normalize_arg"
    )
  }

  # Centralized validation
  req <- c("G0","I0","TG","HDL_c","FFA","waist","bmi")

  cm      <- .hm_build_col_map(data, col_map, req, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # Emit test-aligned error before any generic helper
  nm <- names(col_map)
  if (is.null(nm)) nm <- character(0)
  missing_keys <- setdiff(req, nm)
  if (length(missing_keys)) {
    rlang::abort(
      paste0("adipo_is(): missing col_map entries for: ",
             paste(missing_keys, collapse = ", ")),
      class = "healthmarkers_adipo_is_error_missing_map"
    )
  }

  # Ensure mapped columns exist in data
  mapped_cols <- unname(unlist(col_map[req], use.names = FALSE))
  not_found <- setdiff(mapped_cols, names(data))
  if (length(not_found)) {
    rlang::abort(
      paste0("adipo_is(): columns not found in data: ", paste(not_found, collapse = ", ")),
      class = "healthmarkers_adipo_is_error_missing_columns"
    )
  }

  # Progress message — verbose col_map
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: computing markers list
  if (isTRUE(verbose)) {
    indices <- list(
      c("Revised_QUICKI",   "G0, I0, FFA"),
      c("VAI_Men_inv",      "waist, bmi, TG, HDL_c"),
      c("VAI_Women_inv",    "waist, bmi, TG, HDL_c"),
      c("TG_HDL_C_inv",     "TG, HDL_c"),
      c("TyG_inv",          "TG, G0"),
      c("LAP_Men_inv",      "waist, TG"),
      c("LAP_Women_inv",    "waist, TG"),
      c("McAuley_index",    "I0, TG"),
      c("Adipo_inv",        "FFA, I0"),
      c("Belfiore_inv_FFA", "I0, FFA")
    )
    status <- vapply(indices, function(x) sprintf("  %-20s [%s]", x[1], x[2]), character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Coerce to numeric when needed (warn once per column on NA introduction)
  for (key in req) {
    cn <- col_map[[key]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      intro <- sum(is.na(data[[cn]]) & !is.na(old))
      if (intro > 0) {
        rlang::warn(
          sprintf("adipo_is(): column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
          class = "healthmarkers_adipo_is_warn_na_coercion"
        )
      }
    }
  }

  # NA handling
  if (na_action == "error") {
    any_na <- Reduce(`|`, lapply(mapped_cols, function(cn) is.na(data[[cn]])))
    if (any(any_na)) {
      rlang::abort(
        "adipo_is(): required inputs contain missing values (na_action='error').",
        class = "healthmarkers_adipo_is_error_missing_values"
      )
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(mapped_cols, function(cn) is.na(data[[cn]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }
  if (nrow(data) == 0L) {
    return(tibble::tibble(
      Revised_QUICKI = numeric(),
      VAI_Men_inv = numeric(),
      VAI_Women_inv = numeric(),
      TG_HDL_C_inv = numeric(),
      TyG_inv = numeric(),
      LAP_Men_inv = numeric(),
      LAP_Women_inv = numeric(),
      McAuley_index = numeric(),
      Adipo_inv = numeric(),
      Belfiore_inv_FFA = numeric()
    ))
  }

  # Extract and convert units
  # Raw mmol/L values kept separately for formulas that require them (VAI, LAP, McAuley)
  TG_raw  <- as.numeric(data[[col_map$TG]])      # mmol/L — for VAI, LAP, McAuley
  HDL_raw <- as.numeric(data[[col_map$HDL_c]])   # mmol/L — for VAI

  G0    <- as.numeric(data[[col_map$G0]]) * 18   # mmol/L -> mg/dL (TyG, Revised_QUICKI)
  I0    <- as.numeric(data[[col_map$I0]]) / 6    # pmol/L -> mU/L
  TG    <- TG_raw * 88.57                        # mmol/L -> mg/dL (TyG, TG_HDL_C_inv)
  HDL   <- HDL_raw * 38.67                       # mmol/L -> mg/dL (TG_HDL_C_inv)
  FFA   <- as.numeric(data[[col_map$FFA]])
  waist <- as.numeric(data[[col_map$waist]])
  bmi   <- as.numeric(data[[col_map$bmi]])

  # Safe helpers
  safe_div <- function(num, den) { out <- num / den; out[!is.finite(out)] <- NA_real_; out }
  safe_log <- function(x) { out <- rep(NA_real_, length(x)); ok <- is.finite(x) & x > 0; out[ok] <- log(x[ok]); out }
  safe_log10 <- function(x) { out <- rep(NA_real_, length(x)); ok <- is.finite(x) & x > 0; out[ok] <- log10(x[ok]); out }

  # Indices
  Revised_QUICKI <- safe_div(1, safe_log10(I0) + safe_log10(G0) + safe_log10(FFA))

  # VAI: reference constants (1.03, 1.31, 0.81, 1.52) are in mmol/L units (Amato 2010)
  VAI_Men_inv <- -(
    safe_div(waist, (39.68 + 1.88 * bmi)) *
      safe_div(TG_raw, 1.03) *
      safe_div(1.31, HDL_raw)
  )
  VAI_Women_inv <- -(
    safe_div(waist, (36.58 + 1.89 * bmi)) *
      safe_div(TG_raw, 0.81) *
      safe_div(1.52, HDL_raw)
  )

  TG_HDL_C_inv  <- -safe_div(TG, HDL)         # ratio: both converted to mg/dL, scale-invariant
  TyG_inv       <- -safe_log(TG * G0 / 2)     # TG and G0 in mg/dL (Guerrero-Romero 2018)
  LAP_Men_inv   <- -((waist - 65) * TG_raw)   # TG in mmol/L (Kahn 2005)
  LAP_Women_inv <- -((waist - 58) * TG_raw)   # TG in mmol/L
  McAuley_index <- exp(2.63 - 0.28 * safe_log(I0) - 0.31 * safe_log(TG_raw))  # TG in mmol/L (McAuley 2001)
  Adipo_inv <- -(FFA * I0)
  Belfiore_inv_FFA <- -safe_div(2, (I0 * FFA + 1))

  out <- tibble::tibble(
    Revised_QUICKI = as.numeric(Revised_QUICKI),
    VAI_Men_inv = as.numeric(VAI_Men_inv),
    VAI_Women_inv = as.numeric(VAI_Women_inv),
    TG_HDL_C_inv = as.numeric(TG_HDL_C_inv),
    TyG_inv = as.numeric(TyG_inv),
    LAP_Men_inv = as.numeric(LAP_Men_inv),
    LAP_Women_inv = as.numeric(LAP_Women_inv),
    McAuley_index = as.numeric(McAuley_index),
    Adipo_inv = as.numeric(Adipo_inv),
    Belfiore_inv_FFA = as.numeric(Belfiore_inv_FFA)
  )

  # Optional normalization
  if (normalize != "none" && exists("normalize_vec", where = asNamespace("HealthMarkers"))) {
    norm_fun <- getExportedValue("HealthMarkers", "normalize_vec")
    out[] <- lapply(out, function(x) do.call(norm_fun, list(x = x, method = normalize)))
  }

  # Results summary
  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }
  # --- Verbose: results summary
  hm_inform(
    level = if (isTRUE(verbose)) "inform" else "debug",
    msg   = hm_result_summary(out, fn_name)
  )
  out
}

# tests/testthat/test_adipo_is.R

library(testthat)
library(tibble)
library(dplyr)
library(withr)

run_adipo <- function(df, normalize = "none", ...) {
  adipo_is(
    df,
    col_map = list(
      G0    = "G0",
      I0    = "I0",
      TG    = "TG",
      HDL_c = "HDL_c",
      FFA   = "FFA",
      waist = "waist",
      bmi   = "bmi"
    ),
    normalize = normalize,
    ...
  )
}

base_df <- tibble(
  G0    = 5.5,  # mmol/L
  I0    = 60,   # pmol/L
  TG    = 1.2,  # mmol/L
  HDL_c = 1.0,  # mmol/L
  FFA   = 0.45, # mmol/L
  waist = 80,   # cm
  bmi   = 24    # kg/m^2
)

test_that("partial col_map is supplemented by dictionary inference", {
  skip_on_cran()
  expect_no_error(
    adipo_is(base_df, col_map = list(G0 = "G0"))
  )
})

test_that("returns all 10 adipose-based indices", {
  skip_on_cran()
  out <- run_adipo(base_df)
  expected <- c(
    "Revised_QUICKI","VAI_Men_inv","VAI_Women_inv",
    "TG_HDL_C_inv","TyG_inv","LAP_Men_inv",
    "LAP_Women_inv","McAuley_index","Adipo_inv",
    "Belfiore_inv_FFA"
  )
  expect_named(out, expected)
  expect_equal(ncol(out), length(expected))
  expect_equal(nrow(out), 1L)
})

test_that("vectorized input produces matching row count", {
  skip_on_cran()
  df2 <- bind_rows(base_df, mutate(base_df, G0 = 6))
  out2 <- run_adipo(df2)
  expect_equal(nrow(out2), 2)
})

test_that("McAuley_index uses TG in mmol/L (not mg/dL)", {
  skip_on_cran()
  I0_u   <- base_df$I0 / 6          # pmol/L -> mU/L
  TG_mmol <- base_df$TG             # mmol/L: use raw, NOT * 88.57
  manual_mc <- exp(2.63 - 0.28 * log(I0_u) - 0.31 * log(TG_mmol))
  out <- run_adipo(base_df)
  expect_equal(out$McAuley_index, manual_mc, tolerance = 1e-8)
  # Sanity: typical adults -> McAuley ~3-10
  expect_gt(out$McAuley_index, 2)
  expect_lt(out$McAuley_index, 15)
})

test_that("LAP uses TG in mmol/L (not mg/dL)", {
  skip_on_cran()
  TG_mmol <- base_df$TG             # mmol/L: use raw
  manual_lap_m <- -((base_df$waist - 65) * TG_mmol)
  manual_lap_f <- -((base_df$waist - 58) * TG_mmol)
  out <- run_adipo(base_df)
  expect_equal(out$LAP_Men_inv,   manual_lap_m, tolerance = 1e-8)
  expect_equal(out$LAP_Women_inv, manual_lap_f, tolerance = 1e-8)
  # Typical men WC=80, TG=1.2: LAP_Men = (80-65)*1.2 = 18; _inv = -18
  expect_lt(out$LAP_Men_inv, 0)
})

test_that("VAI uses TG and HDL in mmol/L (Amato 2010 reference constants)", {
  skip_on_cran()
  TG_mmol  <- base_df$TG      # mmol/L
  HDL_mmol <- base_df$HDL_c   # mmol/L
  manual_vai_m <- -(
    (base_df$waist / (39.68 + 1.88 * base_df$bmi)) *
    (TG_mmol / 1.03) *
    (1.31 / HDL_mmol)
  )
  out <- run_adipo(base_df)
  expect_equal(out$VAI_Men_inv, manual_vai_m, tolerance = 1e-8)
})

test_that("Revised_QUICKI matches manual computation", {
  skip_on_cran()
  I0_u <- base_df$I0 / 6
  G0_mgdL <- base_df$G0 * 18
  FFA_v <- base_df$FFA
  manual <- 1 / (log10(I0_u) + log10(G0_mgdL) + log10(FFA_v))
  out <- run_adipo(base_df)
  expect_equal(out$Revised_QUICKI, manual, tolerance = 1e-8)
})

test_that("normalize='range' maps variable columns to [0,1]; constants allowed", {
  skip_on_cran()
  skip_if_not(exists("normalize_vec", where = asNamespace("HealthMarkers")),
              "normalize_vec not available")
  df2 <- bind_rows(base_df, mutate(base_df, G0 = 6))
  out_r <- suppressWarnings(run_adipo(df2, normalize = "range"))
  for (col in names(out_r)) {
    vals <- out_r[[col]]
    finite <- vals[is.finite(vals)]
    if (length(unique(finite)) > 1) {
      expect_gte(min(finite), 0)
      expect_lte(max(finite), 1)
    } else {
      expect_true(length(unique(finite)) == 1)
    }
  }
})

test_that("normalize='z' gives mean≈0 / sd≈1 when variable; constants -> zeros", {
  skip_on_cran()
  skip_if_not(exists("normalize_vec", where = asNamespace("HealthMarkers")),
              "normalize_vec not available")
  df3 <- bind_rows(
    base_df,
    mutate(base_df, G0 = 6),
    mutate(base_df, G0 = 8)
  )
  out_z <- suppressWarnings(run_adipo(df3, normalize = "z"))
  for (col in names(out_z)) {
    vals <- out_z[[col]]
    finite <- vals[is.finite(vals)]
    if (length(unique(finite)) > 1) {
      expect_equal(mean(finite), 0, tolerance = 1e-6)
      expect_equal(sd(finite), 1, tolerance = 1e-6)
    } else {
      expect_true(all(finite == 0))
    }
  }
})

test_that("invalid normalize argument handling depends on normalize_vec", {
  skip_on_cran()
  if (exists("normalize_vec", where = asNamespace("HealthMarkers"))) {
    expect_error(run_adipo(base_df, normalize = "foo"),
                 "`normalize` must be one of|invalid 'method'")
  } else {
    expect_error(run_adipo(base_df, normalize = "foo"), NA)
  }
})

test_that("na_action policies", {
  skip_on_cran()
  df_na <- bind_rows(base_df, mutate(base_df, TG = NA_real_))
  expect_error(
    suppressWarnings(adipo_is(df_na,
                              col_map = list(G0="G0",I0="I0",TG="TG",HDL_c="HDL_c",FFA="FFA",waist="waist",bmi="bmi"),
                              na_action = "error")),
    "required inputs contain missing values"
  )
  out_omit <- suppressWarnings(adipo_is(df_na,
                                        col_map = list(G0="G0",I0="I0",TG="TG",HDL_c="HDL_c",FFA="FFA",waist="waist",bmi="bmi"),
                                        na_action = "omit"))
  expect_equal(nrow(out_omit), 1L)
})

test_that("extreme inputs produce NA in sensitive outputs", {
  skip_on_cran()
  df_ext <- mutate(base_df, HDL_c = 0, TG = 0)
  out_zero <- run_adipo(df_ext, verbose = FALSE)
  expect_true(is.na(out_zero$TG_HDL_C_inv))
})

test_that("verbose = TRUE emits column map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(run_adipo(base_df, verbose = TRUE), "col_map")
  expect_message(run_adipo(base_df, verbose = TRUE), "adipo_is\\(\\): results")
})

test_that("verbose column map lists all required keys", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(run_adipo(base_df, verbose = TRUE))
  map_msgs <- msgs[grepl("col_map", msgs)]
  expect_gte(length(map_msgs), 1L)
  all_map_text <- paste(map_msgs, collapse = " ")
  for (key in c("G0", "I0", "TG", "HDL_c", "FFA", "waist", "bmi")) {
    expect_true(grepl(key, all_map_text), info = paste("key missing from column map msg:", key))
  }
})

test_that("verbose results line reports correct non-NA counts", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  # base_df has 1 complete row -> expect all columns show 1/1
  msgs <- testthat::capture_messages(run_adipo(base_df, verbose = TRUE))
  res_msg <- msgs[grepl("results:", msgs)]
  expect_length(res_msg, 1L)
  expect_true(grepl("Revised_QUICKI 1/1", res_msg))
  expect_true(grepl("McAuley_index 1/1", res_msg))
})

test_that("verbose = TRUE with inform option emits messages exactly once (no double-fire)", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(run_adipo(base_df, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",      msgs)), 1L)
})

test_that("verbose = FALSE suppresses messages even at inform global level", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_no_message(run_adipo(base_df, verbose = FALSE))
})

# Internal registry & utilities
# -----------------------------
# Helper: normalize choices
.hm_normalize_choice <- function(x, choices) {
  if (length(x) == 0) return(choices[1L])
  x <- match.arg(x, choices)
  x
}

# Helper: safely bind new columns
.hm_bind_new_cols <- function(df, addon) {
  if (is.null(addon) || NROW(addon) == 0) return(df)
  if (NROW(addon) != NROW(df)) {
    hm_inform(level = "debug",
              msg   = sprintf("Addon has different number of rows (%d vs %d); skipping bind.",
                              NROW(addon), NROW(df)))
    return(df)
  }
  keep <- setdiff(names(addon), names(df))
  if (length(keep)) cbind(df, addon[keep]) else df
}

# Fallback for MetS so aggregator never errors
.hm_mets_fallback <- function(data) {
  tibble::tibble(MetS_simple = rep(NA_integer_, nrow(data)))
}

# Dispatcher wrapper for obesity_indices (which uses NSE column arguments)
.hm_obesity_indices_dispatch <- function(data, col_map = NULL, verbose = FALSE,
                                         na_action = "keep", ...) {
  allkv <- c("weight", "height", "waist", "hip", "sex")
  cm2     <- .hm_build_col_map(data, col_map, allkv, fn = "obesity_indices")
  col_map2 <- cm2$col_map
  data     <- cm2$data
  for (k in allkv) {
    if (is.null(col_map2[[k]]) && k %in% names(data)) col_map2[[k]] <- k
  }
  req <- c("weight", "height", "waist", "hip")
  miss <- setdiff(req, names(col_map2))
  if (length(miss)) {
    rlang::abort(
      sprintf("obesity_indices(): missing required columns: %s", paste(miss, collapse = ", ")),
      class = "healthmarkers_obesity_error_missing_columns"
    )
  }
  args <- list(
    data      = data,
    weight    = as.name(col_map2[["weight"]]),
    height    = as.name(col_map2[["height"]]),
    waist     = as.name(col_map2[["waist"]]),
    hip       = as.name(col_map2[["hip"]]),
    na_action = na_action,
    verbose   = verbose
  )
  if (!is.null(col_map2[["sex"]])) args$sex <- as.name(col_map2[["sex"]])
  do.call(obesity_indices, args)
}

.hm_marker_registry <- function(verbose = FALSE) {
  reg <- list()
  add <- function(name, fun_name, needs_col_map) {
    f <- get0(fun_name, mode = "function")
    if (is.null(f)) {
      if (isTRUE(verbose)) {
        hm_inform(level = "debug",
          msg = sprintf("Registry: function '%s' not found; skipping group '%s'.", fun_name, name))
      }
    } else {
      reg[[name]] <<- list(fun = f, needs_col_map = needs_col_map)
    }
  }

  # Insulin-related (require col_map)
  add("insulin_fasting",    "fasting_is",            TRUE)
  add("insulin_ogtt",       "ogtt_is",               TRUE)
  add("insulin_adipose",    "adipo_is",              TRUE)
  add("insulin_tracer_dxa", "tracer_dxa_is",         TRUE)

  # Body composition / obesity
  add("adiposity_sds",        "adiposity_sds",                   FALSE)
  add("adiposity_sds_strat",  "adiposity_sds_strat",             FALSE)
  add("obesity_metrics",      ".hm_obesity_indices_dispatch",    FALSE)
  add("alm_bmi",              "alm_bmi_index",                   FALSE)

  # Lipid and atherogenic
  add("lipid",                  "lipid_markers",               FALSE)
  add("atherogenic_indices",    "atherogenic_indices",         FALSE)
  add("atherogenic",            "atherogenic_indices",         FALSE)  # alias
  add("cvd_aip",                "cvd_marker_aip",              FALSE)
  add("cvd_ldl_particles",      "cvd_marker_ldl_particle_number", FALSE)

  # Cardiovascular risk scores
  add("cvd_risk",           "cvd_risk",              FALSE)
  add("cvd_ascvd",          "cvd_risk_ascvd",        FALSE)
  add("cvd_qrisk3",         "cvd_risk_qrisk3",       FALSE)
  add("cvd_scorescvd",      "cvd_risk_scorescvd",    FALSE)
  add("cvd_stroke",         "cvd_risk_stroke",       FALSE)

  # Liver
  add("liver",              "liver_markers",         FALSE)
  add("liver_fat",          "liver_fat_markers",     FALSE)

  # Glycemic
  add("glycemic",           "glycemic_markers",      FALSE)

  # Metabolic syndrome
  add("mets",               "metss",                 FALSE)
  add("metabolic_risk",     "metabolic_risk_features", FALSE)

  # Pulmonary
  add("pulmo",              "pulmo_markers",         FALSE)
  add("spirometry",         "spirometry_markers",    FALSE)
  add("bode",               "bode_index",            FALSE)

  # Saliva, Sweat, Urine
  add("saliva",             "saliva_markers",        FALSE)
  add("sweat",              "sweat_markers",         FALSE)
  add("urine",              "urine_markers",         FALSE)

  # Renal and CKD
  add("renal",              "renal_markers",         FALSE)
  add("kidney_kfre",        "kidney_failure_risk",   FALSE)
  add("ckd_stage",          "ckd_stage",             FALSE)

  # Nutrients and vitamins
  add("nutrient",           "nutrient_markers",      FALSE)
  add("vitamin",            "vitamin_markers",       FALSE)

  # Hormone and inflammation
  add("hormone",            "hormone_markers",       FALSE)
  add("inflammatory",       "inflammatory_markers",  FALSE)

  # Bone, FRAX, and allostatic load
  add("bone",               "bone_markers",          FALSE)
  add("frax",               "frax_score",            FALSE)
  add("allostatic_load",    "allostatic_load",       FALSE)

  # Oxidative stress
  add("oxidative",          "oxidative_markers",     FALSE)

  # Frailty / comorbidity / functional
  add("frailty_index",      "frailty_index",        FALSE)
  add("charlson",           "charlson_index",       FALSE)
  add("sarc_f",             "sarc_f_score",         FALSE)

  # Psychiatric markers
  add("psych",              "psych_markers",        FALSE)

  # Neuro / aging markers
  add("nfl",                "nfl_marker",           FALSE)
  add("iAge",               "iAge",                 FALSE)
  add("inflammatory_age",   "inflammatory_age",     FALSE)

  # Micronutrient / vitamin sub-panels
  add("vitamin_d_status",   "vitamin_d_status",     FALSE)

  # Single biochemical ratios / corrections
  add("calcium_corrected",  "corrected_calcium",    FALSE)
  add("kyn_trp",            "kyn_trp_ratio",        FALSE)

  reg
}

# Required key helper for group-specific inference
# Returns a character vector of keys we should attempt to infer for the
# requested groups. Groups not listed here either do not require col_map or
# have uncertain inputs, so we avoid over-constraining inference.
.hm_group_required_keys <- function(which_vec, include_insulin = TRUE) {
  key_map <- list(
    insulin_fasting    = c("G0", "I0"),
    insulin_ogtt       = c("G0", "I0", "G30", "I30", "G120", "I120"),
    insulin_adipose    = c("BMI", "waist", "TG", "HDL_c"),
    insulin_tracer_dxa = c("fat_mass", "rate_palmitate", "rate_glycerol"),

    adiposity_sds_strat = c("BMI", "age", "sex"),
    alm_bmi             = c("ALM", "BMI", "age", "sex"),

    lipid                  = c("TG", "HDL_c", "LDL_c", "TC"),
    atherogenic_indices    = c("TG", "HDL_c", "LDL_c", "TC"),
    atherogenic            = c("TG", "HDL_c", "LDL_c", "TC"),
    cvd_aip                = c("TG", "HDL_c"),
    cvd_ldl_particles      = c("ApoB"),

    liver_fat = c("ALT", "AST", "BMI", "sex"),

    metabolic_risk = c("waist", "bp_sys", "bp_dia", "glucose", "TG", "HDL_c", "sex", "race"),

    spirometry = c("fev1", "fvc", "height", "age", "sex"),
    bode       = c("fev1", "fvc", "height", "age", "sex", "BMI"),
    pulmo      = c("fev1", "fvc", "age", "sex"),

    renal        = c("eGFR", "UACR"),
    kidney_kfre  = c("age", "sex", "eGFR", "UACR"),
    ckd_stage    = c("eGFR", "UACR"),

    vitamin_d_status = c("vitamin_d", "vitD", "VitD"),
    calcium_corrected = c("calcium", "albumin"),
    kyn_trp           = c("kynurenine", "tryptophan")
  )

  # Insulin panel is computed up-front when include_insulin = TRUE
  insulin_keys <- if (isTRUE(include_insulin)) {
    unlist(key_map[c("insulin_fasting", "insulin_ogtt", "insulin_adipose", "insulin_tracer_dxa")], use.names = FALSE)
  } else character(0)

  req <- c(insulin_keys, unlist(key_map[intersect(names(key_map), which_vec)], use.names = FALSE))
  unique(req)
}

# Prepare data for specific groups (small convenience fixes)
.hm_prepare_for_group <- function(data, grp, col_map = NULL) {
  out <- data

  # Helper: resolve data column via col_map key or ordered synonym list
  .prep_resolve <- function(key, synonyms = character(0)) {
    if (!is.null(col_map)) {
      cm_val <- col_map[[key]]
      if (!is.null(cm_val) && cm_val %in% names(out)) return(cm_val)
    }
    for (syn in synonyms) if (syn %in% names(out)) return(syn)
    NULL
  }

  if (identical(grp, "mets")) {
    if (!("bp_sys" %in% names(out))) {
      src <- .prep_resolve("sbp", c("sbp", "SBP", "sysbp", "systolic", "sys_bp", "systolic_bp"))
      if (!is.null(src)) out$bp_sys <- out[[src]]
    }
    if (!("bp_dia" %in% names(out))) {
      src <- .prep_resolve("dbp", c("dbp", "DBP", "diabp", "diastolic", "dia_bp", "diastolic_bp"))
      if (!is.null(src)) out$bp_dia <- out[[src]]
    }
    if (!("glucose" %in% names(out))) {
      src <- .prep_resolve("G0", c("G0", "pglu0", "bglu0", "glu0", "fasting_glucose",
                                   "glucose0", "glucose_fasting", "fpg", "FPG", "GLUC", "nglu0"))
      if (!is.null(src)) out$glucose <- out[[src]]
    }
    if (!("TG" %in% names(out))) {
      src <- .prep_resolve("TG", c("trig", "Trig", "TRIG", "triglycerides", "trigl", "TGly"))
      if (!is.null(src)) out$TG <- out[[src]]
    }
    if (!("HDL_c" %in% names(out))) {
      src <- .prep_resolve("HDL_c", c("hdlc", "HDLC", "HDL", "hdl", "HDLc", "hdl_c"))
      if (!is.null(src)) out$HDL_c <- out[[src]]
    }
    if (!("waist" %in% names(out))) {
      src <- .prep_resolve("waist", c("WC", "wc", "waistc", "waist_c", "waist_circumference",
                                      "abdominal_circumference", "abdcirc"))
      if (!is.null(src)) out$waist <- out[[src]]
    }
    if (!("triglycerides" %in% names(out)) && "TG" %in% names(out)) out$triglycerides <- out$TG
    if (!("bp_treated" %in% names(out))) out$bp_treated <- FALSE
    if (!("smoker" %in% names(out))) out$smoker <- FALSE
    if (!("diabetes" %in% names(out))) out$diabetes <- FALSE

    if (!("sex" %in% names(out))) {
      out$sex <- "M"
    } else {
      s <- out$sex
      if (is.numeric(s)) {
        out$sex <- ifelse(!is.na(s) & s == 2, "F", "M")
      } else {
        s <- toupper(as.character(s))
        out$sex <- ifelse(startsWith(s, "F"), "F", "M")
      }
    }

    if (!("race" %in% names(out)) && "ethnicity" %in% names(out)) out$race <- out$ethnicity
    if ("race" %in% names(out)) {
      r <- toupper(as.character(out$race))
      out$race <- ifelse(grepl("^NHW|WHITE|CAUC", r), "NHW",
                  ifelse(grepl("^NHB|BLACK|AFRIC", r), "NHB",
                  ifelse(grepl("HISP|LATIN", r), "HISP", "Other")))
    } else {
      out$race <- "NHW"
    }
  }

  if (identical(grp, "urine")) {
    if (!("urine_albumin" %in% names(out))) {
      src <- .prep_resolve("UACR", c("ualb", "urine_alb", "u_albumin", "alb_urine",
                                     "urinary_albumin", "albumin_urine", "u_alb"))
      if (!is.null(src)) out$urine_albumin <- out[[src]]
    }
    if (!("urine_creatinine" %in% names(out))) {
      src <- .prep_resolve("creatinine", c("ucrea", "urine_crea", "u_creatinine",
                                           "crea_urine", "urinary_creatinine",
                                           "creatinine_urine", "u_crea"))
      if (!is.null(src)) out$urine_creatinine <- out[[src]]
    }
  }

  if (identical(grp, "pulmo")) {
    if (!("ethnicity" %in% names(out))) {
      src <- .prep_resolve("race", c("race", "Race", "ethnic", "ethn", "ethnicity_code"))
      if (!is.null(src)) out$ethnicity <- out[[src]]
      else out$ethnicity <- "Other"   # default so GLI reference works
    }
    if (!("fev1" %in% names(out))) {
      src <- .prep_resolve("FEV1", c("FEV1", "fev1_abs", "FEV1_abs"))
      if (!is.null(src)) out$fev1 <- out[[src]]
    }
    if (!("fvc" %in% names(out))) {
      src <- .prep_resolve("FVC", c("FVC", "fvc_abs", "FVC_abs"))
      if (!is.null(src)) out$fvc <- out[[src]]
    }
    if (!("height" %in% names(out))) {
      src <- .prep_resolve("height", c("ht", "hgt", "height_cm", "HEIGHT", "Ht", "HGT"))
      if (!is.null(src)) out$height <- out[[src]]
    }
  }

  out
}

.hm_safe_call <- function(fun, data, col_map, needs_col_map, verbose, tag, extra_args = list()) {
  hm_inform(level = if (isTRUE(verbose)) "inform" else "debug", msg = paste0("health_markers(): calling ", tag))

  args <- list(data)
  fn_formals <- tryCatch(formals(fun), error = function(e) NULL)
  fn_names <- if (is.null(fn_formals)) character(0) else names(fn_formals)

  if (isTRUE(needs_col_map) && "col_map" %in% fn_names) args$col_map <- col_map

  if (!is.null(fn_formals)) {
    has_dots <- any(fn_names == "...")
    if (!has_dots) {
      extra_args <- extra_args[intersect(names(extra_args), fn_names)]
    }
  } else {
    extra_args <- list()
  }

  if (!is.null(fn_formals)) {
    req <- names(fn_formals)[vapply(fn_formals, function(x) identical(x, quote(expr = )), logical(1))]
    first_arg <- if (length(fn_names)) fn_names[1L] else NULL
    req <- setdiff(req, c("...", first_arg))
    planned <- union(names(args), names(extra_args))
    missing_required <- setdiff(req, planned)
    if (length(missing_required)) {
      if (isTRUE(verbose)) hm_inform(level = "debug",
        msg = sprintf("Skipping '%s': missing required args: %s", tag, paste(missing_required, collapse = ", ")))
      return(NULL)
    }
  }

  call_fun <- function() do.call(fun, c(args, extra_args))
  handle_error <- function(e) {
    msg <- conditionMessage(e)
    hm_inform(level = "debug", msg = sprintf("'%s' failed: %s", tag, msg))
    pkg <- if (!is.null(e$package)) {
      e$package
    } else if (grepl("Package '([A-Za-z0-9._]+)' is required", msg)) {
      sub(".*Package '([A-Za-z0-9._]+)' is required.*", "\\1", msg)
    } else if (grepl("package (.+) is required", msg, ignore.case = TRUE)) {
      sub(".*package ([A-Za-z0-9._]+) is required.*", "\\1", msg)
    } else {
      NULL
    }
    structure(list(), hm_error = msg, hm_error_class = class(e)[1], hm_missing_pkg = pkg)
  }

  if (isTRUE(verbose)) {
    tryCatch(call_fun(), error = handle_error)
  } else {
    tryCatch(suppressWarnings(call_fun()), error = handle_error)
  }
}

# -----------------------------
# 1. All insulin indices
# -----------------------------
#' Compute insulin sensitivity/resistance panels (fasting, OGTT, adipose, tracer/DXA)
#'
#' @note For scholarly references to specific indices (e.g., HOMA-IR, QUICKI,
#' Raynaud, Belfiore, tracer-derived indices, adiposity-related IS metrics),
#' consult the individual function help pages (e.g. ?fasting_is, ?ogtt_is,
#' ?adipo_is, ?tracer_dxa_is). Citations are intentionally not duplicated here.
#'
#' @param data A data.frame or tibble of raw measurements.
#' @param col_map Named list with keys G0,I0,G30,I30,G120,I120,TG,HDL_c,FFA,waist,weight,bmi,age,sex,rate_palmitate,rate_glycerol,fat_mass.
#' @param normalize One of c("none","z","inverse","range","robust").
#' @param mode One of c("IS","IR","both"). "IR" returns only inverted IR, "IS" only the original IS, "both" returns both with IR_ prefix.
#' @param verbose Logical.
#' @param na_action One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).
#' @return A tibble of IS (and/or IR_) columns.
#' @note
#' Aggregator wrapper. See underlying function help pages for full references:
#' fasting_is(), ogtt_is(), adipo_is(), tracer_dxa_is().
#' @references
#' \insertRef{suleman2024is}{HealthMarkers}
#' @export
#' @examples
#' # Quick smoke-test (fasting indices only)
#' df <- data.frame(G0 = 5.2, I0 = 60)
#' all_insulin_indices(df, normalize = "none", mode = "IS",
#'                     verbose = FALSE, na_action = "keep")
#'
#' \donttest{
#' # Full panel with all supported inputs
#' df <- data.frame(
#'   G0 = 5.2, I0 = 60, G30 = 7.5, I30 = 90, G120 = 6.2, I120 = 80,
#'   TG = 1.5, HDL_c = 1.3, FFA = 0.3, waist = 85, weight = 70, bmi = 24,
#'   age = 40, sex = "M", rate_palmitate = 0.1, rate_glycerol = 0.2, fat_mass = 20
#' )
#' all_insulin_indices(df, col_map = list(
#'   G0="G0", I0="I0", G30="G30", I30="I30", G120="G120", I120="I120",
#'   TG="TG", HDL_c="HDL_c", FFA="FFA", waist="waist", weight="weight",
#'   bmi="bmi", age="age", sex="sex", rate_palmitate="rate_palmitate",
#'   rate_glycerol="rate_glycerol", fat_mass="fat_mass"
#' ), normalize = "none", mode = "IS", verbose = FALSE, na_action = "keep")
#' }
all_insulin_indices <- function(
  data,
  col_map = NULL,
  normalize = c("none","z","inverse","range","robust"),
  mode = c("both","IS","IR"),
  verbose = TRUE,
  na_action = c("keep","omit","error")
) {
  normalize <- .hm_normalize_choice(normalize, c("none","z","inverse","range","robust"))
  mode <- .hm_normalize_choice(mode, c("both","IS","IR"))
  na_action <- match.arg(na_action)
  hm_inform("all_insulin_indices(): preparing inputs", level = if (isTRUE(verbose)) "inform" else "debug")

  common_args <- list(normalize = normalize, na_action = na_action, verbose = verbose)

  pieces <- list(
    fasting_is    = .hm_safe_call(fasting_is,    data, col_map, TRUE, verbose, "fasting",    common_args),
    ogtt_is       = .hm_safe_call(ogtt_is,       data, col_map, TRUE, verbose, "OGTT",       common_args),
    adipo_is      = .hm_safe_call(adipo_is,      data, col_map, TRUE, verbose, "adipose",    common_args),
    tracer_dxa_is = .hm_safe_call(tracer_dxa_is, data, col_map, TRUE, verbose, "tracer/DXA", common_args)
  )
  pieces <- Filter(is.data.frame, pieces)
  if (!length(pieces)) return(tibble::tibble())
  # Drop duplicate id columns — keep only from the first piece
  id_nm <- .hm_detect_id_col(data)
  if (!is.null(id_nm)) {
    for (i in seq_along(pieces)[-1L]) {
      pieces[[i]] <- pieces[[i]][, setdiff(names(pieces[[i]]), id_nm), drop = FALSE]
    }
  }
  is_tbl <- dplyr::bind_cols(pieces)

  if (mode == "IS") return(is_tbl)

  # IR inversion -- only numeric columns; skip id/character columns
  numeric_nms <- names(is_tbl)[vapply(is_tbl, is.numeric, logical(1))]
  ir_tbl <- as.data.frame(
    lapply(numeric_nms, function(nm) {
      x <- is_tbl[[nm]]
      ifelse(is.na(x) | x == 0, NA_real_, 1 / x)
    }),
    stringsAsFactors = FALSE
  )
  names(ir_tbl) <- paste0("IR_", numeric_nms)

  if (mode == "IR") return(ir_tbl)

  dplyr::bind_cols(is_tbl, ir_tbl)
}

# -----------------------------
# 2. Metabolic markers (compat)
# -----------------------------
#' Aggregate selected metabolic marker groups
#'
#' @note For references supporting liver, lipid, glycemic, MetS, adiposity and
#' other domain-specific indices, see each underlying function's documentation
#' (e.g. ?liver_markers, ?lipid_markers, ?glycemic_markers, ?metss, ?adiposity_sds).
#' This wrapper omits repeated reference listings to avoid redundancy.
#'
#' @param data A data.frame or tibble.
#' @param col_map Named list for column mapping forwarded to underlying functions.
#' @param which Character vector of groups to compute: c("insulin","adiposity_sds","cardio","lipid","liver","glycemic","mets").
#' @param normalize One of c("none","z","inverse","range","robust").
#' @param mode One of c("both","IS","IR").
#' @param verbose Logical.
#' @param na_action One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).
#' @return Data frame with original columns plus derived markers.
#' @note
#' Aggregator wrapper. See underlying function help pages for full references:
#' all_insulin_indices(), lipid_markers(), liver_markers(), glycemic_markers(), metss().
#' @export
#' @examples
#' \donttest{
#' df <- data.frame(
#'   TC = 200, HDL_c = 50, TG = 150, LDL_c = 120,
#'   ALT = 30, AST = 20, BMI = 25
#' )
#' metabolic_markers(df, col_map = list(), which = c("lipid","liver"),
#'                   normalize = "none", mode = "both", verbose = FALSE, na_action = "keep")
#' }
metabolic_markers <- function(
  data,
  col_map = NULL,
  which = c("insulin","adiposity_sds","cardio","lipid","liver","glycemic","mets"),
  normalize = c("none","z","inverse","range","robust"),
  mode = c("both","IS","IR"),
  verbose = TRUE,
  na_action = c("keep","omit","error")
) {
  normalize <- .hm_normalize_choice(normalize, c("none","z","inverse","range","robust"))
  mode <- .hm_normalize_choice(mode, c("both","IS","IR"))
  which <- match.arg(which, several.ok = TRUE)
  na_action <- match.arg(na_action)
  hm_inform("metabolic_markers(): preparing inputs", level = if (isTRUE(verbose)) "inform" else "debug")

  out <- data

  if ("insulin" %in% which) {
    add <- .hm_safe_call(
      all_insulin_indices, out, col_map, TRUE, verbose, "insulin_panel",
      list(normalize = normalize, mode = mode, verbose = verbose, na_action = na_action)
    )
    if (is.data.frame(add)) out <- .hm_bind_new_cols(out, add)
  }

  if ("adiposity_sds" %in% which) {
    add <- .hm_safe_call(get0("adiposity_sds", mode = "function"), out, col_map, FALSE, verbose, "adiposity_sds",
                         list(verbose = verbose, na_action = na_action))
    if (is.data.frame(add)) out <- .hm_bind_new_cols(out, add)
  }

  if ("cardio" %in% which) {
    cr <- suppressWarnings(try(cvd_risk(out), silent = TRUE))
    if (!inherits(cr, "try-error")) {
      out <- .hm_bind_new_cols(out, cr)
    } else if (isTRUE(verbose)) {
      hm_inform(level = "debug", msg = "health_markers(): cardio skipped (cvd_risk unavailable)")
    }
  }

  if ("lipid" %in% which) {
    add <- .hm_safe_call(lipid_markers, out, col_map, FALSE, verbose, "lipid",
                         list(verbose = verbose, na_action = na_action))
    if (is.data.frame(add)) out <- .hm_bind_new_cols(out, add)
  }

  if ("liver" %in% which) {
    out2 <- .hm_prepare_for_group(out, "liver", col_map)
    add <- .hm_safe_call(liver_markers, out2, col_map, FALSE, verbose, "liver",
                         list(verbose = verbose, na_action = na_action))
    if (is.data.frame(add)) out <- .hm_bind_new_cols(out, add)
  }

  if ("glycemic" %in% which) {
    add <- .hm_safe_call(glycemic_markers, out, col_map, TRUE, verbose, "glycemic",
                         list(verbose = verbose, na_action = na_action))
    if (is.data.frame(add)) out <- .hm_bind_new_cols(out, add)
  }

  if ("mets" %in% which) {
    out_m <- .hm_prepare_for_group(out, "mets", col_map)
    mets_add <- .hm_safe_call(metss, out_m, col_map, FALSE, verbose, "mets",
                              list(verbose = verbose, na_warn_prop = 0, na_action = na_action))
    if (!is.data.frame(mets_add)) mets_add <- .hm_mets_fallback(out_m)
    if (is.data.frame(mets_add)) out <- .hm_bind_new_cols(out, mets_add)
  }

  out
}

# -----------------------------
# 3. Comprehensive aggregator
# -----------------------------
#' Compute all available HealthMarkers categories
#'
#' @note For academic / clinical references tied to each derived marker or
#' index, consult the help pages of the source functions (e.g. ?allostatic_load,
#' ?bone_markers, ?vitamin_markers, ?inflammatory_markers, etc.). This
#' aggregator provides integration only and does not restate citations.
#'
#' @details
#' Common group names for `which` include:
#' \itemize{
#'   \item \code{"lipid"}, \code{"liver"}, \code{"glycemic"}, \code{"mets"}, \code{"oxidative"}
#'   \item \code{"bone"}, \code{"allostatic_load"}, \code{"nutrient"}, \code{"vitamin"}, \code{"vitamin_d_status"}
#'   \item \code{"renal"}, \code{"ckd_stage"}, \code{"kidney_kfre"}
#'   \item \code{"frailty_index"}, \code{"charlson"}, \code{"sarc_f"}
#'   \item \code{"nfl"}, \code{"iAge"}, \code{"calcium_corrected"}, \code{"kyn_trp"}
#' }
#'
#' @param data A data.frame or tibble.
#' @param col_map Named list for column mapping forwarded to underlying functions.
#'   If `col_map` is `NULL` or an empty list, `all_health_markers()` calls
#'   [hm_col_report()] once at the top level to guess a column map from common
#'   synonyms (for example `TG` vs `triglycerides`, `BMI` vs `bmi`,
#'   `HDL_c` vs `HDL`). The inferred `col_map` is then reused for all groups
#'   that require it, and an error is thrown if required keys (e.g. `TG`,
#'   `HDL_c`, `LDL_c`, `TC`, `BMI`, `age`, `sex`) cannot be inferred.
#' @param which "all" or a vector of registry keys (see Details).
#' @param include_insulin Logical; include all_insulin_indices() first.
#' @param normalize One of c("none","z","inverse","range","robust").
#' @param mode One of c("both","IS","IR") passed to insulin indices.
#' @param verbose Logical.
#' @param na_action One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).
#' @param id_col Optional character string naming a column in `data` to include
#'   in the returned output when `return_input = FALSE`. Ignored when
#'   `return_input = TRUE`. Typical use: participant ID or sample barcode.
#' @param return_input Logical (default `TRUE`). When `TRUE` the original input
#'   columns are retained in the output (current behaviour). When `FALSE` only
#'   the newly computed marker columns are returned, plus `id_col` if supplied.
#'   Set to `FALSE` to avoid carrying a large input data frame through the
#'   pipeline — join back later with `cbind()` or `dplyr::left_join()` on `id_col`.
#' @return Data frame. When `return_input = TRUE` (default): original columns
#'   plus all derived markers. When `return_input = FALSE`: only the newly
#'   computed columns (and `id_col` if specified).
#' @note
#' Aggregator wrapper. See underlying function help pages for full references
#' across categories included by `which`.
#' @export
#' @examples
#' # Quick smoke-test (lipid group only, no insulin)
#' df <- data.frame(TC = 200, HDL_c = 50, TG = 150, LDL_c = 120)
#' all_health_markers(df, col_map = list(), which = "lipid",
#'                    include_insulin = FALSE, normalize = "none",
#'                    verbose = FALSE, na_action = "keep")
#'
#' \donttest{
#' # Lipid + liver groups
#' df <- data.frame(
#'   TC = 200, HDL_c = 50, TG = 150, LDL_c = 120,
#'   ALT = 30, AST = 20, BMI = 25
#' )
#' all_health_markers(df, col_map = list(), which = c("lipid","liver"),
#'                    include_insulin = FALSE, normalize = "none", mode = "both",
#'                    verbose = FALSE, na_action = "keep")
#' }
all_health_markers <- function(
  data,
  col_map,
  which = "all",
  include_insulin = TRUE,
  normalize = c("none","z","inverse","range","robust"),
  mode = c("both","IS","IR"),
  verbose = TRUE,
  na_action = c("keep","omit","error"),
  id_col = NULL,
  return_input = TRUE
) {
  orig_col_map <- if (missing(col_map)) NULL else col_map
  orig_cols    <- names(data)

  normalize <- .hm_normalize_choice(normalize, c("none","z","inverse","range","robust"))
  mode <- .hm_normalize_choice(mode, c("both","IS","IR"))
  na_action <- match.arg(na_action)
  hm_inform("all_health_markers(): preparing inputs", level = if (isTRUE(verbose)) "inform" else "debug")

  reg <- .hm_marker_registry(verbose = isTRUE(verbose))
  reg_names <- names(reg)

  if (identical(which, "all")) {
    # Exclude alias entries that duplicate another group's function call
    which_vec <- setdiff(reg_names, c("atherogenic_indices", "cvd_risk"))
  } else {
    unknown <- setdiff(which, reg_names)
    if (length(unknown)) {
      rlang::abort(
        paste0("Unknown marker group(s): ", paste(unknown, collapse = ", ")),
        class = "healthmarkers_health_markers_error_unknown_group"
      )
    }
    which_vec <- which
  }

  # Auto-infer col_map if not supplied -- best-effort, never throws
  if (missing(col_map) || is.null(col_map)) {
    # Use hm_col_report() internally: matches all keys it can, silently skips the rest
    col_map <- tryCatch(
      hm_col_report(data, col_map = NULL, verbose = FALSE, fuzzy = FALSE, show_unmatched = FALSE),
      error = function(e) {
        hm_inform(level = "debug", msg = sprintf(
          "all_health_markers(): column inference failed (%s); proceeding with empty col_map",
          conditionMessage(e)))
        list()
      }
    )
  }

  # --- Tier 0: derive globally useful variables BEFORE the verbose summary
  # so the summary can correctly classify derived vs not-found keys
  out <- data
  gp  <- .hm_global_precompute(out, col_map, verbose = FALSE)
  out <- gp$data
  # Merge newly derived keys into col_map (key -> same-name column)
  for (k in names(gp$derived_map)) {
    if (is.null(col_map[[k]])) col_map[[k]] <- k
  }

  # Track which keys were mapped vs not found (for verbose summary)
  all_pattern_keys <- names(.hm_default_col_patterns_exact())
  user_keys <- if (is.null(orig_col_map)) character(0) else
    names(orig_col_map)[!vapply(orig_col_map, is.null, logical(1))]
  did_auto_infer <- is.null(orig_col_map)  # TRUE when we ran hm_col_report() above
  inferred_keys  <- if (did_auto_infer) setdiff(names(col_map), names(gp$derived_map)) else character(0)
  derived_keys   <- if (did_auto_infer) intersect(names(gp$derived_map), all_pattern_keys) else character(0)
  not_found_keys <- if (did_auto_infer) setdiff(all_pattern_keys, names(col_map)) else character(0)

  if (isTRUE(verbose)) {
    hm_inform(sprintf("all_health_markers(): mapping summary -- %d key(s) mapped",
                      length(names(col_map))), level = "inform")
    rule <- paste(rep("\u2500", 55L), collapse = "")
    cat(sprintf("\u2500\u2500 all_health_markers(): column mapping %s\n", rule))
    cat(sprintf(" Data: %d row%s \u00d7 %d column%s\n",
                nrow(data), if (nrow(data) != 1L) "s" else "",
                ncol(data), if (ncol(data) != 1L) "s" else ""))

    if (length(user_keys)) {
      cat(sprintf("\n User-supplied (%d):\n", length(user_keys)))
      for (k in user_keys)
        cat(sprintf("   \u2714 %-24s -> %s\n", k, col_map[[k]]))
    }
    if (did_auto_infer && length(inferred_keys)) {
      cat(sprintf("\n Auto-detected (%d):\n", length(inferred_keys)))
      for (k in inferred_keys)
        cat(sprintf("   \u2714 %-24s -> %s\n", k, col_map[[k]]))
    }
    if (did_auto_infer && length(derived_keys)) {
      cat(sprintf("\n Derived (%d):\n", length(derived_keys)))
      for (k in derived_keys)
        cat(sprintf("   \u25b6 %-24s (computed from available inputs)\n", k))
    }
    if (did_auto_infer && length(not_found_keys)) {
      cat(sprintf("\n Not found (%d) \u2014 groups needing these keys will be skipped:\n",
                  length(not_found_keys)))
      for (k in not_found_keys)
        cat(sprintf("   \u2718 %s\n", k))
    }
    cat(sprintf(" %s\n\n",
                paste(rep("\u2500", 87L), collapse = "")))
  }

  group_status <- list()

  if (isTRUE(include_insulin)) {
    ins <- .hm_safe_call(
      all_insulin_indices, out, col_map, TRUE, verbose, "insulin_panel",
      extra_args = list(
        normalize = normalize,
        mode      = mode,
        verbose   = verbose,
        na_action = na_action
      )
    )
    status <- list(state = if (is.data.frame(ins)) "ok" else "skipped_or_failed",
                   message = attr(ins, "hm_error", exact = TRUE),
                   missing_pkg = attr(ins, "hm_missing_pkg", exact = TRUE))
    if (is.data.frame(ins)) out <- .hm_bind_new_cols(out, ins)
    group_status[["insulin_panel"]] <- status
  }

  for (grp in which_vec) {
    if (isTRUE(include_insulin) && startsWith(grp, "insulin_")) {
      group_status[[grp]] <- list(state = "skipped", message = "covered by insulin_panel", missing_pkg = NULL)
      next
    }

    entry <- reg[[grp]]
    if (is.null(entry)) {
      group_status[[grp]] <- list(state = "skipped", message = "not in registry", missing_pkg = NULL)
      next
    }

    data2 <- .hm_prepare_for_group(out, grp, col_map)
    addon <- .hm_safe_call(
      entry$fun, data2, col_map, entry$needs_col_map, verbose, grp,
      extra_args = list(
        verbose   = verbose,
        na_action = na_action,
        normalize = normalize,
        `return`  = "data"
      )
    )
    status <- list(state = if (is.data.frame(addon)) "ok" else "skipped_or_failed",
                   message = attr(addon, "hm_error", exact = TRUE),
                   missing_pkg = attr(addon, "hm_missing_pkg", exact = TRUE))
    if (is.data.frame(addon)) out <- .hm_bind_new_cols(out, addon)
    group_status[[grp]] <- status
  }

  if (isTRUE(verbose) && length(group_status)) {
    ok <- names(group_status)[vapply(group_status, function(x) identical(x$state, "ok"), logical(1))]
    other <- names(group_status)[vapply(group_status, function(x) !identical(x$state, "ok"), logical(1))]

    parts <- character()
    if (length(ok)) {
      parts <- c(parts, sprintf("computed: %s", paste(ok, collapse = ", ")))
    }
    if (length(other)) {
      detail <- vapply(other, function(nm) {
        st <- group_status[[nm]]
        extras <- c(
          if (!is.null(st$missing_pkg)) sprintf("missing package: %s", st$missing_pkg) else NULL,
          if (!is.null(st$message)) sprintf("error: %s", st$message) else NULL
        )
        if (length(extras)) sprintf("%s (%s)", nm, paste(extras, collapse = "; ")) else nm
      }, character(1))
      parts <- c(parts, sprintf("skipped/failed: %s", paste(detail, collapse = "; ")))
    }

    if (length(parts)) {
      hm_inform(paste0("all_health_markers(): ", paste(parts, collapse = "; ")), level = "inform")
      cat(sprintf("\u2500\u2500 all_health_markers(): group summary %s\n",
                  paste(rep("\u2500", 55L), collapse = "")))
      if (length(ok)) {
        cat(sprintf(" Computed (%d):\n", length(ok)))
        for (nm in ok) cat(sprintf("   \u2714 %s\n", nm))
      }
      if (length(other)) {
        cat(sprintf("\n Skipped / failed (%d):\n", length(other)))
        for (nm in other) {
          st <- group_status[[nm]]
          extras <- c(
            if (!is.null(st$missing_pkg)) sprintf("missing package: %s", st$missing_pkg) else NULL,
            if (!is.null(st$message))     sprintf("%s", st$message) else NULL
          )
          reason <- if (length(extras)) paste(extras, collapse = "; ") else st$state
          cat(sprintf("   \u2718 %-30s %s\n", nm, reason))
        }
      }
      cat(sprintf(" %s\n", paste(rep("\u2500", 87L), collapse = "")))
    }
  }

  if (!isTRUE(return_input)) {
    new_cols <- setdiff(names(out), orig_cols)
    keep     <- if (!is.null(id_col) && id_col %in% names(out)) c(id_col, new_cols) else new_cols
    return(out[keep])
  }
  out
}

## Integration tests for all_health_markers() — Phase E DAG pipeline
## Verifies that .hm_global_precompute() (Tier 0) feeds derived columns to
## downstream groups within the same all_health_markers() call.

# ---------------------------------------------------------------------------
# eGFR pipeline: creatinine + age + sex -> eGFR (Tier 0) -> ckd_stage (Tier 2)
# ---------------------------------------------------------------------------
test_that("all_health_markers: global precompute derives eGFR and feeds ckd_stage", {
  skip_on_cran()
  df <- data.frame(
    creatinine = c(0.9, 1.5),
    age        = c(45,  70),
    sex        = c("F", "M")
  )
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = c("renal", "ckd_stage"),
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_s3_class(out, "data.frame")
  expect_equal(nrow(out), 2L)
  # Global precompute must have added eGFR
  expect_true("eGFR" %in% names(out))
  expect_true(all(is.finite(out[["eGFR"]])))
  # ckd_stage downstream group must have run
  expect_true("CKD_stage" %in% names(out))
  expect_false(any(is.na(out[["CKD_stage"]])))
})

test_that("all_health_markers: ckd_stage resolves even without the 'renal' group", {
  skip_on_cran()
  # Only creatinine+age+sex — no renal group; eGFR comes solely from global precompute
  df <- data.frame(
    creatinine = c(1.2),
    age        = c(55),
    sex        = c("M")
  )
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = "ckd_stage",
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_true("eGFR"      %in% names(out))
  expect_true("CKD_stage" %in% names(out))
  expect_true(is.finite(out[["eGFR"]]))
})

# ---------------------------------------------------------------------------
# BMI pipeline: weight + height -> BMI (Tier 0) -> used by lipid/glycemic
# ---------------------------------------------------------------------------
test_that("all_health_markers: global precompute derives BMI from weight and height", {
  skip_on_cran()
  df <- data.frame(
    weight = 80.0, height = 175.0,
    TC = 5.2, HDL_c = 1.3, TG = 1.8
  )
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = "lipid",
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_true("BMI" %in% names(out))
  expect_equal(round(out[["BMI"]], 2), round(80 / (1.75^2), 2))
})

# ---------------------------------------------------------------------------
# LDL_c pipeline: TC + HDL_c + TG -> LDL_c (Tier 0) -> lipid_markers ratios
# ---------------------------------------------------------------------------
test_that("all_health_markers: global precompute derives LDL_c and lipid group uses it", {
  skip_on_cran()
  df <- data.frame(TC = 5.2, HDL_c = 1.3, TG = 1.8)
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = "lipid",
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_true("LDL_c" %in% names(out))
  expect_true(is.finite(out[["LDL_c"]]))
  # LDL_c / HDL_c ratio should also exist (produced by lipid_markers)
  expect_true(any(grepl("LDL", names(out))))
})

# ---------------------------------------------------------------------------
# Glucose alias pipeline: G0 -> glucose (Tier 0) -> glycemic group
# ---------------------------------------------------------------------------
test_that("all_health_markers: G0 aliased to glucose by global precompute", {
  skip_on_cran()
  df <- data.frame(G0 = 5.5, I0 = 10.0, age = 35, sex = "M", BMI = 25)
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(G0 = "G0", I0 = "I0"),
    which          = "glycemic",
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_true("glucose" %in% names(out))
  expect_equal(out[["glucose"]], 5.5)
})

# ---------------------------------------------------------------------------
# UACR pipeline: urine_albumin + urine_creatinine -> UACR (Tier 0) -> ckd_stage
# ---------------------------------------------------------------------------
test_that("all_health_markers: UACR derived by global precompute and used by ckd_stage", {
  skip_on_cran()
  df <- data.frame(
    creatinine       = 1.1,
    age              = 60,
    sex              = "F",
    urine_albumin    = 45.0,
    urine_creatinine = 150.0
  )
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = "ckd_stage",
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  expect_true("UACR" %in% names(out))
  expect_equal(out[["UACR"]], 45.0 / 150.0)
  expect_true("Albuminuria_stage" %in% names(out))
  # UACR = 0.3 -> A1 category
  expect_equal(as.character(out[["Albuminuria_stage"]]), "A1")
})

# ---------------------------------------------------------------------------
# No-data graceful skip: ckd_stage absent when no eGFR inputs provided
# ---------------------------------------------------------------------------
test_that("all_health_markers: ckd_stage skipped gracefully when no eGFR data", {
  skip_on_cran()
  df <- data.frame(TC = 5.0, HDL_c = 1.2, TG = 1.8)
  out <- suppressWarnings(all_health_markers(
    df,
    col_map        = list(),
    which          = c("lipid", "ckd_stage"),
    include_insulin = FALSE,
    verbose        = FALSE,
    na_action      = "keep"
  ))
  # lipid should have run
  expect_true("non_HDL_c" %in% names(out) || any(grepl("LDL|HDL|TC", names(out))))
  # ckd_stage should NOT have produced CKD_stage (no creatinine in data)
  expect_false("CKD_stage" %in% names(out))
})

# ---------------------------------------------------------------------------
# .hm_safe_call() error attribute structure
# ---------------------------------------------------------------------------
test_that(".hm_safe_call() returns hm_error attribute when function throws", {
  skip_on_cran()
  bad_fun <- function(data, ...) stop("deliberate test failure")
  result <- HealthMarkers:::.hm_safe_call(
    fun           = bad_fun,
    data          = data.frame(x = 1),
    col_map       = list(),
    needs_col_map = FALSE,
    verbose       = FALSE,
    tag           = "test_tag"
  )
  expect_false(is.null(attr(result, "hm_error")))
  expect_match(attr(result, "hm_error"), "deliberate test failure")
})

test_that(".hm_safe_call() captures missing-package error and sets hm_missing_pkg", {
  skip_on_cran()
  pkg_fun <- function(data, ...) {
    rlang::abort("Package 'fakepackage' is required for this feature.",
                 class = "healthmarkers_missing_package", package = "fakepackage")
  }
  result <- HealthMarkers:::.hm_safe_call(
    fun           = pkg_fun,
    data          = data.frame(x = 1),
    col_map       = list(),
    needs_col_map = FALSE,
    verbose       = FALSE,
    tag           = "pkg_test"
  )
  expect_equal(attr(result, "hm_missing_pkg"), "fakepackage")
})

#' Kidney Failure Risk Equation (KFRE, 2- and 5-year risk)
#'
#' Compute 2- and 5-year risk of end-stage kidney disease using the original
#' 4-variable KFRE (Tangri et al., 2011) with optional data-quality diagnostics,
#' and verbose progress reporting.
#'
#' This function preserves prior behavior by default:
#' - Inputs are taken as-is; NA values propagate to outputs (na_action = "keep").
#' - No capping or out-of-range checks are applied.
#'
#' Units (no automatic conversion):
#' - age: years; sex: 1 = male, 2 = female
#' - eGFR: mL/min/1.73 m^2
#' - UACR: mg/g (albumin-to-creatinine ratio)
#'
#' Details
#' - Prognostic index: `PI = 0.220 x log(age) + (-0.556) x log(eGFR) + 0.451 x log(UACR) + 0.391 x (male)`
#'   where male = 1 if sex == 1, else 0.
#' - Baseline survival: S0(2y) = 0.934, S0(5y) = 0.881 (Tangri 2011).
#' - Risks: KFRE_t = 1 - (S0_t ^ exp(PI)).
#' - The 2016 JAMA study provides a large, multinational validation of the KFRE in humans.
#' - This implementation computes the original 4-variable linear predictor and does not
#'   apply recalibration or alternative coefficient sets.
#'
#' @param data A data.frame or tibble containing at least the columns mapped in `col_map`.
#' @param col_map Named list mapping:
#'   - `age`  -> age in years
#'   - `sex`  -> sex code (1 = male, 2 = female)
#'   - `eGFR` -> estimated GFR (mL/min/1.73 m^2)
#'   - `UACR` -> urine albumin-to-creatinine ratio (mg/g)
#' @param na_action One of c("keep","error","omit","warn"). Default "keep" to preserve previous behavior:
#'   - "keep": propagate NA/NaN through logs and outputs.
#'   - "error": abort if any required input contains missing values.
#'   - "omit": drop rows with NA in required inputs before computation.
#'   - "warn": like "keep" but emits high-missingness warnings.
#' @param na_warn_prop Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness warnings. Default 0.2.
#' @param verbose Logical; if TRUE, prints stepwise messages and a completion summary. Default TRUE.
#'
#' @return A tibble with:
#' - `KFRE_2yr` risk (0-1) at 2 years
#' - `KFRE_5yr` risk (0-1) at 5 years
#' @seealso [inflammatory_markers()], [iAge()], [impute_missing()]
#'
#' @examples
#' library(tibble)
#' df <- tibble(
#'   age  = c(65, 72),
#'   sex  = c(1, 2),          # 1 = male, 2 = female
#'   eGFR = c(45, 22),        # mL/min/1.73 m^2
#'   UACR = c(300, 1200)      # mg/g
#' )
#' # Default behavior (NA propagate, no extreme checks)
#' kidney_failure_risk(
#'   data = df,
#'   col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
#' )
#'
#' # With verbose output
#' \donttest{
#' kidney_failure_risk(
#'   data = df,
#'   col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR"),
#'   verbose = TRUE
#' )
#' }
#'
#' @references
#' \insertRef{tangri2011kfre}{HealthMarkers}
#' \insertRef{tangri2016meta}{HealthMarkers}
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
kidney_failure_risk <- function(data,
                                col_map = list(
                                  age  = "age",
                                  sex  = "sex",
                                  eGFR = "eGFR",
                                  UACR = "UACR"
                                ),
                                na_action = c("keep", "error", "omit", "warn"),
                                na_warn_prop = 0.2,
                                verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "kidney_failure_risk"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  na_action <- match.arg(na_action)
  .kfre_validate_args(data, col_map, na_warn_prop)
  # Optional package-level validation; keep test-facing messages stable
  req_keys <- c("age", "sex", "eGFR", "UACR")
  hm_validate_inputs(data, col_map, required_keys = req_keys, fn = "kidney_failure_risk")
  # Check data columns
  req <- req_keys
  mapped_cols <- unlist(col_map[req_keys], use.names = FALSE)
  missing_cols <- setdiff(mapped_cols, names(data))
  if (length(missing_cols)) {
    rlang::abort(
      message = paste0("missing required columns: ", paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_kfre_error_missing_columns"
    )
  }
  if (isTRUE(verbose)) {
    map_parts <- vapply(req_keys, function(k) sprintf("%s -> '%s'", k, col_map[[k]]), character(1))
    hm_inform(sprintf("%s(): col_map: %s", fn_name, paste(map_parts, collapse = ", ")), level = "inform")
    hm_inform(sprintf("%s(): computing markers:\n  KFRE_2yr  [age, sex, eGFR, UACR -- 2-year kidney failure risk]\n  KFRE_5yr  [age, sex, eGFR, UACR -- 5-year kidney failure risk]", fn_name), level = "inform")
  }
  # Column classes
  if (!is.numeric(data[[col_map$age]]))  rlang::abort("kidney_failure_risk(): 'age' column must be numeric.",  class = "healthmarkers_kfre_error_nonnumeric_age")
  if (!is.numeric(data[[col_map$eGFR]])) rlang::abort("kidney_failure_risk(): 'eGFR' column must be numeric.", class = "healthmarkers_kfre_error_nonnumeric_egfr")
  if (!is.numeric(data[[col_map$UACR]])) rlang::abort("kidney_failure_risk(): 'UACR' column must be numeric.", class = "healthmarkers_kfre_error_nonnumeric_uacr")
  # Normalize sex: accepts M/F/male/female/1/2/0 -> 1=male, 2=female
  sex_orig <- data[[col_map$sex]]
  sex_norm <- .hm_normalize_sex(sex_orig, to = "12", fn = "kidney_failure_risk")
  unrecognized <- !is.na(sex_orig) & is.na(sex_norm)
  if (any(unrecognized)) {
    rlang::abort(
      "kidney_failure_risk(): 'sex' must be coded as 1=male or 2=female (or M/F/male/female); unrecognized value(s) found.",
      class = "healthmarkers_kfre_error_invalid_sex"
    )
  }
  data[[col_map$sex]] <- sex_norm
  # Missingness policy
  .kfre_warn_high_missing(
    data,
    unlist(col_map[req], use.names = FALSE),
    na_warn_prop = na_warn_prop,
    do_warn_high_missing = identical(na_action, "warn")
  )
  keep <- rep(TRUE, nrow(data))
  if (na_action == "error") {
    any_na <- Reduce(`|`, lapply(req, function(k) is.na(data[[col_map[[k]]]])))
    if (any(any_na)) {
      rlang::abort("kidney_failure_risk(): required inputs contain missing values (na_action='error').",
                   class = "healthmarkers_kfre_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(req, function(k) is.na(data[[col_map[[k]]]])))
    hm_inform(sprintf("kidney_failure_risk(): omitting %d rows with NA in required inputs", sum(!keep)), level = if (isTRUE(verbose)) "inform" else "debug")
    data <- data[keep, , drop = FALSE]
  }
  # Extract inputs
  age <- data[[col_map$age]]
  sex <- data[[col_map$sex]]
  eGFR <- data[[col_map$eGFR]]
  UACR <- data[[col_map$UACR]]
  # Sex coding (already normalized to 1/2 above)
  male <- ifelse(sex == 1, 1L, 0L)
  # Prognostic index and risk
  pi <- 0.220 * log(age) +
    (-0.556) * log(eGFR) +
    0.451 * log(UACR) +
    0.391 * male
  S0_2 <- 0.934
  S0_5 <- 0.881
  KFRE_2yr <- 1 - (S0_2^exp(pi))
  KFRE_5yr <- 1 - (S0_5^exp(pi))
  out <- tibble::tibble(
    KFRE_2yr = KFRE_2yr,
    KFRE_5yr = KFRE_5yr
  )
  if (!is.null(id_col)) {
    id_vec <- if (na_action == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }
  return(out)
}
# ---- internal helpers (not exported) -----------------------------------------
.kfre_validate_args <- function(data, col_map, na_warn_prop) {
  if (!is.data.frame(data)) {
    rlang::abort("kidney_failure_risk(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_kfre_error_data_type")
  }
  if (!is.list(col_map) || is.null(names(col_map)) || any(names(col_map) == "")) {
    rlang::abort("kidney_failure_risk(): `col_map` must be a named list.",
                 class = "healthmarkers_kfre_error_colmap_type")
  }
  needed <- c("age","sex","eGFR","UACR")
  missing_keys <- setdiff(needed, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(paste0("kidney_failure_risk(): you must supply col_map entries for: ", paste(missing_keys, collapse = ", ")),
                 class = "healthmarkers_kfre_error_missing_map")
  }
  if (!(is.numeric(na_warn_prop) && length(na_warn_prop) == 1L && is.finite(na_warn_prop) &&
        na_warn_prop >= 0 && na_warn_prop <= 1)) {
    rlang::abort("kidney_failure_risk(): `na_warn_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_kfre_error_na_warn_prop")
  }
  invisible(TRUE)
}
.kfre_warn_high_missing <- function(df, cols, na_warn_prop = 0.2, do_warn_high_missing = TRUE) {
  for (cn in cols) {
    x <- df[[cn]]
    n <- length(x)
    if (n == 0L) next
    pna <- sum(is.na(x)) / n
    # High-missingness warning only when na_action = 'warn'
    if (isTRUE(do_warn_high_missing) && pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("kidney_failure_risk(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
    # Diagnostic warnings for domain issues (always on; tests rely on these)
    if (grepl("eGFR", cn, ignore.case = TRUE)) {
      neg <- sum(is.finite(x) & x <= 0)
      if (neg > 0L) rlang::warn(sprintf("kidney_failure_risk(): '%s' contains %d non-positive values; log() undefined.", cn, neg))
    }
    if (grepl("UACR", cn, ignore.case = TRUE)) {
      nonpos <- sum(is.finite(x) & x <= 0)
      if (nonpos > 0L) rlang::warn(sprintf("kidney_failure_risk(): '%s' contains %d non-positive values; log() undefined.", cn, nonpos))
      if (sum(is.finite(x) & x > 30000) > 0L) {
        rlang::warn(sprintf("kidney_failure_risk(): '%s' has very large values (>30000 mg/g); check units (mg/g vs mg/mmol).", cn))
      }
    }
  }
  invisible(TRUE)
}
.kfre_default_extreme_rules <- function() {
  list(
    age  = c(18, 120),
    eGFR = c(1, 200),     # mL/min/1.73 m^2
    UACR = c(0.1, 10000)  # mg/g
  )
}
.kfre_extreme_scan <- function(age, eGFR, UACR, rules) {
  flags <- list(
    age  = is.finite(age)  & (age  < rules$age[1]  | age  > rules$age[2]),
    eGFR = is.finite(eGFR) & (eGFR < rules$eGFR[1] | eGFR > rules$eGFR[2]),
    UACR = is.finite(UACR) & (UACR < rules$UACR[1] | UACR > rules$UACR[2])
  )
  list(count = sum(flags$age) + sum(flags$eGFR) + sum(flags$UACR), flags = flags)
}
.kfre_cap_inputs <- function(age, eGFR, UACR, flags, rules) {
  if (any(flags$age))  { age[flags$age]   <- pmin(pmax(age[flags$age],   rules$age[1]),  rules$age[2]) }
  if (any(flags$eGFR)) { eGFR[flags$eGFR] <- pmin(pmax(eGFR[flags$eGFR], rules$eGFR[1]), rules$eGFR[2]) }
  if (any(flags$UACR)) { UACR[flags$UACR] <- pmin(pmax(UACR[flags$UACR], rules$UACR[1]), rules$UACR[2]) }
  list(age = age, eGFR = eGFR, UACR = UACR)
}

library(testthat)
library(tibble)

test_that("kidney_failure_risk returns a tibble with 2 risk estimates between 0 and 1", {
  df <- tibble(
    age  = 60,
    sex  = 1,
    eGFR = 45,
    UACR = 300
  )

  out <- kidney_failure_risk(
    data = df,
    col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  )

  expect_s3_class(out, "tbl_df")
  expect_named(out, c("KFRE_2yr", "KFRE_5yr"))
  expect_true(all(is.numeric(out$KFRE_2yr)))
  expect_true(all(is.numeric(out$KFRE_5yr)))
  expect_true(all(out$KFRE_2yr >= 0 & out$KFRE_2yr <= 1))
  expect_true(all(out$KFRE_5yr >= 0 & out$KFRE_5yr <= 1))
})

test_that("kidney_failure_risk errors if mapped columns are missing in data", {
  skip_on_cran()
  df_bad <- tibble(
    age  = 60,
    sex  = 1,
    eGFR = 45
    # UACR missing
  )

  expect_error(
    kidney_failure_risk(
      data = df_bad,
      col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
    ),
    "missing required columns: UACR"
  )
})

test_that("kidney_failure_risk errors if required col_map entries are missing", {
  skip_on_cran()
  df <- tibble(
    age  = 60,
    sex  = 1,
    eGFR = 45,
    UACR = 300
  )
  # drop UACR key from col_map - kidney_kfre requires explicit mapping (no inference)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR")
  expect_error(
    kidney_failure_risk(data = df, col_map = cm),
    "UACR"
  )
})

test_that("kidney_failure_risk respects a custom col_map", {
  skip_on_cran()
  df2 <- tibble(A = 70, SEX = 2, GFR = 90, PROT = 20)

  out2 <- kidney_failure_risk(
    data = df2,
    col_map = list(age = "A", sex = "SEX", eGFR = "GFR", UACR = "PROT")
  )

  expect_named(out2, c("KFRE_2yr", "KFRE_5yr"))
  expect_true(all(out2$KFRE_2yr >= 0 & out2$KFRE_2yr <= 1))
  expect_true(all(out2$KFRE_5yr >= 0 & out2$KFRE_5yr <= 1))
})

test_that("na_action='error' aborts when required inputs contain NA", {
  skip_on_cran()
  df <- tibble(age = 60, sex = 1, eGFR = 45, UACR = NA_real_)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  expect_error(
    suppressWarnings(kidney_failure_risk(df, col_map = cm, na_action = "error")),
    "required inputs contain missing values"
  )
})

test_that("na_action='omit' drops rows with NA", {
  skip_on_cran()
  df <- tibble(
    age = c(60, 70),
    sex = c(1, 2),
    eGFR = c(45, NA_real_),
    UACR = c(300, 100)
  )
  cm <- as.list(names(df)); names(cm) <- names(df)
  out <- suppressWarnings(kidney_failure_risk(df, col_map = cm, na_action = "omit"))
  expect_equal(nrow(out), 1L)
  expect_s3_class(out, "tbl_df")
})

test_that("invalid sex coding errors", {
  skip_on_cran()
  df <- tibble(age = 60, sex = 3, eGFR = 45, UACR = 300)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  expect_error(
    kidney_failure_risk(df, col_map = cm),
    "sex' must be coded as 1=male or 2=female"
  )
})

test_that("extreme input values pass through without error", {
  skip_on_cran()
  df <- tibble(age = 10, sex = 1, eGFR = 0.5, UACR = 20000)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  out <- suppressWarnings(kidney_failure_risk(df, col_map = cm))
  expect_true(is.numeric(out$KFRE_2yr))
  expect_true(is.numeric(out$KFRE_5yr))
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble::tibble(age = 60, sex = 1, eGFR = 45, UACR = 300)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  expect_message(kidney_failure_risk(df, col_map = cm, verbose = TRUE), "kidney_failure_risk")
  expect_message(kidney_failure_risk(df, col_map = cm, verbose = TRUE), "col_map")
  expect_message(kidney_failure_risk(df, col_map = cm, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble::tibble(age = 60, sex = 1, eGFR = 45, UACR = 300)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  msgs <- testthat::capture_messages(
    kidney_failure_risk(df, col_map = cm, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("non-positive eGFR triggers log warning", {
  skip_on_cran()
  df <- tibble(age = 60, sex = 1, eGFR = 0, UACR = 1)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  expect_warning(
    kidney_failure_risk(df, col_map = cm),
    "'eGFR'.*log\\(\\) undefined"
  )
})

test_that("non-positive UACR triggers log warning", {
  skip_on_cran()
  df <- tibble(age = 60, sex = 1, eGFR = 1, UACR = 0)
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  expect_warning(
    kidney_failure_risk(df, col_map = cm),
    "'UACR'.*log\\(\\) undefined"
  )
})

# ---- Multi-row vectorization ------------------------------------------------
test_that("kidney_failure_risk handles multi-row data correctly", {
  skip_on_cran()
  df <- tibble::tibble(
    age  = c(45, 60, 75, 80),
    sex  = c(1, 2, 1, 2),
    eGFR = c(90, 45, 22, 10),
    UACR = c(30, 300, 1000, 3000)
  )
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  out <- kidney_failure_risk(df, col_map = cm)
  expect_equal(nrow(out), 4L)
  expect_true(all(out$KFRE_2yr >= 0 & out$KFRE_2yr <= 1))
  expect_true(all(out$KFRE_5yr >= 0 & out$KFRE_5yr <= 1))
  # Lower eGFR / higher UACR should yield higher risk
  expect_gt(out$KFRE_5yr[4], out$KFRE_5yr[1])
})

# ---- String sex encoding ----------------------------------------------------
test_that("kidney_failure_risk accepts string sex codes M/F", {
  skip_on_cran()
  df_str <- tibble::tibble(
    age  = c(60, 70),
    sex  = c("M", "F"),
    eGFR = c(45, 30),
    UACR = c(300, 500)
  )
  df_int <- tibble::tibble(
    age  = c(60, 70),
    sex  = c(1, 2),
    eGFR = c(45, 30),
    UACR = c(300, 500)
  )
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  out_str <- kidney_failure_risk(df_str, col_map = cm)
  out_int <- kidney_failure_risk(df_int, col_map = cm)
  expect_equal(out_str$KFRE_2yr, out_int$KFRE_2yr)
  expect_equal(out_str$KFRE_5yr, out_int$KFRE_5yr)
})

test_that("kidney_failure_risk accepts 'male'/'female' sex coding", {
  skip_on_cran()
  df <- tibble::tibble(
    age  = 60, sex = "male", eGFR = 45, UACR = 300
  )
  df2 <- tibble::tibble(
    age  = 60, sex = 1L, eGFR = 45, UACR = 300
  )
  cm <- list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
  out  <- kidney_failure_risk(df,  col_map = cm)
  out2 <- kidney_failure_risk(df2, col_map = cm)
  expect_equal(out$KFRE_2yr, out2$KFRE_2yr)
})

#' CKD staging (GFR and albuminuria) and KDIGO risk
#'
#' @description
#' Categorizes eGFR into G1-G5, albuminuria into A1-A3 (by UACR mg/g), and maps KDIGO risk.
#'
#' @param data Data frame with renal measures.
#' @param col_map Named list with required key: eGFR; optional key: UACR.
#' @param na_action One of:
#'   - "keep"  (retain rows; stages become NA where inputs missing)
#'   - "omit"  (drop rows with any missing eGFR/UACR that are mapped)
#'   - "error" (abort if any mapped input missing)
#' @param verbose Logical; if TRUE (default), emits progress messages via `hm_inform()`.
#'
#' @return Tibble with CKD_stage, Albuminuria_stage, KDIGO_risk.
#' @references \insertRef{kdigo2012ckd}{HealthMarkers}
#' @examples
#' df <- data.frame(eGFR = c(95, 50), UACR = c(10, 200))
#' ckd_stage(df, list(eGFR = "eGFR", UACR = "UACR"))
#' @export
ckd_stage <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error"),
  verbose = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "ckd_stage"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  id_col <- .hm_detect_id_col(data)

  # Auto-fill col_map when not supplied
  cm      <- .hm_build_col_map(data, col_map, c("eGFR","UACR"), fn = "ckd_stage")
  data    <- cm$data
  col_map <- cm$col_map
  if (is.null(col_map)) col_map <- list()

  # Validate mapping and columns (explicit; no hm_validate_inputs)
  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("ckd_stage(): `col_map` must be a named list.", class = "healthmarkers_ckd_error_colmap_type")
  }
  if (is.null(col_map$eGFR) || !nzchar(col_map$eGFR)) {
    rlang::abort("ckd_stage(): `col_map$eGFR` is required.", class = "healthmarkers_ckd_error_missing_map")
  }
  if (!col_map$eGFR %in% names(data)) {
    rlang::abort(sprintf("ckd_stage(): eGFR column '%s' not found in data.", col_map$eGFR),
                 class = "healthmarkers_ckd_error_missing_column")
  }
  # Optional UACR: include only if column exists; warn if mapped but missing
  has_uacr_map <- !is.null(col_map$UACR) && nzchar(col_map$UACR)
  has_uacr_col <- has_uacr_map && (col_map$UACR %in% names(data))
  if (has_uacr_map && !has_uacr_col) {
    rlang::warn(sprintf("ckd_stage(): UACR column '%s' not found; albuminuria will be NA.", col_map$UACR),
                class = "healthmarkers_ckd_warn_uacr_missing_column")
  }

  # Identify optional present inputs to include in NA/extreme policies
  keys_for_policy <- c("eGFR", if (has_uacr_col) "UACR")

  hm_inform(level = "debug", msg = "ckd_stage(): computing stages")
  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  CKD_stage          [eGFR G-stage]\n  Albuminuria_stage  [UACR A-stage]\n  KDIGO_risk         [combined KDIGO risk category]", fn_name), level = "inform")

  # Coerce to numeric; warn on NA introduction; non-finite -> NA
  eGFR <- data[[col_map$eGFR]]
  if (!is.numeric(eGFR)) {
    old <- eGFR
    suppressWarnings(eGFR <- as.numeric(eGFR))
    intro <- sum(is.na(eGFR) & !is.na(old))
    if (intro > 0) {
      rlang::warn(sprintf("ckd_stage(): column '%s' coerced to numeric; NAs introduced: %d", col_map$eGFR, intro),
                  class = "healthmarkers_ckd_warn_na_coercion")
    }
  }
  eGFR[!is.finite(eGFR)] <- NA_real_

  if (has_uacr_col) {
    UACR <- data[[col_map$UACR]]
    if (!is.numeric(UACR)) {
      old <- UACR
      suppressWarnings(UACR <- as.numeric(UACR))
      intro <- sum(is.na(UACR) & !is.na(old))
      if (intro > 0) {
        rlang::warn(sprintf("ckd_stage(): column '%s' coerced to numeric; NAs introduced: %d", col_map$UACR, intro),
                    class = "healthmarkers_ckd_warn_na_coercion")
      }
    }
    UACR[!is.finite(UACR)] <- NA_real_
  } else {
    UACR <- rep(NA_real_, length(eGFR))
  }

  # NA policy over mapped inputs (required + present optional)
  if (length(keys_for_policy)) {
    vals <- list(eGFR = eGFR)
    if ("UACR" %in% keys_for_policy) vals$UACR <- UACR
    rows_with_na <- Reduce(`|`, lapply(vals, function(x) is.na(x)))
  } else {
    rows_with_na <- rep(FALSE, length(eGFR))
  }

  keep <- rep(TRUE, nrow(data))

  if (na_action == "error" && any(rows_with_na)) {
    rlang::abort("ckd_stage(): missing/non-finite values in mapped inputs (na_action='error').",
                 class = "healthmarkers_ckd_error_missing_values")
  } else if (na_action == "omit" && any(rows_with_na)) {
    keep <- !rows_with_na
    eGFR <- eGFR[keep]
    UACR <- UACR[keep]
  }

  # G stage
  G <- cut(
    eGFR,
    breaks = c(-Inf, 15, 30, 45, 60, 90, Inf),
    labels = c("G5","G4","G3b","G3a","G2","G1"),
    right = FALSE
  )

  # A stage (if UACR provided; else NA)
  A <- cut(
    UACR,
    breaks = c(-Inf, 30, 300, Inf),
    labels = c("A1","A2","A3"),
    right = FALSE
  )

  # For risk mapping, treat missing albuminuria as A1 by convention
  A_filled <- as.character(A); A_filled[is.na(A_filled)] <- "A1"

  kdigo_map <- function(g, a) {
    if (is.na(g) || is.na(a)) return(NA_character_)
    if (g %in% c("G1","G2") && a == "A1") return("Low")
    if (g %in% c("G1","G2") && a %in% c("A2","A3")) return("Moderate")
    if (g == "G3a" && a == "A1") return("Moderate")
    if (g == "G3a" && a %in% c("A2","A3")) return("High")
    if (g == "G3b") return(ifelse(a == "A1","High","Very High"))
    if (g %in% c("G4","G5")) return("Very High")
    "Moderate"
  }
  KDIGO <- mapply(kdigo_map, as.character(G), A_filled, USE.NAMES = FALSE)

  out <- tibble::tibble(
    CKD_stage = factor(G, levels = c("G1","G2","G3a","G3b","G4","G5")),
    Albuminuria_stage = factor(A, levels = c("A1","A2","A3")),
    KDIGO_risk = factor(KDIGO, levels = c("Low","Moderate","High","Very High"))
  )

  if (!is.null(id_col)) {
    out[[id_col]] <- data[[id_col]][keep]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }

  out
}

test_that("ckd_stage classifies G and A stages", {
  skip_on_cran()
  df <- data.frame(eGFR = c(95, 70, 50, 35, 10), UACR = c(10, 50, 200, 600, 100))
  res <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"))
  expect_s3_class(res, "tbl_df")
  expect_named(res, c("CKD_stage", "Albuminuria_stage", "KDIGO_risk"))
  expect_equal(as.character(res$CKD_stage), c("G1","G2","G3a","G3b","G5"))
  expect_equal(as.character(res$Albuminuria_stage), c("A1","A2","A2","A3","A2"))
})

test_that("eGFR boundary values map to correct G stages", {
  skip_on_cran()
  df <- data.frame(
    eGFR = c(90, 89, 60, 59, 45, 44, 30, 29, 15, 14),
    UACR = rep(10, 10)
  )
  res <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"))
  expect_equal(
    as.character(res$CKD_stage),
    c("G1","G2","G2","G3a","G3a","G3b","G3b","G4","G4","G5")
  )
})

test_that("UACR boundary values map to correct A stages", {
  skip_on_cran()
  df <- data.frame(eGFR = rep(65, 4), UACR = c(0, 29, 30, 300))
  res <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"))
  expect_equal(as.character(res$Albuminuria_stage), c("A1","A1","A2","A3"))
})

test_that("eGFR-only (no UACR) returns NA albuminuria and KDIGO assuming A1", {
  skip_on_cran()
  df <- data.frame(eGFR = c(95, 50, 12))
  res <- ckd_stage(df, col_map = list(eGFR = "eGFR"))
  expect_equal(as.character(res$CKD_stage), c("G1","G3a","G5"))
  expect_true(all(is.na(res$Albuminuria_stage)))
  # G1 + A1 (assumed) -> Low; G3a + A1 -> Moderate; G4/G5 -> Very High
  expect_equal(as.character(res$KDIGO_risk), c("Low","Moderate","Very High"))
})

test_that("ckd_stage default keep retains rows; omit drops any NA in mapped inputs", {
  skip_on_cran()
  df <- data.frame(eGFR = c(NA, 80), UACR = c(10, NA))
  res_keep <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"))
  expect_equal(nrow(res_keep), 2L)
  expect_true(is.na(res_keep$CKD_stage[1]))
  res_omit <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"), na_action = "omit")
  expect_equal(nrow(res_omit), 0L)
})

test_that("ckd_stage na_action = error aborts on missing mapped inputs", {
  skip_on_cran()
  df <- data.frame(eGFR = c(NA, 80), UACR = c(10, NA))
  expect_error(
    ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"), na_action = "error"),
    "missing/non-finite values"
  )
})

test_that("extreme eGFR/UACR values stage correctly without error", {
  skip_on_cran()
  df <- data.frame(eGFR = c(250, 60), UACR = c(10, 6000))
  res <- ckd_stage(df, col_map = list(eGFR = "eGFR", UACR = "UACR"))
  # eGFR 250 > G1 threshold, maps to G1
  expect_equal(as.character(res$CKD_stage[1]), "G1")
  # UACR 6000 > 300 -> A3
  expect_equal(as.character(res$Albuminuria_stage[2]), "A3")
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df  <- data.frame(eGFR = c(95, 50), UACR = c(10, 200))
  cm2 <- list(eGFR = "eGFR", UACR = "UACR")
  expect_message(ckd_stage(df, col_map = cm2, verbose = TRUE), "ckd_stage")
  expect_message(ckd_stage(df, col_map = cm2, verbose = TRUE), "col_map")
  expect_message(ckd_stage(df, col_map = cm2, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df   <- data.frame(eGFR = c(95, 50), UACR = c(10, 200))
  cm2  <- list(eGFR = "eGFR", UACR = "UACR")
  msgs <- testthat::capture_messages(
    ckd_stage(df, col_map = cm2, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Compute sex-stratified standardized scores (SDS) for adiposity measures
#'
#' @description
#' Computes sex-specific SDS (z-scores) for selected anthropometric variables using
#' reference means and SDs provided separately for males and females.
#'
#' @param data Data frame or tibble with variables and a sex column
#' @param col_map Named list mapping:
#'   - sex: column name with sex values ("M","F","m","f", 1, 2)
#'   - vars: optional named list mapping reference variable names -> data column names.
#'           If omitted, expects the reference variable names to exist in `data`.
#' @param ref Named list with elements "M" and "F". Each is a named list of numeric
#'   vectors c(mean=, sd=) keyed by variable name, e.g.:
#'   list(
#'     M = list(BMI = c(mean=23, sd=3.5), waist = c(mean=85, sd=10)),
#'     F = list(BMI = c(mean=21, sd=3.0), waist = c(mean=75, sd=9))
#'   )
#' @param na_action One of:
#'   - "keep"  - keep rows with NA (propagates to outputs)
#'   - "omit"  - drop rows with NA in any required variable
#'   - "error" - abort if any required variable has NA
#' @param allow_partial If TRUE, skip variables absent in data (with a warning); if FALSE error
#' @param prefix Optional prefix for output columns (default "")
#' @param verbose Logical; when TRUE emit progress via package logger; by default logging is controlled by options(healthmarkers.verbose)
#'
#' @return A tibble with one SDS column per retained variable: \code{varname_SDS}, where \code{varname} is the original variable name (optionally prefixed by \code{prefix}).
#'
#' @references
#' \insertRef{who1995anthro}{HealthMarkers}
#'
#' @examples
#' ref <- list(
#'   M = list(BMI = c(mean=24.5, sd=3.8), waist = c(mean=88, sd=12)),
#'   F = list(BMI = c(mean=22.1, sd=4.2), waist = c(mean=76, sd=11))
#' )
#' df <- data.frame(BMI=c(25.2,21.8,27.1), waist=c(85,72,95), sex=c("M","F","M"))
#' adiposity_sds_strat(
#'   df,
#'   col_map = list(sex = "sex", vars = list(BMI = "BMI", waist = "waist")),
#'   ref = ref
#' )
#' @export
adiposity_sds_strat <- function(data,
                                col_map,
                                ref,
                                na_action = c("keep","omit","error"),
                                allow_partial = FALSE,
                                prefix = "",
                                verbose = TRUE) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "adiposity_sds_strat"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  na_action <- match.arg(na_action)

  # Custom checks so tests see the expected error messages
  if (!is.list(col_map) || is.null(col_map$sex) || !nzchar(col_map$sex)) {
    rlang::abort("adiposity_sds_strat(): 'col_map$sex' is required.",
                 class = "healthmarkers_adiposds_error_colmap_sex")
  }
  sex_col <- col_map$sex
  if (!sex_col %in% names(data)) {
    rlang::abort(sprintf("adiposity_sds_strat(): sex column '%s' not found.", sex_col),
                 class = "healthmarkers_adiposds_error_sexcol_missing")
  }

  # Validate ref structure
  if (!is.list(ref) || !all(c("M","F") %in% names(ref))) {
    rlang::abort("adiposity_sds_strat(): `ref` must be a named list with elements 'M' and 'F'.",
                 class = "healthmarkers_adiposds_error_ref_structure")
  }
  vars_M <- names(ref$M); vars_F <- names(ref$F)
  if (is.null(vars_M) || is.null(vars_F) || any(vars_M == "") || any(vars_F == "")) {
    rlang::abort("adiposity_sds_strat(): reference variable names must be non-empty.",
                 class = "healthmarkers_adiposds_error_ref_names")
  }
  if (!identical(sort(vars_M), sort(vars_F))) {
    rlang::abort("adiposity_sds_strat(): ref$M and ref$F must define identical variables.",
                 class = "healthmarkers_adiposds_error_ref_var_mismatch")
  }
  vars <- vars_M

  # Validate each ref component
  check_ref_comp <- function(x) {
    if (!is.numeric(x) || is.null(names(x)) || !all(c("mean","sd") %in% names(x))) {
      rlang::abort("adiposity_sds_strat(): ref component must be numeric with names mean, sd.",
                   class = "healthmarkers_adiposds_error_ref_component")
    }
    if (!is.finite(x[["sd"]]) || x[["sd"]] <= 0) {
      rlang::abort("adiposity_sds_strat(): sd must be >0.",
                   class = "healthmarkers_adiposds_error_ref_sd")
    }
  }
  for (v in vars) {
    check_ref_comp(ref$M[[v]])
    check_ref_comp(ref$F[[v]])
  }

  # Build variable mapping with inference support; backward-compat col_map$vars
  flat_vars_map <- if (!is.null(col_map$vars)) col_map$vars else NULL
  cm      <- .hm_build_col_map(data, flat_vars_map, keys = vars, fn = fn_name)
  data    <- cm$data
  var_map <- cm$col_map

  # Sex normalization
  sex_raw <- data[[sex_col]]
  sex_norm <- ifelse(sex_raw %in% c("M","m",1), "M",
                ifelse(sex_raw %in% c("F","f",2), "F", NA_character_))
  bad_sex <- sum(is.na(sex_norm))
  if (bad_sex > 0) {
    rlang::abort(sprintf("adiposity_sds_strat(): %d rows have unmapped sex values.", bad_sex),
                 class = "healthmarkers_adiposds_error_sex_values")
  }

  # Ensure variable columns exist (respect allow_partial)
  # Check both: key present in var_map AND mapped column exists in data
  missing_in_data <- vars[!vapply(vars, function(v) {
    !is.null(var_map[[v]]) && isTRUE(var_map[[v]] %in% names(data))
  }, logical(1))]
  if (length(missing_in_data)) {
    if (isTRUE(allow_partial)) {
      hm_inform(sprintf(
        "adiposity_sds_strat(): skipping %d ref vars absent in data: %s",
        length(missing_in_data), paste(missing_in_data, collapse = ", ")
      ), level = "inform")
      keep_vars <- vars[vars %in% names(var_map)]
      vars <- keep_vars
      var_map <- var_map[vars]
    } else {
      rlang::abort(sprintf("adiposity_sds_strat(): variables not found in data: %s",
                           paste(missing_in_data, collapse = ", ")),
                   class = "healthmarkers_adiposds_error_missing_vars")
    }
  }
  if (!length(vars)) {
    rlang::abort("adiposity_sds_strat(): no variables remain to process.",
                 class = "healthmarkers_adiposds_error_no_vars")
  }

  .hm_log_cols(cm, var_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  %s", fn_name, paste(paste0(prefix, vars, "_SDS"), collapse = ", ")), level = "inform")

  # Coerce to numeric (warn on NA introduction)
  for (v in vars) {
    cn <- var_map[[v]]
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      intro <- sum(is.na(data[[cn]]) & !is.na(old))
      if (intro > 0) {
        rlang::warn(
          sprintf("adiposity_sds_strat(): column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
          class = "healthmarkers_adiposds_warn_na_coercion"
        )
      }
    }
  }

  # NA policy (across selected variables)
  if (na_action == "error") {
    any_na <- Reduce(`|`, lapply(unname(unlist(var_map)), function(cn) is.na(data[[cn]])))
    if (any(any_na)) {
      rlang::abort("adiposity_sds_strat(): missing values present (na_action='error').",
                   class = "healthmarkers_adiposds_error_missing_values")
    }
  } else if (na_action == "omit") {
    keep <- !Reduce(`|`, lapply(unname(unlist(var_map)), function(cn) is.na(data[[cn]])))
    data <- data[keep, , drop = FALSE]
    sex_norm <- sex_norm[keep]
  }
  if (nrow(data) == 0L) return(tibble::tibble())

  # Compute SDS
  out_list <- lapply(vars, function(v) {
    stats_M <- ref$M[[v]]
    stats_F <- ref$F[[v]]
    mu  <- ifelse(sex_norm == "M", stats_M[["mean"]], stats_F[["mean"]])
    sdv <- ifelse(sex_norm == "M", stats_M[["sd"]],   stats_F[["sd"]])
    (data[[var_map[[v]]]] - mu) / sdv
  })
  names(out_list) <- paste0(prefix, vars, "_SDS")
  out <- tibble::as_tibble(out_list)

  if (!is.null(id_col)) {
    id_vec <- data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }
  out
}

# tests/testthat/test_adiposity_sds_strat.R

library(testthat)
library(tibble)

# Shared reference set
ref_full <- list(
  M = list(
    BMI   = c(mean = 23, sd = 3),
    waist = c(mean = 85, sd = 10)
  ),
  F = list(
    BMI   = c(mean = 21, sd = 3),
    waist = c(mean = 75, sd = 9)
  )
)

cm <- list(sex = "sex", vars = list(BMI = "BMI", waist = "waist"))

test_that("ref must have M and F", {
  skip_on_cran()
  df <- tibble(sex = "M", BMI = 25)
  bad_refs <- list(
    NULL,
    list(M = list(BMI = c(mean = 23, sd = 3))),
    list(F = list(BMI = c(mean = 21, sd = 3))),
    list(X = list(BMI = c(mean = 23, sd = 3)), F = list(BMI = c(mean = 21, sd = 3)))
  )
  for (r in bad_refs) {
    expect_error(
      adiposity_sds_strat(df, col_map = cm, ref = r),
      "`ref` must be a named list with elements 'M' and 'F'"
    )
  }
})

test_that("ref M/F variable sets must match", {
  skip_on_cran()
  df <- tibble(sex = "M", BMI = 25)
  ref_mismatch <- list(
    M = list(BMI = c(mean = 23, sd = 3), waist = c(mean = 80, sd = 10)),
    F = list(BMI = c(mean = 21, sd = 3))
  )
  expect_error(
    adiposity_sds_strat(df, col_map = cm, ref = ref_mismatch),
    "ref\\$M and ref\\$F must define identical variables"
  )
})

test_that("ref component must have mean & sd names", {
  skip_on_cran()
  df <- tibble(sex = "M", BMI = 25)
  bad_ref <- list(
    M = list(BMI = c(23, 3)),
    F = list(BMI = c(mean = 21, sd = 3))
  )
  expect_error(
    adiposity_sds_strat(df, col_map = cm, ref = bad_ref),
    "must be numeric with names mean, sd"
  )
})

test_that("sd must be >0", {
  skip_on_cran()
  df <- tibble(sex = "F", BMI = 20)
  ref_bad_sd <- list(
    M = list(BMI = c(mean = 23, sd = 3)),
    F = list(BMI = c(mean = 21, sd = 0))
  )
  expect_error(
    adiposity_sds_strat(df, col_map = cm, ref = ref_bad_sd),
    "sd must be >0"
  )
})

test_that("missing sex column errors", {
  skip_on_cran()
  df <- tibble(gender = "M", BMI = 25)
  ref <- list(
    M = list(BMI = c(mean = 23, sd = 3)),
    F = list(BMI = c(mean = 21, sd = 3))
  )
  expect_error(
    adiposity_sds_strat(df, col_map = list(sex = "sex", vars = list(BMI = "BMI")), ref = ref),
    "sex column 'sex' not found"
  )
})

test_that("invalid sex values error", {
  skip_on_cran()
  df <- tibble(sex = c("M", "X"), BMI = c(25, 22))
  ref <- list(
    M = list(BMI = c(mean = 23, sd = 3)),
    F = list(BMI = c(mean = 21, sd = 3))
  )
  expect_error(
    adiposity_sds_strat(df, col_map = list(sex = "sex", vars = list(BMI = "BMI")), ref = ref),
    "rows have unmapped sex values"
  )
})

test_that("computes correct SDS (multi-variable, vectorized)", {
  skip_on_cran()
  df <- tibble(
    sex   = c("M", "F", "M"),
    BMI   = c(25, 20, 26),
    waist = c(95, 70, 80)
  )
  out <- adiposity_sds_strat(df, col_map = cm, ref = ref_full)
  expect_equal(out$BMI_SDS[1],  (25 - 23) / 3)
  expect_equal(out$waist_SDS[1], (95 - 85) / 10)
  expect_equal(out$BMI_SDS[2],  (20 - 21) / 3)
  expect_equal(out$waist_SDS[2], (70 - 75) / 9)
  expect_equal(nrow(out), 3)
  expect_named(out, c("BMI_SDS", "waist_SDS"))
})

test_that("custom variable mapping via col_map$vars", {
  skip_on_cran()
  df <- tibble(sex = c("M","F"), BMI_col = c(25, 23))
  ref <- list(
    M = list(BMI = c(mean = 24, sd = 3)),
    F = list(BMI = c(mean = 22, sd = 3))
  )
  out <- adiposity_sds_strat(df, col_map = list(sex = "sex", vars = list(BMI = "BMI_col")), ref = ref)
  expect_equal(out$BMI_SDS, c((25-24)/3, (23-22)/3))
})

test_that("na_action omit vs error", {
  skip_on_cran()
  df <- tibble(sex = c("M","F","M"), BMI = c(25, NA, 24), waist = c(90, 70, 88))
  out_keep <- adiposity_sds_strat(df, col_map = cm, ref = ref_full, na_action = "keep")
  expect_equal(nrow(out_keep), 3)
  out_omit <- adiposity_sds_strat(df, col_map = cm, ref = ref_full, na_action = "omit")
  expect_equal(nrow(out_omit), 2)
  expect_error(
    adiposity_sds_strat(df, col_map = cm, ref = ref_full, na_action = "error"),
    "missing values present"
  )
})

test_that("allow_partial skips missing vars when TRUE", {
  skip_on_cran()
  ref_partial <- list(
    M = list(BMI = c(mean=23, sd=3), waist = c(mean=85, sd=10)),
    F = list(BMI = c(mean=21, sd=3), waist = c(mean=75, sd=9))
  )
  df <- tibble(sex = c("M","F"), BMI = c(24,22))  # waist missing
  out <- adiposity_sds_strat(df, col_map = list(sex = "sex"), ref = ref_partial, allow_partial = TRUE)
  expect_named(out, "BMI_SDS")
  expect_equal(nrow(out), 2)
})

test_that("allow_partial=FALSE errors when data var missing", {
  skip_on_cran()
  ref_partial <- ref_full
  df <- tibble(sex = c("M","F"), BMI = c(24,22))
  expect_error(
    adiposity_sds_strat(df, col_map = list(sex = "sex", vars = list(BMI = "BMI", waist = "waist")), ref = ref_partial, allow_partial = FALSE),
    "variables not found in data"
  )
})

test_that("check_extreme removed: function passes through raw outlier rows", {
  skip_on_cran()
  df <- tibble(
    sex = c("M","F"),
    BMI = c(120, 2),
    waist = c(300, 15)
  )
  ref <- list(
    M = list(BMI = c(mean=23, sd=3), waist = c(mean=85, sd=10)),
    F = list(BMI = c(mean=21, sd=3), waist = c(mean=75, sd=9))
  )
  out <- adiposity_sds_strat(df, col_map = cm, ref = ref)
  expect_equal(nrow(out), 2L)
})

test_that("prefix applied to output columns", {
  skip_on_cran()
  df <- tibble(sex = c("M","F"), BMI = c(24,22))
  ref <- list(
    M = list(BMI = c(mean=23, sd=3)),
    F = list(BMI = c(mean=21, sd=3))
  )
  out <- adiposity_sds_strat(df, col_map = list(sex = "sex", vars = list(BMI = "BMI")), ref = ref, prefix = "x_")
  expect_named(out, "x_BMI_SDS")
})

test_that("numeric sex coding 1/2 accepted", {
  skip_on_cran()
  df <- tibble(sex = c(1L, 2L, 1L), BMI = c(25, 20, 26), waist = c(95, 70, 80))
  out <- adiposity_sds_strat(df, col_map = cm, ref = ref_full)
  # row 1 is male (1), row 2 is female (2)
  expect_equal(out$BMI_SDS[1], (25 - 23) / 3)
  expect_equal(out$BMI_SDS[2], (20 - 21) / 3)
  expect_named(out, c("BMI_SDS", "waist_SDS"))
})

test_that("verbose = TRUE emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(sex = "M", BMI = 25, waist = 95)
  expect_message(adiposity_sds_strat(df, col_map = cm, ref = ref_full, verbose = TRUE), "adiposity_sds_strat")
  expect_message(adiposity_sds_strat(df, col_map = cm, ref = ref_full, verbose = TRUE), "col_map")
  expect_message(adiposity_sds_strat(df, col_map = cm, ref = ref_full, verbose = TRUE), "results:")
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(sex = "M", BMI = 25, waist = 95)
  msgs <- testthat::capture_messages(
    adiposity_sds_strat(df, col_map = cm, ref = ref_full, verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

#' Summarize selected numeric marker columns
#'
#' @description
#' Lightweight summary for numeric columns: n, n_na, mean, sd, median, p25, p75.
#' @param x A data.frame or tibble.
#' @param cols Optional character vector of column names to summarize. If NULL, all numeric columns are summarized.
#' @return A tibble with one row per summarized column.
#' @examples
#' df <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5), c = factor("x"))
#' health_summary(df)
#' @export
health_summary <- function(x, cols = NULL) {
  if (!is.data.frame(x)) rlang::abort("`x` must be a data.frame or tibble.", class = "healthmarkers_health_summary_error_data_type")
  if (is.null(cols)) cols <- names(x)[vapply(x, is.numeric, logical(1))]
  cols <- intersect(cols, names(x))
  if (length(cols) == 0L) {
    return(tibble::tibble(
      measure = character(),
      n = integer(), n_na = integer(),
      mean = numeric(), sd = numeric(),
      median = numeric(), p25 = numeric(), p75 = numeric()
    ))
  }
  rows <- lapply(cols, function(nm) {
    v <- x[[nm]]
    nn <- sum(!is.na(v))
    tibble::tibble(
      measure = nm,
      n = length(v),
      n_na = sum(is.na(v)),
      mean = if (nn) mean(v, na.rm = TRUE) else NA_real_,
      sd = if (nn) stats::sd(v, na.rm = TRUE) else NA_real_,
      median = if (nn) stats::median(v, na.rm = TRUE) else NA_real_,
      p25 = if (nn) stats::quantile(v, 0.25, na.rm = TRUE, names = FALSE) else NA_real_,
      p75 = if (nn) stats::quantile(v, 0.75, na.rm = TRUE, names = FALSE) else NA_real_
    )
  })
  do.call(rbind, rows)
}

test_that("health_summary works on numeric columns", {
  df <- data.frame(a = c(1,2,NA), b = c(3,4,5), c = factor("x"))
  s <- health_summary(df)
  expect_true(all(c("a","b") %in% s$measure))
  expect_true(all(c("mean","sd","median","p25","p75") %in% names(s)))
})

#' Calculate glycemic-, C-peptide-, and additional metabolic markers
#'
#' Given fasting labs and anthropometry, computes:
#'  - SPISE (Single-Point Insulin Sensitivity Estimator)
#'  - METS_IR (Metabolic Score for Insulin Resistance)
#'  - prediabetes flag (HbA1c >= 42 mmol/mol)
#'  - diabetes flag (HbA1c >= 48 mmol/mol)
#'  - HOMA_CP (C-peptide-based HOMA-IR variant; operational formula, see notes)
#'  - LAR (Leptin/Adiponectin Ratio)
#'  - ASI (Adiponectin Sensitivity Index; adiponectin/insulin)
#'  - TyG_index (Triglyceride-Glucose Index)
#'
#' Assumed units (no automatic conversion of inputs except where noted):
#' - HDL_c, TG: mmol/L (TyG internally converts TG to mg/dL via 88.57)
#' - BMI: kg/m^2
#' - glucose, G0: mmol/L (TyG internally converts glucose to mg/dL via 18)
#' - HbA1c: mmol/mol
#' - C_peptide, I0: pmol/L (HOMA_CP uses I-like conversion factor 6 as in insulin muU/mL; see notes)
#' - leptin, adiponectin: ng/mL
#'
#' These indices are intended for research and feature-engineering applications.
#' The prediabetes and diabetes flags apply standard HbA1c cut-offs (WHO/IDF criteria)
#' but this function is not validated as a clinical diagnostic tool.
#'
#' Quality controls and options:
#' - Input validation ensures required variables exist and are numeric-coercible.
#' - Non-numeric inputs are coerced to numeric with a warning (NAs introduced reported).
#' - Missing or non-finite inputs are handled via `na_action`.
#' - Logs and divisions are computed safely (non-positive arguments yield NA).
#' - Physiological range notes are printed when `verbose = TRUE` (values are not altered).
#' - Verbose mode prints step-by-step progress and a completion summary.
#'
#' @param data A data.frame or tibble containing at least:
#'   - HDL_c (mmol/L), TG (mmol/L), BMI (kg/m^2)
#'   Optional if present: glucose (mmol/L), HbA1c (mmol/mol), C_peptide (pmol/L),
#'   G0 (mmol/L), I0 (pmol/L), leptin (ng/mL), adiponectin (ng/mL).
#' @param col_map Optional named list mapping keys to column names in `data`.
#'   Required keys: `HDL_c`, `TG`, `BMI`. Optional keys (if present will be used):
#'   `glucose`, `HbA1c`, `C_peptide`, `G0`, `I0`, `leptin`, `adiponectin`.
#'   If `NULL` (default), the function uses columns named exactly as the keys.
#' @param na_action One of `c("ignore","warn","error","keep","omit")`.
#'   HM-CS aliases: `keep` == `ignore`; `omit` drops rows with any NA/non-finite
#'   in used inputs before computing markers. Default "ignore".
#' @param na_warn_prop Proportion (0-1) threshold for high-missingness warnings
#'   among used columns when `na_action = "warn"`. Default 0.2.
#' @param verbose Logical; if `TRUE` (default), prints step-by-step progress
#'   including column mapping, optional input availability, pre-computation
#'   notes, physiological range information (informational only, values are not
#'   altered), the list of markers being computed with their inputs, and a
#'   per-column results summary.
#'
#' @return A tibble with columns:
#'   - SPISE, METS_IR, prediabetes, diabetes, HOMA_CP, LAR, ASI, TyG_index
#'   If an ID column is detected in `data` (e.g. `id`, `IID`, `participant_id`),
#'   it is prepended as the first output column.
#'
#' @details
#' Optional marker detection and pre-computation (one level deep):
#' When `col_map = NULL` (default), column names are inferred automatically.
#' The seven optional keys (glucose, HbA1c, C_peptide, G0, I0, leptin,
#' adiponectin) are computed only when present in `data`. If `BMI` is absent
#' but `weight` (kg) and `height` (m or cm) are present, BMI is computed
#' automatically. If `glucose` is absent but `G0` column exists (or vice
#' versa), the missing key is derived via alias. With `verbose = TRUE`
#' (default), the function reports: column mapping, what is missing and which
#' raw inputs could provide it, which indices will be NA and why, any
#' physiological range notes, and a per-column results summary.
#'
#' Notes on HOMA_CP:
#' - This function retains the package's existing operational formula:
#'   HOMA_CP = (G0 (mmol/L) * (C_peptide (pmol/L) / 6)) / 22.5
#'   which mirrors HOMA-IR's structure using a 6 pmol/muU scaling used for insulin.
#'   Users should verify unit conventions for their datasets; alternative C-peptide
#'   HOMA implementations exist (e.g., HOMA2-CP).
#' @references
#' \insertRef{paulmichl2016spise}{HealthMarkers}
#' \insertRef{bellogaytan2018metsir}{HealthMarkers}
#' \insertRef{fruhbeck2018adiponectinleptin}{HealthMarkers}
#' \insertRef{matthews1985homa}{HealthMarkers}
#' \insertRef{simentalmendia2008tyg}{HealthMarkers}
#' \insertRef{yang2006adiponectininsulin}{HealthMarkers}
#' 
#' @importFrom tibble tibble
#' @export
#'
#' @examples
#' # Quick smoke-test
#' df <- data.frame(glucose = 5.6, HbA1c = 44, G0 = 5.5, I0 = 60,
#'                  HDL_c = 1.2, TG = 1.5, BMI = 24)
#' glycemic_markers(df, verbose = FALSE)
#'
#' \donttest{
#' df <- tibble::tibble(
#'   HDL_c       = c(1.0, 1.3),
#'   TG          = c(1.3, 2.0),
#'   BMI         = c(24, 30),
#'   glucose     = c(5.6, 7.1),
#'   HbA1c       = c(44, 38),
#'   C_peptide   = c(300, 500),
#'   G0          = c(5.5, 6.2),
#'   I0          = c(60, 120),
#'   leptin      = c(10, 20),
#'   adiponectin = c(8, 5)
#' )
#' glycemic_markers(df)
#' glycemic_markers(df, verbose = FALSE)
#' glycemic_markers(df, na_action = "omit", verbose = FALSE)
#' }
glycemic_markers <- function(
  data,
  col_map      = NULL,
  na_action    = c("ignore", "warn", "error", "keep", "omit"),
  na_warn_prop = 0.2,
  verbose      = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "glycemic_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  if (!is.data.frame(data)) {
    stop(sprintf("%s(): `data` must be a data.frame or tibble.", fn_name), call. = FALSE)
  }

  # --- Detect and preserve ID column -----------------------------------------
  id_col <- .hm_detect_id_col(data)

  # --- Build col_map: infer missing keys, materialize aliases ----------------
  all_gm_keys <- c("HDL_c", "TG", "BMI", "HbA1c", "C_peptide", "G0", "I0",
                   "leptin", "adiponectin", "glucose")
  req_keys <- c("HDL_c", "TG", "BMI")
  opt_keys <- c("glucose", "HbA1c", "C_peptide", "G0", "I0", "leptin", "adiponectin")

  cm      <- .hm_build_col_map(data, col_map, all_gm_keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  # --- Verbose: column mapping (user-provided vs inferred) -------------------
  .hm_log_cols(cm, col_map, fn_name, verbose, keys = all_gm_keys)

  # --- Pre-computation: fill still-missing required keys from raw inputs -----
  is_missing_key <- function(k) is.null(col_map[[k]]) || !(col_map[[k]] %in% names(data))
  missing_req <- req_keys[vapply(req_keys, is_missing_key, logical(1))]
  if (length(missing_req) > 0L) {
    pc   <- .hm_precompute_from_deps(data, .gm_precompute_deps(),
                                     missing_req, fn = fn_name, verbose = verbose)
    data <- pc$data
    for (k in missing_req) {
      if (is.null(col_map[[k]]) && k %in% names(data)) col_map[[k]] <- k
    }
    missing_req <- req_keys[vapply(req_keys, is_missing_key, logical(1))]
  }
  if (length(missing_req)) {
    stop(sprintf("%s(): missing required columns: %s",
                 fn_name, paste(missing_req, collapse = ", ")), call. = FALSE)
  }

  req_cols <- unname(unlist(col_map[req_keys]))
  miss     <- setdiff(req_cols, names(data))
  if (length(miss)) {
    stop(sprintf("%s(): missing required columns: %s",
                 fn_name, paste(miss, collapse = ", ")), call. = FALSE)
  }

  # --- Determine which optional keys are present ------------------------------
  present_opt   <- vapply(opt_keys, function(k) {
    nm <- col_map[[k]]
    !is.null(nm) && nm %in% names(data)
  }, logical(1))
  used_opt_keys <- opt_keys[present_opt]

  # --- Pre-computation: fill derivable optional keys (one-level) -------------
  missing_opt <- setdiff(opt_keys, used_opt_keys)
  if (length(missing_opt) > 0L) {
    pc          <- .hm_precompute_from_deps(data, .gm_precompute_deps(),
                                            missing_opt, fn = fn_name, verbose = verbose)
    newly_added <- setdiff(names(pc$data), names(data))
    if (length(newly_added) > 0L) {
      data <- pc$data
      for (k in newly_added) {
        if (is.null(col_map[[k]]) && k %in% names(data)) col_map[[k]] <- k
      }
      present_opt <- vapply(opt_keys, function(k) {
        nm <- col_map[[k]]
        !is.null(nm) && nm %in% names(data)
      }, logical(1))
      used_opt_keys <- opt_keys[present_opt]
      missing_opt   <- setdiff(opt_keys, used_opt_keys)
    }
  }

  used_keys <- c(req_keys, used_opt_keys)
  used_cols <- unname(unlist(col_map[used_keys]))

  # --- Verbose: optional inputs + what each missing key affects ---------------
  if (isTRUE(verbose)) {
    idx_deps <- list(
      METS_IR     = c("glucose"),
      prediabetes = c("HbA1c"),
      diabetes    = c("HbA1c"),
      HOMA_CP     = c("C_peptide", "G0"),
      LAR         = c("leptin", "adiponectin"),
      ASI         = c("adiponectin", "I0"),
      TyG_index   = c("glucose")
    )
    na_indices <- character(0)
    for (idx in names(idx_deps)) {
      still_missing <- setdiff(idx_deps[[idx]], used_opt_keys)
      if (length(still_missing)) {
        na_indices <- c(na_indices,
          sprintf("  %s -> NA  [missing: %s]", idx, paste(still_missing, collapse = ", ")))
      }
    }
    deps_all        <- .gm_precompute_deps()
    derivable_hints <- character(0)
    for (k in missing_opt) {
      dep <- deps_all[[k]]
      if (!is.null(dep)) {
        prereqs_absent <- setdiff(dep$needs, names(data))
        if (length(prereqs_absent) == 0L) {
          derivable_hints <- c(derivable_hints,
            sprintf("  %s can be derived from: %s", k, paste(dep$needs, collapse = ", ")))
        } else {
          derivable_hints <- c(derivable_hints,
            sprintf("  %s: provide %s to enable", k, paste(dep$needs, collapse = ", ")))
        }
      }
    }
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(used_opt_keys))
      lines <- c(lines, sprintf("  present:  %s", paste(used_opt_keys, collapse = ", ")))
    if (length(missing_opt))
      lines <- c(lines, sprintf("  missing:  %s", paste(missing_opt, collapse = ", ")))
    if (length(na_indices))
      lines <- c(lines, "  indices -> NA:", na_indices)
    if (length(derivable_hints))
      lines <- c(lines, "  derivable:", derivable_hints)
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # --- Coerce used columns to numeric -----------------------------------------
  for (cn in used_cols) {
    if (!is.numeric(data[[cn]])) {
      old_vals <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old_vals))
      introduced_na <- sum(is.na(data[[cn]]) & !is.na(old_vals))
      if (introduced_na > 0) {
        warning(sprintf("Column '%s' coerced to numeric; NAs introduced: %d",
                        cn, introduced_na), call. = FALSE)
      }
    }
  }

  # --- Extract vectors --------------------------------------------------------
  getv <- function(key) if (key %in% used_keys) data[[col_map[[key]]]] else NULL

  HDL_c       <- data[[col_map$HDL_c]]
  TG          <- data[[col_map$TG]]
  BMI         <- data[[col_map$BMI]]
  glucose     <- getv("glucose")
  HbA1c       <- getv("HbA1c")
  C_peptide   <- getv("C_peptide")
  G0          <- getv("G0")
  I0          <- getv("I0")
  leptin      <- getv("leptin")
  adiponectin <- getv("adiponectin")

  # --- NA handling ------------------------------------------------------------
  used_df     <- data[, used_cols, drop = FALSE]
  nonfin_mask <- vapply(used_cols, function(cn) any(!is.finite(data[[cn]])), logical(1))
  if (na_action == "warn" && any(nonfin_mask)) {
    warning(sprintf("Non-finite values detected in: %s; treated as NA.",
                    paste(used_cols[nonfin_mask], collapse = ", ")), call. = FALSE)
  }
  if (na_action == "error" && any(!stats::complete.cases(used_df))) {
    stop(sprintf("%s(): missing or non-finite values in required inputs with na_action='error'.",
                 fn_name), call. = FALSE)
  } else if (na_action == "omit") {
    keep  <- stats::complete.cases(used_df)
    data  <- data[keep, , drop = FALSE]
    HDL_c <- HDL_c[keep]; TG <- TG[keep]; BMI <- BMI[keep]
    if (!is.null(glucose))     glucose     <- glucose[keep]
    if (!is.null(HbA1c))       HbA1c       <- HbA1c[keep]
    if (!is.null(C_peptide))   C_peptide   <- C_peptide[keep]
    if (!is.null(G0))          G0          <- G0[keep]
    if (!is.null(I0))          I0          <- I0[keep]
    if (!is.null(leptin))      leptin      <- leptin[keep]
    if (!is.null(adiponectin)) adiponectin <- adiponectin[keep]
  }

  # --- Verbose: physiological range check (informational, values not altered) --
  if (isTRUE(verbose)) {
    rules         <- .gm_default_extreme_rules()
    remapped_rules <- list()
    for (nm in names(rules)) {
      cn <- if (!is.null(col_map[[nm]])) col_map[[nm]] else nm
      remapped_rules[[cn]] <- rules[[nm]]
    }
    scan_vars <- intersect(used_cols, names(remapped_rules))
    ex        <- .gm_extreme_scan(data, scan_vars, remapped_rules)
    if (ex$count > 0L) {
      details <- vapply(names(ex$flags), function(v) {
        nb <- sum(ex$flags[[v]])
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", v, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: which markers will be computed and from what ------------------
  if (isTRUE(verbose)) {
    status <- c(
      "  SPISE           [HDL_c, TG, BMI]",
      sprintf("  METS_IR         %s",
              if (!is.null(glucose)) "[glucose, TG, BMI, HDL_c]" else "NA [glucose missing]"),
      sprintf("  prediabetes     %s",
              if (!is.null(HbA1c))  "[HbA1c]"                    else "NA [HbA1c missing]"),
      sprintf("  diabetes        %s",
              if (!is.null(HbA1c))  "[HbA1c]"                    else "NA [HbA1c missing]"),
      sprintf("  HOMA_CP         %s",
              if (!is.null(C_peptide) && !is.null(G0)) "[G0, C_peptide]"
              else "NA [C_peptide/G0 missing]"),
      sprintf("  LAR             %s",
              if (!is.null(leptin) && !is.null(adiponectin)) "[leptin, adiponectin]"
              else "NA [leptin/adiponectin missing]"),
      sprintf("  ASI             %s",
              if (!is.null(adiponectin) && !is.null(I0)) "[adiponectin, I0]"
              else "NA [adiponectin/I0 missing]"),
      sprintf("  TyG_index       %s",
              if (!is.null(glucose)) "[TG, glucose]" else "NA [glucose missing]")
    )
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # --- Compute markers --------------------------------------------------------
  lg <- function(x) .gm_log(x)
  dv <- function(a, b) .gm_safe_div(a, b)

  TG_mgdl <- TG * 88.57

  SPISE <- dv(600 * (HDL_c ^ 0.185), ((TG ^ 0.2) * (BMI ^ 1.338)))

  METS_IR <- if (!is.null(glucose)) {
    dv(lg(2 * glucose + TG) * BMI, lg(HDL_c))
  } else {
    rep(NA_real_, NROW(HDL_c))
  }

  prediabetes <- if (!is.null(HbA1c)) as.integer(HbA1c >= 42) else rep(NA_integer_, NROW(HDL_c))
  diabetes    <- if (!is.null(HbA1c)) as.integer(HbA1c >= 48) else rep(NA_integer_, NROW(HDL_c))

  HOMA_CP <- if (!is.null(C_peptide) && !is.null(G0)) {
    dv(G0 * (C_peptide / 6), 22.5)
  } else {
    rep(NA_real_, NROW(HDL_c))
  }

  LAR <- if (!is.null(leptin) && !is.null(adiponectin)) {
    dv(leptin, adiponectin)
  } else {
    rep(NA_real_, NROW(HDL_c))
  }

  ASI <- if (!is.null(adiponectin) && !is.null(I0)) {
    dv(adiponectin, I0)
  } else {
    rep(NA_real_, NROW(HDL_c))
  }

  TyG_index <- if (!is.null(glucose)) {
    lg((TG_mgdl * (glucose * 18)) / 2)
  } else {
    rep(NA_real_, NROW(HDL_c))
  }

  out <- tibble::tibble(
    SPISE       = SPISE,
    METS_IR     = METS_IR,
    prediabetes = prediabetes,
    diabetes    = diabetes,
    HOMA_CP     = HOMA_CP,
    LAR         = LAR,
    ASI         = ASI,
    TyG_index   = TyG_index
  )

  # --- Prepend ID column if detected ------------------------------------------
  if (!is.null(id_col)) {
    id_vec         <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]]  <- id_vec
    out            <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out            <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary -----------------------------------------------
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# ---- internal helpers (not exported) ----------------------------------------

# Dependency map for one-level pre-computation inside glycemic_markers().
# Each entry: list(needs, describe, compute). `needs` are literal column names
# that must exist in data; `compute(data)` returns a numeric vector.
.gm_precompute_deps <- function() {
  list(
    BMI = list(
      needs    = c("weight", "height"),
      describe = "weight_kg / height_m^2",
      compute  = function(data) {
        w   <- as.numeric(data[["weight"]])
        h   <- as.numeric(data[["height"]])
        h_m <- ifelse(is.finite(h) & h > 3, h / 100, h)  # cm -> m if needed
        w / (h_m ^ 2)
      }
    ),
    glucose = list(
      needs    = "G0",
      describe = "alias: fasting glucose (G0)",
      compute  = function(data) as.numeric(data[["G0"]])
    ),
    G0 = list(
      needs    = "glucose",
      describe = "alias: fasting glucose (glucose)",
      compute  = function(data) as.numeric(data[["glucose"]])
    )
  )
}

.gm_default_extreme_rules <- function() {
  list(
    HDL_c = c(0.1, 5),
    TG = c(0.1, 20),
    BMI = c(10, 80),
    glucose = c(2, 30),
    HbA1c = c(20, 200),
    C_peptide = c(0, 5000),
    G0 = c(2, 30),
    I0 = c(0, 3000),
    leptin = c(0, 200),
    adiponectin = c(0, 300)
  )
}

.gm_extreme_scan <- function(df, vars, rules) {
  flags <- lapply(vars, function(v) {
    if (is.null(rules[[v]])) return(rep(FALSE, NROW(df)))
    rng <- rules[[v]]
    x <- df[[v]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    bad
  })
  names(flags) <- vars
  list(count = sum(vapply(flags, sum, integer(1))), flags = flags)
}

.gm_safe_div <- function(num, den) {
  out <- num / den
  out[!is.finite(out)] <- NA_real_
  out
}

.gm_log <- function(x) {
  y <- x
  y[!(is.finite(y) & y > 0)] <- NA_real_
  out <- suppressWarnings(log(y))
  out[!is.finite(out)] <- NA_real_
  out
}

#' Compute a Suite of Nutrient-Based Health Markers
#'
#' Given a data frame or tibble of routine biochemical labs,
#' `nutrient_markers()` returns a set of widely used ratios, products,
#' and simple percentages that summarize iron metabolism, protein status,
#' omega-3 balance, renal excretion, mineral homeostasis, and aromatic
#' amino-acid patterns.
#'
#' Recognized markers (returned as columns):
#' - FerritinTS: Ferritin / Transferrin saturation
#' - AGR: Albumin / Globulin, where Globulin = Total protein - Albumin
#' - Omega3Index: EPA + DHA (percentage points)
#' - Mg_Cr_Ratio: Magnesium / Creatinine
#' - GlycatedAlbuminPct: (Glycated albumin / Albumin) x 100
#' - UA_Cr_Ratio: Uric acid / Creatinine
#' - BUN_Cr_Ratio: BUN / Creatinine
#' - Ca_x_Phosphate: Calcium x Phosphate
#' - AnionGap: (Na + K) - (Cl + HCO3)
#' - Tyr_Phe_Ratio: Tyrosine / Phenylalanine
#'
#' @param data A data frame or tibble containing subject-level data.
#' @param col_map Optional named list mapping variable keys (see Details) to
#'   column names in `data`. You only need to supply the keys you have; any
#'   markers with missing inputs return `NA`. If NULL, defaults to identity
#'   mapping for all known keys.
#' @param na_action One of c("keep","omit","error") controlling missing-data policy
#'   across the columns referenced by `col_map`.
#'   - "keep" (default): keep NA; outputs become NA where inputs are NA.
#'   - "omit": drop rows with NA in any used input column.
#'   - "error": abort if any used input contains NA.
#' @param na_warn_prop Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness
#'   diagnostics on used input columns. Default 0.2.
#' @param verbose Logical; if TRUE, prints stepwise messages and a final summary via hm_inform. Default FALSE.
#'
#' @details
#' Recognized `col_map` keys and expected units (no automatic conversion):
#' - ferritin: Serum ferritin (ng/mL)
#' - transferrin_sat: Transferrin saturation (%)
#' - albumin: Serum albumin (g/L)
#' - total_protein: Total serum protein (g/L)
#' - EPA: Red-cell EPA as % of total fatty acids
#' - DHA: Red-cell DHA as % of total fatty acids
#' - Mg: Serum magnesium (mmol/L)
#' - creatinine: Serum creatinine (umol/L)
#' - glycated_albumin: Glycated albumin (g/L)
#' - uric_acid: Serum uric acid (umol/L)
#' - BUN: Blood urea nitrogen (mg/dL)
#' - phosphate: Serum phosphate (mmol/L)
#' - calcium: Serum calcium (mmol/L)
#' - Na: Serum sodium (mmol/L)
#' - K: Serum potassium (mmol/L)
#' - Cl: Serum chloride (mmol/L)
#' - HCO3: Serum bicarbonate (mmol/L)
#' - Tyr: Serum tyrosine (umol/L)
#' - Phe: Serum phenylalanine (umol/L)
#'
#' **Unit-mixing notes:**
#' - `BUN_Cr_Ratio` divides BUN (mg/dL) by creatinine (umol/L). This is NOT
#'   numerically equivalent to the standard clinical BUN:Creatinine ratio
#'   (reference range ~10-20), which requires both in mg/dL. The result here
#'   is approximately 88.4x smaller than the standard ratio. Provide creatinine
#'   in mg/dL and adjust `col_map` if you require the standard clinical ratio.
#' - `Mg_Cr_Ratio` divides Mg (mmol/L) by creatinine (umol/L). The result is
#'   1/1000 of the standard Mg/Cr ratio in mmol/mmol. Typically applied to urine;
#'   serum Mg/Cr is not a standard clinical metric.
#'
#' Default `extreme_rules` (inputs) are broad and intended for unit/entry checks:
#' ferritin (0, 2000), transferrin_sat (0, 100), albumin (10, 60), total_protein (40, 100),
#' EPA (0, 20), DHA (0, 20), Mg (0.2, 3), creatinine (20, 2000), glycated_albumin (0, 60),
#' uric_acid (50, 1000), BUN (1, 150), phosphate (0.1, 5), calcium (0.5, 4),
#' Na (100, 200), K (2, 8), Cl (70, 130), HCO3 (5, 45), Tyr (10, 300), Phe (20, 300).
#'
#' @return A tibble with one row per input row and these columns:
#' FerritinTS, AGR, Omega3Index, Mg_Cr_Ratio, GlycatedAlbuminPct,
#' UA_Cr_Ratio, BUN_Cr_Ratio, Ca_x_Phosphate, AnionGap, Tyr_Phe_Ratio.
#'
#' @references
#' \insertRef{harris2004omega3}{HealthMarkers}
#' \insertRef{koga2010glyalb}{HealthMarkers}
#' \insertRef{block1998calcp}{HealthMarkers}
#' \insertRef{waikar2009creak}{HealthMarkers} (creatinine kinetics context)
#' @examples
#' # Quick smoke-test
#' df <- data.frame(ferritin = 50, albumin = 45, uric_acid = 300, Na = 140)
#' nutrient_markers(df, verbose = FALSE)
#'
#' \donttest{
#' df <- tibble::tibble(
#'   ferritin         = c(50, 100),
#'   transferrin_sat  = c(30, 50),
#'   albumin          = c(45, 40),
#'   total_protein    = c(70, 75),
#'   EPA              = c(2.0, 2.5),
#'   DHA              = c(4.0, 4.5),
#'   Mg               = c(0.85, 0.90),
#'   creatinine       = c(80, 90),
#'   glycated_albumin = c(12, 14),
#'   uric_acid        = c(300, 400),
#'   BUN              = c(14, 16),
#'   phosphate        = c(1.0, 1.2),
#'   calcium          = c(2.3, 2.4),
#'   Na               = c(140, 138),
#'   K                = c(4.2, 4.0),
#'   Cl               = c(100, 102),
#'   HCO3             = c(24, 26),
#'   Tyr              = c(60, 70),
#'   Phe              = c(50, 55)
#' )
#' nutrient_markers(df, verbose = TRUE)
#' }
#'
#' @importFrom rlang abort warn inform
#' @importFrom tibble tibble
#' @export
nutrient_markers <- function(
  data,
  col_map      = NULL,
  na_action    = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose      = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "nutrient_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  # HM-CS v2 validation hook; no strictly required keys for this summarizer
  hm_validate_inputs(data, col_map, required_keys = character(0), fn = fn_name)

  keys <- c(
    "ferritin","transferrin_sat","albumin","total_protein",
    "EPA","DHA","Mg","creatinine","glycated_albumin","uric_acid",
    "BUN","phosphate","calcium","Na","K","Cl","HCO3","Tyr","Phe"
  )

  if (is.list(col_map) && length(col_map) > 0L) {
    extra <- setdiff(names(col_map), keys)
    if (length(extra)) {
      rlang::warn(
        sprintf("nutrient_markers(): ignoring unrecognized keys in col_map: %s", paste(extra, collapse = ", ")),
        class = "healthmarkers_nutrient_warn_unrecognized_keys"
      )
      col_map[extra] <- NULL
    }
  }
  cm      <- .hm_build_col_map(data, col_map, keys = keys, fn = fn_name)
  data    <- cm$data
  col_map <- cm$col_map

  mapped    <- unlist(col_map, use.names = TRUE)
  used_cols <- intersect(unname(mapped), names(data))

  # --- Verbose: column mapping
  .hm_log_cols(cm, col_map, fn_name, verbose)

  # --- Verbose: optional inputs
  if (isTRUE(verbose)) {
    avail_keys <- names(col_map)[vapply(names(col_map), function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    absent_keys <- setdiff(keys, avail_keys)
    lines <- sprintf("%s(): optional inputs", fn_name)
    if (length(avail_keys))
      lines <- c(lines, sprintf("  present:  %s", paste(avail_keys, collapse = ", ")))
    if (length(absent_keys))
      lines <- c(lines, sprintf("  missing:  %s", paste(absent_keys, collapse = ", ")))
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # Coerce used inputs to numeric; NA on non-finite
  for (cn in used_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::warn(sprintf("nutrient_markers(): column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # High-missingness diagnostics (debug verbosity)
  .nm_high_missing_diag(data, used_cols, na_warn_prop = na_warn_prop)

  # NA policy over used inputs
  if (na_action == "error") {
    has_na <- Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (any(has_na)) {
      rlang::abort("nutrient_markers(): used input columns contain missing values (na_action='error').",
                   class = "healthmarkers_nm_error_missing_values")
    }
  } else if (na_action == "omit" && length(used_cols)) {
    keep <- !Reduce(`|`, lapply(used_cols, function(cn) is.na(data[[cn]])))
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in used inputs", fn_name, sum(!keep)),
                level = "inform")
    data <- data[keep, , drop = FALSE]
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    rules_info <- .nm_default_extreme_rules()
    ex_info    <- .nm_extreme_scan(data, col_map, rules_info, keys)
    if (ex_info$count > 0L) {
      details <- vapply(names(ex_info$flags), function(cn) {
        nb <- sum(ex_info$flags[[cn]], na.rm = TRUE)
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", cn, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    marker_deps <- list(
      FerritinTS         = c("ferritin", "transferrin_sat"),
      AGR                = c("albumin", "total_protein"),
      Omega3Index        = c("EPA", "DHA"),
      Mg_Cr_Ratio        = c("Mg", "creatinine"),
      GlycatedAlbuminPct = c("glycated_albumin", "albumin"),
      UA_Cr_Ratio        = c("uric_acid", "creatinine"),
      BUN_Cr_Ratio       = c("BUN", "creatinine"),
      Ca_x_Phosphate     = c("calcium", "phosphate"),
      AnionGap           = c("Na", "K", "Cl", "HCO3"),
      Tyr_Phe_Ratio      = c("Tyr", "Phe")
    )
    avail_keys2 <- names(col_map)[vapply(names(col_map), function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    status <- vapply(names(marker_deps), function(m) {
      miss_k <- setdiff(marker_deps[[m]], avail_keys2)
      if (length(miss_k) == 0L)
        sprintf("  %-20s [%s]", m, paste(marker_deps[[m]], collapse = ", "))
      else
        sprintf("  %-20s NA [missing: %s]", m, paste(miss_k, collapse = ", "))
    }, character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  n <- nrow(data)
  getcol <- function(key) {
    nm <- col_map[[key]]
    if (!is.null(nm) && nm %in% names(data)) data[[nm]] else NULL
  }

  denom_zero <- list()

  safe_div <- function(num, den, label) {
    if (is.null(num) || is.null(den)) return(rep(NA_real_, n))
    z <- (!is.na(den)) & (den == 0)
    denom_zero[[label]] <<- sum(z, na.rm = TRUE)
    out <- num / den
    out[!is.finite(out)] <- NA_real_
    out
  }

  ferr <- getcol("ferritin")
  tsat <- getcol("transferrin_sat")
  alb  <- getcol("albumin")
  tprot <- getcol("total_protein")
  epa  <- getcol("EPA")
  dha  <- getcol("DHA")
  mg   <- getcol("Mg")
  cr   <- getcol("creatinine")
  galb <- getcol("glycated_albumin")
  ua   <- getcol("uric_acid")
  bun  <- getcol("BUN")
  phos <- getcol("phosphate")
  ca   <- getcol("calcium")
  na_  <- getcol("Na")
  k_   <- getcol("K")
  cl_  <- getcol("Cl")
  hco3 <- getcol("HCO3")
  tyr  <- getcol("Tyr")
  phe  <- getcol("Phe")

  FerritinTS <- safe_div(ferr, tsat, "FerritinTS")

  AGR <- if (!is.null(alb) && !is.null(tprot)) {
    glob <- tprot - alb
    safe_div(alb, glob, "AGR")
  } else rep(NA_real_, n)

  Omega3Index <- if (!is.null(epa) && !is.null(dha)) epa + dha else rep(NA_real_, n)

  Mg_Cr_Ratio <- safe_div(mg, cr, "Mg_Cr_Ratio")
  GlycatedAlbuminPct <- {
    x <- safe_div(galb, alb, "GlycatedAlbuminPct")
    if (all(is.na(x))) x else x * 100
  }
  UA_Cr_Ratio <- safe_div(ua, cr, "UA_Cr_Ratio")
  BUN_Cr_Ratio <- safe_div(bun, cr, "BUN_Cr_Ratio")

  Ca_x_Phosphate <- if (!is.null(ca) && !is.null(phos)) ca * phos else rep(NA_real_, n)

  AnionGap <- if (!is.null(na_) && !is.null(k_) && !is.null(cl_) && !is.null(hco3)) {
    (na_ + k_) - (cl_ + hco3)
  } else rep(NA_real_, n)

  Tyr_Phe_Ratio <- safe_div(tyr, phe, "Tyr_Phe_Ratio")

  out <- tibble::tibble(
    FerritinTS = FerritinTS,
    AGR = AGR,
    Omega3Index = Omega3Index,
    Mg_Cr_Ratio = Mg_Cr_Ratio,
    GlycatedAlbuminPct = GlycatedAlbuminPct,
    UA_Cr_Ratio = UA_Cr_Ratio,
    BUN_Cr_Ratio = BUN_Cr_Ratio,
    Ca_x_Phosphate = Ca_x_Phosphate,
    AnionGap = AnionGap,
    Tyr_Phe_Ratio = Tyr_Phe_Ratio
  )

  dz_total <- sum(unlist(denom_zero), na.rm = TRUE)
  if (dz_total > 0L) {
    nz <- unlist(denom_zero) > 0
    which_str <- paste(sprintf("%s=%d", names(denom_zero)[nz], unlist(denom_zero)[nz]), collapse = ", ")
    rlang::warn(sprintf("nutrient_markers(): zero denominators detected in %d cases (%s).", dz_total, which_str))
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    out           <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }
  out
}

# ---- internal helpers ---------------------------------------------------------

.nm_high_missing_diag <- function(df, cols, na_warn_prop = 0.2) {
  if (!length(cols)) return(invisible(TRUE))
  for (cn in cols) {
    x <- df[[cn]]; n <- length(x); if (n == 0L) next
    pna <- sum(is.na(x)) / n
    if (pna >= na_warn_prop && pna > 0) {
      hm_inform(level = "debug", msg = sprintf("nutrient_markers(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
  }
  invisible(TRUE)
}

.nm_default_extreme_rules <- function() {
  list(
    ferritin = c(0, 2000),
    transferrin_sat = c(0, 100),
    albumin = c(10, 60),
    total_protein = c(40, 100),
    EPA = c(0, 20),
    DHA = c(0, 20),
    Mg = c(0.2, 3),
    creatinine = c(20, 2000),
    glycated_albumin = c(0, 60),
    uric_acid = c(50, 1000),
    BUN = c(1, 150),
    phosphate = c(0.1, 5),
    calcium = c(0.5, 4),
    Na = c(100, 200),
    K = c(2, 8),
    Cl = c(70, 130),
    HCO3 = c(5, 45),
    Tyr = c(10, 300),
    Phe = c(20, 300)
  )
}

.nm_extreme_scan <- function(df, col_map, rules, keys) {
  count <- 0L
  flags <- list()
  for (key in intersect(names(col_map), names(rules))) {
    cn <- col_map[[key]]
    if (is.null(cn) || !cn %in% names(df)) next
    x <- df[[cn]]
    rng <- rules[[key]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    flags[[cn]] <- bad
    count <- count + sum(bad, na.rm = TRUE)
  }
  list(count = count, flags = flags)
}

# tests/testthat/test-nutrient_markers.R

library(testthat)
library(tibble)
library(HealthMarkers)

cm_id <- function(df) { m <- as.list(names(df)); names(m) <- names(df); m }

test_that("nutrient_markers computes all indices correctly", {
  skip_on_cran()
  df <- tibble(
    ferritin         = 100,
    transferrin_sat  = 50,
    albumin          = 40,
    total_protein    = 70,
    EPA              = 1,
    DHA              = 2,
    Mg               = 1.0,
    creatinine       = 50,
    glycated_albumin = 4,
    uric_acid        = 400,
    BUN              = 10,
    phosphate        = 1.2,
    calcium          = 2.2,
    Na               = 140,
    K                = 4,
    Cl               = 100,
    HCO3             = 24,
    Tyr              = 60,
    Phe              = 30
  )
  out <- nutrient_markers(df, col_map = cm_id(df))
  expect_s3_class(out, "tbl_df")
  expect_named(out, c(
    "FerritinTS", "AGR", "Omega3Index", "Mg_Cr_Ratio",
    "GlycatedAlbuminPct", "UA_Cr_Ratio", "BUN_Cr_Ratio",
    "Ca_x_Phosphate", "AnionGap", "Tyr_Phe_Ratio"
  ))
  expect_equal(out$FerritinTS, 100 / 50, tolerance = 1e-12)
  expect_equal(out$AGR, 40 / (70 - 40), tolerance = 1e-8)
  expect_equal(out$Omega3Index, 1 + 2, tolerance = 1e-12)
  expect_equal(out$Mg_Cr_Ratio, 1.0 / 50, tolerance = 1e-8)
  expect_equal(out$GlycatedAlbuminPct, (4 / 40) * 100, tolerance = 1e-8)
  expect_equal(out$UA_Cr_Ratio, 400 / 50, tolerance = 1e-8)
  expect_equal(out$BUN_Cr_Ratio, 10 / 50, tolerance = 1e-8)
  expect_equal(out$Ca_x_Phosphate, 2.2 * 1.2, tolerance = 1e-8)
  expect_equal(out$AnionGap, (140 + 4) - (100 + 24), tolerance = 1e-12)
  expect_equal(out$Tyr_Phe_Ratio, 60 / 30, tolerance = 1e-8)
})

test_that("nutrient_markers is vectorized over rows", {
  skip_on_cran()
  df <- tibble(
    ferritin         = c(80, 120),
    transferrin_sat  = c(40, 60),
    albumin          = c(42, 38),
    total_protein    = c(74, 72)
  )
  out <- nutrient_markers(df, col_map = cm_id(df))
  expect_equal(nrow(out), 2L)
  expect_equal(out$FerritinTS, c(80 / 40, 120 / 60), tolerance = 1e-8)
  expect_equal(out$AGR, c(42 / (74 - 42), 38 / (72 - 38)), tolerance = 1e-8)
  expect_true(all(is.na(out$Omega3Index)))
})

test_that("missing inputs yield NA rather than error", {
  skip_on_cran()
  df <- tibble(
    ferritin        = 90,
    transferrin_sat = 45,
    albumin         = 42
  )
  out <- nutrient_markers(df, col_map = cm_id(df))
  expect_equal(out$FerritinTS, 90 / 45, tolerance = 1e-8)
  expect_true(is.na(out$AGR))
  expect_true(is.na(out$Omega3Index))
  expect_true(is.na(out$Mg_Cr_Ratio))
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(ferritin = 50, transferrin_sat = 25)
  expect_message(nutrient_markers(df, col_map = cm_id(df), verbose = TRUE), "nutrient_markers")
  expect_message(nutrient_markers(df, col_map = cm_id(df), verbose = TRUE), "col_map")
  expect_message(nutrient_markers(df, col_map = cm_id(df), verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(ferritin = 50, transferrin_sat = 25)
  msgs <- testthat::capture_messages(
    nutrient_markers(df, col_map = cm_id(df), verbose = TRUE)
  )
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("errors if `data` is not a data.frame", {
  skip_on_cran()
  expect_error(nutrient_markers("not a df", col_map = list()), "data\\.frame|tibble")
})

test_that("na_action='error' aborts when used inputs contain NA", {
  skip_on_cran()
  df <- tibble(
    ferritin        = c(100, 90),
    transferrin_sat = c(50, NA_real_)
  )
  expect_error(
    suppressWarnings(nutrient_markers(df, col_map = cm_id(df), na_action = "error")),
    "used input columns contain missing values"
  )
})

test_that("na_action='omit' drops rows with NA in used inputs", {
  skip_on_cran()
  df <- tibble(
    ferritin        = c(100, 90),
    transferrin_sat = c(50, NA_real_)
  )
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    out <- nutrient_markers(df, col_map = cm_id(df), na_action = "omit", verbose = TRUE),
    "omitting 1 rows with NA in used inputs"
  )
  expect_equal(nrow(out), 1L)
  expect_equal(out$FerritinTS, 100 / 50, tolerance = 1e-8)
})

test_that("extreme values produce range note in verbose and no warnings in silent mode", {
  skip_on_cran()
  df <- tibble(ferritin = 5000, transferrin_sat = 50)
  # No warning emitted in non-verbose mode
  expect_no_warning(
    nutrient_markers(df, col_map = cm_id(df), verbose = FALSE)
  )
  # Verbose mode emits a range note informational message
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    nutrient_markers(df, col_map = cm_id(df), verbose = TRUE),
    "range note"
  )
})

test_that("denominator zero emits a single summary warning and yields NA outputs", {
  skip_on_cran()
  df <- tibble(
    Mg         = 0.9,
    uric_acid  = 300,
    BUN        = 14,
    creatinine = 0
  )
  out <- NULL
  expect_warning(
    out <- nutrient_markers(df, col_map = cm_id(df)),
    "zero denominators detected"
  )
  expect_true(all(is.na(out$Mg_Cr_Ratio)))
  expect_true(all(is.na(out$UA_Cr_Ratio)))
  expect_true(all(is.na(out$BUN_Cr_Ratio)))
})

test_that("col_map with unrecognized keys warns and ignores them", {
  skip_on_cran()
  df <- tibble(ferritin = 100, transferrin_sat = 50)
  cm <- list(ferritin = "ferritin", transferrin_sat = "transferrin_sat", LDL = "LDL")
  expect_warning(
    nutrient_markers(df, col_map = cm),
    "ignoring unrecognized keys in col_map"
  )
})

#' Compute a suite of hormone ratio markers with QA and verbose summaries
#'
#' Ratios computed:
#' - FAI = (total_testosterone / SHBG) * 100
#' - LH_FSH = LH / FSH
#' - E2_P = estradiol / progesterone
#' - T3_T4 = free_T3 / free_T4
#' - ARR = aldosterone / renin
#' - Ins_Glu = insulin / glucagon
#' - GH_IGF1 = GH / IGF1
#' - PRL_T = prolactin / total_testosterone
#' - CAR_slope = (cortisol_30 - cortisol_0) / 30
#'
#' @param data Data frame or tibble with mapped hormone inputs.
#' @param col_map Named list mapping the required keys to column names:
#'   total_testosterone, SHBG, LH, FSH, estradiol, progesterone, free_T3, free_T4,
#'   aldosterone, renin, insulin, glucagon, GH, IGF1, prolactin, cortisol_0, cortisol_30.
#' @param na_action One of `c("keep","omit","error","warn","ignore")`. `"keep"`/`"ignore"` leave NAs;
#'   `"omit"` drops rows with any NA in used inputs; `"error"` aborts;
#'   `"warn"` also warns about high missingness.
#' @param na_warn_prop Proportion in \eqn{[0,1]} for high-missingness warnings
#'   when `na_action = "warn"`. Default 0.2.
#' @param verbose Logical; if `TRUE` (default), prints column mapping, input
#'   availability, inference notes, physiological range information
#'   (informational only, values not altered), computing markers, and a
#'   per-column results summary.
#' @details
#' Some inputs may be inferred when missing (for example, `free_T3` from
#' `TSH` + `free_T4`, or `GH` from `IGF1`) using internal heuristics. These
#' inferred values are intended for exploratory feature engineering only and
#' must not be treated as clinical substitutes for directly measured assays.
#' Ratios such as FAI, ARR, and CAR_slope have established literature usage;
#' several other outputs are simple arithmetic composites included for
#' feature-engineering convenience and may not have a single canonical
#' derivation paper.
#' @references
#' \insertRef{sowers2009fai}{HealthMarkers}
#' \insertRef{funder2016primaryaldosteronism}{HealthMarkers}
#' \insertRef{clow2004car}{HealthMarkers}
#' @return Tibble with one column per computable ratio. If an ID column is
#'   detected in `data` (e.g. `id`, `IID`, `participant_id`), it is prepended
#'   as the first output column.
#'
#' @examples
#' df <- data.frame(
#'   TT = c(15, 12), SHBG = c(40, 35), LH = c(5, 6), FSH = c(4, 5),
#'   E2 = c(100, 120), Prog = c(0.5, 0.6), fT3 = c(4.5, 4.2),
#'   fT4 = c(15, 14), Aldo = c(200, 180), Renin = c(10, 12),
#'   Ins = c(60, 70), Gluc = c(8, 9), GH = c(1.2, 1.0),
#'   IGF1 = c(180, 160), Prl = c(10, 12), Cort0 = c(400, 380),
#'   Cort30 = c(600, 580)
#' )
#' col_map <- list(
#'   total_testosterone = "TT", SHBG = "SHBG", LH = "LH", FSH = "FSH",
#'   estradiol = "E2", progesterone = "Prog", free_T3 = "fT3",
#'   free_T4 = "fT4", aldosterone = "Aldo", renin = "Renin",
#'   insulin = "Ins", glucagon = "Gluc", GH = "GH", IGF1 = "IGF1",
#'   prolactin = "Prl", cortisol_0 = "Cort0", cortisol_30 = "Cort30"
#' )
#' hormone_markers(df, col_map = col_map)
#' @export
hormone_markers <- function(
  data,
  col_map      = NULL,
  na_action    = c("keep", "omit", "error", "warn", "ignore"),
  na_warn_prop = 0.2,
  verbose      = TRUE
) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name   <- "hormone_markers"
  .hm_log_input(data, data_name, fn_name, verbose)
  na_action <- match.arg(na_action)
  if (na_action == "keep") na_action <- "ignore"

  # --- Detect and preserve ID column
  id_col <- .hm_detect_id_col(data)

  # Per-ratio input dependencies
  ratio_defs <- list(
    FAI       = c("total_testosterone", "SHBG"),
    LH_FSH    = c("LH", "FSH"),
    E2_P      = c("estradiol", "progesterone"),
    E2_T      = c("estradiol", "total_testosterone"),
    T3_T4     = c("free_T3", "free_T4"),
    TSH_fT4   = c("TSH", "free_T4"),
    ARR       = c("aldosterone", "renin"),
    Ins_Glu   = c("insulin", "glucagon"),
    GH_IGF1   = c("GH", "IGF1"),
    PRL_T     = c("prolactin", "total_testosterone"),
    CAR_slope = c("cortisol_0", "cortisol_30")
  )

  # Inference rules: derive a missing key from available mapped keys.
  # Each entry: list(target_key  = name to inject into col_map,
  #                   needs       = mapped keys required,
  #                   compute     = function(data, col_map) -> vector,
  #                   label       = short description for messaging)
  infer_rules <- list(
    list(
      target  = "free_T3",
      needs   = c("TSH", "free_T4"),
      compute = function(d, cm) {
        # Internal heuristic estimate: fT3 ~ fT4 * 0.33 * TSH^(-0.20)
        # Source DOI not yet verified; keep as heuristic estimate.
        fT4 <- d[[cm[["free_T4"]]]]
        tsh <- d[[cm[["TSH"]]]]
        ifelse(is.finite(tsh) & tsh > 0 & is.finite(fT4),
               fT4 * 0.33 * tsh^(-0.20), NA_real_)
      },
      label   = "free_T3 estimated from TSH + free_T4 (heuristic)"
    ),
    list(
      target  = "GH",
      needs   = c("IGF1"),
      compute = function(d, cm) {
        # Rough equivalence: median GH ~ IGF1 / 22.5 (nmol-based scale)
        # Use only as a last resort; flag as estimated
        d[[cm[["IGF1"]]]] / 22.5
      },
      label   = "GH estimated from IGF1 (IGF1 / 22.5, approximate)"
    )
  )

  # Validate inputs (HM-CS v3)
  if (!is.data.frame(data)) {
    rlang::abort("hormone_markers(): `data` must be a data.frame or tibble.", class = "healthmarkers_horm_error_data_type")
  }

  all_hm_keys <- c("total_testosterone","SHBG","LH","FSH","estradiol","progesterone",
    "free_T3","free_T4","TSH","aldosterone","renin",
    "insulin","IGF1","prolactin","cortisol_0","cortisol_30")
  cm_hm   <- .hm_build_col_map(data, col_map, all_hm_keys, fn = fn_name)
  data    <- cm_hm$data
  col_map <- cm_hm$col_map

  if (is.null(col_map) || !is.list(col_map)) {
    rlang::abort("hormone_markers(): `col_map` must be a named list.", class = "healthmarkers_horm_error_colmap_type")
  }

  # --- Inference pass: derive missing keys from available mapped data --------
  inferred_keys <- character(0)
  for (rule in infer_rules) {
    if (rule$target %in% names(col_map)) next          # already explicitly mapped
    if (!all(rule$needs %in% names(col_map))) next      # prerequisites not available
    tmp_col <- paste0(".hm_inferred_", rule$target)
    data[[tmp_col]] <- rule$compute(data, col_map)
    col_map[[rule$target]] <- tmp_col
    inferred_keys <- c(inferred_keys, rule$target)
    hm_inform(level = if (isTRUE(verbose)) "inform" else "debug",
              msg   = sprintf("hormone_markers(): inferred '%s' \u2014 %s", rule$target, rule$label))
  }
  # --------------------------------------------------------------------------

  # Determine which ratios can be computed from the supplied col_map
  avail_ratios <- vapply(ratio_defs, function(keys) all(keys %in% names(col_map)), logical(1))
  skipped_ratios <- names(ratio_defs)[!avail_ratios]
  if (length(skipped_ratios) > 0L && isTRUE(verbose))
    hm_inform(sprintf("%s(): skipping %d ratio(s) with unmapped inputs: %s",
                      fn_name, length(skipped_ratios), paste(skipped_ratios, collapse = ", ")),
              level = "inform")
  if (!any(avail_ratios))
    rlang::abort("hormone_markers(): no computable ratios; supply at least one pair of mapped inputs.",
                 class = "healthmarkers_horm_error_no_ratios")

  used_keys <- unique(unlist(ratio_defs[avail_ratios], use.names = FALSE))

  # Ensure each used mapping is a single non-empty character string
  bad_keys <- vapply(used_keys, function(k) {
    v <- col_map[[k]]
    !is.character(v) || length(v) != 1L || !nzchar(v)
  }, logical(1))
  if (any(bad_keys))
    rlang::abort(paste0("missing required columns: ", paste(used_keys[bad_keys], collapse = ", ")),
                 class = "healthmarkers_horm_error_bad_map_values")

  # Mapped columns must exist in data
  used_cols <- unname(vapply(used_keys, function(k) col_map[[k]], character(1)))
  missing_cols <- setdiff(used_cols, names(data))
  if (length(missing_cols))
    rlang::abort(paste0("missing required columns: ", paste(missing_cols, collapse = ", ")),
                 class = "healthmarkers_horm_error_missing_columns")
  # na_warn_prop sanity
  if (!(is.numeric(na_warn_prop) && length(na_warn_prop) == 1L && is.finite(na_warn_prop) && na_warn_prop >= 0 && na_warn_prop <= 1)) {
    rlang::abort("hormone_markers(): `na_warn_prop` must be a single number in [0,1].",
                 class = "healthmarkers_horm_error_na_warn_prop")
  }

  # --- Verbose: col_map
  .hm_log_cols(cm_hm, col_map, fn_name, verbose)

  # --- Verbose: optional inputs (ratio availability)
  if (isTRUE(verbose)) {
    all_needed <- unique(unlist(ratio_defs, use.names = FALSE))
    avail_k    <- names(col_map)[vapply(names(col_map), function(k) {
      !is.null(col_map[[k]]) && col_map[[k]] %in% names(data)
    }, logical(1))]
    absent_k   <- setdiff(all_needed, avail_k)
    lines      <- sprintf("%s(): optional inputs", fn_name)
    if (length(avail_k))
      lines <- c(lines, sprintf("  present:  %s", paste(intersect(all_needed, avail_k), collapse = ", ")))
    if (length(absent_k))
      lines <- c(lines, sprintf("  missing:  %s", paste(absent_k, collapse = ", ")))
    if (length(skipped_ratios))
      lines <- c(lines, sprintf("  ratios skipped (missing inputs): %s",
                                paste(skipped_ratios, collapse = ", ")))
    hm_inform(paste(lines, collapse = "\n"), level = "inform")
  }

  # Coerce to numeric; warn if NAs introduced; sanitize non-finite to NA
  for (cn in used_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(data[[cn]] <- as.numeric(old))
      introduced_na <- sum(is.na(data[[cn]]) & !is.na(old))
      if (introduced_na > 0L) {
        rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, introduced_na),
                    class = "healthmarkers_horm_warn_na_coercion")
      }
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  # HM-CS: omit rows with any NA in used inputs
  if (na_action == "omit") {
    keep_rows <- stats::complete.cases(data[, used_cols, drop = FALSE])
    if (isTRUE(verbose))
      hm_inform(sprintf("%s(): omitting %d rows with NA in required inputs", fn_name, sum(!keep_rows)),
                level = "inform")
    data <- data[keep_rows, , drop = FALSE]
  }

  # Missingness QA
  qa <- .hor_quality_scan(data, used_cols, na_warn_prop)
  if (na_action == "warn") {
    if (length(qa$any_na)) {
      rlang::warn(sprintf("Missing values in: %s.", paste(qa$any_na, collapse = ", ")),
                  class = "healthmarkers_horm_warn_any_na")
    }
    if (length(qa$high_na)) {
      rlang::warn(sprintf("High missingness (>= %.0f%%): %s.", 100 * na_warn_prop, paste(qa$high_na, collapse = ", ")),
                  class = "healthmarkers_horm_warn_high_na")
    }
  }
  if (na_action == "error" && length(qa$any_na)) {
    rlang::abort("hormone_markers(): missing or non-finite values in required inputs with na_action='error'.",
                 class = "healthmarkers_horm_error_missing_values")
  }

  # --- Verbose: physiological range check (informational, values not altered)
  if (isTRUE(verbose)) {
    rules_info   <- .hor_default_extreme_rules()
    flagged_info <- .hor_extreme_scan(data, col_map, rules_info)
    total_flagged_info <- sum(vapply(flagged_info, function(x) sum(x, na.rm = TRUE), integer(1)))
    if (total_flagged_info > 0L) {
      details <- vapply(names(flagged_info), function(nm) {
        nb <- sum(flagged_info[[nm]], na.rm = TRUE)
        if (nb > 0L) sprintf("  %s: %d value(s) outside plausible range", nm, nb) else ""
      }, character(1))
      details <- details[nzchar(details)]
      hm_inform(
        paste(c(sprintf("%s(): range note (informational, values not altered):", fn_name),
                details), collapse = "\n"),
        level = "inform"
      )
    }
  }

  # --- Verbose: computing markers
  if (isTRUE(verbose)) {
    status <- vapply(names(ratio_defs), function(idx) {
      keys_for_idx <- ratio_defs[[idx]]
      if (avail_ratios[idx])
        sprintf("  %-12s [%s]", idx, paste(keys_for_idx, collapse = ", "))
      else {
        miss_k <- setdiff(keys_for_idx, names(col_map))
        sprintf("  %-12s NA [missing: %s]", idx, paste(miss_k, collapse = ", "))
      }
    }, character(1))
    hm_inform(
      paste(c(sprintf("%s(): computing markers:", fn_name), status), collapse = "\n"),
      level = "inform"
    )
  }

  # Extract only the vectors actually needed
  v <- function(k) data[[col_map[[k]]]]

  sdiv <- function(a, b) { z <- a / b; z[!is.finite(z)] <- NA_real_; z }

  # Compute only the available ratios
  out_list <- list()
  if (avail_ratios["FAI"])       out_list$FAI       <- sdiv(v("total_testosterone"), v("SHBG")) * 100
  if (avail_ratios["LH_FSH"])    out_list$LH_FSH    <- sdiv(v("LH"), v("FSH"))
  if (avail_ratios["E2_P"])      out_list$E2_P      <- sdiv(v("estradiol"), v("progesterone"))
  if (avail_ratios["E2_T"])      out_list$E2_T      <- sdiv(v("estradiol"), v("total_testosterone"))
  if (avail_ratios["T3_T4"])     out_list$T3_T4     <- sdiv(v("free_T3"), v("free_T4"))
  if (avail_ratios["TSH_fT4"])   out_list$TSH_fT4   <- sdiv(v("TSH"), v("free_T4"))
  if (avail_ratios["ARR"])       out_list$ARR       <- sdiv(v("aldosterone"), v("renin"))
  if (avail_ratios["Ins_Glu"])   out_list$Ins_Glu   <- sdiv(v("insulin"), v("glucagon"))
  if (avail_ratios["GH_IGF1"])   out_list$GH_IGF1   <- sdiv(v("GH"), v("IGF1"))
  if (avail_ratios["PRL_T"])     out_list$PRL_T     <- sdiv(v("prolactin"), v("total_testosterone"))
  if (avail_ratios["CAR_slope"]) out_list$CAR_slope <- sdiv(v("cortisol_30") - v("cortisol_0"), 30)

  # Tag inferred columns so downstream callers know
  out <- tibble::as_tibble(out_list)
  if (length(inferred_keys) > 0L) {
    inferred_ratios <- names(ratio_defs)[vapply(ratio_defs, function(keys) any(keys %in% inferred_keys), logical(1))]
    inferred_ratios <- intersect(inferred_ratios, names(out))
    attr(out, "inferred_inputs") <- inferred_keys
    attr(out, "inferred_ratios") <- inferred_ratios
  }

  # --- Prepend ID column if detected
  if (!is.null(id_col)) {
    id_vec        <- data[[id_col]][seq_len(nrow(out))]
    out[[id_col]] <- id_vec
    keep_cols     <- c(id_col, setdiff(names(out), id_col))
    out           <- out[, keep_cols, drop = FALSE]
    out           <- tibble::as_tibble(out)
  }

  # --- Verbose: results summary
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

# --- internal helpers ---------------------------------------------------------

.hor_quality_scan <- function(df, cols, na_warn_prop = 0.2) {
  any_na <- character(0); high_na <- character(0)
  for (cn in cols) {
    x <- df[[cn]]
    if (any(is.na(x))) any_na <- c(any_na, cn)
    prop <- mean(is.na(x))
    if (is.finite(prop) && prop >= na_warn_prop) high_na <- c(high_na, cn)
  }
  list(any_na = unique(any_na), high_na = unique(high_na))
}

.hor_default_extreme_rules <- function() {
  # Plausible broad ranges for lab units; tuned to tests where needed
  list(
    total_testosterone = c(0, 100),
    SHBG = c(0, 200),
    LH = c(0, 200),
    FSH = c(0, 200),
    estradiol = c(0, 10000),
    progesterone = c(0, 1000),
    TSH = c(0, 100),
    free_T3 = c(0, 20),
    free_T4 = c(0, 50),
    aldosterone = c(0, 1000),
    renin = c(0, 50),
    insulin = c(0, 1000),
    glucagon = c(0, 500),
    GH = c(0, 200),
    IGF1 = c(0, 2000),
    prolactin = c(0, 500),
    cortisol_0 = c(0, 2000),
    cortisol_30 = c(0, 2000)
  )
}

.hor_extreme_scan <- function(df, col_map, rules) {
  flags <- list()
  for (nm in names(rules)) {
    cn <- col_map[[nm]]
    if (is.null(cn) || !(cn %in% names(df))) next
    rng <- rules[[nm]]
    x <- df[[cn]]
    flags[[nm]] <- is.finite(x) & (x < rng[1] | x > rng[2])
  }
  flags
}

library(testthat)
library(tibble)
library(HealthMarkers)

test_that("errors when no ratio can be computed from supplied col_map", {
  skip_on_cran()
  df <- tibble(x = 1)
  # Only one key provided; no ratio has both its inputs mapped
  expect_error(
    hormone_markers(df, col_map = list(total_testosterone = "x")),
    "no computable ratios"
  )
})

test_that("computes all eleven ratios correctly", {
  skip_on_cran()
  df <- tibble(
    total_testosterone = 10, SHBG = 2,
    LH = 8, FSH = 4,
    estradiol = 100, progesterone = 50,
    TSH = 2, free_T3 = 5, free_T4 = 10,
    aldosterone = 20, renin = 4,
    insulin = 20, glucagon = 10,
    GH = 2, IGF1 = 4,
    prolactin = 12,
    cortisol_0 = 200, cortisol_30 = 260
  )
  cm <- setNames(as.list(names(df)), names(df))
  out <- hormone_markers(df, col_map = cm, na_action = "keep")
  expect_named(out, c(
    "FAI", "LH_FSH", "E2_P", "E2_T", "T3_T4", "TSH_fT4",
    "ARR", "Ins_Glu", "GH_IGF1", "PRL_T", "CAR_slope"
  ))
  expect_equal(out$FAI,       (10 / 2) * 100)
  expect_equal(out$LH_FSH,    8 / 4)
  expect_equal(out$E2_P,      100 / 50)
  expect_equal(out$E2_T,      100 / 10)
  expect_equal(out$T3_T4,     5 / 10)
  expect_equal(out$TSH_fT4,   2 / 10)
  expect_equal(out$ARR,       20 / 4)
  expect_equal(out$Ins_Glu,   20 / 10)
  expect_equal(out$GH_IGF1,   2 / 4)
  expect_equal(out$PRL_T,     12 / 10)
  expect_equal(out$CAR_slope, (260 - 200) / 30)
})

test_that("verbose emits preparing, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    total_testosterone = 10, SHBG = 2, LH = 8, FSH = 4,
    estradiol = 100, progesterone = 50, TSH = 2, free_T3 = 5, free_T4 = 10,
    aldosterone = 20, renin = 4, insulin = 20, glucagon = 10,
    GH = 2, IGF1 = 4, prolactin = 12, cortisol_0 = 200, cortisol_30 = 260
  )
  cm <- setNames(as.list(names(df)), names(df))
  expect_message(hormone_markers(df, col_map = cm, verbose = TRUE, na_action = "keep"), "hormone_markers")
  expect_message(hormone_markers(df, col_map = cm, verbose = TRUE, na_action = "keep"), "col_map")
  expect_message(hormone_markers(df, col_map = cm, verbose = TRUE, na_action = "keep"), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    total_testosterone = 10, SHBG = 2, LH = 8, FSH = 4,
    estradiol = 100, progesterone = 50, TSH = 2, free_T3 = 5, free_T4 = 10,
    aldosterone = 20, renin = 4, insulin = 20, glucagon = 10,
    GH = 2, IGF1 = 4, prolactin = 12, cortisol_0 = 200, cortisol_30 = 260
  )
  cm <- setNames(as.list(names(df)), names(df))
  msgs <- testthat::capture_messages(hormone_markers(df, col_map = cm, verbose = TRUE, na_action = "keep"))
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("handles NA inputs gracefully", {
  skip_on_cran()
  df <- tibble(
    total_testosterone = NA, SHBG = 2,
    LH = 1, FSH = 1,
    estradiol = 1, progesterone = 1,
    TSH = 2, free_T3 = 1, free_T4 = 1,
    aldosterone = 1, renin = 1,
    insulin = 1, glucagon = 1,
    GH = 1, IGF1 = 1,
    prolactin = 1,
    cortisol_0 = 1, cortisol_30 = 1
  )
  cm <- setNames(as.list(names(df)), names(df))
  out <- hormone_markers(df, col_map = cm, na_action = "keep")
  expect_true(is.na(out$FAI))  # NA numerator propagates
  expect_equal(out$LH_FSH, 1 / 1)
})

test_that("inference: free_T3 derived from TSH + free_T4 enables T3_T4", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- tibble(
    total_testosterone = 10, SHBG = 2,
    estradiol = 100,
    TSH = 2, free_T4 = 10,       # free_T3 NOT mapped
    aldosterone = 20, renin = 4,
    IGF1 = 100,                   # GH NOT mapped
    cortisol_0 = 200, cortisol_30 = 260
  )
  cm <- setNames(as.list(names(df)), names(df))
  msgs <- testthat::capture_messages(
    out <- hormone_markers(df, col_map = cm, verbose = TRUE, na_action = "keep")
  )
  # T3_T4 should be present via inferred free_T3
  expect_true("T3_T4" %in% names(out))
  # GH_IGF1 should be present via inferred GH
  expect_true("GH_IGF1" %in% names(out))
  # E2_T uses estradiol + total_testosterone (both mapped)
  expect_true("E2_T" %in% names(out))
  # TSH_fT4 from TSH + free_T4 (both mapped)
  expect_true("TSH_fT4" %in% names(out))
  # Inference messages emitted
  expect_true(any(grepl("inferred", msgs)))
  # Attributes mark which keys were inferred
  expect_true(!is.null(attr(out, "inferred_inputs")))
})

#' FRAX 10-Year Fracture Risk Score (simplified placeholder)
#'
#' Estimates the 10-year probabilities of major osteoporotic and hip fracture
#' using a simplified, non-validated approximation based on FRAX risk factors.
#' This is for development/demo only and does not implement the proprietary FRAX algorithm.
#'
#' @details
#' Important: this function is an educational placeholder and must not be used for
#' clinical decision-making, patient counseling, or guideline-based treatment selection.
#'
#' @param data A data frame or tibble with inputs.
#' @param col_map Named list mapping required and optional inputs:
#'   - Required: age, sex
#'   - Optional binary risk factors: prior_fracture, parent_fracture, steroids,
#'     rheumatoid, secondary_op, smoker, alcohol
#'   - Optional: bmd (T-score; if present and in (-5,0], used to adjust risk)
#' @param country optional country/region code for FRAX calibration (accepted, currently unused).
#' @param na_action One of c("keep","omit","error","ignore","warn").
#' @param verbose Logical; if TRUE, emits progress via rlang::inform.
#' @return Tibble with frax_major_percent and frax_hip_percent.
#'
#' @examples
#' df <- data.frame(Age = c(65, 72, 58), Sex = c("female", "female", "male"))
#' frax_score(df)
#' @export
frax_score <- function(
  data,
  col_map = NULL,
  na_action = c("keep","omit","error","ignore","warn"),
  country = NULL,
  verbose = TRUE
) {
  # Accept but do not use; normalize without emitting warnings
  if (!is.null(country)) {
    country <- as.character(country)[1L]
  }

  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "frax_score"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col  <- .hm_detect_id_col(data)
  .na <- .hm_normalize_na_action(match.arg(na_action))
  na_action_raw <- .na$na_action_raw
  na_action_eff <- .na$na_action_eff

  # Validate inputs
  if (!is.data.frame(data)) {
    rlang::abort("frax_score(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_frax_error_data_type")
  }
  req <- c("age","sex")
  all_frax_keys <- c("age","sex","bmd_t","prior_fracture","parent_fracture","steroids","smoker","alcohol")
  cm      <- .hm_build_col_map(data, col_map, all_frax_keys, fn = "frax_score")
  data    <- cm$data
  col_map <- cm$col_map
  if (!is.list(col_map) || is.null(names(col_map))) {
    rlang::abort("frax_score(): `col_map` must be a named list.",
                 class = "healthmarkers_frax_error_colmap_type")
  }
  missing_keys <- setdiff(req, names(col_map))
  if (length(missing_keys)) {
    rlang::abort(paste0("frax_score(): missing col_map entries for: ",
                        paste(missing_keys, collapse = ", ")),
                 class = "healthmarkers_frax_error_missing_map")
  }
  mapped_req <- unname(unlist(col_map[req], use.names = FALSE))
  if (any(!nzchar(mapped_req))) {
    bad <- req[!nzchar(unname(unlist(col_map[req])))]
    rlang::abort(paste0("frax_score(): missing col_map entries for: ",
                        paste(bad, collapse = ", ")),
                 class = "healthmarkers_frax_error_bad_map_values")
  }
  missing_cols <- setdiff(mapped_req, names(data))
  if (length(missing_cols)) {
    rlang::abort(paste0("frax_score(): missing required columns in data: ",
                        paste(missing_cols, collapse = ", ")),
                 class = "healthmarkers_frax_error_missing_columns")
  }

  .hm_log_cols(cm, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  frax_major_percent [age, sex, risk factors, bmd]\n  frax_hip_percent [age, sex, risk factors, bmd]", fn_name), level = "inform")

  # Build list of used columns that are present in data
  opt_keys <- c("prior_fracture","parent_fracture","steroids","rheumatoid",
                "secondary_op","smoker","alcohol","bmd")
  mapped_all <- unlist(col_map[c(req, opt_keys)], use.names = TRUE)
  mapped_all <- mapped_all[!is.na(mapped_all) & nzchar(mapped_all)]
  used_cols <- intersect(unname(mapped_all), names(data))

  # Coerce numeric-like where appropriate; keep sex as character/factor
  numeric_like_keys <- c("age","prior_fracture","parent_fracture","steroids",
                         "rheumatoid","secondary_op","smoker","alcohol","bmd")
  numeric_like_cols <- unname(unlist(col_map[intersect(names(col_map), numeric_like_keys)]))
  numeric_like_cols <- numeric_like_cols[!is.na(numeric_like_cols) & nzchar(numeric_like_cols)]
  numeric_like_cols <- intersect(numeric_like_cols, used_cols)

  for (cn in numeric_like_cols) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings(new <- as.numeric(old))
      intro <- sum(is.na(new) & !is.na(old))
      if (intro > 0L) {
        rlang::warn(sprintf("Column '%s' coerced to numeric; NAs introduced: %d", cn, intro),
                    class = "healthmarkers_frax_warn_na_coercion")
      }
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  age <- data[[col_map$age]]
  sex <- data[[col_map$sex]]
  # Map sex
  sex_chr <- tolower(as.character(sex))
  sex_norm <- ifelse(substr(sex_chr, 1, 1) %in% c("m","1"), "male",
                ifelse(substr(sex_chr, 1, 1) %in% c("f","0"), "female", NA_character_))
  if (any(is.na(sex_norm))) {
    rlang::warn("frax_score(): unknown sex codes encountered; related rows set to NA.",
                class = "healthmarkers_frax_warn_unknown_sex")
  }
  # Optional risk factors (numeric 0/1)
  as01 <- function(x) {
    if (is.null(x)) return(NULL)
    if (is.logical(x)) return(as.integer(x))
    if (is.numeric(x)) return(as.integer(x > 0))
    xv <- suppressWarnings(as.numeric(x))
    as.integer(ifelse(is.na(xv), NA, xv > 0))
  }
  getn <- function(key) {
    nm <- col_map[[key]]
    if (!is.null(nm) && nzchar(nm) && nm %in% names(data)) data[[nm]] else NULL
  }
  prior_fx <- as01(getn("prior_fracture")); parent_fx <- as01(getn("parent_fracture"))
  steroids <- as01(getn("steroids")); rheumatoid <- as01(getn("rheumatoid"))
  secondary <- as01(getn("secondary_op")); smoker <- as01(getn("smoker")); alcohol <- as01(getn("alcohol"))
  bmd <- getn("bmd")

  # NA policy on required inputs
  any_na_req <- is.na(age) | is.na(sex_norm)
  if (na_action_raw == "warn" && any(any_na_req)) {
    rlang::warn("Missing age or sex values; FRAX outputs will be NA for those entries.",
                class = "healthmarkers_frax_warn_missing_inputs")
  }
  if (na_action_eff == "error" && any(any_na_req)) {
    rlang::abort("frax_score(): required inputs contain missing values (na_action='error').",
                 class = "healthmarkers_frax_error_missing_values")
  }
  keep <- if (na_action_eff == "omit") !any_na_req else rep(TRUE, length(age))

  # Subset for computation
  d_age <- age[keep]
  d_sex <- sex_norm[keep]
  d_prior <- if (!is.null(prior_fx)) prior_fx[keep] else rep(0L, sum(keep))
  d_parent <- if (!is.null(parent_fx)) parent_fx[keep] else rep(0L, sum(keep))
  d_ster <- if (!is.null(steroids)) steroids[keep] else rep(0L, sum(keep))
  d_rheu <- if (!is.null(rheumatoid)) rheumatoid[keep] else rep(0L, sum(keep))
  d_sec  <- if (!is.null(secondary)) secondary[keep] else rep(0L, sum(keep))
  d_smok <- if (!is.null(smoker)) smoker[keep] else rep(0L, sum(keep))
  d_alc  <- if (!is.null(alcohol)) alcohol[keep] else rep(0L, sum(keep))
  d_bmd  <- if (!is.null(bmd)) bmd[keep] else rep(NA_real_, sum(keep))

  # Domain warnings
  if (any(is.finite(d_age) & (d_age < 40 | d_age > 90))) {
    rlang::warn(
      "frax_score(): age outside typical FRAX range (40-90 years) detected.",
      class = "healthmarkers_frax_warn_age_range"
    )
  }
  if (!is.null(d_bmd) && any(is.finite(d_bmd) & (d_bmd < -6 | d_bmd > 2))) {
    hm_inform(level = "debug", msg = "frax_score(): BMD T-scores outside plausible range (-6 to +2) detected.")
  }

  hm_inform("frax_score(): computing risk", level = "debug")

  n <- length(d_age)
  risk_major <- rep(NA_real_, n)

  # Baseline by sex and age (placeholder logic)
  is_female <- !is.na(d_sex) & d_sex == "female"
  is_male   <- !is.na(d_sex) & d_sex == "male"

  base <- rep(NA_real_, n)
  base[is_female] <- 2 + 0.3 * pmax(d_age[is_female] - 50, 0)
  base[is_male]   <- 1 + 0.2 * pmax(d_age[is_male] - 50, 0)

  # Add risk-factor contributions (binary 0/1; NA contributes 0)
  nz <- function(x) { x[is.na(x)] <- 0L; x }
  contrib <- 5 * nz(d_prior) + 3 * nz(d_parent) + 4 * nz(d_ster) +
            2 * nz(d_rheu) + 2 * nz(d_sec) + 2 * nz(d_smok) + 2 * nz(d_alc)

  risk_major <- base + contrib

  # BMD as T-score in [-5, 0] reduces/increases risk
  if (!all(is.na(d_bmd))) {
    tscore <- ifelse(is.finite(d_bmd) & d_bmd <= 0 & d_bmd >= -5, d_bmd, NA_real_)
    adj <- pmax(-tscore - 1, 0) * 2
    adj[!is.finite(adj)] <- 0
    risk_major <- risk_major + adj
  }

  # Cap major risk and set to NA where sex is unknown
  risk_major[is.na(d_sex)] <- NA_real_
  risk_major <- pmin(risk_major, 95)

  # Hip fracture risk as a fraction of major, increasing with age
  frac <- pmin(0.4 + 0.01 * pmax(d_age - 50, 0), 0.8)
  risk_hip <- pmin(risk_major * frac, 95)

  # Round to one decimal place
  risk_major <- round(risk_major, 1)
  risk_hip <- round(risk_hip, 1)

  out <- tibble::tibble(
    frax_major_percent = risk_major,
    frax_hip_percent   = risk_hip,
    frax_sex_norm      = d_sex,
    frax_age_used      = d_age,
    frax_bmd_tscore    = d_bmd
  )

  if (na_action_eff != "omit") {
    res <- tibble::tibble(
      frax_major_percent = rep(NA_real_, nrow(data)),
      frax_hip_percent   = rep(NA_real_, nrow(data)),
      frax_sex_norm      = rep(NA_character_, nrow(data)),
      frax_age_used      = rep(NA_real_, nrow(data)),
      frax_bmd_tscore    = rep(NA_real_, nrow(data))
    )
    res[keep, ] <- out
    out <- res
  }

  if (!is.null(id_col)) {
    id_vec <- if (na_action_eff == "omit") data[[id_col]][keep] else data[[id_col]]
    out[[id_col]] <- id_vec
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) {
    hm_inform(hm_result_summary(out, fn_name), level = "inform")
  }

  out
}

library(testthat)

cm <- list(
  age = "Age", sex = "Sex",
  prior_fracture = "PriorFx", parent_fracture = "ParentFx", steroids = "Steroids",
  rheumatoid = "RA", secondary_op = "SecOP", smoker = "Smoker", alcohol = "Alcohol",
  bmd = "BMD"
)

test_that("mapping validation and missing columns error", {
  skip_on_cran()
  df <- data.frame(Age = 65, Sex = "Female")
  # partial col_map is supplemented by inference (sex inferred from 'Sex' column)
  expect_no_error(frax_score(df, list(age="Age")))
  expect_error(frax_score(df, list(age="Age", sex="SexX")), class = "healthmarkers_frax_error_missing_columns")
})

test_that("verbose emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(Age = 65, Sex = "Female")
  expect_message(frax_score(df, list(age = "Age", sex = "Sex"), verbose = TRUE), "frax_score")
  expect_message(frax_score(df, list(age = "Age", sex = "Sex"), verbose = TRUE), "col_map")
  expect_message(frax_score(df, list(age = "Age", sex = "Sex"), verbose = TRUE), "results:")
})

test_that("verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df <- data.frame(Age = 65, Sex = "Female")
  msgs <- testthat::capture_messages(
    frax_score(df, list(age = "Age", sex = "Sex"), verbose = TRUE)
  )
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

test_that("NA policies: keep, omit, error, warn", {
  skip_on_cran()
  df <- data.frame(Age = c(70, NA), Sex = c("Male", "Female"))
  out_keep <- frax_score(df, list(age="Age", sex="Sex"), na_action = "keep")
  expect_equal(nrow(out_keep), 2L)
  expect_true(is.na(out_keep$frax_major_percent[2]))
  out_omit <- frax_score(df, list(age="Age", sex="Sex"), na_action = "omit")
  expect_equal(nrow(out_omit), 1L)
  expect_error(frax_score(df, list(age="Age", sex="Sex"), na_action = "error"),
               class = "healthmarkers_frax_error_missing_values")
  expect_warning(frax_score(df, list(age="Age", sex="Sex"), na_action = "warn"),
                 class = "healthmarkers_frax_warn_missing_inputs")
})

test_that("coercion warnings on numeric-like inputs", {
  skip_on_cran()
  df <- data.frame(Age = c("70", "oops"), Sex = c("Male", "Female"))
  expect_warning(frax_score(df, list(age="Age", sex="Sex")),
                 class = "healthmarkers_frax_warn_na_coercion")
})

test_that("sex normalization warning only", {
  skip_on_cran()
  df <- data.frame(Age = c(60, 70), Sex = c("X", "u"))  # ages in range
  expect_warning(
    frax_score(df, list(age = "Age", sex = "Sex")),
    class = "healthmarkers_frax_warn_unknown_sex"
  )
})

test_that("age range warning only", {
  skip_on_cran()
  df <- data.frame(Age = c(35, 95), Sex = c("male","female"))  # valid sex
  expect_warning(
    frax_score(df, list(age = "Age", sex = "Sex")),
    class = "healthmarkers_frax_warn_age_range"
  )
})

# If you want to assert both warnings together instead of splitting:
test_that("sex and age warnings together (optional)", {
  skip_on_cran()
  df <- data.frame(Age = c(35, 95), Sex = c("x","u"))
  w <- expect_warning(
    expect_warning(
      frax_score(df, list(age = "Age", sex = "Sex")),
      class = "healthmarkers_frax_warn_unknown_sex"
    ),
    class = "healthmarkers_frax_warn_age_range"
  )
})

test_that("risk increases with added risk factors; hip fraction behaved", {
  skip_on_cran()
  df <- data.frame(
    Age = rep(70, 3), Sex = rep("Female", 3),
    PriorFx = c(0,1,1), ParentFx = c(0,0,1),
    Steroids = 0, RA = 0, SecOP = 0, Smoker = 0, Alcohol = 0
  )
  out <- frax_score(df, list(age="Age", sex="Sex",
                             prior_fracture="PriorFx", parent_fracture="ParentFx"))
  expect_lt(out$frax_major_percent[1], out$frax_major_percent[2])
  expect_lt(out$frax_major_percent[2], out$frax_major_percent[3])
  # Hip is a fraction of major; at age 70 => frac = 0.4 + 0.2 = 0.6
  expect_equal(out$frax_hip_percent, round(out$frax_major_percent * 0.6, 1))
})

test_that("BMD T-score adjustment applies when in [-5, 0]", {
  skip_on_cran()
  df <- data.frame(Age = 70, Sex = "Female", BMD = -2.5)
  out_low <- frax_score(df, list(age="Age", sex="Sex", bmd="BMD"))
  df2 <- data.frame(Age = 70, Sex = "Female", BMD = -1.0)
  out_mild <- frax_score(df2, list(age="Age", sex="Sex", bmd="BMD"))
  expect_gt(out_low$frax_major_percent, out_mild$frax_major_percent)
})

test_that("capping at 95% works", {
  skip_on_cran()
  df <- data.frame(
    Age = 90, Sex = "Female",
    PriorFx = 1, ParentFx = 1, Steroids = 1, RA = 1, SecOP = 1, Smoker = 1, Alcohol = 1
  )
  out <- frax_score(df, cm)
  expect_lte(out$frax_major_percent, 95)
  expect_lte(out$frax_hip_percent, 95)
})

test_that("check_extreme removed: function passes through outlier rows", {
  skip_on_cran()
  cm <- list(age = "Age", sex = "Sex", bmd_t = "BMD")
  df <- data.frame(
    Age = c(20, 85, 100, 60),
    Sex = c("male","female","male","female"),
    BMD = c(-7, 0, 3, -10)
  )
  # age-range warning still fires; no extreme-scan error
  out <- suppressWarnings(frax_score(df, cm))
  expect_equal(nrow(out), 4L)
})

test_that("padding preserved for keep/warn", {
  skip_on_cran()
  df <- data.frame(Age = c(70, NA), Sex = c("Female", "Female"))
  out_keep <- frax_score(df, list(age="Age", sex="Sex"), na_action = "keep")
  out_warn <- suppressWarnings(frax_score(df, list(age="Age", sex="Sex"), na_action = "warn"))
  expect_equal(nrow(out_keep), 2L)
  expect_equal(nrow(out_warn), 2L)
})

test_that("country argument is accepted and does not error", {
  skip_on_cran()
  df <- data.frame(Age = 70, Sex = "Female")
  out <- frax_score(df, list(age="Age", sex="Sex"), country = "UK")
  expect_equal(nrow(out), 1L)
})

#' Infer column names from user data based on flexible patterns, with logging

#'

#' Given a data.frame and a named mapping spec (e.g., list(G0 = NULL, I0 = NULL)),

#' infer the source column names for each key using a set of regex patterns.

#' You can supply your own patterns and "preferred" names to deterministically

#' resolve ambiguous matches. A structured log is kept and can be written to disk.

#'

#' This helper produces a col_map you can pass to HealthMarkers functions

#' (e.g., fasting_is(), lipid-derived indices).

#'

#' Backward compatibility:

#' - By default, strict = TRUE and strategy = "error" keep prior behavior:

#'   - Error if no match found.

#'   - Error if multiple candidates found.

#' - You can opt into smarter resolution via strategy = "prefer" or "first".

#'

#' @param data A data.frame or tibble whose column names are scanned.

#' @param map Named list where names are target keys (e.g., "G0","I0","TG") and

#'   values are NULL (to infer) or a user-supplied column name (to keep as-is).

#' @param verbose Logical; if TRUE, messages are printed for each mapping decision.

#'   Default TRUE.

#' @param log_file Optional file path; if supplied, a human-readable mapping log is written there.

#' @param patterns Optional named character vector of regex patterns keyed by the

#'   same names as `map`. If NULL, a built-in dictionary is used.

#' @param prefer Optional named list of character vectors with preferred column

#'   names for each key, used to resolve multiple matches deterministically.

#'   Matching is case-insensitive and exact against the provided names.

#' @param strategy One of c("error","prefer","first","stable") controlling resolution when

#'   there are multiple candidates. Default "error" (backward compatible).

#'   - "prefer": use `prefer` names first; else fall back to "stable" tie-break.

#'   - "first": take the first match in data's column order.

#'   - "stable": choose shortest name, then alphabetical.

#' @param strict Logical; if TRUE (default), missing matches error. If FALSE, missing

#'   matches leave \code{map[[key]]} as NULL and issue a warning.

#' @param ignore_case Logical; pass to grep(ignore.case = ...). Default TRUE.

#' @param fuzzy Logical; if TRUE and no regex matches are found, attempt a fuzzy match

#'   with agrep using `max_distance`. Default FALSE.

#' @param max_distance Numeric in \eqn{[0,1]} passed to agrep when fuzzy = TRUE. Default 0.1.

#' @param return One of c("map","list"). "map" (default) invisibly returns the

#'   filled mapping list. "list" returns a list(map = ..., log = tibble) for auditing.

#'

#' @return By default, invisibly returns the filled `map`. If return = "list",

#'   returns a list(map = \code{named list}, log = \code{tibble}).

#' @export

#'

#' @examples

#' df <- tibble::tibble(

#'   fasting_glucose = c(5.5, 6.1),

#'   fasting_insulin = c(60, 88),

#'   TG = c(120, 150),

#'   `HDL-c` = c(50, 45),

#'   age = c(55, 60)

#' )

#' spec <- list(G0 = NULL, I0 = NULL, TG = NULL, HDL_c = NULL)

#' # Backward-compatible: strict and "error" strategy

#' res1 <- infer_cols(df, spec, verbose = FALSE)

#' # Prefer/resolve ties deterministically

#' res2 <- infer_cols(df, spec, strategy = "prefer", verbose = TRUE)

#' # Get structured log

#' res3 <- infer_cols(df, spec, return = "list")

infer_cols <- function(data,

                       map,

                       verbose = TRUE,

                       log_file = NULL,

                       patterns = NULL,

                       prefer = NULL,

                       strategy = c("error", "prefer", "first", "stable"),

                       strict = TRUE,

                       ignore_case = TRUE,

                       fuzzy = FALSE,

                       max_distance = 0.1,

                       return = c("map", "list")) {

  strategy <- match.arg(strategy)

  return <- match.arg(return)



  # ---- validations ----

  if (!is.data.frame(data)) stop("HealthMarkers::infer_cols: `data` must be a data.frame or tibble.")

  if (!is.list(map) || is.null(names(map)) || any(names(map) == "")) {

    stop("HealthMarkers::infer_cols: `map` must be a named list (e.g., list(G0=NULL, I0=NULL)).")

  }

  keys <- names(map)

  cn <- names(data)



  # Built-in regex dictionary (separator-tolerant: space/_/-/.)

  if (is.null(patterns)) {

    sep <- "[-_. ]"             # flexible separators

    opt <- function(x) paste0("(?:", x, ")?")  # optional group (non-capturing)

    # Tokens for common fields

    patterns <- c(

      # ---- Insulin / OGTT / Fasting ----

      G0   = paste0("^(?:G", opt(sep), "0(?:\\b|$)|glu(?:cose)?", opt(sep), "0(?:\\b|$)|",

                    "fast(?:ing)?", opt(sep), opt("plasma"), opt(sep), "glu(?:cose)?(?:\\b|$)|",

                    "FPG(?:\\b|$))"),

      I0   = paste0("^(?:I", opt(sep), "0(?:\\b|$)|ins(?:ulin)?", opt(sep), "0(?:\\b|$)|",

                    "fast(?:ing)?", opt(sep), "ins(?:ulin)?(?:\\b|$))"),

      G30  = paste0("^(?:G", opt(sep), "30(?:\\b|$)|glu(?:cose)?", opt(sep), "30(?:\\b|$))"),

      I30  = paste0("^(?:I", opt(sep), "30(?:\\b|$)|ins(?:ulin)?", opt(sep), "30(?:\\b|$))"),

      G120 = paste0("^(?:G", opt(sep), "120(?:\\b|$)|glu(?:cose)?", opt(sep), "120(?:\\b|$))"),

      I120 = paste0("^(?:I", opt(sep), "120(?:\\b|$)|ins(?:ulin)?", opt(sep), "120(?:\\b|$))"),



      # ---- Anthropometry ----

      height    = paste0("^(?:height|Height|height_m|body_height|stature|ht_cm)(?:\\b|$)"),

      weight    = paste0("^(?:weight|body", sep, "?weight)(?:\\b|$)"),

      bmi       = paste0("^(?:BMI|body", sep, "?mass", sep, "?index)(?:\\b|$)"),

      BMI       = paste0("^(?:BMI|body", sep, "?mass", sep, "?index)(?:\\b|$)"),

      waist     = paste0("^(?:WC|waist|waist", sep, "?circumference)(?:\\b|$)"),

      WC        = paste0("^(?:WC|waist|waist", sep, "?circumference)(?:\\b|$)"),

      age       = "^(?:age|years?)(?:\\b|$)",

      sex       = "^(?:sex|gender)(?:\\b|$)",

      SBP       = paste0("^(?:SBP|sys(?:tolic)?", sep, "?(?:blood", sep, "?)?pres(?:sure)?)(?:\\b|$)"),

      DBP       = paste0("^(?:DBP|dia(?:stolic)?", sep, "?(?:blood", sep, "?)?pres(?:sure)?)(?:\\b|$)"),

      FFA       = paste0("^(?:FFA|NEFA|free", sep, "?fatty", sep, "?acids?)(?:\\b|$)"),

      fat_mass  = paste0("^(?:fat", sep, "?mass|FM)(?:\\b|$)"),

      lean_mass = paste0("^(?:lean", sep, "?mass|LM|lean", sep, "?body", sep, "?mass)(?:\\b|$)"),

      ALM       = paste0("^(?:ALM|alm|ALM_kg|appendicular", sep, "?lean", sep, "?mass)(?:\\b|$)"),



      # ---- Lipids ----

      TG     = "^(?:TG|tri(?:acyl)?glyceri?des?)(?:\\b|$)",

      HDL_c  = paste0("^(?:HDL(?:", sep, "?c)?|HDL", sep, "?chol(?:esterol)?|",

                      "high", sep, "?density", sep, "?lipoprotein)(?:\\b|$)"),

      LDL_c  = paste0("^(?:LDL(?:", sep, "?c)?|LDL", sep, "?chol(?:esterol)?|",

                      "low", sep, "?density", sep, "?lipoprotein)(?:\\b|$)"),

      TC     = paste0("^(?:TC|total", sep, "?chol(?:esterol)?)(?:\\b|$)"),

      ApoB   = paste0("^(?:Apo", sep, "?B|apoB|apolipoprotein", sep, "?B|ApoB100)(?:\\b|$)"),

      ApoA1  = paste0("^(?:Apo", sep, "?A1|apoA1|apolipoprotein", sep, "?A1)(?:\\b|$)"),



      # ---- Cardiometabolic aliases ----

      chol_total    = paste0("^(?:chol", sep, "?total|total", sep, "?chol(?:esterol)?)(?:\\b|$)"),

      chol_ldl      = paste0("^(?:chol", sep, "?ldl|ldl)(?:\\b|$)"),

      chol_hdl      = paste0("^(?:chol", sep, "?hdl|hdl)(?:\\b|$)"),

      triglycerides = "^(?:TG|tri(?:acyl)?glyceri?des?)(?:\\b|$)",



      # ---- Liver ----

      AST        = paste0("^(?:AST|aspartate", sep, "?aminotransferase)(?:\\b|$)"),

      ALT        = paste0("^(?:ALT|alanine", sep, "?aminotransferase)(?:\\b|$)"),

      GGT        = paste0("^(?:GGT|gamma", sep, "?glutamyl", sep, "?transferase)(?:\\b|$)"),

      platelets  = paste0("^(?:platelets?|platelet", sep, "?count)(?:\\b|$)"),

      albumin    = paste0("^(?:albumin|serum", sep, "?albumin)(?:\\b|$)"),

      bilirubin  = "^(?:bilirubin|bili)(?:\\b|$)",

      creatinine = "^(?:creatinine|creat)(?:\\b|$)",



      # ---- Electrolytes & minerals ----

      calcium   = paste0("^(?:calcium|ca|Ca|ca_mgdl)(?:\\b|$)"),

      phosphate = paste0("^(?:phosphate|phos)(?:\\b|$)"),

      magnesium = paste0("^(?:magnesium|Mg|mg)(?:\\b|$)"),



      # ---- Renal / Urine ----

      eGFR             = paste0("^(?:eGFR|estimated", sep, "?GFR|CKD", sep, "?EPI)(?:\\b|$)"),

      UACR             = paste0("^(?:UACR|ACR|urine", sep, "?albumin", sep, "?creatinine", sep, "?ratio)(?:\\b|$)"),

      urine_albumin    = paste0("^(?:urine", sep, "?albumin)(?:\\b|$)"),

      urine_creatinine = paste0("^(?:urine", sep, "?creatinine)(?:\\b|$)"),

      plasma_Na        = paste0("^(?:plasma", sep, "?Na|serum", sep, "?Na)(?:\\b|$)"),

      urine_Na         = paste0("^(?:urine", sep, "?Na)(?:\\b|$)"),

      serum_creatinine = paste0("^(?:serum", sep, "?creatinine)(?:\\b|$)"),



      # ---- Sweat ----

      sweat_chloride   = paste0("^(?:sweat", sep, "?chloride)(?:\\b|$)"),

      sweat_Na         = paste0("^(?:sweat", sep, "?Na)(?:\\b|$)"),

      sweat_K          = paste0("^(?:sweat", sep, "?K)(?:\\b|$)"),

      sweat_lactate    = paste0("^(?:sweat", sep, "?lactate)(?:\\b|$)"),

      weight_before    = paste0("^(?:weight", sep, "?before)(?:\\b|$)"),

      weight_after     = paste0("^(?:weight", sep, "?after)(?:\\b|$)"),

      duration         = paste0("^(?:duration|time", sep, "?h)(?:\\b|$)"),

      body_surface_area = paste0("^(?:body", sep, "?surface", sep, "?area|BSA)(?:\\b|$)"),



      # ---- Tracer / metabolic flux ----

      rate_glycerol  = paste0("^(?:rate", sep, "?glycerol|glycerol", sep, "?fm)(?:\\b|$)"),

      rate_palmitate = paste0("^(?:rate", sep, "?palmitate|palmitate", sep, "?fm)(?:\\b|$)"),



      # ---- Tryptophan-kynurenine pathway ----

      tryptophan  = paste0("^(?:tryptophan|Trp", sep, "?uM|tryptophan", sep, "?umolL)(?:\\b|$)"),

      kynurenine  = paste0("^(?:kynurenine|Kyn", sep, "?nM|kynurenine", sep, "?nmolL)(?:\\b|$)"),



      # ---- Saliva ----

      saliva_cort1   = paste0("^(?:saliva", sep, "?cort1|cortisol", sep, "?wake)(?:\\b|$)"),

      saliva_cort2   = paste0("^(?:saliva", sep, "?cort2|cortisol", sep, "?30)(?:\\b|$)"),

      saliva_cort3   = paste0("^(?:saliva", sep, "?cort3|cortisol", sep, "?60)(?:\\b|$)"),

      saliva_amylase = paste0("^(?:saliva", sep, "?amylase)(?:\\b|$)"),

      saliva_glucose = paste0("^(?:saliva", sep, "?glucose)(?:\\b|$)")

    )

  }



  # default preferences (exact names to pick first if multiple matches)

  if (is.null(prefer)) {

    prefer <- list(

      G0 = c("G0", "glucose_0", "glucose0", "fasting_glucose", "fpg"),

      I0 = c("I0", "insulin_0", "insulin0", "fasting_insulin"),

      TG = c("TG", "triglycerides"),

      HDL_c = c("HDL_c", "HDL-c", "HDLc", "HDL", "hdl_chol"),

      LDL_c = c("LDL_c", "LDL-c", "LDLc", "LDL", "ldl_chol"),

      TC = c("TC", "total_cholesterol"),

      ApoB = c("ApoB", "apoB", "apolipoprotein_B", "ApoB100"),

      ApoA1 = c("ApoA1", "apoA1", "apolipoprotein_A1")

    )

  }



  logs <- list()

  add_log <- function(key, selected, candidates, reason) {

    logs[[length(logs) + 1L]] <<- list(

      key = key,

      selected = if (is.null(selected)) NA_character_ else selected,

      candidates = paste(candidates, collapse = ", "),

      reason = reason

    )

    if (isTRUE(verbose)) {

      msg <- sprintf("HealthMarkers::infer_cols - %s -> %s (%s)",

                     key, ifelse(is.null(selected), "<none>", selected), reason)

      message(msg)

    }

  }



  `%||%` <- function(a, b) if (is.null(a)) b else a



  pick_preferred <- function(cands, prefs) {

    if (length(cands) == 0L || length(prefs) == 0L) return(NULL)

    lc <- tolower(cands); lp <- tolower(prefs)

    for (p in lp) {

      hit <- which(lc == p)[1]

      if (length(hit) == 1L && !is.na(hit)) return(cands[hit])

    }

    NULL

  }



  pick_stable <- function(cands) {

    if (length(cands) <= 1L) return(cands)

    lens <- nchar(cands)

    ord <- order(lens, cands)

    cands[ord][1L]

  }



  # iterate over keys

  for (nm in keys) {

    # keep user-specified mapping

    if (!is.null(map[[nm]])) {

      if (map[[nm]] %in% cn) {

        add_log(nm, map[[nm]], map[[nm]], "user-supplied")

        next

      } else {

        msg <- sprintf("HealthMarkers::infer_cols: user-supplied column '%s' for key '%s' not found in data.", map[[nm]], nm)

        if (isTRUE(strict)) stop(msg) else { warning(msg, call. = FALSE); map[[nm]] <- NULL }

      }

    }



    pat <- if (nm %in% names(patterns)) patterns[[nm]] else NULL

    if (is.null(pat) || is.na(pat)) {

      msg <- sprintf("HealthMarkers::infer_cols: no pattern defined for '%s'.", nm)

      if (isTRUE(strict)) stop(msg) else { warning(msg, call. = FALSE); add_log(nm, NULL, character(0), "no pattern"); next }

    }



    hits <- grep(pat, cn, ignore.case = isTRUE(ignore_case), value = TRUE)

    # Fuzzy fallback if enabled and no hits

    if (length(hits) == 0L && isTRUE(fuzzy)) {

      idx <- tryCatch(agrep(nm, cn, max.distance = max_distance, ignore.case = isTRUE(ignore_case)), error = function(e) integer(0))

      if (length(idx)) {

        hits <- cn[idx]

      }

    }



    if (length(hits) == 0L) {

      msg <- sprintf("HealthMarkers::infer_cols: no match for '%s'.", nm)

      if (isTRUE(strict)) stop(msg) else { warning(msg, call. = FALSE); add_log(nm, NULL, hits, "no match"); next }

    }



    if (length(hits) == 1L) {

      map[[nm]] <- hits

      add_log(nm, hits, hits, "unique match")

      next

    }



    # resolve multiple candidates

    sel <- NULL

    reason <- NULL

    if (strategy == "prefer" && !is.null(prefer[[nm]])) {

      sel <- pick_preferred(hits, prefer[[nm]])

      if (!is.null(sel)) reason <- "preferred match"

    }

    if (is.null(sel)) {

      if (strategy == "first") {

        sel <- hits[1L]

        reason <- "first in data order"

      } else if (strategy == "stable" || (strategy == "prefer" && is.null(reason))) {

        sel <- pick_stable(hits)

        reason <- "stable tie-break (shortest, then alphabetical)"

      } else if (strategy == "error") {

        stop(sprintf("HealthMarkers::infer_cols: multiple candidates for '%s': %s",

                     nm, paste(hits, collapse = ", ")))

      }

    }



    map[[nm]] <- sel

    add_log(nm, sel, hits, reason %||% "resolved")

  }



  # write log if requested

  if (!is.null(log_file)) {

    lines <- vapply(logs, function(x) {

      sprintf("%s -> %s | candidates: [%s] | %s", x$key, x$selected, x$candidates, x$reason)

    }, character(1))

    writeLines(lines, con = log_file)

    if (isTRUE(verbose)) message("HealthMarkers::infer_cols - wrote inference log to ", log_file)

  }



  if (return == "list") {

    log_tbl <- if (length(logs)) {

      tibble::tibble(

        key = vapply(logs, `[[`, character(1), "key"),

        selected = vapply(logs, `[[`, character(1), "selected"),

        candidates = vapply(logs, `[[`, character(1), "candidates"),

        reason = vapply(logs, `[[`, character(1), "reason")

      )

    } else {

      tibble::tibble(key = character(0), selected = character(0), candidates = character(0), reason = character(0))

    }

    return(list(map = map, log = log_tbl))

  }



  invisible(map)

}



#' Simplified column inference for HealthMarkers aggregators

#'

#' Exact-name matching helper used by `all_health_markers()` and related wrappers.

#' It picks the first matching candidate for each key, logs decisions via

#' `hm_inform()` when `verbose = TRUE`, and errors if required keys cannot be

#' resolved.

#'

#' @param data Data frame whose column names are scanned.

#' @param patterns Named list of character vectors, each giving candidate column

#'   names for a key (first match wins).

#' @param required_keys Character vector of keys that must resolve; otherwise an

#'   error is raised.

#' @param verbose Logical; if TRUE, emits hm_inform() messages for matches and

#'   unresolved keys.

#' @return Named list mapping keys to column names; unresolved non-required keys

#'   become `NA_character_`.

#' @rdname infer_cols

#' @keywords internal

hm_infer_cols <- function(data, patterns, required_keys = names(patterns), verbose = FALSE) {

  if (!is.data.frame(data)) {

    rlang::abort("hm_infer_cols(): `data` must be a data.frame or tibble.",

                 class = "healthmarkers_infer_error_data_type")

  }

  if (!is.list(patterns) || is.null(names(patterns)) || any(names(patterns) == "")) {

    rlang::abort("hm_infer_cols(): `patterns` must be a named list of candidate names per key.",

                 class = "healthmarkers_infer_error_patterns_type")

  }



  hm_inform(level = if (isTRUE(verbose)) "inform" else "debug", msg = "hm_infer_cols(): inferring column map")



  resolved <- list()

  for (key in names(patterns)) {

    cands <- unique(na.omit(as.character(patterns[[key]])))

    hit <- intersect(cands, names(data))

    if (length(hit) >= 1L) {

      resolved[[key]] <- hit[[1L]]

      if (isTRUE(verbose)) hm_inform(level = "debug", msg = sprintf("hm_infer_cols(): key '%s' -> '%s'", key, hit[[1L]]))

    } else {

      resolved[[key]] <- NA_character_

      hm_inform(level = "debug", msg = sprintf("hm_infer_cols(): key '%s' unresolved", key))

    }

  }



  # Ensure required keys are found

  missing_req <- required_keys[is.na(rlang::`%||%`(unlist(resolved[required_keys]), NA_character_))]

  if (length(missing_req)) {

    rlang::abort(

      sprintf("hm_infer_cols(): could not infer columns for required keys: %s",

              paste(missing_req, collapse = ", ")),

      class = "healthmarkers_infer_error_missing_required"

    )

  }



  resolved

}



#' Internal: default exact-name patterns for hm_infer_cols()

#' Not exported; used by all_health_markers() when col_map is missing.

#' Each vector lists every column-name spelling observed in the package's

#' simulated data, the Inter99/ADDITION real phenotype dataset, and common

#' biobank / cohort study conventions.  The first element is the canonical

#' internal key name.

#' @keywords internal

.hm_default_col_patterns_exact <- function() {

  list(



  ## =========================================================

  ## Demographics & basic

  ## =========================================================

  age = c("age","Age","AGE","age_year","age_years","age_0","baseline_age",

          "participant_age","age_at_visit","age_at_recruitment","age_enrol",

          ## UKB

          "age_when_attended_assessment_centre_0_0",

          "age_when_attended_assessment_centre_1_0",

          "age_at_recruitment_0_0",

          ## NHANES

          "RIDAGEYR","ridageyr",

          ## Danish (Statistics Denmark, SUSY, GESUS, Inter99)

          "alder","ALDER","alderen","pt_alder","f_alder",

          ## Norwegian (HUNT, Tromso) -- same word as Danish

          ## Swedish (SCAPIS, TwinGene) -- same word as Danish

          ## Finnish (FinnGen, THL)

          "ika","ik\u00e4","syntymavuosi",

          ## Estonian (EstBB)

          "vanus",

          ## Dutch (LifeLines, Rotterdam Study)

          "leeftijd",

          ## LOINC (All of Us / OMOP flat exports)

          "LOINC_30525_0","loinc_30525_0"),

  sex = c("sex","Sex","SEX","gender","Gender","gender_qrisk",

          "biological_sex","sex_at_birth","male","Male","female","Female",

          "woman","women",

          ## UKB

          "sex_0_0","sex_f31_0_0","genetic_sex_0_0","genetic_sex_f22001_0_0",

          ## NHANES

          "RIAGENDR","riagendr",

          ## Danish (CPR register: koen 1=male 2=female)

          "koen","KOEN","kon","KON","koensvar","pt_koen",

          ## Norwegian (HUNT, Tromso)

          "kjonn","kj\u00f8nn","Kjonn",

          ## Swedish (SCAPIS, TwinGene)

          ## "kon" already above; also:

          "k\u00f6n",

          ## Finnish (FinnGen)

          "sukupuoli","sp",

          ## Estonian (EstBB)

          "sugu",

          ## Dutch (LifeLines, Rotterdam)

          "geslacht",

          ## Generation Scotland (GS)

          "genetic_sex"),

  ethnicity = c("ethnicity","Ethnicity","ethiniciy","race","Race",

                "ethnicity_qrisk","ASIAN","asian","BLACK","black",

                "HISPANIC","hispanic","WHITE","white","Caucasian","caucasian",

                "NHW","NHW_M","NHB","HISP","latino",

                ## UKB

                "ethnic_background_0_0","ethnic_background_1_0",

                "ethnic_background_f21000_0_0",

                ## NHANES

                "RIDRETH1","ridreth1","RIDRETH3","ridreth3",

                ## Generation Scotland

                "ethnic_group",

                ## Dutch (LifeLines)

                "etniciteit"),



  ## =========================================================

  ## Anthropometry & obesity

  ## =========================================================

  height = c("height","Height","HEIGHT","height_m","height_5",

             "body_height","stature","standing_height","ht_cm","cm",

             ## UKB

             "standing_height_0_0","standing_height_1_0",

             ## NHANES

             "BMXHT","bmxht",

             ## Danish (SUSY, DANHES, Inter99, GESUS)

             "hoejde","HOEJDE","hojde","hoejde_cm","kropshoejde",

             ## Norwegian (HUNT, Tromso) -- hoyde / hoyde

             "hoyde","hoyden","hoyde_cm",

             ## Swedish (SCAPIS) -- langd / langd

             "langd","langd_cm",

             ## Finnish (FinnGen)

             "pituus","pituus_cm",

             ## Estonian (EstBB)

             "pikkus",

             ## Dutch (LifeLines, Rotterdam)

             "lengte","lengte_cm",

             ## LOINC

             "LOINC_8302_2","loinc_8302_2"),

  weight = c("weight","Weight","WEIGHT","weight_kg","weight_0","weight_1",

             "weight_3","weight_5","weight_before","weight_after","wt","wt_kg","kg",

             ## UKB

             "weight_0_0","weight_1_0",

             ## NHANES

             "BMXWT","bmxwt",

             ## Danish

             "vaegt","VAEGT","kropsv\u00e6gt","kropsvaegt","vaegt_kg",

             ## Norwegian (HUNT, Tromso)

             "vekt","vekt_kg",

             ## Swedish (SCAPIS)

             "vikt","vikt_kg",

             ## Finnish (FinnGen)

             "paino","kehonpaino","paino_kg",

             ## Estonian (EstBB)

             "kaal","kaal_kg",

             ## Dutch (LifeLines, Rotterdam)

             "gewicht","gewicht_kg",

             ## LOINC

             "LOINC_29463_7","loinc_29463_7"),

  BMI    = c("BMI","bmi","bmi_0","bmi_1","bmi_3","bmi_5",

             "BMI_value","bmi_calc","BMIkgm2","bmi_kg_m2","body_mass_index",

             "BMI3025",

             ## UKB (also caught by Layer 3 via "body_mass_index")

             "body_mass_index_bmi_0_0","body_mass_index_bmi_1_0",

             ## NHANES

             "BMXBMI","bmxbmi",

             ## Finnish (FinnGen)

             "painoindeksi",

             ## Dutch (LifeLines)

             "lichaamsgewichtsindex",

             ## LOINC

             "LOINC_39156_5","loinc_39156_5"),

  waist  = c("waist","Waist","waist_cm","waist_0","waist_1","waist_3","waist_5",

             "waist_circ","waist_circumference","waist_measure","WC","wc",

             ## UKB

             "waist_circumference_0_0","waist_circumference_1_0",

             ## NHANES

             "BMXWAIST","bmxwaist",

             ## Danish

             "talje","TALJE","taljemaal","taljeomkreds","talje_cm",

             "taille","midje",

             ## Norwegian (HUNT, Tromso)

             "midjeomkrets","midjemaal","livvidde",

             ## Swedish (SCAPIS)

             "midjeomkrets","midjematt","midja",

             ## Finnish (FinnGen)

             "vyotaronymp\u00e4rys","vyotaronymparys","vyotaro",

             ## Estonian (EstBB)

             "vootkond",

             ## Dutch (LifeLines, Rotterdam)

             "tailleomtrek","buikomvang","middel_omtrek"),

  hip    = c("hip","Hip","hip_cm","hip_0","hip_1","hip_3","hip_5",

             "hip_circ","hip_circumference",

             ## UKB

             "hip_circumference_0_0","hip_circumference_1_0",

             ## NHANES

             "BMXHIP","bmxhip",

             ## Danish

             "hofte","HOFTE","hofteomkreds","hofte_cm",

             ## Norwegian (HUNT, Tromso)

             "hofteomkrets","hoftemaal",

             ## Swedish (SCAPIS)

             "hoftomkrets","hoftomkrets_cm",

             ## Finnish (FinnGen)

             "lantionymp\u00e4rys","lantionymp\u00e4rys_cm",

             ## Dutch (LifeLines, Rotterdam)

             "heupomtrek","heup_omtrek"),

  whr    = c("whr","WHR","whratio","whratio_5","waist_hip_ratio",

             "waist_to_BMI_ratio","waist_to_height_ratio","WHRadjBMI"),

  ABSI   = c("ABSI","absi"),

  BAI    = c("BAI","bai"),

  BRI    = c("BRI","bri"),

  RFM    = c("RFM","rfm"),

  WHRadjBMI   = c("WHRadjBMI"),

  obesity_metrics = c("obesity_metrics","metabolic_risk_features"),



  ## =========================================================

  ## Body composition / DXA

  ## =========================================================

  fat_mass = c("fat_mass","fatmass","FM","fm","fm_kg","fm_wt","fat_kg","body_fat_mass",

               "fatpct","fat_percent","fat_percentage","fatfreemass",

               "fatfreemassindex","fatmassindex","body_fat_perc"),

  lean_mass = c("lean_mass","leanmass","LM","lm","lm_kg","fat_free_mass",

                "fatfreemass","FFM","ffm"),

  ALM   = c("ALM","alm","ALM_kg","appendicular_lean_mass"),

  VAT   = c("VAT","vat","visceral_fat","visceral_adipose_tissue"),

  SAT   = c("SAT","SAT_VAT_ratio","subcutaneous_adipose_tissue",

            "subcutaneous_fat","SAT_kg","SAT_cm2"),

  BMD   = c("BMD","bmd","bmd_t","bone_mineral_density"),

  BMD_ref_mean = c("BMD_ref_mean"),

  BMD_ref_sd   = c("BMD_ref_sd"),

  liver_fat    = c("liver_fat","liver_fat_pct"),



  ## =========================================================

  ## Glycemic markers (glucose / HbA1c)

  ## =========================================================

  fasting_glucose = c(

    "G0","pglu0","pglu0_0","glu0","fasting_glucose","glucose_fasting",

    "glucose0","glucose_0","fpg","FPG","fasting_plasma_glucose",

    "fasting_bg","fbg","FBG","fast_glu","bg0","BG0",

    "plasma_glucose","serum_glucose","glucose_f","GLUC","glu","Glu",

    "GLUCOSE","glucose","nglu0","p_glucose0",

    ## UKB

    "glucose_0_0","glucose_1_0",

    ## NHANES

    "LBXGLU","lbxglu","LBXSGL","lbxsgl","LBDGLUSI","lbdglusi",

    ## Danish (NPU01370; glukose)

    "NPU01370","NPU21572",

    "glukose","p_glukose","s_glukose","fastende_glukose","faste_glukose",

    "blodsukker","fasteblodsukker","fP_glukose",

    ## Norwegian (HUNT, Tromso) -- same as Danish

    ## Swedish (SCAPIS) -- note: glukos (no final e)

    "glukos","p_glukos","s_glukos","fasteglukos",

    ## Finnish (FinnGen)

    "glukoosi","p_glukoosi","paastoglukoosi",

    ## Estonian (EstBB)

    "glukoos","p_glukoos",

    ## Dutch (LifeLines, Rotterdam)

    "glucose_nuchter","nuchtere_glucose",

    ## LOINC

    "LOINC_2345_7","loinc_2345_7","LOINC_14749_6","loinc_14749_6"),

  glucose_30m  = c("G30","pglu30","pglu30_0","glu30","glucose_30m",

                   "glucose_30","bg30","G30_0"),

  glucose_120m = c("G120","pglu120","pglu120_0","glu120","glucose_120",

                   "glucose_2h","glucose_120m","bg120","glu120m","G120_0"),

  glucose_generic = c("glucose"),

  HbA1c = c("hba1c","HbA1c","hba1c_0","hba1c_1","hba1c_3","hba1c_5",

             "A1c","a1c","HBA1C","HbA1C","glycated_hba1c","hba1c_mmol",

             "hemoglobin_a1c","HBAIC","ghb","GHB","hgba1c",

             "glycohemoglobin","HbA1c_ifcc",

             ## UKB

             "glycated_haemoglobin_hba1c_0_0","glycated_haemoglobin_hba1c_1_0",

             ## NHANES

             "LBXGH","lbxgh","LBDGH","lbdgh",

             ## Danish (NPU22089)

             "NPU22089","NPU27300",

             "hba1c_ifcc","hba1c_mmolmol","glykeret_hb","glykeret_haemoglobin",

             ## Finnish (FinnGen)

             "hemoglobiini_a1c","hba1c_hemoglobiini",

             ## Dutch (LifeLines, Rotterdam)

             "geglycosyleerd_hemoglobine",

             ## LOINC

             "LOINC_4548_4","loinc_4548_4","LOINC_59261_8","loinc_59261_8"),

  glycated_albumin = c("glycated_albumin","GlycatedAlbuminPct"),



  ## =========================================================

  ## Insulin / OGTT / C-peptide

  ## =========================================================

  fasting_insulin = c(

    "I0","insu0","insu0_0","ins0","insulin0","insulin_0","fasting_insulin",

    "insulin_fasting","ins_f","basal_insulin","FI","fi","fast_ins",

    "Insulin","INS","ins","plasma_insulin","serum_insulin","Xinsulin",

    "insulin_f","ins_fasting","insulin_bas","homair"),

  insulin_30m  = c("I30","insu30","insu30_0","ins30","insulin_30",

                   "ins_30","insulin_30m"),

  insulin_120m = c("I120","insu120","insu120_0","ins120","insulin_120",

                   "ins_2h","insulin_120m"),

  insulin_ogtt    = c("insulin_ogtt","insulin_panel","insulin_adipose",

                      "insulin_tracer_dxa","insulin"),

  c_peptide_0     = c("cp0","cpep0","C_peptide","c_peptide","cp0_0"),

  c_peptide_30    = c("cp30","cpep30"),

  c_peptide_120   = c("cp120","cpep120"),

  fast_is_index   = c("fasting_is","fasting"),

  glucose_markers = c("glycemic","glycemic_markers"),



  ## Derived IR / IS indices stored as columns

  HOMA_IR   = c("HOMA_IR","homair","z_HOMA","HOMA","homa_ir","homa","IR","IR_"),

  ISI       = c("ISI_matsuda","ISIstum0_120","isi","IS","insulin_sensitivity_index",

                "BIG_SI","BIG_AIR"),

  AISI      = c("AISI"),

  Avignon_Si0   = c("Avignon_Si0"),

  Avignon_Si120 = c("Avignon_Si120"),

  OGTT_ISI      = c("ogtt_is","OGTT","ISI_matsuda","ISIstum0_120"),



  ## =========================================================

  ## Lipids

  ## =========================================================

  total_cholesterol = c(

    "TC","tc","chol","Chol","CHOL","chol_0","chol_1","chol_3","chol_5",

    "total_cholesterol","total_chol","chol_total","chol_t","cholesterol",

    "Cholesterol","TCHOL","tot_chol","totalchol","TotalChol","total.chol",

    "s_chol","serum_cholesterol","plasma_cholesterol","TC_mmol","CHOLE",

    "tcholesterol","cholesterol_total",

    ## UKB

    "cholesterol_0_0","cholesterol_1_0",

    ## NHANES

    "LBXTC","lbxtc","LBDTCSI","lbdtcsi",

    ## Danish (NPU01567; kolesterol)

    "NPU01567",

    "kolesterol","p_kolesterol","s_kolesterol","total_kolesterol",

    "p_total_kolesterol","totalkolesterol",

    ## Finnish (FinnGen) -- note -i ending

    "kolesteroli","p_kolesteroli","s_kolesteroli","kokonaiskolesteroli",

    ## Estonian (EstBB) -- note double-o

    "kolesterool","p_kolesterool","kogukolesterool",

    ## German (NAKO, KORA, Gutenberg) -- Cholesterin (not cholesterol!)

    "Cholesterin","cholesterin","Gesamtcholesterin","gesamtcholesterin",

    ## Dutch (LifeLines, Rotterdam)

    "totaal_cholesterol","cholesterol_totaal",

    ## LOINC

    "LOINC_2093_3","loinc_2093_3"),

  HDL_c = c(

    "HDL_c","HDL","hdlc","hdlc_0","hdlc_1","hdlc_3","hdlc_5",

    "HDLc","hdl_chol","chol_hdl","hdl_c","HDL_cholesterol",

    "hdl_cholesterol","hdlchol","HDLC","s_HDL","p_hdl","HDL_mmol",

    "hdl_mmol","HDLCHOLEST","chol_hdl_c","high_density_lipoprotein",

    "total.hdl","cholesterol_HDL_ratio",

    ## UKB

    "hdl_cholesterol_0_0","hdl_cholesterol_1_0",

    ## NHANES

    "LBDHDD","lbdhdd","LBXHDD","lbxhdd","LBDHDSI","lbdhdsi",

    ## Danish (NPU01637)

    "NPU01637",

    "hdl_kolesterol","p_hdl_kolesterol","s_hdl_kolesterol",

    ## Finnish (FinnGen)

    "hdl_kolesteroli","p_hdl_kolesteroli",

    ## Estonian (EstBB)

    "hdl_kolesterool",

    ## German (NAKO, KORA)

    "HDL_Cholesterin","hdl_cholesterin",

    ## Dutch (LifeLines, Rotterdam)

    "hdl_cholesterol_nuchter",

    ## LOINC

    "LOINC_2085_9","loinc_2085_9"),

  LDL_c = c(

    "LDL_c","LDL","ldl","ldl_5","LDLc","ldl_c","ldl_chol",

    "ldlcalc","LDLcalc","LDL_PN","ldl_cholesterol","LDL_cholesterol",

    "LDLC","ldlchol","ldl_calc","s_LDL","p_ldl","LDL_mmol","ldl_mmol",

    "LDLCHOLEST","ldl_direct","LDL_direct","low_density_lipoprotein",

    ## UKB

    "ldl_direct_0_0","ldl_direct_1_0",

    ## NHANES

    "LBDLDL","lbdldl","LBXLDL","lbxldl","LBDLDLSI","lbdldlsi",

    ## Danish (NPU01568)

    "NPU01568",

    "ldl_kolesterol","p_ldl_kolesterol","s_ldl_kolesterol",

    ## Finnish (FinnGen)

    "ldl_kolesteroli","p_ldl_kolesteroli",

    ## Estonian (EstBB)

    "ldl_kolesterool",

    ## German (NAKO, KORA)

    "LDL_Cholesterin","ldl_cholesterin",

    ## LOINC

    "LOINC_13457_7","loinc_13457_7","LOINC_18262_6","loinc_18262_6"),

  TG    = c(

    "TG","tg","trig","trig_0","trig_1","trig_3","trig_5",

    "triglycerides","Triglycerides","TRIGLYCERIDES","trigs","tgs",

    "TRIG","TryG","tryg","triacylglycerol","triacylglycerols",

    "TAG","TAGs","tag","fasting_tg","fasting_triglycerides",

    "plasma_tg","serum_tg","tg_mmol","tg_mg","tg_0","TG_fasting",

    "s_TG","p_TG","lipid_tg","tg_baseline","triglycerid",

    "TGL","tgl","non_fasting_tg","TG_mgdl","uNMR_TRIG",

    ## UKB

    "triglycerides_0_0","triglycerides_1_0",

    ## NHANES

    "LBXTR","lbxtr","LBDTRSI","lbdtrsi",

    ## Danish (NPU01566; triglycerider)

    "NPU01566",

    "triglycerider","p_triglycerider","s_triglycerider",

    "p_triglycerid","triglyceridkonc",

    ## Norwegian (HUNT, Tromso) -- triglyserider (different ending)

    "triglyserider","p_triglyserider",

    ## Finnish (FinnGen)

    "triglyseridit","p_triglyseridit",

    ## Estonian (EstBB)

    "triglutseriiidid","triglutseriidid",

    ## German (NAKO, KORA) -- Triglyzeride

    "Triglyzeride","triglyzeride",

    ## Dutch (LifeLines, Rotterdam) -- triglyceriden

    "triglyceriden","p_triglyceriden",

    ## LOINC

    "LOINC_2571_8","loinc_2571_8"),

  VLDL  = c("VLDL","vldl","vldl_c","vldl_chol","vldlcalc","VLDLC",

            "vldl_cholesterol"),

  remnant_c = c("remnant_c","remCHOL","remnant_cholesterol"),

  non_HDL   = c("non_HDL_c","nonHDL","non_hdl","nonhdl","non_HDL_cholesterol",

               "nonHDL_cholesterol","non_hdl_chol"),

  apoA1 = c("apoA1","ApoA1","APOA1","apolipoprotein_A1","apo_A1","apo_a1",

             ## UKB (ukbtools uses "apolipoprotein_a" not "apolipoprotein_a1")

             "apolipoprotein_a","apolipoprotein_a_0_0","apolipoprotein_a_1_0"),

  apoB  = c("apoB","ApoB","APOB","apolipoprotein_B","ApoB100","apo_B","apo_b",

             ## UKB

             "apolipoprotein_b","apolipoprotein_b_0_0","apolipoprotein_b_1_0"),



  ## Atherogenic / lipid-derived indices

  AIP          = c("AIP","AIP_denHDL","ratio_TG_HDL","atherogenic","atherogenic_indices"),

  CRI_I        = c("CRI_I_denHDL"),

  CRI_II       = c("CRI_II_denHDL"),

  HDL_TG_ratio = c("ratio_TG_HDL"),

  TC_HDL_ratio = c("ratio_TC_HDL","cholesterol_HDL_ratio"),

  LDL_HDL_ratio = c("ratio_LDL_HDL"),



  ## =========================================================

  ## Blood pressure & heart rate

  ## =========================================================

  sbp = c("sbp","SBP","sysbp","sysbp_0","sysbp_1","sysbp_3","sysbp_5",

          "systolic","systolic_bp","bp_sys","bp_sys_z","systolic_blood_pressure",

          "blood_pressure_systolic","std_systolic_blood_pressure",

          ## UKB

          "systolic_blood_pressure_automated_reading_0_0",

          "systolic_blood_pressure_automated_reading_0_1",

          ## NHANES

          "BPXOSY1","bpxosy1",

          ## Danish (SUSY, DANHES, clinical EHR)

          "systolisk","systolisk_bt","systolisk_blodtryk",

          "sbt","SBT","sys_bt","sbt1","sbt2",

          ## Norwegian (HUNT, Tromso) -- note double k: blodtrykk

          "systolisk_blodtrykk",

          ## Swedish (SCAPIS)

          "systoliskt_blodtryck",

          ## Dutch (LifeLines)

          "systolische_bloeddruk","sbd",

          ## Generation Scotland (GS)

          "SBP_mean","sbp_mean",

          ## LOINC

          "LOINC_8480_6","loinc_8480_6"),

  dbp = c("dbp","DBP","diabp","diabp_0","diabp_1","diabp_3","diabp_5",

          "diastolic","diastolic_bp","bp_dia","bp_dia_z",

          "blood_pressure_diastolic",

          ## UKB

          "diastolic_blood_pressure_automated_reading_0_0",

          ## NHANES

          "BPXODI1","bpxodi1",

          ## Danish

          "diastolisk","diastolisk_bt","diastolisk_blodtryk",

          "dbt","DBT","dia_bt","dbt1","dbt2",

          ## Norwegian (HUNT, Tromso)

          "diastolisk_blodtrykk",

          ## Swedish (SCAPIS)

          "diastoliskt_blodtryck",

          ## Dutch (LifeLines)

          "diastolische_bloeddruk","dbd",

          ## Generation Scotland (GS)

          "DBP_mean","dbp_mean",

          ## LOINC

          "LOINC_8462_4","loinc_8462_4"),

  pulse = c("pulse","Pulse","bpm","heart_rate","hr","heart.rate",

            "AvgRRInterval","pulse_rate","pulse_0",

            ## UKB

            "pulse_rate_automated_reading_0_0",

            ## NHANES

            "BPXOPLS1","bpxopls1",

            ## Danish

            "puls","PULS","hjertefrekvens","pulsfrekvens",

            ## Swedish (SCAPIS)

            "hj\u00e4rtfrekvens","hjartfrekvens",

            ## Finnish (FinnGen)

            "syke","pulssi",

            ## Dutch (LifeLines)

            "hartslag","polsslag",

            ## LOINC

            "LOINC_8867_4","loinc_8867_4"),

  MAP   = c("MAP","map","mean_arterial_pressure","PP","pp"),



  ## =========================================================

  ## Liver function

  ## =========================================================

  ALT = c("ALT","alt","alat","ALAT","GPT","gpt","GPT",

          "alanine_aminotransferase","alanine_transaminase",

          "SGPT","sgpt","liver_alt","alt_ul","ALT_UL","alt_iu","ALT_IU",

          "alt_u","alanine_aminotransferasa",

          ## UKB

          "alanine_aminotransferase_0_0","alanine_aminotransferase_1_0",

          ## NHANES

          "LBXSATSI","lbxsatsi","LBDSATSI","lbdsatsi",

          ## Danish (NPU03429)

          "NPU03429","NPU19651",

          "p_alat","s_alat","alanin_aminotransferase",

          ## Finnish (FinnGen)

          "alaniiniaminotransferaasi",

          ## German (NAKO, KORA)

          "GPT_ALAT","Alanin_Aminotransferase",

          ## Dutch (LifeLines)

          "alanine_transaminase_alat",

          ## LOINC

          "LOINC_1742_6","loinc_1742_6"),

  AST = c("AST","ast","asat","ASAT","GOT","got","SGOT","sgot",

          "aspartate_aminotransferase","aspartate_transaminase",

          "liver_ast","ast_ul","AST_UL","ast_iu","AST_IU",

          ## UKB

          "aspartate_aminotransferase_0_0","aspartate_aminotransferase_1_0",

          ## NHANES

          "LBXSASSI","lbxsassi","LBDSASSI","lbdsassi",

          ## Danish (NPU03631)

          "NPU03631",

          "p_asat","s_asat","aspartat_aminotransferase",

          ## Finnish (FinnGen)

          "aspartaattiaminotransferaasi",

          ## German (NAKO, KORA)

          "GOT_ASAT","Aspartat_Aminotransferase",

          ## LOINC

          "LOINC_1920_8","loinc_1920_8"),

  ALP = c("ALP","alp","ap","AP","alk_phos","alkaline_phosphatase",

          "alk_phosphatase","ALP_UL","alp_ul",

          ## UKB

          "alkaline_phosphatase_0_0","alkaline_phosphatase_1_0",

          ## NHANES

          "LBXSAPSI","lbxsapsi","LBDSAPSI","lbdsapsi",

          ## Danish (NPU01408 = P-Alkalisk fosfatase)

          "NPU01408",

          "alkalisk_fosfatase","p_alkalisk_fosfatase"),

  GGT = c("GGT","ggt","gamma_gt","gamma_glutamyltransferase",

          "gammaGT","gamma_GT","GGT_UL","ggt_ul",

          ## UKB (gamma_glutamyltransferase_0_0 caught by Layer 3)

          "gamma_glutamyltransferase_0_0","gamma_glutamyltransferase_1_0",

          ## NHANES

          "LBXSGTSI","lbxsgtsi","LBDSGTSI","lbdsgtsi",

          ## Danish (NPU01817)

          "NPU01817",

          "p_ggt","gamma_glutamyl_transferase",

          ## Finnish (FinnGen)

          "gammaglutamyylitransferaasi",

          ## German (NAKO, KORA)

          "Gamma_GT","GGT_gamma_GT",

          ## LOINC

          "LOINC_2324_2","loinc_2324_2"),

  bilirubin = c("bilirubin","bili","tbili","total_bilirubin","BILI",

                "Bilirubin","bilirubin_total","tbil","direct_bilirubin",

                ## UKB (total_bilirubin_0_0 caught by Layer 3)

                "total_bilirubin_0_0","total_bilirubin_1_0",

                ## NHANES

                "LBXSTB","lbxstb","LBDSTBSI","lbdstbsi",

                ## Danish (NPU01556 = P-Bilirubin; bilirubin = same in Danish)

                "NPU01556",

                "p_bilirubin","total_bilirubin_umol"),

  albumin   = c("albumin","Albumin","ALBUMIN","Alb","alb","ALB",

                "alb_gdl","serum_albumin","alb_s","plasma_albumin",

                "alb_serum","s_albumin",

                ## UKB

                "albumin_0_0","albumin_1_0",

                ## NHANES

                "LBXSAL","lbxsal","LBDSAL","lbdsal",

                ## Danish (NPU04998)

                "NPU04998",

                "p_albumin","s_albumin_gl","albumin_gl",

                ## Finnish (FinnGen)

                "albumiini","p_albumiini",

                ## Estonian (EstBB)

                "albumiin",

                ## Dutch (LifeLines) -- albumine (with e)

                "albumine",

                ## LOINC

                "LOINC_1751_7","loinc_1751_7"),

  total_protein = c("total_protein","total_prot","tot_protein","TP","tp",

                    ## UKB

                    "total_protein_0_0","total_protein_1_0",

                    ## NHANES

                    "LBXSTP","lbxstp","LBDSTP","lbdstp"),

  sev_liver = c("sev_liver","mild_liver"),



  ## =========================================================

  ## Kidney / renal (serum)

  ## =========================================================

  creatinine = c("creatinine","Creatinine","CREATININE","crea","crea_s",

                 "Cr","cr","SCr","sCr","serum_creatinine","creatinine_s",

                 "creatinine_serum","creat","scr","s_creatinine",

                 "p_creatinine","cr_mgdl","crea_umol",

                 ## UKB

                 "creatinine_0_0","creatinine_1_0",

                 ## NHANES

                 "LBXSCR","lbxscr","LBDSCR","lbdscr","LBDSCRSI","lbdscrsi",

                 ## Danish (NPU01994; kreatinin)

                 "NPU01994","NPU18016",

                 "kreatinin","p_kreatinin","s_kreatinin","kreatinin_umol",

                 ## Finnish (FinnGen) -- note double i: kreatiniini

                 "kreatiniini","p_kreatiniini",

                 ## Estonian (EstBB) -- kreatiniin

                 "kreatiniin",

                 ## German (NAKO, KORA)

                 "Kreatinin","Serumkreatinin",

                 ## Dutch (LifeLines, Rotterdam)

                 "creatinine_serum",

                 ## LOINC

                 "LOINC_2160_0","loinc_2160_0","LOINC_38483_4","loinc_38483_4"),

  BUN        = c("BUN","bun","blood_urea_nitrogen","urea_plasma",

                 ## NHANES

                 "LBXSBU","lbxsbu","LBDSBUSI","lbdsbusi",

                 ## Danish/Norwegian (carbamid = Scandinavian term for urea)

                 "carbamid","p_carbamid",

                 ## Dutch (LifeLines) -- ureum is the Dutch/German clinical term

                 "ureum","serum_ureum",

                 ## German (NAKO, KORA)

                 "Harnstoff","harnstoff",

                 ## LOINC

                 "LOINC_3094_0","loinc_3094_0","LOINC_6299_2","loinc_6299_2"),

  urea_serum = c("urea_serum","urea_s","serum_urea","urea",

                 ## UKB

                 "urea_0_0","urea_1_0",

                 ## Danish/Norwegian

                 "NPU01927",

                 "carbamid","p_carbamid","s_carbamid","carbamid_mmol",

                 ## Norwegian (HUNT)

                 "karbamid","p_karbamid",

                 ## Dutch (LifeLines, Rotterdam) -- ureum

                 "ureum","serum_ureum","plasma_ureum",

                 ## German (NAKO, KORA)

                 "Harnstoff","harnstoff","Serumharnstoff",

                 ## LOINC

                 "LOINC_22664_7","loinc_22664_7"),

  cystatin_C = c("cystatin_C","cystatinC","cystatin_c","CysC","cysc",

                 ## UKB (cystatin_c_0_0 caught by Layer 3)

                 "cystatin_c_0_0","cystatin_c_1_0",

                 ## NHANES

                 "LBXCYSTAT","lbxcystat","LBDCYTAT","lbdcytat",

                 ## Danish (NPU18454 = P-Cystatin C)

                 "NPU18454",

                 "p_cystatin_c","cystatin_c_mgl"),

  eGFR       = c("eGFR","egfr","gfr","estimated_gfr","GFR","mdrd_gfr",

                 "ckd_epi_gfr"),

  uric_acid  = c("uric_acid","uricacid","serum_urate","urate","UA","ua",

                 ## UKB

                 "urate_0_0","urate_1_0",

                 ## NHANES

                 "LBXSUA","lbxsua","LBDSUASI","lbdsuasi",

                 ## Danish/Norwegian (NPU01937; urinsyre)

                 "NPU01937",

                 "urinsyre","p_urinsyre","s_urinsyre",

                 ## Swedish (SCAPIS) -- urinsyra (note -a not -e)

                 "urinsyra","p_urinsyra",

                 ## Finnish (FinnGen)

                 "virtsahappo","uraatti",

                 ## Estonian (EstBB)

                 "kusihape",

                 ## German (NAKO, KORA)

                 "Harns\u00e4ure","Harnsaeure","harnsaeure",

                 ## Dutch (LifeLines, Rotterdam)

                 "urinezuur","serum_urinezuur",

                 ## LOINC

                 "LOINC_3084_1","loinc_3084_1"),



  ## Derived renal

  BUN_Cr_ratio = c("BUN_Cr_ratio","BUN_Cr_Ratio"),

  CKD_stage    = c("ckd_stage","kidney_failure_risk","kidney_kfre"),



  ## =========================================================

  ## Urine markers

  ## =========================================================

  u_albumin    = c("u_albumin","ualb","urine_albumin","u_albumin_mgL",

                   "urinary_albumin","alb_urine",

                   ## UKB

                   "microalbumin_in_urine_0_0","microalbumin_in_urine_1_0",

                   ## NHANES

                   "URXUMA","urxuma","URDUMA","urduma","URXUMASI","urxumasi",

                   ## Danish (NPU17550; mikroalbumin)

                   "NPU17550",

                   "u_albumin_mgl","urin_albumin","mikroalbumin","u_mikroalbumin",

                   ## Finnish (FinnGen)

                   "virtsa_albumiini","mikroalbuminuria",

                   ## Dutch (LifeLines)

                   "albumine_urine","micro_albumine",

                   ## LOINC

                   "LOINC_1754_1","loinc_1754_1"),

  u_creatinine = c("u_creatinine","ucrea","u_crea_mgdl","urine_creatinine",

                   "urinary_creatinine","uNMR_CREA",

                   ## UKB

                   "creatinine_enzymatic_in_urine_0_0",

                   "creatinine_enzymatic_in_urine_1_0",

                   ## NHANES

                   "URXUCR","urxucr","URDCR","urdcr","URXUCRSI","urxucrsi",

                   ## Danish (NPU17997; u_kreatinin)

                   "NPU17997",

                   "u_kreatinin","urin_kreatinin","u_kreatinin_mmol",

                   ## Finnish (FinnGen)

                   "virtsa_kreatiniini",

                   ## Dutch (LifeLines)

                   "creatinine_urine",

                   ## LOINC

                   "LOINC_2161_8","loinc_2161_8"),

  UACR         = c("UACR","ualbcrea","ualb_ucrea","UA_Cr_Ratio",

                   "albumin_creatinine_ratio","alb_cre_ratio",

                   "UACR_creatinine","urinary_ACR","uACR"),

  urine_protein = c("urine_protein","u_protein","urinary_protein","uprotein"),

  urine_Na      = c("urine_Na","uNa","u_Na","urinary_Na","u_sodium",

                    "urinary_sodium","uNMR_CREA",

                    ## UKB

                    "sodium_in_urine_0_0","sodium_in_urine_1_0",

                    ## NHANES

                    "URXUNA","urxuna","URDUNA","urduna"),

  urine_K       = c("urine_K","uK","u_K","urinary_K","u_potassium",

                    "urinary_potassium",

                    ## UKB

                    "potassium_in_urine_0_0","potassium_in_urine_1_0",

                    ## NHANES

                    "URXUPHS","urxuphs"),

  urine_Ca      = c("urine_Ca","uCa","u_Ca","urinary_Ca","u_calcium",

                    "urine_calcium","urinary_calcium"),

  urine_phos    = c("urine_phos","uPhos","u_phos","urinary_phosphate",

                    "u_phosphate","urine_phosphate"),

  urine         = c("urine_markers"),



  ## NMR urine metabolomics (Inter99/ADDITION naming)

  uNMR_CRTI    = c("uNMR_CRTI"),

  uNMR_GLUC    = c("uNMR_GLUC"),

  uNMR_ALAN    = c("uNMR_ALAN"),

  uNMR_LACT    = c("uNMR_LACT"),

  uNMR_ACTA    = c("uNMR_ACTA"),

  uNMR_SUCC    = c("uNMR_SUCC"),

  uNMR_CITR    = c("uNMR_CITR"),

  uNMR_DIME    = c("uNMR_DIME"),

  uNMR_TRIM    = c("uNMR_TRIM"),

  uNMR_BETA    = c("uNMR_BETA"),

  uNMR_GLYI    = c("uNMR_GLYI"),

  uNMR_FUMA    = c("uNMR_FUMA"),

  uNMR_FORM    = c("uNMR_FORM"),

  uNMR_X1MN    = c("uNMR_X1MN"),

  uNMR_NNDI    = c("uNMR_NNDI"),

  uNMR_HIPP    = c("uNMR_HIPP"),



  ## =========================================================

  ## Renal tubular injury markers (urine +/- gCr normalized)

  ## =========================================================

  A1M      = c("a1_micro","A1M_gCr","a1_microglobulin"),

  B2M      = c("beta2_micro","B2M_gCr","beta2_microglobulin","B2M"),

  KIM1     = c("KIM1","KIM1_gCr","kim1","kidney_injury_molecule_1"),

  KIM1_gCr = c("KIM1_gCr"),

  NGAL     = c("NGAL","NGAL_gCr","ngal","neutrophil_gelatinase"),

  NGAL_gCr = c("NGAL_gCr"),

  NAG      = c("NAG","nag"),

  NAG_gCr  = c("NAG_gCr"),

  L_FABP   = c("L_FABP","l_fabp","liver_fabp"),

  L_FABP_gCr = c("L_FABP_gCr"),

  IL18     = c("IL18","il18","IL_18","interleukin_18"),

  IL18_gCr = c("IL18_gCr"),



  ## =========================================================

  ## Neurology / neurofilament

  ## =========================================================

  nfl      = c("nfl","NfL","NFL","neurofilament_light","neurofilament_light_chain",

               "nfl_pgml","NfL_pgml"),



  ## =========================================================

  ## Endocrine / GI hormones

  ## =========================================================

  glucagon = c("glucagon","Glucagon","GLUCAGON","glucagon_pgml"),

  GH       = c("GH","gh","growth_hormone","GH_ngml","somatotropin"),

  PIVKA_II = c("PIVKA_II","PIVKA2","pivka_ii","des_gamma_carboxyprothrombin","DCP"),



  ## =========================================================

  ## Electrolytes & minerals

  ## =========================================================

  sodium    = c("sodium","Na","na","natrium","Na_plasma","serum_Na","plasma_sodium",

                ## UKB

                "sodium_0_0","sodium_1_0",

                ## NHANES

                "LBXSNASI","lbxsnasi","LBDSNASI","lbdsnasi",

                ## Danish/Norwegian (NPU01960; natrium)

                "NPU01960",

                "p_natrium","s_natrium","natrium_mmol",

                ## Estonian (EstBB) -- naatrium (extra 'a')

                "naatrium","p_naatrium",

                ## German (NAKO, KORA)

                "Natrium","Serumnatrium",

                ## Dutch (LifeLines)

                "natrium_serum",

                ## LOINC

                "LOINC_2951_2","loinc_2951_2"),

  potassium = c("potassium","K","k","Kalium","K_plasma","serum_K","plasma_potassium",

                ## UKB

                "potassium_0_0","potassium_1_0",

                ## NHANES

                "LBXSKSI","lbxsksi","LBDSKSI","lbdsksi",

                ## Danish/Norwegian (NPU01961; kalium)

                "NPU01961",

                "p_kalium","s_kalium","kalium_mmol",

                ## Estonian (EstBB) -- kaalium

                "kaalium","p_kaalium",

                ## German (NAKO, KORA)

                "Kalium","Serumkalium",

                ## Dutch (LifeLines)

                "kalium_serum",

                ## LOINC

                "LOINC_2823_3","loinc_2823_3"),

  chloride  = c("chloride","Cl","cl","serum_chloride",

                ## NHANES

                "LBXSCLSI","lbxsclsi","LBDSCLSI","lbdsclsi"),

  bicarbonate = c("bicarbonate","HCO3","hco3","bicarb",

                  ## NHANES

                  "LBXSC3SI","lbxsc3si","LBDSC3SI","lbdsc3si"),

  sodium_potassium_ratio = c("Na_K_ratio","U_Na_K_ratio_denK"),

  calcium   = c("calcium","Calcium","CALCIUM","ca","Ca","ca_mgdl",

                "serum_calcium","plasma_Ca",

                ## UKB

                "calcium_0_0","calcium_1_0",

                ## NHANES

                "LBXSCA","lbxsca","LBDSCASI","lbdscasi",

                ## Danish (NPU02497)

                "NPU02497",

                "p_calcium","s_calcium","calcium_mmol",

                ## Finnish (FinnGen)

                "kalsium","p_kalsium",

                ## German (NAKO, KORA)

                "Kalzium","Serumcalcium",

                ## LOINC

                "LOINC_17861_6","loinc_17861_6"),

  phosphate = c("phosphate","Phosphate","phos","PHOS","phosphorus",

                "inorganic_phosphate","serum_phos","p_phos",

                ## UKB

                "phosphate_0_0","phosphate_1_0",

                ## NHANES

                "LBXSPH","lbxsph","LBDSPHSI","lbdsphsi",

                ## Danish (NPU02319; fosfat)

                "NPU02319",

                "fosfat","p_fosfat","s_fosfat","fosfat_mmol",

                ## German (NAKO, KORA)

                "Phosphat","Serumphosphat","anorganisches_Phosphat",

                ## LOINC

                "LOINC_2777_1","loinc_2777_1"),

  corrected_calcium = c("corrected_calcium","calcium_corrected"),

  magnesium = c("magnesium","Magnesium","Mg","mg","serum_Mg","plasma_Mg",

                ## NHANES

                "LBXSMG","lbxsmg","LBDSMGSI","lbdsmgsi",

                ## Danish (NPU03792)

                "NPU03792",

                "p_magnesium","s_magnesium","magnesium_mmol",

                ## Finnish (FinnGen)

                "magnesium_seerumissa",

                ## German (NAKO, KORA)

                "Magnesium","Serummagnesium",

                ## LOINC

                "LOINC_19123_9","loinc_19123_9"),

  zinc      = c("zinc","Zinc","Zn","zn","serum_zinc","plasma_Zn"),

  copper    = c("copper","Copper","Cu","cu","serum_copper"),

  Mg_Zn_den = c("Mg_Zn_den"),

  Cu_Zn_den = c("Cu_Zn_den"),



  ## =========================================================

  ## Inflammatory markers & hematology

  ## =========================================================

  CRP = c("CRP","crp","crp_hs","hs_crp","hsCRP","hs_CRP",

          "CRP_tethys","high_sensitivity_crp","CRP_category",

          "CRP_mgL","crp_mgL","crp_mg_L","CRP_mg_L",

          "C_reactive_protein","c_reactive_protein","hscrp",

          "CRP_hs","CRP_high_sens",

          ## UKB

          "c_reactive_protein_0_0","c_reactive_protein_1_0",

          ## NHANES

          "LBXHSCRP","lbxhscrp","LBDHRP","lbdhrp",

          "LBXCRP","lbxcrp",

          ## Danish (NPU10438)

          "NPU10438",

          "p_crp","s_crp","hs_crp_mgl","crp_mgl",

          ## Finnish (FinnGen)

          "c_reaktiivinen_proteiini",

          ## German (NAKO, KORA)

          "CRP_hsCRP","hochsensitives_CRP","hs_CRP_mgl",

          ## Dutch (LifeLines)

          "c_reactief_proteine",

          ## LOINC

          "LOINC_30522_7","loinc_30522_7","LOINC_1988_5","loinc_1988_5"),

  IL6  = c("IL6","il6","IL_6","interleukin_6","il6_pgml","IL6_pgml"),

  TNFa = c("TNFa","tnfa","TNF_alpha","tnf_alpha","TNFalpha"),

  ESR  = c("ESR","esr","erythrocyte_sedimentation_rate"),

  dNLR = c("dNLR","dnlr"),

  NLR  = c("NLR","nlr","neutrophil_lymphocyte_ratio"),

  PLR  = c("PLR","plr","platelet_lymphocyte_ratio"),

  SII  = c("SII","sii"),

  SIRI = c("SIRI","siri"),

  WBC  = c("WBC","wbc","leukocytes","white_blood_cells","leucocytes",

           ## UKB

           "white_blood_cell_leucocyte_count_0_0",

           "white_blood_cell_leucocyte_count_1_0",

           ## NHANES

           "LBXWBCSI","lbxwbcsi","LBDWBCSI","lbdwbcsi",

           ## Danish (NPU02593; leukocytter)

           "NPU02593",

           "leukocytter","hvide_blodlegemer","b_leukocytter",

           ## Norwegian (HUNT) -- same as Danish

           ## Swedish (SCAPIS) -- leukocyter (single t)

           "leukocyter","b_leukocyter",

           ## Finnish (FinnGen)

           "leukosyytit","valkosolujen_lukumaara",

           ## Estonian (EstBB)

           "leukotsyydid",

           ## German (NAKO, KORA)

           "Leukozyten","Leukozytenzahl",

           ## Dutch (LifeLines)

           "leukocyten",

           ## LOINC

           "LOINC_6690_2","loinc_6690_2"),

  neutrophils = c("neutrophils","Neutrophils","NEUT","neut","neutro",

                  ## singular (UKB: neutrophil_count_0_0)

                  "neutrophil","neutrophil_count_0_0","neutrophil_count_1_0",

                  ## NHANES

                  "LBDNENO","lbdneno","LBXNEPCT","lbxnepct",

                  ## Danish (NPU04578; neutrofile = Danish plural)

                  "NPU04578",

                  "neutrofile","neutrofil","b_neutrofile"),

  lymphocytes = c("lymphocytes","Lymphocytes","LYMPH","lymph",

                  ## singular (UKB: lymphocyte_count_0_0)

                  "lymphocyte","lymphocyte_count_0_0","lymphocyte_count_1_0",

                  ## NHANES

                  "LBDLYMNO","lbdlymno","LBXLYPCT","lbxlypct",

                  ## Danish (NPU01455; lymfocytter = Danish)

                  "NPU01455",

                  "lymfocytter","lymfocyt","b_lymfocytter"),

  eosinophils = c("eosinophils","Eosinophils","eos","EOS","EOSIN",

                  ## singular (UKB: eosinophil_count_0_0)

                  "eosinophil","eosinophil_count_0_0","eosinophil_count_1_0",

                  ## NHANES

                  "LBDEOSNNO","lbdeosnno","LBXEOPCT","lbxeopct",

                  ## Danish (eosinofile = Danish)

                  "eosinofile","eosinofil","b_eosinofile"),

  monocytes   = c("monocytes","Monocytes","MONO","mono","monocyte_count",

                  ## UKB (monocyte_count_0_0 caught by Layer 3 via "monocyte_count")

                  "monocyte_count_0_0","monocyte_count_1_0",

                  ## NHANES

                  "LBDMONO","lbdmono","LBXMOPCT","lbxmopct",

                  ## Danish (monocytter = Danish)

                  "monocytter","monocyt","b_monocytter"),

  platelets   = c("platelets","Platelets","PLT","plt","thrombocytes","platelet_count","Platelet_Count","PLATELET",

                  ## UKB

                  "platelet_count_0_0","platelet_count_1_0",

                  ## NHANES

                  "LBXPLTSI","lbxpltsi","LBDPLTSI","lbdpltsi",

                  ## Danish (NPU03568; trombocytter)

                  "NPU03568",

                  "trombocytter","trombocyt","thrombocytter","b_trombocytter",

                  ## Swedish (SCAPIS) -- trombocyter

                  "trombocyter","b_trombocyter",

                  "TPK","tpk",

                  ## Finnish (FinnGen)

                  "trombosyytit",

                  ## Estonian (EstBB)

                  "trombotsyydid",

                  ## German (NAKO, KORA)

                  "Thrombozyten","Thrombozytenzahl",

                  ## Dutch (LifeLines)

                  "trombocyten","bloedplaatjes",

                  ## LOINC

                  "LOINC_777_3","loinc_777_3"),



  Hgb         = c("Hgb","hgb","Hb","hb","hemoglobin","haemoglobin","HGB","HB",

                  ## UKB

                  "haemoglobin_concentration_0_0","haemoglobin_concentration_1_0",

                  ## NHANES

                  "LBXHGB","lbxhgb","LBDHGB","lbdhgb",

                  ## Danish (NPU03609)

                  "NPU03609",

                  "b_haemoglobin","b_hgb","hb_konc",

                  ## Finnish (FinnGen)

                  "hemoglobiini","b_hemoglobiini",

                  ## Estonian (EstBB)

                  "hemoglobiin",

                  ## German (NAKO, KORA)

                  "H\u00e4moglobin","Haemoglobin","haemoglobin_konz",

                  ## Dutch (LifeLines) -- hemoglobine

                  "hemoglobine",

                  ## LOINC

                  "LOINC_718_7","loinc_718_7"),

  inflammatory_age     = c("inflammatory_age","iAge_raw","inf_age",

                           "immune_age","inflammaging_score"),

  inflammatory_markers = c("inflammatory","inflammatory_markers"),



  ## Cytokines / proteins from multiplex panel (ADDITION)

  ADIPOQ   = c("ADIPOQ","adiponectin","Adiponectin"),

  LEP      = c("LEP","leptin","Leptin"),

  RETN     = c("RETN","resistin","Resistin"),

  IGFBP1   = c("IGFBP1","igfbp1","IGFBP_1"),

  IGFBP2   = c("IGFBP2","igfbp2","IGFBP_2"),

  IGFBP3   = c("IGFBP3","igfbp3","IGFBP_3"),

  FTH1     = c("FTH1","fth1"),

  HSPA1B   = c("HSPA1B","hspa1b","HSP70"),

  DPP4     = c("DPP4","dpp4","DPP_4","dipeptidyl_peptidase_4"),

  GH1      = c("GH1","gh1","GH","gh","growth_hormone"),

  Proinsulin = c("Proinsulin","proinsulin","pro_insulin"),

  ntproBNP   = c("ntproBNP","NT_proBNP","nt_probnp","BNP","proBNP"),



  ## =========================================================

  ## Vitamins, micronutrients & iron

  ## =========================================================

  vitaminD = c(

    "vitd","VitD","VITD","vitd25","VitD25","vitd25_immu","vitd_level",

    "25OHD","25OHD3","25(OH)D","25_OHD","25OH_D","25OH_D3",

    "vitamin_d","vitamin_D","VITAMIN_D","oh25d","serum_25OHD",

    "25OHD_nmol","vitd_nmol","calcidiol","vitamin_d_25oh",

    "VitD_ref_mean","VitD_ref_sd","vitamin_d_status",

    "vit_d","vit_d25","s_vitd","plasma_vitd",

    ## UKB

    "vitamin_d_0_0","vitamin_d_1_0",

    ## NHANES

    "LBXVD2","lbxvd2","LBDVD2","lbdvd2",

    "LBDVDMSN","lbdvdmsn",

    ## Danish (NPU10501; d_vitamin)

    "NPU10501",

    "d_vitamin","dvitamin","25_oh_d_vitamin","p_25ohd",

    "d_vit","vitd_nmolL",

    ## Swedish (SCAPIS)

    "d_vitamin",

    ## Finnish (FinnGen)

    "D_vitamiini","d_vitamiini","kolekalsiferoli",

    ## German (NAKO, KORA)

    "Vitamin_D","25_OH_Vitamin_D","Calcidiol",

    ## Dutch (LifeLines)

    "vitamine_D","25_hydroxy_vitamine_D",

    ## LOINC

    "LOINC_62292_8","loinc_62292_8","LOINC_1989_3","loinc_1989_3"),

  vitaminB12 = c("vitb12","vitb12_immu","vitamin_b12","B12","b12","b12_level",

                 "vitaminB12","cobalamin","Cobalamin","cyanocobalamin",

                 ## NHANES

                 "LBXB12","lbxb12","LBDB12","lbdb12",

                 ## Danish (NPU01454 = P-Cobalamin; kobalamin = Danish spelling)

                 "NPU01454",

                 "kobalamin","p_kobalamin","p_b12","s_b12"),

  MMA        = c("MMA","mma","methylmalonic_acid"),

  folate     = c("folate","Folate","folate_immu","folic_acid","vitb9",

                 "folate_serum","serum_folate","folicacid",

                 ## NHANES

                 "LBXFOLSI","lbxfolsi","LBDFOLSI","lbdfolsi",

                 ## Danish (NPU01894 = P-Folat; folsyre = Danish for folic acid)

                 "NPU01894",

                 "folsyre","p_folat","s_folat","p_folsyre"),

  ferritin   = c("ferritin","Ferritin","ferri","ferri_immu","ferritin_s",

                 "serum_ferritin","FerritinTS","s_ferritin",

                 ## NHANES

                 "LBXFER","lbxfer","LBDFER","lbdfer",

                 ## Danish (NPU04698)

                 "NPU04698",

                 "p_ferritin","ferritin_mgl","ferritin_ngl",

                 ## Finnish (FinnGen)

                 "ferritiini","p_ferritiini",

                 ## Estonian (EstBB)

                 "ferritiin",

                 ## German (NAKO, KORA)

                 "Ferritin","Serumferritin",

                 ## Dutch (LifeLines)

                 "ferritine",

                 ## LOINC

                 "LOINC_2276_4","loinc_2276_4"),

  iron       = c("iron","Iron","iron_s","serum_iron","s_iron","Fe","fe",

                 ## NHANES

                 "LBXIRN","lbxirn","LBDIRNSI","lbdirnsi",

                 ## Danish (NPU01581 = P-Jern; jern = Danish for iron)

                 "NPU01581",

                 "jern","p_jern","s_jern","jern_umol"),

  transferrin     = c("transferrin","Transferrin","tf","transferrin_s",

                      ## NHANES

                      "LBXTFN","lbxtfn"),

  transferrin_sat = c("TSat","transferrin_sat","Ferr_TSat_den",

                      "transferrin_saturation","iron_saturation"),

  Retinol    = c("Retinol","retinol","vitamin_A","Retinol_ref_mean",

                 "Retinol_ref_sd"),

  VitC       = c("VitC","vitc","vitamin_c","Vitamin_C","ascorbate",

                 "ascorbic_acid","vit_c"),

  Tocopherol = c("Tocopherol","tocopherol","vitamin_E","alpha_tocopherol"),

  DHA        = c("DHA","dha","docosahexaenoic_acid"),

  EPA        = c("EPA","epa","eicosapentaenoic_acid"),

  FFA        = c("FFA","ffa","free_fatty_acids","NEFA","nefa"),

  Homocysteine = c("Homocysteine","homocysteine","hcy","HCY","tHcy",

                   ## NHANES

                   "LBXHCY","lbxhcy","LBDHCYSI","lbdhcysi",

                   ## Danish (NPU19404; homocystein)

                   "NPU19404",

                   "homocystein","p_homocystein","s_homocystein",

                   ## Finnish (FinnGen)

                   "homokysteiini",

                   ## German (NAKO, KORA)

                   "Homozystein","Homocystein",

                   ## Dutch (LifeLines)

                   "homocysteine_plasma",

                   ## LOINC

                   "LOINC_13965_9","loinc_13965_9"),

  Total_lipids = c("Total_lipids","total_lipids"),



  ## =========================================================

  ## Hormones

  ## =========================================================

  PTH          = c("pth","PTH","pth_immu","parathyroid_hormone","iPTH",

                   ## NHANES

                   "LBXPTH","lbxpth","LBDPTHSI","lbdpthsi"),

  TSH          = c("tsh","TSH","tsh_immu","thyroid_stimulating_hormone",

                   "thyrotropin","TSH_mIU",

                   ## UKB

                   "thyroid_stimulating_hormone_tsh_0_0",

                   ## Danish (NPU01407)

                   "NPU01407",

                   "p_tsh","s_tsh","tsh_miu_l",

                   ## Finnish (FinnGen)

                   "tyreotropiini","TSH_arvo",

                   ## German (NAKO, KORA)

                   "TSH_basal","Thyreotropin",

                   ## Dutch (LifeLines)

                   "thyreotroop_hormoon",

                   ## LOINC

                   "LOINC_3016_3","loinc_3016_3"),

  FT4          = c("ft4","FT4","free_t4","free_T4","free_thyroxine",

                   "fT4","FT4_pmol"),

  free_T3      = c("free_T3","fT3","ft3","FT3","free_triiodothyronine",

                   "FT3_pmol"),

  testosterone = c("testosterone","Testosterone","testo",

                   "total_testosterone","TT","tt_nmol",

                   ## UKB

                   "testosterone_0_0","testosterone_1_0",

                   ## Danish (NPU27281; testosteron)

                   "NPU27281",

                   "testosteron","p_testosteron","s_testosteron",

                   ## Finnish (FinnGen)

                   "testosteroni","p_testosteroni",

                   ## German (NAKO, KORA)

                   "Testosteron","Gesamttestosteron",

                   ## Dutch (LifeLines)

                   "testosteron_totaal",

                   ## LOINC

                   "LOINC_2986_3","loinc_2986_3"),

  estradiol    = c("estradiol","Estradiol","e2","E2","estrad",

                   "oestradiol","estradiol_pmol",

                   ## UKB (oestradiol_0_0 caught by Layer 3 via "oestradiol")

                   "oestradiol_0_0","oestradiol_1_0"),

  progesterone = c("progesterone","Progesterone","PROG","prog"),

  prolactin    = c("prolactin","Prolactin","PRL","prl"),

  FSH          = c("FSH","fsh","follicle_stimulating_hormone"),

  LH           = c("LH","lh","luteinising_hormone","luteinizing_hormone"),

  DHEAS        = c("DHEAS","dheas","dhaes","dehydroepiandrosterone_s",

                   "dhea_s","DHEA_S"),

  Cortisol     = c("Cortisol","cortisol","cort1","cort2","cort3",

                   "cortisol_0","cortisol_30","cortisol_am",

                   "morning_cortisol","cortisol_nmol"),

  cortisol_0   = c("cortisol_0","cort1","cort_wake","saliva_cort1"),

  cortisol_30  = c("cortisol_30","cort2","cort_30min","saliva_cort2"),

  renin        = c("renin","Renin","plasma_renin","renin_activity","PRA"),

  aldosterone  = c("aldosterone","Aldosterone","plasma_aldosterone"),

  SHBG         = c("SHBG","shbg","sex_hormone_binding_globulin","SHBG_nmol",

                   ## UKB

                   "sex_hormone_binding_globulin_0_0",

                   "sex_hormone_binding_globulin_1_0"),

  IGF1         = c("IGF1","igf1","IGF_1","insulin_growth_factor",

                   "insulin_like_growth_factor","IGF1_ngml",

                   ## UKB

                   "insulin_like_growth_factor_igf_1_0_0",

                   "insulin_like_growth_factor_igf_1_1_0"),

  tpoab        = c("tpoab","TPOAB","TPO_Ab","anti_TPO","anti_thyroid_peroxidase"),



  ## Hormone ratios (derived)

  Cort_DHEAS_den = c("Cort_DHEAS_den"),

  T_E2_den       = c("T_E2_den"),

  TSH_fT4_den    = c("TSH_fT4_den"),



  ## =========================================================

  ## Pulmonary / spirometry

  ## =========================================================

  FEV1     = c("FEV1","fev1","FEV1_pred","fev1_pred","fev1_post","FEV1_L",

               ## UKB

               "forced_expiratory_volume_in_1_second_fev1_0_0",

               "forced_expiratory_volume_in_1_second_fev1_0_2"),

  FEV1pct  = c("FEV1pct","fev1_pct","fev1_pp","FEV1FVCratio"),

  FVC      = c("FVC","fvc","fvc_post","FVC_L",

               ## UKB

               "forced_vital_capacity_fvc_0_0",

               "forced_vital_capacity_fvc_0_2"),

  FEV1FVC  = c("FEV1FVC","FEV1FVCratio","fev1_fvc","fev1fvc",

               ## UKB

               "fev1_fvc_ratio_0_0","fev1_fvc_ratio_1_0"),

  mmrc     = c("mmrc","mMRC","MMRC","mrc_dyspnoea","mrc_score"),

  sixmwd   = c("sixmwd","six_minute_walk","6mwd","6MWD","walk_6min"),

  COPD     = c("copd","COPD","chronic_obstructive_pulmonary"),

  pulmo_markers = c("pulmo","pulmo_markers"),



  ## =========================================================

  ## Cardiovascular / ECG

  ## =========================================================

  qtcf     = c("qtcf","QTcF","qtc","QTc","QTcf_ms"),

  qt_interval = c("qt_interval","QT","qt","QT_ms"),

  qrs_duration = c("qrs_duration","QRS","qrs","QRS_ms"),

  PR_interval  = c("PR_Interval","pr_interval","PR","pr"),

  ntproBNP_cv  = c("ntproBNP","BNP","NT_proBNP"),



  ## =========================================================

  ## Bone markers

  ## =========================================================

  BSAP      = c("BSAP","bsap","bone_specific_alkaline_phosphatase"),

  CTX       = c("CTX","ctx","CTX_I","C_telopeptide"),

  PINP      = c("PINP","pinp","procollagen_type1","P1NP"),

  Osteocalcin = c("Osteocalcin","osteocalcin","OC","oc","bone_gla_protein"),

  TBS       = c("TBS","tbs","trabecular_bone_score"),

  parent_fracture = c("parent_fracture","parental_fracture",

                      "family_fracture","hip_fracture_parent"),

  prior_fracture  = c("prior_fracture","previous_fracture",

                      "fracture_history","fracture_hx"),



  ## =========================================================

  ## Frailty / sarcopenia

  ## =========================================================

  frailty_index = c("frailty_index","frailty","fi_score","rockwood"),

  sarc_f   = c("sarc_f","sarc_f_score","SarcF","sarcf","SARC_F"),

  strength = c("strength","Strength","grip_strength","handgrip",

               "handgripMax","handgrip_kg","grip"),

  chair    = c("chair","Chair","chair_stand","five_chair_stands",

               "chair_stand_time"),

  stairs   = c("stairs","Stairs","stair_climb"),

  walking  = c("walking","Walking","Walk_m","gait_speed","walk_speed",

               "usual_gait_speed"),

  falls    = c("falls","Falls","fall_history","number_of_falls"),



  ## =========================================================

  ## Lifestyle & comorbidities

  ## =========================================================

  smoking  = c("smoking","smoke","smoker","Smoking","smoke_daily_gr",

               "smoke_packyrs","smoking_heavy","current_smoker","ever_smoker"),

  alcohol  = c("alcohol","alko_unit0","alko_class0","alko_binge",

               "alcohol_units","drinks_per_week","alcohol_consumption"),

  physical_activity = c("FYSAKT0","physical_activity","PA","pa",

                        "exercise","stepKondi","stepIlt"),

  diabetes  = c("diabetes","T2D_NGT","T2D_NFG","glu_tol","diabetes1",

                "diabetes2","dm","DM","T2D","t2d"),

  hypertension = c("hypertension","HT_NT","HT_NT2","hyptreat",

                   "blood_pressure_treatment","bp_treated"),

  liptreat  = c("liptreat","lipid_treatment","statin","cholesterol_treatment"),

  instreat  = c("instreat","insulin_treatment","insulin_therapy"),

  pulsetreat = c("pulsetreat","pulse_treatment","beta_blocker"),



  ## =========================================================

  ## Sweat biomarkers

  ## =========================================================

  sweat           = c("sweat","sweat_markers"),

  sweat_chloride  = c("sweat_chloride"),

  sweat_Na        = c("sweat_Na"),

  sweat_K         = c("sweat_K"),

  sweat_lactate   = c("sweat_lactate"),

  sweat_rate_bsa       = c("sweat_rate_bsa"),

  sweat_rate_duration  = c("sweat_rate_duration"),



  ## =========================================================

  ## Saliva biomarkers

  ## =========================================================

  saliva        = c("saliva","saliva_markers"),

  saliva_amylase = c("saliva_amylase","amylase","salivary_amylase"),

  saliva_glucose = c("saliva_glucose"),

  saliva_cort1   = c("saliva_cort1","cort1"),

  saliva_cort2   = c("saliva_cort2","cort2"),

  saliva_cort3   = c("saliva_cort3","cort3"),



  ## =========================================================

  ## Tracer / metabolic flux markers

  ## =========================================================

  tracer_dxa_is  = c("tracer_dxa_is","insulin_tracer_dxa"),

  rate_glycerol  = c("rate_glycerol","glycerol_fm"),

  rate_palmitate = c("rate_palmitate","palmitate_fm"),



  ## =========================================================

  ## Tryptophan-kynurenine pathway

  ## =========================================================

  tryptophan    = c("tryptophan","Trp_uM","tryptophan_umolL","Trp","trp"),

  kynurenine    = c("kynurenine","Kyn_nM","kynurenine_nmolL","Kyn","kyn"),

  kyn_trp_ratio = c("kyn_trp","kyn_trp_ratio","KTR","ktr"),

  Tyr           = c("Tyr","tyr","tyrosine","Tyrosine"),

  Phe           = c("Phe","phe","phenylalanine","Phenylalanine"),

  Tyr_Phe_Ratio = c("Tyr_Phe_Ratio","tyr_phe","TyrPhe"),



  ## =========================================================

  ## Metabolic / cardio risk scores (derived columns)

  ## =========================================================

  ASCVD        = c("ASCVD","cvd_risk_ascvd","ascvd_10yr"),

  QRISK3       = c("QRISK3","qrisk3","QRISK3_score","q_risk3"),

  PooledCohort = c("PooledCohort","pooled_cohort"),

  CVrisk       = c("CVrisk","RiskScorescvd","score2","SCORE2"),



  ## =========================================================

  ## Allostatic load & composite indices

  ## =========================================================

  allostatic_load = c("allostatic_load","AllostaticLoad","AL_score"),



  ## =========================================================

  ## iAge / inflammatory clock

  ## =========================================================

  iAge = c("iAge","iage","inflammatory_age_clock","iAge_score")

)

}

#' Compute a simplified Inflammatory Age Index (iAge) with QA and verbose summaries
#'
#' Implements a linear proxy for immunosenescence based on key inflammatory biomarkers,
#' conceptually inspired by the inflammatory aging clock (iAge) literature.
#' This simplified iAge is computed as a weighted sum of C-reactive protein (CRP),
#' interleukin-6 (IL6), and tumor necrosis factor-alpha (TNFa).
#'
#' By default, rows with any missing required marker return NA in the index
#' (consistent with the default behavior of other package functions).
#' Optional diagnostics can warn on high missingness and implausible negative values.
#' Verbose mode prints step-by-step progress and a final summary.
#'
#' Assumed units (no automatic unit conversion):
#' - CRP: mg/L
#' - IL6: pg/mL
#' - TNFa: pg/mL
#'
#' Note:
#' - The original iAge model in Sayed et al. (Nature Aging, 2021) is a multi-marker
#'   machine learning model. This function provides a simple, linear proxy using
#'   three canonical inflammatory biomarkers. It is not identical to the original
#'   published iAge but is inspired by its rationale.
#' - This proxy is intended for exploratory feature engineering and cohort-level
#'   analyses. It must not be treated as a validated replacement for the published
#'   iAge model or used as a standalone clinical decision metric.
#'
#' @param data A data.frame or tibble containing the biomarker columns mapped by `col_map`.
#' @param col_map Named list mapping:
#'   - CRP  -> column name for C-reactive protein (mg/L)
#'   - IL6  -> column name for interleukin-6 (pg/mL)
#'   - TNFa -> column name for tumor necrosis factor-alpha (pg/mL)
#' @param weights Named numeric vector of weights for each marker (must sum to 1).
#'   Defaults to c(CRP = 0.33, IL6 = 0.33, TNFa = 0.34).
#' @param verbose Logical; if TRUE, prints column mapping and computing messages.
#' @param na_action One of c("keep","omit","error","ignore","warn") controlling how missing inputs affect iAge:
#'   - "keep": return NA for rows where any required marker is NA (default; consistent with other functions).
#'   - "omit": ignore NAs in the weighted sum (treat missing markers as 0).
#'   - "error": abort if any required marker contains NA.
#'   - "ignore": alias of "omit".
#'   - "warn": alias of "omit" but emits missingness warnings (per na_warn_prop).
#' @param na_warn_prop Proportion in \eqn{[0,1]} above which a high-missingness warning
#'   is emitted when `na_action = "warn"`. Default 0.2.
#'
#' @return A tibble with one column:
#'   - iAge (numeric): the computed inflammatory age index.
#'
#' @references
#' \insertRef{sayed2021iage}{HealthMarkers} (conceptual background; not method-identical to this implementation)
#' \insertRef{harris1999il6crp}{HealthMarkers}
#' \insertRef{bruunsgaard2003death}{HealthMarkers}
#'
#' @seealso [impute_missing()], [glycemic_markers()]
#'
#' @examples
#' library(tibble)
#' df <- tibble(
#'   CRP  = c(1.2, 3.5, NA),  # mg/L
#'   IL6  = c(2.0, 4.1, 1.5), # pg/mL
#'   TNFa = c(1.0, 1.8, 0.9)  # pg/mL
#' )
#' # Default behavior (rows with any missing marker return NA)
#' iAge(
#'   df,
#'   col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa")
#' )
#'
#' # Keep NA if any marker missing in a row
#' iAge(
#'   df,
#'   col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa"),
#'   na_action = "keep"
#' )
#'
#' # Verbose output
#' iAge(
#'   df,
#'   col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa"),
#'   verbose = TRUE
#' )
#'
#' @importFrom tibble tibble
#' @importFrom rlang abort warn inform
#' @export
iAge <- function(data,
                 col_map = NULL,
                 weights = c(CRP = 0.33, IL6 = 0.33, TNFa = 0.34),
                 verbose = TRUE,
                 na_action = c("keep", "omit", "error", "ignore", "warn"),
                 na_warn_prop = 0.2) {
  data_name <- (function(.e) if (is.symbol(.e)) as.character(.e) else "data")(substitute(data))
  fn_name <- "iAge"
  .hm_log_input(data, data_name, fn_name, verbose)
  id_col <- .hm_detect_id_col(data)
  na_action_raw <- match.arg(na_action)
  na_action <- if (na_action_raw %in% c("ignore","warn")) "omit" else na_action_raw

  if (!is.data.frame(data)) rlang::abort("iAge(): `data` must be a data.frame or tibble.")

  markers <- c("CRP","IL6","TNFa")
  cm_iage <- .hm_build_col_map(data, col_map, markers, fn = "iAge")
  data    <- cm_iage$data
  col_map <- cm_iage$col_map
  if (is.null(col_map) || !is.list(col_map)) rlang::abort("iAge(): `col_map` must be a named list.")
  missing_keys <- setdiff(markers, names(col_map))
  if (length(missing_keys)) rlang::abort(paste0("missing required columns: ", paste(missing_keys, collapse = ", ")))

  used_cols <- vapply(markers, function(m) col_map[[m]], character(1))
  if (any(!nzchar(used_cols))) {
    bad <- markers[!nzchar(used_cols)]
    rlang::abort(paste0("missing required columns: ", paste(bad, collapse = ", ")))
  }
  missing_cols <- setdiff(unname(used_cols), names(data))
  if (length(missing_cols)) {
    rlang::abort(sprintf("iAge(): column '%s' not found in data.", missing_cols[1]))
  }
  # Validate weights
  if (is.null(weights)) rlang::abort("iAge(): `weights` must be a numeric vector.")
  if (!is.numeric(weights)) rlang::abort("iAge(): `weights` must be numeric.")
  if (length(weights) != length(markers)) rlang::abort("iAge(): `weights` must have length 3.")
  if (!is.null(names(weights))) {
    if (all(markers %in% names(weights))) {
      weights <- weights[markers]
    }
  }
  if (any(!is.finite(weights))) rlang::abort("iAge(): `weights` must be finite numeric values.")
  if (abs(sum(weights) - 1) > 1e-8) rlang::abort("iAge(): `weights` must sum to 1.")

  .hm_log_cols(cm_iage, col_map, fn_name, verbose)
  if (isTRUE(verbose))
    hm_inform(sprintf("%s(): computing markers:\n  iAge  [weighted sum: CRP*%.2f + IL6*%.2f + TNFa*%.2f]",
      fn_name, weights[1], weights[2], weights[3]), level = "inform")

  for (cn in unname(used_cols)) {
    if (!is.numeric(data[[cn]])) {
      old <- data[[cn]]
      suppressWarnings({
        new <- as.numeric(old)
      })
      introduced_na <- sum(is.na(new) & !is.na(old))
      if (introduced_na > 0L) rlang::abort(sprintf("iAge(): mapped column '%s' must be numeric.", cn))
      data[[cn]] <- new
    }
    data[[cn]][!is.finite(data[[cn]])] <- NA_real_
  }

  .ia_quality_scan(data, unname(used_cols), na_warn_prop = na_warn_prop, do_warn = (na_action_raw == "warn"))
  if (na_action == "error") {
    if (!all(stats::complete.cases(data[, unname(used_cols), drop = FALSE]))) {
      rlang::abort("iAge(): missing or non-finite values in required inputs with na_action='error'.")
    }
  }

  M <- as.matrix(data[, unname(used_cols), drop = FALSE])
  w <- as.numeric(weights)
  if (na_action == "keep") {
    ok <- stats::complete.cases(M)
    iage <- rep(NA_real_, nrow(M))
    if (any(ok)) {
      M_ok <- M[ok, , drop = FALSE]
      M_ok[is.na(M_ok)] <- 0
      iage[ok] <- as.numeric(M_ok %*% w)
    }
  } else {
    M[is.na(M)] <- 0
    iage <- as.numeric(M %*% w)
  }

  out <- tibble::tibble(iAge = iage)
  if (!is.null(id_col)) {
    out[[id_col]] <- data[[id_col]]
    out <- out[, c(id_col, setdiff(names(out), id_col)), drop = FALSE]
    out <- tibble::as_tibble(out)
  }
  if (isTRUE(verbose)) { hm_inform(hm_result_summary(out, fn_name), level = "inform") }
  out
}

.ia_quality_scan <- function(df, cols, na_warn_prop = 0.2, do_warn = TRUE) {
  for (cn in cols) {
    x <- df[[cn]]
    if (!length(x)) next
    pna <- mean(is.na(x))
    if (do_warn && is.finite(pna) && pna >= na_warn_prop && pna > 0) {
      rlang::warn(sprintf("iAge(): column '%s' has high missingness (%.1f%%).", cn, 100 * pna))
    }
    neg_n <- sum(is.finite(x) & x < 0)
    if (do_warn && neg_n > 0L) {
      rlang::warn(sprintf("iAge(): column '%s' contains %d negative values; check units.", cn, neg_n))
    }
  }
  invisible(TRUE)
}

.ia_default_extreme_rules <- function() {
  list(
    CRP = c(0, 300),
    IL6 = c(0, 1000),
    TNFa = c(0, 2000)
  )
}

.ia_extreme_scan <- function(df, used_cols_named, rules) {
  flags <- list(); total <- 0L
  for (mk in names(used_cols_named)) {
    cn <- used_cols_named[[mk]]
    if (!mk %in% names(rules)) next
    rng <- rules[[mk]]
    x <- df[[cn]]
    bad <- is.finite(x) & (x < rng[1] | x > rng[2])
    flags[[mk]] <- bad
    total <- total + sum(bad)
  }
  list(count = total, flags = flags)
}

.ia_cap_inputs <- function(df, flags, used_cols_named, rules) {
  for (mk in names(flags)) {
    cn <- used_cols_named[[mk]]
    if (!mk %in% names(rules)) next
    bad <- flags[[mk]]
    rng <- rules[[mk]]
    x <- df[[cn]]
    x[bad & is.finite(x) & x < rng[1]] <- rng[1]
    x[bad & is.finite(x) & x > rng[2]] <- rng[2]
    df[[cn]] <- x
  }
  df
}

# tests/testthat/test_inflammatory_age.R

library(testthat)
library(tibble)

# Fixture data
df <- tibble(
  CRP  = c(1.0, 2.0, NA_real_),
  IL6  = c(5.0, NA_real_, 3.0),
  TNFa = c(1.5, 2.5, 3.5)
)
col_map <- list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa")

# 1) Default na_action='keep': NAs propagate to output
test_that("iAge default na_action='keep' propagates NA", {
  skip_on_cran()
  out <- iAge(df, col_map = col_map)
  # Row 2: IL6 is NA -> iAge should be NA
  expect_true(is.na(out$iAge[2]))
  # Row 3: CRP is NA -> iAge should be NA
  expect_true(is.na(out$iAge[3]))
  # Row 1: all present -> numeric
  expect_equal(out$iAge[1], 0.33 * 1.0 + 0.33 * 5.0 + 0.34 * 1.5)
})

# 1b) With na_action='omit': NAs treated as 0
test_that("iAge computes correct weighted sum with na_action='omit'", {
  skip_on_cran()
  out <- suppressWarnings(iAge(df, col_map = col_map, na_action = "omit"))
  expected <- rowSums(cbind(
    0.33 * df$CRP,
    0.33 * df$IL6,
    0.34 * df$TNFa
  ), na.rm = TRUE)
  expect_s3_class(out, "tbl_df")
  expect_named(out, "iAge")
  expect_equal(out$iAge, expected)
})

# 2) Custom weights
test_that("iAge respects custom weights and sums to correct values", {
  skip_on_cran()
  w <- c(CRP = 0.5, IL6 = 0.3, TNFa = 0.2)
  out2 <- suppressWarnings(iAge(df, col_map = col_map, weights = w, na_action = "omit"))
  expected2 <- rowSums(cbind(
    0.5 * df$CRP,
    0.3 * df$IL6,
    0.2 * df$TNFa
  ), na.rm = TRUE)
  expect_equal(out2$iAge, expected2)
})

# 3) NA handling: if all markers NA for a row, sum gives 0 (omit)
test_that("iAge NA handling gives zero for all-NA row with na_action='omit'", {
  skip_on_cran()
  df_na <- tibble(CRP = NA_real_, IL6 = NA_real_, TNFa = NA_real_)
  out3 <- suppressWarnings(iAge(df_na, col_map = col_map, na_action = "omit"))
  expect_equal(out3$iAge, 0)
})

# 4) Verbose message
test_that("iAge verbose emits col_map and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_ok <- tibble(CRP = 1, IL6 = 2, TNFa = 3)
  expect_message(iAge(df_ok, col_map = col_map, verbose = TRUE), "iAge")
  expect_message(iAge(df_ok, col_map = col_map, verbose = TRUE), "col_map")
  expect_message(iAge(df_ok, col_map = col_map, verbose = TRUE), "results:")
})

test_that("iAge verbose double-fire guard", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  df_ok <- tibble(CRP = 1, IL6 = 2, TNFa = 3)
  msgs <- testthat::capture_messages(iAge(df_ok, col_map = col_map, verbose = TRUE))
  expect_gte(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

# 5) Errors on invalid inputs
test_that("iAge errors on non-data.frame input", {
  skip_on_cran()
  expect_error(iAge("not_df", col_map), "data.frame or tibble")
})

test_that("partial col_map is supplemented by dictionary inference", {
  skip_on_cran()
  bad_map <- list(CRP = "CRP", IL6 = "IL6")
  expect_no_error(iAge(df, col_map = bad_map))
})

test_that("iAge errors on non-numeric weights", {
  skip_on_cran()
  expect_error(
    iAge(df, col_map = col_map, weights = c(CRP = "a", IL6 = 0.5, TNFa = 0.5)),
    "weights.*numeric"
  )
})

test_that("iAge errors on weights not summing to 1", {
  skip_on_cran()
  w2 <- c(CRP = 0.2, IL6 = 0.2, TNFa = 0.2)
  expect_error(
    iAge(df, col_map = col_map, weights = w2),
    "weights.*sum to 1"
  )
})

# 6) Errors on missing columns in data
test_that("iAge errors if data is missing required columns", {
  skip_on_cran()
  df_bad <- tibble(CRP = 1, IL6 = 2)
  expect_error(iAge(df_bad, col_map = col_map), "column 'TNFa'.*not found in data")
})

# 7) Errors if mapped column is non-numeric
test_that("iAge errors if mapped column is not numeric", {
  skip_on_cran()
  df_char <- tibble(CRP = c("x"), IL6 = 2, TNFa = 3)
  expect_error(iAge(df_char, col_map = col_map), "must be numeric")
})

#' Import Rdpack for Rd macros
#'
#' Ensures Rd macros from Rdpack are available without triggering an unused
#' import note during R CMD check.
#' @name healthmarkers-rdpack-import
#' @keywords internal
#' @import Rdpack
NULL

#' Calculate Standard Deviation Scores (SDS; z-scores) for health markers
#'
#' Computes per-variable SDS as (x - mean) / sd using supplied reference statistics.
#' Includes input validation, NA/error handling, data quality warnings, and
#' verbose progress via the package logger (hm_inform), aligned with HM-CS v3.
#'
#' Derivation: for each requested variable, SDS is calculated as
#' `(observed_value - reference_mean) / reference_sd` using user-supplied
#' reference statistics in `ref`.
#'
#' Usage note: this function does not derive normative reference values internally.
#' Users should provide `ref` from the target population (preferred), or from an
#' external source matched as closely as possible on age/sex/ethnicity/context.
#' Interpretability depends on the quality and relevance of those supplied
#' reference means and SDs.
#'
#' By default, returns a tibble with added `[var]_sds` columns (tidyverse-friendly).
#' For backward compatibility, you can request the previous list output.
#'
#' @param data A data.frame/tibble containing the variables.
#' @param vars Character vector of variable names in `data` to compute SDS for.
#'   Must be non-empty and unique.
#' @param ref A data.frame with columns: `variable`, `mean`, `sd` supplying reference
#'   statistics for each variable listed in `vars`. Must contain exactly one row per
#'   requested variable, with finite `mean` and `sd > 0`.
#' @param id_col Optional character scalar naming an ID column in `data` (used only
#'   in messages and duplicate-ID notes).
#' @param sds_cap Numeric scalar; absolute cap for SDS when `extreme_strategy = "cap"`.
#'   Must be positive. Default 6.
#' @param na_strategy One of c("omit","error","keep"):
#'   - "omit": drop rows with missing values across any of `vars` (default)
#'   - "error": stop if any missing values among `vars`
#'   - "keep": keep rows; SDS for missing inputs will be NA
#' @param extreme_strategy One of c("cap","warn","error","NA"):
#'   - "cap": cap |SDS| at `sds_cap` and warn (default)
#'   - "warn": keep extreme SDS, but warn
#'   - "error": stop if any |SDS| > `sds_cap`
#'   - "NA": set extreme SDS to NA
#' @param na_action Optional HM-CS alias for `na_strategy`
#'   (accepted values: keep/omit/error; used if provided).
#' @param check_extreme Logical; if TRUE, enables SDS extreme handling (alias for
#'   legacy behavior; defaults to TRUE in this implementation).
#' @param extreme_action Optional HM-CS alias for `extreme_strategy`
#'   (accepted values: cap/NA/error; used if provided).
#' @param warn_thresholds Named list controlling warnings (proportions in \eqn{[0,1]}):
#'   - na_prop: warn if proportion of rows with NA among `vars` exceeds this (default 0.05)
#'   - extreme_prop: warn if proportion of extreme SDS (cells) exceeds this (default 0.01)
#' @param return One of c("data","list"). "data" returns a tibble with added
#'   `[var]_sds` columns (default). "list" returns a list with components
#'   `data`, `summary`, and `warnings` (backward compatible).
#' @param verbose Logical; if TRUE, emit progress via hm_inform(). Also controlled by
#'   options(healthmarkers.verbose = "none"|"inform"|"debug").
#'
#' @return
#' - If `return = "data"` (default): a tibble with added `[var]_sds` columns.
#' - If `return = "list"`: a list with:
#'   - data: tibble with added SDS columns
#'   - summary: list with input/output row counts, omitted rows, total extremes, and per-variable summary
#'   - warnings: character vector of warning messages emitted
#'
#' @examples
#' ref <- data.frame(
#'   variable = c("bmi","sbp"),
#'   mean     = c(25, 120),
#'   sd       = c(4, 15)
#' )
#' df <- data.frame(
#'   id  = 1:6,
#'   bmi = c(24, 30, NA, 29, 10, 26),
#'   sbp = c(118, 200, 119, 121, 500, 120)
#' )
#' out_tbl <- calc_sds(
#'   data = df,
#'   vars = c("bmi","sbp"),
#'   ref = ref,
#'   id_col = "id",
#'   na_strategy = "omit",
#'   extreme_strategy = "cap",
#'   sds_cap = 6,
#'   verbose = FALSE
#' )
#' @importFrom tibble as_tibble
#' @export
calc_sds <- function(
  data,
  vars,
  ref,
  id_col = NULL,
  sds_cap = 6,
  na_strategy = c("omit","error","keep"),
  extreme_strategy = c("cap","warn","error","NA"),
  warn_thresholds = list(na_prop = 0.05, extreme_prop = 0.01),
  return = c("data","list"),
  verbose = TRUE,
  # HM-CS aliases (back-compat)
  na_action = NULL,
  check_extreme = TRUE,
  extreme_action = NULL
) {
  # HM-CS alias mapping (keep messages gated at debug)
  if (!is.null(na_action)) {
    na_strategy <- match.arg(na_action, c("keep","omit","error"))
    hm_inform("calc_sds(): using 'na_action' alias of 'na_strategy'.", level = "debug")
  }
  if (!is.null(extreme_action)) {
    extreme_strategy <- match.arg(extreme_action, c("cap","NA","error"))
    hm_inform("calc_sds(): using 'extreme_action' alias of 'extreme_strategy'.", level = "debug")
  }

  na_strategy <- match.arg(na_strategy)
  extreme_strategy <- match.arg(extreme_strategy)
  return <- match.arg(return)

  # --- Input validation ---
  if (!is.data.frame(data)) {
    rlang::abort("calc_sds(): `data` must be a data.frame or tibble.",
                 class = "healthmarkers_calc_sds_error_data_type")
  }
  if (!is.character(vars)) {
    rlang::abort("calc_sds(): `vars` must be a character vector.",
                 class = "healthmarkers_calc_sds_error_vars_type")
  }
  if (length(vars) == 0) {
    rlang::abort("calc_sds(): `vars` must be non-empty.",
                 class = "healthmarkers_calc_sds_error_vars_empty")
  }
  if (anyNA(vars) || any(vars == "")) {
    rlang::abort("calc_sds(): `vars` must not contain NA or empty names.",
                 class = "healthmarkers_calc_sds_error_vars_bad_names")
  }
  if (anyDuplicated(vars)) {
    rlang::abort("calc_sds(): `vars` must be unique (no duplicates).",
                 class = "healthmarkers_calc_sds_error_vars_duplicate")
  }
  .ensure_cols_exist(data, vars, label = "data")
  if (!is.null(id_col)) .ensure_cols_exist(data, id_col, label = "data (id_col)")

  ref_ok <- .validate_ref(ref, vars)
  if (!ref_ok$ok) {
    rlang::abort(ref_ok$msg, class = "healthmarkers_calc_sds_error_ref")
  }

  if (!is.numeric(sds_cap) || length(sds_cap) != 1 || !is.finite(sds_cap) || sds_cap <= 0) {
    rlang::abort("calc_sds(): `sds_cap` must be a positive finite number.",
                 class = "healthmarkers_calc_sds_error_sds_cap")
  }
  if (!is.list(warn_thresholds)) {
    rlang::abort("calc_sds(): `warn_thresholds` must be a list with elements `na_prop` and `extreme_prop`.",
                 class = "healthmarkers_calc_sds_error_warn_thresholds_type")
  }
  na_prop <- warn_thresholds$na_prop %||% 0.05
  extreme_prop <- warn_thresholds$extreme_prop %||% 0.01
  if (!(is.numeric(na_prop) && length(na_prop) == 1 && is.finite(na_prop) && na_prop >= 0 && na_prop <= 1)) {
    rlang::abort("calc_sds(): `warn_thresholds$na_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_calc_sds_error_warn_na_prop")
  }
  if (!(is.numeric(extreme_prop) && length(extreme_prop) == 1 && is.finite(extreme_prop) && extreme_prop >= 0 && extreme_prop <= 1)) {
    rlang::abort("calc_sds(): `warn_thresholds$extreme_prop` must be a single numeric in [0, 1].",
                 class = "healthmarkers_calc_sds_error_warn_extreme_prop")
  }
  if (!(is.logical(verbose) && length(verbose) == 1L && !is.na(verbose))) {
    rlang::abort("calc_sds(): `verbose` must be a single logical value.",
                 class = "healthmarkers_calc_sds_error_verbose_type")
  }

  # --- Start messages ---
  if (isTRUE(verbose)) {
    total_rows <- nrow(data)
    msg_id <- if (is.null(id_col)) "" else paste0(" (id: ", id_col, ")")
    hm_inform(sprintf("calc_sds: starting on %d row(s), %d variable(s)%s",
                      total_rows, length(vars), msg_id), level = "inform")
    map_parts <- vapply(vars, function(k) sprintf("%s -> '%s'", k, k), character(1))
    hm_inform(sprintf("calc_sds(): col_map: %s", paste(map_parts, collapse = ", ")), level = "inform")
  }

  warns <- character(0)
  out <- data

  # --- Coerce non-numeric vars with warning ---
  for (v in vars) {
    if (!is.numeric(out[[v]])) {
      old <- out[[v]]
      suppressWarnings(out[[v]] <- as.numeric(out[[v]]))
      introduced_na <- sum(is.na(out[[v]]) & !is.na(old))
      w <- sprintf("variable `%s` was coerced to numeric; NAs introduced: %d", v, introduced_na)
      rlang::warn(w, class = "healthmarkers_calc_sds_warn_na_coercion")
      warns <- c(warns, w)
      if (isTRUE(verbose)) hm_inform(sprintf("- coerced `%s` to numeric", v), level = "debug")
    }
  }

  # --- NA handling strategy ---
  row_has_na <- rowSums(is.na(out[, vars, drop = FALSE])) > 0
  n_rows_in <- nrow(out)
  n_rows_na <- sum(row_has_na)
  prop_na_rows <- if (n_rows_in > 0) n_rows_na / n_rows_in else 0

  if (prop_na_rows > na_prop) {
    w <- sprintf("calc_sds(): high NA proportion among `vars`: %.1f%% of rows", 100 * prop_na_rows)
    rlang::warn(w, class = "healthmarkers_calc_sds_warn_na_prop_high")
    warns <- c(warns, w)
  }

  if (na_strategy == "error" && n_rows_na > 0) {
    rlang::abort(sprintf("Missing values found in `vars` for %d row(s); na_strategy='error'", n_rows_na),
                 class = "healthmarkers_calc_sds_error_na_strategy_missing")
  } else if (na_strategy == "omit" && n_rows_na > 0) {
    if (isTRUE(verbose)) hm_inform(sprintf("- omitting %d row(s) with NA across `vars`", n_rows_na), level = "inform")
    out <- out[!row_has_na, , drop = FALSE]
  } else if (isTRUE(verbose) && n_rows_na > 0) {
    hm_inform(sprintf("- keeping %d row(s) with NA across `vars`", n_rows_na), level = "inform")
  }

  # Recompute after omission
  n_rows_out <- nrow(out)
  omitted_rows <- n_rows_in - n_rows_out

  # --- Prepare reference lookups ---
  ref_map_mean <- setNames(ref$mean, ref$variable)
  ref_map_sd   <- setNames(ref$sd,   ref$variable)

  # Validate SD > 0 for all vars
  zero_sd_vars <- vars[ref_map_sd[vars] <= 0 | !is.finite(ref_map_sd[vars])]
  if (length(zero_sd_vars)) {
    rlang::abort(sprintf("Reference SD must be > 0 and finite for: %s", paste(zero_sd_vars, collapse = ", ")),
                 class = "healthmarkers_calc_sds_error_ref_sd")
  }

  # --- Compute SDS per variable ---
  if (isTRUE(verbose)) hm_inform(sprintf("Computing SDS for %d variable(s)...", length(vars)), level = "inform")
  per_var_summary <- data.frame(
    variable = vars,
    n_missing = integer(length(vars)),
    n_extreme = integer(length(vars)),
    stringsAsFactors = FALSE
  )

  total_extreme <- 0L

  for (i in seq_along(vars)) {
    v <- vars[i]
    mu <- ref_map_mean[[v]]
    sdv <- ref_map_sd[[v]]

    z <- (out[[v]] - mu) / sdv
    # Count missing (after strategy)
    per_var_summary$n_missing[i] <- sum(is.na(out[[v]]))

    if (isTRUE(check_extreme)) {
      is_extreme <- is.finite(z) & abs(z) > sds_cap
      n_extreme <- sum(is_extreme, na.rm = TRUE)
      total_extreme <- total_extreme + n_extreme

      if (n_extreme > 0) {
        if (identical(extreme_strategy, "error")) {
          rlang::abort(sprintf("Found %d SDS beyond +/-%g for `%s`", n_extreme, sds_cap, v),
                       class = "healthmarkers_calc_sds_error_extreme_sds")
        } else if (identical(extreme_strategy, "cap")) {
          z[is_extreme] <- sds_cap * sign(z[is_extreme])
          w <- sprintf("Capped %d SDS beyond +/-%g for `%s`", n_extreme, sds_cap, v)
           rlang::warn(w, class = "healthmarkers_calc_sds_warn_cap_extreme")
           warns <- c(warns, w)
        } else if (identical(extreme_strategy, "warn")) {
          w <- sprintf("Detected %d SDS beyond +/-%g for `%s` (not capped)", n_extreme, sds_cap, v)
           rlang::warn(w, class = "healthmarkers_calc_sds_warn_detect_extreme")
           warns <- c(warns, w)
        } else if (identical(extreme_strategy, "NA")) {
          z[is_extreme] <- NA_real_
        }
      }

      per_var_summary$n_extreme[i] <- n_extreme
    }

    out[[paste0(v, "_sds")]] <- z
  }

  # --- Data quality warning for extremes proportion ---
  prop_extreme <- if (n_rows_out > 0) total_extreme / (n_rows_out * length(vars)) else 0
  if (prop_extreme > extreme_prop) {
    w <- sprintf("calc_sds(): high proportion of extreme SDS: %.2f%% of computed cells", 100 * prop_extreme)
    rlang::warn(w, class = "healthmarkers_calc_sds_warn_extreme_prop_high")
    warns <- c(warns, w)
  }

  # --- Verbose completion summary ---
  if (isTRUE(verbose)) {
    hm_inform(level = "inform", msg = "calc_sds: completed")
    hm_inform(sprintf("- rows in:    %d", n_rows_in), level = "debug")
    hm_inform(sprintf("- rows out:   %d", n_rows_out), level = "debug")
    hm_inform(sprintf("- omitted:    %d (na_strategy = '%s')", omitted_rows, na_strategy), level = "debug")
    hm_inform(sprintf("- extremes:   %d (strategy = '%s', cap = %s)", total_extreme, extreme_strategy, as.character(sds_cap)), level = "debug")
    if (!is.null(id_col) && n_rows_out > 0 && anyDuplicated(out[[id_col]]) > 0) {
      hm_inform("- note: duplicate IDs detected", level = "debug")
    }
    hm_inform(level = "inform", msg = hm_result_summary(
      tibble::as_tibble(out)[paste0(vars, "_sds")], "calc_sds"
    ))
  }

  out_tbl <- tibble::as_tibble(out)

  if (return == "data") {
    return(out_tbl)
  } else {
    return(list(
      data = out_tbl,
      summary = list(
        rows_in = n_rows_in,
        rows_out = n_rows_out,
        omitted_rows = omitted_rows,
        total_extreme = total_extreme,
        per_var = per_var_summary
      ),
      warnings = unique(warns)
    ))
  }
}

# Internal helpers ----------------------------------------------------------

# Safe null-coalescing for lists
`%||%` <- function(x, y) if (is.null(x)) y else x

.ensure_cols_exist <- function(df, cols, label = "data") {
  missing_cols <- setdiff(cols, names(df))
  if (length(missing_cols)) {
    rlang::abort(
      sprintf("Missing column(s) in `%s`: %s", label, paste(missing_cols, collapse = ", ")),
      class = "healthmarkers_calc_sds_error_missing_columns"
    )
  }
  invisible(TRUE)
}

.validate_ref <- function(ref, vars) {
  if (!is.data.frame(ref)) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref` must be a data.frame with columns: variable, mean, sd"))
  }
  req <- c("variable","mean","sd")
  if (!all(req %in% names(ref))) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref` must have columns: variable, mean, sd"))
  }
  if (!is.character(ref$variable)) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref$variable` must be character"))
  }
  if (!is.numeric(ref$mean) || !is.numeric(ref$sd)) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref$mean` and `ref$sd` must be numeric"))
  }
  if (anyDuplicated(ref$variable)) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref` must have at most one row per variable"))
  }
  missing_vars <- setdiff(vars, ref$variable)
  if (length(missing_vars)) {
    return(list(ok = FALSE, msg = sprintf("calc_sds(): `ref` is missing stats for: %s", paste(missing_vars, collapse = ", "))))
  }
  rsub <- ref[match(vars, ref$variable), , drop = FALSE]
  if (any(!is.finite(rsub$mean)) || any(!is.finite(rsub$sd))) {
    return(list(ok = FALSE, msg = "calc_sds(): `ref` contains non-finite mean or sd"))
  }
  list(ok = TRUE, msg = "")
}

# File: tests/testthat/test_calc_sds.R

ref <- data.frame(
  variable = c("bmi", "sbp"),
  mean     = c(25, 120),
  sd       = c(5, 15),
  stringsAsFactors = FALSE
)

df <- data.frame(
  id  = 1:6,
  bmi = c(24, 30, 26, 29, 20, 26),
  sbp = c(118, 130, 119, 121, 140, 120)
)

test_that("calc_sds computes z-scores and returns structure", {
  res <- calc_sds(
    data = df,
    vars = c("bmi", "sbp"),
    ref = ref,
    id_col = "id",
    na_action = "keep",
    check_extreme = TRUE,
    extreme_action = "cap",
    warn_thresholds = list(na_prop = 1, extreme_prop = 1),
    return = "list",
    verbose = FALSE
  )
  expect_type(res, "list")
  expect_true(all(c("data", "summary", "warnings") %in% names(res)))
  out <- res$data
  expect_true(all(c("bmi_sds", "sbp_sds") %in% names(out)))
  expect_equal(out$bmi_sds, (df$bmi - 25) / 5)
  expect_equal(out$sbp_sds, (df$sbp - 120) / 15)
  expect_equal(res$summary$rows_in, nrow(df))
  expect_equal(res$summary$rows_out, nrow(df))
})

test_that("NA actions: keep, omit, error", {
  skip_on_cran()
  df_na <- df
  df_na$bmi[c(2, 5)] <- NA_real_
  # keep -> same rows
  res_keep <- calc_sds(
    data = df_na, vars = c("bmi", "sbp"), ref = ref,
    na_action = "keep", warn_thresholds = list(na_prop = 1, extreme_prop = 1),
    return = "list", verbose = FALSE
  )
  expect_equal(res_keep$summary$rows_out, nrow(df_na))
  expect_true(is.na(res_keep$data$bmi_sds[2]))
  # omit -> fewer rows
  res_omit <- calc_sds(
    data = df_na, vars = c("bmi", "sbp"), ref = ref,
    na_action = "omit", warn_thresholds = list(na_prop = 1, extreme_prop = 1),
    return = "list", verbose = FALSE
  )
  expect_equal(res_omit$summary$omitted_rows, 2)
  # error -> throws
  expect_error(
    calc_sds(data = df_na, vars = c("bmi", "sbp"), ref = ref,
             na_action = "error", warn_thresholds = list(na_prop = 1, extreme_prop = 1),
             verbose = FALSE),
    "Missing values found in `vars`"
  )
})

test_that("SDS extremes: cap, NA, error; and disabling via check_extreme", {
  skip_on_cran()
  df_ext <- df
  df_ext$bmi[1] <- 50  # mean=25, sd=5 -> z=5
  # cap -> warning and capped to 3
  expect_warning(
    res_cap <- calc_sds(df_ext, vars = c("bmi"),
                        ref = ref[ref$variable == "bmi",],
                        sds_cap = 3, extreme_action = "cap", check_extreme = TRUE,
                        warn_thresholds = list(na_prop = 1, extreme_prop = 1),
                        return = "list", verbose = FALSE),
    "Capped 1 SDS beyond"
  )
  expect_equal(res_cap$data$bmi_sds[1], 3)
  # NA -> no cap; becomes NA
  res_na <- calc_sds(df_ext, vars = c("bmi"),
                     ref = ref[ref$variable == "bmi",],
                     sds_cap = 3, extreme_action = "NA", check_extreme = TRUE,
                     warn_thresholds = list(na_prop = 1, extreme_prop = 1),
                     return = "list", verbose = FALSE)
  expect_true(is.na(res_na$data$bmi_sds[1]))
  # error -> stops
  expect_error(
    calc_sds(df_ext, vars = c("bmi"),
             ref = ref[ref$variable == "bmi",], sds_cap = 3,
             extreme_action = "error", check_extreme = TRUE,
             warn_thresholds = list(na_prop = 1, extreme_prop = 1),
             verbose = FALSE),
    "Found 1 SDS beyond"
  )
  # check_extreme = FALSE -> leave value unchanged (> cap)
  res_off <- calc_sds(df_ext, vars = c("bmi"),
                      ref = ref[ref$variable == "bmi",],
                      sds_cap = 3, extreme_action = "cap", check_extreme = FALSE,
                      warn_thresholds = list(na_prop = 1, extreme_prop = 1),
                      return = "list", verbose = FALSE)
  expect_gt(res_off$data$bmi_sds[1], 3)
})

test_that("Input validation: refs and args", {
  skip_on_cran()
  # missing column in data
  expect_error(
    calc_sds(df, vars = c("bmi", "sbp", "missing"), ref = ref, verbose = FALSE),
    "Missing column\\(s\\) in `data`"
  )
  # missing stats for a var in ref
  ref_missing <- subset(ref, variable == "bmi")
  expect_error(
    calc_sds(df, vars = c("bmi", "sbp"), ref = ref_missing, verbose = FALSE),
    "`ref` is missing stats for: sbp"
  )
  # zero or non-finite SD
  bad_ref <- ref; bad_ref$sd[bad_ref$variable == "bmi"] <- 0
  expect_error(
    calc_sds(df, vars = c("bmi"), ref = bad_ref, verbose = FALSE),
    "Reference SD must be > 0 and finite"
  )
  # sds_cap must be positive finite
  expect_error(
    calc_sds(df, vars = c("bmi"), ref = ref, sds_cap = 0, verbose = FALSE),
    "must be a positive finite"
  )
  # warn_thresholds can omit elements; defaults are applied -> no error
  expect_error(
    calc_sds(df, vars = c("bmi"), ref = ref, warn_thresholds = list(na_prop = 0.1), verbose = FALSE),
    NA
  )
})

test_that("Coercion of non-numeric vars warns and introduces NA", {
  skip_on_cran()
  df_chr <- df
  df_chr$bmi <- c("24", "30", "not-a-number", "29", "20", "26")
  expect_warning(
    res <- calc_sds(df_chr, vars = c("bmi"),
                    ref = ref[ref$variable == "bmi",],
                    warn_thresholds = list(na_prop = 1, extreme_prop = 1),
                    na_action = "keep",
                    return = "list", verbose = FALSE),
    "was coerced to numeric"
  )
  expect_true(is.na(res$data$bmi_sds[3]))
})

test_that("verbose = TRUE emits starting, column map, and results messages", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  expect_message(
    calc_sds(df, vars = c("bmi", "sbp"), ref = ref, verbose = TRUE,
             warn_thresholds = list(na_prop = 1, extreme_prop = 1)),
    "calc_sds"
  )
  expect_message(
    calc_sds(df, vars = c("bmi", "sbp"), ref = ref, verbose = TRUE,
             warn_thresholds = list(na_prop = 1, extreme_prop = 1)),
    "col_map"
  )
  expect_message(
    calc_sds(df, vars = c("bmi", "sbp"), ref = ref, verbose = TRUE,
             warn_thresholds = list(na_prop = 1, extreme_prop = 1)),
    "results:"
  )
})

test_that("verbose double-fire guard: each message fires exactly once", {
  skip_on_cran()
  withr::local_options(healthmarkers.verbose = "inform")
  msgs <- testthat::capture_messages(
    calc_sds(df, vars = c("bmi", "sbp"), ref = ref, verbose = TRUE,
             warn_thresholds = list(na_prop = 1, extreme_prop = 1))
  )
  expect_equal(sum(grepl("col_map", msgs)), 1L)
  expect_equal(sum(grepl("results:",   msgs)), 1L)
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz-rdpack.R
\name{healthmarkers-rdpack-import}
\alias{healthmarkers-rdpack-import}
\title{Import Rdpack for Rd macros}
\description{
Ensures Rd macros from Rdpack are available without triggering an unused
import note during R CMD check.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{spq_score}
\alias{spq_score}
\title{Schizotypal Personality Questionnaire (key-driven)}
\usage{
spq_score(
  data,
  col_map = list(),
  key,
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "SPQ",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{key}{List with \code{items}, \code{min_val}, \code{max_val}, optional \code{reverse} and \code{subscales}.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only (total and optional subscales). Input columns are not included.
}
\description{
Schizotypal Personality Questionnaire (key-driven)
}
\examples{
spq_key <- list(items = sprintf("spq_\%02d", 1:5), min_val = 0, max_val = 1)
df <- data.frame(spq_01 = 0, spq_02 = 1, spq_03 = 0, spq_04 = 1, spq_05 = 0)
spq_score(df, key = spq_key)
}
\references{
\insertRef{raine1991spq}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_risk_ascvd}
\alias{cvd_risk_ascvd}
\title{ASCVD risk (ACC/AHA Pooled Cohort Equations)}
\usage{
cvd_risk_ascvd(
  data,
  year = 10,
  col_map = NULL,
  na_warn_prop = 0.2,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{A data frame with the required cardiovascular risk columns.}

\item{year}{Risk horizon: 10 or 30.}

\item{col_map}{Optional named list mapping internal keys (\code{age}, \code{sex},
\code{race}, \code{smoker}, \code{total_chol}, \code{HDL_c}, \code{sbp},
\code{bp_treated}, \code{diabetes}, \code{bmi}) to actual column names in
\code{data}. If \code{NULL} (default), column names are auto-inferred then fall
back to the key names themselves. \code{sex} accepts \code{1}/\code{0},
\code{"m"}/\code{"f"}, or \code{"male"}/\code{"female"} (case-insensitive).}

\item{na_warn_prop}{Proportion (0-1) to flag high missingness warnings (default 0.2).
Only used when \code{verbose = TRUE}; underlying backend handles NA as per its API.}

\item{verbose}{Logical; if TRUE, prints progress and a short summary.}

\item{...}{Passed to the underlying PooledCohort function.}
}
\value{
A tibble with columns \code{model}, \code{year}, \code{risk} (percentage).
}
\description{
Wrapper around the PooledCohort ASCVD calculators with added input validation,
optional data-quality warnings, and quiet failure to NA if the backend errors.
}
\examples{
df <- tibble::tibble(
  age = 55, sex = 1, race = "white", smoker = FALSE,
  total_chol = 200, HDL_c = 50, sbp = 140, bp_treated = FALSE,
  diabetes = FALSE, bmi = 27
)
if (requireNamespace("PooledCohort", quietly = TRUE)) {
  cvd_risk_ascvd(df, year = 10, verbose = TRUE)
}

}
\references{
\insertRef{goff2014accaha}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{normalize_vec}
\alias{normalize_vec}
\title{Normalize a numeric vector}
\usage{
normalize_vec(
  x,
  method = c("none", "z", "inverse", "range", "robust"),
  na_rm = TRUE,
  feature_range = c(0, 1),
  invnorm_denominator = c("n", "n+1", "blom"),
  ties = c("average", "first", "last", "random", "max", "min"),
  warn_constant = TRUE
)
}
\arguments{
\item{x}{A numeric (or numeric-coercible) vector.}

\item{method}{One of c("none","z","inverse","range","robust"). Default "none".}

\item{na_rm}{Logical; remove NAs when estimating statistics (mean, sd, etc.). Default TRUE.}

\item{feature_range}{Numeric length-2 vector giving the target range for method = "range".
Default c(0, 1).}

\item{invnorm_denominator}{One of c("n","n+1","blom") controlling the denominator
of the inverse-normal transform:
\itemize{
\item "n": p = (r - 0.5) / n  (Rankit; default)
\item "n+1": p = (r - 0.5) / (n + 1)
\item "blom": p = (r - 3/8) / (n + 1/4)  (Blom, 1958)
}}

\item{ties}{Ties method passed to base::rank for method = "inverse".
One of c("average","first","last","random","max","min"). Default "average".}

\item{warn_constant}{Logical; if TRUE, warn when input is constant and a zero vector
(or lower bound for range) is returned. Default TRUE.}
}
\value{
A numeric vector of the same length as x.
}
\description{
Utility used across HealthMarkers to normalize numeric vectors with several
common schemes while handling edge cases (constant vectors, all-NA, non-finite
values) robustly. NA positions are preserved.
}
\details{
Methods:
\itemize{
\item "none": return input as-is (no coercion; fully backward compatible).
\item "z": z-score (mean 0, sd 1). Constant vectors return zeros (non-NA entries).
\item "range": min-max to a target interval (default \eqn{[0,1]}). Constant vectors return
the lower bound (mapped from zeros).
\item "robust": median/MAD scaling. Constant vectors (MAD=0) return zeros.
\item "inverse": rank-based inverse normal transform (normal scores).
}
}
\examples{
x <- c(1, 2, 3, NA, 5)
normalize_vec(x, "none")
normalize_vec(x, "z")
normalize_vec(x, "range", feature_range = c(-1, 1))
normalize_vec(x, "robust")
normalize_vec(x, "inverse")               # Rankit (default)
normalize_vec(x, "inverse", invnorm_denominator = "blom")

}
\references{
\insertRef{beasley2009invnorm}{HealthMarkers}
\insertRef{leys2013mad}{HealthMarkers}
\insertRef{bland1996sd}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frax_score.R
\name{frax_score}
\alias{frax_score}
\title{FRAX 10-Year Fracture Risk Score (simplified placeholder)}
\usage{
frax_score(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  country = NULL,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame or tibble with inputs.}

\item{col_map}{Named list mapping required and optional inputs:
\itemize{
\item Required: age, sex
\item Optional binary risk factors: prior_fracture, parent_fracture, steroids,
rheumatoid, secondary_op, smoker, alcohol
\item Optional: bmd (T-score; if present and in (-5,0], used to adjust risk)
}}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{country}{optional country/region code for FRAX calibration (accepted, currently unused).}

\item{verbose}{Logical; if TRUE, emits progress via rlang::inform.}
}
\value{
Tibble with frax_major_percent and frax_hip_percent.
}
\description{
Estimates the 10-year probabilities of major osteoporotic and hip fracture
using a simplified, non-validated approximation based on FRAX risk factors.
This is for development/demo only and does not implement the proprietary FRAX algorithm.
}
\details{
Important: this function is an educational placeholder and must not be used for
clinical decision-making, patient counseling, or guideline-based treatment selection.
}
\examples{
df <- data.frame(Age = c(65, 72, 58), Sex = c("female", "female", "male"))
frax_score(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{psych_markers}
\alias{psych_markers}
\title{Psychometric markers dispatcher}
\usage{
psych_markers(
  data,
  col_map = list(),
  which = c("phq9", "gad7", "k6", "k10", "ghq12_likert", "ghq12_binary", "who5", "isi",
    "mdq", "asrs", "bis", "spq", "cognitive", "dx_flags", "med_flags"),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  bis_key = NULL,
  spq_key = NULL,
  cognitive_method = c("z_mean", "pca1"),
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Nested list of mappings per instrument (e.g., col_map$phq9, col_map$bis, col_map$dx_flags, ...).}

\item{which}{Vector of modules to compute (e.g., "phq9", "gad7", "bis").}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{bis_key}{SPQ/BIS key list passed to \code{bis_score} when requested.}

\item{spq_key}{SPQ key list passed to \code{spq_score} when requested.}

\item{cognitive_method}{Method passed to \code{cognitive_score} ("z_mean" or "pca1").}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of computed score columns from all requested modules, bound together. No input columns are included in the output.
}
\description{
Psychometric markers dispatcher
}
\examples{
df <- data.frame(
  phq9_01 = 0, phq9_02 = 1, phq9_03 = 2, phq9_04 = 1, phq9_05 = 0,
  phq9_06 = 1, phq9_07 = 2, phq9_08 = 1, phq9_09 = 0,
  gad7_01 = 0, gad7_02 = 1, gad7_03 = 2, gad7_04 = 1, gad7_05 = 0,
  gad7_06 = 1, gad7_07 = 2
)
psych_markers(df, which = c("phq9", "gad7"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liver_fat_markers.R
\name{liver_fat_markers}
\alias{liver_fat_markers}
\title{Liver fat surrogates: HSI and NAFLD Liver Fat Score}
\usage{
liver_fat_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame with needed columns (see col_map).}

\item{col_map}{Named list mapping:
\itemize{
\item Required for HSI: ALT, AST, BMI
\item Optional direct inputs: sex, diabetes, MetS, insulin
\item Optional to derive MetS or insulin: I0, waist, TG, HDL_c, sbp, bp_sys, bp_treated, glucose, G0
}}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{na_warn_prop}{Proportion in \eqn{[0,1]} for high-missingness warnings when na_action = "warn". Default 0.2.}

\item{verbose}{Logical; if TRUE, prints column mapping and computing messages.}
}
\value{
A tibble with columns HSI and NAFLD_LFS.
}
\description{
Computes:
\itemize{
\item HSI = 8 * (ALT/AST) + BMI + 2 (if female) + 2 (if diabetes)
\item NAFLD-LFS = -2.89 + 1.18\emph{MetS + 0.45}Type2DM + 0.15\emph{Insulin_u + 0.04}AST - 0.94*(AST/ALT)
}
}
\details{
Assumptions/units:
\itemize{
\item ALT, AST in U/L; BMI in kg/m^2.
\item \code{insulin} is expected in muU/mL; if unavailable and \code{I0} is provided,
\code{I0} is interpreted in pmol/L and converted to muU/mL via /6.
\item MetS is taken directly if provided; otherwise derived using a simplified
NCEP-ATP III style rule when sufficient inputs exist.
\item Type2DM is taken from \code{diabetes} (logical or 0/1).
}

These scores are surrogate indices for research/feature-engineering use and
are not validated as standalone clinical diagnostic tools.
}
\examples{
df <- data.frame(ALT=20, AST=25, BMI=27, sex="female", diabetes=FALSE, I0=60)
liver_fat_markers(
  df,
  col_map = list(ALT="ALT", AST="AST", BMI="BMI",
                 sex="sex", diabetes="diabetes", I0="I0")
)
}
\references{
\insertRef{lee2010hsi}{HealthMarkers}
\insertRef{kotronen2009nafldlfs}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{.hm_build_col_map}
\alias{.hm_build_col_map}
\title{Build a complete col_map with dictionary inference and alias materialization}
\usage{
.hm_build_col_map(data, col_map, keys, fn = "")
}
\arguments{
\item{data}{The data.frame being analysed (modified locally; original
is not altered outside the function call).}

\item{col_map}{Named list of user-supplied key -> column-name mappings, or
\code{NULL} for full auto-inference.}

\item{keys}{Character vector of short internal keys to resolve.}

\item{fn}{Calling function name string (for messages).}
}
\value{
A list with four elements:
\describe{
\item{data}{The (possibly augmented) data.frame.}
\item{col_map}{Complete named list of key -> resolved column name.}
\item{user_keys}{Keys that were explicitly provided in the original
\code{col_map} argument.}
\item{inferred_keys}{Keys resolved by dictionary inference or literal
fallback (i.e., not in \code{user_keys}).}
}
}
\description{
Combines three previously separate steps into one call:
\enumerate{
\item User-supplied entries in \code{col_map} are treated as authoritative
Layer 0 seeds -- they are never overwritten.
\item \code{hm_col_report()} fills any remaining unmatched keys via the
full synonym dictionary (exact -> case-insensitive -> contains layers).
\item Literal-name fallback: if a key in \code{keys} still has no match,
but the key name exists verbatim as a column in \code{data}, it is
used directly.
\item Materialization: for every mapped key \code{k} where
\code{col_map[[k]] != k} and \code{k} is \emph{not} already a column
in \code{data}, a copy \code{data[[k]] <- data[[col_map[[k]]]]} is
added. This ensures that downstream pre-computation helpers (which
check \code{names(data)} for dependency names) and their
\code{compute()} closures (which use literal key names) both resolve
correctly even when the physical column has a non-standard name.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vitamin_markers.R
\name{vitamin_markers}
\alias{vitamin_markers}
\title{Compute composite vitamin and endocrine marker ratios and z-scores}
\usage{
vitamin_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with vitamin/analyte columns.}

\item{col_map}{Named list mapping required keys to column names:
VitD, VitD_ref_mean, VitD_ref_sd, B12, Folate, Ferritin, TSat,
Cortisol, DHEAS, Testosterone, Estradiol, TSH, free_T4,
Retinol, Retinol_ref_mean, Retinol_ref_sd, Tocopherol, Total_lipids,
PIVKA_II, VitC, Homocysteine, MMA, Magnesium, Zinc, Copper.}

\item{na_action}{One of c("keep","omit","error") for required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness debug notices. Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, input
availability, physiological range information (informational only, values
not altered), the list of markers being computed with their inputs, and a
per-column results summary.}
}
\value{
A tibble with columns:
VitD_Z, B12_Fol_Ratio, Ferr_TSat_R, Cort_DHEA_R, T_E2_Ratio, TSH_fT4_R,
Retinol_Z, Toco_Lip_R, PIVKA_II, VitC, Homocysteine, MMA, Mg_Zn_R, Cu_Zn_R.
If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}),
it is prepended as the first output column.
}
\description{
Given serum/plasma vitamins and related analytes, \code{vitamin_markers()} computes:
\itemize{
\item VitD_Z: z-score of 25(OH)D using provided reference mean/sd
\item B12_Fol_Ratio: vitamin B12 / folate
\item Ferr_TSat_R: ferritin / transferrin saturation (TSat)
\item Cort_DHEA_R: cortisol / DHEA-S
\item T_E2_Ratio: testosterone / estradiol
\item TSH_fT4_R: TSH / free T4
\item Retinol_Z: z-score of retinol using provided reference mean/sd
\item Toco_Lip_R: alpha-tocopherol / total lipids
\item Mg_Zn_R: magnesium / zinc
\item Cu_Zn_R: copper / zinc
Plus pass-through: PIVKA_II, VitC, Homocysteine, MMA
}
}
\details{
HM-CS v2:
\itemize{
\item Validation via \code{hm_validate_inputs(data, col_map, required_keys, fn)}
\item User errors via \code{rlang::abort(..., class=...)}
\item Verbosity via \code{hm_inform(level)} controlled by \code{options(healthmarkers.verbose)}
\item High-missingness diagnostics at debug level only
}
}
\note{
\code{VitD_Z} and \code{Retinol_Z} are z-scores using \strong{user-supplied} reference
mean and SD; no population reference equations are applied. All ratio markers
(\code{B12_Fol_Ratio}, \code{Ferr_TSat_R}, \code{Cort_DHEA_R}, \code{T_E2_Ratio}, \code{TSH_fT4_R},
\code{Toco_Lip_R}, \code{Mg_Zn_R}, \code{Cu_Zn_R}) are simple numerator/denominator
divisions. \code{PIVKA_II}, \code{VitC}, \code{Homocysteine}, and \code{MMA} are
\strong{pass-through} columns; no formula is applied.
}
\examples{
\donttest{
# All 25 required columns must be supplied
df <- data.frame(
  VitD = 50, VitD_ref_mean = 40, VitD_ref_sd = 5,
  B12 = 300, Folate = 15, Ferritin = 80, TSat = 0.25,
  Cortisol = 200, DHEAS = 100, Testosterone = 12, Estradiol = 120,
  TSH = 2, free_T4 = 14, Retinol = 0.8, Retinol_ref_mean = 0.9,
  Retinol_ref_sd = 0.2, Tocopherol = 30, Total_lipids = 3,
  PIVKA_II = 5, VitC = 60, Homocysteine = 10, MMA = 0.3,
  Magnesium = 0.8, Zinc = 15, Copper = 15
)
vitamin_markers(df, verbose = FALSE)
}

}
\references{
\insertRef{holick2007vitddeficiency}{HealthMarkers} (vitamin D deficiency review; background)
\insertRef{oleary2010vitb12}{HealthMarkers} (vitamin B12 in health and disease; background)
\insertRef{ganz2015iron}{HealthMarkers} (iron homeostasis and ferritin; background)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{ghq12_score}
\alias{ghq12_score}
\title{GHQ-12 scoring (Likert or binary)}
\usage{
ghq12_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "GHQ12",
  method = c("likert", "binary"),
  case_cutoff_binary = 3,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{method}{Scoring method: \code{likert} (0-3 per item) or \code{binary} (0/1 per item).}

\item{case_cutoff_binary}{Cut-off for case status when using binary scoring.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{GHQ12_total_likert} (likert method) or \code{GHQ12_total_binary} and \code{GHQ12_case_binary} (binary method). Input columns are not included.
}
\description{
GHQ-12 scoring (Likert or binary)
}
\examples{
df <- data.frame(ghq12_01 = 0, ghq12_02 = 1, ghq12_03 = 2, ghq12_04 = 1,
                 ghq12_05 = 0, ghq12_06 = 1, ghq12_07 = 0, ghq12_08 = 1,
                 ghq12_09 = 2, ghq12_10 = 1, ghq12_11 = 0, ghq12_12 = 1)
ghq12_score(df, method = "likert")
}
\references{
\insertRef{goldberg1988ghq12}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/atherogenic_indices.R
\name{atherogenic_indices}
\alias{atherogenic_indices}
\title{Compute atherogenic indices}
\usage{
atherogenic_indices(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  normalize = c("none", "log10"),
  verbose = TRUE
)
}
\arguments{
\item{data}{data.frame/tibble with lipid columns.}

\item{col_map}{named list mapping keys to columns, e.g. list(TG="TG", HDL_c="HDL_c", TC="TC", LDL_c="LDL_c").}

\item{na_action}{one of c("keep","omit","error").}

\item{normalize}{one of c("none","log10"). Reserved; AIP always uses log10(TG/HDL_c).}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, the list
of indices being computed, and a per-column results summary.}
}
\value{
tibble with columns AIP, CRI_I, CRI_II. If an ID column is detected
in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}), it is prepended.
}
\description{
Calculates:
\itemize{
\item AIP: Atherogenic Index of Plasma = log10(TG / HDL_c)
\item CRI_I: Castelli Risk Index I = TC / HDL_c
\item CRI_II: Castelli Risk Index II = LDL_c / HDL_c
}
}
\details{
Behavior:
\itemize{
\item Required keys: TG, HDL_c. Optional: TC, LDL_c.
\item NA policy via \code{na_action}: "keep" (default), "omit" (drop rows with any NA in used lipids), "error".
\item Extreme screening via \code{check_extreme} and \code{extreme_action} ("warn","cap","error","ignore","NA").
Default bounds (mg/dL) used only for screening: TG (0, 10000), HDL_c (0, 1000), LDL_c (0, 10000), TC (0, 10000).
Note: All indices are unitless ratios; units cancel in computations.
\item Emits progress via \code{hm_inform()} when \code{verbose = TRUE} or when package option enables logs.
}
}
\examples{
df <- tibble::tibble(
  TG = c(150, 200),
  HDL_c = c(50, 40),
  TC = c(200, 220),
  LDL_c = c(120, 150)
)
cm <- list(TG = "TG", HDL_c = "HDL_c", TC = "TC", LDL_c = "LDL_c")
atherogenic_indices(df, col_map = cm)

}
\references{
\insertRef{dobiasova2004aip}{HealthMarkers}
\insertRef{castelli1977framingham}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{.hm_normalize_na_action}
\alias{.hm_normalize_na_action}
\title{Normalise na_action aliases to canonical values}
\usage{
.hm_normalize_na_action(na_action)
}
\arguments{
\item{na_action}{character(1) already matched via match.arg()}
}
\description{
Converts the backward-compat aliases "ignore" and "warn" to "keep" so that
the rest of each function only needs to handle c("keep","omit","error").
Returns a list with:
\itemize{
\item na_action_eff: the canonical value ("keep", "omit", or "error")
\item na_action_raw: the original matched value (for warn-diagnostic checks)
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frailty_index.R
\name{frailty_index}
\alias{frailty_index}
\title{Compute Frailty (Deficit) Index using di::di with QA and verbose summaries}
\usage{
frailty_index(
  data,
  cols = NULL,
  invert = NULL,
  rescale = TRUE,
  age = NULL,
  rescale.custom = NULL,
  rescale.avoid = NULL,
  bins = 7,
  visible = FALSE,
  na_action = c("ignore", "warn", "error", "keep", "omit"),
  na_warn_prop = 0.2,
  return = c("list", "data"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble of health deficits (ideally binary/logical or
scaled to \eqn{[0,1]}). Non-binary numeric columns can be rescaled by di::di when
rescale = TRUE.}

\item{cols}{Character vector of deficit column names to use. If NULL (default),
all numeric columns are used except age (if supplied).}

\item{invert}{Character vector of column names whose values should be inverted
by di::di (e.g., where higher values indicate better health).}

\item{rescale}{Logical; if TRUE, non-binary columns will be rescaled to \eqn{[0,1]}
by di::di. Default TRUE.}

\item{age}{Optional name of the column holding age (used by di for plotting and
optional age-binned outputs; excluded from auto-selected cols).}

\item{rescale.custom}{Advanced argument passed through to di::di.
See di::di documentation for syntax.}

\item{rescale.avoid}{Advanced argument passed through to di::di;
see di::di documentation for syntax.}

\item{bins}{Integer; number of age bins for FI-by-age plots. Default 7.}

\item{visible}{Logical; if TRUE and age is provided, di will draw a plot (via
plot.di()). Default FALSE.}

\item{na_action}{One of \code{c("keep","omit","error","warn","ignore")}. Controls
handling of missing values in selected deficit columns. \code{"keep"} (and its
backward-compatible alias \code{"ignore"}) passes NAs through to di::di.
\code{"warn"} emits a warning and then keeps NAs (alias for \code{"keep"} with a
missingness warning). \code{"omit"} drops rows with any NA in selected deficits
before computing. \code{"error"} stops if any NA is detected.
Default \code{"ignore"} (retained for backward compatibility; equivalent to \code{"keep"}).}

\item{na_warn_prop}{Proportion in \eqn{[0,1]} above which a high-missingness warning
is emitted (per column) when na_action = "warn". Default 0.2.}

\item{return}{One of c("list","data"). "list" (default) returns the original
di::di result (backward compatible). "data" returns a tibble with one row
per individual, columns: di (the frailty index) plus the selected deficit
columns (post-capping if applied). Age is included if present.}

\item{verbose}{Logical; if TRUE, prints progress and a completion summary.}
}
\value{
\itemize{
\item If return = "list" (default): the object returned by di::di (typically a list
with di and columns).
\item If return = "data": a tibble with di and the selected columns.
}
}
\description{
Thin wrapper around the di package's di() that:
\itemize{
\item Validates inputs and arguments.
\item Coerces tibbles to base data.frames (for di()'s class checks).
\item Auto-selects numeric deficit columns when cols = NULL (excluding age if supplied).
\item Optionally scans for missing/out-of-range values with warnings or errors.
\item Provides step-by-step verbose output and a completion summary.
\item Optionally returns a tidy tibble instead of the original list.
}
}
\details{
Background
The Frailty Index (FI) is computed as the proportion of health deficits present
in an individual across a set of candidate deficits. The approach was introduced
and formalized by Rockwood and Mitnitski and subsequently standardized for
construction and reporting.
}
\examples{
# Minimal example (runs only if the 'di' package is installed)
if (requireNamespace("di", quietly = TRUE)) {
  df <- data.frame(
    age = c(70, 75, 80),
    d1 = c(0, 1, 1),
    d2 = c(0.2, 0.8, 1.0),
    d3 = c(TRUE, FALSE, TRUE)
  )
  # Auto-select numeric deficits; returns list (di, columns)
  res <- frailty_index(df, cols = NULL, age = "age", verbose = TRUE)
  # Tidy tibble return
  tb  <- frailty_index(df, cols = c("d1","d2","d3"), age = "age",
                       return = "data", verbose = TRUE)
}

}
\references{
\insertRef{mitnitski2001deficit}{HealthMarkers}
\insertRef{rockwood2007frailty}{HealthMarkers}
\insertRef{searle2008frailtyindex}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adipo_is.R
\name{adipo_is}
\alias{adipo_is}
\title{Adipose insulin sensitivity indices (QUICKI, VAI, LAP, TyG, TG/HDL, Belfiore)}
\usage{
adipo_is(
  data,
  col_map = NULL,
  normalize = "none",
  na_action = c("keep", "omit", "error"),
  verbose = TRUE,
  ...
)
}
\arguments{
\item{data}{Data frame or tibble with required columns mapped by \code{col_map}}

\item{col_map}{Named list mapping keys to columns: G0, I0, TG, HDL_c, FFA, waist, bmi}

\item{normalize}{One of c("none","z","inverse","range","robust"); default "none"}

\item{na_action}{One of c("keep","omit","error"); default "keep"}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, the list
of indices being computed, and a per-column results summary.}

\item{...}{Reserved}
}
\value{
A tibble with columns:
\code{Revised_QUICKI}, \code{VAI_Men_inv}, \code{VAI_Women_inv}, \code{TG_HDL_C_inv}, \code{TyG_inv},
\code{LAP_Men_inv}, \code{LAP_Women_inv}, \code{McAuley_index}, \code{Adipo_inv}, \code{Belfiore_inv_FFA}.
If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}), it
is prepended as the first output column.
}
\description{
Computes adipose-related insulin sensitivity/resistance indices from fasting inputs.
Expected input units (converted internally):
\itemize{
\item Glucose G0 mmol/L -> mg/dL (* 18)
\item Insulin I0 pmol/L -> muU/mL (/ 6)
\item TG mmol/L -> mg/dL (* 88.57)
\item HDL mmol/L -> mg/dL (* 38.67)
}

Reported indices (higher magnitude of negative "_inv" values implies worse adipose IR):
\itemize{
\item Revised_QUICKI = 1 / (log10(I0 (muU/mL)) + log10(G0 (mg/dL)) + log10(FFA (mmol/L)))
\item VAI (sex-specific; inverted as VAI_*_inv so larger negative = worse)
\item TG_HDL_C_inv = -(TG/HDL) in mg/dL
\item TyG_inv = -ln(TG (mg/dL) * G0 (mg/dL) / 2)
\item LAP (sex-specific; inverted)
\item McAuley_index = exp(2.63 - 0.28 ln(I0 (muU/mL)) - 0.31 ln(TG (mmol/L)))
\item Adipo_inv = -(FFA * I0 (muU/mL))
\item Belfiore_inv_FFA = - 2 / (I0 (muU/mL) * FFA + 1)
}

\strong{Inversion Note}: Most indices (VAI, LAP, TyG, TG/HDL, Adipo, Belfiore) are algebraically
inverted from their original insulin RESISTANCE definitions so that more negative values
consistently indicate worse adipose insulin sensitivity. Revised_QUICKI and McAuley_index
are retained in their original orientation (already sensitivity indices; higher = better).
See the vignette for detailed interpretation guidance.
}
\examples{
df <- tibble::tibble(
  G0 = c(5.2, 6.1),      # mmol/L
  I0 = c(60, 110),       # pmol/L
  TG = c(1.2, 1.8),      # mmol/L
  HDL_c = c(1.3, 1.0),   # mmol/L
  FFA = c(0.4, 0.6),     # mmol/L
  waist = c(85, 102),    # cm
  bmi = c(24, 31)        # kg/m^2
)
cm <- as.list(names(df)); names(cm) <- names(df)
out <- adipo_is(df, cm, verbose = FALSE, na_action = "keep")
}
\references{
\insertRef{katz2000quicki}{HealthMarkers}
\insertRef{amato2010vai}{HealthMarkers}
\insertRef{kahn2005lap}{HealthMarkers}
\insertRef{guerreroromero2018tyg}{HealthMarkers}
\insertRef{dobiasova2001atherogenic}{HealthMarkers}
\insertRef{belfiore1998insulin}{HealthMarkers}
\insertRef{raynaud1999fasting}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{.hm_global_precompute}
\alias{.hm_global_precompute}
\title{Derive "Tier 0" variables before any marker function runs.}
\usage{
.hm_global_precompute(data, col_map = NULL, verbose = FALSE)
}
\arguments{
\item{data}{data.frame after col_map materialization.}

\item{col_map}{Named list of key -> column mappings (used only to resolve
the physical column names when materialization has NOT been run yet).}

\item{verbose}{Logical; emit one-line summary per derived variable.}
}
\value{
A list:
\describe{
\item{data}{data.frame with new columns appended.}
\item{precomputed}{Character vector of variable names that were derived.}
}
}
\description{
Computes the following variables \strong{only when they are absent} from \code{data}:
\describe{
\item{BMI}{from \code{weight} (kg) and \code{height} (m or cm) via \code{weight / height_m^2}}
\item{glucose / G0}{bidirectional alias: whichever is absent is filled from the other}
\item{insulin / I0}{bidirectional alias (pmol/L <-> muU/mL unchanged -- same unit assumed)}
\item{eGFR}{CKD-EPI 2009 creatinine equation from \code{creatinine}, \code{age}, \code{sex}
(and optionally \code{race}). Written as \code{eGFR} only; downstream
functions that need \code{eGFR_cr} receive it via col_map.}
\item{UACR}{\code{urine_albumin} / \code{urine_creatinine} (both in mg/mmol or the
ratio already in mg/g -- no unit conversion; caller's responsibility)}
\item{LDL_c}{Friedewald: \code{TC - HDL_c - TG/2.2} (mmol/L); skipped if TG > 4.5}
}
}
\details{
col_map keys are respected: if the user has mapped e.g.\ \code{creatinine -> "Cr_serum"},
the materialized \code{data[["creatinine"]]} column (placed there by \code{.hm_build_col_map()})
is used transparently.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/alm_bmi_index.R
\name{alm_bmi_index}
\alias{alm_bmi_index}
\title{Appendicular Lean Mass to BMI Index}
\usage{
alm_bmi_index(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with ALM, BMI, and sex columns.}

\item{col_map}{Named list with:
\itemize{
\item alm: appendicular lean mass column name (kg)
\item bmi: body mass index column name (kg/m^2)
\item sex: sex column name ("Male"/"Female" or m/f; case-insensitive)
}}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if TRUE (default), emits progress messages.}
}
\value{
A tibble with:
\itemize{
\item alm_bmi_ratio   (numeric)
\item low_muscle_mass (logical; TRUE if below sex-specific cut-point;
NA if sex unknown or ratio NA)
}
}
\description{
Calculates the ratio of appendicular lean mass (ALM) to body mass index (BMI),
and flags low muscle mass based on FNIH Sarcopenia Project cut-points.
}
\details{
ALM/BMI reflects muscle mass relative to body size. FNIH cut-points:
\itemize{
\item Men:    ALM/BMI < 0.789
\item Women:  ALM/BMI < 0.512
}

ALM should be in kilograms and BMI in kg/m^2.
}
\examples{
df <- data.frame(ALM_kg = c(7.2, 5.8, 6.5), BMI = c(24, 28, 22),
                 Sex = c("male", "female", "male"))
alm_bmi_index(df)
}
\references{
\insertRef{mclean2014fnih}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nfl_marker.R
\name{nfl_marker}
\alias{nfl_marker}
\title{Neurofilament Light Chain (NfL) Biomarker}
\usage{
nfl_marker(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with an NfL concentration column.}

\item{col_map}{Named list with \code{nfl} indicating the NfL column name.}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if TRUE (default), emits progress messages.}
}
\value{
A tibble with one column: nfl_value (numeric; same units as input).
}
\description{
Incorporates a neurofilament light chain (NfL) measurement into the analysis pipeline.
Placeholder for future NfL-based computations; returns provided values with input checks.
}
\details{
NfL is released during neuroaxonal injury; elevated levels in CSF or blood indicate
neuroaxonal damage and typically increase with age and in neurological diseases.
Interpretation requires context-specific and age-adjusted references. This function
simply returns the input NfL values (assumed in a single matrix/fluid, e.g., plasma pg/mL)
without classification.
}
\examples{
df <- data.frame(NfL = c(8.5, 14.2, 22.1))
nfl_marker(df)

}
\references{
\insertRef{simren2022nfl}{HealthMarkers}
\insertRef{disanto2017nfl}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metabolic_risk_features.R
\name{metabolic_risk_features}
\alias{metabolic_risk_features}
\title{Calculate metabolic risk feature flags (pediatric-friendly thresholds)}
\usage{
metabolic_risk_features(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing at least these numeric columns:
\itemize{
\item chol_total, chol_ldl, chol_hdl, triglycerides (mmol/L)
\item age_year (years)
\item z_HOMA (standardized HOMA-IR)
\item glucose (mmol/L)
\item HbA1c (mmol/mol; IFCC units)
\item bp_sys_z, bp_dia_z (BP z-scores)
}}

\item{col_map}{Optional named list to map required keys to column names in \code{data}.
Keys: c("chol_total","chol_ldl","chol_hdl","triglycerides","age_year",
"z_HOMA","glucose","HbA1c","bp_sys_z","bp_dia_z"). Default NULL (use same names).}

\item{na_action}{One of c("keep","omit","error","ignore","warn") controlling missing-data policy.
\itemize{
\item "keep": keep NA; outputs become NA where inputs are NA.
\item "omit": drop rows with NA in any required input.
\item "error": abort if any required input contains NA.
\item "ignore"/"warn": aliases of "keep"; "warn" also emits missingness diagnostics.
}}

\item{na_warn_prop}{Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness warnings. Default 0.2.}

\item{verbose}{Logical; if TRUE, prints column mapping and computing messages.}
}
\value{
A tibble with four factor columns (levels c("0","1")):
\itemize{
\item dyslipidemia
\item insulin_resistance
\item hyperglycemia
\item hypertension
}
}
\description{
Compute four binary risk flags from routine clinical measures:
\itemize{
\item dyslipidemia
\item insulin_resistance
\item hyperglycemia (prediabetes-range glycemia)
\item hypertension (BP >=95th percentile via z > 1.64)
}
}
\details{
By default, behavior matches prior implementation: required columns are
validated, NA values are kept (propagate to outputs), no extreme-value
checks or capping are applied, and a tibble with 0/1 factor flags is returned.

Units and criteria (no automatic unit conversion):
\itemize{
\item Lipids (mmol/L): total cholesterol > 5.2 OR LDL-C > 3.4 OR HDL-C < 1.0 OR
triglycerides > 1.1 (age 0-9) OR > 1.5 (age 10-19) => dyslipidemia = 1.
Note: no TG cutoff is applied for adults aged >= 20 years.
\item Insulin resistance: z_HOMA > 1.28 (~=90th percentile) => insulin_resistance = 1.
z_HOMA is a within-sample or external z-score of HOMA-IR (see Matthews et al. 1985).
\item Hyperglycemia: fasting glucose in (5.6, 6.9) mmol/L OR HbA1c in (39, 47) mmol/mol
=> hyperglycemia = 1. Boundaries are EXCLUSIVE (open intervals); boundary values
(exactly 5.6 or 6.9 mmol/L; exactly 39 or 47 mmol/mol) are not flagged.
ADA criteria use inclusive lower bound (>= 5.6 mmol/L, >= 39 mmol/mol).
\item Hypertension: either BP z-score > 1.64 (~=95th percentile) for systolic or diastolic
=> hypertension = 1.
}
}
\note{
These flags are heuristic screening rules derived from published clinical
guidelines. They are not validated diagnostic criteria and should not replace
clinical judgment. The dyslipidemia and hypertension thresholds are designed
for pediatric populations (ages 0-19); for adults >= 20, only TC, LDL-C, and
HDL-C criteria contribute to the dyslipidemia flag.
}
\examples{
df <- data.frame(
  chol_total = c(5.2, 6.4), chol_ldl = c(3.2, 4.1), chol_hdl = c(1.3, 1.0),
  triglycerides = c(1.8, 2.5), age_year = c(45, 60), z_HOMA = c(0.5, 1.2),
  glucose = c(5.5, 6.8), HbA1c = c(38, 46), bp_sys_z = c(0.2, 1.1),
  bp_dia_z = c(0.1, 0.9)
)
metabolic_risk_features(df)
}
\references{
\insertRef{nhlbi2011lipids}{HealthMarkers}
\insertRef{ada2024standards}{HealthMarkers}
\insertRef{flynn2017bp}{HealthMarkers}
\insertRef{matthews1985homa}{HealthMarkers}
}
\seealso{
\code{\link[=liver_markers]{liver_markers()}}, \code{\link[=lipid_markers]{lipid_markers()}}, \code{\link[=kidney_failure_risk]{kidney_failure_risk()}}, \code{\link[=inflammatory_markers]{inflammatory_markers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrected_calcium.R
\name{corrected_calcium}
\alias{corrected_calcium}
\title{Albumin-Corrected Calcium}
\usage{
corrected_calcium(
  data,
  col_map = NULL,
  units = c("auto", "conventional", "si"),
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing serum calcium and albumin.}

\item{col_map}{Named list with \code{calcium} and \code{albumin} indicating column names.}

\item{units}{One of c("auto","conventional","si"). "auto" attempts unit detection.}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if TRUE (default), emits progress via hm_inform().}
}
\value{
A tibble with one column: corrected_calcium (numeric, in mg/dL for
conventional input or mmol/L for SI / auto-SI input).
}
\description{
Calculates the albumin-adjusted (corrected) serum calcium level, accounting for hypoalbuminemia,
using the Payne formula.
}
\details{
Payne formula (conventional units): Corrected Ca (mg/dL) = measured Ca (mg/dL) + 0.8 * (4.0 - albumin (g/dL)).
If inputs appear to be in SI units (calcium mmol/L, albumin g/L), they are converted to mg/dL and g/dL
(using 1 mmol/L ~= 4 mg/dL; 1 g/L = 0.1 g/dL) for the correction and converted back to mmol/L for output.
}
\examples{
df <- data.frame(Ca = c(2.3, 2.5, 2.1), Alb = c(38, 42, 30))
corrected_calcium(df)
}
\references{
\insertRef{payne1973}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sweat_markers.R
\name{sweat_markers}
\alias{sweat_markers}
\title{Calculate sweat-based ionic & metabolic markers}
\usage{
sweat_markers(
  data,
  col_map = NULL,
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2
)
}
\arguments{
\item{data}{A data.frame or tibble containing sweat assay and anthropometrics.}

\item{col_map}{Named list mapping required inputs (defaults assume same names):
\itemize{
\item sweat_chloride, sweat_Na, sweat_K, sweat_lactate,
weight_before, weight_after, duration, body_surface_area
}}

\item{verbose}{Logical; if \code{TRUE}, prints progress messages and a completion summary.}

\item{na_action}{One of \code{c("keep","omit","error")} for handling missing values in required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness diagnostics for required inputs (debug level). Default 0.2.}
}
\value{
A tibble with columns:
\code{sweat_chloride}, \code{Na_K_ratio}, \code{sweat_lactate}, \code{sweat_rate}
}
\description{
Computes:
\itemize{
\item sweat_chloride (mmol/L)
\item Na_K_ratio     (sweat Na+/K+)
\item sweat_lactate  (mmol/L)
\item sweat_rate     (L/m^2/h) from body mass loss per hour per m^2
}
}
\details{
Inputs are validated, missingness handled via \code{na_action}, safe divisions
are used to avoid Inf/NaN, and an optional extremes scan/cap is available.

Expected units:
\itemize{
\item sweat_chloride, sweat_Na, sweat_K: mmol/L
\item sweat_lactate: mmol/L
\item weight_before, weight_after: kg
\item duration: hours
\item body_surface_area: m^2
}
}
\note{
\code{sweat_chloride} and \code{sweat_lactate} are \strong{pass-through} columns; no formula
is applied. \code{Na_K_ratio} is a simple Na/K division. \code{sweat_rate} uses the
mass-loss method: (weight_before - weight_after) / duration / body_surface_area
(units: L/m\ifelse{html}{\out{&sup2;}}{^2}/h; assumes 1 kg \eqn{\approx} 1 L).
Dill & Costill 1974 describes haematocrit-based blood-volume change, not sweat
rate directly; cited here as background context only.
}
\examples{
df <- tibble::tibble(
  sweat_chloride    = 45,
  sweat_Na          = 55,
  sweat_K           = 5,
  sweat_lactate     = 4.8,
  weight_before     = 70.0,
  weight_after      = 69.5,
  duration          = 1.0,
  body_surface_area = 1.9
)
sweat_markers(df)

}
\references{
\insertRef{gibson1959sweat}{HealthMarkers} (pilocarpine sweat chloride test origin; background)
\insertRef{dill1974dehydration}{HealthMarkers} (dehydration and fluid loss context; background)
\insertRef{farrell2017cf}{HealthMarkers} (CF diagnostic sweat chloride cutoffs; background)
\insertRef{sawka2015hypo}{HealthMarkers} (sweat rate and hypohydration context; background)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{hm_result_summary}
\alias{hm_result_summary}
\title{Summarise a result tibble: count non-NA rows per output column}
\usage{
hm_result_summary(result, fn = NULL)
}
\arguments{
\item{result}{A tibble / data.frame returned by a HealthMarkers function.}

\item{fn}{Optional function name prefixed to the message.}
}
\description{
Produces a single string like:
"fn(): results: col_a 28/30, col_b 30/30, col_c 25/30"
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/health_markers.R
\name{all_health_markers}
\alias{all_health_markers}
\title{Compute all available HealthMarkers categories}
\usage{
all_health_markers(
  data,
  col_map,
  which = "all",
  include_insulin = TRUE,
  normalize = c("none", "z", "inverse", "range", "robust"),
  mode = c("both", "IS", "IR"),
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  id_col = NULL,
  return_input = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble.}

\item{col_map}{Named list for column mapping forwarded to underlying functions.
If \code{col_map} is \code{NULL} or an empty list, \code{all_health_markers()} calls
\code{\link[=hm_col_report]{hm_col_report()}} once at the top level to guess a column map from common
synonyms (for example \code{TG} vs \code{triglycerides}, \code{BMI} vs \code{bmi},
\code{HDL_c} vs \code{HDL}). The inferred \code{col_map} is then reused for all groups
that require it, and an error is thrown if required keys (e.g. \code{TG},
\code{HDL_c}, \code{LDL_c}, \code{TC}, \code{BMI}, \code{age}, \code{sex}) cannot be inferred.}

\item{which}{"all" or a vector of registry keys (see Details).}

\item{include_insulin}{Logical; include all_insulin_indices() first.}

\item{normalize}{One of c("none","z","inverse","range","robust").}

\item{mode}{One of c("both","IS","IR") passed to insulin indices.}

\item{verbose}{Logical.}

\item{na_action}{One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).}

\item{id_col}{Optional character string naming a column in \code{data} to include
in the returned output when \code{return_input = FALSE}. Ignored when
\code{return_input = TRUE}. Typical use: participant ID or sample barcode.}

\item{return_input}{Logical (default \code{TRUE}). When \code{TRUE} the original input
columns are retained in the output (current behaviour). When \code{FALSE} only
the newly computed marker columns are returned, plus \code{id_col} if supplied.
Set to \code{FALSE} to avoid carrying a large input data frame through the
pipeline — join back later with \code{cbind()} or \code{dplyr::left_join()} on \code{id_col}.}
}
\value{
Data frame. When \code{return_input = TRUE} (default): original columns
plus all derived markers. When \code{return_input = FALSE}: only the newly
computed columns (and \code{id_col} if specified).
}
\description{
Compute all available HealthMarkers categories
}
\details{
Common group names for \code{which} include:
\itemize{
\item \code{"lipid"}, \code{"liver"}, \code{"glycemic"}, \code{"mets"}, \code{"oxidative"}
\item \code{"bone"}, \code{"allostatic_load"}, \code{"nutrient"}, \code{"vitamin"}, \code{"vitamin_d_status"}
\item \code{"renal"}, \code{"ckd_stage"}, \code{"kidney_kfre"}
\item \code{"frailty_index"}, \code{"charlson"}, \code{"sarc_f"}
\item \code{"nfl"}, \code{"iAge"}, \code{"calcium_corrected"}, \code{"kyn_trp"}
}
}
\note{
For academic / clinical references tied to each derived marker or
index, consult the help pages of the source functions (e.g. ?allostatic_load,
?bone_markers, ?vitamin_markers, ?inflammatory_markers, etc.). This
aggregator provides integration only and does not restate citations.

Aggregator wrapper. See underlying function help pages for full references
across categories included by \code{which}.
}
\examples{
# Quick smoke-test (lipid group only, no insulin)
df <- data.frame(TC = 200, HDL_c = 50, TG = 150, LDL_c = 120)
all_health_markers(df, col_map = list(), which = "lipid",
                   include_insulin = FALSE, normalize = "none",
                   verbose = FALSE, na_action = "keep")

\donttest{
# Lipid + liver groups
df <- data.frame(
  TC = 200, HDL_c = 50, TG = 150, LDL_c = 120,
  ALT = 30, AST = 20, BMI = 25
)
all_health_markers(df, col_map = list(), which = c("lipid","liver"),
                   include_insulin = FALSE, normalize = "none", mode = "both",
                   verbose = FALSE, na_action = "keep")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vitamin_d_status.R
\name{vitamin_d_status}
\alias{vitamin_d_status}
\title{Vitamin D Status Category}
\usage{
vitamin_d_status(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with a 25-hydroxyvitamin D concentration column.}

\item{col_map}{A named list with \code{vitamin_d} giving the column name in \code{data} for 25(OH)D.}

\item{na_action}{One of c("keep","omit","error","ignore","warn").
\itemize{
\item keep/ignore: compute and propagate NA
\item omit: drop rows with NA in required input
\item error: abort if required input contains NA
\item warn: like keep, but emit missingness warnings
}}

\item{verbose}{Logical; if TRUE, emits progress via rlang::inform.}
}
\value{
A tibble with one column: vitamin_d_status (ordered factor with levels "Deficient","Insufficient","Sufficient").
}
\description{
Categorizes vitamin D status based on serum 25-hydroxyvitamin D (25(OH)D) levels.
}
\details{
Serum 25(OH)D is the standard biomarker for vitamin D status. This function classifies
levels (assumed in ng/mL) into categories:
\itemize{
\item Deficient (< 20 ng/mL)
\item Insufficient (20-29 ng/mL)
\item Sufficient (>= 30 ng/mL)
}

Note: Ensure input units are ng/mL. If values appear extremely high (e.g., median > 150),
they might be in nmol/L (divide by 2.5 to convert to ng/mL).
}
\examples{
df <- data.frame(VitD = c(18, 45, 72))
vitamin_d_status(df)

}
\references{
\insertRef{iomm2011calciumvitd}{HealthMarkers}
\insertRef{holick2011vitd}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_missing.R
\name{impute_missing}
\alias{impute_missing}
\title{Impute missing values in a data.frame or tibble (simple, column-wise)}
\usage{
impute_missing(
  data,
  method = c("mean", "median", "zero", "constant"),
  cols = NULL,
  constant = NULL,
  na_warn_prop = 0.2,
  verbose = FALSE
)
}
\arguments{
\item{data}{A data.frame or tibble containing missing values.}

\item{method}{Character; one of c("mean","median","zero","constant").}

\item{cols}{Optional character vector of column names to impute. Defaults to all
numeric columns in \code{data} that contain at least one NA.}

\item{constant}{Numeric; single value to use when \code{method = "constant"}.}

\item{na_warn_prop}{Numeric in \eqn{[0,1]}; threshold for high-missingness warnings per column.
Default 0.2.}

\item{verbose}{Logical; if TRUE, prints progress and a completion summary. Default FALSE.}
}
\value{
A data.frame/tibble of the same dimensions as \code{data}, with the specified
columns' missing values imputed.
}
\description{
Performs deterministic, per-column imputation for numeric variables:
\itemize{
\item "mean": replace NAs with the column mean
\item "median": replace NAs with the column median
\item "zero": replace NAs with 0
\item "constant": replace NAs with the single value given in \code{constant}
}
}
\details{
Non-numeric columns are left untouched. If \code{cols = NULL}, all numeric columns
that have at least one NA are selected automatically. NA positions are the only
values modified; non-NA entries are preserved as-is.

Quality checks:
\itemize{
\item Warns for high-missingness columns (>= \code{na_warn_prop}).
\item Warns and skips imputation when a column has no non-NA values (mean/median undefined).
\item Coerces only numeric columns; non-numerics in \code{cols} are skipped with a warning.
}
}
\examples{
df <- tibble::tibble(a = c(1, NA, 3), b = c(NA, NA, 2), c = letters[1:3])
impute_missing(df, method = "mean")
impute_missing(df, method = "median", verbose = TRUE)
impute_missing(df, method = "constant", constant = -1, cols = "a")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_missing.R
\name{impute_missforest}
\alias{impute_missforest}
\title{Impute missing values via random forest (missForest)}
\usage{
impute_missforest(data, ntree = 100, cols = NULL, verbose = FALSE, ...)
}
\arguments{
\item{data}{A data.frame or tibble containing missing values.}

\item{ntree}{Integer; number of trees to grow in each forest (passed to missForest). Default 100.}

\item{cols}{Optional character vector of numeric columns to impute.
Defaults to all numeric columns with at least one NA.}

\item{verbose}{Logical; if TRUE, prints progress and a completion summary. Default FALSE.}

\item{...}{Additional arguments passed to missForest::missForest().}
}
\value{
A data.frame/tibble with numeric columns imputed by missForest (or mean fallback).
}
\description{
Wraps missForest to impute numeric columns using non-parametric random forests.
Requires the missForest package to be installed (Suggests). Non-numeric columns
are untouched. If no numeric columns contain NAs, \code{data} is returned unchanged.
}
\details{
Notes:
\itemize{
\item missForest uses iterative RF training; it can be slow on wide/high-NA data.
\item Errors (e.g., insufficient unique values) are caught and a deterministic mean
imputation fallback is applied to the selected numeric columns.
}
}
\examples{
\donttest{
if (requireNamespace("missForest", quietly = TRUE)) {
  df <- tibble::tibble(a = c(1, NA, 3), b = c(2, 4, NA), c = 5)
  impute_missforest(df, ntree = 50, verbose = TRUE)
}
}

}
\references{
\insertRef{stekhoven2012missforest}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/urine_markers.R
\name{urine_markers}
\alias{urine_markers}
\title{Calculate urine-only renal and tubular markers (research-ready)}
\usage{
urine_markers(
  data,
  col_map = NULL,
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2
)
}
\arguments{
\item{data}{A data.frame or tibble with at least urine_albumin and urine_creatinine.}

\item{col_map}{Optional named list mapping canonical keys (e.g., \code{urine_albumin},
\code{urine_creatinine}) to actual column names in \code{data}. If \code{NULL}, column names
are inferred automatically.}

\item{verbose}{Logical; if \code{TRUE}, prints progress messages and a completion summary. Default FALSE.}

\item{na_action}{One of \code{c("keep","omit","error")} for handling missing values in required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness warnings for required inputs. Default 0.2.}
}
\value{
A tibble with columns:
UACR, albuminuria_stage, microalbuminuria, UPCR, U_Na_K_ratio,
NGAL_per_gCr, KIM1_per_gCr, NAG_per_gCr, Beta2Micro_per_gCr,
A1Micro_per_gCr, IL18_per_gCr, L_FABP_per_gCr
}
\description{
Computes (urine-only):
\itemize{
\item UACR (Albumin-to-Creatinine Ratio, mg/g)
\item albuminuria_stage (KDIGO A1/A2/A3 by UACR)
\item microalbuminuria flag ("normal" vs "micro")
\item UPCR (Urine Protein-to-Creatinine Ratio, mg/g; if urine_protein available)
\item U_Na_K_ratio (urine Na+/K+; if urine_Na and urine_K available)
\item Creatinine-normalized tubular markers (if present, per g creatinine):
NGAL_per_gCr, KIM1_per_gCr, NAG_per_gCr, Beta2Micro_per_gCr,
A1Micro_per_gCr, IL18_per_gCr, L_FABP_per_gCr
}
}
\details{
Inputs are validated, missingness handled via \code{na_action}, divisions are
safeguarded (Inf/NaN -> NA) with a consolidated zero-denominator warning,
and an optional extremes scan/cap is available.

Expected units:
\itemize{
\item urine_albumin: mg/L
\item urine_protein: mg/L (optional)
\item urine_creatinine: mg/dL
\item urine_Na, urine_K: mmol/L (optional)
\item Optional tubular markers above assumed mg/L when normalized per g creatinine
}
}
\note{
UACR formula: albumin (mg/L) / creatinine (g/L) = albumin (mg/L) \eqn{\times} 100 /
creatinine (mg/dL). UPCR and per-gCr tubular markers use the same creatinine
denominator: \verb{gCr_den = creatinine (mg/dL) \\times 0.01} (= g/L).
Tubular markers (NGAL, KIM-1, NAG, Beta-2-microglobulin, alpha-1-microglobulin,
IL-18, L-FABP) are \strong{pass-through} columns normalised per g creatinine; no
formula other than creatinine adjustment is applied.
}
\examples{
df <- tibble::tibble(
  urine_albumin    = 30,
  urine_creatinine = 1.2,
  serum_creatinine = 0.9,
  plasma_Na        = 140,
  urine_Na         = 100,
  age              = 55,
  sex              = 2,
  urine_protein    = 150
)
urine_markers(df)
}
\references{
\insertRef{mogensen1984uacr}{HealthMarkers}
\insertRef{ginsberg1983upcr}{HealthMarkers}
\insertRef{kdigo2012ckd}{HealthMarkers} (albuminuria staging A1–A3 UACR cutoffs)
\insertRef{dezeeuw2006uacr}{HealthMarkers} (prognostic UACR validation; background)
\insertRef{ichimura2004kim1}{HealthMarkers} (KIM-1 tubular biomarker; pass-through normalization, background)
\insertRef{portilla2008lfabp}{HealthMarkers} (L-FABP tubular biomarker; pass-through normalization, background)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/health_summary.R
\name{health_summary}
\alias{health_summary}
\title{Summarize selected numeric marker columns}
\usage{
health_summary(x, cols = NULL)
}
\arguments{
\item{x}{A data.frame or tibble.}

\item{cols}{Optional character vector of column names to summarize. If NULL, all numeric columns are summarized.}
}
\value{
A tibble with one row per summarized column.
}
\description{
Lightweight summary for numeric columns: n, n_na, mean, sd, median, p25, p75.
}
\examples{
df <- data.frame(a = c(1, 2, NA), b = c(3, 4, 5), c = factor("x"))
health_summary(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/obesity_metrics.R
\name{obesity_indices}
\alias{obesity_indices}
\title{Compute anthropometric obesity & adiposity indices}
\usage{
obesity_indices(
  data,
  weight,
  height,
  waist,
  hip,
  sex = NULL,
  weight_unit = c("kg", "lb"),
  height_unit = c("cm", "m"),
  adjust_WHR = FALSE,
  include_RFM = FALSE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing the input columns.}

\item{weight}{Unquoted column name for weight.}

\item{height}{Unquoted column name for height.}

\item{waist}{Unquoted column name for waist circumference.}

\item{hip}{Unquoted column name for hip circumference.}

\item{sex}{(Optional) Unquoted column name for sex, coded 0=male, 1=female; required if include_RFM=TRUE.}

\item{weight_unit}{One of c("kg","lb"); if "lb", converts weight to kg by *0.45359237.}

\item{height_unit}{One of c("cm","m"); if "cm", converts height to metres by /100.}

\item{adjust_WHR}{Logical; if TRUE, adds a column WHRadjBMI as residuals from WHR ~ BMI.}

\item{include_RFM}{Logical; if TRUE, computes Relative Fat Mass (requires sex column).}

\item{na_action}{One of c("keep","omit","error") for handling NA in required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness warnings for required inputs. Default 0.2.}

\item{verbose}{Logical; if TRUE, prints column mapping and computing messages.}
}
\value{
A tibble with only the computed indices (slim output):
\itemize{
\item weight_kg, height_m (unit-normalised intermediates),
\item BMI, BMI_cat,
\item WHR, WHRadjBMI (if \code{adjust_WHR = TRUE}),
\item waist_to_height_ratio, waist_to_BMI_ratio, weight_to_height_ratio,
\item AVI, BAI, ABSI, BRI, CI,
\item RFM (if \code{include_RFM = TRUE}).
}
}
\description{
Calculates a comprehensive set of body shape and adiposity indices:
\itemize{
\item BMI and WHO BMI categories
\item Waist-to-hip ratio (WHR) and optional WHR adjusted for BMI (WHRadjBMI)
\item Waist-to-height ratio (WHtR)
\item Abdominal Volume Index (AVI)
\item Body Adiposity Index (BAI)
\item A Body Shape Index (ABSI)
\item Body Roundness Index (BRI)
\item Conicity Index (CI)
\item (Optional) Relative Fat Mass (RFM)
}
}
\details{
Units assumed (no automatic conversion beyond the specified weight/height options):
\itemize{
\item weight: kg (or lb if weight_unit = "lb")
\item height: m (or cm if height_unit = "cm")
\item waist, hip: cm
}

Note: WHtR, ABSI, BRI, CI, and RFM all require waist in the same unit as
height (metres). The function converts waist internally (waist_cm / 100)
for these five indices; users should always supply waist in cm.
\itemize{
\item sex: 0 = male, 1 = female (only required if include_RFM = TRUE)
}
}
\examples{
library(tibble)
df <- tibble(
  wt     = c(70, 80),  # kg
  ht     = c(175, 165),# cm
  waist  = c(80, 90),  # cm
  hip    = c(100, 95), # cm
  sex    = c(0, 1)
)
obesity_indices(
  df,
  weight       = wt,
  height       = ht,
  waist        = waist,
  hip          = hip,
  sex          = sex,
  weight_unit  = "kg",
  height_unit  = "cm",
  adjust_WHR   = TRUE,
  include_RFM  = TRUE,
  verbose      = TRUE
)
}
\references{
\insertRef{quetelet1842}{HealthMarkers}
\insertRef{who1995physicalstatus}{HealthMarkers}
\insertRef{guerrero1999avi}{HealthMarkers}
\insertRef{bergman2011bai}{HealthMarkers}
\insertRef{krakauer2012absi}{HealthMarkers}
\insertRef{thomas2013bri}{HealthMarkers}
\insertRef{valdez1991ci}{HealthMarkers}
\insertRef{woolcott2018rfm}{HealthMarkers}
\insertRef{calle1999bmi}{HealthMarkers}
\insertRef{freedman2012bai}{HealthMarkers}
\insertRef{he2013absi}{HealthMarkers}
\insertRef{maessen2014bri}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zzz-validate_inputs.R
\name{validate_inputs}
\alias{validate_inputs}
\title{Validate required inputs for a calling function}
\usage{
validate_inputs(data, col_map, fun_name, required_keys = NULL)
}
\arguments{
\item{data}{data.frame or tibble}

\item{col_map}{named list mapping keys to column names}

\item{fun_name}{character scalar naming the calling function (e.g., "lipid_markers").
Used to look up built-in required keys when \code{required_keys} is not supplied.}

\item{required_keys}{optional character vector of required col_map keys. When
supplied, this takes precedence over the \code{fun_name} built-in lookup, making
the function useful for any caller regardless of \code{fun_name}.}
}
\value{
invisibly TRUE on success; otherwise aborts
}
\description{
Ensures required keys exist in \code{col_map} and have non-empty mappings.
Missing keys are reported in a stable order aligned with tests.
}
\examples{
df <- data.frame(TG = c(1.5, 2.0), HDL_c = c(1.2, 1.0),
  LDL_c = c(2.0, 2.5), TC = c(4.5, 5.0))
# Using built-in lookup
validate_inputs(df,
  list(TG = "TG", HDL_c = "HDL_c", LDL_c = "LDL_c", TC = "TC"),
  fun_name = "lipid_markers")
# Using explicit required_keys (works for any function)
validate_inputs(df,
  list(TG = "TG", HDL_c = "HDL_c"),
  fun_name = "my_function",
  required_keys = c("TG", "HDL_c"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/renal_markers.R
\name{renal_markers}
\alias{renal_markers}
\title{Calculate a Suite of Renal Function, Injury, and Excretion Markers}
\usage{
renal_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with renal lab data.}

\item{col_map}{Named list mapping:
\itemize{
\item creatinine -> serum creatinine (mg/dL)
\item age -> age (years)
\item sex -> sex indicator (1 = male, 0 = female). Also accepts "male"/"female".
\item race -> race ("white", "black", or "other"). Also accepts common aliases.
\item BUN -> blood urea nitrogen (mg/dL)
\item optional cystatin_C -> serum cystatin C (mg/L)
\item optional urea_serum -> serum urea (mg/dL)
\item optional creatinine_urine -> urine creatinine (mg/dL)
\item optional urea_urine -> urine urea (mg/dL)
\item optional NGAL, KIM1, NAG, beta2_micro, IL18, L_FABP -> urine injury markers
}}

\item{na_action}{One of \code{c("keep","omit","error")} for handling missing values in
required inputs. Default "keep".}

\item{na_warn_prop}{Proportion (0-1) threshold for high-missingness diagnostics.
Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} (default), prints step-by-step progress
including column mapping, optional input availability, physiological range
information (informational only, values are not altered), the list of
markers being computed, and a per-column results summary.}
}
\value{
A tibble with computed renal markers:
\code{eGFR_cr}, \code{eGFR_cys}, \code{eGFR_combined}, \code{BUN_Cr_ratio}, \code{FE_Urea},
\code{NGAL}, \code{KIM1}, \code{NAG}, \code{Beta2Micro}, \code{IL18}, \code{L_FABP}.
If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}),
it is prepended as the first output column.
}
\description{
Given routine blood and urine assays, \code{renal_markers()} computes:
\itemize{
\item eGFR_cr: CKD-EPI creatinine equation (2009 variant; race factor retained to preserve prior behavior)
\item eGFR_cys: CKD-EPI cystatin C equation (if \code{cystatin_C} provided)
\item eGFR_combined: CKD-EPI combined creatinine+cystatin C (if both provided)
\item BUN_Cr_ratio: Blood urea nitrogen / serum creatinine
\item FE_Urea: Fractional excretion of urea (\%)
\item NGAL, KIM1, NAG, Beta2Micro, IL18, L_FABP: pass-through urinary injury markers (if mapped)
}
}
\details{
Robust validation is applied, including NA handling (\code{na_action}), high-missingness
diagnostics, safe divisions with a consolidated zero-denominator warning, and an
optional input extremes scan/cap. New arguments are appended for backward compatibility.

Expected units (no automatic conversion performed):
\itemize{
\item creatinine (serum): mg/dL
\item cystatin C (serum): mg/L
\item BUN (serum): mg/dL
\item urea_serum, urea_urine: mg/dL
\item creatinine_urine: mg/dL
}
}
\note{
\code{eGFR_cr} uses the 2009 CKD-EPI creatinine equation (Levey et al. 2009) with the
Black-race multiplier (\eqn{\times 1.159}) retained. The 2021 race-free CKD-EPI
equations (Inker et al., NEJM 2021) are not yet implemented; the \code{race} input
is accepted for forward compatibility and used only for the 2009 race factor.
\code{eGFR_cys} and \code{eGFR_combined} use Inker et al. (2012); note that
\code{eGFR_combined} applies its own sex (\eqn{\times 1.008} female) and race
(\eqn{\times 1.145} Black) multipliers, which differ from those of \code{eGFR_cr}.
\code{NGAL}, \code{KIM1}, \code{NAG}, \code{Beta2Micro}, \code{IL18}, and \code{L_FABP} are
\strong{pass-through} columns — values are returned as-is with no formula applied.
}
\examples{
df <- tibble::tibble(Cr = 1.0, Age = 40, Sex = 1, Race = "white", BUN = 14)
cm <- list(creatinine = "Cr", age = "Age", sex = "Sex", race = "Race", BUN = "BUN")
renal_markers(df, cm)

}
\references{
\insertRef{levey2009ckdepi}{HealthMarkers}
\insertRef{inker2012cys}{HealthMarkers}
\insertRef{waikar2009feu}{HealthMarkers} (FE_Urea formula source; bib content: Kaplan and Kohn 1992)
\insertRef{parikh2011ngal}{HealthMarkers} (clinical context; NGAL is a pass-through biomarker)
\insertRef{vaidya2010kim1}{HealthMarkers} (KIM-1 biomarker qualification; pass-through)
\insertRef{portilla2008lfabp}{HealthMarkers} (L-FABP as AKI biomarker; pass-through)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/marker_summary.R
\name{marker_summary}
\alias{marker_summary}
\title{Summarize marker outputs}
\usage{
marker_summary(x, verbose = FALSE)
}
\arguments{
\item{x}{Data frame returned by marker functions}

\item{verbose}{Logical; if TRUE, prints progress messages}
}
\value{
Tibble with columns: variable, mean, sd, iqr
}
\description{
Computes simple summaries (mean, SD, IQR) for numeric columns.
}
\examples{
df <- data.frame(glucose = c(5.5, 6.1, 4.9), insulin = c(60, 88, 55),
                 bmi = c(24, 27, 22))
marker_summary(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{asrs_score}
\alias{asrs_score}
\title{Adult ADHD Self-Report Scale scoring}
\usage{
asrs_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "ASRS",
  partA_items = sprintf("asrs_\%02d", 1:6),
  partA_thresholds = c(3, 3, 3, 4, 4, 4),
  partA_cutoff = 4,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{partA_items}{Vector of Part A item IDs.}

\item{partA_thresholds}{Numeric thresholds applied to Part A items.}

\item{partA_cutoff}{Count threshold for Part A positivity.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{ASRS_total}, \code{ASRS_partA_count}, \code{ASRS_partA_positive}. Input columns are not included.
}
\description{
Adult ADHD Self-Report Scale scoring
}
\note{
Items are expected on a 0--4 scale (Never=0, Rarely=1, Sometimes=2,
Often=3, Very Often=4). The default \code{partA_thresholds} follow the
official WHO ASRS v1.1 guide: items 1--3 are positive at \eqn{\geq 3}
(Often or Very Often) and items 4--6 at \eqn{\geq 4} (Very Often only).
}
\examples{
df <- data.frame(matrix(2, nrow = 1, ncol = 18))
names(df) <- sprintf("asrs_\%02d", 1:18)
asrs_score(df)
}
\references{
\insertRef{adler2006asrs}{HealthMarkers}
\insertRef{kessler2007asrs}{HealthMarkers} (prevalence study; ASRS used as instrument)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{hm_fmt_col_map}
\alias{hm_fmt_col_map}
\title{Format a column-map resolution line for verbose output}
\usage{
hm_fmt_col_map(col_map, fn = NULL)
}
\arguments{
\item{col_map}{Named list of key -> column-name mappings (only the keys
actually used by the function need to be supplied).}

\item{fn}{Optional function name prefixed to the message.}
}
\description{
Produces a single string like:
"fn(): column map: G0 -> 'col_a', I0 -> 'col_b', ..."
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{mdq_score}
\alias{mdq_score}
\title{Mood Disorder Questionnaire scoring}
\usage{
mdq_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  prefix = "MDQ",
  symptom_cutoff = 7,
  require_clustering = TRUE,
  require_impairment = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{prefix}{Prefix for output column names.}

\item{symptom_cutoff}{Minimum symptom count for a positive screen.}

\item{require_clustering}{Require clustering item == 1 to be positive.}

\item{require_impairment}{Require impairment item == 1 to be positive.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{MDQ_symptom_count}, \code{MDQ_clustering}, \code{MDQ_impairment}, \code{MDQ_positive_screen}. Input columns are not included.
}
\description{
Mood Disorder Questionnaire scoring
}
\note{
The impairment column is expected to be \strong{binary} (1 = impaired, 0 = not).
The original MDQ impairment question uses a 4-category scale
(1 = no problem, 2 = minor, 3 = moderate, 4 = serious); if the raw
4-category response is passed, the caller must recode it to binary
(e.g., \code{impair_binary = as.integer(impair_raw >= 3)}) before scoring.
}
\examples{
df <- data.frame(matrix(0, nrow = 1, ncol = 13))
names(df) <- sprintf("mdq_\%02d", 1:13)
df$mdq_cluster <- 1; df$mdq_impair <- 1
mdq_score(df)
}
\references{
\insertRef{hirschfeld2000mdq}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kidney_kfre.R
\name{kidney_failure_risk}
\alias{kidney_failure_risk}
\title{Kidney Failure Risk Equation (KFRE, 2- and 5-year risk)}
\usage{
kidney_failure_risk(
  data,
  col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR"),
  na_action = c("keep", "error", "omit", "warn"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing at least the columns mapped in \code{col_map}.}

\item{col_map}{Named list mapping:
\itemize{
\item \code{age}  -> age in years
\item \code{sex}  -> sex code (1 = male, 2 = female)
\item \code{eGFR} -> estimated GFR (mL/min/1.73 m^2)
\item \code{UACR} -> urine albumin-to-creatinine ratio (mg/g)
}}

\item{na_action}{One of c("keep","error","omit","warn"). Default "keep" to preserve previous behavior:
\itemize{
\item "keep": propagate NA/NaN through logs and outputs.
\item "error": abort if any required input contains missing values.
\item "omit": drop rows with NA in required inputs before computation.
\item "warn": like "keep" but emits high-missingness warnings.
}}

\item{na_warn_prop}{Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness warnings. Default 0.2.}

\item{verbose}{Logical; if TRUE, prints stepwise messages and a completion summary. Default TRUE.}
}
\value{
A tibble with:
\itemize{
\item \code{KFRE_2yr} risk (0-1) at 2 years
\item \code{KFRE_5yr} risk (0-1) at 5 years
}
}
\description{
Compute 2- and 5-year risk of end-stage kidney disease using the original
4-variable KFRE (Tangri et al., 2011) with optional data-quality diagnostics,
and verbose progress reporting.
}
\details{
This function preserves prior behavior by default:
\itemize{
\item Inputs are taken as-is; NA values propagate to outputs (na_action = "keep").
\item No capping or out-of-range checks are applied.
}

Units (no automatic conversion):
\itemize{
\item age: years; sex: 1 = male, 2 = female
\item eGFR: mL/min/1.73 m^2
\item UACR: mg/g (albumin-to-creatinine ratio)
}

Details
\itemize{
\item Prognostic index: \verb{PI = 0.220 x log(age) + (-0.556) x log(eGFR) + 0.451 x log(UACR) + 0.391 x (male)}
where male = 1 if sex == 1, else 0.
\item Baseline survival: S0(2y) = 0.934, S0(5y) = 0.881 (Tangri 2011).
\item Risks: KFRE_t = 1 - (S0_t ^ exp(PI)).
\item The 2016 JAMA study provides a large, multinational validation of the KFRE in humans.
\item This implementation computes the original 4-variable linear predictor and does not
apply recalibration or alternative coefficient sets.
}
}
\examples{
library(tibble)
df <- tibble(
  age  = c(65, 72),
  sex  = c(1, 2),          # 1 = male, 2 = female
  eGFR = c(45, 22),        # mL/min/1.73 m^2
  UACR = c(300, 1200)      # mg/g
)
# Default behavior (NA propagate, no extreme checks)
kidney_failure_risk(
  data = df,
  col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR")
)

# With verbose output
\donttest{
kidney_failure_risk(
  data = df,
  col_map = list(age = "age", sex = "sex", eGFR = "eGFR", UACR = "UACR"),
  verbose = TRUE
)
}

}
\references{
\insertRef{tangri2011kfre}{HealthMarkers}
\insertRef{tangri2016meta}{HealthMarkers}
}
\seealso{
\code{\link[=inflammatory_markers]{inflammatory_markers()}}, \code{\link[=iAge]{iAge()}}, \code{\link[=impute_missing]{impute_missing()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_risk}
\alias{cvd_risk}
\title{Compute cardiovascular risk or marker by selected model}
\usage{
cvd_risk(
  data,
  model = c("ALL", "ASCVD", "QRISK3", "Stroke", "RiskScorescvd", "AIP", "LDL_PN"),
  year = 10,
  ...,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame required by your chosen model.}

\item{model}{One of:
\itemize{
\item "ALL"
\item Risk calculators: "ASCVD","QRISK3","Stroke","RiskScorescvd"
\item Lipid markers: "AIP","LDL_PN"
}}

\item{year}{Risk horizon (10 or 30) for applicable models; ignored for lipid markers.}

\item{...}{Forwarded to underlying wrappers (e.g., col_map, na_action).}

\item{verbose}{Logical; if TRUE, prints progress (legacy; messages now routed via hm_inform).}
}
\value{
A tibble.
}
\description{
Dispatch to the appropriate risk or marker function, or run all of them.
Includes basic argument validation and robust fallback to NA rows if
individual calculators fail.
}
\examples{
df <- data.frame(TG = c(150, 200), HDL_c = c(50, 40))
cvd_risk(df, model = "AIP")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hormone_markers.R
\name{hormone_markers}
\alias{hormone_markers}
\title{Compute a suite of hormone ratio markers with QA and verbose summaries}
\usage{
hormone_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "warn", "ignore"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame or tibble with mapped hormone inputs.}

\item{col_map}{Named list mapping the required keys to column names:
total_testosterone, SHBG, LH, FSH, estradiol, progesterone, free_T3, free_T4,
aldosterone, renin, insulin, glucagon, GH, IGF1, prolactin, cortisol_0, cortisol_30.}

\item{na_action}{One of \code{c("keep","omit","error","warn","ignore")}. \code{"keep"}/\code{"ignore"} leave NAs;
\code{"omit"} drops rows with any NA in used inputs; \code{"error"} aborts;
\code{"warn"} also warns about high missingness.}

\item{na_warn_prop}{Proportion in \eqn{[0,1]} for high-missingness warnings
when \code{na_action = "warn"}. Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, input
availability, inference notes, physiological range information
(informational only, values not altered), computing markers, and a
per-column results summary.}
}
\value{
Tibble with one column per computable ratio. If an ID column is
detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}), it is prepended
as the first output column.
}
\description{
Ratios computed:
\itemize{
\item FAI = (total_testosterone / SHBG) * 100
\item LH_FSH = LH / FSH
\item E2_P = estradiol / progesterone
\item T3_T4 = free_T3 / free_T4
\item ARR = aldosterone / renin
\item Ins_Glu = insulin / glucagon
\item GH_IGF1 = GH / IGF1
\item PRL_T = prolactin / total_testosterone
\item CAR_slope = (cortisol_30 - cortisol_0) / 30
}
}
\details{
Some inputs may be inferred when missing (for example, \code{free_T3} from
\code{TSH} + \code{free_T4}, or \code{GH} from \code{IGF1}) using internal heuristics. These
inferred values are intended for exploratory feature engineering only and
must not be treated as clinical substitutes for directly measured assays.
Ratios such as FAI, ARR, and CAR_slope have established literature usage;
several other outputs are simple arithmetic composites included for
feature-engineering convenience and may not have a single canonical
derivation paper.
}
\examples{
df <- data.frame(
  TT = c(15, 12), SHBG = c(40, 35), LH = c(5, 6), FSH = c(4, 5),
  E2 = c(100, 120), Prog = c(0.5, 0.6), fT3 = c(4.5, 4.2),
  fT4 = c(15, 14), Aldo = c(200, 180), Renin = c(10, 12),
  Ins = c(60, 70), Gluc = c(8, 9), GH = c(1.2, 1.0),
  IGF1 = c(180, 160), Prl = c(10, 12), Cort0 = c(400, 380),
  Cort30 = c(600, 580)
)
col_map <- list(
  total_testosterone = "TT", SHBG = "SHBG", LH = "LH", FSH = "FSH",
  estradiol = "E2", progesterone = "Prog", free_T3 = "fT3",
  free_T4 = "fT4", aldosterone = "Aldo", renin = "Renin",
  insulin = "Ins", glucagon = "Gluc", GH = "GH", IGF1 = "IGF1",
  prolactin = "Prl", cortisol_0 = "Cort0", cortisol_30 = "Cort30"
)
hormone_markers(df, col_map = col_map)
}
\references{
\insertRef{sowers2009fai}{HealthMarkers}
\insertRef{funder2016primaryaldosteronism}{HealthMarkers}
\insertRef{clow2004car}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{psych_med_flags}
\alias{psych_med_flags}
\title{Psychiatric medication flags aggregator}
\usage{
psych_med_flags(
  data,
  col_map = list(),
  prefix = "med",
  na_action = c("keep", "omit", "error")
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list \code{med} mapping medication classes (e.g., ssri, snri) to columns of boolean/numeric flags.}

\item{prefix}{Prefix for output flag columns.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}
}
\value{
A tibble of flag columns only: \code{med_any_psych}, \code{med_count}. Input columns are not included.
}
\description{
Psychiatric medication flags aggregator
}
\examples{
df <- data.frame(med_ssri = c(1, 0), med_antipsychotic = c(0, 1))
cm <- list(med = list(
  ssri = "med_ssri",
  antipsychotic = "med_antipsychotic"
))
psych_med_flags(df, col_map = cm)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/impute_missing.R
\name{impute_mice}
\alias{impute_mice}
\title{Impute missing values via Multiple Imputation (mice)}
\usage{
impute_mice(data, m = 5, cols = NULL, verbose = FALSE, ...)
}
\arguments{
\item{data}{A data.frame or tibble containing missing values.}

\item{m}{Integer; number of imputations to run (passed to mice). Default 5.}

\item{cols}{Optional character vector of numeric columns to impute.
Defaults to all numeric columns with at least one NA.}

\item{verbose}{Logical; if TRUE, prints progress and a completion summary. Default FALSE.}

\item{...}{Additional arguments passed to mice::mice().}
}
\value{
A data.frame/tibble with numeric columns imputed by mice.
}
\description{
Wraps mice to impute only numeric columns; non-numeric columns are untouched.
Requires the mice package to be installed (Suggests). If no numeric columns
contain NAs, \code{data} is returned unchanged.
}
\details{
Notes:
\itemize{
\item At least two numeric columns are typically needed by mice to borrow strength.
\item This function runs \code{m} imputations and returns the first completed dataset.
\item Messages from mice are suppressed; use \code{verbose = TRUE} here for high-level progress.
}
}
\examples{
\donttest{
if (requireNamespace("mice", quietly = TRUE)) {
  df <- tibble::tibble(a = c(1, NA, 3), b = c(2, 4, NA), c = 5)
  impute_mice(df, m = 2, verbose = TRUE)
}
}

}
\references{
\insertRef{rubin1987mi}{HealthMarkers}
\insertRef{vanbuuren2011mice}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adiposity_sds_strat.R
\name{adiposity_sds_strat}
\alias{adiposity_sds_strat}
\title{Compute sex-stratified standardized scores (SDS) for adiposity measures}
\usage{
adiposity_sds_strat(
  data,
  col_map,
  ref,
  na_action = c("keep", "omit", "error"),
  allow_partial = FALSE,
  prefix = "",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame or tibble with variables and a sex column}

\item{col_map}{Named list mapping:
\itemize{
\item sex: column name with sex values ("M","F","m","f", 1, 2)
\item vars: optional named list mapping reference variable names -> data column names.
If omitted, expects the reference variable names to exist in \code{data}.
}}

\item{ref}{Named list with elements "M" and "F". Each is a named list of numeric
vectors c(mean=, sd=) keyed by variable name, e.g.:
list(
M = list(BMI = c(mean=23, sd=3.5), waist = c(mean=85, sd=10)),
F = list(BMI = c(mean=21, sd=3.0), waist = c(mean=75, sd=9))
)}

\item{na_action}{One of:
\itemize{
\item "keep"  - keep rows with NA (propagates to outputs)
\item "omit"  - drop rows with NA in any required variable
\item "error" - abort if any required variable has NA
}}

\item{allow_partial}{If TRUE, skip variables absent in data (with a warning); if FALSE error}

\item{prefix}{Optional prefix for output columns (default "")}

\item{verbose}{Logical; when TRUE emit progress via package logger; by default logging is controlled by options(healthmarkers.verbose)}
}
\value{
A tibble with one SDS column per retained variable: \code{varname_SDS}, where \code{varname} is the original variable name (optionally prefixed by \code{prefix}).
}
\description{
Computes sex-specific SDS (z-scores) for selected anthropometric variables using
reference means and SDs provided separately for males and females.
}
\examples{
ref <- list(
  M = list(BMI = c(mean=24.5, sd=3.8), waist = c(mean=88, sd=12)),
  F = list(BMI = c(mean=22.1, sd=4.2), waist = c(mean=76, sd=11))
)
df <- data.frame(BMI=c(25.2,21.8,27.1), waist=c(85,72,95), sex=c("M","F","M"))
adiposity_sds_strat(
  df,
  col_map = list(sex = "sex", vars = list(BMI = "BMI", waist = "waist")),
  ref = ref
)
}
\references{
\insertRef{who1995anthro}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nutrient_markers.R
\name{nutrient_markers}
\alias{nutrient_markers}
\title{Compute a Suite of Nutrient-Based Health Markers}
\usage{
nutrient_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame or tibble containing subject-level data.}

\item{col_map}{Optional named list mapping variable keys (see Details) to
column names in \code{data}. You only need to supply the keys you have; any
markers with missing inputs return \code{NA}. If NULL, defaults to identity
mapping for all known keys.}

\item{na_action}{One of c("keep","omit","error") controlling missing-data policy
across the columns referenced by \code{col_map}.
\itemize{
\item "keep" (default): keep NA; outputs become NA where inputs are NA.
\item "omit": drop rows with NA in any used input column.
\item "error": abort if any used input contains NA.
}}

\item{na_warn_prop}{Numeric in \eqn{[0,1]}; per-variable threshold for high-missingness
diagnostics on used input columns. Default 0.2.}

\item{verbose}{Logical; if TRUE, prints stepwise messages and a final summary via hm_inform. Default FALSE.}
}
\value{
A tibble with one row per input row and these columns:
FerritinTS, AGR, Omega3Index, Mg_Cr_Ratio, GlycatedAlbuminPct,
UA_Cr_Ratio, BUN_Cr_Ratio, Ca_x_Phosphate, AnionGap, Tyr_Phe_Ratio.
}
\description{
Given a data frame or tibble of routine biochemical labs,
\code{nutrient_markers()} returns a set of widely used ratios, products,
and simple percentages that summarize iron metabolism, protein status,
omega-3 balance, renal excretion, mineral homeostasis, and aromatic
amino-acid patterns.
}
\details{
Recognized markers (returned as columns):
\itemize{
\item FerritinTS: Ferritin / Transferrin saturation
\item AGR: Albumin / Globulin, where Globulin = Total protein - Albumin
\item Omega3Index: EPA + DHA (percentage points)
\item Mg_Cr_Ratio: Magnesium / Creatinine
\item GlycatedAlbuminPct: (Glycated albumin / Albumin) x 100
\item UA_Cr_Ratio: Uric acid / Creatinine
\item BUN_Cr_Ratio: BUN / Creatinine
\item Ca_x_Phosphate: Calcium x Phosphate
\item AnionGap: (Na + K) - (Cl + HCO3)
\item Tyr_Phe_Ratio: Tyrosine / Phenylalanine
}

Recognized \code{col_map} keys and expected units (no automatic conversion):
\itemize{
\item ferritin: Serum ferritin (ng/mL)
\item transferrin_sat: Transferrin saturation (\%)
\item albumin: Serum albumin (g/L)
\item total_protein: Total serum protein (g/L)
\item EPA: Red-cell EPA as \% of total fatty acids
\item DHA: Red-cell DHA as \% of total fatty acids
\item Mg: Serum magnesium (mmol/L)
\item creatinine: Serum creatinine (umol/L)
\item glycated_albumin: Glycated albumin (g/L)
\item uric_acid: Serum uric acid (umol/L)
\item BUN: Blood urea nitrogen (mg/dL)
\item phosphate: Serum phosphate (mmol/L)
\item calcium: Serum calcium (mmol/L)
\item Na: Serum sodium (mmol/L)
\item K: Serum potassium (mmol/L)
\item Cl: Serum chloride (mmol/L)
\item HCO3: Serum bicarbonate (mmol/L)
\item Tyr: Serum tyrosine (umol/L)
\item Phe: Serum phenylalanine (umol/L)
}

\strong{Unit-mixing notes:}
\itemize{
\item \code{BUN_Cr_Ratio} divides BUN (mg/dL) by creatinine (umol/L). This is NOT
numerically equivalent to the standard clinical BUN:Creatinine ratio
(reference range ~10-20), which requires both in mg/dL. The result here
is approximately 88.4x smaller than the standard ratio. Provide creatinine
in mg/dL and adjust \code{col_map} if you require the standard clinical ratio.
\item \code{Mg_Cr_Ratio} divides Mg (mmol/L) by creatinine (umol/L). The result is
1/1000 of the standard Mg/Cr ratio in mmol/mmol. Typically applied to urine;
serum Mg/Cr is not a standard clinical metric.
}

Default \code{extreme_rules} (inputs) are broad and intended for unit/entry checks:
ferritin (0, 2000), transferrin_sat (0, 100), albumin (10, 60), total_protein (40, 100),
EPA (0, 20), DHA (0, 20), Mg (0.2, 3), creatinine (20, 2000), glycated_albumin (0, 60),
uric_acid (50, 1000), BUN (1, 150), phosphate (0.1, 5), calcium (0.5, 4),
Na (100, 200), K (2, 8), Cl (70, 130), HCO3 (5, 45), Tyr (10, 300), Phe (20, 300).
}
\examples{
# Quick smoke-test
df <- data.frame(ferritin = 50, albumin = 45, uric_acid = 300, Na = 140)
nutrient_markers(df, verbose = FALSE)

\donttest{
df <- tibble::tibble(
  ferritin         = c(50, 100),
  transferrin_sat  = c(30, 50),
  albumin          = c(45, 40),
  total_protein    = c(70, 75),
  EPA              = c(2.0, 2.5),
  DHA              = c(4.0, 4.5),
  Mg               = c(0.85, 0.90),
  creatinine       = c(80, 90),
  glycated_albumin = c(12, 14),
  uric_acid        = c(300, 400),
  BUN              = c(14, 16),
  phosphate        = c(1.0, 1.2),
  calcium          = c(2.3, 2.4),
  Na               = c(140, 138),
  K                = c(4.2, 4.0),
  Cl               = c(100, 102),
  HCO3             = c(24, 26),
  Tyr              = c(60, 70),
  Phe              = c(50, 55)
)
nutrient_markers(df, verbose = TRUE)
}

}
\references{
\insertRef{harris2004omega3}{HealthMarkers}
\insertRef{koga2010glyalb}{HealthMarkers}
\insertRef{block1998calcp}{HealthMarkers}
\insertRef{waikar2009creak}{HealthMarkers} (creatinine kinetics context)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_risk_qrisk3}
\alias{cvd_risk_qrisk3}
\title{QRISK3 10-year risk (UK QRISK3-2017)}
\usage{
cvd_risk_qrisk3(data, ..., patid = NULL, na_warn_prop = 0.2, verbose = TRUE)
}
\arguments{
\item{data}{A data frame with variables required by \code{QRISK3::QRISK3_2017()}.}

\item{...}{Passed to \code{QRISK3::QRISK3_2017()}.}

\item{patid}{Optional vector of patient IDs (default: \code{1:nrow(data)}).}

\item{na_warn_prop}{Proportion (0-1) to flag high missingness warnings (default 0.2).
Only used when \code{verbose = TRUE}.}

\item{verbose}{Logical; if TRUE, prints progress and a short summary.}
}
\value{
A tibble with columns \code{model}, \code{year}, \code{risk}.
}
\description{
Wrapper around \code{QRISK3::QRISK3_2017()} that auto-generates a \code{patid}
if one is not supplied. Adds input validation and quiet failure to NA on backend error.
}
\examples{
\donttest{
  if (requireNamespace("QRISK3", quietly = TRUE)) {
    df <- data.frame(
      gender = 1L, age = 55L,
      atrial_fibrillation = 0L, atypical_antipsy = 0L,
      regular_steroid_tablets = 0L, erectile_disfunction = 0L,
      migraine = 0L, rheumatoid_arthritis = 0L,
      chronic_kidney_disease = 0L, severe_mental_illness = 0L,
      systemic_lupus_erythematosis = 0L, blood_pressure_treatment = 0L,
      diabetes1 = 0L, diabetes2 = 0L,
      weight = 80, height = 175,
      ethnicity = 1L, heart_attack_relative = 0L,
      cholesterol_HDL_ratio = 4.2, systolic_blood_pressure = 130,
      std_systolic_blood_pressure = 7, smoke = 0L, townsend = 0
    )
    cvd_risk_qrisk3(df)
  }
}

}
\references{
\insertRef{hippisleycox2017qrisk3}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_risk_stroke}
\alias{cvd_risk_stroke}
\title{Stroke 10-year risk}
\usage{
cvd_risk_stroke(data, col_map = NULL, na_warn_prop = 0.2, verbose = TRUE, ...)
}
\arguments{
\item{data}{A data frame with the required cardiovascular risk columns.}

\item{col_map}{Optional named list mapping internal keys (\code{age}, \code{sex},
\code{race}, \code{smoker}, \code{total_chol}, \code{HDL_c}, \code{sbp},
\code{bp_treated}, \code{diabetes}, \code{bmi}) to actual column names in
\code{data}. If \code{NULL} (default), column names are auto-inferred then fall
back to the key names themselves. \code{sex} accepts \code{1}/\code{0},
\code{"m"}/\code{"f"}, or \code{"male"}/\code{"female"} (case-insensitive).}

\item{na_warn_prop}{Proportion (0-1) to flag high missingness warnings (default 0.2).
Only used when \code{verbose = TRUE}.}

\item{verbose}{Logical; if TRUE, prints progress and a short summary.}

\item{...}{Passed to \code{PooledCohort::predict_10yr_stroke_risk()}.}
}
\value{
A tibble with \code{model}, \code{year}, \code{risk}.
}
\description{
Wrapper around \code{PooledCohort::predict_10yr_stroke_risk()} with quiet
fallback to NA if the backend errors.
}
\examples{
if (requireNamespace("PooledCohort", quietly = TRUE)) {
  df <- data.frame(age = 55, sex = 1, race = "white", smoker = FALSE,
    total_chol = 200, HDL_c = 50, sbp = 140, bp_treated = FALSE,
    diabetes = FALSE, bmi = 27)
  cvd_risk_stroke(df)
}
}
\references{
\insertRef{goff2014accaha}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liver_markers.R
\name{liver_markers}
\alias{liver_markers}
\title{Compute liver-related indices (FLI, NFS, APRI, FIB-4, BARD, ALBI, MELD-XI) with validation and diagnostics}
\usage{
liver_markers(
  data,
  col_map = NULL,
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2
)
}
\arguments{
\item{data}{A data.frame or tibble containing your liver and anthropometry data.}

\item{col_map}{Named list mapping these keys -> column names in \code{data}:
\itemize{
\item \code{BMI} (kg/m^2), \code{waist} (cm), \code{TG} (mg/dL), \code{GGT} (U/L),
\item \code{age} (years), \code{AST} (U/L), \code{ALT} (U/L), \code{platelets} (10^9/L),
\item \code{albumin} (g/L), \code{diabetes} (0/1 or logical),
\item \code{bilirubin} (mg/dL), \code{creatinine} (mg/dL).
}}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, input
availability, physiological range information (informational only, values
not altered), the list of markers being computed with their inputs, and a
per-column results summary.}

\item{na_action}{One of \code{c("keep","omit","error")} controlling missing-data
policy. Default "keep".}

\item{na_warn_prop}{Numeric in \eqn{[0,1]}; per-variable threshold for
high-missingness warnings. Default 0.2.}
}
\value{
A tibble with one column per marker: \code{FLI}, \code{NFS}, \code{APRI}, \code{FIB4},
\code{BARD}, \code{ALBI}, \code{MELD_XI}. If an ID column is detected in \code{data} (e.g.
\code{id}, \code{IID}, \code{participant_id}), it is prepended as the first output column.
}
\description{
Given routine labs and anthropometry, computes:
\itemize{
\item FLI      - Fatty Liver Index (Bedogni et al. 2006)
\item NFS      - NAFLD Fibrosis Score (Angulo et al. 2007)
\item APRI     - AST-to-Platelet Ratio Index
\item FIB4     - Fibrosis-4 Index
\item BARD     - BMI-AST/ALT-Diabetes score
\item ALBI     - Albumin-Bilirubin score
\item MELD_XI  - MELD excluding INR
}
}
\details{
Enhancements:
\itemize{
\item Robust input validation (columns present, types) with informative errors.
\item Configurable NA policy and optional extreme-value scanning/capping.
\item Data-quality warnings (high missingness, non-positive logs, zero denominators).
\item Verbose stepwise progress and completion summary.
}

Units (no automatic conversion):
\itemize{
\item BMI: kg/m^2; Waist: cm; TG: mg/dL; GGT/AST/ALT: U/L; Platelets: 10^9/L; Albumin: g/L; Bilirubin: mg/dL; Creatinine: mg/dL.
\item ALBI uses bilirubin in mumol/L internally (converted as bilirubin (mg/dL) * 17.1).
}

Formulas
\itemize{
\item FLI      = logistic(0.953\emph{ln(TG) + 0.139}BMI + 0.718\emph{ln(GGT) + 0.053}waist - 15.745) * 100
\item NFS      = -1.675 + 0.037\emph{age + 0.094}BMI + 1.13\emph{diabetes + 0.99}(AST/ALT) - 0.013\emph{platelets - 0.066}albumin
(albumin in g/L; Angulo 2007 published coefficient -0.66 was for g/dL, divided by 10 here)
\item APRI     = (AST / 40) / platelets * 100; assumes AST upper limit of normal = 40 U/L
\item FIB-4    = (age * AST) / (platelets * sqrt(ALT))
\item BARD     = +1 if BMI>=28, +2 if AST/ALT>=0.8, +1 if diabetes present; sum in 0-4
\item ALBI     = 0.66\emph{log10(bilirubin (mumol/L)) - 0.0852}albumin (g/L)
\item MELD-XI  = 5.11\emph{ln(bilirubin (mg/dL)) + 11.76}ln(creatinine (mg/dL)) + 9.44
}
}
\examples{
# Quick smoke-test
df <- data.frame(ALT = 25, AST = 20, BMI = 24, platelets = 250)
liver_markers(df, verbose = FALSE)

\donttest{
library(tibble)
df <- tibble(
  BMI           = 24,
  waist         = 80,
  TG = 150,
  GGT           = 30,
  age           = 45,
  AST           = 25,
  ALT           = 20,
  platelets     = 250,
  albumin       = 42,
  diabetes      = FALSE,
  bilirubin     = 1.0,
  creatinine    = 0.9
)
liver_markers(df)
liver_markers(df, verbose = FALSE)
}

}
\references{
\insertRef{bedogni2006fli}{HealthMarkers}
\insertRef{angulo2007nfs}{HealthMarkers}
\insertRef{wai2003apri}{HealthMarkers}
\insertRef{sterling2006fib4}{HealthMarkers}
\insertRef{harrison2008bard}{HealthMarkers}
\insertRef{johnson2015albi}{HealthMarkers}
\insertRef{heuman2007meldxi}{HealthMarkers}
}
\seealso{
\code{\link[=inflammatory_markers]{inflammatory_markers()}}, \code{\link[=kidney_failure_risk]{kidney_failure_risk()}}, \code{\link[=iAge]{iAge()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/glycemic_markers.R
\name{glycemic_markers}
\alias{glycemic_markers}
\title{Calculate glycemic-, C-peptide-, and additional metabolic markers}
\usage{
glycemic_markers(
  data,
  col_map = NULL,
  na_action = c("ignore", "warn", "error", "keep", "omit"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing at least:
\itemize{
\item HDL_c (mmol/L), TG (mmol/L), BMI (kg/m^2)
Optional if present: glucose (mmol/L), HbA1c (mmol/mol), C_peptide (pmol/L),
G0 (mmol/L), I0 (pmol/L), leptin (ng/mL), adiponectin (ng/mL).
}}

\item{col_map}{Optional named list mapping keys to column names in \code{data}.
Required keys: \code{HDL_c}, \code{TG}, \code{BMI}. Optional keys (if present will be used):
\code{glucose}, \code{HbA1c}, \code{C_peptide}, \code{G0}, \code{I0}, \code{leptin}, \code{adiponectin}.
If \code{NULL} (default), the function uses columns named exactly as the keys.}

\item{na_action}{One of \code{c("ignore","warn","error","keep","omit")}.
HM-CS aliases: \code{keep} == \code{ignore}; \code{omit} drops rows with any NA/non-finite
in used inputs before computing markers. Default "ignore".}

\item{na_warn_prop}{Proportion (0-1) threshold for high-missingness warnings
among used columns when \code{na_action = "warn"}. Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} (default), prints step-by-step progress
including column mapping, optional input availability, pre-computation
notes, physiological range information (informational only, values are not
altered), the list of markers being computed with their inputs, and a
per-column results summary.}
}
\value{
A tibble with columns:
\itemize{
\item SPISE, METS_IR, prediabetes, diabetes, HOMA_CP, LAR, ASI, TyG_index
If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}),
it is prepended as the first output column.
}
}
\description{
Given fasting labs and anthropometry, computes:
\itemize{
\item SPISE (Single-Point Insulin Sensitivity Estimator)
\item METS_IR (Metabolic Score for Insulin Resistance)
\item prediabetes flag (HbA1c >= 42 mmol/mol)
\item diabetes flag (HbA1c >= 48 mmol/mol)
\item HOMA_CP (C-peptide-based HOMA-IR variant; operational formula, see notes)
\item LAR (Leptin/Adiponectin Ratio)
\item ASI (Adiponectin Sensitivity Index; adiponectin/insulin)
\item TyG_index (Triglyceride-Glucose Index)
}
}
\details{
Assumed units (no automatic conversion of inputs except where noted):
\itemize{
\item HDL_c, TG: mmol/L (TyG internally converts TG to mg/dL via 88.57)
\item BMI: kg/m^2
\item glucose, G0: mmol/L (TyG internally converts glucose to mg/dL via 18)
\item HbA1c: mmol/mol
\item C_peptide, I0: pmol/L (HOMA_CP uses I-like conversion factor 6 as in insulin muU/mL; see notes)
\item leptin, adiponectin: ng/mL
}

These indices are intended for research and feature-engineering applications.
The prediabetes and diabetes flags apply standard HbA1c cut-offs (WHO/IDF criteria)
but this function is not validated as a clinical diagnostic tool.

Quality controls and options:
\itemize{
\item Input validation ensures required variables exist and are numeric-coercible.
\item Non-numeric inputs are coerced to numeric with a warning (NAs introduced reported).
\item Missing or non-finite inputs are handled via \code{na_action}.
\item Logs and divisions are computed safely (non-positive arguments yield NA).
\item Physiological range notes are printed when \code{verbose = TRUE} (values are not altered).
\item Verbose mode prints step-by-step progress and a completion summary.
}

Optional marker detection and pre-computation (one level deep):
When \code{col_map = NULL} (default), column names are inferred automatically.
The seven optional keys (glucose, HbA1c, C_peptide, G0, I0, leptin,
adiponectin) are computed only when present in \code{data}. If \code{BMI} is absent
but \code{weight} (kg) and \code{height} (m or cm) are present, BMI is computed
automatically. If \code{glucose} is absent but \code{G0} column exists (or vice
versa), the missing key is derived via alias. With \code{verbose = TRUE}
(default), the function reports: column mapping, what is missing and which
raw inputs could provide it, which indices will be NA and why, any
physiological range notes, and a per-column results summary.

Notes on HOMA_CP:
\itemize{
\item This function retains the package's existing operational formula:
HOMA_CP = (G0 (mmol/L) * (C_peptide (pmol/L) / 6)) / 22.5
which mirrors HOMA-IR's structure using a 6 pmol/muU scaling used for insulin.
Users should verify unit conventions for their datasets; alternative C-peptide
HOMA implementations exist (e.g., HOMA2-CP).
}
}
\examples{
# Quick smoke-test
df <- data.frame(glucose = 5.6, HbA1c = 44, G0 = 5.5, I0 = 60,
                 HDL_c = 1.2, TG = 1.5, BMI = 24)
glycemic_markers(df, verbose = FALSE)

\donttest{
df <- tibble::tibble(
  HDL_c       = c(1.0, 1.3),
  TG          = c(1.3, 2.0),
  BMI         = c(24, 30),
  glucose     = c(5.6, 7.1),
  HbA1c       = c(44, 38),
  C_peptide   = c(300, 500),
  G0          = c(5.5, 6.2),
  I0          = c(60, 120),
  leptin      = c(10, 20),
  adiponectin = c(8, 5)
)
glycemic_markers(df)
glycemic_markers(df, verbose = FALSE)
glycemic_markers(df, na_action = "omit", verbose = FALSE)
}
}
\references{
\insertRef{paulmichl2016spise}{HealthMarkers}
\insertRef{bellogaytan2018metsir}{HealthMarkers}
\insertRef{fruhbeck2018adiponectinleptin}{HealthMarkers}
\insertRef{matthews1985homa}{HealthMarkers}
\insertRef{simentalmendia2008tyg}{HealthMarkers}
\insertRef{yang2006adiponectininsulin}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{.hm_detect_id_col}
\alias{.hm_detect_id_col}
\title{Detect a participant/sample ID column in a data frame}
\usage{
.hm_detect_id_col(data)
}
\description{
Checks column names against common patterns (case-insensitive exact match):
id, iid, participant_id, subject_id, sample_id, pid, sid, record_id.
Returns the first matching column name, or NULL if none found.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adiposity_sds.R
\name{adiposity_sds}
\alias{adiposity_sds}
\title{Calculate standardized scores (SDS) for adiposity measures}
\usage{
adiposity_sds(
  data,
  col_map = NULL,
  ref,
  na_action = c("keep", "omit", "error"),
  extreme_action = c("cap", "NA", "error", "warn", "ignore"),
  sds_cap = 6,
  diagnostics = FALSE,
  warn_thresholds = list(na_prop = 0.05, extreme_prop = 0.01),
  id_col = NULL,
  return_summary = FALSE,
  verbose = TRUE,
  na_strategy = NULL,
  extreme_strategy = NULL
)
}
\arguments{
\item{data}{data.frame or tibble containing the measurement columns.}

\item{col_map}{Optional named list mapping reference variable names to column names
in \code{data}. If NULL, identity mapping is assumed (names(ref) must be in data).}

\item{ref}{Named list where each element is a numeric vector with names
\code{mean} and \code{sd}, e.g. \code{list(BMI = c(mean = 23, sd = 4))}.}

\item{na_action}{One of \code{c("keep","omit","error")} for row handling when any
mapped variable is missing.}

\item{extreme_action}{One of \code{c("cap","NA","error","warn","ignore")} for
SDS values exceeding \code{sds_cap}.}

\item{sds_cap}{Positive numeric; absolute cap used when \code{extreme_action = "cap"}.}

\item{diagnostics}{Logical; if TRUE emit informational/warning messages
(coercions, missingness, extremes). FALSE suppresses non-critical warnings.}

\item{warn_thresholds}{Named list with optional elements \code{na_prop} (default 0.05)
and \code{extreme_prop} (default 0.01) used for proportion-based diagnostic warnings.}

\item{id_col}{Optional column name used only in verbose summaries.}

\item{return_summary}{Logical; if TRUE return a list with elements \code{data},
\code{summary}, and \code{warnings} instead of just the SDS tibble.}

\item{verbose}{Logical; if TRUE print progress and completion summaries.}

\item{na_strategy}{Soft-deprecated alias for \code{na_action} (if provided and
\code{na_action} missing, it is used).}

\item{extreme_strategy}{Soft-deprecated alias for \code{extreme_action} (if provided
and \code{extreme_action} missing, it is used).}
}
\value{
A tibble with one \verb{<var>_SDS} column per reference variable, or a
list when \code{return_summary = TRUE}.
}
\description{
Computes standard deviation (z) scores for anthropometric variables relative to
a single (non-sex-stratified) reference set of means and standard deviations.
Includes input validation, optional raw-value extreme screening/capping,
configurable handling of extreme SDS values, NA row policies, optional concise
summary output, and optional verbose progress messages.
}
\details{
SDS are computed as: (observed - mean) / sd. Rows are removed only when
\code{na_action = 'omit'}. Raw-value extreme screening (if enabled) is applied
before SDS computation. Extreme SDS handling (cap / warn / error / ignore)
is controlled by \code{extreme_action}. Legacy argument aliases (\code{na_strategy},
\code{extreme_strategy}) are soft-deprecated but still accepted.
}
\examples{
ref <- list(BMI = c(mean = 23, sd = 4), waist = c(mean = 80, sd = 12))
df <- data.frame(BMI = c(25, NA, 60, 18), waist = c(85, 70, 300, 55))
adiposity_sds(df, ref)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fasting_is.R
\name{fasting_is}
\alias{fasting_is}
\title{Calculate fasting-based insulin sensitivity indices}
\usage{
fasting_is(
  data,
  col_map = NULL,
  normalize = c("none", "z", "inverse", "range", "robust"),
  na_action = c("keep", "omit", "error", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame with required inputs.}

\item{col_map}{Named list mapping required keys:
\itemize{
\item G0: fasting glucose (mmol/L)
\item I0: fasting insulin (pmol/L)
}}

\item{normalize}{One of: "none","z","inverse","range","robust".}

\item{na_action}{One of:
\itemize{
\item "keep"  (retain all rows; indices become NA where inputs missing/non-finite)
\item "omit"  (drop rows with any missing/non-finite required inputs)
\item "error" (abort if any required input is missing/non-finite)
\item "warn"  (emit a warning for rows with missing inputs, then keep them)
}}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, the list
of indices being computed, and a per-column results summary.}
}
\value{
Tibble with 10 columns (indices listed above). If an ID column is
detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}), it is prepended
as the first output column.
}
\description{
Compute 10 fasting indices from glucose (mmol/L) and insulin (pmol/L):
Fasting_inv, Raynaud, HOMA_IR_inv, FIRI, QUICKI, Belfiore_basal, Ig_ratio_basal,
Isi_basal, Bennett, HOMA_IR_rev_inv. Units converted internally:
G0_mg = G0*18 (mg/dL), I0_u = I0/6 (muU/mL).
}
\examples{
# Minimal example (units: G0 in mmol/L, I0 in pmol/L)
df <- data.frame(G0 = c(5.2, 6.1, 4.8), I0 = c(60, 120, 80))
res <- fasting_is(df, col_map = list(G0 = "G0", I0 = "I0"))
head(res)

# With NA handling
df2 <- data.frame(G0 = c(5.0, NA), I0 = c(90, 150))
fasting_is(df2, col_map = list(G0 = "G0", I0 = "I0"), na_action = "keep")
}
\references{
\insertRef{matthews1985homa}{HealthMarkers}
\insertRef{katz2000quicki}{HealthMarkers}
\insertRef{raynaud1999fasting}{HealthMarkers}
\insertRef{avignon1999ogtt}{HealthMarkers}
\insertRef{belfiore1998insulin}{HealthMarkers}
\insertRef{sluiter1976gtolerance}{HealthMarkers}
\insertRef{hanson2000evaluation}{HealthMarkers}
\insertRef{anderson1995exploration}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/saliva_markers.R
\name{saliva_markers}
\alias{saliva_markers}
\title{Calculate saliva-based stress & glycemic markers}
\usage{
saliva_markers(
  data,
  col_map = NULL,
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  times = c(0, 30, 60)
)
}
\arguments{
\item{data}{A data.frame or tibble with salivary markers.}

\item{col_map}{Named list mapping required inputs. Defaults assume same names:
\itemize{
\item cort1   -> "saliva_cort1" (nmol/L at wake)
\item cort2   -> "saliva_cort2" (nmol/L ~30 min)
\item cort3   -> "saliva_cort3" (nmol/L ~60 min)
\item amylase -> "saliva_amylase" (U/mL)
\item glucose -> "saliva_glucose" (mg/dL)
}}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, input
availability, physiological range information (informational only, values
not altered), the list of markers being computed with their inputs, and a
per-column results summary.}

\item{na_action}{One of \code{c("keep","omit","error")} for required-input NA handling. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness diagnostics (debug). Default 0.2.}

\item{times}{Numeric vector of sampling times (minutes) for CAR AUC. Must align with cort1/2/3. Default c(0,30,60).}
}
\value{
A tibble with columns:
\itemize{
\item \code{log_cortisol_wake}
\item \code{CAR_AUC}
\item \code{log_amylase}
\item \code{saliva_glucose}
}

If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}),
it is prepended as the first output column.
}
\description{
Computes:
\itemize{
\item log_cortisol_wake (log-transformed waking cortisol)
\item CAR_AUC           (Cortisol Awakening Response, trapezoidal AUC over 0-60 min by default)
\item log_amylase       (log-transformed salivary alpha-amylase)
\item saliva_glucose    (raw salivary glucose)
}
}
\details{
Inputs are validated, missingness handled via \code{na_action}, logs made safe
(<= 0 -> NA), and optional extremes scan/cap is available.
}
\note{
\code{log_cortisol_wake} and \code{log_amylase} use the \strong{natural logarithm} (\code{log()}).
\code{CAR_AUC} is the trapezoidal area under the cortisol-time curve (Pruessner
et al. 2003, AUC with respect to ground). \code{saliva_glucose} is a
\strong{pass-through} column; no formula is applied.
}
\examples{
df <- tibble::tibble(
  saliva_cort1    = 12.5,
  saliva_cort2    = 18.0,
  saliva_cort3    = 16.2,
  saliva_amylase  = 85,
  saliva_glucose  = 4.2
)
saliva_markers(df)  # uses default col_map

}
\references{
\insertRef{pruessner2003auc}{HealthMarkers}
\insertRef{kirschbaum1994cortisol}{HealthMarkers} (salivary cortisol methods; background)
\insertRef{clow2004car}{HealthMarkers} (CAR methodological review; background)
\insertRef{nater2009amylase}{HealthMarkers} (salivary alpha-amylase SNS biomarker; background)
\insertRef{scales1987salgluc}{HealthMarkers} (salivary glucose application; pass-through, no formula)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hm_col_report.R
\name{hm_col_report}
\alias{hm_col_report}
\title{Diagnose and report column name mapping for your data}
\usage{
hm_col_report(
  data,
  col_map = NULL,
  verbose = TRUE,
  fuzzy = FALSE,
  show_unmatched = FALSE
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{tibble}.}

\item{col_map}{Optional named list of manually specified mappings
(\code{list(TG = "trig_my_col")}).  These always take priority.}

\item{verbose}{Logical (default \code{TRUE}).  Print the formatted report.}

\item{fuzzy}{Logical (default \code{FALSE}).  Enable fuzzy / approximate
matching as a final layer.  May produce false positives; review results
with care.}

\item{show_unmatched}{Logical (default \code{FALSE}).  If \code{TRUE},
also list every key that could not be matched.  Useful when you expect a
variable to be found but it isn't.}
}
\value{
Invisibly returns a named list: internal key \eqn{\to} matched
data-column name.  Keys that were not found are omitted.  You can assign
this to \code{col_map} and pass it to any \pkg{HealthMarkers} function.
}
\description{
Scans the column names of your data frame against the internal synonym
dictionary used by all \pkg{HealthMarkers} functions and prints a
formatted report showing which internal keys were matched automatically and
which were not found.  The function uses five matching layers in order:
\enumerate{
\item \strong{User-supplied} (via \code{col_map}) — always wins.
\item \strong{Exact synonym match} — column name is in the synonym list.
\item \strong{Case-insensitive exact} — same, ignoring upper/lower case.
\item \strong{Column contains synonym} — data column name contains a synonym as a
whole word (e.g.\ \code{"trig_baseline"} matches \code{"trig"}).
\item \strong{Synonym contains column} — a synonym contains the column name as a
whole word (e.g.\ \code{"TG_fasting"} synonym matches column \code{"TG"}).
\item \strong{Fuzzy} (opt-in via \code{fuzzy = TRUE}) — Levenshtein-based approximate
matching as a last resort.
}

The function returns the matched mappings invisibly as a named list that
can be passed directly as the \code{col_map} argument to any
\pkg{HealthMarkers} function.
}
\examples{
\dontrun{
# Basic diagnostic
hm_col_report(my_data)

# Fuzzy matching + show all unmatched
hm_col_report(my_data, fuzzy = TRUE, show_unmatched = TRUE)

# Capture the result as a ready-to-use col_map
cm <- hm_col_report(my_data, verbose = FALSE)
all_health_markers(my_data, col_map = cm)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/health_markers.R
\name{metabolic_markers}
\alias{metabolic_markers}
\title{Aggregate selected metabolic marker groups}
\usage{
metabolic_markers(
  data,
  col_map = NULL,
  which = c("insulin", "adiposity_sds", "cardio", "lipid", "liver", "glycemic", "mets"),
  normalize = c("none", "z", "inverse", "range", "robust"),
  mode = c("both", "IS", "IR"),
  verbose = TRUE,
  na_action = c("keep", "omit", "error")
)
}
\arguments{
\item{data}{A data.frame or tibble.}

\item{col_map}{Named list for column mapping forwarded to underlying functions.}

\item{which}{Character vector of groups to compute: c("insulin","adiposity_sds","cardio","lipid","liver","glycemic","mets").}

\item{normalize}{One of c("none","z","inverse","range","robust").}

\item{mode}{One of c("both","IS","IR").}

\item{verbose}{Logical.}

\item{na_action}{One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).}
}
\value{
Data frame with original columns plus derived markers.
}
\description{
Aggregate selected metabolic marker groups
}
\note{
For references supporting liver, lipid, glycemic, MetS, adiposity and
other domain-specific indices, see each underlying function's documentation
(e.g. ?liver_markers, ?lipid_markers, ?glycemic_markers, ?metss, ?adiposity_sds).
This wrapper omits repeated reference listings to avoid redundancy.

Aggregator wrapper. See underlying function help pages for full references:
all_insulin_indices(), lipid_markers(), liver_markers(), glycemic_markers(), metss().
}
\examples{
\donttest{
df <- data.frame(
  TC = 200, HDL_c = 50, TG = 150, LDL_c = 120,
  ALT = 30, AST = 20, BMI = 25
)
metabolic_markers(df, col_map = list(), which = c("lipid","liver"),
                  normalize = "none", mode = "both", verbose = FALSE, na_action = "keep")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_marker_ldl_particle_number}
\alias{cvd_marker_ldl_particle_number}
\title{LDL Particle Number Estimate (via ApoB)}
\usage{
cvd_marker_ldl_particle_number(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame with numeric column ApoB (mg/dL).}

\item{col_map}{Named list mapping required key:
\itemize{
\item ApoB
}}

\item{na_action}{One of:
\itemize{
\item "keep"  (retain rows; value is NA where input missing/non-finite)
\item "omit"  (drop rows with missing/non-finite input)
\item "error" (abort if input missing/non-finite)
}}

\item{verbose}{Logical; if TRUE, emit hm_inform() progress messages.}
}
\value{
A tibble with columns model = "LDL_PN" and value.
}
\description{
Returns ApoB as a proxy for LDL particle number with HM-CS NA handling.
}
\examples{
df <- data.frame(ApoB = c(80, 120, 100))
cvd_marker_ldl_particle_number(df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{.hm_precompute_from_deps}
\alias{.hm_precompute_from_deps}
\title{One-level pre-computation of missing keys from existing raw columns}
\usage{
.hm_precompute_from_deps(data, deps, missing_keys, fn = "", verbose = FALSE)
}
\arguments{
\item{data}{data.frame to augment.}

\item{deps}{Named list of dependency definitions, each with:
\describe{
\item{needs}{character vector of prerequisite column names}
\item{describe}{short human-readable formula string}
\item{compute}{function(data) returning a numeric vector}
}}

\item{missing_keys}{character vector of key names to attempt to compute.}

\item{fn}{calling function name (for messages).}

\item{verbose}{logical.}
}
\value{
A list with \code{$data} (augmented data.frame) and
\code{$log} (character vector of per-key messages).
}
\description{
Given a dependency map (\code{deps}), for each key in \code{missing_keys}
checks whether its prerequisite columns are present in \code{data}. If yes,
the key is computed and added as a new column; otherwise an informational
message is emitted (when \code{verbose = TRUE}) indicating what to provide.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{hm_validate_inputs}
\alias{hm_validate_inputs}
\title{HM-CS v2 input validation hook}
\usage{
hm_validate_inputs(data, col_map, required_keys, fn)
}
\arguments{
\item{data}{data.frame}

\item{col_map}{named list or NULL}

\item{required_keys}{character vector of required keys}

\item{fn}{function name (string)}
}
\description{
HM-CS v2 input validation hook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inflammatory_markers.R
\name{inflammatory_markers}
\alias{inflammatory_markers}
\title{Compute inflammatory indices (classic, eosinophil, or both)}
\usage{
inflammatory_markers(
  data,
  col_map = NULL,
  panel = c("auto", "classic", "eos", "both"),
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{data.frame or tibble}

\item{col_map}{named list mapping keys to column names in \code{data}.
Keys: neutrophils, lymphocytes, monocytes, platelets, WBC, CRP, albumin, eosinophils, ESR.}

\item{panel}{one of c("auto","classic","eos","both"). "auto" uses presence of eosinophils key.}

\item{na_action}{one of c("keep","omit","error"). Default "keep" propagates NA
in outputs where inputs are missing.}

\item{verbose}{logical; if \code{TRUE} (default), prints column mapping, optional input
availability, physiological range information (informational only), the list of
markers being computed, and a results summary.}
}
\value{
tibble with selected inflammatory indices, with ID column prepended if
detected (e.g. \code{id}, \code{IID}, \code{participant_id}).
}
\description{
Panels:
\itemize{
\item classic: NLR, PLR, LMR, dNLR, SII, SIRI, AISI, CRP_category
\item eos:     NLR, PLR, LMR, NER, SII, SIRI, PIV, CLR, CAR, PCR, mGPS, ESR (if mapped)
\item both:    union of classic and eos panels
}
}
\details{
Derived markers:
\itemize{
\item NLR  = neutrophils / lymphocytes
\item PLR  = platelets / lymphocytes
\item LMR  = lymphocytes / monocytes
\item dNLR = neutrophils / (WBC - neutrophils) when WBC available
\item SII  = platelets * neutrophils / lymphocytes
\item SIRI = neutrophils * monocytes / lymphocytes
\item AISI = neutrophils * monocytes * platelets / lymphocytes
\item CRP_category: "low" (<1 mg/L), "moderate" (1-3 mg/L), "high" (>3 mg/L) when CRP available
\item Eosinophil-panel extras: NER = neutrophils / eosinophils; PIV = platelets * neutrophils * monocytes / lymphocytes;
CLR = CRP/lymphocytes; CAR = CRP/albumin; PCR = platelets/CRP; mGPS (CRP, albumin); ESR passthrough.
}

Note:
\itemize{
\item These outputs are deterministic algebraic indices computed from the mapped
laboratory variables. They are intended for feature engineering and descriptive
analyses, not as standalone diagnosis/prognosis tools.
\item References below document commonly used index definitions or interpretation
conventions directly used in this implementation.
}
}
\examples{
# Quick smoke-test
df <- data.frame(neutrophils = 4, lymphocytes = 2, monocytes = 0.5,
                 platelets = 200, WBC = 7, CRP = 2.5)
inflammatory_markers(df, panel = "classic", na_action = "keep", verbose = FALSE)

\donttest{
df <- data.frame(
  neutrophils = c(4, 2),
  lymphocytes = c(2, 0),
  monocytes   = c(0.5, 0.3),
  platelets   = c(200, 150),
  WBC         = c(7, 4.5),
  CRP         = c(2.5, 0.8),
  albumin     = c(40, 42),
  eosinophils = c(0.2, 0.1),
  ESR         = c(12, 15)
)
cm <- list(
  neutrophils = "neutrophils", lymphocytes = "lymphocytes", monocytes = "monocytes",
  platelets = "platelets", WBC = "WBC", CRP = "CRP", albumin = "albumin",
  eosinophils = "eosinophils", ESR = "ESR"
)
classic_cm <- cm; classic_cm$eosinophils <- NULL; classic_cm$ESR <- NULL
inflammatory_markers(df, classic_cm, panel = "classic", na_action = "keep")
inflammatory_markers(df, cm, panel = "eos", na_action = "keep", verbose = TRUE)
}

}
\references{
\insertRef{zahorec2001}{HealthMarkers}
\insertRef{hu2014sii}{HealthMarkers}
\insertRef{qi2016siri}{HealthMarkers}
\insertRef{proctor2011mgps}{HealthMarkers}
\insertRef{pearson2003markers}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_infer-cols.R
\name{.hm_default_col_patterns_exact}
\alias{.hm_default_col_patterns_exact}
\title{Internal: default exact-name patterns for hm_infer_cols()
Not exported; used by all_health_markers() when col_map is missing.
Each vector lists every column-name spelling observed in the package's
simulated data, the Inter99/ADDITION real phenotype dataset, and common
biobank / cohort study conventions.  The first element is the canonical
internal key name.}
\usage{
.hm_default_col_patterns_exact()
}
\description{
Internal: default exact-name patterns for hm_infer_cols()
Not exported; used by all_health_markers() when col_map is missing.
Each vector lists every column-name spelling observed in the package's
simulated data, the Inter99/ADDITION real phenotype dataset, and common
biobank / cohort study conventions.  The first element is the canonical
internal key name.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/kyn_trp_ratio.R
\name{kyn_trp_ratio}
\alias{kyn_trp_ratio}
\title{Kynurenine/Tryptophan Ratio (KTR)}
\usage{
kyn_trp_ratio(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with kynurenine and tryptophan concentrations.}

\item{col_map}{Named list with:
\itemize{
\item kynurenine: column for kynurenine (nmol/L)
\item tryptophan: column for tryptophan (mumol/L)
}}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping and a
per-column results summary.}
}
\value{
A tibble with one column: kyn_trp_ratio (numeric). If an ID column
is detected, it is prepended.
}
\description{
Computes the ratio of kynurenine to tryptophan, a marker of IDO activity and immune activation.
}
\details{
KTR is calculated as Kyn (nmol/L) divided by Trp (mumol/L). Elevated KTR indicates
increased tryptophan catabolism via the kynurenine pathway, often reflecting
inflammation and cell-mediated immune activation.

Inputs should already be in Kyn (nmol/L) and Trp (mumol/L).
}
\examples{
# columns named exactly as the required keys (auto-detected)
df <- data.frame(kynurenine = c(2500, 3100, 2700), tryptophan = c(55, 48, 62))
kyn_trp_ratio(df, verbose = FALSE)

# non-standard column names require explicit col_map
df2 <- data.frame(Kyn_nM = c(2500, 3100, 2700), Trp_uM = c(55, 48, 62))
kyn_trp_ratio(df2, col_map = list(kynurenine = "Kyn_nM", tryptophan = "Trp_uM"),
             verbose = FALSE)

}
\references{
\insertRef{fuchs1998ktr}{HealthMarkers};
\insertRef{damerell2025kyn}{HealthMarkers} (clinical application in colorectal cancer)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/charlson_index.R
\name{charlson_index}
\alias{charlson_index}
\title{Charlson Comorbidity Index (CCI)}
\usage{
charlson_index(
  data,
  col_map = list(mi = "mi", chf = "chf", pvd = "pvd", stroke = "stroke", dementia =
    "dementia", copd = "copd", rheum = "rheum", ulcer = "ulcer", mild_liver =
    "mild_liver", diabetes = "diabetes", diab_comp = "diab_comp", hemiplegia =
    "hemiplegia", renal = "renal", cancer = "cancer", leukemia = "leukemia", lymphoma =
    "lymphoma", sev_liver = "sev_liver", metastatic_cancer = "metastatic_cancer", hiv =
    "hiv"),
  verbose = TRUE,
  na_action = c("keep", "omit", "error", "ignore", "warn")
)
}
\arguments{
\item{data}{A data.frame or tibble with binary indicators (0/1) for each comorbidity.}

\item{col_map}{Named list mapping keys to columns in \code{data}:
mi, chf, pvd, stroke, dementia, copd, rheum, ulcer, mild_liver, diabetes, diab_comp,
hemiplegia, renal, cancer, leukemia, lymphoma, sev_liver, metastatic_cancer, hiv.}

\item{verbose}{Logical; if TRUE, emits progress messages.}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}
}
\value{
A tibble with one column: charlson_index (integer total score; NA if any
required input is NA and na_action != "omit").
}
\description{
Computes the Charlson Comorbidity Index by summing weighted comorbidities.
}
\details{
The Charlson Index predicts 10-year mortality by summing weighted comorbidities.
We implement the canonical 19-condition, weight scheme:
\itemize{
\item 1 point: myocardial infarction, congestive heart failure, peripheral vascular disease,
cerebrovascular disease (stroke), dementia, chronic pulmonary disease (COPD),
rheumatologic disease, peptic ulcer disease
\item 2 points: hemiplegia/paraplegia, moderate/severe renal disease, any malignancy (non-metastatic),
leukemia, lymphoma, diabetes with complications
\item 3 points: moderate/severe liver disease
\item 6 points: metastatic solid tumor, AIDS/HIV
}

To avoid double counting paired conditions, the following use the maximum applicable weight:
\itemize{
\item Diabetes: max(1 * diabetes without complications, 2 * diabetes with complications)
\item Liver disease: max(1 * mild liver disease, 3 * moderate/severe liver disease)
\item Cancer: max(2 * non-metastatic solid tumor, 6 * metastatic solid tumor)
}

Age points are not included here and can be added separately if needed.
}
\examples{
patient <- tibble::tibble(
  mi=0, chf=0, pvd=0, stroke=0, dementia=0, copd=0, rheum=0, ulcer=0,
  mild_liver=0, diabetes=0, diab_comp=1, hemiplegia=0, renal=1,
  cancer=0, leukemia=0, lymphoma=0, sev_liver=0, metastatic_cancer=0, hiv=0
)
charlson_index(
  patient,
  col_map = as.list(stats::setNames(names(patient), names(patient)))
)

}
\references{
\insertRef{charlson1987cci}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/allostatic_load.R
\name{allostatic_load}
\alias{allostatic_load}
\title{Allostatic Load Index}
\usage{
allostatic_load(
  data,
  thresholds,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  return_summary = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{data}{data.frame or tibble of numeric biomarker columns.}

\item{thresholds}{named list of scalar numeric cutoffs (names must match columns).}

\item{col_map}{optional named list mapping keys in \code{thresholds} to column names in \code{data}.}

\item{na_action}{one of c("keep","omit","error") ("keep" treats NA as zero contribution).}

\item{return_summary}{logical; TRUE returns list(data, summary, warnings).}

\item{verbose}{logical; print progress messages via hm_inform() (also gated by options(healthmarkers.verbose)).}
}
\value{
tibble with AllostaticLoad or list when return_summary = TRUE.
}
\description{
Computes a composite Allostatic Load (AL) score by flagging biomarkers that
exceed user-specified high-risk thresholds (strict > when multiple biomarkers;
inclusive >= when only one biomarker). Aligned with HM-CS v3: structured
validation, diagnostic control, verbose reporting, and optional summary output.
}
\details{
\strong{API pattern note:} Unlike most HealthMarkers functions which follow the
standard \verb{(data, col_map, ...)} signature, \code{allostatic_load()} uses
\verb{(data, thresholds, col_map = NULL, ...)} because \code{thresholds} is a
domain-specific required argument that cannot be inferred from column names
alone. As a result, this function is not directly dispatchable through the
standard \code{all_health_markers()} registry; a custom wrapper that supplies
\code{thresholds} explicitly would be required.
}
\examples{
df <- tibble::tibble(
  SBP = c(118, 142, 130),
  DBP = c(76, 92, 85),
  CRP = c(1.2, 4.8, 2.1)
)
thr <- list(SBP = 130, DBP = 85, CRP = 3)
allostatic_load(df, thresholds = thr, na_action = "keep", verbose = FALSE)

# Single biomarker uses inclusive >= rule
allostatic_load(df, thresholds = list(CRP = 3))

}
\references{
\insertRef{seeman1997price}{HealthMarkers}
}
\seealso{
\code{\link{adiposity_sds}}, \code{\link{adiposity_sds_strat}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{gad7_score}
\alias{gad7_score}
\title{GAD-7 scoring}
\usage{
gad7_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "GAD7",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{GAD7_total} and \code{GAD7_severity} (factor). Input columns are not included.
}
\description{
GAD-7 scoring
}
\examples{
df <- data.frame(gad7_01 = 0, gad7_02 = 1, gad7_03 = 2, gad7_04 = 1,
                 gad7_05 = 0, gad7_06 = 1, gad7_07 = 2)
gad7_score(df)
}
\references{
\insertRef{spitzer2006gad7}{HealthMarkers}
\insertRef{plummer2016gad7}{HealthMarkers} (validation meta-analysis)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_helpers.R
\name{hm_bind_cols_quiet}
\alias{hm_bind_cols_quiet}
\title{Quiet name repair binder (if you need it elsewhere)}
\usage{
hm_bind_cols_quiet(...)
}
\description{
Quiet name repair binder (if you need it elsewhere)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_risk_scorescvd}
\alias{cvd_risk_scorescvd}
\title{RiskScorescvd calculator}
\usage{
cvd_risk_scorescvd(data, ...)
}
\arguments{
\item{data}{Data required by \code{RiskScorescvd::calc_scores()}.}

\item{...}{Passed to \code{RiskScorescvd::calc_scores()}.}
}
\value{
Object returned by \code{RiskScorescvd::calc_scores()}.
}
\description{
Passthrough to \code{RiskScorescvd::calc_scores()} with graceful fallback to NA.
}
\examples{
\donttest{
  if (requireNamespace("RiskScorescvd", quietly = TRUE)) {
    df <- data.frame(
      Age = 55, Sex = 0, Smoking_status = 1,
      systolic.bp = 140, Total_cholesterol = 5.5,
      HDL.cholesterol = 1.3
    )
    cvd_risk_scorescvd(df)
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spirometry_markers.R
\name{spirometry_markers}
\alias{spirometry_markers}
\title{Spirometry markers: FEV1/FVC, LLN-based obstruction, GOLD grade, bronchodilator response}
\usage{
spirometry_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame with spirometry inputs.}

\item{col_map}{Named list:
fev1, fvc, fev1_post, fvc_post, age, height, sex, ethnicity}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if TRUE (default), emits progress via rlang::inform.}
}
\value{
Tibble with ratio_pre, ratio_post, copd_flag_fixed, obstruction_lln, fev1_pp, fvc_pp, fev1_z, fvc_z, ratio_z, gold_grade, bdr_fev1, bdr_fvc.
}
\description{
Spirometry markers: FEV1/FVC, LLN-based obstruction, GOLD grade, bronchodilator response
}
\examples{
df <- data.frame(FEV1 = c(3.2, 2.1, 1.5), FVC = c(4.0, 3.0, 2.5))
spirometry_markers(df)
}
\references{
\insertRef{miller2005spirometry}{HealthMarkers} (spirometry standardisation methodology; background)
\insertRef{quanjer2012}{HealthMarkers}
\insertRef{gold2025copd}{HealthMarkers}
}
\concept{respiratory-markers}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ckd_stage.R
\name{ckd_stage}
\alias{ckd_stage}
\title{CKD staging (GFR and albuminuria) and KDIGO risk}
\usage{
ckd_stage(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame with renal measures.}

\item{col_map}{Named list with required key: eGFR; optional key: UACR.}

\item{na_action}{One of:
\itemize{
\item "keep"  (retain rows; stages become NA where inputs missing)
\item "omit"  (drop rows with any missing eGFR/UACR that are mapped)
\item "error" (abort if any mapped input missing)
}}

\item{verbose}{Logical; if TRUE (default), emits progress messages via \code{hm_inform()}.}
}
\value{
Tibble with CKD_stage, Albuminuria_stage, KDIGO_risk.
}
\description{
Categorizes eGFR into G1-G5, albuminuria into A1-A3 (by UACR mg/g), and maps KDIGO risk.
}
\examples{
df <- data.frame(eGFR = c(95, 50), UACR = c(10, 200))
ckd_stage(df, list(eGFR = "eGFR", UACR = "UACR"))
}
\references{
\insertRef{kdigo2012ckd}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{k10_score}
\alias{k10_score}
\title{K10 scoring}
\usage{
k10_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "K10",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{K10_total}. Input columns are not included.
}
\description{
K10 scoring
}
\note{
K10 items are summed as provided. The original scale uses 1--5 coding
(total 10--50); some implementations subtract 1 (0--4, total 0--40).
The function accepts either coding, as no reverse-scored items are
involved and \code{min_val}/\code{max_val} only affect reversal.
}
\examples{
df <- data.frame(k10_01 = 0, k10_02 = 1, k10_03 = 2, k10_04 = 1, k10_05 = 0,
                 k10_06 = 1, k10_07 = 2, k10_08 = 1, k10_09 = 0, k10_10 = 1)
k10_score(df)
}
\references{
\insertRef{kessler2002k6k10}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ogtt_is.R
\name{ogtt_is}
\alias{ogtt_is}
\title{Calculate OGTT-based insulin sensitivity indices}
\usage{
ogtt_is(
  data,
  col_map = NULL,
  normalize = "none",
  verbose = TRUE,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2
)
}
\arguments{
\item{data}{A data.frame or tibble containing at least the columns mapped by \code{col_map}.}

\item{col_map}{Named list mapping:
\itemize{
\item G0, G30, G120 -> glucose at 0, 30, 120 min (mmol/L)
\item I0, I30, I120 -> insulin at 0, 30, 120 min (pmol/L)
\item weight -> body weight (kg)
\item bmi -> body-mass index (kg/m^2)
\item age -> age (years)
\item sex -> sex (1 = male, 2 = female)
}}

\item{normalize}{One of c("none","z","inverse","range","robust") used by normalize_vec().}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping, the list
of indices being computed, and a per-column results summary.}

\item{na_action}{One of c("keep","omit","error") for missing/non-finite required inputs. Default "keep".}

\item{na_warn_prop}{Proportion (0-1) for high-missingness diagnostics (debug). Default 0.2.}
}
\value{
A tibble with the OGTT-based index columns listed above. If an ID
column is detected in \code{data} (e.g. \code{id}, \code{IID}, \code{participant_id}), it is
prepended as the first output column.
}
\description{
Given glucose & insulin at 0, 30, 120 min (plus weight, BMI, age, sex),
computes:
\itemize{
\item Isi_120
\item Cederholm_index
\item Gutt_index
\item Avignon_Si0
\item Avignon_Si120
\item Avignon_Sim
\item Modified_stumvoll
\item Stumvoll_Demographics
\item Matsuda_AUC
\item Matsuda_ISI
\item BigttSi
\item Ifc_inv
\item HIRI_inv
\item Belfiore_isi_gly
}
}
\details{
Units assumed:
\itemize{
\item OGTT glucose in mmol/L (internally converted to mg/dL via *18 for select indices)
\item OGTT insulin in pmol/L (internally converted to muU/mL via /6 for select indices)
\item weight in kg; BMI in kg/m^2; age in years; sex coded 1 = male, 2 = female
}

Notes
\itemize{
\item Conversions mirror existing implementation to preserve outputs. Some
formulas intentionally use unconverted inputs (as in prior code).
\item Modified_stumvoll and Stumvoll_Demographics use raw pmol/L and mmol/L
as published in Stumvoll et al. (2000). BigttSi likewise uses raw units.
\item Matsuda_AUC is a non-standard AUC-based variant; the original Matsuda
index (Matsuda_ISI) uses time-point means, not AUCs.
\item Cederholm_index uses log(I0 + I120) as implemented (sum, not mean);
Gutt_index uses log((I0 + I120)/2) (mean). This mirrors published
implementations; the difference is a constant log(2) offset.
\item Ifc_inv and HIRI_inv are derived composite proxies not attributed to a
single formula publication; treat as research tools.
\item Logs are safe: log(x) becomes NA when x <= 0 or non-finite.
}
}
\examples{
df <- tibble::tibble(
  G0 = 5.5, I0 = 60,
  G30 = 7.8, I30 = 90,
  G120 = 6.2, I120 = 50,
  weight = 70, bmi = 24, age = 30, sex = 1
)
ogtt_is(
  df,
  col_map = list(
    G0 = "G0", I0 = "I0",
    G30 = "G30", I30 = "I30",
    G120 = "G120", I120 = "I120",
    weight = "weight", bmi = "bmi",
    age = "age", sex = "sex"
  ),
  normalize = "none",
  verbose = TRUE
)
}
\references{
\insertRef{matsuda1999ogtt}{HealthMarkers}
\insertRef{gutt2000isi}{HealthMarkers}
\insertRef{stumvoll2000ogtt}{HealthMarkers}
\insertRef{hansen2007bigtt}{HealthMarkers}
\insertRef{avignon1999ogtt}{HealthMarkers}
\insertRef{belfiore1998insulin}{HealthMarkers}
\insertRef{matthews1985homa}{HealthMarkers}
\insertRef{suleman2024is}{HealthMarkers} (genetic epidemiology study reviewing IS indices)
}
\seealso{
\code{\link[=fasting_is]{fasting_is()}}, \code{\link[=normalize_vec]{normalize_vec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{bis_score}
\alias{bis_score}
\title{Barratt Impulsiveness Scale (key-driven)}
\usage{
bis_score(
  data,
  col_map = list(),
  key,
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "BIS",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{key}{List with \code{items}, \code{min_val}, \code{max_val}, optional \code{reverse} and \code{subscales}.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only (total and optional subscales). Input columns are not included.
}
\description{
Barratt Impulsiveness Scale (key-driven)
}
\examples{
bis_key <- list(items = sprintf("bis_\%02d", 1:5), min_val = 1, max_val = 4)
df <- data.frame(bis_01 = 1, bis_02 = 2, bis_03 = 3, bis_04 = 4, bis_05 = 2)
bis_score(df, key = bis_key)
}
\references{
\insertRef{patton1995bis}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/health_markers.R
\name{all_insulin_indices}
\alias{all_insulin_indices}
\title{Compute insulin sensitivity/resistance panels (fasting, OGTT, adipose, tracer/DXA)}
\usage{
all_insulin_indices(
  data,
  col_map = NULL,
  normalize = c("none", "z", "inverse", "range", "robust"),
  mode = c("both", "IS", "IR"),
  verbose = TRUE,
  na_action = c("keep", "omit", "error")
)
}
\arguments{
\item{data}{A data.frame or tibble of raw measurements.}

\item{col_map}{Named list with keys G0,I0,G30,I30,G120,I120,TG,HDL_c,FFA,waist,weight,bmi,age,sex,rate_palmitate,rate_glycerol,fat_mass.}

\item{normalize}{One of c("none","z","inverse","range","robust").}

\item{mode}{One of c("IS","IR","both"). "IR" returns only inverted IR, "IS" only the original IS, "both" returns both with IR_ prefix.}

\item{verbose}{Logical.}

\item{na_action}{One of c("keep","omit","error"); forwarded to underlying calculators (HM-CS v2).}
}
\value{
A tibble of IS (and/or IR_) columns.
}
\description{
Compute insulin sensitivity/resistance panels (fasting, OGTT, adipose, tracer/DXA)
}
\note{
For scholarly references to specific indices (e.g., HOMA-IR, QUICKI,
Raynaud, Belfiore, tracer-derived indices, adiposity-related IS metrics),
consult the individual function help pages (e.g. ?fasting_is, ?ogtt_is,
?adipo_is, ?tracer_dxa_is). Citations are intentionally not duplicated here.

Aggregator wrapper. See underlying function help pages for full references:
fasting_is(), ogtt_is(), adipo_is(), tracer_dxa_is().
}
\examples{
# Quick smoke-test (fasting indices only)
df <- data.frame(G0 = 5.2, I0 = 60)
all_insulin_indices(df, normalize = "none", mode = "IS",
                    verbose = FALSE, na_action = "keep")

\donttest{
# Full panel with all supported inputs
df <- data.frame(
  G0 = 5.2, I0 = 60, G30 = 7.5, I30 = 90, G120 = 6.2, I120 = 80,
  TG = 1.5, HDL_c = 1.3, FFA = 0.3, waist = 85, weight = 70, bmi = 24,
  age = 40, sex = "M", rate_palmitate = 0.1, rate_glycerol = 0.2, fat_mass = 20
)
all_insulin_indices(df, col_map = list(
  G0="G0", I0="I0", G30="G30", I30="I30", G120="G120", I120="I120",
  TG="TG", HDL_c="HDL_c", FFA="FFA", waist="waist", weight="weight",
  bmi="bmi", age="age", sex="sex", rate_palmitate="rate_palmitate",
  rate_glycerol="rate_glycerol", fat_mass="fat_mass"
), normalize = "none", mode = "IS", verbose = FALSE, na_action = "keep")
}
}
\references{
\insertRef{suleman2024is}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/globals.R
\name{hm_normalize}
\alias{hm_normalize}
\title{Normalise marker columns in a data frame}
\usage{
hm_normalize(
  data,
  cols = NULL,
  method = c("z", "inverse", "range", "robust"),
  skip_cols = NULL,
  ...
)
}
\arguments{
\item{data}{A data frame (or tibble) containing marker columns to normalise.}

\item{cols}{Character vector of column names to normalise.  If \code{NULL}
(default), \strong{all numeric columns} in \code{data} are normalised.}

\item{method}{One of \code{c("z","inverse","range","robust")}.  Passed directly to
\code{\link[=normalize_vec]{normalize_vec()}}:
\describe{
\item{\code{"z"}}{z-score (mean 0, sd 1).}
\item{\code{"inverse"}}{Rank-based inverse-normal transform (Rankit; default).}
\item{\code{"range"}}{Min-max scaling to \code{feature_range} (default \verb{[0, 1]}).}
\item{\code{"robust"}}{Median/MAD scaling.}
}}

\item{skip_cols}{Character vector of column names to leave untouched even if
they are numeric (e.g. \code{c("age", "BMI")}).  Ignored when \code{cols} is
explicitly supplied.}

\item{...}{Additional arguments forwarded to \code{\link[=normalize_vec]{normalize_vec()}} (e.g.
\code{feature_range}, \code{invnorm_denominator}, \code{ties}).}
}
\value{
The input \code{data} with the selected columns replaced by their
normalised values.  Class (tibble, data.frame, etc.) is preserved.
}
\description{
A convenience wrapper around \code{\link[=normalize_vec]{normalize_vec()}} that applies a normalisation
method to every selected numeric column in a data frame and returns the
modified data frame. The most common use-case is to normalise the output of
any HealthMarkers function — especially domain functions such as
\code{glycemic_markers()}, \code{lipid_markers()}, or \code{renal_markers()} whose internal
\code{normalize} argument currently has no effect.
}
\examples{
# Build a tiny data frame with pre-computed marker columns
df <- data.frame(
  age    = c(45, 52, 61),
  HOMA_IR = c(1.2, 3.4, 2.1),
  TyG     = c(8.1, 9.0, 8.6),
  NLR     = c(2.1, 3.5, 1.8)
)

marker_cols <- c("HOMA_IR", "TyG", "NLR")

# z-score normalise marker columns only
hm_normalize(df, cols = marker_cols, method = "z")

# Inverse-normal transform, leaving age untouched
hm_normalize(df, method = "inverse", skip_cols = "age")
}
\seealso{
\code{\link[=normalize_vec]{normalize_vec()}} for single-vector normalisation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{who5_score}
\alias{who5_score}
\title{WHO-5 scoring}
\usage{
who5_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  impute = c("none", "mean"),
  prefix = "WHO5",
  low_cutoff_percent = 50,
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list mapping canonical item IDs to column names; defaults assume items are already named.}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{impute}{Imputation strategy for missing items when under the threshold: \code{none} or \code{mean} (row-wise mean).}

\item{prefix}{Prefix for output column names.}

\item{low_cutoff_percent}{Percentage threshold for low well-being flag.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \code{WHO5_raw}, \code{WHO5_percent}, \code{WHO5_low_wellbeing}. Input columns are not included.
}
\description{
WHO-5 scoring
}
\examples{
df <- data.frame(who5_01 = 0, who5_02 = 1, who5_03 = 2, who5_04 = 3, who5_05 = 4)
who5_score(df)
}
\references{
\insertRef{topp2015who5}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sarc_f_score.R
\name{sarc_f_score}
\alias{sarc_f_score}
\title{SARC-F Sarcopenia Screening Score}
\usage{
sarc_f_score(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with SARC-F questionnaire responses.}

\item{col_map}{Named list mapping the five SARC-F components to columns:
strength, walking, chair, stairs, falls.}

\item{na_action}{One of c("keep","omit","error","ignore","warn").}

\item{verbose}{Logical; if TRUE (default), emits progress messages.}
}
\value{
A tibble with:
\itemize{
\item sarc_f_score (numeric 0-10; NA if any component is NA)
\item sarc_f_high_risk (logical; TRUE if score >= 4, NA if score is NA)
}
}
\description{
Computes the SARC-F questionnaire score, a quick screening tool for sarcopenia risk.
}
\details{
SARC-F has 5 items: Strength, Assistance in walking, Rise from a chair,
Climb stairs, and Falls. Each item is scored 0 (no difficulty) to 2 (high difficulty).
Total SARC-F score ranges 0-10. A score >= 4 indicates high risk of sarcopenia
and suggests further assessment.
}
\examples{
df <- data.frame(Strength = c(1, 2, 0), Walking = c(0, 1, 2),
                 Chair = c(1, 1, 2), Stairs = c(0, 2, 2), Falls = c(0, 1, 1))
sarc_f_score(df)
}
\references{
\insertRef{malmstrom2013sarcf}{HealthMarkers}
\insertRef{malmstrom2016sarcf}{HealthMarkers} (SARC-F validation and functional outcome prediction; background)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bone_markers.R
\name{bone_markers}
\alias{bone_markers}
\title{Compute Bone Health & Body-Composition Markers (HM-CS v3)}
\usage{
bone_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A \code{data.frame} or tibble with subject-level DXA/anthropometry data.}

\item{col_map}{Named list mapping keys to column names. Required keys:
\itemize{
\item \code{age}, \code{weight}, \code{height}, \code{ALM}, \code{FM}, \code{BMD}, \code{BMD_ref_mean}, \code{BMD_ref_sd}
Optional (passed-through if present and found in data): \code{TBS}, \code{HSA}, \code{PINP}, \code{CTX}, \code{BSAP}, \code{Osteocalcin}.
}}

\item{na_action}{One of "keep", "omit", or "error" controlling how
missing/non-finite input values are treated.}

\item{verbose}{Logical; if TRUE (default), emits progress messages via \code{hm_inform()}.}
}
\value{
A tibble with columns: \code{OSTA}, \code{ALMI}, \code{FMI}, \code{BMD_Tscore}, and
optionally \code{TBS}, \code{HSA}, \code{PINP}, \code{CTX}, \code{BSAP}, \code{Osteocalcin} (in that order).
}
\description{
Given DXA, anthropometry, and optional bone-turnover markers, computes:
\itemize{
\item OSTA: (weight - age) x 0.2
\item ALMI: Appendicular Lean Mass Index = ALM / height^2
\item FMI: Fat Mass Index = FM / height^2
\item BMD_Tscore: (BMD - ref_mean) / ref_sd
and (if in \code{col_map} + data) passes through: TBS, HSA, PINP, CTX, BSAP, Osteocalcin.
}
}
\details{
Notes:
\itemize{
\item Units: height in meters; ALM, FM, weight in kilograms; BMD in g/cm^2; ALMI/FMI in kg/m^2.
\item Non-finite values are treated as NA; division by zero is prevented by input checks.
\item \code{BMD_ref_mean} and \code{BMD_ref_sd} must be supplied by the user from an appropriate
reference population (for example, study-specific values or external norms such as NHANES).
}
}
\examples{
library(tibble)
df <- tibble(
  age = c(60, 72), weight = c(65, 50), height = c(1.65, 1.58),
  ALM = c(18.2, 14.7), FM = c(22.0, 20.5),
  BMD = c(0.95, 0.80), BMD_ref_mean = c(1.00, 1.00), BMD_ref_sd = c(0.12, 0.12)
)
col_map <- list(
  age = "age", weight = "weight", height = "height",
  ALM = "ALM", FM = "FM", BMD = "BMD",
  BMD_ref_mean = "BMD_ref_mean", BMD_ref_sd = "BMD_ref_sd"
)
bone_markers(df, col_map)

}
\references{
\insertRef{woo2002osta}{HealthMarkers}
\insertRef{who1994osteoporosis}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pulmo_markers.R
\name{pulmo_markers}
\alias{pulmo_markers}
\title{Calculate pulmonary function markers (FEV1/FVC, z-scores, percent predicted, LLN, etc.)}
\usage{
pulmo_markers(
  data,
  col_map = NULL,
  equation = c("GLI", "GLIgl", "NHANES3"),
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble with columns:
\itemize{
\item \code{age}       (numeric): years
\item \code{sex}       (character or numeric): "male"/"female" (case-insensitive) or codes 1/2 or 0/1
\item \code{height}    (numeric): cm or m (auto-detected)
\item \code{ethnicity} (character): e.g. "Caucasian", "African-American", "NE Asian", "SE Asian", "Other/Mixed"
\item \code{fev1}      (numeric): observed FEV1 in L
\item \code{fvc}       (numeric): observed FVC in L
}}

\item{col_map}{Optional named list mapping canonical keys (\code{age}, \code{sex}, \code{height},
\code{ethnicity}, \code{fev1}, \code{fvc}) to actual column names in \code{data}. If \code{NULL},
column names are inferred automatically.}

\item{equation}{One of \code{c("GLI","GLIgl","NHANES3")} (see \code{rspiro} for details).
GLIgl ignores ethnicity.}

\item{na_action}{One of \code{c("keep","omit","error")} for handling missing values in
required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness warnings on
required inputs. Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} prints progress messages and a completion summary.}
}
\value{
A tibble with columns:
\itemize{
\item \code{fev1_pred}, \code{fev1_z},   \code{fev1_pctpred}, \code{fev1_LLN}
\item \code{fvc_pred},  \code{fvc_z},    \code{fvc_pctpred},  \code{fvc_LLN}
\item \code{fev1_fvc_ratio}, \code{fev1_fvc_pred}, \code{fev1_fvc_z}, \code{fev1_fvc_pctpred},
\code{fev1_fvc_LLN} (\code{NA} if the equation lacks native FEV1/FVC support in rspiro)
}
}
\description{
Uses the \code{rspiro} reference equations to compute predicted normals,
z-scores, percent predicted and lower limits of normal (LLN) for FEV1, FVC,
and the FEV1/FVC ratio.
}
\details{
Inputs are validated, missingness is handled via \code{na_action}, and heights are
auto-detected as cm when any height > 3; otherwise interpreted as metres (no
automatic unit conversion beyond that heuristic, preserving prior behavior).
}
\note{
\code{fev1_fvc_z}, \code{fev1_fvc_pred}, and \code{fev1_fvc_LLN} are computed via
rspiro's native FEV1/FVC parameter (equivalent to \code{param = "FEV1FVC"}).
If that parameter is not supported by the installed rspiro version or
equation, these columns fall back gracefully: \code{fev1_fvc_z} and
\code{fev1_fvc_LLN} become \code{NA}; \code{fev1_fvc_pred} falls back to
\code{fev1_pred / fvc_pred}.
}
\examples{
if (requireNamespace("rspiro", quietly = TRUE)) {
  df <- data.frame(
    age = c(40, 55), sex = c("male", "female"),
    height = c(175, 162), ethnicity = c("Caucasian", "Caucasian"),
    fev1 = c(3.5, 2.4), fvc = c(4.4, 3.1)
  )
  pulmo_markers(df)
}
}
\references{
\insertRef{quanjer2012}{HealthMarkers}
\insertRef{hankinson1999spirometry}{HealthMarkers}
\insertRef{bowerman2023gli}{HealthMarkers} (race-neutral GLI global equations and interpretation framework; used by rspiro's \code{GLIgl} equation)
}
\seealso{
rspiro
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tracer_dxa_is.R
\name{tracer_dxa_is}
\alias{tracer_dxa_is}
\title{Compute tracer/DXA-based insulin sensitivity indices}
\usage{
tracer_dxa_is(
  data,
  col_map = NULL,
  normalize = NULL,
  na_action = c("keep", "omit", "error"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A data.frame or tibble containing raw measurements.}

\item{col_map}{Named list with entries (depending on mode):
Adipose-only required:
- I0: fasting insulin (pmol/L)
- rate_glycerol, rate_palmitate: tracer rates (mumol/min)
- fat_mass, weight, bmi: body composition
- HDL_c: HDL cholesterol (mmol/L)
Full mode additionally requires:
- G0, G30, G120: glucose (mmol/L)
- I30, I120: insulin (pmol/L)
- TG: triglycerides (mmol/L)
- FFA: free fatty acids (mmol/L)}

\item{normalize}{Ignored (kept for backward compatibility).}

\item{na_action}{One of c("keep","omit","error") for NA handling on required inputs. Default "keep".}

\item{na_warn_prop}{Proportion \eqn{[0,1]} to trigger high-missingness warnings on required inputs. Default 0.2.}

\item{verbose}{Logical; if TRUE, prints progress messages and a completion summary.}
}
\value{
\itemize{
\item Adipose-only tibble columns: LIRI_inv, Lipo_inv, ATIRI_inv
\item Full-mode tibble columns: I_AUC, FFA_AUC, tracer_palmitate_SI, tracer_glycerol_SI, LIRI_inv, Lipo_inv, ATIRI_inv
}
}
\description{
Uses stable isotope tracer infusion rates and DXA-measured fat mass
to compute peripheral and adipose insulin sensitivity and related metrics.
}
\details{
Modes:
\itemize{
\item Adipose-only indices when only adipose-related keys are mapped (no OGTT glucose/insulin time series)
\item Full indices otherwise
}

Expected units:
\itemize{
\item Glucose: mmol/L (internally converted to mg/dL when needed)
\item Insulin: pmol/L (internally converted to muU/mL via /6)
\item TG: mmol/L (to mg/dL via *88.57); HDL-c: mmol/L (to mg/dL via *38.67)
\item Tracer rates: mumol/min
\item Fat mass, weight: kg; BMI: kg/m^2
}
}
\note{
\code{tracer_palmitate_SI} and \code{tracer_glycerol_SI} are simple rate/fat_mass
ratios; the Steele (1959) non-steady-state tracer equation is \strong{not}
implemented here. The LIRI formula coefficients (-0.091, 0.4, 0.346,
-0.408, 0.435) are attributed to Gastaldelli et al. but the paper cited
(\code{gastaldelli2004betacell}) covers beta-cell dysfunction, not LIRI
derivation; the primary LIRI source should be verified. In adipose-only
mode (I30 absent) the mean-insulin term uses I0 twice as a fallback.
}
\examples{
df <- data.frame(
  I0 = c(60, 75), rate_glycerol = c(2.1, 2.8), rate_palmitate = c(1.8, 2.3),
  fat_mass = c(18, 24), weight = c(72, 85), BMI = c(24, 29),
  HDL_c = c(1.3, 1.1)
)
col_map <- list(I0="I0", rate_glycerol="rate_glycerol",
                rate_palmitate="rate_palmitate", fat_mass="fat_mass",
                weight="weight", bmi="BMI", HDL_c="HDL_c")
tracer_dxa_is(df, col_map = col_map)

}
\references{
\insertRef{groop1989tracer}{HealthMarkers} (tracer lipolysis methodology; background)
\insertRef{steele1959glucose}{HealthMarkers} (tracer dilution theory; Steele equation not directly implemented — background)
\insertRef{roden1996ffa}{HealthMarkers} (FFA-induced insulin resistance mechanism; background)
\insertRef{gastaldelli2004betacell}{HealthMarkers} (beta-cell dysfunction context; LIRI formula source unverified — background)
\insertRef{karpe2011ffa}{HealthMarkers} (FFA and insulin resistance review; background)
\insertRef{petersen2007muscleinsulin}{HealthMarkers} (muscle insulin resistance and metabolic syndrome; background)
\insertRef{santomauro1999acipimox}{HealthMarkers} (FFA lowering and insulin sensitivity; background)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_sds.R
\name{calc_sds}
\alias{calc_sds}
\title{Calculate Standard Deviation Scores (SDS; z-scores) for health markers}
\usage{
calc_sds(
  data,
  vars,
  ref,
  id_col = NULL,
  sds_cap = 6,
  na_strategy = c("omit", "error", "keep"),
  extreme_strategy = c("cap", "warn", "error", "NA"),
  warn_thresholds = list(na_prop = 0.05, extreme_prop = 0.01),
  return = c("data", "list"),
  verbose = TRUE,
  na_action = NULL,
  check_extreme = TRUE,
  extreme_action = NULL
)
}
\arguments{
\item{data}{A data.frame/tibble containing the variables.}

\item{vars}{Character vector of variable names in \code{data} to compute SDS for.
Must be non-empty and unique.}

\item{ref}{A data.frame with columns: \code{variable}, \code{mean}, \code{sd} supplying reference
statistics for each variable listed in \code{vars}. Must contain exactly one row per
requested variable, with finite \code{mean} and \code{sd > 0}.}

\item{id_col}{Optional character scalar naming an ID column in \code{data} (used only
in messages and duplicate-ID notes).}

\item{sds_cap}{Numeric scalar; absolute cap for SDS when \code{extreme_strategy = "cap"}.
Must be positive. Default 6.}

\item{na_strategy}{One of c("omit","error","keep"):
\itemize{
\item "omit": drop rows with missing values across any of \code{vars} (default)
\item "error": stop if any missing values among \code{vars}
\item "keep": keep rows; SDS for missing inputs will be NA
}}

\item{extreme_strategy}{One of c("cap","warn","error","NA"):
\itemize{
\item "cap": cap |SDS| at \code{sds_cap} and warn (default)
\item "warn": keep extreme SDS, but warn
\item "error": stop if any |SDS| > \code{sds_cap}
\item "NA": set extreme SDS to NA
}}

\item{warn_thresholds}{Named list controlling warnings (proportions in \eqn{[0,1]}):
\itemize{
\item na_prop: warn if proportion of rows with NA among \code{vars} exceeds this (default 0.05)
\item extreme_prop: warn if proportion of extreme SDS (cells) exceeds this (default 0.01)
}}

\item{return}{One of c("data","list"). "data" returns a tibble with added
\verb{[var]_sds} columns (default). "list" returns a list with components
\code{data}, \code{summary}, and \code{warnings} (backward compatible).}

\item{verbose}{Logical; if TRUE, emit progress via hm_inform(). Also controlled by
options(healthmarkers.verbose = "none"|"inform"|"debug").}

\item{na_action}{Optional HM-CS alias for \code{na_strategy}
(accepted values: keep/omit/error; used if provided).}

\item{check_extreme}{Logical; if TRUE, enables SDS extreme handling (alias for
legacy behavior; defaults to TRUE in this implementation).}

\item{extreme_action}{Optional HM-CS alias for \code{extreme_strategy}
(accepted values: cap/NA/error; used if provided).}
}
\value{
\itemize{
\item If \code{return = "data"} (default): a tibble with added \verb{[var]_sds} columns.
\item If \code{return = "list"}: a list with:
\itemize{
\item data: tibble with added SDS columns
\item summary: list with input/output row counts, omitted rows, total extremes, and per-variable summary
\item warnings: character vector of warning messages emitted
}
}
}
\description{
Computes per-variable SDS as (x - mean) / sd using supplied reference statistics.
Includes input validation, NA/error handling, data quality warnings, and
verbose progress via the package logger (hm_inform), aligned with HM-CS v3.
}
\details{
Derivation: for each requested variable, SDS is calculated as
\code{(observed_value - reference_mean) / reference_sd} using user-supplied
reference statistics in \code{ref}.

Usage note: this function does not derive normative reference values internally.
Users should provide \code{ref} from the target population (preferred), or from an
external source matched as closely as possible on age/sex/ethnicity/context.
Interpretability depends on the quality and relevance of those supplied
reference means and SDs.

By default, returns a tibble with added \verb{[var]_sds} columns (tidyverse-friendly).
For backward compatibility, you can request the previous list output.
}
\examples{
ref <- data.frame(
  variable = c("bmi","sbp"),
  mean     = c(25, 120),
  sd       = c(4, 15)
)
df <- data.frame(
  id  = 1:6,
  bmi = c(24, 30, NA, 29, 10, 26),
  sbp = c(118, 200, 119, 121, 500, 120)
)
out_tbl <- calc_sds(
  data = df,
  vars = c("bmi","sbp"),
  ref = ref,
  id_col = "id",
  na_strategy = "omit",
  extreme_strategy = "cap",
  sds_cap = 6,
  verbose = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metss.R
\name{metss}
\alias{metss}
\title{Metabolic Syndrome Severity Score (MetSSS)}
\usage{
metss(
  data,
  params = list(NHW_M = list(intercept = -2.344, waist = c(mean = 94, sd = 12.4, coef =
    0.846), TG = c(mean = 1.5, sd = 0.6, coef = 0.701), HDL = c(mean = 1.1, sd = 0.3,
    coef = -0.663), glucose = c(mean = 5.3, sd = 0.6, coef = 0.658), MAP = c(mean = 97,
    sd = 11, coef = 0.466)), NHW_F = list(intercept = -2.381, waist = c(mean = 89.7, sd =
    14.8, coef = 0.817), TG = c(mean = 1.28, sd = 0.91, coef = 0.679), HDL = c(mean =
    1.5, sd = 0.4, coef = -0.727), glucose = c(mean = 5.08, sd = 0.52, coef = 0.622), MAP
    = c(mean = 91, 
     sd = 11, coef = 0.557)), NHB_M = list(intercept = -2.399, waist
    = c(mean = 92.8, sd = 13.1, coef = 0.83), TG = c(mean = 1.18, sd = 0.75, coef =
    0.551), HDL = c(mean = 1.27, sd = 0.37, coef = -0.598), glucose = c(mean = 5.55, sd =
    0.85, coef = 0.702), MAP = c(mean = 98, sd = 13, coef = 0.564)), NHB_F =
    list(intercept = -2.395, waist = c(mean = 96.4, sd = 16.4, coef = 0.858), TG = c(mean
    = 1.14, sd = 0.7, coef = 0.57), HDL = c(mean = 1.36, sd = 0.39, coef = -0.634),
    glucose = c(mean = 5.42, sd = 0.84, 
     coef = 0.687), MAP = c(mean = 95, sd = 13,
    coef = 0.577)), HW_M = list(intercept = -2.377, waist = c(mean = 98.5, sd = 11.5,
    coef = 0.864), TG = c(mean = 1.95, sd = 1.19, coef = 0.724), HDL = c(mean = 1.13, sd
    = 0.3, coef = -0.62), glucose = c(mean = 5.67, sd = 0.9, coef = 0.624), MAP = c(mean
    = 97, sd = 11, coef = 0.448)), HW_F = list(intercept = -2.388, waist = c(mean = 97.9,
    sd = 14.2, coef = 0.858), TG = c(mean = 1.66, sd = 1.06, coef = 0.715), HDL = c(mean
    = 1.29, sd = 0.35, coef = -0.657), 
     glucose = c(mean = 5.53, sd = 0.87, coef =
    0.644), MAP = c(mean = 91, sd = 11, coef = 0.512))),
  col_map = NULL,
  verbose = TRUE,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2,
  diagnostics = TRUE
)
}
\arguments{
\item{data}{data.frame / tibble.}

\item{params}{Named list keyed by "RACE_SEX" (e.g. "NHW_M"). Each element:
list(intercept, waist, TG, HDL, glucose, MAP) where each component (except intercept)
is a named numeric vector c(mean=, sd=, coef=). Default parameters are provided
for NHW, NHB, and HW (male and female); no default HA parameters are included
because Gurka et al. (2014) did not publish HA-specific coefficients. Passing
\code{race = "HA"} with default params will raise an error; supply custom params if needed.}

\item{col_map}{Optional named list mapping canonical keys (\code{waist}, \code{TG}, \code{HDL},
\code{glucose}, \code{MAP}) to actual column names in \code{data}. If \code{NULL}, column names
are inferred automatically.}

\item{verbose}{Logical; if TRUE, prints column mapping and computing messages.}

\item{na_action}{One of c("keep","omit","error","ignore","warn") for required-input NAs. Default "keep".}

\item{na_warn_prop}{Proportion (0-1) above which high-missingness warning fires when na_action='warn'. Default 0.2.}

\item{diagnostics}{Logical; if TRUE (default) emit value/range diagnostic warnings
(negative, out-of-range checks).}
}
\value{
tibble with one numeric column: MetSSS
}
\description{
Computes a continuous metabolic syndrome severity z-score using sex- and
race-specific standardized components and coefficients (factor-loading style).

Behavior note:
\itemize{
\item Parameters are selected per-row based on each row's (race, sex) key.
\item All unique keys present in the data must have a matching entry in \code{params}.
}

Required columns (no unit conversion performed):
\itemize{
\item waist (cm), bp_sys (mmHg), bp_dia (mmHg)
\item TG, HDL_c, glucose (mmol/L)
\item sex (1=male, 2=female)
\item race (one of "NHW","NHB","HW", or accepted synonyms; "HA" is recognised
by the normaliser but has no default params — see \code{params} argument)
}
}
\details{
Calculate Metabolic Syndrome Severity Score (MetSSS)
}
\examples{
df <- data.frame(
  waist = 95, bp_sys = 120, bp_dia = 80, TG = 1.5, HDL_c = 1.2,
  glucose = 5.5, sex = 1, race = "NHW", age = 45
)
metss(df)

}
\references{
\insertRef{gurka2014metsss}{HealthMarkers}
\insertRef{deboer2015metss}{HealthMarkers} (clinical application)
\insertRef{gurka2017metss}{HealthMarkers} (clinical application)
\insertRef{deboer2018metss}{HealthMarkers} (clinical application)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/psych_scores.R
\name{cognitive_score}
\alias{cognitive_score}
\title{Cognitive composite (z-mean or PCA1)}
\usage{
cognitive_score(
  data,
  col_map = list(),
  na_action = c("keep", "omit", "error"),
  missing_prop_max = 0.2,
  method = c("z_mean", "pca1"),
  prefix = "cog",
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame containing questionnaire item columns.}

\item{col_map}{Named list with \code{tasks} mapping task IDs to column names (>= 2 tasks required).}

\item{na_action}{How to handle rows with missing items: \code{keep}, \code{omit}, or \code{error}.}

\item{missing_prop_max}{Maximum allowed proportion of missing items per row before the score is set to \code{NA}.}

\item{method}{Aggregation method: \code{z_mean} (average of z-scores) or \code{pca1} (first PC).}

\item{prefix}{Prefix for output column names.}

\item{verbose}{Logical; if \code{TRUE}, emits informational messages about column resolution and scoring progress via \code{hm_inform()}.}
}
\value{
A tibble of score columns only: \verb{\{prefix\}_z_mean} or \verb{\{prefix\}_pca1}. Input columns are not included.
}
\description{
Cognitive composite (z-mean or PCA1)
}
\examples{
df <- data.frame(task_a = c(1, 2), task_b = c(2, 3), task_c = c(3, 4))
cm <- list(tasks = list(
  task_a = "task_a",
  task_b = "task_b",
  task_c = "task_c"
))
cognitive_score(df, col_map = cm, method = "z_mean")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cvd_risk.R
\name{cvd_marker_aip}
\alias{cvd_marker_aip}
\title{Atherogenic Index of Plasma (AIP)}
\usage{
cvd_marker_aip(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{A data frame with numeric columns TG and HDL_c (mg/dL).}

\item{col_map}{Named list mapping required keys:
\itemize{
\item TG: triglycerides
\item HDL_c: HDL cholesterol
}}

\item{na_action}{One of:
\itemize{
\item "keep"  (retain rows; AIP is NA where inputs missing/non-finite)
\item "omit"  (drop rows with any missing/non-finite inputs)
\item "error" (abort if any required input missing/non-finite)
}}

\item{verbose}{Logical; if TRUE, emit hm_inform() progress messages.}
}
\value{
A tibble with columns model = "AIP" and value.
}
\description{
Computes log10(TG / HDL_c) with input validation and HM-CS NA handling.
}
\examples{
df <- data.frame(TG = c(150, 200), HDL_c = c(50, 40))
cvd_marker_aip(df)
}
\references{
\insertRef{dobiasova2004aip}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bode_index.R
\name{bode_index}
\alias{bode_index}
\title{BODE Index (BMI, Obstruction, Dyspnea, Exercise capacity)}
\usage{
bode_index(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  verbose = TRUE
)
}
\arguments{
\item{data}{data.frame/tibble with required columns.}

\item{col_map}{named list with keys:
fev1_pct OR (fev1 and fev1_pred) OR fev1_pp; plus sixmwd, mmrc, bmi.
Example minimal: list(fev1_pct="FEV1pct", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")
Example derive: list(fev1="FEV1", fev1_pred="FEV1_pred", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")
Example from spirometry_markers: list(fev1_pp="fev1_pp", sixmwd="Walk_m", mmrc="mMRC", bmi="BMI")}

\item{na_action}{one of c("keep","omit","error","ignore","warn").}

\item{verbose}{logical; TRUE (default) emits messages.}
}
\value{
tibble with bode_index (integer). NA if any required input missing (unless omitted).
}
\description{
Computes the BODE index (0-10) using FEV1 \% predicted, 6-minute walk distance (6MWD),
mMRC dyspnea scale, and BMI. Higher scores indicate worse prognosis in COPD.
}
\details{
Scoring components:
FEV1 \% predicted: >=65 = 0; 50-64 = 1; 36-49 = 2; <=35 = 3
6MWD (meters): >=350 = 0; 250-349 = 1; 150-249 = 2; <=149 = 3
mMRC dyspnea: 0-1 = 0; 2 = 1; 3 = 2; 4 = 3
BMI: >21 = 0; <=21 = 1
}
\examples{
df <- data.frame(FEV1pct = c(68, 45, 30), Walk_m = c(400, 280, 140),
                 mMRC = c(1, 2, 3), BMI = c(24, 19, 18))
bode_index(df, col_map = list(fev1_pct = "FEV1pct", sixmwd = "Walk_m",
                              mmrc = "mMRC", bmi = "BMI"))
}
\references{
\insertRef{celli2004bode}{HealthMarkers}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oxidative_markers.R
\name{oxidative_markers}
\alias{oxidative_markers}
\title{Oxidative stress markers}
\usage{
oxidative_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error"),
  verbose = TRUE
)
}
\arguments{
\item{data}{Data frame with columns for GSH and GSSG (per col_map).}

\item{col_map}{Named list with required keys \code{GSH} and \code{GSSG}. Defaults assume
column names match keys. Both columns must be in the same units (e.g., µmol/L).}

\item{na_action}{One of c("keep","omit","error").}

\item{verbose}{Logical; if \code{TRUE} (default), prints column mapping and a
per-column results summary.}
}
\value{
A tibble with column GSH_GSSG_Ratio. If an ID column is detected,
it is prepended.
}
\description{
Computes GSH_GSSG_Ratio = reduced glutathione (GSH) / oxidized glutathione (GSSG).
}
\note{
\code{GSH_GSSG_Ratio} is dimensionless only when \code{GSH} and \code{GSSG} are supplied in
the same units (typically µmol/L). The formula \code{GSH / GSSG} is a standard
biochemical redox ratio; no unit conversion is applied.
}
\examples{
df <- data.frame(GSH = c(5, 3), GSSG = c(1, 0.5))
oxidative_markers(df, col_map = list(GSH="GSH", GSSG="GSSG"))

}
\references{
\insertRef{glutathione_redox_review}{HealthMarkers} (background review;
GSH/GSSG is a standard biochemical redox ratio, not a formula from this paper)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lipid_markers.R
\name{lipid_markers}
\alias{lipid_markers}
\title{Calculate lipid-panel markers, Visceral Adiposity Index (VAI),
Lipid Accumulation Product (LAP), and TyG-BMI index}
\usage{
lipid_markers(
  data,
  col_map = NULL,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2,
  verbose = TRUE
)
}
\arguments{
\item{data}{A \code{data.frame} or \code{tibble} containing your lipid
(and optional anthropometry/glucose) data.}

\item{col_map}{Named list mapping:
\itemize{
\item \code{TC}    -> total cholesterol
\item \code{HDL_c} -> HDL-C
\item \code{TG}    -> triglycerides
\item \code{LDL_c} -> (optional) LDL-C; if absent, estimated via Friedewald
\item \code{ApoB}, \code{ApoA1} -> (optional) apolipoproteins
\item \code{waist} -> (optional) waist circumference (cm)
\item \code{BMI}   -> (optional) body mass index (kg/m^2)
\item \code{glucose} -> (optional) fasting glucose (mmol/L); used for TyG_BMI
}}

\item{na_action}{One of \code{c("keep","omit","error","ignore","warn")}.
\itemize{
\item keep/ignore: compute and propagate NA in outputs
\item omit: drop rows with NA in required inputs (TC, HDL_c, TG)
\item error: abort if any required input contains NA
\item warn: like keep, but emit missingness warnings
}}

\item{na_warn_prop}{Proportion (0-1) threshold for high-missingness
warnings when \code{na_action = "warn"}. Default 0.2.}

\item{verbose}{Logical; if \code{TRUE} (default), prints step-by-step progress
including column mapping, optional input availability, pre-computation
notes, physiological range information (informational only, values are
not altered), the list of markers being computed with their inputs, and a
per-column results summary.}
}
\value{
A tibble with computed lipid markers.
Required outputs (always present):
\code{non_HDL_c}, \code{remnant_c}, \code{ratio_TC_HDL}, \code{ratio_TG_HDL},
\code{ratio_LDL_HDL}, \code{ApoB_ApoA1}.
Optional outputs (present when inputs available):
\code{VAI_Men}, \code{VAI_Women} (waist + BMI required);
\code{LAP_Men}, \code{LAP_Women} (waist required);
\code{TyG_BMI} (BMI + glucose required).
If an ID column is detected in \code{data} (e.g. \code{id}, \code{IID},
\code{participant_id}), it is prepended as the first output column.
}
\description{
Given total cholesterol, HDL, TG (and optionally LDL, ApoB/ApoA1,
waist, BMI, glucose), computes:
\itemize{
\item \code{non_HDL_c}, \code{remnant_c}
\item \code{ratio_TC_HDL}, \code{ratio_TG_HDL}, \code{ratio_LDL_HDL}
\item \code{ApoB_ApoA1}
\item \code{VAI_Men}, \code{VAI_Women}
\item \code{LAP_Men}, \code{LAP_Women}
\item \code{TyG_BMI}
}
}
\details{
Assumed units (no automatic conversion except where noted):
\itemize{
\item TC, HDL_c, TG, LDL_c: mmol/L
\item glucose: mmol/L (converted to mg/dL internally for TyG_BMI)
\item waist: cm
\item BMI: kg/m^2
}

Pre-computation (one level deep):
\itemize{
\item If \code{BMI} is absent but \code{weight} (kg) and \code{height} (m or cm) are
present, BMI is computed automatically.
\item If \code{glucose} is absent but \code{G0} is present (or vice versa), the alias
is derived automatically.
\item If \code{LDL_c} is absent, it is always estimated via Friedewald
(TC - HDL - TG/2.2, mmol/L form). An informational message is emitted when
\code{verbose = TRUE}.
}
}
\examples{
df <- data.frame(TC = c(5.2, 6.1), HDL_c = c(1.3, 1.1), TG = c(1.8, 2.3),
                 LDL_c = c(3.2, 4.1), waist = c(88, 95), BMI = c(26, 29))
# Full verbose output (default)
lipid_markers(df)
# Suppress messaging for batch use
lipid_markers(df, verbose = FALSE)
# Pre-compute BMI from weight and height
df2 <- data.frame(TC = 5.2, HDL_c = 1.3, TG = 1.8, weight = 70, height = 175)
lipid_markers(df2, verbose = FALSE)
}
\references{
\insertRef{friedewald1972}{HealthMarkers}
\insertRef{amato2010vai}{HealthMarkers}
\insertRef{kahn2005lap}{HealthMarkers}
\insertRef{lee2016tyg}{HealthMarkers}
\insertRef{lee2020tygbmi}{HealthMarkers} (clinical application)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inflammatory_age.R
\name{iAge}
\alias{iAge}
\title{Compute a simplified Inflammatory Age Index (iAge) with QA and verbose summaries}
\usage{
iAge(
  data,
  col_map = NULL,
  weights = c(CRP = 0.33, IL6 = 0.33, TNFa = 0.34),
  verbose = TRUE,
  na_action = c("keep", "omit", "error", "ignore", "warn"),
  na_warn_prop = 0.2
)
}
\arguments{
\item{data}{A data.frame or tibble containing the biomarker columns mapped by \code{col_map}.}

\item{col_map}{Named list mapping:
\itemize{
\item CRP  -> column name for C-reactive protein (mg/L)
\item IL6  -> column name for interleukin-6 (pg/mL)
\item TNFa -> column name for tumor necrosis factor-alpha (pg/mL)
}}

\item{weights}{Named numeric vector of weights for each marker (must sum to 1).
Defaults to c(CRP = 0.33, IL6 = 0.33, TNFa = 0.34).}

\item{verbose}{Logical; if TRUE, prints column mapping and computing messages.}

\item{na_action}{One of c("keep","omit","error","ignore","warn") controlling how missing inputs affect iAge:
\itemize{
\item "keep": return NA for rows where any required marker is NA (default; consistent with other functions).
\item "omit": ignore NAs in the weighted sum (treat missing markers as 0).
\item "error": abort if any required marker contains NA.
\item "ignore": alias of "omit".
\item "warn": alias of "omit" but emits missingness warnings (per na_warn_prop).
}}

\item{na_warn_prop}{Proportion in \eqn{[0,1]} above which a high-missingness warning
is emitted when \code{na_action = "warn"}. Default 0.2.}
}
\value{
A tibble with one column:
\itemize{
\item iAge (numeric): the computed inflammatory age index.
}
}
\description{
Implements a linear proxy for immunosenescence based on key inflammatory biomarkers,
conceptually inspired by the inflammatory aging clock (iAge) literature.
This simplified iAge is computed as a weighted sum of C-reactive protein (CRP),
interleukin-6 (IL6), and tumor necrosis factor-alpha (TNFa).
}
\details{
By default, rows with any missing required marker return NA in the index
(consistent with the default behavior of other package functions).
Optional diagnostics can warn on high missingness and implausible negative values.
Verbose mode prints step-by-step progress and a final summary.

Assumed units (no automatic unit conversion):
\itemize{
\item CRP: mg/L
\item IL6: pg/mL
\item TNFa: pg/mL
}

Note:
\itemize{
\item The original iAge model in Sayed et al. (Nature Aging, 2021) is a multi-marker
machine learning model. This function provides a simple, linear proxy using
three canonical inflammatory biomarkers. It is not identical to the original
published iAge but is inspired by its rationale.
\item This proxy is intended for exploratory feature engineering and cohort-level
analyses. It must not be treated as a validated replacement for the published
iAge model or used as a standalone clinical decision metric.
}
}
\examples{
library(tibble)
df <- tibble(
  CRP  = c(1.2, 3.5, NA),  # mg/L
  IL6  = c(2.0, 4.1, 1.5), # pg/mL
  TNFa = c(1.0, 1.8, 0.9)  # pg/mL
)
# Default behavior (rows with any missing marker return NA)
iAge(
  df,
  col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa")
)

# Keep NA if any marker missing in a row
iAge(
  df,
  col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa"),
  na_action = "keep"
)

# Verbose output
iAge(
  df,
  col_map = list(CRP = "CRP", IL6 = "IL6", TNFa = "TNFa"),
  verbose = TRUE
)

}
\references{
\insertRef{sayed2021iage}{HealthMarkers} (conceptual background; not method-identical to this implementation)
\insertRef{harris1999il6crp}{HealthMarkers}
\insertRef{bruunsgaard2003death}{HealthMarkers}
}
\seealso{
\code{\link[=impute_missing]{impute_missing()}}, \code{\link[=glycemic_markers]{glycemic_markers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_infer-cols.R
\name{infer_cols}
\alias{infer_cols}
\alias{hm_infer_cols}
\title{Infer column names from user data based on flexible patterns, with logging}
\usage{
infer_cols(
  data,
  map,
  verbose = TRUE,
  log_file = NULL,
  patterns = NULL,
  prefer = NULL,
  strategy = c("error", "prefer", "first", "stable"),
  strict = TRUE,
  ignore_case = TRUE,
  fuzzy = FALSE,
  max_distance = 0.1,
  return = c("map", "list")
)

hm_infer_cols(data, patterns, required_keys = names(patterns), verbose = FALSE)
}
\arguments{
\item{data}{Data frame whose column names are scanned.}

\item{map}{Named list where names are target keys (e.g., "G0","I0","TG") and
values are NULL (to infer) or a user-supplied column name (to keep as-is).}

\item{verbose}{Logical; if TRUE, emits hm_inform() messages for matches and
unresolved keys.}

\item{log_file}{Optional file path; if supplied, a human-readable mapping log is written there.}

\item{patterns}{Named list of character vectors, each giving candidate column
names for a key (first match wins).}

\item{prefer}{Optional named list of character vectors with preferred column
names for each key, used to resolve multiple matches deterministically.
Matching is case-insensitive and exact against the provided names.}

\item{strategy}{One of c("error","prefer","first","stable") controlling resolution when
there are multiple candidates. Default "error" (backward compatible).
\itemize{
\item "prefer": use \code{prefer} names first; else fall back to "stable" tie-break.
\item "first": take the first match in data's column order.
\item "stable": choose shortest name, then alphabetical.
}}

\item{strict}{Logical; if TRUE (default), missing matches error. If FALSE, missing
matches leave \code{map[[key]]} as NULL and issue a warning.}

\item{ignore_case}{Logical; pass to grep(ignore.case = ...). Default TRUE.}

\item{fuzzy}{Logical; if TRUE and no regex matches are found, attempt a fuzzy match
with agrep using \code{max_distance}. Default FALSE.}

\item{max_distance}{Numeric in \eqn{[0,1]} passed to agrep when fuzzy = TRUE. Default 0.1.}

\item{return}{One of c("map","list"). "map" (default) invisibly returns the
filled mapping list. "list" returns a list(map = ..., log = tibble) for auditing.}

\item{required_keys}{Character vector of keys that must resolve; otherwise an
error is raised.}
}
\value{
By default, invisibly returns the filled \code{map}. If return = "list",
returns a list(map = \code{named list}, log = \code{tibble}).

Named list mapping keys to column names; unresolved non-required keys
become \code{NA_character_}.
}
\description{
Given a data.frame and a named mapping spec (e.g., list(G0 = NULL, I0 = NULL)),
infer the source column names for each key using a set of regex patterns.
You can supply your own patterns and "preferred" names to deterministically
resolve ambiguous matches. A structured log is kept and can be written to disk.

Exact-name matching helper used by \code{all_health_markers()} and related wrappers.
It picks the first matching candidate for each key, logs decisions via
\code{hm_inform()} when \code{verbose = TRUE}, and errors if required keys cannot be
resolved.
}
\details{
This helper produces a col_map you can pass to HealthMarkers functions
(e.g., fasting_is(), lipid-derived indices).

Backward compatibility:
\itemize{
\item By default, strict = TRUE and strategy = "error" keep prior behavior:
\itemize{
\item Error if no match found.
\item Error if multiple candidates found.
}
\item You can opt into smarter resolution via strategy = "prefer" or "first".
}
}
\examples{
df <- tibble::tibble(
  fasting_glucose = c(5.5, 6.1),
  fasting_insulin = c(60, 88),
  TG = c(120, 150),
  `HDL-c` = c(50, 45),
  age = c(55, 60)
)
spec <- list(G0 = NULL, I0 = NULL, TG = NULL, HDL_c = NULL)
# Backward-compatible: strict and "error" strategy
res1 <- infer_cols(df, spec, verbose = FALSE)
# Prefer/resolve ties deterministically
res2 <- infer_cols(df, spec, strategy = "prefer", verbose = TRUE)
# Get structured log
res3 <- infer_cols(df, spec, return = "list")
}
\keyword{internal}
